/*
 * last modified---
 * 	08-06-14/jet/ Spark plugin version
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide utility methods for dealing with encryption objects in base64 form
 *
 * notes---
 * 	An Openfire version of this class exists in all plugin packages.
 */

package org.silentvault.spark;

import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.Element;
import org.dom4j.io.XPP3Reader;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.mxp1.MXParser;

import org.jivesoftware.smack.util.Base64;

import org.jivesoftware.spark.util.log.Log;

import java.util.*;
import java.security.*;
import java.security.cert.CertificateFactory;
import java.security.cert.CertificateException;
import java.security.cert.CertificateParsingException;
import java.security.cert.X509Certificate;
import java.security.spec.X509EncodedKeySpec;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.InvalidKeySpecException;
import javax.crypto.*;
import javax.crypto.spec.SecretKeySpec;
import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.StringReader;


/**
 * Holds static methods to obtain encryption objects from base64-encoded strings
 * and perform other encryption and document encoding tasks.
 *
 * @author Kevin Wilkerson
 */
public final class EncodingUtils {
	/**
	 * block size for encrypting RSA blocks (must be smaller than key modulus,
	 * by 11 bytes minimum padding, so (2048 / 8) - 11 = 245)
	 */
	private static final int	M_EncBlockSz = 245;

	/**
	 * block size for decrypting RSA blocks (takes into account padding)
	 */
	private static final int	M_DecBlockSz = 256;

	/**
	 * the code for an email address name type
	 * (must be equal to org.bouncycastle.asn1.x509.GeneralName.rfc822Name)
	 */
	private static final int	M_rfc822Name = 1;

	/**
	 * the tag which must start a certificate
	 */
	public static final String	M_CertSt = "-----BEGIN CERTIFICATE-----";

	/**
	 * the tag which must end a certificate
	 */
	public static final String	M_CertEnd = "-----END CERTIFICATE-----";

	// BEGIN methods
	/**
	 * method to obtain a X509 certificate from a base64-encoded string
	 * @param b64Str the encoded string
	 * @return the certificate, or null if the string does not parse properly
	 */
	public static X509Certificate getCertFromBase64Str(String b64Str) {
		if (b64Str == null || b64Str.isEmpty()) {
			return null;
		}

		// if the string doesn't start with the marker, then we'll need to add it
		String input = null;
		if (!b64Str.startsWith(M_CertSt)) {
			input = M_CertSt + "\n" + b64Str.trim().replace(' ', '\n') + "\n"
						+ M_CertEnd;
		}
		else {
			input = b64Str;
		}
		ByteArrayInputStream certInput
			= new ByteArrayInputStream(input.getBytes());
		CertificateFactory cf = null;
		X509Certificate cert = null;
		try {
			cf = CertificateFactory.getInstance("X.509");
			cert = (X509Certificate) cf.generateCertificate(certInput);
		}
		catch (CertificateException ce) {
			Log.error("Could not get X.509 cert factory, or bad cert spec", ce);
			return null;
		}
		return cert;
	}

	/**
	 * method to obtain a public RSA key from a base64-encoded string
	 * @param b64Str the encoded string
	 * @return the public key, or null if the string does not parse properly
	 */
	public static PublicKey getPubkeyFromBase64Str(String b64Str) {
		if (b64Str == null || b64Str.isEmpty()) {
			return null;
		}

		byte[] pubKeyBytes = Base64.decode(b64Str);
		X509EncodedKeySpec eks = new X509EncodedKeySpec(pubKeyBytes);
		PublicKey pubKey = null;
		KeyFactory keyFactory = null;
		try {
			keyFactory = KeyFactory.getInstance("RSA");
		}
		catch (NoSuchAlgorithmException nsae) {
			Log.error("Bad key factory algorithm", nsae);
			return null;
		}
		try {
			pubKey = keyFactory.generatePublic(eks);
		}
		catch (InvalidKeySpecException ikse) {
			Log.error("Bad pubkey specification", ikse);
			return null;
		}
		return pubKey;
	}

	/**
	 * method to obtain a private RSA key from a base64-encoded string
	 * @param b64Str the encoded string
	 * @return the private key, or null if the string does not parse properly
	 */
	public static PrivateKey getPrivkeyFromBase64Str(String b64Str) {
		if (b64Str == null || b64Str.isEmpty()) {
			return null;
		}

		byte[] privKeyBytes = Base64.decode(b64Str);
		PKCS8EncodedKeySpec eks = new PKCS8EncodedKeySpec(privKeyBytes);
		PrivateKey privKey = null;
		KeyFactory keyFactory = null;
		try {
			keyFactory = KeyFactory.getInstance("RSA");
		}
		catch (NoSuchAlgorithmException nsae) {
			Log.error("Bad key factory algorithm", nsae);
			return null;
		}
		try {
			privKey = keyFactory.generatePrivate(eks);
		}
		catch (InvalidKeySpecException ikse) {
			Log.error("Bad privkey specification", ikse);
			return null;
		}
		return privKey;
	}

	/**
	 * method to obtain the VS number from a certificate's extension
	 * @param cert the X509 certificate with the rfc822 name extension
	 * @return the VS number, or null if the cert does not contain the VS number
	 */
	public static String getVSNumFromCert(X509Certificate cert) {
		Collection<List<?>> nameExts = null;
		try {
			nameExts = cert.getSubjectAlternativeNames();
		}
		catch (CertificateParsingException cpe) {
			Log.error("error getting VS# from cert", cpe);
			return null;
		}
		Iterator<List<?>> nameIt = nameExts.iterator();
		String vsNum = null;
		while (nameIt.hasNext()) {
			List<?> nameExt = nameIt.next();
			Integer nameType = (Integer) nameExt.get(0);
			// make sure it's an email address
			if (nameType.intValue() == M_rfc822Name) {
				vsNum = (String) nameExt.get(1);
				// use the first one as the VS#, even if more are available
				break;
			}
		}
		if (vsNum == null) {
			Log.error("VS# not found in certificate");
		}
		return vsNum;
	}

	/**
	 * method to encrypt and base64-encode a string with a given public RSA key
	 * @param data the string to be encrypted
	 * @param key the public RSA key to use
	 * @return the encrypted and encoded string, or null if encrypting failed
	 */
	public static String makeBase64PubkeyEncStr(String data, PublicKey key) {
		if (data == null || data.isEmpty() || key == null) {
			return null;
		}

		Cipher pubkeyCipher = null;
		boolean gotErr = false;
		KeyFactory keyFactory = null;
		try {
			// init the cipher object with the public key
			keyFactory = KeyFactory.getInstance("RSA");
			pubkeyCipher = Cipher.getInstance("RSA/ECB/PKCS1Padding");
			pubkeyCipher.init(Cipher.ENCRYPT_MODE, key);
		}
		catch (NoSuchAlgorithmException nsae) {
			Log.error("Unable to encrypt string with pubkey", nsae);
			gotErr = true;
		}
		catch (NoSuchPaddingException nspe) {
			Log.error("Unable to encrypt string with pubkey", nspe);
			gotErr = true;
		}
		catch (InvalidKeyException ike) {
			Log.error("Bad pubkey for encrypting string", ike);
			gotErr = true;
		}
		if (gotErr) {
			return null;
		}
		
		/* Perform the encrypt.  Because the size of the data being 
		 * encrypted must be slightly less than the modulus, we are
		 * compelled to encrypt it a bit at a time.
		 */
		ByteArrayInputStream encInput = new ByteArrayInputStream(data.getBytes());
		ByteArrayOutputStream encOutput = new ByteArrayOutputStream(1024);
		byte[] ibuff = new byte[M_EncBlockSz];
		int cnt = encInput.read(ibuff, 0, M_EncBlockSz);
		while (cnt != -1) {
			byte[] obuff = null;
			gotErr = false;

			// encrypt this block
			try {
				obuff = pubkeyCipher.doFinal(ibuff);
			}
			catch (IllegalBlockSizeException ibse) {
				Log.error("Bad block size encrypting string", ibse);
				gotErr = true;
			}
			catch (BadPaddingException bpe) {
				Log.error("Bad padding encrypting string", bpe);
				gotErr = true;
			}
			if (gotErr) {
				return null;
			}
			encOutput.write(obuff, 0, obuff.length);
			ibuff = new byte[M_EncBlockSz];
			cnt = encInput.read(ibuff, 0, M_EncBlockSz);
		}

		byte[] encBlob = encOutput.toByteArray();
		return Base64.encodeBytes(encBlob, Base64.DONT_BREAK_LINES);
	}

	/**
	 * method to encrypt a String to a secret key using tripleDES (DESede)
	 * @param encData the data to be encrypted
	 * @param secKey the secret key (will be zero-padded to 24 bytes if less)
	 * @return the encrypted bytes, base64-encoded, or null if encrypt failed
	 */
	public static String makeBase64DESkeyEncStr(String encData, byte[] secKey) {
		if (encData == null || encData.isEmpty()
			 || secKey == null || secKey.length == 0) {
			return null;
		}
		
		// Pad key to 24 bytes (3x8), because triple-DES wants 192 bit keys.
		byte[] encKey = new byte[24];
		int len = secKey.length > 24 ? 24 : secKey.length;
		System.arraycopy(secKey, 0, encKey, 0, len);
		SecretKeySpec key = new SecretKeySpec(encKey, "DESede");
		Cipher passphraseCipher = null;
		boolean gotErr = false;

		// Set up cipher object for encryption using this key.
		try {
			passphraseCipher = Cipher.getInstance("DESede/ECB/PKCS5Padding");
			passphraseCipher.init(Cipher.ENCRYPT_MODE, key);
		}
		catch (NoSuchAlgorithmException nsae) {
			Log.error("Unable to encrypt string with 3DES", nsae);
			gotErr = true;
		}
		catch (NoSuchPaddingException nspe) {
			Log.error("Unable to encrypt string with 3DES", nspe);
			gotErr = true;
		}
		catch (InvalidKeyException ike) {
			Log.error("Bad key for encrypting string with 3DES", ike);
			gotErr = true;
		}
		if (gotErr) {
			return null;
		}

		// perform the encrypt
		byte[] encBlob = null;
		gotErr = false;
		try {
			encBlob = passphraseCipher.doFinal(encData.getBytes());
		}
		catch (IllegalBlockSizeException ibse) {
			Log.error("Bad block size 3DES encrypting string", ibse);
			gotErr = true;
		}
		catch (BadPaddingException bpe) {
			Log.error("Bad padding 3DES encrypting string", bpe);
			gotErr = true;
		}
		if (gotErr) {
			return null;
		}
		return Base64.encodeBytes(encBlob, Base64.DONT_BREAK_LINES);
	}

	/**
	 * method to encrypt a String to a secret key using tripleDES (DESede)
	 * @param encData the data to be encrypted
	 * @param secKey the secret key, in base64-encoded String form
	 * @return the encrypted bytes, base64-encoded, or null if encrypt failed
	 */
	public static String makeBase64DESkeyEncStr(String encData, String secKey) {
		if (encData == null || encData.isEmpty()
			 || secKey == null || secKey.isEmpty()) {
			return null;
		}

		// simply invoke the method above
		return makeBase64DESkeyEncStr(encData, Base64.decode(secKey));
	}
		
	/**
	 * method to decrypt an input base64-encoded string with a 3DES key
	 * @param encData the encoded and enrypted data
	 * @param secKey the secret key (passphrase), will be padded to 24 bytes
	 * @return the decrypted data
	 */
	public static String getStrFromBase64DESEnc(String encData, byte[] secKey) {
		if (encData == null || encData.isEmpty()
			 || secKey == null || secKey.length == 0) {
			return null;
		}

		// Pad key to 24 bytes (3x8), because triple-DES wants 192 bit keys.
		byte[] decKey = new byte[24];
		int len = secKey.length > 24 ? 24 : secKey.length;
		System.arraycopy(secKey, 0, decKey, 0, len);
		SecretKeySpec key = new SecretKeySpec(decKey, "DESede");
		Cipher passphraseCipher = null;
		boolean gotErr = false;

		// Set up cipher object for decryption using this key.
		try {
			passphraseCipher = Cipher.getInstance("DESede/ECB/PKCS5Padding");
			passphraseCipher.init(Cipher.DECRYPT_MODE, key);
		}
		catch (NoSuchAlgorithmException nsae) {
			Log.error("Unable to decrypt string with 3DES", nsae);
			gotErr = true;
		}
		catch (NoSuchPaddingException nspe) {
			Log.error("Unable to decrypt string with 3DES", nspe);
			gotErr = true;
		}
		catch (InvalidKeyException ike) {
			Log.error("Bad key for decrypting string with 3DES", ike);
			gotErr = true;
		}
		if (gotErr) {
			return null;
		}

		// perform the decrypt
		byte[] clearBlob = null;
		gotErr = false;
		try {
			clearBlob = passphraseCipher.doFinal(Base64.decode(encData));
		}
		catch (IllegalBlockSizeException ibse) {
			Log.error("Bad block size 3DES decrypting string", ibse);
			gotErr = true;
		}
		catch (BadPaddingException bpe) {
			gotErr = true;
		}
		if (gotErr) {
			return null;
		}
		return new String(clearBlob);
	}

	/**
	 * method to decrypt an input base64-encoded string with a 3DES key
	 * @param encData the encoded and enrypted data
	 * @param secKey the secret key, in base64-encoded String form
	 * @return the decrypted data
	 */
	public static String getStrFromBase64DESEnc(String encData, String secKey) {
		if (encData == null || encData.isEmpty()
			 || secKey == null || secKey.isEmpty()) {
			return null;
		}

		// simply invoke the method above
		return getStrFromBase64DESEnc(encData, Base64.decode(secKey));
	}

	/**
	 * method to use a private key to decrypt a base64-encoded blob encrypted
	 * using the corresponding public key
	 * @param encData the encoded and encrypted data
	 * @param privKey the private key to use decrypting
	 * @return the clear text, or null if decryption failed
	 */
	public static String getStrFromBase64PubkeyEnc(String encData,
																  PrivateKey privKey)
	{
		if (encData == null || encData.isEmpty() || privKey == null) {
			return null;
		}

		// init the cipher
		Cipher privkeyCipher = null;
      try {
      	privkeyCipher = Cipher.getInstance("RSA/ECB/PKCS1Padding");
			privkeyCipher.init(Cipher.DECRYPT_MODE, privKey);
      }
      catch (NoSuchAlgorithmException nsae) {
      	Log.error("Unable to decrypt string with private key", nsae);
      	return null;
      }
      catch (NoSuchPaddingException nspe) {
      	Log.error("Unable to decrypt string with private key", nspe);
      	return null;
      }
		catch (InvalidKeyException ike) {
			Log.error("Bad private key for decrypting string", ike);
			return null;
		}

		// decrypt the input
		byte[] encDataBytes = Base64.decode(encData);
		if (encDataBytes == null) {
			return null;
		}
		ByteArrayInputStream decInput = new ByteArrayInputStream(encDataBytes);
		ByteArrayOutputStream decOutput
			= new ByteArrayOutputStream(encDataBytes.length / 2);
		byte[] ibuff = new byte[M_DecBlockSz];
		int cnt = decInput.read(ibuff, 0, M_DecBlockSz);
		while (cnt != -1) {
			
			// decrypt this block
			byte[] obuff = null;
			try {
				obuff = privkeyCipher.doFinal(ibuff);
			}
			catch (IllegalBlockSizeException ibse) {
				Log.error("Bad block size decrypting with privkey", ibse);
				return null;
			}
			catch (BadPaddingException bpe) {
				//Log.error("Bad padding decrypting with privkey", bpe);
				return null;
			}
			decOutput.write(obuff, 0, obuff.length);
			ibuff = new byte[M_DecBlockSz];
			cnt = decInput.read(ibuff, 0, M_DecBlockSz);
		}
		byte[] decBytes = decOutput.toByteArray();
		String rawDecStr = new String(decBytes);

		// chop off any trailing padding chars
		int trailIdx = rawDecStr.indexOf(Character.MIN_VALUE);
		if (trailIdx != -1) {
			return rawDecStr.substring(0, trailIdx);
		}
		return rawDecStr;
	}

	/**
	 * method to create a Dom4j Element from a string containing the XML
	 * @param xmlStr the input XML
	 * @return the root Element produced, or null on parse failure
	 */
	public static final Element buildElement(String xmlStr) {
		if (xmlStr == null || xmlStr.isEmpty()) {
			return null;
		}
		boolean builtOk = true;
		XPP3Reader reader = new XPP3Reader();
		Document doc = null;
		try {
			doc = reader.read(xmlStr.toCharArray());
		}
		catch (DocumentException de) {
			Log.error("parse error", de);
			builtOk = false;
		}
		catch (IOException ioe) {
			Log.error("parse error", ioe);
			builtOk = false;
		}
		catch (XmlPullParserException xppe) {
			Log.error("parse error", xppe);
			builtOk = false;
		}
		if (builtOk) {
			return doc.getRootElement();
		}
		return null;
	}

	// END methods
}
