/*
 * last modified---
 * 	08-15-18/kbw/ ensure listener is always removed
 * 	01-10-18/kbw/ handle multiple units for the same asset in <TokenValues/>
 * 	07-11-15/jet/ use new static methods in VPAssetConfig
 * 	04-02-15/jet/ restore web message filter in resetConnFilter()
 * 	08-06-14/jet/ Spark plugin version
 * 	02-26-14/jet/ allow an OFS host to be a raw IP address
 * 	12-28-13/jet/ for SDS/DHT/PKS paths, favor the same domain as the OFS
 * 					  when there are multiple paths presented
 * 	10-12-13/jet/ support multiple paths to SDS/PKS/DHT in network disco
 * 	08-09-13/jet/ force logout if VP pubkey or voucher/token fetch fails
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	listen for messages from the OFS which are replies to IQ packets we sent
 *
 * notes---
 * 	see underlying messages in vsc_ofs.x
 */

package org.silentvault.spark;

import org.dom4j.Element;
import org.dom4j.DocumentException;

import org.jivesoftware.smack.PacketListener;
import org.jivesoftware.smack.util.StringUtils;
import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.*;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.filter.OrFilter;
import org.jivesoftware.smack.filter.PacketFilter;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.ui.wallet.*;
import org.silentvault.spark.ui.svx.XTabManager;
import org.silentvault.spark.ui.svm.MTabManager;

import java.util.List;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.HashMap;
import java.util.Locale;
import java.util.Iterator;
import java.net.InetSocketAddress;
import java.io.StringReader;
import java.text.NumberFormat;
import java.text.ParseException;
import java.security.PublicKey;


/**
 * class to handle processing of OFS reply packets, within the context of the
 * jabber:iq:voucher-safe#vsc-ofs namespace.
 *
 * @author Kevin Wilkerson
 */
public final class OFSListener implements PacketListener, Runnable {
	// BEGIN data members

	/**
	 * reference to the Plugin or application object
	 */
	private WalletClient		m_Plugin;

	/**
	 * a map of currently outstanding listen filters, representing all requests
	 * for which we are awaiting replies, keyed by their packet IDs
	 */
	private Hashtable<String, PacketIDFilter>	m_FilterMap;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the reference to the plugin or app of which we are a part
	 */
	public OFSListener(WalletClient plugin) {
		m_Plugin = plugin;
		m_FilterMap = new Hashtable<String, PacketIDFilter>();
	}

	// method required to implement the PacketListener interface
	public void processPacket(Packet packet) {
		final WTabManager tm = m_Plugin.getWalletTabManager();

		// verify we got what's expected
		if (packet instanceof OFSMessage) {
			OFSMessage thisMsg = (OFSMessage) packet;
			String opcode = thisMsg.getOpcode();
			//Log.debug("Listener got OFS message opcode " + opcode);

			// record message
			VscState stateObj = m_Plugin.getVscState();
			stateObj.setLastActivity(thisMsg);

			// see if it's an error and handle appropriately
			if (thisMsg.getType() == IQ.Type.ERROR) {
				// take actions based on the opcode
				if (opcode.equalsIgnoreCase("REP_long-phrase_challenge")) {
					tm.showError("no match using entered wallet ID and PIN",
									thisMsg.getErrmsg(), "please try again");
				}
				else if (opcode.equalsIgnoreCase("REP_login-blob")) {
					tm.showError("could not log in",
									thisMsg.getErrmsg(), "please try again");
				}
				else if (opcode.equalsIgnoreCase("REP_publisher_list")) {
					tm.showError("could not obtain publisher list");
				}
				else if (opcode.equalsIgnoreCase("REP_key_params")) {
					tm.showError("could not obtain publisher's key parameters",
									thisMsg.getErrmsg());

					// transition back to publisher selection
					tm.queueTransition(new Integer(WTabManager.M_NEW_SAFE1));
					new Thread(this).start();
				}
				else if (opcode.equalsIgnoreCase("REP_safe_data")) {
					tm.showError("could not obtain new safe data",
									thisMsg.getErrmsg(),
									"please change input or try again later");

					// go back to publisher selection screen
					tm.queueTransition(new Integer(WTabManager.M_NEW_SAFE1));
					new Thread(this).start();
				}
				else if (opcode.equalsIgnoreCase("REP_safe_init")) {
					tm.getNewSafePane().setInitRep(thisMsg);

					// proceed to confirmation screen (will show error)
					tm.queueTransition(new Integer(WTabManager.M_NEW_SAFE4));
					new Thread(this).start();
				}
				else if (opcode.equalsIgnoreCase("REP_network_details")) {
					tm.showError("cannot complete login",
									thisMsg.getErrmsg(),
									"please log out and try again later");

					// back to second page of login
					tm.queueTransition(new Integer(WTabManager.M_LOGIN2));
					new Thread(this).start();
				}
				else if (opcode.equalsIgnoreCase("REP_questions")) {
					tm.showError("cannot recover safe data",
									"publisher list unavailable",
									"please try again later");
					
					// back to first page of login
					tm.queueTransition(new Integer(WTabManager.M_LOGIN1));
					new Thread(this).start();
				}
				else if (opcode.equalsIgnoreCase("REP_recovery_blob")) {
					if (thisMsg.getErrcode()
						 == java.net.HttpURLConnection.HTTP_GATEWAY_TIMEOUT)
					{
						// timeout at SDS
						tm.showError("cannot recover safe data",
										thisMsg.getErrmsg(), "please try again later");
					}
					else {
						// user must have guessed wrong
						tm.showError("cannot recover safe data",
										"incorrect recovery answers, or "
										+ "incorrect safe number",
										"please try again");
					}
					tm.getRecoveryPane().setCursor(null);
					// leave existing answers up for modification
				}
				else if (opcode.equalsIgnoreCase("REP_safe_closed")) {
					// send error to close safe pane
					tm.getCloseSafePane().processCloseReply(thisMsg);
				}
				else if (opcode.equalsIgnoreCase("REP_payment_sent")) {
					// send error case to payment pane
					tm.getPaymentPane().processPaymentReply(thisMsg);
				}
				else if (opcode.equalsIgnoreCase("REP_payment_confirmed")) {
					// send error case to incoming pane
					tm.getIncomingPane().processPickupReply(thisMsg);
				}
				else if (opcode.equalsIgnoreCase("REP_validation")) {
					// validate, split, or merge failed

					// uncommit the vouchers and tokens committed to this operation
					tm.getHomePane().processValidateReply(thisMsg);

					// update token count header
					tm.queueTransition(new Integer(WTabManager.M_HOME_TOKEN_REGEN));
					new Thread(this).start();

					// put up error message
					tm.showError("validate/split/merge voucher operation failed",
									"code " + thisMsg.getErrcode() + ": "
									+ thisMsg.getErrmsg());
					// (leave the user on the validate/split/merge pane)
				}
				else if (opcode.equalsIgnoreCase("REP_payment_reverted")) {
					// pass this message to the unclaimed pane
					tm.getUnclaimedPane().processReclaimReply(thisMsg);
				}
				else if (opcode.equalsIgnoreCase("REP_tokens_purged")) {
					// this does not come here (see {@see SpentTokenCache})
					Log.error("Inappropriate opcode: " + opcode);
				}
				else if (opcode.equalsIgnoreCase("REP_tokens_purchased")) {
					// send error case to tokens pane
					tm.getTokensPane().processPurchTokenReply(thisMsg);
				}
				else if (opcode.equalsIgnoreCase("REP_voucher_redeemed")) {
					// send error case to expired pane
					tm.getExpiredPane().processRedeemReply(thisMsg);
				}
				else if (opcode.equalsIgnoreCase("REP_voucher_deleted")) {
					// send error case to expired pane
					tm.getExpiredPane().processDeleteReply(thisMsg);
				}
				else {
					Log.error("Error on unsupported opcode: " + opcode);
				}
				removeFilter(packet.getPacketID());
				return;
			}

			// success -- take certain further actions based on the opcode
			if (opcode.equalsIgnoreCase("REP_long-phrase_challenge")) {
				// ask user for the 3 indicated chars from the long-phrase
				tm.getLoginPane().setChallenge(thisMsg);

				// go to next step
				tm.queueTransition(new Integer(WTabManager.M_LOGIN2));
				new Thread(this).start();
			}
			else if (opcode.equalsIgnoreCase("REP_login-blob")) {
				// record the blob with the login pane object
				LoginPane lp = tm.getLoginPane();
				lp.setLoginblob(thisMsg.getLoginblob());

				// Decrypt and parse this and store the pieces in VsSecrets object.
				// Blob will be encrypted with our passphrase.

				if (lp.decodeLoginblob()) {
					// mark the VS as logged in now
					stateObj.setLoggedIn(true);

					// transition to logged-in home
					tm.queueTransition(new Integer(WTabManager.M_HOME_INIT));

					// update connect panes on other tabs to enable buttons
					final XTabManager xtm = m_Plugin.getExchangeTabManager();
					xtm.queueTransition(new Integer(XTabManager.M_CONNECT));
					final MTabManager mtm = m_Plugin.getMarketplaceTabManager();
					mtm.queueTransition(new Integer(MTabManager.M_LOGIN));

					// do all these transitions in separate thread, in order
					Runnable goTrans = new Runnable() {
						public void run() {
							tm.makeTransition();
							xtm.makeTransition();
							mtm.makeTransition();
						}
					};
					new Thread(goTrans).start();
				}
			}
			else if (opcode.equalsIgnoreCase("REP_publisher_list")) {
				// supply the returned list to the new safe pane
				List<String> pubs = thisMsg.getVoucherPublishers();
				if (!tm.getNewSafePane().setPubList(pubs)) {
					Log.error("Voucher publisher list was empty!");
					tm.showError("cannot select publisher",
									"returned voucher publisher list was empty");
				}

				// transition to publisher selection
				tm.queueTransition(new Integer(WTabManager.M_NEW_SAFE1));
				new Thread(this).start();
			}
			else if (opcode.equalsIgnoreCase("REP_key_params")) {
				String encAlgo = thisMsg.getAlgorithm();
				if (encAlgo.isEmpty()) {
					Log.error("Key algorithm for VP missing!");
					tm.showError("cannot generate key pair",
									"missing key algorithm");
					removeFilter(packet.getPacketID());
					return;
				}
				int keyLen = thisMsg.getMin_length();
				if (keyLen == 0) {
					Log.error("Key length for VP missing!");
					tm.showError("cannot generate key pair", "missing key length");
					removeFilter(packet.getPacketID());
					return;
				}

				// tell the new safe pane about these specifications
				tm.getNewSafePane().setKeySpecs(thisMsg);

				// transition to key generation
				tm.queueTransition(new Integer(WTabManager.M_NEW_SAFE2));
				new Thread(this).start();
			}
			else if (opcode.equalsIgnoreCase("REP_safe_data")) {
				// pass this message to the new safe pane
				tm.getNewSafePane().setSafeData(thisMsg);

				// transition to decryption and prompting user for input
				tm.queueTransition(new Integer(WTabManager.M_NEW_SAFE3));
				new Thread(this).start();
			}
			else if (opcode.equalsIgnoreCase("REP_safe_init")) {
				// just echo out the status message received back
				tm.getNewSafePane().setInitRep(thisMsg);

				// proceed to confirmation screen
				tm.queueTransition(new Integer(WTabManager.M_NEW_SAFE4));
				new Thread(this).start();
			}
			else if (opcode.equalsIgnoreCase("REP_network_details")) {
				String encNetBlob = thisMsg.getNetwork_conf_blob();
				byte[] netBlob = StringUtils.decodeBase64(encNetBlob);
				String netConf = new String(netBlob);
				//Log.debug("Got disco conf: " + netConf);

				/* We should now have a blob that looks like this:
				 * <network_conf>
				 * 	<sds-gateway-list>
				 * 		<sds_gateway domain="XMPP domain">host:port</sds_gateway>
				 * 		...
				 * 	</sds-gateway-list>
				 * 	<dht-node-list>
				 * 		<dht_node domain="XMPP domain">host:port</dht-node>
				 * 		...
				 * 	</dht-node-list>
				 * 	<key-server-list>
				 * 		<key_server domain="XMPP domain">host:port</key_server>
				 * 		...
				 * 	</key-server-list>
				 * 	<VoucherAssets>asset1:asset2:...</VoucherAssets>
				 * 	<VoucherUnits>unit1:unit2:...</VoucherUnits>
				 * 	<asset1>
				 * 		<unit1>
				 * 			<conversion>ratio to base units (1.0 base)</conversion>
				 * 			<display>units to be displayed to users</display>
				 * 			<lookup>index for price table (asset:unit)</lookup>
				 * 		</unit1>
				 * 		... repeat for each additonal unit defined for asset1
				 * 	</asset1>
				 * 	... repeat for each additional asset in VoucherAssets
				 * 	<TokenValues>
				 * 		<asset1>value:unit1</asset1>
				 * 		<asset1>value:unit2</asset1>
				 * 		<asset2>value:unit3</asset2>
				 * 		...
				 * 	</TokenValues>
				 * </network_conf>
				 * All of these items must be present.  There should be at least
				 * one item present in each section.
				 */
				List<InetSocketAddress> sdsList
					= new ArrayList<InetSocketAddress>();
				List<InetSocketAddress> dhtList
					= new ArrayList<InetSocketAddress>();
				HashMap<String, ArrayList<InetSocketAddress>> pksList
					= new HashMap<String, ArrayList<InetSocketAddress>>();
				List<String> assetList = new ArrayList<String>();
				List<String> vUnitList = new ArrayList<String>();
				List<String> tUnitList = new ArrayList<String>();
				InetSocketAddress isa = null;
				String sdsDomain = "";
				String dhtDomain = "";
				String ofsDomain = m_Plugin.getOFSConnection().getHost();
				try {
					Element netConfElt = EncodingUtils.buildElement(netConf);
					if (netConfElt == null) {
						throw new DocumentException("<network_conf/> parse error");
					}

					// get SDS addresses
					Element sdsListElt = netConfElt.element("sds-gateway-list");
					if (sdsListElt == null) {
						throw new DocumentException("Missing <sds-gateway-list>");
					}
					Iterator sdsElts = sdsListElt.elementIterator("sds_gateway");
					String addr = null;
					while (sdsElts.hasNext()) {
						Element sdsGtwyElt = (Element) sdsElts.next();
						addr = sdsGtwyElt.getTextTrim();
						if (addr != null && !addr.isEmpty()) {
							String[] address = addr.split(":");
							if (address.length == 2) {
								isa = new InetSocketAddress(address[0],
																Integer.parseInt(address[1]));
								// add to list first if domain matches that of the OFS
								if (isa.getHostString().equals(ofsDomain)) {
									sdsList.add(0, isa);
								}
								else {
									sdsList.add(isa);
								}
							}
							else {
								throw new DocumentException("Improper SDS address: "
																	+ addr);
							}
						}
						else {
							throw new DocumentException("Null SDS address");
						}
						// get XMPP domain as attribute
						String domain = sdsGtwyElt.attributeValue("domain");
						if (domain == null || domain.isEmpty()) {
							throw new DocumentException("Missing SDS domain");
						}
						if (sdsDomain.isEmpty()) {
							sdsDomain = domain;
						}
						else if (!sdsDomain.equals(domain)) {
							Log.error("Inconsistent SDS XMPP domain: " + domain);
						}
					}

					// get DHT addresses
					Element dhtListElt = netConfElt.element("dht-node-list");
					if (dhtListElt == null) {
						throw new DocumentException("Missing <dht-node-list>");
					}
					Iterator dhtElts = dhtListElt.elementIterator("dht_node");
					while (dhtElts.hasNext()) {
						Element dhtNodeElt = (Element) dhtElts.next();
						addr = dhtNodeElt.getTextTrim();
						if (addr != null && !addr.isEmpty()) {
							String[] address = addr.split(":");
							if (address.length == 2) {
								isa = new InetSocketAddress(address[0],
																Integer.parseInt(address[1]));
								// add to list first if domain matches that of the OFS
								if (isa.getHostString().equals(ofsDomain)) {
									dhtList.add(0, isa);
								}
								else {
									dhtList.add(isa);
								}
							}
							else {
								throw new DocumentException("Improper DHT address: "
																	+ addr);
							}
						}
						else {
							throw new DocumentException("Null DHT address");
						}
						// get XMPP domain as attribute
						String domain = dhtNodeElt.attributeValue("domain");
						if (domain == null || domain.isEmpty()) {
							throw new DocumentException("Missing DHT domain");
						}
						if (dhtDomain.isEmpty()) {
							dhtDomain = domain;
						}
						else if (!dhtDomain.equals(domain)) {
							Log.error("Inconsistent DHT XMPP domain: " + domain);
						}
					}

					// get PKS address(es)
					Element pksListElt = netConfElt.element("key-server-list");
					if (pksListElt == null) {
						throw new DocumentException("Missing <key-server-list>");
					}
					Iterator pksElts = pksListElt.elementIterator("key_server");
					while (pksElts.hasNext()) {
						// element should look like: host1:port1::host2:port2 etc.
						Element pksElt = (Element) pksElts.next();
						String pksXMPP = pksElt.attributeValue("domain");
						if (pksXMPP == null || pksXMPP.isEmpty()) {
							throw new DocumentException("Missing PKS domain");
						}
						addr = pksElt.getTextTrim();
						String[] pksPaths = addr.split("::");
						for (int ppp = 0; ppp < pksPaths.length; ppp++) {
							String path = pksPaths[ppp];
							if (path != null && !path.isEmpty()) {
								String[] address = path.split(":");
								if (address.length == 2) {
									isa = new InetSocketAddress(address[0],
																Integer.parseInt(address[1]));
								}
								else {
									throw new DocumentException("Improper PKS address: "
																		+ path);
								}
							}
							else {
								throw new DocumentException("Null PKS address");
							}
							ArrayList<InetSocketAddress> paths = pksList.get(pksXMPP);
							if (paths == null) {
								paths = new ArrayList<InetSocketAddress>();
								pksList.put(pksXMPP, paths);
							}
							// add to list first if domain matches that of the OFS
							if (isa.getHostString().equals(ofsDomain)) {
								paths.add(0, isa);
							}
							else {
								paths.add(isa);
							}
						}
					}

					// get voucher asset list
					Element assetsElt = netConfElt.element("VoucherAssets");
					if (assetsElt == null) {
						throw new DocumentException("Missing <VoucherAssets>");
					}
					String assets = assetsElt.getTextTrim();
					if (assets.isEmpty()) {
						throw new DocumentException("Empty voucher asset list");
					}
					String[] allAssets = assets.split(":");
					for (int iii = 0; iii < allAssets.length; iii++) {
						assetList.add(allAssets[iii]);
					}

					// get voucher unit list (provided only for historical purposes)
					Element unitsElt = netConfElt.element("VoucherUnits");
					if (unitsElt == null) {
						// this is no longer fatal
						Log.warning("Missing <VoucherUnits/>");
					}
					String units = unitsElt.getTextTrim();
					if (units.isEmpty()) {
						// this is no longer fatal
						Log.warning("Empty voucher unit list");
					}
					else {
						String[] allUnits = units.split(":");
						for (int jjj = 0; jjj < allUnits.length; jjj++) {
							if (!vUnitList.contains(allUnits[jjj])) {
								vUnitList.add(allUnits[jjj]);
							}
						}
					}

					/* Now for everything in the asset list, we will have a units
					 * and token specification.  We have to fill in the token value
					 * only from the <TokenValues/> section, processed below.
					 */
					for (String vAsset : assetList) {
						Element assetElt = netConfElt.element(vAsset);
						if (assetElt == null) {
							throw new DocumentException("Missing <" + vAsset + "/>");
						}
						// record config for asset
						VPAssetConfig.AssetConfig aConf
							= new VPAssetConfig.AssetConfig();
						aConf.m_Asset = new String(vAsset);

						// a number of units will be listed
						Iterator unitIt = assetElt.elementIterator();
						while (unitIt.hasNext()) {
							Element unitElt = (Element) unitIt.next();
							String unit = unitElt.getName();

							// find <conversion/>
							Element convElt = unitElt.element("conversion");
							if (convElt == null) {
								throw new DocumentException("Missing <conversion/> in "
														+ "<" + vAsset + "><" + unit + ">");
							}
							Double baseConv = null;
							try {
								baseConv = Double.parseDouble(convElt.getText());
							}
							catch (NumberFormatException nfe) {
								throw new DocumentException("Invalid conversion value "
													+ " for <" + vAsset + "><" + unit + ">");
							}

							// find <display/>
							Element dispElt = unitElt.element("display");
							if (dispElt == null) {
								throw new DocumentException("Missing <display/> in "
														+ "<" + vAsset + "><" + unit + ">");
							}
							String disp = dispElt.getTextTrim();

							// find <lookup/>
							Element lookElt = unitElt.element("lookup");
							if (lookElt == null) {
								throw new DocumentException("Missing <lookup/> in "
														+ "<" + vAsset + "><" + unit + ">");
							}
							String lookup = lookElt.getTextTrim();

							// add unit to config for asset
							VPAssetConfig.AssetConfig.UnitConfig uConf
								= aConf.new UnitConfig();
							uConf.m_Unit = new String(unit);
							uConf.m_Conversion = baseConv.doubleValue();
							uConf.m_Display = disp;
							uConf.m_Lookup = lookup;
							aConf.m_Units.put(unit, uConf);
						}
						VPAssetConfig.m_AssetConfig.put(vAsset, aConf);
					}

					// get token definitions list
					Element tokListElt = netConfElt.element("TokenValues");
					if (tokListElt == null) {
						throw new DocumentException("Missing <TokenValues>");
					}
					for (String vAsset : assetList) {
						/* allow for possibility that there are multiple token units 
						 * for a given listed asset
						 */
						Iterator tokUnitIt = tokListElt.elementIterator(vAsset);
						if (tokUnitIt == null || !tokUnitIt.hasNext()) {
							throw new DocumentException("Missing token definition for "
																+ vAsset);
						}
						while (tokUnitIt.hasNext()) {
							Element tokDefElt = (Element) tokUnitIt.next();
							String defn = tokDefElt.getTextTrim();
							if (defn != null && !defn.isEmpty()) {
								String[] tokVal = defn.split(":");
								if (tokVal.length == 2) {
									String val = tokVal[0];
									String unit = tokVal[1];
									tUnitList.add(unit);
									NumberFormat nf
										= NumberFormat.getInstance(Locale.US);
									nf.setMinimumFractionDigits(XMLToken.M_Precision);
									nf.setMaximumFractionDigits(XMLToken.M_Precision);
									Number value = null;
									try {
										value = nf.parse(val);
									}
									catch (ParseException pe) {
										throw new DocumentException(
															"Illegal token value, " + val, pe);
									}

									// add def'n to appropriate spot in VPAssetConfig
									if (!VPAssetConfig.addTokenDefinition(vAsset, unit,
																			value.doubleValue()))
									{
										Log.error("Error adding token value "
													+ nf.format(value.doubleValue())
													+ " for " + vAsset + ":" + unit);
									}
								/*
									else {
										Log.debug("Added token value "
													+ nf.format(value.doubleValue())
													+ " for " + vAsset + ":" + unit);
									}
								 */
								}
								else {
									throw new DocumentException("Illegal token value "
																		+ "specification, "
																		+ defn);
								}
							}
							else {
								throw new DocumentException("Empty token value spec "
																	+ "for " + vAsset);
							}
						}
					}

					// check we got everything required thus far
					if (assetList.isEmpty()) {
						throw new DocumentException("Empty voucher asset list");
					}
					if (vUnitList.isEmpty()) {
						// no longer fatal
						Log.debug("Empty voucher unit list");
					}
					if (tUnitList.isEmpty()) {
						throw new DocumentException("Empty token definition list");
					}
				}
				catch (DocumentException de) {
					Log.error("Error parsing <network_conf_blob/> data, XML was:\n"
								+ netConf, de);
					tm.showError("Could not initialize publisher network",
									"Error in VP network configuration details");
					removeFilter(packet.getPacketID());
					return;
				}

				// add network elements to the stored network config
				VsSecrets secrets = m_Plugin.getLoginSecrets();
				if (!secrets.initNetwork(sdsDomain, sdsList, dhtDomain, dhtList,
												 pksList))
				{
					Log.error("Error recording <network_conf_blob/> data");
					tm.showError("could not record network configuration details");
					removeFilter(packet.getPacketID());
					return;
				}

				// initialize the SpentTokensCache
				if (!m_Plugin.getTokenCache().initSpentTokens()) {
					Log.error("Unable to initialize saved spent tokens from file");
					tm.showError("Unable to initialize saved spent tokens file");
					// do not allow running without this!
					removeFilter(packet.getPacketID());
					return;
				}

				// obtain the pub key for the VP off the first indicated keyserver
				// (retrieved key will be stored in secrets)

				HomePane hp = tm.getHomePane();
				PublicKey vpKey = m_Plugin.getPKSListener().obtainVPpubkey(null);
				if (vpKey == null) {
					Log.error("Could not get public key for VP following network "
								+ "discovery");

					// reset the progress bar
					hp.setCursor(null);
					m_Plugin.setDefaultStatusBar();

					// really can't run without this
					m_Plugin.doLogout();
					tm.queueTransition(new Integer(WTabManager.M_LOGIN1));
					new Thread(this).start();
					removeFilter(packet.getPacketID());
					return;
				}
				secrets.setVPKey(vpKey);

				// tell the voucher and token tables about the VP's key
				hp.getVoucherModel().setSigKey(secrets.getVPKey());
				hp.getTokenModel().setSigKey(secrets.getVPKey());

				// go on to query SDS async for the list of vouchers/tokens in safe
				if (!tm.querySafeContents(false)) {
					// logout to avoid showing user alarming "safe is emtpy" message
					hp.setCursor(null);
					m_Plugin.setDefaultStatusBar();
					m_Plugin.doLogout();
					tm.queueTransition(new Integer(WTabManager.M_LOGIN1));
					new Thread(this).start();
				}
				else {
					// also query SDS for the list of pending (outgoing) payments
					HousekeepingPane hkp = tm.getHousekeepingPane();
					hkp.queryPayments(false);
				}
			}
			else if (opcode.equalsIgnoreCase("REP_questions")) {
				// pass this message to recovery pane
				RecoveryPane rp = tm.getRecoveryPane();
				rp.setRecoveryQues(thisMsg);

				// go to phase 2 of recovery
				tm.queueTransition(new Integer(WTabManager.M_RECOV2));
				new Thread(this).start();
			}
			else if (opcode.equalsIgnoreCase("REP_recovery_blob")) {
				// pass this message to recovery pane
				RecoveryPane rp = tm.getRecoveryPane();
				rp.setRecoveryBlob(thisMsg);

				// go to phase 3 of recovery
				tm.queueTransition(new Integer(WTabManager.M_RECOV3));
				new Thread(this).start();
			}
			else if (opcode.equalsIgnoreCase("REP_safe_closed")) {
				// send reply to close safe pane
				tm.getCloseSafePane().processCloseReply(thisMsg);
			}
			else if (opcode.equalsIgnoreCase("REP_payment_sent")) {
				// pass this message to the payment pane, which will process confirm
				tm.getPaymentPane().processPaymentReply(thisMsg);

				// redraw token count header
				tm.queueTransition(new Integer(WTabManager.M_HOME_TOKEN_REGEN));
				new Thread(this).start();
			}
			else if (opcode.equalsIgnoreCase("REP_payment_confirmed")) {
				// send message back to incoming pane
				tm.getIncomingPane().processPickupReply(thisMsg);

				// redraw token count header
				tm.queueTransition(new Integer(WTabManager.M_HOME_TOKEN_REGEN));
				new Thread(this).start();
			}
			else if (opcode.equalsIgnoreCase("REP_validation")) {
				// validate, split, or merge succeeded

				// uncommit and "spend" the vouchers and tokens committed to
				// this operation, and record the new voucher(s) received

				tm.getHomePane().processValidateReply(thisMsg);

				// send the user back to the main menu
				tm.queueTransition(new Integer(WTabManager.M_GO_MAIN_MENU));
				new Thread(this).start();
			}
			else if (opcode.equalsIgnoreCase("REP_payment_reverted")) {
				// pass this message to the unclaimed pane
				tm.getUnclaimedPane().processReclaimReply(thisMsg);
			}
			else if (opcode.equalsIgnoreCase("REP_tokens_purged")) {
				// this does not come here (see {@see SpentTokenCache})
				Log.error("Inappropriate opcode: " + opcode);
			}
			else if (opcode.equalsIgnoreCase("REP_tokens_purchased")) {
				// send to tokens pane, which will process confirmation
				tm.getTokensPane().processPurchTokenReply(thisMsg);
			}
			else if (opcode.equalsIgnoreCase("REP_voucher_redeemed")) {
				// send reply to expired pane
				tm.getExpiredPane().processRedeemReply(thisMsg);
			}
			else if (opcode.equalsIgnoreCase("REP_voucher_deleted")) {
				// send reply to expired pane
				tm.getExpiredPane().processDeleteReply(thisMsg);
			}
			else {
				Log.error("Unsupported OFS reply opcode " + opcode + ", XML: "
							+ packet.toXML());
				tm.showError("unsupported OFS reply opcode", opcode);
			}
		}
		else {
			// not for us, probably a payment message from OnionPay or a website
			Log.debug("Non-OFSMessage packet received, with XML: "
						+ packet.toXML());
		}

		// retire this ID from the list of active packet filters
		removeFilter(packet.getPacketID());
	}

	/**
	 * Method to implement Runnable, so that we can invoke UI elements in an
	 * asynchronous manner, without risking the termination or hijacking of
	 * this thread by Swing.  Also, we need this callback to return quickly so
	 * that further messages may be processed, and graphics updates can be slow.
	 */
	public void run() {
		// obtain UI tab manager from plugin and do the queued state transition
		m_Plugin.getWalletTabManager().makeTransition();
	}

	/**
	 * method to add a filter to the list of packets we're accepting replies for
	 * @param ID the packet ID we're expecting in a reply
	 * @param filter the filter to add, which watches for the ID
	 */
	public void addFilter(String ID, PacketIDFilter filter) {
		if (ID == null || ID.isEmpty() || filter == null
			|| m_FilterMap.containsKey(ID))
		{
			// redundant
			return;
		}
		m_FilterMap.put(ID, filter);
		resetConnFilter();
	}

	/**
	 * method to remove a filter from the list of replies we're listening for
	 * @param ID the ID of the packet filter to remove
	 */
	public void removeFilter(String ID) {
		if (ID == null || ID.isEmpty() || !m_FilterMap.containsKey(ID)) {
			// nothing to do
			return;
		}
		m_FilterMap.remove(ID);
		resetConnFilter();
	}

	/**
	 * private utility method to rebuild and reset the filter on this connection,
	 * ORing together all filters extant in the map
	 */
	private synchronized void resetConnFilter() {
		// rebuild filter of all outstanding requests
		OrFilter of = new OrFilter();
		for (String packetId : m_FilterMap.keySet()) {
			PacketIDFilter filt = m_FilterMap.get(packetId);
			of.addFilter(filt);
		}
		XMPPConnection conn = m_Plugin.getOFSConnection();
		conn.removePacketListener(this);
		// add back the web Message filter if we have one
		PacketFilter msgFilter = m_Plugin.getWebMsgFilter();
		if (msgFilter != null) {
			conn.addPacketListener(this, msgFilter);
		}

		// add new listener if map is non-empty
		if (!m_FilterMap.isEmpty()) {
			conn.addPacketListener(this, of);
		}
	}

	// END methods
}
