/*
 * last modified---
 * 	08-06-14/jet/ Spark plugin version
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	Parse {@link OFSMessage} packets.  Parses the <tt>vsc_ofs</tt>
 * 	sub-document and creates an instance of {@link OFSMessage}.
 *
 * notes---
 * 	see underlying message definitions in vsc_ofs.x
 */

package org.silentvault.spark;

import org.xmlpull.v1.XmlPullParser;

import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.packet.DefaultPacketExtension;
import org.jivesoftware.smack.provider.IQProvider;

import org.jivesoftware.spark.util.log.Log;

import java.util.ArrayList;


/**
 * Implements the IQProvider interface to receive and parse replies from an OFS
 * using the vsc_ofs element within the jabber:iq:voucher-safe#vsc-ofs
 * namespace.
 * Uses the XmlPullParser to interpret each message according to its REP_*
 * opcode, and creates a {@link OFSMessage} object to hold it.
 *
 * @author Kevin Wilkerson
 */
public final class OFSMessageProvider implements IQProvider {
	// members
	private static final String	M_EndTag = "vsc_ofs";
	
	/**
	 * constructor
	 */
	public OFSMessageProvider() {
	}

	/**
	 * parse the IQ packet's child element
	 * @param parser the parser, positioned at the start of the child element
	 */
	public IQ parseIQ(XmlPullParser parser) throws Exception {
		OFSMessage ofsMsg = new OFSMessage();
		// NB: these values should be vsc_ofs and jabber:iq:voucher-safe#vsc-ofs
		ofsMsg.addExtension(new DefaultPacketExtension(parser.getName(),
																	  parser.getNamespace()));
		ofsMsg.setType(IQ.Type.RESULT);

		// first, determine the opcode
		int eventType = parser.next();
		String opcode = "";
		if (eventType == XmlPullParser.START_TAG) {
			if (parser.getName().equals("opcode")) {
				opcode = parser.nextText().trim();
			}
			else {
				Log.error("Missing opcode");
				throw new Exception("missing " + M_EndTag + " opcode");
			}
		}
		else {
			Log.error("Missing opcode");
			throw new Exception("missing " + M_EndTag + " opcode");
		}
		ofsMsg.setOpcode(opcode);

		// Now based on the opcode, pull out the other data.
		// NB: any reply can have errcode/errmsg, and we cannot tell (from here)
		// what the overall type of the IQ packet is.

		if (opcode.equalsIgnoreCase("REP_long-phrase_challenge")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					int idx = 0;
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("long_phrase_len")) {
						idx = Integer.parseInt(parser.nextText());
						ofsMsg.setLong_phrase_len(idx);
					}
					else if (parser.getName().equals("long_phrase_index1")) {
						idx = Integer.parseInt(parser.nextText());
						ofsMsg.setLong_phrase_index1(idx);
					}
					else if (parser.getName().equals("long_phrase_index2")) {
						idx = Integer.parseInt(parser.nextText());
						ofsMsg.setLong_phrase_index2(idx);
					}
					else if (parser.getName().equals("long_phrase_index3")) {
						idx = Integer.parseInt(parser.nextText());
						ofsMsg.setLong_phrase_index3(idx);
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_login-blob")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("loginblob")) {
						ofsMsg.setLoginblob(parser.nextText().trim());
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_publisher_list")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("voucher-publisher-list")) {
						// go parse the complete list separately (see below)
						parsePublisherList(parser, ofsMsg);
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_price_data")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("priceData")) {
						ofsMsg.setPriceData(parser.nextText().trim());
						// NB: this field still needs to be base64-decoded
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_key_params")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("algorithm")) {
						ofsMsg.setAlgorithm(parser.nextText().trim());
					}
					else if (parser.getName().equals("min_length")) {
						ofsMsg.setMin_length(Integer.parseInt(parser.nextText()));
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_safe_data")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("certificate")) {
						ofsMsg.setCertificate(parser.nextText().trim());
					}
					else if (parser.getName().equals("key_server_addr")) {
						ofsMsg.setKey_server_addr(parser.nextText().trim());
					}
					else if (parser.getName().equals("cap_blob")) {
						ofsMsg.setCap_blob(parser.nextText().trim());
					}
					else if (parser.getName().equals("pin_length")) {
						ofsMsg.setPin_length(Short.parseShort(parser.nextText()));
					}
					else if (parser.getName().equals("recovery_ques_1")) {
						ofsMsg.setRecovery_ques_1(parser.nextText().trim());
					}
					else if (parser.getName().equals("recovery_ques_2")) {
						ofsMsg.setRecovery_ques_2(parser.nextText().trim());
					}
					else if (parser.getName().equals("recovery_ques_3")) {
						ofsMsg.setRecovery_ques_3(parser.nextText().trim());
					}
					else if (parser.getName().equals("recovery_ques_4")) {
						ofsMsg.setRecovery_ques_4(parser.nextText().trim());
					}
					else if (parser.getName().equals("recovery_ques_5")) {
						ofsMsg.setRecovery_ques_5(parser.nextText().trim());
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_safe_init")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("status_msg")) {
						ofsMsg.setStatus_msg(parser.nextText());
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_network_details")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("network_conf_blob")) {
						ofsMsg.setNetwork_conf_blob(parser.nextText().trim());
						// NB: this blob still needs to be decoded and sub-parsed
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_safe_closed")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("status_msg")) {
						ofsMsg.setStatus_msg(parser.nextText());
					}
					else if (parser.getName().equals("syncPending")) {
						String sync = parser.nextText().toLowerCase();
						ofsMsg.setSyncPending(Boolean.parseBoolean(sync));
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_payment_sent")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("syncPending")) {
						String sync = parser.nextText().toLowerCase().trim();
						ofsMsg.setSyncPending(Boolean.parseBoolean(sync));
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_payment_confirmed")
					|| opcode.equalsIgnoreCase("REP_payment_reverted")) {

			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("confirmation")) {
						ofsMsg.setConfirmation(parser.nextText().trim());
						// NB: this blob still needs to be decrypted and sub-parsed
					}
					else if (parser.getName().equals("syncPending")) {
						String sync = parser.nextText().toLowerCase().trim();
						ofsMsg.setSyncPending(Boolean.parseBoolean(sync));
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_validation")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("validation")) {
						ofsMsg.setValidation(parser.nextText().trim());
						// NB: this blob still needs to be decrypted and sub-parsed
					}
					else if (parser.getName().equals("syncPending")) {
						String sync = parser.nextText().toLowerCase().trim();
						ofsMsg.setSyncPending(Boolean.parseBoolean(sync));
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_tokens_purged")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("purged")) {
						ofsMsg.setPurged(parser.nextText().trim());
						// NB: this blob still needs to be decrypted and sub-parsed
					}
					else if (parser.getName().equals("syncPending")) {
						String sync = parser.nextText().toLowerCase().trim();
						ofsMsg.setSyncPending(Boolean.parseBoolean(sync));
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_tokens_purchased")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("purchased")) {
						ofsMsg.setPurchased(parser.nextText().trim());
						// NB: this blob still needs to be decrypted and sub-parsed
					}
					else if (parser.getName().equals("syncPending")) {
						String sync = parser.nextText().toLowerCase().trim();
						ofsMsg.setSyncPending(Boolean.parseBoolean(sync));
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_questions")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					int idx = 0;
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("recovery_ques_1")) {
						ofsMsg.setRecovery_ques_1(parser.nextText().trim());
					}
					else if (parser.getName().equals("recovery_ques_2")) {
						ofsMsg.setRecovery_ques_2(parser.nextText().trim());
					}
					else if (parser.getName().equals("recovery_ques_3")) {
						ofsMsg.setRecovery_ques_3(parser.nextText().trim());
					}
					else if (parser.getName().equals("recovery_ques_4")) {
						ofsMsg.setRecovery_ques_4(parser.nextText().trim());
					}
					else if (parser.getName().equals("recovery_ques_5")) {
						ofsMsg.setRecovery_ques_5(parser.nextText().trim());
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_recovery_blob")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					int idx = 0;
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("recovery_blob")) {
						ofsMsg.setRecovery_blob(parser.nextText().trim());
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_voucher_redeemed")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("confirmation")) {
						ofsMsg.setConfirmation(parser.nextText().trim());
						// NB: this blob still needs to be decrypted and sub-parsed
					}
					else if (parser.getName().equals("syncPending")) {
						String sync = parser.nextText().toLowerCase().trim();
						ofsMsg.setSyncPending(Boolean.parseBoolean(sync));
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("REP_voucher_deleted")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrcode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrmsg(parser.nextText());
					}
					else if (parser.getName().equals("syncPending")) {
						String sync = parser.nextText().toLowerCase().trim();
						ofsMsg.setSyncPending(Boolean.parseBoolean(sync));
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else {
			Log.error("Unsupported opcode " + opcode);
			throw new Exception("Unsupported " + M_EndTag + " opcode " + opcode);
		}
		return ofsMsg;
	}

	/**
	 * private sub-method to parse the supplied list of voucher publishers
	 * @param parser the parser, positioned at the start of the list
	 * @param ofsMsg the message being built
	 */
	private void parsePublisherList(XmlPullParser parser, OFSMessage ofsMsg)
		throws Exception
	{
		if (parser == null || ofsMsg == null) {
			throw new Exception("Null input passed");
		}
		boolean done = false;
		ArrayList<String> publishers = new ArrayList<String>();
		while (!done) {
			int eventType = parser.next();
			if (eventType == XmlPullParser.START_TAG) {
				if (parser.getName().equals("voucher_publisher")) {
					publishers.add(parser.nextText().trim());
				}
			}
			else if (eventType == XmlPullParser.END_TAG) {
				if (parser.getName().equals("voucher-publisher-list")) {
					done = true;
				}
			}
		}
		ofsMsg.setVoucherPublishers(publishers);
	}
}
