/*
 * last modified---
 * 	08-15-18/kbw/ explicitly cancel PacketCollectors after use
 * 	02-28-18/kbw/ add compression and other security checks on connect
 * 	08-23-14/jet/ invoke reorderGateways() appropriately, to speed up logins
 * 	08-06-14/jet/ Spark plugin version
 * 	11-21-13/jet/ on PKS connect failures, retry once after :80 connect
 * 	10-12-13/jet/ support multiple paths to the same PKS XMPP domain; pick
 * 					  domains at random but try paths in listed order
 * 	06-13-13/jet/ modify obtainVPpubkey() to also handle SVX/SSM key requests
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	listen for messages from the PKS which are replies to IQ packets we sent
 *
 * notes---
 * 	see underlying messages in sys_pks.x
 */

package org.silentvault.spark;

import javax.swing.*;

import org.jivesoftware.smack.PacketListener;
import org.jivesoftware.smack.util.StringUtils;
import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.ConnectionConfiguration;
import org.jivesoftware.smack.XMPPException;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.SmackConfiguration;
import org.jivesoftware.smack.packet.Packet;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.filter.PacketIDFilter;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.ui.wallet.WTabManager;

import java.security.PublicKey;
import java.security.cert.X509Certificate;
import java.net.InetSocketAddress;
import java.util.Random;
import java.util.HashMap;
import java.util.ArrayList;


/**
 * class to handle processing of PKS reply packets, within the context of the
 * jabber:iq:voucher-safe#sys-pks namespace.
 *
 * @author Kevin Wilkerson
 */
public final class PKSListener implements PacketListener {
	// BEGIN data members

	/**
	 * back pointer reference to the Plugin object
	 */
	private WalletClient		m_Plugin;

	/**
	 * holder for reference to the message pending to the OFS once the VP
	 * pubkey is received a (REQ_init_safe)
	 */
	private OFSMessage		m_PendInitSafe;

	/**
	 * the live connection to the PKS on which we are listening
	 */
	private XMPPConnection	m_PKSConn;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the reference to the plugin of which we are a part
	 */
	public PKSListener(WalletClient plugin) {
		m_Plugin = plugin;
	}

	// method required to implement the PacketListener interface
	/**
	 * deal with a reply packet from the PKS
	 * @param packet the packet (will really be a PKSMessage)
	 */
	public void processPacket(Packet packet) {
		final XMPPConnection OFSconn = m_Plugin.getOFSConnection();
		final WTabManager tm = m_Plugin.getWalletTabManager();

		// verify we got what's expected
		if (packet instanceof PKSMessage) {
			PKSMessage thisMsg = (PKSMessage) packet;
			String opcode = thisMsg.getOpcode();
			//Log.debug("Listener got PKS message opcode " + opcode);

			// record message
			VscState stateObj = m_Plugin.getVscState();
			stateObj.setLastActivity(thisMsg);

			// see if it's an error
			if (thisMsg.getType() == IQ.Type.ERROR) {
				Log.error("Got an error: code " + thisMsg.getErrcode() + ", "
							+ thisMsg.getErrmsg());
				tm.showError("PKS error code " + thisMsg.getErrcode() + ", "
								+ thisMsg.getErrmsg());
				m_PKSConn.removePacketListener(this);
				return;
			}

			// take further actions based on the opcode
			if (opcode.equalsIgnoreCase("REP_pubkey")) {

				PublicKey vpPubkey
					= EncodingUtils.getPubkeyFromBase64Str(thisMsg.getPublic_key());
				if (vpPubkey != null) {
					// encrypt the details blob using the VP's pubkey
					String encDetBlob = EncodingUtils.makeBase64PubkeyEncStr(
															m_PendInitSafe.getDetails_blob(),
															vpPubkey);
					if (encDetBlob == null) {
						Log.error("Could not encrypt details blob for VP");
						tm.showError("error encrypting details blob with VP pubkey");
						m_PKSConn.removePacketListener(this);
						return;
					}
					m_PendInitSafe.setDetails_blob(encDetBlob);
				}
				else {
					Log.error("VP pubkey transmitted was not valid");
					tm.showError("Invalid publisher pubkey");
					m_PKSConn.removePacketListener(this);
					return;
				}

				// Now send the constructed REQ_init_safe to the OFS.
				// First, register a listener for the OFS's reply.

				PacketIDFilter pFilter
					= new PacketIDFilter(m_PendInitSafe.getPacketID());
				OFSconn.addPacketListener(m_Plugin.getOFSListener(), pFilter);

				// send it
				OFSconn.sendPacket(m_PendInitSafe);
			/*
				Log.debug(m_PendInitSafe.getOpcode() + " sent to voucher publisher "
							+ m_PendInitSafe.getVoucherPublisher());
			 */

				// record outgoing packet and then toss it
				m_Plugin.getVscState().setLastActivity(m_PendInitSafe);
				m_PendInitSafe = null;
			}
			else {
				Log.error("Unrecognized PKS opcode " + opcode + ", XML: "
							+ packet.toXML());
				tm.showError("PKS reply opcode unrecognized, " + opcode);
			}
		}
		else {
			Log.error("Non-PKSMessage packet received, with XML: "
						+ packet.toXML());
		}

		// in any event, remove ourselves as listener
		m_PKSConn.removePacketListener(this);
	}

	/**
	 * record the pending message to OFS
	 * @param msg the message to the OFS under construction
	 */
	public void setPendingMsg(OFSMessage msg) { m_PendInitSafe = msg; }

	/**
	 * record the connection to the PKS, so we can close it later
	 * @param conn the connection (must be currently live)
	 */
	public void setPKSConnection(XMPPConnection conn) {
		if (conn != null && conn.isConnected()) {
			if (m_PKSConn != null) {
				// blow away old one first
				m_PKSConn.disconnect();
			}
			m_PKSConn = conn;
		}
	}

	/**
	 * utility method to pick a PKS domain at random from the list of keyservers
	 * @return the randomly selected XMPP domain
	 */
	private String getRandomPKSDomain() {
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		HashMap<String, ArrayList<InetSocketAddress>> keyServers
			= secrets.getKeyServers();
		if (keyServers.isEmpty()) {
			Log.error("No key servers defined");
			return null;
		}

		// pick a domain at random
		ArrayList<String> pksDomains = new ArrayList<String>(keyServers.keySet());
		int numDom = pksDomains.size();
		int domIdx = 0;
		if (numDom > 1) {
			// make random pick
			Random rng = new Random();
			domIdx = rng.nextInt(numDom);
		}
		String pickedDom = pksDomains.get(domIdx);
		return pickedDom;
	}

	/**
	 * utility method to obtain a public key off a randomly selected PKS.
	 * NB: this method blocks pending the key server's reply
	 * @param keyId the ID of the VP, SVX, or SSM whose key we want
	 * @return the corresponding public key
	 */
	public PublicKey obtainVPpubkey(String keyId) {
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		if (keyId == null || keyId.isEmpty()) {
			// assume we want the VP's pubkey
			keyId = secrets.getPublisher();
		}
		VscState vsState = m_Plugin.getVscState();
		WTabManager tm = m_Plugin.getWalletTabManager();

		// obtain a random pick from the key server list
		String pksDomain = getRandomPKSDomain();
		if (pksDomain == null) {
			Log.error("Cannot look up pubkey Id " + keyId
						+ ", no key servers defined");
			tm.showError("cannot look up public key for key Id " + keyId,
							"no key servers defined");
			return null;
		}

		// build message
		PKSMessage getVPkey = new PKSMessage();
		getVPkey.setType(IQ.Type.GET);
		getVPkey.setTo(pksDomain);
		getVPkey.setOpcode("REQ_get_pubkey");
		getVPkey.setUser_id(keyId);
		getVPkey.setPacketID("get_" + vsState.getNextPKSid());

		// see whether we need to open a new connection
		boolean needNewConn = true;
		if (m_PKSConn != null && m_PKSConn.isConnected()
			 && m_PKSConn.getServiceName().equals(pksDomain))
		{
			needNewConn = false;
		}

		if (needNewConn) {
			if (m_PKSConn != null && m_PKSConn.isConnected()) {
				// must have connection to another domain previously picked
				closePKSConnection();
			}

			// reduce connect times (on connect, Smack seems to use 3x reply time)
			int orgTimeout = SmackConfiguration.getPacketReplyTimeout();
			SmackConfiguration.setPacketReplyTimeout(10000);

			// take the list in the order given
			ArrayList<InetSocketAddress> paths
				= secrets.getKeyServers().get(pksDomain);
			InetSocketAddress usedServ = null;
			for (InetSocketAddress keyServ : paths) {
				// make a connection to the PKS Openfire server via the given path
				ConnectionConfiguration cc
					= new ConnectionConfiguration(keyServ.getHostString(),
															keyServ.getPort(),
															pksDomain);
				cc.setSecurityMode(ConnectionConfiguration.SecurityMode.required);
				cc.setSelfSignedCertificateEnabled(true);
				cc.setCompressionEnabled(true);
				cc.setNotMatchingDomainCheckEnabled(true);
				XMPPConnection pksConn = new XMPPConnection(cc);

				/* If the host we're about to connect to is the same as the OFS
				 * host, AND if it's more than two minutes since we were initialized
				 * by Spark, then just skip over it.  Don't even bother pinging it
				 * first (won't work).  This is due to a peculilarity of how
				 * PageKites handle raw port connections vs. HTTP proxies.
				 */
				String ofsHost = m_Plugin.getOFSConnection().getHost();
				if (m_Plugin.getTimeSinceInit() >= 120
					 && keyServ.getHostString().equals(ofsHost))
				{
					// skip it
					Log.debug("Skipped over PKS host " + keyServ.getHostString()
								+ " because it's same as OFS domain " + ofsHost
								+ " which is stale by " + m_Plugin.getTimeSinceInit()
								+ " seconds.");
					continue;
				}
				
				// if we're not on the first iteration, try a pagekite 80 ping first
				boolean pinged = false;
				if (paths.indexOf(keyServ) != 0) {
					if (!m_Plugin.port80KitePing(keyServ.getHostString())) {
						Log.error("Ping " + keyServ.getHostString() + ":80 failed");
						// presumably, connection also won't work
						continue;
					}
					pinged = true;
				}
				try {
					pksConn.connect();
				}
				catch (XMPPException xe) {
					Log.error("Could not connect to PKS " + pksDomain + " at "
								+ keyServ.getHostString() + ":" + keyServ.getPort());

					// attempt retry after pinging pagekite 80 port
					if (!pinged
						 && m_Plugin.port80KitePing(keyServ.getHostString()))
					{
						try {
							pksConn.connect();
						}
						catch (XMPPException xe2) {
							Log.error("Also failed after :80 ping");
							continue;
						}
					}
					else {
						continue;
					}
				}
				// perform anon login
				try {
					pksConn.loginAnonymously();
					Log.debug("Successful connect to PKS " + pksDomain + " at path "
								+ keyServ.getHostString() + ":" + keyServ.getPort());
					setPKSConnection(pksConn);
					usedServ = keyServ;
					break;
				}
				catch (XMPPException xe) {
					Log.error("Could not login anon to PKS " + pksDomain
								+ " at path " + keyServ.getHostString() + ":"
								+ keyServ.getPort(), xe);
					pksConn.disconnect();
				}
			}
			SmackConfiguration.setPacketReplyTimeout(orgTimeout);
			if (m_PKSConn == null || !m_PKSConn.isConnected()) {
				tm.showError("unable to fetch " + keyId + " pubkey from key server",
								"PKS connect failed", "log out and try again later");
				return null;
			}

			// if needed, resort the gateway lists to the connected hostname is 1st
			String connHost = m_PKSConn.getHost();
			if (paths.indexOf(usedServ) != 0) {
				m_Plugin.reorderGateways(paths, connHost);
				//Log.debug("Shuffled host " + connHost + " to top of PKS list");
				// do the same for SDS and DHT if we don't have connections yet
				if (m_Plugin.getSDSListener().getSDSConnection() == null) {
					m_Plugin.reorderGateways(
								m_Plugin.getLoginSecrets().getSDSGateways(), connHost);
					//Log.debug("Shuffled host " + connHost + " to top of SDS list");
				}
				if (m_Plugin.getDHTListener().getDHTConnection() == null) {
					m_Plugin.reorderGateways(
									m_Plugin.getLoginSecrets().getDHTNodes(), connHost);
					//Log.debug("Shuffled host " + connHost + " to top of DHT list");
				}
			}
		}

		// create packet filter to back collector, and install collector
		PacketIDFilter pksFilt
			= new PacketIDFilter(getVPkey.getPacketID());
		PacketCollector repColl = m_PKSConn.createPacketCollector(pksFilt);

		// send the request to the PKS with a blocking read
		getVPkey.setFrom(m_PKSConn.getUser());
		m_PKSConn.sendPacket(getVPkey);
		vsState.setLastActivity(getVPkey);
		//Log.debug("Fetching pubkey for " + keyId);
		Packet pksRep = repColl.nextResult(m_Plugin.getTimeout());
		repColl.cancel();
		if (pksRep == null) {
			// timed out
			Log.error("PKS not responding");
			tm.showError("unable to fetch " + keyId + " pubkey from key server",
							"timeout, PKS not responding", "try again later");
			closePKSConnection();
			return null;
		}

		// parse reply from PKS
		if (!(pksRep instanceof PKSMessage)) {
			Log.error("Weird non-PKSMessage IQ received from PKS: "
						+ pksRep.toXML());
			return null;
		}

		PKSMessage ksRep = (PKSMessage) pksRep;
		String pksOp = ksRep.getOpcode();

		// record message as last activity from this partner
		vsState.setLastActivity(ksRep);

		// see if it's an error
		if (ksRep.getType() == IQ.Type.ERROR) {
			Log.error("Got an error: code " + ksRep.getErrcode() + ", "
						+ ksRep.getErrmsg());
			tm.showError("unable to fetch " + keyId + " pubkey from key server",
							"received error code " + ksRep.getErrcode() + " , "
							+ ksRep.getErrmsg());
			return null;
		}

		// take further actions based on the opcode
		PublicKey retKey = null;
		if (pksOp.equalsIgnoreCase("REP_pubkey")) {
			// now make this into a real key
			retKey = EncodingUtils.getPubkeyFromBase64Str(ksRep.getPublic_key());
			if (retKey == null) {
				Log.error("Null pubkey after parsing");
				tm.showError("unable to fetch " + keyId + " pubkey from key server",
								"unparseable key received");
			}
		}
		else {
			Log.error("Aberrant opcode " + pksOp + " on PKS reply");
		}
		return retKey;
	}

	/**
	 * utility method to obtain a VS's public key off a randomly selected PKS.
	 * NB: this method blocks pending the key server's reply
	 * @param vsNum the voucher safe number whose key we want to look up
	 * @return the key associated with the retrieved certificate
	 */
	public PublicKey obtainVSpubkey(String vsNum) {
		VscState vsState = m_Plugin.getVscState();
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		WTabManager tm = m_Plugin.getWalletTabManager();

		// obtain a random pick from the key server list
		String pksDomain = getRandomPKSDomain();
		if (pksDomain == null) {
			Log.error("Cannot look up wallet key, no key servers defined");
			tm.showError("cannot look up public key for wallet Id " + vsNum,
							"no key servers defined");
			return null;
		}

		// build message
		PKSMessage getCert = new PKSMessage();
		getCert.setType(IQ.Type.GET);
		getCert.setTo(pksDomain);
		getCert.setPacketID("get_" + vsState.getNextPKSid());
		getCert.setOpcode("REQ_get_cert");
		getCert.setUser_id(vsNum);

		// see whether we need to open a new connection
		boolean needNewConn = true;
		if (m_PKSConn != null && m_PKSConn.isConnected()
			 && m_PKSConn.getServiceName().equals(pksDomain))
		{
			needNewConn = false;
		}

		if (needNewConn) {
			if (m_PKSConn != null && m_PKSConn.isConnected()) {
				// must have connection to another domain previously picked
				closePKSConnection();
			}

			// reduce connect times (on connect, Smack seems to use 3x reply time)
			int orgTimeout = SmackConfiguration.getPacketReplyTimeout();
			SmackConfiguration.setPacketReplyTimeout(10000);

			// take the list in the order given
			ArrayList<InetSocketAddress> paths
				= secrets.getKeyServers().get(pksDomain);
			InetSocketAddress usedServ = null;
			for (InetSocketAddress keyServ : paths) {
				// make a connection to the PKS Openfire server via the given path
				ConnectionConfiguration cc
					= new ConnectionConfiguration(keyServ.getHostString(),
															keyServ.getPort(),
															pksDomain);
				cc.setSecurityMode(ConnectionConfiguration.SecurityMode.required);
				cc.setSelfSignedCertificateEnabled(true);
				cc.setCompressionEnabled(true);
				cc.setNotMatchingDomainCheckEnabled(true);
				XMPPConnection pksConn = new XMPPConnection(cc);

				/* If the host we're about to connect to is the same as the OFS
				 * host, AND if it's more than two minutes since we were initialized
				 * by Spark, then just skip over it.  Don't even bother pinging it
				 * first (won't work).  This is due to a peculilarity of how
				 * PageKites handle raw port connections vs. HTTP proxies.
				 */
				String ofsHost = m_Plugin.getOFSConnection().getHost();
				if (m_Plugin.getTimeSinceInit() >= 120
					 && keyServ.getHostString().equals(ofsHost))
				{
					// skip it
					Log.debug("Skipped over PKS host " + keyServ.getHostString()
								+ " because it's same as OFS domain " + ofsHost
								+ " which is stale by " + m_Plugin.getTimeSinceInit()
								+ " seconds.");
					continue;
				}
				
				// if we're not on the first iteration, try a pagekite 80 ping first
				boolean pinged = false;
				if (paths.indexOf(keyServ) != 0) {
					if (!m_Plugin.port80KitePing(keyServ.getHostString())) {
						Log.error("Ping " + keyServ.getHostString() + ":80 failed");
						// presumably, connection also won't work
						continue;
					}
					pinged = true;
				}
				try {
					pksConn.connect();
				}
				catch (XMPPException xe) {
					Log.error("Could not connect to PKS " + pksDomain + " at "
								+ keyServ.getHostString() + ":" + keyServ.getPort());

					// attempt retry after pinging pagekite 80 port
					if (!pinged
						 && m_Plugin.port80KitePing(keyServ.getHostString()))
					{
						try {
							pksConn.connect();
						}
						catch (XMPPException xe2) {
							Log.error("Also failed after :80 ping");
							continue;
						}
					}
					else {
						continue;
					}
				}
				// perform anon login
				try {
					pksConn.loginAnonymously();
					Log.debug("Successful connect to PKS " + pksDomain + " at path "
								+ keyServ.getHostString() + ":" + keyServ.getPort());
					setPKSConnection(pksConn);
					usedServ = keyServ;
					break;
				}
				catch (XMPPException xe) {
					Log.error("Could not login anon to PKS " + pksDomain
								+ " at path " + keyServ.getHostString() + ":"
								+ keyServ.getPort(), xe);
					pksConn.disconnect();
				}
			}
			SmackConfiguration.setPacketReplyTimeout(orgTimeout);
			if (m_PKSConn == null || !m_PKSConn.isConnected()) {
				tm.showError("unable to fetch wallet Id " + vsNum
								+ " pubkey from key server",
								"could not connect to PKS",
								"log out and try again later");
				return null;
			}

			// if needed, resort the gateway lists to the connected hostname is 1st
			String connHost = m_PKSConn.getHost();
			if (paths.indexOf(usedServ) != 0) {
				m_Plugin.reorderGateways(paths, connHost);
				//Log.debug("Shuffled host " + connHost + " to top of PKS list");
				// do the same for SDS and DHT if we don't have connections yet
				if (m_Plugin.getSDSListener().getSDSConnection() == null) {
					m_Plugin.reorderGateways(
								m_Plugin.getLoginSecrets().getSDSGateways(), connHost);
					//Log.debug("Shuffled host " + connHost + " to top of SDS list");
				}
				if (m_Plugin.getDHTListener().getDHTConnection() == null) {
					m_Plugin.reorderGateways(
									m_Plugin.getLoginSecrets().getDHTNodes(), connHost);
					//Log.debug("Shuffled host " + connHost + " to top of DHT list");
				}
			}
		}

		// create packet filter to back collector, and install collector
		PacketIDFilter pksFilt
			= new PacketIDFilter(getCert.getPacketID());
		PacketCollector repColl = m_PKSConn.createPacketCollector(pksFilt);

		// send the request to the PKS with a blocking read
		getCert.setFrom(m_PKSConn.getUser());
		m_PKSConn.sendPacket(getCert);
		vsState.setLastActivity(getCert);
		Packet pksRep = repColl.nextResult(m_Plugin.getTimeout());
		repColl.cancel();
		if (pksRep == null) {
			// timed out
			Log.error("PKS not responding");
			tm.showError("unable to fetch wallet Id " + vsNum
							+ " pubkey from key server",
							"timeout, PKS not responding", "try again later");
			closePKSConnection();
			return null;
		}

		// parse reply from PKS
		if (!(pksRep instanceof PKSMessage)) {
			Log.error("Weird non-PKSMessage IQ received from PKS: "
						+ pksRep.toXML());
			return null;
		}

		PKSMessage ksRep = (PKSMessage) pksRep;
		String pksOp = ksRep.getOpcode();

		// record message as last activity from this partner
		vsState.setLastActivity(ksRep);

		// see if it's an error
		if (ksRep.getType() == IQ.Type.ERROR) {
			tm.showError("unable to fetch wallet Id " + vsNum
							+ " pubkey from key server",
							"received error code " + ksRep.getErrcode() + ", "
							+ ksRep.getErrmsg());
			return null;
		}

		// take further actions based on the opcode
		PublicKey vsPubKey = null;
		if (pksOp.equalsIgnoreCase("REP_cert")) {

			// now make this into a real key
			X509Certificate vsCert
				= EncodingUtils.getCertFromBase64Str(ksRep.getCertificate());
			if (vsCert != null) {
				vsPubKey = vsCert.getPublicKey();
			}
			else {
				Log.error("Null VS cert after parsing");
				tm.showError("unable to fetch wallet Id " + vsNum
								+ " pubkey from key server",
								"unparseable certificate received");
				return null;
			}
		}
		else {
			Log.error("Aberrant opcode " + pksOp + " on PKS reply");
			tm.showError("unable to fetch wallet Id " + vsNum
							+ " pubkey from key server", "bad reply opcode");
			return null;
		}
		return vsPubKey;
	}

	/**
	 * close the connection to the PKS
	 */
	public void closePKSConnection() {
		if (m_PKSConn != null) {
			m_PKSConn.disconnect();
			m_PKSConn = null;
		}
	}

	// END methods
}
