/*
 * last modified---
 * 	10-02-14/jet/ SSM -> SVM
 * 	08-06-14/jet/ Spark plugin version
 * 	07-03-13/jet/ split IQ extension into SVGSSMIQ
 * 	06-06-13/jet/ new
 *
 * purpose---
 * 	class to manipulate messages received to/from a OFS-SVG to be forwarded
 * 	to a SVM
 *
 * notes---
 * 	see underlying message definitions in vsc_ofs-svm.xml
 *
 * 	If this class is changed, org.silentvault.openfire.VSCSVMMessage must
 * 	also be updated.
 */

package org.silentvault.spark;

import org.jivesoftware.smack.packet.Message;
import org.jivesoftware.smack.packet.PacketExtension;

import org.jivesoftware.spark.util.log.Log;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Locale;


/**
 * Holds all the pieces of a message which is part of the protocol using the
 * jabber:message:silent-market#vsc_ofs namespace between an OFS (SVG) and a
 * VSC (Voucher Safe Client) using the element vsc_ofs.  This class knows how
 * to emit itself as packet extension XML and parse extension XML into a
 * Dom4j Element structure.
 *
 * @author Justin Turrell
 */
public final class SVGSVMMessage extends Message implements PacketExtension {
	// BEGIN data members
	/**
	 * The opcode of this request (begins with NFY_*).
	 * This value determines which of the other fields will contain valid data.
	 */
	private String					m_Opcode;

	/**
	 * seriesId is used with opcode NFY_ticker_event.  It represents the ID of
	 * a series being reported in a ticker event, following a market transaction.
	 */
	private String					m_SeriesId;

	/**
	 * quantity is used with opcode NFY_ticker_event.  It represents the amount
	 * of a market clearing transaction being reported in a ticker event.
	 */
	private double					m_Quantity;

	/**
	 * price is used with opcode NFY_ticker_event.  It represents the price at
	 * which the market transaction being reported cleared.
	 */
	private float					m_Price;

	/**
	 * timestamp is used with opcode NFY_ticker_event.  It represents the date
	 * and time at which the transaction being reported was settled, expressed
	 * as a Unix timestamp (milliseconds since the epoch).
	 */
	private long					m_Timestamp;

	/**
	 *	orderIds is used with opcode NFY_ticker_event.  It represents the two
	 *	order ID numbers which settled against each other, separated by a colon:
	 *	buyer's:seller's.
	 */
	private String					m_OrderIds;

	/**
	 * signature is used with opcode NFY_ticker_event.  It represents the privkey
	 * signature of the SVM on the ticker fields, base64-encoded.
	 */
	private String					m_Signature;

	/**
	 * level is used with opcode NFY_sys_broadcast.  It represents the type or
	 * severity of the broadcast message: info, warning, error, or alert.
	 */
	private String					m_Level;

	/**
	 * text is used with opcode NFY_sys_broadcast.  It represents the text of
	 * the message to be broadcast to the indicated VSCs.
	 */
	private String					m_Text;

	// END data members

	// BEGIN methods
	/**
	 * nullary constructor
	 */
	public SVGSVMMessage() {
		super();
		m_Opcode = m_SeriesId = m_OrderIds = m_Signature = m_Level = m_Text = "";
	}


	// get accessor methods
	/**
	 * obtain the opcode
	 * @return the opcode
	 */
	public String getOpcode() { return m_Opcode; }

	/**
	 * obtain the series ID (valid when opcode = NFY_ticker_event)
	 * @return the series to which the ticker event pertains
	 */
	public String getSeriesId() { return m_SeriesId; }

	/**
	 * obtain the quantity (valid when opcode = NFY_ticker_event)
	 * @return the quantity for the ticker event
	 */
	public double getQuantity() { return m_Quantity; }

	/**
	 * obtain the clearing price (valid when opcode = NFY_ticker_event)
	 * @return the price at which the trade settled
	 */
	public float getPrice() { return m_Price; }

	/**
	 * obtain the clearing timestamp (valid when opcode = NFY_ticker_event)
	 * @return the settle date/time, in msec since the epoch
	 */
	public long getTimestamp() { return m_Timestamp; }

	/**
	 * obtain the order IDs which settled against each other (valid when opcode
	 * = NFY_ticker_event)
	 * @return the orders, in format buyer:seller
	 */
	public String getOrderIds() { return m_OrderIds; }

	/**
	 * obtain the SVM's signature (valid when opcode = NFY_ticker_event)
	 * @return the privkey signature, base64-encoded
	 */
	public String getSignature() { return m_Signature; }

	/**
	 * obtain the message severity level (valid when opcode = NFY_sys_broadcast)
	 * @return the level (one of info, warning, error, or alert)
	 */
	public String getLevel() { return m_Level; }

	/**
	 * obtain the message text (valid when opcode = NFY_sys_broadcast)
	 * @return the text to be broadcast to the user
	 */
	public String getText() { return m_Text; }


	// set access methods
	/**
	 * set the opcode
	 * @param opcode the opcode
	 */
	public void setOpcode(String opcode) {
		if (opcode != null && !opcode.isEmpty()) {
			m_Opcode = new String(opcode);
		}
	}

	/**
	 * set the series ID (needed when opcode = NFY_ticker_event)
	 * @param series the ID of the series the ticker event concerns
	 */
	public void setSeriesId(String series) {
		if (series != null && !series.isEmpty()) {
			m_SeriesId = new String(series);
		}
	}

	/**
	 * set the quantity (needed when opcode = NFY_ticker_event)
	 * @param amt the quantity involved in the ticker event
	 */
	public void setQuantity(double amt) {
		if (amt > 0.0) {
			m_Quantity = amt;
		}
	}

	/**
	 * set the price (needed when opcode = NFY_ticker_event)
	 * @param price the settle price of the transaction being reported
	 */
	public void setPrice(float price) {
		if (price > 0.0F) {
			m_Price = price;
		}
	}

	/**
	 * set the timestamp (needed when opcode = NFY_ticker_event)
	 * @param time the transaction timestamp, in msec since the epoch
	 */
	public void setTimestamp(long time) {
		if (time > 0L) {
			m_Timestamp = time;
		}
	}

	/**
	 * set the settled order IDs (needed when opcode = NFY_ticker_event)
	 * @param ids the order IDs, in format buyer:seller
	 */
	public void setOrderIds(String ids) {
		if (ids != null && ids.indexOf(':') != -1) {
			m_OrderIds = new String(ids);
		}
	}

	/**
	 * set the SVM signature (needed when opcode = NFY_ticker_event)
	 * @param sig the privkey signature, base64-encoded
	 */
	public void setSignature(String sig) {
		if (sig != null && !sig.isEmpty()) {
			m_Signature = new String(sig);
		}
	}

	/**
	 * set the message severity level (needed when opcode = NFY_sys_broadcast)
	 * @param level the level to set (one of info, warning, error, or alert)
	 */
	public void setLevel(String level) {
		if (level != null && !level.isEmpty()) {
			m_Level = new String(level);
		}
	}

	/**
	 * set the broadcast message text (needed when opcode = NFY_sys_broadcast)
	 * @param text the text to be sent
	 */
	public void setText(String text) {
		if (text != null && !text.isEmpty()) {
			m_Text = new String(text);
		}
	}


	// methods required to implement the PacketExtension interface
	/**
	 * provide the root element name
	 * @return the name
	 */
	public String getElementName() { return "vsc_ofs"; }

	/**
	 * provide the XML namespace of the root element
	 * @return the namespace
	 */
	public String getNamespace() {
		return "jabber:message:silent-market#vsc-ofs";
	}

	/**
	 * provide the extension element as XML
	 * @return the extended child element as a XML string
	 */
	public String toXML() {
		StringBuilder xml = new StringBuilder(2048);
		xml.append("<" + getElementName() + " xmlns=\"" + getNamespace() + "\">");
		xml.append("<opcode>" + m_Opcode + "</opcode>");

		// based on the value of the opcode, emit additional fields
		if (m_Opcode.equalsIgnoreCase("NFY_ticker_event")) {
			xml.append("<ticker>");
			xml.append("<seriesId>" + m_SeriesId + "</seriesId>");
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
			df.setMaximumFractionDigits(XMLVoucher.M_Precision);
			df.setMinimumFractionDigits(XMLVoucher.M_Precision);
			df.setGroupingSize(0);
			xml.append("<quantity>" + df.format(m_Quantity) + "</quantity>");
			xml.append("<price>" + df.format(m_Price) + "</price>");
			xml.append("<timestamp>" + m_Timestamp + "</timestamp>");
			xml.append("<orderIds>" + m_OrderIds + "</orderIds>");
			xml.append("</ticker>");
			xml.append("<signature>" + m_Signature + "</signature>");
		}
		else if (m_Opcode.equalsIgnoreCase("NFY_sys_broadcast")) {
			xml.append("<level>" + m_Level + "</level>");
			xml.append("<text>" + m_Text + "</text>");
		}
		else {
			Log.error("Improper Message opcode " + m_Opcode);
		}
		xml.append("</" + getElementName() + ">");
		return xml.toString();
	}

	// END methods
}
