/*
 * last modified---
 * 	10-02-14/jet/ SSM -> SVM
 * 	08-06-14/jet/ Spark plugin version
 * 	06-06-13/jet/ new
 *
 * purpose---
 * 	Parse {@link SVGSVMMessage} and {@link SVGSVMIQ} packets.  Parses the
 * 	<tt>vsc_ofs</tt> sub-document and creates an instance of the message.
 *
 * notes---
 * 	see underlying message definitions in vsc_ofs-svm.xml
 */

package org.silentvault.spark;

import org.xmlpull.v1.XmlPullParser;

import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.packet.PacketExtension;
import org.jivesoftware.smack.packet.DefaultPacketExtension;
import org.jivesoftware.smack.provider.IQProvider;
import org.jivesoftware.smack.provider.PacketExtensionProvider;

import org.jivesoftware.spark.util.log.Log;


/**
 * Implements the IQProvider interface to receive and parse replies from an OFS
 * using the vsc_ofs element within the jabber:iq:silent-market#vsc-ofs
 * namespace.  Implements the PacketExtensionProvider interface to receive and
 * parse incoming messages from a OFS-SVM, according to the vsc_ofs element of
 * the jabber:message:silent-market#vsc-ofs namespace.
 * Uses the XmlPullParser to interpret each message according to its opcode,
 * and creates a {@link SVGSVMMessage} object to hold it, or to represent the
 * packet extention.
 *
 * @author Justin Turrell
 */
public final class SVGSVMMessageProvider
	implements IQProvider, PacketExtensionProvider
{
	// members
	private static final String	M_EndTag = "vsc_ofs";
	
	// methods
	/**
	 * constructor
	 */
	public SVGSVMMessageProvider() {
	}

	/**
	 * parse the IQ packet's child element
	 * @param parser the parser, positioned at the start of the child element
	 */
	public IQ parseIQ(XmlPullParser parser) throws Exception {
		SVGSVMIQ ofsMsg = new SVGSVMIQ();
		// NB: these values should be vsc_ofs and jabber:iq:silent-market#vsc-ofs
		ofsMsg.addExtension(new DefaultPacketExtension(parser.getName(),
																	  parser.getNamespace()));
		ofsMsg.setType(IQ.Type.RESULT);

		// first, determine the opcode
		int eventType = parser.next();
		String opcode = "";
		if (eventType == XmlPullParser.START_TAG) {
			if (parser.getName().equals("opcode")) {
				opcode = parser.nextText().trim();
			}
			else {
				Log.error("Missing opcode");
				throw new Exception("missing " + M_EndTag + " opcode");
			}
		}
		else {
			Log.error("Missing opcode");
			throw new Exception("missing " + M_EndTag + " opcode");
		}
		ofsMsg.setOpcode(opcode);

		/* Now based on the opcode, pull out the other data.
		 * NB: any reply can have errcode/errmsg, and we cannot tell (from here)
		 * what the overall type of the IQ packet is.
		 */
		if (opcode.equalsIgnoreCase("REP_svm_message")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					int idx = 0;
					if (parser.getName().equals("errcode")) {
						ofsMsg.setErrCode(Integer.parseInt(parser.nextText()));
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("errmsg")) {
						ofsMsg.setErrMsg(parser.nextText());
						ofsMsg.setType(IQ.Type.ERROR);
					}
					else if (parser.getName().equals("rep_blob")) {
						ofsMsg.setRepBlob(parser.nextText());
						// NB: this blob still needs to be decrypted and sub-parsed
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else {
			Log.error("Unsupported opcode " + opcode);
			throw new Exception("Unsupported " + M_EndTag + " opcode " + opcode);
		}
		return ofsMsg;
	}

	/**
	 * parse the packet extension (used with a Message packet)
	 * @param parser the parser, positioned at the start of the extension
	 * @return the extension element
	 */
	public PacketExtension parseExtension(XmlPullParser parser) throws Exception
	{
		SVGSVMMessage ofsMsg = new SVGSVMMessage();
		// NB: these values should be vsc_ofs and
		// 	 jabber:message:silent-market#vsc-ofs
		ofsMsg.addExtension(new DefaultPacketExtension(parser.getName(),
																	  parser.getNamespace()));

		// first, determine the opcode
		int eventType = parser.next();
		String opcode = "";
		if (eventType == XmlPullParser.START_TAG) {
			if (parser.getName().equals("opcode")) {
				opcode = parser.nextText().trim();
			}
			else {
				Log.error("Missing opcode");
				throw new Exception("missing " + M_EndTag + " opcode");
			}
		}
		else {
			Log.error("Missing opcode");
			throw new Exception("missing " + M_EndTag + " opcode");
		}
		ofsMsg.setOpcode(opcode);

		/* Now based on the opcode, pull out the other data.
		 * NB: no errors are passed, these are actually Message types not IQs.
		 */
		if (opcode.equalsIgnoreCase("NFY_ticker_event")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					int idx = 0;
					if (parser.getName().equals("ticker")) {
						// go parse the ticker item separately (see below)
						parseTickerEvent(parser, ofsMsg);
					}
					else if (parser.getName().equals("signature")) {
						ofsMsg.setSignature(parser.nextText().trim());
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else if (opcode.equalsIgnoreCase("NFY_sys_broadcast")) {
			boolean done = false;
			while (!done) {
				eventType = parser.next();
				if (eventType == XmlPullParser.START_TAG) {
					int idx = 0;
					if (parser.getName().equals("level")) {
						ofsMsg.setLevel(parser.nextText());
					}
					else if (parser.getName().equals("text")) {
						ofsMsg.setText(parser.nextText());
					}
				}
				else if (eventType == XmlPullParser.END_TAG) {
					if (parser.getName().equals(M_EndTag)) {
						done = true;
					}
				}
			}
		}
		else {
			Log.error("Unsupported opcode " + opcode);
			throw new Exception("Unsupported " + M_EndTag + " opcode " + opcode);
		}
		return ofsMsg;
	}

	/**
	 * private sub-method to parse the supplied ticker event
	 * @param parser the parser, positioned at the start of the list
	 * @param ofsMsg the message being built
	 */
	private void parseTickerEvent(XmlPullParser parser, SVGSVMMessage ofsMsg)
		throws Exception
	{
		if (parser == null || ofsMsg == null) {
			throw new Exception("Null input passed");
		}
		boolean done = false;
		while (!done) {
			int eventType = parser.next();
			if (eventType == XmlPullParser.START_TAG) {
				if (parser.getName().equals("seriesId")) {
					ofsMsg.setSeriesId(parser.nextText().trim());
				}
				else if (parser.getName().equals("quantity")) {
					ofsMsg.setQuantity(Double.parseDouble(parser.nextText()));
				}
				else if (parser.getName().equals("price")) {
					ofsMsg.setPrice(Float.parseFloat(parser.nextText()));
				}
				else if (parser.getName().equals("timestamp")) {
					ofsMsg.setPrice(Long.parseLong(parser.nextText()));
				}
				else if (parser.getName().equals("orderIds")) {
					ofsMsg.setOrderIds(parser.nextText().trim());
				}
			}
			else if (eventType == XmlPullParser.END_TAG) {
				if (parser.getName().equals("ticker")) {
					done = true;
				}
			}
		}
	}
}
