/*
 * last modified---
 * 	03-28-18/kbw/ permit REP_svx_message in getChildElementXML()
 * 	08-06-14/jet/ Spark plugin version
 * 	07-02-13/jet/ new, split from SVGSVXMessage
 *
 * purpose---
 * 	class to manipulate IQ messages received to/from a OFS-SVG to be forwarded
 * 	to a SVX
 *
 * notes---
 * 	see underlying message definitions in vsc_ofs-svx.xml
 *
 * 	If this class is changed, org.silentvault.openfire.VSCSVXMessage must
 * 	also be updated.
 */

package org.silentvault.spark;

import org.jivesoftware.smack.packet.IQ;

import org.jivesoftware.spark.util.log.Log;
 

/**
 * Holds all the pieces of a message which is part of the protocol using the
 * jabber:iq:silent-vault#vsc-ofs namespace between an OFS (SVG) and a VSC
 * (Voucher Safe Client) using the element vsc_ofs.  This class knows how to
 * emit itself as XML and parse XML into a Dom4j Element structure.  It is
 * used as a utility class for recording incoming VSC messages and outgoing
 * replies, and processing result/error stanzas.
 *
 * @author Justin Turrell
 */
public final class SVGSVXIQ extends IQ {
	// BEGIN data members
	/**
	 * The opcode of this request (begins with REQ_) or reply (begins with REP_).
	 * This value determines which of the other fields will contain valid data.
	 */
	private String					m_Opcode;

	/**
	 * errcode is used when the IQ.Type is ERROR.  It represents the nature of
	 * the error, and uses standard HTTP return values (see HttpURLConnection).
	 * Thus 4xx represents a fatal permanent error while 5xx represents a
	 * temporary failure.
	 */
	private int						m_ErrCode;

	/**
	 * errMsg is used when the IQ.Type is ERROR.  It represents the textual
	 * description of the error condition.
	 */
	private String					m_ErrMsg;

	/**
	 * sv_exchange is used with opcode REQ_svx_message.  It represents the
	 * ID of the SVX to which we are forwarding requests.  This ID can be used
	 * to obtain the SVX's pubkey from a PKS.
	 */
	private String					m_SVXId;

	/**
	 * msg_blob is used with opcode REQ_svx_message.  It represents a blob
	 * of data, base64-encoded and encrypted with the SVX's pubkey, which was
	 * generated by the connected VSC and to be forwarded to it SVX by us.  When
	 * decrypted and parsed, this data will be instantiated as a ClientBlock,
	 * but it is opaque to the gateway.
	 */
	private String					m_MsgBlob;

	/**
	 * rep_blob is used with opcode REP_svx_message.  It represents a blob of
	 * data, base64-encoded and encrypted with the VSC's pubkey, which was
	 * generated by the SVX and sent back to the VSC via us.  This data
	 * is instantiated as a SVXBlock, prior to encryption.
	 */
	private String					m_RepBlob;

	// END data members

	// BEGIN methods
	/**
	 * nullary constructor
	 */
	public SVGSVXIQ() {
		super();
		m_Opcode = m_ErrMsg = m_SVXId = m_MsgBlob = m_RepBlob = "";
	}


	// get accessor methods
	/**
	 * obtain the opcode
	 * @return the opcode
	 */
	public String getOpcode() { return m_Opcode; }

	/**
	 * obtain current error code, valid only when the IQ.Type = ERROR
	 * @return the error code
	 */
	public int getErrCode() { return m_ErrCode; }

	/**
	 * obtain the textual representation of the error, valid only when the
	 * IQ.Type = ERROR
	 * @return the error message
	 */
	public String getErrMsg() { return m_ErrMsg; }

	/**
	 * obtain the SVX's ID (valid when opcode = REQ_svx_message)
	 * @return the SVX ID
	 */
	public String getSVXId() { return m_SVXId; }

	/**
	 * obtain the raw client block (valid when opcode = REQ_svx_message)
	 * @return the VSC's data block, still base64-encoded and encrypted
	 */
	public String getMsgBlob() { return m_MsgBlob; }

	/**
	 * obtain the raw server reply block (valid when opcode = REP_svx_message)
	 * @return the SVX's data block, encrypted and base64-encoded
	 */
	public String getRepBlob() { return m_RepBlob; }


	// set access methods
	/**
	 * set the opcode
	 * @param opcode the opcode
	 */
	public void setOpcode(String opcode) {
		if (opcode != null && !opcode.isEmpty()) {
			m_Opcode = new String(opcode);
		}
	}

	/**
	 * set the error code value (needed when IQ.Type = ERROR)
	 * @param ecode the error code
	 */
	public void setErrCode(int ecode) { m_ErrCode = ecode; }

	/**
	 * set the error message value (needed when IQ.Type = ERROR)
	 * @param emsg the error message
	 */
	public void setErrMsg(String emsg) {
		if (emsg != null) {
			m_ErrMsg = new String(emsg);
		}
	}

	/**
	 * set the SVX ID (needed when opcode = REQ_svx_message)
	 * @param svxId the ID of the SVX
	 */
	public void setSVXId(String svxId) {
		if (svxId != null && !svxId.isEmpty()) {
			m_SVXId = new String(svxId);
		}
	}

	/**
	 * set the client block data (needed when opcode = REQ_svx_message)
	 * @param cBlock the data from the VSC, encrypted and base64-encoded
	 */
	public void setMsgBlob(String cBlock) {
		if (cBlock != null && !cBlock.isEmpty()) {
			m_MsgBlob = new String(cBlock);
		}
	}

	/**
	 * set the server block data (needed when opcode = REP_svx_message)
	 * @param sBlock the data from the SVX, encrypted and base64-encoded
	 */
	public void setRepBlob(String sBlock) {
		if (sBlock != null && !sBlock.isEmpty()) {
			m_RepBlob = new String(sBlock);
		}
	}

	/**
	 * provide the child element as XML (required to extend IQ)
	 * @return the extended child element as a XML string
	 */
	public String getChildElementXML() {
		StringBuilder xml = new StringBuilder(10240);
		xml.append("<vsc_ofs xmlns=\"jabber:iq:silent-vault#vsc-ofs\">");
		xml.append("<opcode>" + m_Opcode + "</opcode>");

		// if this is an error stanza, supply code and message
		if (getType() == IQ.Type.ERROR) {
			xml.append("<errcode>" + m_ErrCode + "</errcode>");
			xml.append("<errmsg>" + m_ErrMsg + "</errmsg>");
			xml.append("</vsc_ofs>");
			return xml.toString();
		}

		// Based on the value of the opcode, emit additional fields.
		if (m_Opcode.equalsIgnoreCase("REQ_svx_message")) {
			xml.append("<sv_exchange>" + m_SVXId + "</sv_exchange>");
			xml.append("<msg_blob>" + m_MsgBlob + "</msg_blob>");
		}
		else if (m_Opcode.equalsIgnoreCase("REP_svx_message")) {
			xml.append("<rep_blob>" + m_RepBlob + "</rep_blob>");
		}
		else {
			Log.error("Unsupported outgoing OFS opcode, " + m_Opcode);
		}
		xml.append("</vsc_ofs>");
		return xml.toString();
	}

	// END methods
}
