/*
 * last modified---
 * 	08-06-14/jet/ Spark plugin version
 * 	07-02-13/jet/ break out IQ implementation to SVGSVXIQ
 * 	06-06-13/jet/ new
 *
 * purpose---
 * 	class to manipulate messages received to/from a OFS-SVG to be forwarded
 * 	to a SVX
 *
 * notes---
 * 	see underlying message definitions in vsc_ofs-svx.xml
 *
 * 	If this class is changed, org.silentvault.openfire.VSCSVXMessage must
 * 	also be updated.
 */

package org.silentvault.spark;

import org.jivesoftware.smack.packet.Message;
import org.jivesoftware.smack.packet.PacketExtension;

import org.jivesoftware.spark.util.log.Log;
 

/**
 * Holds all the pieces of a message which is part of the protocol using the
 * jabber:message:silent-vault#vsc_ofs namespace between an OFS (SVG) and a
 * VSC (Voucher Safe Client) using the element vsc_ofs.  This class knows how
 * to emit itself as XML as a packet extension and parse XML for extensions
 * into a Dom4j Element structure.  It is used as a utility class for recording
 * incoming VSC messages.
 *
 * @author Justin Turrell
 */
public final class SVGSVXMessage extends Message implements PacketExtension {
	// BEGIN data members
	/**
	 * The opcode of this request (begins with NFY_).
	 * This value determines which of the other fields will contain valid data.
	 */
	private String					m_Opcode;

	/**
	 * level is used with opcode NFY_sys_broadcast.  It represents the type or
	 * severity of the broadcast message: info, warning, error, or alert.
	 */
	private String					m_Level;

	/**
	 * text is used with opcode NFY_sys_broadcast.  It represents the text of
	 * the message to be broadcast to the indicated VSCs.
	 */
	private String					m_Text;

	// END data members

	// BEGIN methods
	/**
	 * nullary constructor
	 */
	public SVGSVXMessage() {
		super();
		m_Opcode = m_Level = m_Text = "";
	}


	// get accessor methods
	/**
	 * obtain the opcode
	 * @return the opcode
	 */
	public String getOpcode() { return m_Opcode; }

	/**
	 * obtain the message severity level (valid when opcode = NFY_sys_broadcast)
	 * @return the level (one of info, warning, error, or alert)
	 */
	public String getLevel() { return m_Level; }

	/**
	 * obtain the message text (valid when opcode = NFY_sys_broadcast)
	 * @return the text to be broadcast to the user
	 */
	public String getText() { return m_Text; }


	// set access methods
	/**
	 * set the opcode
	 * @param opcode the opcode
	 */
	public void setOpcode(String opcode) {
		if (opcode != null && !opcode.isEmpty()) {
			m_Opcode = new String(opcode);
		}
	}

	/**
	 * set the message severity level (needed when opcode = NFY_sys_broadcast)
	 * @param level the level to set (one of info, warning, error, or alert)
	 */
	public void setLevel(String level) {
		if (level != null && !level.isEmpty()) {
			m_Level = new String(level);
		}
	}

	/**
	 * set the broadcast message text (needed when opcode = NFY_sys_broadcast)
	 * @param text the text to be sent
	 */
	public void setText(String text) {
		if (text != null && !text.isEmpty()) {
			m_Text = new String(text);
		}
	}


	// methods required to implement the PacketExtension interface
	/**
	 * provide the root element name
	 * @return the name
	 */
	public String getElementName() { return "vsc_ofs"; }

	/**
	 * provide the XML namespace of the root element
	 * @return the namespace
	 */
	public String getNamespace() {
		return "jabber:message:silent-vault#vsc-ofs";
	}

	/**
	 * provide the extension element as XML
	 * @return the extended child element as a XML string
	 */
	public String toXML() {
		StringBuilder xml = new StringBuilder(2048);
		xml.append("<" + getElementName() + " xmlns=\"" + getNamespace() + "\">");
		xml.append("<opcode>" + m_Opcode + "</opcode>");

		// based on the value of the opcode, emit additional fields
		if (m_Opcode.equalsIgnoreCase("NFY_sys_broadcast")) {
			xml.append("<level>" + m_Level + "</level>");
			xml.append("<text>" + m_Text + "</text>");
		}
		else {
			Log.error("Improper Message opcode " + m_Opcode);
		}
		xml.append("</" + getElementName() + ">");
		return xml.toString();
	}

	// END methods
}
