/*
 * last modified---
 * 	08-06-14/jet/ Spark plugin version
 * 	08-23-13/jet/ add m_ExtWallet to REQ_voucher_purchase, for refunds
 * 	06-06-13/jet/ new
 *
 * purpose---
 * 	class to instantiate a message from the VSC to the SVX
 *
 * notes---
 * 	maintain in parallel with org.silentvault.exchange.ClientBlock
 *
 * 	these messages are delivered embedded inside a <msg_blob/> element
 * 	within a REQ_svx_message message routed through the OFS-SVG
 * 	
 * 	see underlying message definitions in vsc_svx.xml
 */

package org.silentvault.spark;

import org.jivesoftware.smack.util.Base64;

import org.jivesoftware.spark.util.log.Log;

import java.util.Date;
import java.util.Locale;
import java.util.ArrayList;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.security.PrivateKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.SignatureException;


/**
 * This class instantiates client requests sent to the SVX via the OFS gateway.
 * It knows how to emit the various message fields based on the designated
 * opcode, and can sign with the client's private key.
 *
 * @author Justin Turrell
 */
public final class SVXClientBlock {
	// BEGIN data members
	/**
	 * native XML as encrypted in the sv_exchange element for the server
	 */
	private String				m_ClientBlockXML;

	/**
	 * opcode sent by the client
	 */
	private String				m_Opcode;

	/**
	 * privkey signature of client, used with all request opcodes except
	 * REQ_discover_config and REQ_get_offers
	 */
	private String				m_Signature;

	/**
	 * walletId of the VSC (used with all request opcodes except REQ_get_offers)
	 */
	private String				m_WalletId;

	/**
	 * JID at which VSC is connected to the OFS (used with opcode
	 * REQ_discover_config)
	 */
	private String				m_JID;

	/**
	 * an asset type to be dealt with (used with opcodes REQ_voucher_purchase,
	 * REQ_voucher_sale, REQ_make_offer, REQ_accept_offer)
	 */
	private String				m_Asset;

	/**
	 * a currency to be spent or received as an external (non-voucher) payment
	 * (used with opcodes REQ_voucher_purchase and REQ_voucher_sale)
	 */
	private String				m_Currency;

	/**
	 * an external wallet identifier, such as bitcoin hash, at which a VSC
	 * will be receiving an external payment from the SVX
	 * (used with opcodes REQ_voucher_sale and REQ_voucher_purchase)
	 */
	private String				m_ExtWallet;

	/**
	 * the order ID involved in a request (used with opcode REQ_voucher_sold)
	 */
	private Integer			m_OrderId;

	/**
	 * the offer/order quantity in voucher asset units (used with opcodes
	 * REQ_voucher_purchase, REQ_voucher_sale, REQ_make_offer)
	 */
	private Double				m_Quantity;

	/**
	 * colon-separated list of assets wanted in a trade (used with opcodes
	 * REQ_make_offer and REQ_get_offers)
	 */
	private String				m_WantedAssets;

	/**
	 * the ratio of offered asset to wanted asset which VSC wants (used with
	 * opcodes REQ_make_offer and REQ_get_offers)
	 */
	private Float				m_Price;

	/**
	 * designated timestamp at which an offer made by the VSC shall expire
	 * (used with opcode REQ_make_offer)
	 */
	private Date				m_Expiration;

	/**
	 * an offer ID (unique DB sequence number) referencing an existing offer
	 * (used with opcodes REQ_fund_offer, REQ_accept_offer, REQ_offer_pickup,
	 * and REQ_cancel_offer)
	 */
	private Integer			m_OfferId;

	/**
	 * colon-separated list of offered assets to which VSC wants a search to be
	 * restricted (used with opcode REQ_get_offers)
	 */
	private String				m_OfferedAssets;

	/**
	 * minimum quantity size stipulated to restrict a search (used with opcode
	 * REQ_get_offers)
	 */
	private Double				m_MinAmount;

	/**
	 * maximum quantity size stipulated to restrict a search (used with opcode
	 * REQ_get_offers)
	 */
	private Double				m_MaxAmount;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 */
	public SVXClientBlock() {
		m_Opcode = m_WalletId = m_Signature = m_Asset = m_Currency = "";
		m_ExtWallet = m_WantedAssets = m_OfferedAssets = m_ClientBlockXML = "";
		m_OrderId = new Integer(0);
		m_OfferId = new Integer(0);
		m_Quantity = new Double(0.0);
		m_Price = new Float(0.0F);
		m_MinAmount = new Double(0.0);
		m_MaxAmount = new Double(0.0);
	}


	// get access methods
	/**
	 * obtain original XML sent from client (before encryption)
	 * @return the unparsed XML
	 */
	public String getClientBlockXML() { return m_ClientBlockXML; }

	/**
	 * obtain the opcode
	 * @return the opcode
	 */
	public String getOpcode() { return m_Opcode; }

	/**
	 * obtain the client's signature
	 * @return the digital privkey signature provided on m_ClientBlockXML
	 */
	public String getSignature() { return m_Signature; }

	/**
	 * obtain the wallet ID
	 * @return the ID of the wallet the VSC is running
	 */
	public String getWalletId() { return m_WalletId; }

	/**
	 * obtain the connected JID
	 * @return the user@server/resource at which the user is connected at the OFS
	 */
	public String getJID() { return m_JID; }

	/**
	 * obtain the asset type
	 * @return the asset type requested
	 */
	public String getAsset() { return m_Asset; }

	/**
	 * obtain the currency type
	 * @return the external currency type requested
	 */
	public String getCurrency() { return m_Currency; }

	/**
	 * obtain the external wallet ID
	 * @return the wallet Id to be used for an external payment
	 */
	public String getExtWallet() { return m_ExtWallet; }

	/**
	 * obtain the order ID requested
	 * @return the ID of the order in the database
	 */
	public int getOrderId() { return m_OrderId.intValue(); }

	/**
	 * obtain the order or offer quantity requested
	 * @return the quantity
	 */
	public double getQuantity() { return m_Quantity.doubleValue(); }

	/**
	 * obtain the list of wanted assets
	 * @return the list, colon-separated
	 */
	public String getWantedAssets() { return m_WantedAssets; }

	/**
	 * obtain the price requested
	 * @return the price (as a ratio of OFFERED to WANTED, or a market price)
	 */
	public float getPrice() { return m_Price.floatValue(); }

	/**
	 * obtain the expiry date, if specified
	 * @return the date, or null if not set
	 */
	public Date getExpiration() { return m_Expiration; }

	/**
	 * obtain the offer ID requested
	 * @return the ID of the offer in the database
	 */
	public int getOfferId() { return m_OfferId.intValue(); }

	/**
	 * obtain the list of offered assets
	 * @return the list, colon-separated
	 */
	public String getOfferedAssets() { return m_OfferedAssets; }

	/**
	 * obtain the minimum offer amount requested
	 * @return the min amount
	 */
	public double getMinAmount() { return m_MinAmount.doubleValue(); }

	/**
	 * obtain the maximum offer amount requested
	 * @return the max amount
	 */
	public double getMaxAmount() { return m_MaxAmount.doubleValue(); }


	// set access methods
	/**
	 * re-record (override) the XML to be sent in the client message
	 * @param xml the original unparsed XML, exclusive of delimiters
	 */
	public void setClientBlockXML(String xml) {
		if (xml != null && !xml.isEmpty()) {
			m_ClientBlockXML = new String(xml);
		}
	}

	/**
	 * set the opcode for the client message
	 * @param op the opcode to set (assumed to be valid)
	 */
	public void setOpcode(String op) {
		if (op != null && op.startsWith("REQ_")) {
			m_Opcode = new String(op);
		}
		else {
			Log.error("Improper incoming opcode, " + op);
		}
	}

	/**
	 * set the client's signature
	 * @param sig the privkey signature of the client on other fields
	 */
	public void setSignature(String sig) {
		if (sig != null && !sig.isEmpty()) {
			m_Signature = new String(sig);
		}
	}

	/**
	 * set the wallet ID for a request
	 * @param wId the wallet ID (VS#) for which we're sending a request
	 */
	public void setWalletId(String wId) {
		if (wId != null && wId.indexOf('-') != -1) {
			m_WalletId = new String(wId);
		}
		else {
			Log.error("Illegal wallet Id, " + wId);
		}
	}

	/**
	 * set the JID at which client is connected to the OFS
	 * @param jid the JID of the VSC (assumed to be valid)
	 */
	public void setJID(String jid) {
		if (jid != null && !jid.isEmpty()) {
			m_JID = new String(jid);
		}
	}

	/**
	 * set the asset type for a request
	 * @param asset the voucher asset type we're dealing with
	 */
	public void setAsset(String asset) {
		if (asset != null && !asset.isEmpty()) {
			m_Asset = new String(asset);
		}
	}

	/**
	 * set the currency type for a request
	 * @param curr the external currency type we're dealing with
	 */
	public void setCurrency(String curr) {
		if (curr != null && !curr.isEmpty()) {
			m_Currency = new String(curr);
		}
	}

	/**
	 * set the external wallet ID
	 * @param extWallet the external wallet to receive payment from SVX
	 */
	public void setExtWallet(String extWallet) {
		if (extWallet != null && !extWallet.isEmpty()) {
			m_ExtWallet = new String(extWallet.trim());
		}
	}

	/**
	 * set the order ID wanted
	 * @param oId the order ID in the database
	 */
	public void setOrderId(int oId) {
		if (oId > 0) {
			m_OrderId = new Integer(oId);
		}
	}

	/**
	 * set the order or order quantity
	 * @param amt the amount of the particular asset involved, in voucher units
	 */
	public void setQuantity(double amt) {
		if (amt > 0.0) {
			m_Quantity = new Double(amt);
		}
	}

	/**
	 * set the wanted asset types
	 * @param wanted the colon-separated list
	 */
	public void setWantedAssets(String wanted) {
		if (wanted != null && !wanted.isEmpty()) {
			m_WantedAssets = new String(wanted);
		}
	}

	/**
	 * set the desired price
	 * @param price the price to set, either a ratio or market price
	 */
	public void setPrice(float price) {
		if (price >= 0.0F) {
			m_Price = new Float(price);
		}
	}

	/**
	 * set the offer expiration date
	 * @param date the timestamp to set, in msec since the epoch
	 */
	public void setExpiration(long date) {
		// suitability (i.e. at least 24 hours in future) is not validated here
		if (date > 0L) {
			m_Expiration = new Date(date);
		}
	}

	/**
	 * set the offer ID wanted
	 * @param oId the offer ID in the database
	 */
	public void setOfferId(int oId) {
		if (oId > 0) {
			m_OfferId = new Integer(oId);
		}
	}

	/**
	 * set the offered asset types
	 * @param offered the colon-separated list
	 */
	public void setOfferedAssets(String offered) {
		if (offered != null && !offered.isEmpty()) {
			m_OfferedAssets = new String(offered);
		}
	}

	/**
	 * set the minimum offer quantity
	 * @param amt the min amount
	 */
	public void setMinAmount(double amt) {
		if (amt > 0.0) {
			m_MinAmount = new Double(amt);
		}
	}

	/**
	 * set the maximum offer quantity
	 * @param amt the max amount
	 */
	public void setMaxAmount(double amt) {
		if (amt > 0.0) {
			m_MaxAmount = new Double(amt);
		}
	}

	/**
	 * generate the data which would be encrypted (and possibly signed)
	 */
	public void buildSigData() {
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(0);

		// build XML data based on the opcode
		StringBuilder sigData = new StringBuilder(1024);
		sigData.append("<opcode>" + m_Opcode + "</opcode>");
		if (m_Opcode.equalsIgnoreCase("REQ_discover_config")) {
			sigData.append("<walletId>" + m_WalletId + "</walletId>");
			sigData.append("<JID>" + m_JID + "</JID>");
			// NB: this opcode is not signed
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_voucher_purchase")) {
			sigData.append("<asset>" + m_Asset + "</asset>");
			sigData.append("<currency>" + m_Currency + "</currency>");
			sigData.append("<quantity>" + df.format(m_Quantity.doubleValue()));
			sigData.append("</quantity>");
			sigData.append("<walletId>" + m_WalletId + "</walletId>");
			sigData.append("<extWallet>" + m_ExtWallet + "</extWallet>");
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_voucher_sale")) {
			sigData.append("<asset>" + m_Asset + "</asset>");
			sigData.append("<currency>" + m_Currency + "</currency>");
			sigData.append("<quantity>" + df.format(m_Quantity.doubleValue()));
			sigData.append("</quantity>");
			sigData.append("<walletId>" + m_WalletId + "</walletId>");
			sigData.append("<extWallet>" + m_ExtWallet + "</extWallet>");
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_voucher_sold")) {
			sigData.append("<orderId>" + m_OrderId + "</orderId>");
			sigData.append("<walletId>" + m_WalletId + "</walletId>");
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_make_offer")) {
			sigData.append("<asset>" + m_Asset + "</asset>");
			sigData.append("<quantity>" + df.format(m_Quantity.doubleValue()));
			sigData.append("</quantity>");
			sigData.append("<wantedAssets>" + m_WantedAssets + "</wantedAssets>");
			sigData.append("<price>" + df.format(m_Price) + "</price>");
			sigData.append("<expiration>" + m_Expiration.getTime());
			sigData.append("</expiration>");
			sigData.append("<walletId>" + m_WalletId + "</walletId>");
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_fund_offer")) {
			sigData.append("<offerId>" + m_OfferId + "</offerId>");
			sigData.append("<walletId>" + m_WalletId + "</walletId>");
			sigData.append("<asset>" + m_Asset + "</asset>");
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_cancel_offer")
					|| m_Opcode.equalsIgnoreCase("REQ_offer_pickup"))
		{
			sigData.append("<offerId>" + m_OfferId + "</offerId>");
			sigData.append("<walletId>" + m_WalletId + "</walletId>");
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_offer_status")) {
			sigData.append("<walletId>" + m_WalletId + "</walletId>");
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_accept_offer")) {
			sigData.append("<offerId>" + m_OfferId + "</offerId>");
			sigData.append("<asset>" + m_Asset + "</asset>");
			sigData.append("<walletId>" + m_WalletId + "</walletId>");
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_get_offers")) {
			sigData.append("<offeredAssets>" + m_OfferedAssets);
			sigData.append("</offeredAssets>");
			sigData.append("<wantedAssets>" + m_WantedAssets + "</wantedAssets>");
			sigData.append("<price>" + df.format(m_Price) + "</price>");
			sigData.append("<minAmount>" + df.format(m_MinAmount));
			sigData.append("</minAmount>");
			sigData.append("<maxAmount>" + df.format(m_MaxAmount));
			sigData.append("</maxAmount>");
			// NB: this opcode is not signed
		}
		else {
			Log.error("Unsupported outgoing ClientBlock opcode, " + m_Opcode);
		}
		m_ClientBlockXML = sigData.toString();
	}

	/**
	 * apply the signature to the block, given a private key
	 * @param privKey the private key required to create the signature
	 * @return true if the signature was applied, false otherwise
	 */
	public boolean signXML(PrivateKey privKey) {
		if (m_ClientBlockXML.isEmpty()) {
			Log.error("No client block XML to sign");
			return false;
		}
		// these opcodes are unsigned
		if (m_Opcode.equalsIgnoreCase("REQ_get_offers")
			 || m_Opcode.equalsIgnoreCase("REQ_discover_config"))
		{
			return true;
		}

		Signature vSig = null;
		byte[] sig = null;
		try {
			vSig = Signature.getInstance("SHA1withRSA");
			vSig.initSign(privKey);
			vSig.update(m_ClientBlockXML.getBytes());
			sig = vSig.sign();
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("No such sig algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid pubkey for signature validation", ike);
		}
		catch (SignatureException se) {
			Log.error("Unable to validate client signature", se);
		}
		if (sig == null) {
			return false;
		}
		setSignature(Base64.encodeBytes(sig, Base64.DONT_BREAK_LINES));
		return true;
	}

	// END methods
}
