/*
 * last modified---
 * 	08-15-18/kbw/ explicitly cancel PacketCollectors after use; ensure any
 * 					  PacketFilter is always removed
 * 	04-02-15/jet/ restore SVX message filter in resetConnFilter()
 * 	08-06-14/jet/ Spark plugin version
 * 	06-11-13/jet/ new
 *
 * purpose---
 * 	handle interface to the OFS-SVG plugin which proxies the SVX
 *
 * notes---
 * 	see underlying messages in vsc_ofs-svx.xml and vsc_svx.xml
 */

package org.silentvault.spark;

import org.jivesoftware.smack.PacketListener;
import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.*;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.filter.OrFilter;
import org.jivesoftware.smack.filter.PacketFilter;
import org.jivesoftware.smack.util.Base64;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.ui.svx.*;

import java.util.ArrayList;
import java.util.Hashtable;
import java.security.PublicKey;
import java.net.HttpURLConnection;


/**
 * Handles messaging to/from the SVX, which is proxied by the SVG plugin at
 * the OFS.  The actual namespace is jabber:iq:silent-vault#vsc-ofs, but the
 * 'virtual namespace' wrapped by REQ_svx_message / REP_svx_message exchanges
 * is jabber:iq:silent-vault#vsc-svx.  This class acts as a listener for
 * requests sent from the SVX via the OFS, for replies to its own requests
 * made to the SVX, and for broadcast messages sent from the SVX.
 *
 * @author Justin Turrell
 */
public final class SVXListener implements PacketListener, Runnable {
	// data members
	/**
	 * reference to client application
	 */
	private WalletClient				m_Plugin;

	/**
	 * a map of currently outstanding listen filters, representing all requests
	 * for which we are awaiting replies, keyed by the packet IDs
	 */
	private Hashtable<String, PacketIDFilter>	m_FilterMap;

	// methods
	/**
	 * constructor
	 * @param plugin the reference to the app of which we are a part
	 */
	public SVXListener(WalletClient plugin) {
		m_Plugin = plugin;
		m_FilterMap = new Hashtable<String, PacketIDFilter>();
	}

	// implement the PacketListener interface
	public void processPacket(Packet packet) {
		XTabManager xtm = m_Plugin.getExchangeTabManager();
		VscState stData = m_Plugin.getVscState();

		// check for a Message notify
		if (packet instanceof Message) {
			/* Look for the extension we recognize.  Due to the filtering, *only*
			 * messages with this extension should be sent here.
			 */
			PacketExtension pExt
				= packet.getExtension("vsc_ofs",
											"jabber:message:silent-vault#vsc-ofs");
			if (pExt == null || !(pExt instanceof SVGSVXMessage)) {
				Log.error("Message from OFS without valid extension");
				return;
			}
			// the extension will be embodied in a SVGSVXMessage
			SVGSVXMessage broadCst = (SVGSVXMessage) pExt;
			if (!broadCst.getOpcode().equalsIgnoreCase("NFY_sys_broadcast")) {
				Log.error("Weird opcode on Message packet extension, "
							+ broadCst.getOpcode());
				return;
			}

			// based on the severity level, display an appropriate popup
			String level = broadCst.getLevel();
			if (level.equals("info")) {
				xtm.showInfo(broadCst.getText());
			}
			else if (level.equals("warning")) {
				xtm.showWarning(broadCst.getText());
			}
			else {
				xtm.showError(broadCst.getText());

				// if this is a shutdown, disconnect forcibly from the SVX
				if (broadCst.getText().endsWith("shutting down NOW")) {
					xtm.getMenuPane().pushDisconnect();
				}
			}
		}
		// check for IQ reply
		else if (packet instanceof IQ) {
			if (!(packet instanceof SVGSVXIQ)) {
				Log.error("Unexpected IQ packet, " + packet.toXML());
				removeFilter(packet.getPacketID());
				return;
			}
			SVGSVXIQ svxMsg = (SVGSVXIQ) packet;
			stData.setLastActivity(svxMsg);

			// get the SVX's reply blob
			String repBlob = svxMsg.getRepBlob();
			if (repBlob.isEmpty() && svxMsg.getType() != IQ.Type.ERROR) {
				Log.debug("Empty SVX reply blob, cannot process");
				removeFilter(packet.getPacketID());
				return;
			}
			SVXBlock svxReply = null;
			String reply = null;
			String opcode = null;

			// attempt decrypt
			VsSecrets secrets = m_Plugin.getLoginSecrets();
			String decBlob = EncodingUtils.getStrFromBase64PubkeyEnc(repBlob,
																			secrets.getPrivKey());
			if (decBlob == null) {
				/* In these cases we will not have any encryption:
				 * 1. An error which occurred before the SVX had our key available.
				 * 2. A REP_offers response (these are sent in clear).
				 */
				if (svxMsg.getType() == IQ.Type.ERROR) {
					// try to parse as error in clear
					svxReply = new SVXBlock(true);
					reply = "<svxBlock>" + repBlob + "</svxBlock>";
					if (!svxReply.parseBlock(reply)) {
						//Log.error("Unable to parse SVX error reply block");
						// show error to user
						xtm.showError(svxMsg.getErrMsg(),
											"code " + svxMsg.getErrCode());
						removeFilter(packet.getPacketID());
						return;
					}
					opcode = svxReply.getOpcode();
				}
				else {
					// attempt to parse as a regular reply, Base64-encoded
					svxReply = new SVXBlock(false);
					reply = "<svxBlock>" + new String(Base64.decode(repBlob))
								+ "</svxBlock>";
					if (!svxReply.parseBlock(reply)) {
						Log.error("Unable to parse clear SVX reply");
						xtm.showError("SVX clear reply message did not parse");
						removeFilter(packet.getPacketID());
						return;
					}
					opcode = svxReply.getOpcode();
					// only this one opcode is legal without encryption
					if (!opcode.equalsIgnoreCase("REP_offers")) {
						Log.error("SVX reply opcode " + opcode + " not encrypted!");
						xtm.showError("SVX reply message wasn't encrypted, ignored");
						removeFilter(packet.getPacketID());
						return;
					}
				}
			}
			else {
				// parse what the SVX sent in its block -- might be an error
				svxReply = new SVXBlock(false);
				reply = "<svxBlock>" + decBlob + "</svxBlock>";
				if (!svxReply.parseBlock(reply)) {
					Log.error("Unable to parse SVX reply block");
					xtm.showError("SVX encrypted reply message did not parse");
					removeFilter(packet.getPacketID());
					return;
				}
				opcode = svxReply.getOpcode();
			}

			if (!svxReply.isError()) {
				// validate SVX's signature
				PublicKey svxKey = secrets.getSVXKey();
				if (svxKey == null) {
					Log.error("No SVX pubkey available for sig check");
					removeFilter(packet.getPacketID());
					return;
				}
				if (!svxReply.sigVerify(svxKey)) {
					Log.error("SVX signature check failed on " + opcode);
					removeFilter(packet.getPacketID());
					return;
				}
			}
			//Log.debug("Valid SVX reply seen, opcode " + opcode);

			// evaluate opcode and pass to appropriate ui handler
			if (opcode.equalsIgnoreCase("REP_config")) {
				// pass to ConnectPane
				xtm.getConnectPane().processConfigReply(svxReply);
				// if not an error, move to home init state
				if (!svxReply.isError()) {
					xtm.queueTransition(new Integer(XTabManager.M_HOME_INIT));
					new Thread(this).run();
				}
			}
			else if (opcode.equalsIgnoreCase("REP_purchase_terms")) {
				// pass to BuyPane
				xtm.getBuyPane().processPurchaseTerms(svxReply);
			}
			else if (opcode.equalsIgnoreCase("REP_sale_terms")) {
				// pass to SellPane
				xtm.getSellPane().processSaleTerms(svxReply);
			}
			else if (opcode.equalsIgnoreCase("REP_voucher_sold")) {
				// pass to SellPane
				xtm.getSellPane().processFundingReply(svxReply);
			}
			else if (opcode.equalsIgnoreCase("REP_offer_entered")) {
				// pass to TradePane
				xtm.getTradePane().processOfferReply(svxReply);
			}
			else if (opcode.equalsIgnoreCase("REP_offer_funded")) {
				/* pass to one of TradePane, OffersPane, or AcceptPane, depending
				 * upon context
				 */
				int xTabState = xtm.getCurrentState();
				if (xTabState == XTabManager.M_GO_TRADE) {
					xtm.getTradePane().processFundingReply(svxReply);
				}
				else if (xTabState == XTabManager.M_GO_OFFERS) {
					xtm.getOffersPane().processFundingReply(svxReply);
				}
				else if (xTabState == XTabManager.M_GO_ACCEPT) {
					xtm.getAcceptPane().processFundingReply(svxReply);
				}
				else {
					Log.error("Unexpected XTab state on " + opcode + ": "
								+ xTabState);
				}
			}
			else if (opcode.equalsIgnoreCase("REP_offers")) {
				xtm.getSearchPane().processOffersReply(svxReply);
			}
			else if (opcode.equalsIgnoreCase("REP_accept_entered")) {
				// pass to AcceptPane
				xtm.getAcceptPane().processAcceptReply(svxReply);
			}
			else if (opcode.equalsIgnoreCase("REP_offer_status")) {
				// pass to OffersPane
				xtm.getOffersPane().processOffersReply(svxReply);
			}
			else if (opcode.equalsIgnoreCase("REP_offer_pickup")) {
				// pass to OffersPane
				xtm.getOffersPane().processPickupReply(svxReply);
			}
			else if (opcode.equalsIgnoreCase("REP_offer_canceled")) {
				// pass to OffersPane
				xtm.getOffersPane().processCancelReply(svxReply);
			}
			else {
				Log.error("Unrecognized SVX reply opcode, " + opcode);
			}
		}
		else {
			// must be Presence, not supported in this namespace
			Log.error("Unexpected (Presence?) packet: " + packet.toXML());
		}

		// retire this ID from the list of active packet filters
		removeFilter(packet.getPacketID());
	}

	/**
	 * prepare and send a request message to the SVX (via OFS), given a client
	 * block.  NB: this method blocks on the reply from the OFS/SVX.
	 * @param cBlock the client block to send
	 * @return true if message sent successfully
	 */
	public boolean sendRequest(SVXClientBlock cBlock) {
		if (cBlock == null || !cBlock.getOpcode().startsWith("REQ_")) {
			Log.error("Missing or improper SVX client block");
			return false;
		}

		// build output XML and sign it (if required)
		cBlock.buildSigData();
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		if (!cBlock.signXML(secrets.getPrivKey())) {
			Log.error("Unable to sign output XML");
			return false;
		}

		// encrypt block, including any signature, with SVX's pubkey
		StringBuilder clearBlock = new StringBuilder(5120);
		clearBlock.append("<clientBlock>");
		clearBlock.append(cBlock.getClientBlockXML());
		if (!cBlock.getSignature().isEmpty()) {
			clearBlock.append("<signature>" + cBlock.getSignature());
			clearBlock.append("</signature>");
		}
		clearBlock.append("</clientBlock>");
		String encBlock
			= EncodingUtils.makeBase64PubkeyEncStr(clearBlock.toString(),
																secrets.getSVXKey());
		if (encBlock == null || encBlock.isEmpty()) {
			Log.error("Error encrypting client block for SVX");
			return false;
		}

		// prepare wrapper message to OFS
		XMPPConnection conn = m_Plugin.getOFSConnection();
		VscState stData = m_Plugin.getVscState();
		String opcode = cBlock.getOpcode();
		SVGSVXIQ svxReq = new SVGSVXIQ();
		if (opcode.equals("REQ_discover_config")
			 || opcode.equals("REQ_get_offers")
			 || opcode.equals("REQ_offer_status"))
		{
			svxReq.setType(IQ.Type.GET);
		}
		else {
			svxReq.setType(IQ.Type.SET);
		}
		svxReq.setFrom(conn.getUser());
		svxReq.setTo(conn.getServiceName());
		String pId = "svx_" + stData.getNextOFSid();
		svxReq.setPacketID(pId);
		svxReq.setOpcode("REQ_svx_message");
		svxReq.setSVXId(secrets.getSVXId());
		svxReq.setMsgBlob(encBlock);

		// construct a packet collector to listen for reply
		PacketIDFilter pFilter = new PacketIDFilter(pId);
		PacketCollector repColl = conn.createPacketCollector(pFilter);

		// send it to the server
		//Log.debug("Sending " + opcode + " to SVG, " + svxReq.toXML());
		conn.sendPacket(svxReq);
		stData.setLastActivity(svxReq);

		// wait for reply
		SVGSVXIQ rep = null;
		IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
		if (reply == null) {
			// got timeout; build error
			rep = new SVGSVXIQ();
			rep.setOpcode("REP_svx_message");
			rep.setFrom(svxReq.getTo());
			rep.setTo(svxReq.getFrom());
			rep.setPacketID(pId);
			rep.setType(IQ.Type.ERROR);
			SVXBlock fakeRep = new SVXBlock(true);
			fakeRep.setErrCode(HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
			fakeRep.setErrMsg("timed out");
			fakeRep.setOpcode("REP_timeout");
			rep.setRepBlob(fakeRep.getXML());
		}
		else {
			rep = (SVGSVXIQ) reply;
		}
		repColl.cancel();
		this.processPacket(rep);
		return true;
	}

	/**
	 * Method to implement Runnable, so that we can invoke UI elements in an
	 * asynchronous manner, without risking the termination or hijacking of
	 * this thread by Swing.  Also, we need this callback to return quickly so
	 * that further messages may be processed.
	 */
	public void run() {
		// obtain UI tab manager from plugin and do the queued state transition
		m_Plugin.getExchangeTabManager().makeTransition();
	}

	/**
	 * add a filter to the list of packets for which we're accepting replies.
	 * NB: asynchronous use only
	 * @param ID the packet ID we're expecting in a reply
	 * @param filter the filter to add, which watches for the ID
	 */
	public void addFilter(String ID, PacketIDFilter filter) {
		if (ID == null || ID.isEmpty() || filter == null
			 || m_FilterMap.containsKey(ID))
		{
			return;
		}
		m_FilterMap.put(ID, filter);
		resetConnFilter();
	}

	/**
	 * remove a filter from the list of replies for which we're listening
	 * @param ID the ID of the packet filter to remove
	 */
	public void removeFilter(String ID) {
		if (ID == null || ID.isEmpty() || !m_FilterMap.containsKey(ID)) {
			// nothing to do
			return;
		}
		m_FilterMap.remove(ID);
		resetConnFilter();
	}

	/**
	 * utility method to rebuild and reset the filter on this connection, by
	 * ORing together all filters extant in the map.  (async requests only)
	 */
	private synchronized void resetConnFilter() {
		// rebuild filter of all outstanding requests
		OrFilter of = new OrFilter();
		for (String packetId : m_FilterMap.keySet()) {
			PacketIDFilter filt = m_FilterMap.get(packetId);
			of.addFilter(filt);
		}
		XMPPConnection conn = m_Plugin.getOFSConnection();
		conn.removePacketListener(this);
		// add back the generic Message filter if we have one
		PacketFilter msgFilter = m_Plugin.getSVXMsgFilter();
		if (msgFilter != null) {
			conn.addPacketListener(this, msgFilter);
		}

		// add new listener if map is non-empty
		if (!m_FilterMap.isEmpty()) {
			conn.addPacketListener(this, of);
		}
	}
}
