/*
 * last modified---
 * 	08-30-18/kbw/ uncommit purged tokens in TokenTableModel
 * 	08-15-18/kbw/ explicitly cancel PacketCollectors after use
 * 	08-06-14/jet/ Spark plugin version
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	class to manage the list of tokens spent to a DHT or SDS node, but not
 * 	yet reported spent to the VP
 */

package org.silentvault.spark;

import java.awt.Cursor;

import org.dom4j.Element;
import org.dom4j.DocumentException;

import org.jivesoftware.smack.util.StringUtils;
import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.packet.Packet;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.filter.PacketIDFilter;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.ui.wallet.WTabManager;
import org.silentvault.spark.ui.wallet.HomePane;
import org.silentvault.spark.ui.TokenTableModel;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.concurrent.ConcurrentHashMap;
import java.io.*;
import java.security.*;


/**
 * Spent token manager class for wallet clients.
 * This class implements a cache which keeps track of tokens which have 
 * been spent by a client, in order to prevent double-spending of tokens.
 * It also facilitates reporting those tokens as spent, by initiating a
 * REQ_delete_spent_tokens / REP_tokens_purged exchange with the OFS/VP.
 * (This is done whenever tokens are bought, or on VS logout.)
 *
 * @author Kevin Wilkerson
 */
public final class SpentTokenCache {
	// BEGIN data members
	/**
	 * the owning plugin
	 */
	private WalletClient				m_Plugin;

	/**
	 * the list of tokens we've presented, but have not yet reported spent,
	 * backed by a disc file.  Tokens are mapped to the VS# they come from.
	 */
	private ConcurrentHashMap<XMLToken, String>	m_SpentTokens;

	/**
	 * the path to the disc file backing m_SpentTokens. (user install directory
	 * + spentTokens.xml)
	 */
	private String						m_SpentTokensPath;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the plugin which owns us
	 */
	public SpentTokenCache(WalletClient plugin) {
		m_Plugin = plugin;
		m_SpentTokensPath = m_Plugin.getSparkUserHome() + File.separator
								+ "spentTokens.xml";
		m_SpentTokens = new ConcurrentHashMap<XMLToken, String>();
	}

	/**
	 * initialization method; constructs spent tokens list from data in disc file
	 * @return true on success, false otherwise
	 */
	public boolean initSpentTokens () {
		if (m_SpentTokensPath == null || m_SpentTokensPath.isEmpty()) {
			Log.error("Spent tokens cache file not set");
			return false;
		}

		// if file does not exist, create an empty one
		File stFile = new File(m_SpentTokensPath);
		if (!stFile.exists()) {
			try {
				if (!stFile.createNewFile()) {
					Log.error("Could not create empty spent tokens file " + stFile);
				}
			}
			catch (IOException ioe) {
				Log.error("Error creating empty spent tokens file", ioe);
			}
		}

		// open reader on file
		FileReader fr = null;
		try {
			fr = new FileReader(m_SpentTokensPath);
		}
		catch (FileNotFoundException fnfe) {
			Log.error("Could not open spent tokens file", fnfe);
			return false;
		}
		BufferedReader br = new BufferedReader(fr);

		String line = null;
		XMLToken token;
		boolean ok = true;
		do {
			// each line should be one token, in format: VS#::<token>...</token>
			try {
				line = br.readLine();
				// allow for empty file
				if (line == null || line.isEmpty()) {
					break;
				}
				String tokLine[] = line.split("::");
				if (tokLine.length != 2) {
					throw new DocumentException("Spent token corrupt line: " + line);
				}
				token = XMLToken.getTokenRep(tokLine[1], true);
				m_SpentTokens.put(token, tokLine[0]);
			}
			catch (IOException ioe) {
				Log.error("Spent token read error", ioe);
				ok = false;
			}
			catch (DocumentException de) {
				Log.error("Spent token parse problem", de);
				ok = false;
			}
		} while (line != null);
		try {
			fr.close();
		}
		catch (IOException ioe) { /* ignore */ }
	/*
		Log.debug("SpentTokenCache initialized with " + m_SpentTokens.size()
					+ " tokens in it");
	 */
		return ok;
	}

	/**
	 * sync the tokens spent before list to the backing file
	 * @return true on success, false otherwise
	 */
	public synchronized boolean syncSpentTokens() {
		// open writer on file
		PrintWriter pw = null;
		try {
			pw = new PrintWriter(m_SpentTokensPath);
		}
		catch (FileNotFoundException fnfe) {
			Log.error("Could not open spent tokens file for write", fnfe);
			return false;
		}

		// output the XML for each token in our list
		for (XMLToken tok : m_SpentTokens.keySet()) {
			pw.println(m_SpentTokens.get(tok) + "::" + tok.toXML());
		}
		pw.flush();
		pw.close();
		return true;
	}

	/**
	 * private helper method to determine whether a token has been spent before
	 * @param token the token to check
	 * @return true if token is on "spent" list, false otherwise
	 */
	public boolean spentBefore(XMLToken token) {
		if (token == null) {
			return false;
		}
		return m_SpentTokens.containsKey(token);
	}

	/**
	 * method to add a token to the list of tokens we've spent once already
	 * (normally called right after spentBefore() above)
	 * @param token the token to add to the list
	 * @param vsNum the VS# to which this token belongs
	 */
	public void recordToken(XMLToken token, String vsNum) {
		if (token != null) {
			m_SpentTokens.put(token, vsNum);
			syncSpentTokens();
		}

		// if we now have 10 or more tokens, trigger a purge
		ArrayList<XMLToken> ourSpentToks = getSpentTokens(vsNum);
		if (ourSpentToks.size() >= 10) {
			if (!purgeSpentTokens()) {
				Log.error("Could not auto-purge spent tokens from " + vsNum);
			}
		}
	}

	/**
	 * tell the OFS/VP about our spent tokens (done synchronously)
	 * @return true on success, false otherwise
	 */
	public boolean purgeSpentTokens() {
		VscState vsState = m_Plugin.getVscState();
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		String vsNum = secrets.getVSnumber();
		if (m_SpentTokens.isEmpty() || !m_SpentTokens.containsValue(vsNum)) {
			// nothing to do, just sync
			syncSpentTokens();
			return true;
		}

		final XMPPConnection conn = m_Plugin.getOFSConnection();
		if (conn == null || !conn.isConnected()) {
			// can't do this
			return false;
		}

		// construct <purgeDetails/>
		StringBuilder pDets = new StringBuilder(2048);
		pDets.append("<purgeDetails>");
		pDets.append("<vs_number>" + vsNum + "</vs_number>");
		pDets.append("<sdshash>" + new String(secrets.getVoucherIndex()));
		pDets.append("</sdshash>");
		pDets.append("<cap>" + secrets.getVouchRWCap() + "</cap>");
		pDets.append("<spentTokens>");
		int purgeTokCnt = 0;
		for (XMLToken tok : m_SpentTokens.keySet()) {
			if (m_SpentTokens.get(tok).equals(vsNum)) {
				pDets.append(tok.toXML());
				purgeTokCnt++;
			}
		}
		pDets.append("</spentTokens>");

		// add our private key signature
		byte[] sig = null;
		boolean ok = false;
		Signature pSig = null;
		try {
			pSig = Signature.getInstance("SHA1withRSA");
			pSig.initSign(secrets.getPrivKey());
			pSig.update(pDets.toString().getBytes());
			sig = pSig.sign();
			ok = true;
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("No such signature algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid privkey for token purge signature", ike);
		}
		catch (SignatureException se) {
			Log.error("Unable to create token purge signature", se);
		}
		if (!ok) {
			Log.error("Could not place signature on token purge <purgeDetails/>");
			return false;
		}
		pDets.append("<signature>" + StringUtils.encodeBase64(sig, false));
		pDets.append("</signature>");
		pDets.append("</purgeDetails>");

		// encrypt the works to VP's key
		String encDets = EncodingUtils.makeBase64PubkeyEncStr(pDets.toString(),
																				secrets.getVPKey());
		if (encDets == null) {
			return false;
		}

		// build message to send
		OFSMessage purgeToks = new OFSMessage();
		purgeToks.setType(IQ.Type.SET);
		purgeToks.setFrom(conn.getUser());
		purgeToks.setTo(conn.getServiceName());
		purgeToks.setPacketID("purge_" + vsState.getNextOFSid());
		purgeToks.setOpcode("REQ_delete_spent_tokens");
		purgeToks.setVoucher_publisher(secrets.getPublisher());
		purgeToks.setPurgeDetails(encDets);

		// create packet filter to back collector, and install collector
		PacketIDFilter pFilt = new PacketIDFilter(purgeToks.getPacketID());
		PacketCollector repColl = conn.createPacketCollector(pFilt);

		// send the request to the OFS with a blocking read
		WTabManager tabm = m_Plugin.getWalletTabManager();
		tabm.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		conn.sendPacket(purgeToks);
	/*
		Log.debug("Sent REQ_delete_spent_tokens to OFS purging "
					+ purgeTokCnt + " used tokens");
	 */
		vsState.setLastActivity(purgeToks);
		Packet pRep = repColl.nextResult(m_Plugin.getTimeout());
		repColl.cancel();
		if (pRep == null) {
			// timed out
			Log.error("Timeout purging spent tokens");
			tabm.setCursor(null);
			tabm.showError("cannot purge spent tokens", "timed out");
			return false;
		}

		OFSMessage purgeRep = (OFSMessage) pRep;
		vsState.setLastActivity(purgeRep);

		// check for error
		if (purgeRep.getType() == IQ.Type.ERROR) {
			Log.error("Error purging spent tokens, code " + purgeRep.getErrcode()
						+ ": " + purgeRep.getErrmsg());
			tabm.setCursor(null);
			tabm.showError("warning, cannot purge spent tokens",
								purgeRep.getErrcode() + ": " + purgeRep.getErrmsg(),
								"will retry automatically at next logout");
			return false;
		}

		/* Parse the reply.  We should have something which looks like this when
		 * decrypted:
		 * <purged>
		 * 	<token>echoed token</token>
		 * 	...
		 * 	<signature>sig of VP on token list</signature>
		 * </purged>
		 */
		String encRep = purgeRep.getPurged();
		String decRep = EncodingUtils.getStrFromBase64PubkeyEnc(encRep,
																			secrets.getPrivKey());
		if (decRep == null || decRep.isEmpty()) {
			Log.error("Token purge reply would not decrypt");
			return false;
		}
		Element purgedElt = EncodingUtils.buildElement(decRep);
		if (purgedElt == null) {
			Log.error("Token purge reply was null after decrypt");
			return false;
		}
		// iterate through sub-elements
		ArrayList<XMLToken> confirmedPurged
			= new ArrayList<XMLToken>(m_SpentTokens.size());
		StringBuilder signedDets = new StringBuilder(10240);
		Iterator eltIt = purgedElt.elementIterator();
		while (eltIt.hasNext()) {
			Element thisElt = (Element) eltIt.next();
			String eltName = thisElt.getName().toLowerCase();
			XMLToken token = null;
			if (eltName.equals("token")) {
				// parse token
				try {
					token = XMLToken.getTokenRep(thisElt, false);
					confirmedPurged.add(token);
					signedDets.append(token.toXML());
				}
				catch (DocumentException de) {
					Log.error("Unparseable token on purge confirm list: "
								+ thisElt.getText(), de);
				}
			}
			else if (eltName.equals("signature")) {
				// check signature of VP on tokens we've seen listed
				Signature vSig = null;
				ok = false;
				try {
					vSig = Signature.getInstance("SHA1withRSA");
					vSig.initVerify(secrets.getVPKey());
					vSig.update(signedDets.toString().getBytes());
					ok
					= vSig.verify(StringUtils.decodeBase64(thisElt.getTextTrim()));
				}
				catch (NoSuchAlgorithmException nsae) {
					// impossible
					Log.error("No such sig verify algorithm", nsae);
				}
				catch (InvalidKeyException ike) {
					Log.error("Invalid key for token purge confirm sig check", ike);
				}
				catch (SignatureException se) {
					Log.error("Unable to validate token purge confirmation sig", se);
				}
				if (!ok) {
					Log.error("Bad signature on token purge confirmation");
					return false;
				}
				else {
					/* Knock the confirmed tokens off the list.  Should be all, but
					 * a partial completion is possible.
					 */
					for (XMLToken tok : confirmedPurged) {
						m_SpentTokens.remove(tok);
					}

					/* Also remove all of these tokens, including any marked as
					 * "committed" in the TokenTableModel, because whenever we do a
					 * queryTokens() it causes any tokens which were in m_SpentTokens
					 * to get marked as committed, and we need to undo that.
					 */
					HomePane hp = tabm.getHomePane();
					TokenTableModel ttm = hp.getTokenModel();
					ttm.spendTokens(confirmedPurged);
				}
			}
			else {
				Log.error("spurious element in token purge confirm: " + eltName);
			}
		}

		// clear cursor and sync to disc
		tabm.setCursor(null);
		syncSpentTokens();
		return true;
	}

	/**
	 * method to remove tokens from our list of spent tokens
	 * NB: usually called when auditing this list against actual folder contents
	 * @param delTokens the tokens to remove
	 * @return true if the list changed as a result
	 */
	public boolean removeTokens(ArrayList<XMLToken> delTokens) {
		if (delTokens == null || delTokens.isEmpty()) {
			return false;
		}
		boolean changed = false;
		for (XMLToken tok : delTokens) {
			if (m_SpentTokens.remove(tok) != null) {
				changed = true;
			}
		}
		return changed;
	}

	/**
	 * access method for our list of spent tokens
	 * @param vsNum the VS number for which we want the tokens
	 * @return the list
	 */
	public ArrayList<XMLToken> getSpentTokens(String vsNum) {
		ArrayList<XMLToken> retList
			= new ArrayList<XMLToken>(m_SpentTokens.size());
		for (XMLToken tok : m_SpentTokens.keySet()) {
			if (m_SpentTokens.get(tok).equals(vsNum)) {
				retList.add(tok);
			}
		}
		return retList;
	}


	/**
	 * finalize object when garbage-collected, for data security
	 */
	@Override
	protected void finalize() throws Throwable {
		try {
			// nix sensitive data if there is any
			if (m_SpentTokens != null && !m_SpentTokens.isEmpty()) {
				m_SpentTokens.clear();
			}
		} finally {
			super.finalize();
		}
	}
	// END methods
}
