/*
 * last modified---
 * 	02-10-18/kbw/ fix setExQuantity() to actually set value
 * 	08-06-14/jet/ Spark plugin version
 * 	07-18-13/jet/ add m_EscrowWallet
 * 	07-09-13/jet/ new
 *
 * purpose---
 * 	provide a class representation of a trade offer record on the SVX
 */

package org.silentvault.spark;

import org.jivesoftware.spark.util.log.Log;

import java.util.Date;


/**
 * This class supplies an instantiation of a trade offer fetched off the SVX.
 * It has methods to get and set all fields, along with some minimal sanity
 * checking logic.
 *
 * @author Justin Turrell
 */
public final class TradeOffer {
	// BEGIN data members

	// offer status settings constants
	/**
	 * status: offer has been placed by originator but not yet escrow funded
	 */
	public static final String			M_StatusOpen = "open";
	/**
	 * status: offer has been escrow funded by originator
	 */
	public static final String			M_StatusFunded = "funded";
	/**
	 * status: offer has been accepted and filled by an acceptor
	 */
	public static final String			M_StatusFilled = "filled";
	/**
	 * status: offer has expired, and escrow refunded to originator
	 */
	public static final String			M_StatusExpired = "expired";
	/**
	 * status: offer was canceled by originator, and escrow refunded
	 */
	public static final String			M_StatusCanceled = "canceled";

	/**
	 * offer ID sequence number (unique index)
	 */
	private int								m_OfferId;

	/**
	 * individual voucher asset type posted to escrow by originator
	 */
	private String							m_Offered;

	/**
	 * quantity of asset offered by originator
	 */
	private double							m_Quantity;

	/**
	 * colon-separated list of exchange assets acceptable to originator
	 */
	private String							m_Wanted;

	/**
	 * offer price requirement, the percentage ratio at which OFFERED will
	 * be turned into WANTED
	 */
	private float							m_Price;

	/**
	 * offer status (see M_Status* constants above).  Note that in the SVX DB
	 * there are also expired, closed, and canceled offers, which are never
	 * retrieved or shown on the client side.
	 */
	private String							m_Status;

	/**
	 * the wallet on the SVX designated for use as the escrow wallet
	 */
	private String							m_EscrowWallet;

	/**
	 * for open or funded offers, expiration date/time of offer
	 */
	private Date							m_Expiration;

	/**
	 * the asset type provided to escrow by the acceptor (will be in WANTED list)
	 */
	private String							m_ExAsset;

	/**
	 * quantity of m_ExAsset actually provided by acceptor
	 */
	private double							m_ExQuantity;

	/**
	 * date/time at which offer was filled, and escrow voucher payouts made
	 */
	private Date							m_ExTimestamp;

	// END data members

	// BEGIN methods
	/**
	 * nullary constructor
	 */
	public TradeOffer() {
		m_Offered = m_Wanted = m_ExAsset = m_EscrowWallet = "";
		m_Status = M_StatusOpen;
	}

	// get methods
	/**
	 * obtain the offer ID sequence number (unique index)
	 * @return the offer ID
	 */
	public int getOfferId() { return m_OfferId; }

	/**
	 * obtain the voucher asset type offered by originator
	 * @return the asset type
	 */
	public String getOffered() { return m_Offered; }

	/**
	 * obtain the quantity of asset offered
	 * @return the quantity entered by the originator
	 */
	public double getQuantity() { return m_Quantity; }

	/**
	 * obtain the list of asset types wanted by originator in exchange
	 * @return the list, colon-separated
	 */
	public String getWanted() { return m_Wanted; }

	/**
	 * obtain the offer price specified by the originator (== spot for market)
	 * @return the offer price, as a ratio of OFFERED:WANTED
	 */
	public float getPrice() { return m_Price; }

	/**
	 * obtain the escrow wallet
	 * @return the Id of the wallet on the SVX designated for holding escrow
	 */
	public String getEscrowWallet() { return m_EscrowWallet; }

	/**
	 * obtain the offer status
	 * @return the status (one of the M_Status* constants defined above)
	 */
	public String getStatus() { return m_Status; }

	/**
	 * obtain the expiration date for an offer
	 * @return the date/time at which the offer expires if not filled
	 */
	public Date getExpiration() { return m_Expiration; }

	/**
	 * obtain the actual single exchange asset type provided by the acceptor
	 * @return the asset (will be an element of WANTED)
	 */
	public String getExAsset() { return m_ExAsset; }

	/**
	 * obtain the quantity of exchange asset actually provided by the acceptor
	 * @return the quantity paid to escrow
	 */
	public double getExQuantity() { return m_ExQuantity; }

	/**
	 * obtain the date at which the offer was filled or refunded
	 * @return the date/time, or null if it is still open
	 */
	public Date getExTimestamp() { return m_ExTimestamp; }


	// set methods
	/**
	 * set the offer ID
	 * @param oId the offer sequence number to set
	 * @return true if the value was set
	 */
	public boolean setOfferId(int oId) {
		if (oId > 0) {
			// immutable; only set once
			if (m_OfferId == 0) {
				m_OfferId = oId;
				return true;
			}
		}
		return false;
	}

	/**
	 * set the offered asset type
	 * @param asset the voucher asset type offered by the originator
	 * @return true if the value was set
	 */
	public boolean setOffered(String asset) {
		if (asset == null || asset.isEmpty()) {
			return false;
		}
		// immutable; set only once
		if (m_Offered.isEmpty()) {
			m_Offered = new String(asset);
			return true;
		}
		return false;
	}

	/**
	 * set the offer quantity
	 * @param quan the amount of the asset being offered by originator
	 * @return true if the value was set
	 */
	public boolean setQuantity(double quan) {
		if (quan > 0.0) {
			// immutable; only set once
			if (m_Quantity == 0.0) {
				m_Quantity = quan;
				return true;
			}
		}
		return false;
	}

	/**
	 * set the desired asset type
	 * @param asset the voucher asset type wanted in exchange by the originator
	 * @return true if the value was set
	 */
	public boolean setWanted(String asset) {
		if (asset == null || asset.isEmpty()) {
			return false;
		}
		// immutable; set only once
		if (m_Wanted.isEmpty()) {
			m_Wanted = new String(asset);
			return true;
		}
		return false;
	}

	/**
	 * set the offer price, the ratio of OFFERED:WANTED.  NB: if this is not 1.0
	 * then WANTED must be restricted to a single asset type
	 * @param price the ratio to set
	 * @return true if the value was set
	 */
	public boolean setPrice(float price) {
		if (price > 0.0F) {
			if (price != 1.0F) {
				if (m_Wanted.indexOf(':') != -1) {
					Log.error("Cannot set unequal ratio, " + price + ", of "
								+ m_Offered + " to multiple wanted assets, "
								+ m_Wanted);
					return false;
				}
			}
			m_Price = price;
			return true;
		}
		else {
			Log.error("Illegal price ratio, " + price);
		}
		return false;
	}

	/**
	 * set the status for the offer
	 * @param stat the status to set (must be a M_Status* value above)
	 * @return true if the value was set
	 */
	public boolean setStatus(String stat) {
		if (stat == null) {
			return false;
		}
		if (stat.equals(M_StatusOpen) || stat.equals(M_StatusFunded)
			 || stat.equals(M_StatusFilled) || stat.equals(M_StatusCanceled)
			 || stat.equals(M_StatusExpired))
		{
			m_Status = new String(stat);
			return true;
		}
		return false;
	}

	/**
	 * set the escrow wallet for the offer
	 * @param walletId the Id of the SVX wallet designated as escrow for offer
	 */
	public void setEscrowWallet(String walletId) {
		if (walletId != null && !walletId.isEmpty()) {
			m_EscrowWallet = new String(walletId);
		}
	}

	/**
	 * set the expiration date for a limit offer
	 * @param expiry the date/time at which the offer is to expire if not filled
	 * @return true if the value was set
	 */
	public boolean setExpiration(Date expiry) {
		if (expiry != null) {
			m_Expiration = expiry;
			return true;
		}
		return false;
	}

	/**
	 * set the asset type actually paid out to originator from escrow
	 * @param asset the voucher asset type (must be in WANTED, or OFFERED)
	 * @return true if the value was set
	 */
	public boolean setExAsset(String asset) {
		if (asset == null || asset.isEmpty()) {
			return false;
		}
		if (m_Wanted.contains(asset) || asset.equals(m_Offered)) {
			m_ExAsset = new String(asset);
		}
		else {
			Log.error("Exchange asset " + asset + " is not of acceptable type");
		}
		return false;
	}

	/**
	 * set the quantity of exchange asset actually paid out of escrow
	 * @param quant the quantity of asset acceptor placed into escrow, or
	 * amount of escrow refund
	 * @return true if the value was set
	 */
	public boolean setExQuantity(double quant) {
		if (quant > 0.0) {
			// immutable; set only once
			if (m_ExQuantity == 0.0) {
				m_ExQuantity = quant;
				return true;
			}
		}
		return false;
	}

	/**
	 * set the time at which escrow was paid out (filled or refunded)
	 * @param fillTime the date/time at which the offer was settled
	 * @return true if the value was set
	 */
	public boolean setExTimestamp(Date fillTime) {
		if (fillTime != null) {
			// immutable; set only once
			if (m_ExTimestamp == null) {
				m_ExTimestamp = fillTime;
				return true;
			}
		}
		return false;
	}

	/**
	 * override Object::equals() so that ArrayList.contains() will work
	 * @param obj the object to compare against
	 * @return true if this object is the same, false if not
	 */
	@Override
	public boolean equals(Object obj) {
		// if it's not an offer, cannot be equal
		if (!(obj instanceof TradeOffer)) {
			return false;
		}
		TradeOffer offer = (TradeOffer) obj;

		// an offer is equal to this one if its ID, amount, and offered matches
		if (m_OfferId == offer.m_OfferId
			 && m_Quantity == offer.m_Quantity
			 && m_Offered.equals(offer.m_Offered))
		{
			return true;
		}
		return false;
	}

	// END methods
}
