/*
 * last modified---
 * 	08-06-14/jet/ Spark plugin version
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	class to hold VSC state-tracking information
 */

package org.silentvault.spark;

import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smackx.packet.DiscoverItems;

import org.jivesoftware.spark.util.log.Log;

import java.util.Date;


/**
 * State-keeping engine for the VSC.  Keeps track of messages exchanged with
 * other voucher-safe.org network entities, last client activity, and the like.
 * Data in this object is only relevant when the user is logged in to a valid
 * voucher safe.
 *
 * @author Kevin Wilkerson
 */
public final class VscState {
	// BEGIN data members
	
	/**
	 * Default inactivity timeout in seconds before safe is considered logged
	 * out, in seconds.  Default is ten (10) minutes, which can be overriden by
	 * user's plugin config value.
	 */
	public static final int		M_INACTIVITY_TIMEOUT = 600;

	/**
	 * whether the client is logged in to a VS or not
	 */
	private boolean		m_LoggedIn;

	/**
	 * date/time of last activity that is voucher-related
	 */
	private Date			m_LastActivity;

	/**
	 * last message sent to OFS (IQ because it's really one of three things:
	 * OFSMessage, SVXMessage, or SSMMessage)
	 */
	private IQ				m_OFSlastSent;

	/**
	 * last message received from OFS (IQ because it's really one of three
	 * things: OFSMessage, SVXMessage, or SSMMessage)
	 */
	private IQ				m_OFSlastRecv;

	/**
	 * last message sent to public key server (PKS)
	 */
	private PKSMessage	m_PKSlastSent;

	/**
	 * last message received from PKS
	 */
	private PKSMessage	m_PKSlastRecv;

	/**
	 * last message sent to SDS
	 */
	private SDSMessage	m_SDSlastSent;

	/**
	 * last message received from SDS
	 */
	private SDSMessage	m_SDSlastRecv;

	/**
	 * last message sent to DHT
	 */
	private DHTMessage	m_DHTlastSent;

	/**
	 * last message received from DHT
	 */
	private DHTMessage	m_DHTlastRecv;

	/**
	 * the current value for constructing IDs of outbound IQ packets to the OFS
	 * (IDs are Strings, so this value will be concatenated onto something else)
	 */
	private long			m_NextOFSid;

	/**
	 * the current value for constructing IDs of outbound IQ packets to the PKS
	 * (IDs are Strings, so this value will be concatenated onto something else)
	 */
	private long			m_NextPKSid;

	/**
	 * the current value for constructing IDs of outbound IQ packets to the DHT
	 * (IDs are Strings, so this value will be concatenated onto something else)
	 */
	private long			m_NextDHTid;

	/**
	 * the current value for constructing IDs of outbound IQ packets to the SDS
	 * (IDs are Strings, so this value will be concatenated onto something else)
	 */
	private long			m_NextSDSid;

	// END data members


	// BEGIN methods
	
	/**
	 * constructor
	 */
	public VscState() {
		m_LastActivity = new Date();
		// default initializations are good for everything else
	}

	// get accesssor methods
	/**
	 * determine whether the client is logged into a VS
	 * @return true if logged in, false otherwise
	 */
	public boolean isLoggedIn() { return m_LoggedIn; }

	/**
	 * get the time of the last voucher-related activity
	 * @return the date/time last voucher-related message was sent/received
	 */
	public java.util.Date getLastActivity() { return m_LastActivity; }

	/**
	 * determine whether the time limit on last voucher-related activity is up
	 * @param timeout the maximum expiration interval in seconds
	 * @return true if timeout interval has expired, false otherwise
	 */
	public boolean isExpired(int timeout) {
		Date timeNow = new Date();
		long diff = timeNow.getTime() - m_LastActivity.getTime();
		return (diff > (timeout * 1000L) ? true : false);
	}

	/**
	 * determine whether M_INACTIVITY_TIMEOUT on voucher-related activity is up
	 * @return true if default timeout interval has expired, false otherwise
	 */
	public boolean isExpired() {
		return isExpired(M_INACTIVITY_TIMEOUT);
	}

	/**
	 * obtain the last message sent to the OFS
	 * @return the message (actually a OFSMessage, SVXMessage, or SSMMessage)
	 */
	public IQ getLastSentOFS() { return m_OFSlastSent; }

	/**
	 * obtain the last message received from the OFS
	 * @return the message (actually a OFSMessage, SVXMessage, or SSMMessage)
	 */
	public IQ getLastRecvOFS() { return m_OFSlastRecv; }

	/**
	 * obtain the last message sent to the PKS
	 * @return the message object
	 */
	public PKSMessage getLastSentPKS() { return m_PKSlastSent; }

	/**
	 * obtain the last message received from the PKS
	 * @return the message object
	 */
	public PKSMessage getLastRecvPKS() { return m_PKSlastRecv; }

	/**
	 * obtain the last message sent to the DHT
	 * @return the message object
	 */
	public DHTMessage getLastSentDHT() { return m_DHTlastSent; }

	/**
	 * obtain the last message received from the DHT
	 * @return the message object
	 */
	public DHTMessage getLastRecvDHT() { return m_DHTlastRecv; }

	/**
	 * obtain the last message sent to the SDS
	 * @return the message object
	 */
	public SDSMessage getLastSentSDS() { return m_SDSlastSent; }

	/**
	 * obtain the last message received from the SDS
	 * @return the message object
	 */
	public SDSMessage getLastRecvSDS() { return m_SDSlastRecv; }

	/**
	 * obtain the next available Id for an IQ packet outgoing to the OFS
	 * @return the unique series number
	 */
	public synchronized long getNextOFSid() { return ++m_NextOFSid; }

	/**
	 * obtain the next available Id for an IQ packet outgoing to the PKS
	 * @return the unique series number
	 */
	public synchronized long getNextPKSid() { return ++m_NextPKSid; }

	/**
	 * obtain the next available Id for an IQ packet outgoing to the DHT
	 * @return the unique series number
	 */
	public synchronized long getNextDHTid() { return ++m_NextDHTid; }

	/**
	 * obtain the next available Id for an IQ packet outgoing to the SDS
	 * @return the unique series number
	 */
	public synchronized long getNextSDSid() { return ++m_NextSDSid; }

	// set access methods
	/**
	 * set login status
	 * @param li true if logged in, false if not
	 */
	public void setLoggedIn(boolean li) { m_LoggedIn = li; }

	/**
	 * record new activity, along with the IQ stanza exchanged
	 * @param msgContext the last voucher IQ message sent/received
	 */
	public void setLastActivity(IQ msgContext) {
		m_LastActivity = new Date();
		if (msgContext == null) {
			// give up here, in the interest of efficiency
			return;
		}
		String opcode = "";
		if (msgContext instanceof OFSMessage) {
			OFSMessage ofsMsg = (OFSMessage) msgContext;
			opcode = ofsMsg.getOpcode();
			// examine the opcode to see whether it's to or from
			if (opcode.startsWith("REQ")) {
				setLastSentOFS(ofsMsg);
			}
			else if (opcode.startsWith("REP")) {
				setLastRecvOFS(ofsMsg);
			}
		}
		else if (msgContext instanceof SVGSVXIQ) {
			// this counts as OFS input
			SVGSVXIQ svxMsg = (SVGSVXIQ) msgContext;
			opcode = svxMsg.getOpcode();
			// examine the opcode to see whether it's to or from
			if (opcode.startsWith("REQ")) {
				setLastSentOFS(svxMsg);
			}
			else if (opcode.startsWith("REP")) {
				setLastRecvOFS(svxMsg);
			}
		}
		else if (msgContext instanceof SVGSVMIQ) {
			// this counts as OFS input
			SVGSVMIQ svmMsg = (SVGSVMIQ) msgContext;
			opcode = svmMsg.getOpcode();
			// examine the opcode to see whether it's to or from
			if (opcode.startsWith("REQ")) {
				setLastSentOFS(svmMsg);
			}
			else if (opcode.startsWith("REP")) {
				setLastRecvOFS(svmMsg);
			}
		}
		else if (msgContext instanceof SVGSVMIQ) {
			// this counts as OFS input
			SVGSVMIQ svmMsg = (SVGSVMIQ) msgContext;
			opcode = svmMsg.getOpcode();
			// examine the opcode to see whether it's to or from
			if (opcode.startsWith("REQ")) {
				setLastSentOFS(svmMsg);
			}
			else if (opcode.startsWith("REP")) {
				setLastRecvOFS(svmMsg);
			}
		}
		else if (msgContext instanceof PKSMessage) {
			PKSMessage pksMsg = (PKSMessage) msgContext;
			opcode = pksMsg.getOpcode();
			// examine the opcode to see whether it's to or from
			if (opcode.startsWith("REQ")) {
				setLastSentPKS(pksMsg);
			}
			else if (opcode.startsWith("REP")) {
				setLastRecvPKS(pksMsg);
			}
		}
		else if (msgContext instanceof DHTMessage) {
			DHTMessage dhtMsg = (DHTMessage) msgContext;
			opcode = dhtMsg.getOpcode();
			// examine the opcode to see whether it's to or from
			if (opcode.startsWith("REQ")) {
				setLastSentDHT(dhtMsg);
			}
			else if (opcode.startsWith("REP")) {
				setLastRecvDHT(dhtMsg);
			}
		}
		else if (msgContext instanceof SDSMessage) {
			SDSMessage sdsMsg = (SDSMessage) msgContext;
			opcode = sdsMsg.getOpcode();
			// examine the opcode to see whether it's to or from
			if (opcode.startsWith("REQ")) {
				setLastSentSDS(sdsMsg);
			}
			else if (opcode.startsWith("REP")) {
				setLastRecvSDS(sdsMsg);
			}
		}
		else if (msgContext instanceof DiscoverItems) {
			// silently ignore
		}
		else {
			Log.error("Unrecognized last message type seen, "
						+ msgContext.getClass().getName());
		}
	}

	/**
	 * record new activity, without recording a message exchange context
	 */
	public void setLastActivity() {
		m_LastActivity = new Date();
	}

	/**
	 * set the last message sent to the OFS
	 * @param msg the message object (can be null, for erasure)
	 */
	private void setLastSentOFS(IQ msg) {
		m_OFSlastSent = msg;
	}

	/**
	 * set the last message received from the OFS
	 * @param msg the message object (can be null, for erasure)
	 */
	private void setLastRecvOFS(IQ msg) {
		m_OFSlastRecv = msg;
	}

	/**
	 * set the last message sent to the PKS
	 * @param msg the message object (can be null, for erasure)
	 */
	private void setLastSentPKS(PKSMessage msg) {
		m_PKSlastSent = msg;
	}

	/**
	 * set the last message received from the PKS
	 * @param msg the messsage object (can be null, for erasure)
	 */
	private void setLastRecvPKS(PKSMessage msg) {
		m_PKSlastRecv = msg;
	}

	/**
	 * set the last message sent to the DHT
	 * @param msg the message object (can be null, for erasure)
	 */
	private void setLastSentDHT(DHTMessage msg) {
		m_DHTlastSent = msg;
	}

	/**
	 * set the last message received from the DHT
	 * @param msg the messsage object (can be null, for erasure)
	 */
	private void setLastRecvDHT(DHTMessage msg) {
		m_DHTlastRecv = msg;
	}

	/**
	 * set the last message sent to the SDS
	 * @param msg the message object (can be null, for erasure)
	 */
	private void setLastSentSDS(SDSMessage msg) {
		m_SDSlastSent = msg;
	}

	/**
	 * set the last message received from the SDS
	 * @param msg the messsage object (can be null, for erasure)
	 */
	private void setLastRecvSDS(SDSMessage msg) {
		m_SDSlastRecv = msg;
	}

	// END methods
}
