/*
 * last modified---
 * 	08-31-18/kbw/ 0.7.3
 * 	02-17-18/kbw/ 0.7.2
 * 	03-14-17/jet/ 0.7.1
 * 	01-24-17/jet/ 0.7.0
 * 	07-15-15/jet/ clear asset/token config on logout
 * 	07-08-15/jet/ version for SVSpark 2.7.1, with ABC in separate plugin
 * 	04-30-15/jet/ merge SVWebApp changes for ABC beta
 * 	08-23-14/jet/ add reorderGateways() and its usage, to speed up logins
 * 	08-01-14/jet/ version for Spark plugin
 * 	03-19-14/jet/ 0.4.3: new OFS gateway and PXGOLD asset support; fix bad
 * 					  voucher sig error for certain sigs created with Java-6;
 * 					  add manual entry for OFS gateway, and connect on choice;
 *						  add "Get JID" menu item which copies to clipboard;
 *						  add doDisconnect()
 * 	01-29-14/jet/ when logging in after timeout, auto-reconnect first
 * 	01-14-14/jet/ add handling for -p option to main() for manual invocation
 * 	11-21-13/jet/ require a connect to :80 at the OFS prior to XMPP c2s open;
 * 					  when SDS or DHT connects fail, retry with prior :80 connect
 * 	11-09-13/jet/ 0.3, silence connection debug and remove beta designation,
 * 					  support TestNet property
 * 	10-12-13/jet/ support separate XMPP domain definitions for OFS/SDS/DHT
 * 					  along with multiple host:port paths
 * 	06-13-13/jet/ add doDisco()
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	main class for Spark plugin implementation of Silent-Vault client
 */

package org.silentvault.spark;

import javax.swing.*;
import java.awt.GridLayout;
import java.awt.Dimension;
import java.awt.Cursor;
import java.awt.Font;
import java.awt.Color;
import java.awt.Point;
import java.awt.Component;
import java.awt.GraphicsEnvironment;
import java.awt.event.KeyEvent;
import java.awt.event.ActionEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Clipboard;
import java.awt.Toolkit;

import org.jivesoftware.resource.SparkRes;
import org.jivesoftware.spark.plugin.Plugin;
import org.jivesoftware.spark.*;
import org.jivesoftware.Spark;
import org.jivesoftware.MainWindow;
import org.jivesoftware.spark.preference.PreferenceManager;
import org.jivesoftware.spark.util.log.Log;
import org.jivesoftware.spark.component.tabbedPane.*;

import org.jivesoftware.smack.provider.*;
import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.XMPPException;
import org.jivesoftware.smack.ConnectionConfiguration;
import org.jivesoftware.smack.ConnectionListener;
import org.jivesoftware.smack.SmackConfiguration;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.packet.Packet;
import org.jivesoftware.smack.packet.Message;
import org.jivesoftware.smack.filter.PacketFilter;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smackx.packet.DiscoverItems;

import org.silentvault.spark.ui.*;
import org.silentvault.spark.ui.wallet.*;
import org.silentvault.spark.ui.svx.*;
import org.silentvault.spark.ui.svm.*;

import java.security.*;
import java.util.*;
import java.net.InetSocketAddress;
import java.net.URL;
import java.net.URLConnection;
import java.net.MalformedURLException;
import java.net.HttpURLConnection;
import java.net.ProtocolException;
import java.io.File;
import java.io.IOException;


/**
 * Implements the SilentVault wallet client as a Spark plugin.
 * This allows the application to act as a voucher safe client
 * within the context of the vouchi.com voucher payment network.
 * It also provides support for the Exchanges and Marketplaces tabs.
 *
 * @author Justin Turrell
 */
public final class WalletClient extends TimerTask implements Plugin {
	// BEGIN data members
	/**
	 * version string displayed in Help/About
	 */
	private static final String	M_Version
										= "Spark Plugin SilentVault Client rev. 0.7.3\n"
											+ " Built 31 August, 2018\n";

	/**
	 * path to help icon
	 */
	private static final String	M_HelpIconName = "images/help16x16Flat.png";

	/**
	 * path to info icon
	 */
	public static final	String	M_InfoIconName = "images/info16x16Flat.png";
	
	/**
	 * path to Voucher-Safe mini icon
	 */
	private static final	String	M_VSIconName = "images/VS_Logo.png";
	
	/**
	 * path to SV mini icon
	 */
	private static final String	M_SVIconName = "images/SV_Padlock.png";

	/**
	 * the actual help icon, inside a label
	 */
	public final JLabel		M_HelpIcon;

	/**
	 * the actual info icon, inside a label
	 */
	public final JLabel		M_InfoIcon;

	/**
	 * the actual VS project icon, inside a label
	 */
	public final JLabel		M_VSIcon;

	/**
	 * the actual SV project icon, inside a label
	 */
	public final JLabel		M_SVIcon;

	/**
	 * default font to use on buttons
	 */
	public final Font			M_ButtonFont;

	/**
	 * default background to use on non-interactive (label) text fields
	 */
	public Color				m_LabelBackColor;

	/**
	 * the voucher system state-tracking object
	 */
	private VscState			m_StateData;

	/**
	 * the object for storing login data (secrets)
	 */
	private VsSecrets			m_LoginSecrets;

	/**
	 * the time at which this plugin was initialized by Spark
	 */
	private Date				m_InitTimestamp;

	/**
	 * the IQ provider manager object (from Smack)
	 */
	private ProviderManager	m_ProviderManager;

	/**
	 * the object that listens for incoming OFS packets (replies)
	 */
	private OFSListener		m_OFSListener;

	/**
	 * the object that listens for incoming SVX packets (replies and broadcasts)
	 */
	private SVXListener		m_SVXListener;

	/**
	 * the object that listens for incoming SVM packets (replies and broadcasts)
	 */
	private SVMListener		m_SVMListener;

	/**
	 * the object that listens for incoming PKS packets (replies)
	 */
	private PKSListener		m_PKSListener;

	/**
	 * the object that listens for incoming DHT packets (replies)
	 */
	private DHTListener		m_DHTListener;

	/**
	 * the object that listens for incoming SDS packets (replies)
	 */
	private SDSListener		m_SDSListener;

	/**
	 * the object that listens for incoming website order messages
	 */
	private WebListener		m_WebListener;

	/**
	 * the filter which identifies website order messages
	 */
	private PacketFilter		m_WebMsgFilter;

	/**
	 * the filter which identifies SVM ticker and broadcast messages
	 */
	private PacketFilter		m_SVMMsgFilter;

	/**
	 * the filter which identifies SVX broadcast messages
	 */
	private PacketFilter		m_SVXMsgFilter;

	/**
	 * UI screen manager and main container for VS wallet tab
	 */
	private WTabManager		m_WalletTabManager;

	/**
	 * UI screen manager and main container for Exchanges tab
	 */
	private XTabManager		m_ExchangeTabManager;

	/**
	 * UI screen manager and main container for Marketplaces tab
	 */
	private MTabManager		m_MarketplaceTabManager;

	/**
	 * spent token list, to track tokens offered to SDS/DHT nodes but not yet
	 * cleared out of our folder
	 */
	private SpentTokenCache	m_SpentCache;

	/**
	 * the object which fetches and caches current asset prices
	 */
	private AssetPriceCache	m_PricingModule;

	/**
	 * app-specific user preferences
	 */
	private VSPreferences	m_UserPrefs;

	/**
	 * the timer which handles logging out on inactivity
	 */
	private java.util.Timer	m_InactivityTimer;

	/**
	 * action for making a connection to the OFS
	 */
	private AbstractAction	m_ConnectAction;

	/**
	 * action for disconnecting from the OFS
	 */
	private AbstractAction	m_DisconnectAction;

	/**
	 * action for exiting
	 */
	private AbstractAction	m_ExitAction;

	/**
	 * action for going to preferences screen
	 */
	private AbstractAction	m_PreferencesAction;

	/**
	 * action for logout from a VS
	 */
	private AbstractAction	m_LogoutAction;

	/**
	 * action for JID display
	 */
	private AbstractAction	m_ShowJIDAction;

	/**
	 * action for JID get (to clipboard)
	 */
	private AbstractAction	m_GetJIDAction;

	/**
	 * action for closing a VS
	 */
	private AbstractAction	m_CloseSafeAction;

	/**
	 * action for displaying tokens
	 */
	private AbstractAction	m_ShowTokensAction;

	/**
	 * action for displaying asset prices
	 */
	private AbstractAction	m_ShowPricesAction;

	/**
	 * action for displaying tutorial help
	 */
	private AbstractAction	m_TutorialHelpAction;

	/**
	 * action to validate a voucher
	 */
	private AbstractAction	m_ValidateAction;

	/**
	 * action to split a voucher
	 */
	private AbstractAction	m_SplitAction;

	/**
	 * action to merge 2 or more vouchers
	 */
	private AbstractAction	m_MergeAction;

	/**
	 * action to deal with expired vouchers
	 */
	private AbstractAction	m_ExpiredAction;

	/**
	 * list of popup windows which must all be closed on VS logout
	 */
	private HashSet<JFrame>	m_PopupList;

	/**
	 * inner class to provide OFS connection listener for status bar updates
	 */
	private class OFSConnListener implements ConnectionListener {
		public final Font		M_LabFont;
		private JLabel			m_ConLab;
		private WalletClient	m_Plugin;

		/**
		 * constructor
		 * @param plugin the application of which we're a part
		 */
		public OFSConnListener(WalletClient plugin) {
			m_Plugin = plugin;
			M_LabFont = new Font(Font.SANS_SERIF, Font.BOLD, 13);
			m_ConLab = new JLabel();
			m_ConLab.setFont(M_LabFont);
		}

		/**
		 * handle reconnection abort notification and close on fatal error
		 */
		public void connectionClosed() {
			/* NB: regular connect closed does not come here, because we remove
			 * the listener first.  So this can only be called from within Spark
			 * and/or Smack.
			 */
			connectionClosedOnError(null);
		}

		/**
		 * handle abnormal connection closed notification
		 * @param excep the exception which caused the closure, if any
		 */
		public void connectionClosedOnError(Exception excep) {
			if (m_StateData != null && m_StateData.isLoggedIn()) {
				doLogout();
				// show disconnected screen
				m_WalletTabManager.queueTransition(
													new Integer(WTabManager.M_DISCONNECTED));
				m_WalletTabManager.makeTransition();
			}
			m_ConLab.setText(" OFS connection lost - Not ready");
			JPanel statBar = getStatusBar();
			if (statBar != null) {
				statBar.removeAll();
				statBar.add(m_ConLab);
				statBar.add(Box.createHorizontalStrut(300));
				statBar.revalidate();
				statBar.repaint();
			}
		}

		/**
		 * handle pending reconnection notification
		 * @param secs the number of seconds till reconnect attempt is made
		 */
		public void reconnectingIn(int secs) {
			if (secs <= 7) {
				m_ConLab.setText(" Reconnecting to OFS in " + secs + " seconds");
				JPanel statBar = getStatusBar();
				if (statBar != null) {
					statBar.removeAll();
					statBar.add(m_ConLab);
					statBar.add(Box.createHorizontalStrut(250));
					statBar.revalidate();
					statBar.repaint();
				}
			}
		}

		/**
		 * handle successful reconnection notification
		 */
		public void reconnectionSuccessful() {
			setDefaultStatusBar();
			if (m_StateData != null && !m_StateData.isLoggedIn()) {
				m_WalletTabManager.queueTransition(
															new Integer(WTabManager.M_LOGIN1));
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						m_WalletTabManager.makeTransition();
					}
				});
			}
			// reset init timestamp
			m_InitTimestamp = new Date();
		}

		/**
		 * handle reconnection failure notification
		 * @param excep the exception which caused the failure
		 */
		public void reconnectionFailed(Exception excep) {
			m_ConLab.setText(" OFS reconnection failed - Not ready");
			JPanel statBar = getStatusBar();
			if (statBar != null) {
				statBar.removeAll();
				statBar.add(m_ConLab);
				statBar.add(Box.createHorizontalStrut(240));
				statBar.revalidate();
				statBar.repaint();
			}

			// now force a logout if needed, and show disconnected screen
			if (m_StateData != null && m_StateData.isLoggedIn()) {
				doLogout();
				m_WalletTabManager.queueTransition(
													new Integer(WTabManager.M_DISCONNECTED));
				m_WalletTabManager.makeTransition();
			}
		}
	}

	/**
	 * listener for connect events on the OFS port
	 */
	private OFSConnListener	m_OFSConnectListener;

	/**
	 * whether we're running on a test network (@voucher-safe.org)
	 */
	private boolean			m_TestNet;

	/**
	 * the plugin specifying the SVM interface to the currently connected SVM
	 */
	private SVMPlugin			m_SVMPlugin;

	// END data members

	// BEGIN methods
	/**
	 * contructor
	 */
	public WalletClient() {
		super();
		m_ProviderManager = ProviderManager.getInstance();
		M_ButtonFont = new Font(Font.SANS_SERIF, Font.BOLD, 12);
		m_LabelBackColor = new Color(214, 217, 223);
		m_PopupList = new HashSet<JFrame>();
		m_OFSConnectListener = new OFSConnListener(this);

		// load standard icons
		ImageIcon helpIcon = null;
		ImageIcon projIcon = null;
		ImageIcon svIcon = null;
		ImageIcon infoIcon = null;
		// this one comes from plugin's images/
		URL imgURL = getClass().getResource(M_HelpIconName);
		if (imgURL == null) {
			Log.error("Could not find " + M_HelpIconName);
		}
		else {
			helpIcon = new ImageIcon(imgURL, "help icon");
		}
		M_HelpIcon = new JLabel(helpIcon);
		// this one comes from plugin's images/
		imgURL = getClass().getResource(M_InfoIconName);
		if (imgURL == null) {
			Log.error("Could not find " + M_InfoIconName);
		}
		else {
			infoIcon = new ImageIcon(imgURL, "info icon");
		}
		M_InfoIcon = new JLabel(infoIcon);
		// pull this one from Spark's images/
		imgURL = getClass().getClassLoader().getResource(M_VSIconName);
		if (imgURL == null) {
			Log.error("Could not find " + M_VSIconName);
		}
		else {
			projIcon = new ImageIcon(imgURL, "Voucher-Safe icon");
		}
		M_VSIcon = new JLabel(projIcon);
		// pull this one from Spark's images/
		imgURL = getClass().getClassLoader().getResource(M_SVIconName);
		if (imgURL == null) {
			Log.error("Could not find " + M_SVIconName);
		}
		else {
			svIcon = new ImageIcon(imgURL, "SilentVault icon");
		}
		M_SVIcon = new JLabel(svIcon);

		// provide standard border and label color
		Color stndGrey = Color.LIGHT_GRAY.darker();
		UIManager.put("standardGray", stndGrey);
		// set grey color for Nimbus progress bars also, by overriding orange
		UIManager.put("nimbusOrange", stndGrey);
	}

	// method to extend the TimerTask class
	/**
	 * perform the action to be run by the inactivity timer
	 */
	public void run() {
		if (m_StateData == null) {
			Log.error("Inactivity timer has no state data available");
			return;
		}
		if (m_StateData == null || !m_StateData.isLoggedIn()) {
			// nothing to do
			return;
		}

		// get the timeout interval from user preferences
		int inactMin = m_UserPrefs.getInactivity();
		if (m_StateData.isExpired(inactMin * 60)) {
			// perform logout and return to login screen
			doLogout();
			m_WalletTabManager.queueTransition(
												new Integer(WTabManager.M_LOGOUT_SAFE));
			m_WalletTabManager.makeTransition();
		}
	}

	// BEGIN methods to implement the Plugin interface
   /**
	 * called after Spark is loaded to initialize the plugin
	 */
   public void initialize() {
      //System.setProperty("debug.mode", "true");
		Log.debug("Welcome to " + M_Version + ".");

		// init state data
		m_StateData = new VscState();

		// init login blob and network config data
		m_LoginSecrets = new VsSecrets();

		// initialze provider manager to know about our IQ provider classes
		OFSMessageProvider ofsProvider = new OFSMessageProvider();
		m_ProviderManager.addIQProvider(	"vsc_ofs",
													"jabber:iq:voucher-safe#vsc-ofs",
													ofsProvider	);
		PKSMessageProvider pksProvider = new PKSMessageProvider();
		m_ProviderManager.addIQProvider(	"sys_pks",
													"jabber:iq:voucher-safe#sys-pks",
													pksProvider	);
		DHTMessageProvider dhtProvider = new DHTMessageProvider();
		m_ProviderManager.addIQProvider(	"vsc_dht",
													"jabber:iq:voucher-safe#vsc-dht",
													dhtProvider	);
		SDSMessageProvider sdsProvider = new SDSMessageProvider();
		m_ProviderManager.addIQProvider(	"vsc_sds",
													"jabber:iq:voucher-safe#vsc-sds",
													sdsProvider	);
		SVGSVMMessageProvider svmProvider = new SVGSVMMessageProvider();
		m_ProviderManager.addIQProvider(	"vsc_ofs",
													"jabber:iq:silent-market#vsc-ofs",
													svmProvider);
		SVGSVXMessageProvider svxProvider = new SVGSVXMessageProvider();
		m_ProviderManager.addIQProvider(	"vsc_ofs",
													"jabber:iq:silent-vault#vsc-ofs",
													svxProvider);

		// initialize provider manager to know about our PacketExtension providers
		WebMessageProvider webProvider = new WebMessageProvider();
		m_ProviderManager.addExtensionProvider("web_vsc",
														"jabber:message:voucher-safe#web-vsc",
															webProvider );
		m_ProviderManager.addExtensionProvider("vsc_ofs",
													"jabber:message:silent-market#vsc-ofs",
															svmProvider );
		m_ProviderManager.addExtensionProvider("vsc_ofs",
														"jabber:message:silent-vault#vsc-ofs",
															svxProvider );

		// initialize listeners for IQ packets from partners
		m_OFSListener = new OFSListener(this);
		m_PKSListener = new PKSListener(this);
		m_DHTListener = new DHTListener(this);
		m_SDSListener = new SDSListener(this);
		m_SVXListener = new SVXListener(this);
		// supply default SVM listener
		m_SVMListener = new SVMListener(this);

		// init listener for website packets
		m_WebListener = new WebListener();

		// set up filter to listen for Message packets from websites
		m_WebMsgFilter = new PacketFilter() {
			public boolean accept(Packet packet) {
				if (packet instanceof Message) {
					Message msg = (Message) packet;
					if (msg.getType() == Message.Type.normal) {
						// see if we have an extension present
						if (msg.getExtension("web_vsc",
													"jabber:message:voucher-safe#web-vsc")
							!= null)
						{
							return true;
						}
					}
				}
				return false;
			}
		};

		// set up filter to listen for broadcast and ticker packets from an SVM
		m_SVMMsgFilter = new PacketFilter() {
			public boolean accept(Packet packet) {
				if (packet instanceof Message) {
					Message msg = (Message) packet;
					if (msg.getType() == Message.Type.normal) {
						// see if we have the SVM extension present
						if (msg.getExtension("vsc_ofs",
													"jabber:message:silent-market#vsc-ofs")
							!= null)
						{
							return true;
						}
					}
				}
				return false;
			}
		};

		// set up filter to listen for broadcast packets from an SVX
		m_SVXMsgFilter = new PacketFilter() {
			public boolean accept(Packet packet) {
				if (packet instanceof Message) {
					Message msg = (Message) packet;
					if (msg.getType() == Message.Type.normal) {
						// see if we have the SVM extension present
						if (msg.getExtension("vsc_ofs",
													"jabber:message:silent-vault#vsc-ofs")
							!= null)
						{
							return true;
						}
					}
				}
				return false;
			}
		};
		final XMPPConnection conn = getOFSConnection();
		conn.addPacketListener(m_WebListener, m_WebMsgFilter);
		conn.addPacketListener(m_SVXListener, m_SVXMsgFilter);
		conn.addPacketListener(m_SVMListener, m_SVMMsgFilter);

		// register connect listener, to keep stat bar up to date
		conn.addConnectionListener(m_OFSConnectListener);

		/* Search the discovered services to make sure the VP and SDS interface
		 * components are available.  (Without them, we cannot log in or create
		 * a wallet.)  Also look for the list of SVXen and SVMs, and note them.
		 */
		SessionManager sm = SparkManager.getSessionManager();
		DiscoverItems discoItems = sm.getDiscoveredItems();
		Iterator<DiscoverItems.Item> dIt = discoItems.getItems();
		boolean foundVP = false;
		boolean foundSDS = false;
		boolean foundSVX = false;
		ArrayList<String> svxIDs = new ArrayList<String>();
		boolean foundSVM = false;
		ArrayList<String> svmIDs = new ArrayList<String>();
		while (dIt.hasNext()) {
			DiscoverItems.Item di = dIt.next();
			/* NB: these strings much match the value of M_Description in the
			 * 	 code for the OFS's Components.
			 */
			String name = di.getName();
			if (name.equals("Voucher Publisher (VP) Interface Component")) {
				foundVP = true;
			}
			else if (name.equals("Secure Data Store (SDS) Interface Component")) {
				foundSDS = true;
			}
			else if (name.startsWith(
									"Silent Vault Exchange (SVX) Interface Component"))
			{
				foundSVX = true;
				// add the ID to the list
				int delimiterIdx = name.indexOf("::");
				if (delimiterIdx != -1) {
					String svxId = name.substring(delimiterIdx+2);
					svxIDs.add(svxId);
				}
			}
			else if (name.startsWith(
										"Silent Vault Market (SVM) Interface Component"))
			{
				foundSVM = true;
				// add the ID to the list
				int delimiterIdx = name.indexOf("::");
				if (delimiterIdx != -1) {
					String svmId = name.substring(delimiterIdx+2);
					svmIDs.add(svmId);
				}
			}
		}

		if (foundVP && foundSDS) {
			// define preferences and register with manager
			m_PricingModule = new AssetPriceCache(this);
			m_UserPrefs = new VSPreferences(this);
			m_UserPrefs.load();
			SparkManager.getPreferenceManager().addPreference(m_UserPrefs);

			// set the Smack timeout to prefs for use with opening connections
			SmackConfiguration.setPacketReplyTimeout(
												((int)m_UserPrefs.getTimeout() + 2) * 1000);

			// init the tab managers (this actually creates the 3 tabs)
			m_WalletTabManager = new WTabManager(this);
			m_ExchangeTabManager = new XTabManager(this);
			if (foundSVX) {
				// tell the ConnectPane about the SVX list
				m_ExchangeTabManager.getConnectPane().recordSVXList(svxIDs);
			}
			m_MarketplaceTabManager = new MTabManager(this);
			if (foundSVM) {
				// tell the LoginPane about the SVM list
				m_MarketplaceTabManager.getLoginPane().recordSVMList(svmIDs);
			}
			// supply default SVM plugin configuration
			m_SVMPlugin = new SVMPlugin(this);
			m_SVMPlugin.setSVMListener(m_SVMListener);
			m_SVMPlugin.setClientBlock("org.silentvault.spark.SVMClientBlock");
			m_SVMPlugin.setReplyBlock("org.silentvault.spark.SVMBlock");
			m_SVMPlugin.setTabManager(m_MarketplaceTabManager);

			// define spent tokens cache
			m_SpentCache = new SpentTokenCache(this);
			if (!m_SpentCache.initSpentTokens()) {
				Log.error("Unable to initialize saved spent tokens from file");
				// do not run without this
				shutdown();
			}

			// activate inactivity timer, checking once a minute
			m_InactivityTimer = new java.util.Timer("inactivity");
			m_InactivityTimer.schedule(this, 60000L, 60000L);

			// determine whether we're running on a test network
			setTestNetwork(conn.getHost().contains("voucher-safe.org"));

			// init the prices and the UI, then go to login screen
			m_WalletTabManager.queueTransition(new Integer(WTabManager.M_LOGIN1));
			SwingUtilities.invokeLater(new Runnable() {
				public void run() {
					// fetch initial current asset prices and start fetch timer
					m_PricingModule.run();
					m_PricingModule.initPriceFetching();

					// add UI elements and transition to login screen
					addTabsToSpark();
					m_WalletTabManager.makeTransition();
				}
			});
		}
		else {
			Log.error("XMPP server does not support voucher safe operations");
		}

		// record when initialized by Spark
		m_InitTimestamp = new Date();
		Log.debug("SilentVault Client Plugin Initialized");
   }

   /**
	 * Called when Spark is shutting down, to allow for persistence of
	 * information or releasing of resources.
 	 */
   public void shutdown() {
		// nix inactivity timer
		if (m_InactivityTimer != null) {
			m_InactivityTimer.cancel();
			m_InactivityTimer = null;
		}
		m_StateData = null;

		// abolish specific namespace IQ providers
		if (m_ProviderManager == null) {
			m_ProviderManager = ProviderManager.getInstance();
		}
		m_ProviderManager.removeIQProvider("vsc_ofs",
														"jabber:iq:voucher-safe#vsc-ofs");
		m_ProviderManager.removeIQProvider("sys_pks",
														"jabber:iq:voucher-safe#sys-pks");
		m_ProviderManager.removeIQProvider("vsc_dht",
														"jabber:iq:voucher-safe#vsc-dht");
		m_ProviderManager.removeIQProvider("vsc_sds",
														"jabber:iq:voucher-safe#vsc-sds");
		m_ProviderManager.removeIQProvider("vsc_ofs",
														"jabber:iq:silent-market#vsc-ofs");
		m_ProviderManager.removeIQProvider("vsc_ofs",
														"jabber:iq:silent-vault#vsc-ofs");
		// abolish extension providers
		m_ProviderManager.removeExtensionProvider("web_vsc",
													"jabber:message:voucher-safe#web-vsc");
		m_ProviderManager.removeExtensionProvider("vsc_ofs",
													"jabber:message:silent-market#vsc-ofs");
		m_ProviderManager.removeExtensionProvider("vsc_ofs",
													"jabber:message:silent-vault#vsc-ofs");

		// stop listening for OFS reply packets
		final XMPPConnection conn = getOFSConnection();
		if (conn != null && conn.isConnected()) {
			conn.removePacketListener(m_WebListener);
			conn.removePacketListener(m_SVMListener);
			conn.removePacketListener(m_SVXListener);
		}
		m_OFSListener = null;
		if (m_PKSListener != null) {
			m_PKSListener.closePKSConnection();
		}
		m_PKSListener = null;
		if (m_DHTListener != null) {
			m_DHTListener.closeDHTConnection();
		}
		m_DHTListener = null;
		if (m_SDSListener != null) {
			m_SDSListener.closeSDSConnection();
		}
		m_SDSListener = null;
		m_WebListener = null;
		m_WebMsgFilter = null;
		m_SVMMsgFilter = null;
		m_SVXMsgFilter = null;
		m_SVMListener = null;
		m_SVXListener = null;
		m_ProviderManager = null;
		if (m_PricingModule != null) {
			m_PricingModule.cancelPriceFetching();
			m_PricingModule = null;
		}

		// get rid of any extant UI elements
		if (m_WalletTabManager != null) {
			m_WalletTabManager.deallocatePanes();
			m_WalletTabManager = null;
		}
		if (m_ExchangeTabManager != null) {
			m_ExchangeTabManager.deallocatePanes();
			m_ExchangeTabManager = null;
		}
		if (m_MarketplaceTabManager != null) {
			m_MarketplaceTabManager.deallocatePanes();
			m_MarketplaceTabManager = null;
		}

		// sync spent token cache and then clear secrets
		if (m_SpentCache != null) {
			m_SpentCache.syncSpentTokens();
		}
		resetSecrets();
		Log.debug("Silent-Vault Client Plugin De-initialized");
   }

	/**
	 * return true if Spark can shutdown on user's request
	 * @return true always
	 */
	public boolean canShutDown() { return true; }

	/**
	 * called when a user explicitly asked to uninstall this plugin.
	 */
	public void uninstall() {
		// Remove all resources belonging to this plugin.
		// This would include silentvault.properties, spentTokens.xml, etc.
		// TBD.
	}
	// END methods to implement the Plugin interface

	/**
	 * reset the secrets object
	 */
	public void resetSecrets() { m_LoginSecrets = new VsSecrets(); }

	/**
	 * Adds all 3 tabs to Spark.  Initiates tab manager objects and tells them
	 * to start with login screens.  Also defines plugin menu items.
	 */
	private void addTabsToSpark() {
		// specify preference for proportional fonts, to help make layouts
		// consistent across various platforms
		GraphicsEnvironment g = GraphicsEnvironment.getLocalGraphicsEnvironment();
		g.preferProportionalFonts();

		// add the 3 tab managers to the Tabbed Pane from Spark
		Workspace workspace = SparkManager.getWorkspace();
		SparkTabbedPane tabbedPane = workspace.getWorkspacePane();
		tabbedPane.addTab("Wallet",
								SparkRes.getImageIcon(SparkRes.LOGIN_KEY_IMAGE),
								m_WalletTabManager,
								"Perform wallet operations");
		tabbedPane.addTab("Exchanges", M_SVIcon.getIcon(),
								m_ExchangeTabManager,
								"SilentVault Exchange operations");
		tabbedPane.addTab("Marketplaces", M_SVIcon.getIcon(),
								m_MarketplaceTabManager,
								"SilentVault Market operations");
		/* NB: setting no dragNdrop target mitigates a Sun Java bug which
		 * 	 causes this exception to be thrown (on Linux only):
		 * 	 sun.awt.X11.XException: Cannot write XdndAware property
		 */
		m_WalletTabManager.setDropTarget(null);
		m_ExchangeTabManager.setDropTarget(null);
		m_MarketplaceTabManager.setDropTarget(null);

		// prepare advanced functions menu
		final MainWindow mainWindow = SparkManager.getMainWindow();
		JMenu advMenu = new JMenu("Wallet");
		advMenu.setMnemonic(KeyEvent.VK_W);
		final XMPPConnection conn = getOFSConnection();

		// define action to show JID
		m_ShowJIDAction = new AbstractAction("Show JID") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (conn != null && conn.isConnected()) {
					// create a popup window which displays the user's JID
					JFrame jPop = new JFrame("Your Jabber ID");
					jPop.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);
					JTextArea jid = new JTextArea(9, 30);
					jid.setEditable(false);
					jid.setLineWrap(true);
					// use same font as HomePane.M_DataFont
					jid.setFont(new Font(Font.SANS_SERIF, Font.PLAIN, 12));
					jid.setWrapStyleWord(true);
					StringBuilder jidTxt = new StringBuilder(128);
					jidTxt.append("Your current Jabber ID (JID) is:\n\n");
					String userJID = conn.getUser();
					int slash = userJID.lastIndexOf('/');
					jidTxt.append(userJID.substring(0, slash) + "\n\n");
					jidTxt.append("You may be asked for this to make a payment on ");
					jidTxt.append("a website which accepts vouchers.\n");
					jid.setText(jidTxt.toString());
					jPop.add(jid);
					jPop.setSize(new Dimension(230, 150));
					jPop.setLocationRelativeTo(mainWindow);
					jPop.setVisible(true);
					registerPopup(jPop);
				}
				else {
					m_WalletTabManager.showInfo("You are not presently connected");
				}
			}
		};
		JMenuItem jidItem = new JMenuItem(m_ShowJIDAction);
		jidItem.setMnemonic(KeyEvent.VK_J);
		jidItem.setIcon(SparkRes.getImageIcon(SparkRes.SMALL_USER1_INFORMATION));
		jidItem.setToolTipText("Display your anonymous Jabber connect ID");
		advMenu.add(jidItem);

		// define action to copy JID to clipboard
		m_GetJIDAction = new AbstractAction("Get JID") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (conn != null && conn.isConnected()) {
					String userJID = conn.getUser();
					int slash = userJID.lastIndexOf('/');
					StringSelection ss
						= new StringSelection(userJID.substring(0, slash));
					Toolkit tk = Toolkit.getDefaultToolkit();
					Clipboard sysClip = tk.getSystemClipboard();
					sysClip.setContents(ss, ss);
					m_WalletTabManager.showInfo("Your JID has been copied to the "
														+ "clipboard.\nYou may use this to "
														+ "make a payment on a website which "
														+ "accepts vouchers.");
				}
				else {
					m_WalletTabManager.showInfo("You are not presently connected");
				}
			}
		};
		JMenuItem jidItem2 = new JMenuItem(m_GetJIDAction);
		jidItem2.setMnemonic(KeyEvent.VK_G);
		jidItem2.setIcon(SparkRes.getImageIcon(SparkRes.NOTEBOOK_IMAGE));
		jidItem2.setToolTipText("Copy your anonymous Jabber connect ID to the "
										+ "clipboard");
		advMenu.add(jidItem2);

		// define action to show prices
		m_ShowPricesAction = new AbstractAction("Show Prices") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_StateData != null && m_StateData.isLoggedIn()) {
					// go to prices screen
					m_WalletTabManager.queueTransition(
															new Integer(WTabManager.M_PRICES));
					m_WalletTabManager.makeTransition();
				}
				else {
					m_WalletTabManager.showError("cannot display prices",
														  "not logged in",
														"please log into your wallet first");
				}
			}
		};
		JMenuItem pricesItem = new JMenuItem(m_ShowPricesAction);
		pricesItem.setMnemonic(KeyEvent.VK_P);
		pricesItem.setToolTipText("Display current gold prices");
		pricesItem.setIcon(SparkRes.getImageIcon(SparkRes.DOWN_ARROW_IMAGE));
		advMenu.add(pricesItem);

		// define action to show tokens
		m_ShowTokensAction = new AbstractAction("Show Tokens") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_StateData != null && m_StateData.isLoggedIn()) {
					// show token table in a popup
					JFrame tPop = new JFrame("Snapshot of Tokens in "
													+ m_LoginSecrets.getVSnumber());
					tPop.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
					JTable tokTab = m_WalletTabManager.getHomePane().getTokenTable();
					JScrollPane js = new JScrollPane(tokTab);
					tPop.setMinimumSize(new Dimension(455, 250));
					tPop.setPreferredSize(new Dimension(610, 400));
					tPop.add(js);
					tPop.setLocationRelativeTo(m_WalletTabManager);
					tPop.setVisible(true);
					registerPopup(tPop);
				}
				else {
					m_WalletTabManager.showInfo("You are not logged in to a wallet");
				}
			}
		};
		JMenuItem tokensItem = new JMenuItem(m_ShowTokensAction);
		tokensItem.setMnemonic(KeyEvent.VK_T);
		tokensItem.setToolTipText("Display the actual tokens in your safe");
		tokensItem.setIcon(SparkRes.getImageIcon(SparkRes.SMALL_DOCUMENT_VIEW));
		advMenu.add(tokensItem);

		// define action to validate a voucher
		m_ValidateAction = new AbstractAction("Validate") {
			public void actionPerformed(ActionEvent actionEvent) {
				// get rid of the voucher list popup if one is displayed
				MenuPane mp = m_WalletTabManager.getMenuPane();
				if (mp != null) {
					mp.closeVoucherPopup();
				}

				if (m_StateData != null && m_StateData.isLoggedIn()) {
					// switch to validate screen
					m_WalletTabManager.queueTransition(
														new Integer(WTabManager.M_VALIDATE));
					m_WalletTabManager.makeTransition();
				}
				else {
					m_WalletTabManager.showError("cannot validate", "not logged in",
													"you must log into your wallet first");
				}
			}
		};
		JMenuItem validateItem = new JMenuItem(m_ValidateAction);
		validateItem.setMnemonic(KeyEvent.VK_V);
		validateItem.setToolTipText("Re-validate a voucher");
		validateItem.setIcon(SparkRes.getImageIcon(SparkRes.CIRCLE_CHECK_IMAGE));
		advMenu.add(validateItem);

		// define action to split a voucher
		m_SplitAction = new AbstractAction("Split") {
			public void actionPerformed(ActionEvent actionEvent) {
				// get rid of the voucher list popup if one is displayed
				MenuPane mp = m_WalletTabManager.getMenuPane();
				if (mp != null) {
					mp.closeVoucherPopup();
				}

				if (m_StateData != null && m_StateData.isLoggedIn()) {
					// switch to split screen
					m_WalletTabManager.queueTransition(
															new Integer(WTabManager.M_SPLIT));
					m_WalletTabManager.makeTransition();
				}
				else {
					m_WalletTabManager.showError("cannot split", "not logged in",
													"you must log into your wallet first");
				}
			}
		};
		JMenuItem splitItem = new JMenuItem(m_SplitAction);
		splitItem.setMnemonic(KeyEvent.VK_S);
		splitItem.setToolTipText("Split a voucher into smaller vouchers");
		splitItem.setIcon(SparkRes.getImageIcon(SparkRes.SMALL_DOCUMENT_ADD));
		advMenu.add(splitItem);

		// define action to merge vouchers
		m_MergeAction = new AbstractAction("Merge") {
			public void actionPerformed(ActionEvent actionEvent) {
				// get rid of the voucher list popup if one is displayed
				MenuPane mp = m_WalletTabManager.getMenuPane();
				if (mp != null) {
					mp.closeVoucherPopup();
				}

				if (m_StateData != null && m_StateData.isLoggedIn()) {
					// switch to merge screen
					m_WalletTabManager.queueTransition(
															new Integer(WTabManager.M_MERGE));
					m_WalletTabManager.makeTransition();
				}
				else {
					m_WalletTabManager.showError("cannot merge", "not logged in",
													"you must log into your wallet first");
				}
			}
		};
		JMenuItem mergeItem = new JMenuItem(m_MergeAction);
		mergeItem.setMnemonic(KeyEvent.VK_M);
		mergeItem.setToolTipText("Consolidate vouchers into one");
		mergeItem.setIcon(SparkRes.getImageIcon(SparkRes.FUNNEL_DOWN_16x16));
		advMenu.add(mergeItem);

		// define action to buy tokens manually
		AbstractAction purchTok = new AbstractAction("Buy Tokens") {
			public void actionPerformed(ActionEvent actionEvent) {
				// get rid of the voucher list popup if one is displayed
				MenuPane mp = m_WalletTabManager.getMenuPane();
				if (mp != null) {
					mp.closeVoucherPopup();
				}

				if (m_StateData != null && m_StateData.isLoggedIn()) {
					// switch to token purchase screen
					m_WalletTabManager.queueTransition(
												new Integer(WTabManager.M_PURCHASE_TOKENS));
					m_WalletTabManager.makeTransition();
				}
				else {
					m_WalletTabManager.showError("cannot buy tokens",
														  "not logged in",
													"you must log into your wallet first");
				}
			}
		};
		JMenuItem purchItem = new JMenuItem(purchTok);
		purchItem.setMnemonic(KeyEvent.VK_B);
		purchItem.setToolTipText("Purchase tokens manually");
		purchItem.setIcon(SparkRes.getImageIcon(SparkRes.SMALL_AGENT_IMAGE));
		advMenu.add(purchItem);

		// define action to deal with any expired vouchers
		m_ExpiredAction = new AbstractAction("Expired Vouchers") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_StateData != null && m_StateData.isLoggedIn()) {
					// go to expired screen
					m_WalletTabManager.queueTransition(
														new Integer(WTabManager.M_EXPIRED));
					m_WalletTabManager.makeTransition();
				}
				else {
					m_WalletTabManager.showError("cannot process expired vouchers",
														  "not logged in",
														"please log into your wallet first");
				}
			}
		};
		JMenuItem expiredItem = new JMenuItem(m_ExpiredAction);
		expiredItem.setMnemonic(KeyEvent.VK_X);
		expiredItem.setToolTipText("Deal with any expired vouchers");
		expiredItem.setIcon(SparkRes.getImageIcon(SparkRes.RED_FLAG_16x16));
		advMenu.add(expiredItem);

		// define action to close a wallet
		m_CloseSafeAction = new AbstractAction("Close Wallet") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_StateData != null && m_StateData.isLoggedIn()) {
					// go to close wallet screen
					m_WalletTabManager.queueTransition(
													new Integer(WTabManager.M_CLOSE_SAFE1));
					m_WalletTabManager.makeTransition();
				}
				else {
					m_WalletTabManager.showError("cannot close wallet",
														  "not logged in",
													"you must log into your wallet first");
				}
			}
		};
		JMenuItem closeItem = new JMenuItem(m_CloseSafeAction);
		closeItem.setMnemonic(KeyEvent.VK_C);
		closeItem.setToolTipText("Close out your wallet");
		closeItem.setIcon(SparkRes.getImageIcon(SparkRes.BRICKWALL_IMAGE));
		advMenu.add(closeItem);

		// define action to log out of safe without disconnecting
		m_LogoutAction = new AbstractAction("Logout") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_StateData != null && m_StateData.isLoggedIn()) {
					doLogout();
					// return to login screen

					m_WalletTabManager.queueTransition(
															new Integer(WTabManager.M_LOGIN1));
					m_WalletTabManager.makeTransition();
				}
				else {
					m_WalletTabManager.showInfo("You are not logged into a wallet");
				}
			}
		};

		JMenuItem logoutItem = new JMenuItem(m_LogoutAction);
		logoutItem.setToolTipText("Log out of your wallet");
		logoutItem.setOpaque(false);
		logoutItem.setIcon(SparkRes.getImageIcon(SparkRes.LINK_DELETE_16x16));
		logoutItem.setMnemonic(KeyEvent.VK_L);
		advMenu.add(logoutItem);

		// add the Wallet (formerly Advanced) menu to Spark
		JMenuBar jmb = mainWindow.getJMenuBar();
		jmb.add(advMenu);

		// define action for tutorials help, and add to existing Help menu
		m_TutorialHelpAction = new AbstractAction("Tutorials") {
			public void actionPerformed(ActionEvent actionEvent) {
				m_WalletTabManager.queueTransition(
														new Integer(WTabManager.M_TUTORIALS));
				m_WalletTabManager.makeTransition();
			}
		};
		JMenuItem tutorialsItem = new JMenuItem(m_TutorialHelpAction);
		tutorialsItem.setMnemonic(KeyEvent.VK_T);
		tutorialsItem.setToolTipText("Go to user tutorials");
		tutorialsItem.setIcon(SparkRes.getImageIcon(
																SparkRes.SMALL_DOCUMENT_VIEW));

		// get the existing Spark Help menu
		JMenu helpMenu = mainWindow.getMenuByName("Help");
		if (helpMenu != null) {
			helpMenu.addSeparator();
			helpMenu.add(tutorialsItem);
		}
		else {
			Log.error("Could not find Spark Help menu to add Tutorials");
		}

		// go to appropriate first screen
		if (conn != null) {
			m_WalletTabManager.queueTransition(new Integer(WTabManager.M_LOGIN1));
		}
		else {
			m_WalletTabManager.queueTransition(
													new Integer(WTabManager.M_DISCONNECTED));
		}
		m_WalletTabManager.makeTransition();
	}

	// get methods
	/**
	 * get state tracking object
	 * @return state-tracking object, of type {@link VscState}
	 */
	public VscState getVscState() { return m_StateData; }

	/**
	 * get login secrets object
	 * @return login secrets object, of type {@link VsSecrets}
	 */
	public VsSecrets getLoginSecrets() { return m_LoginSecrets; }

	/**
	 * get the Smack provider manager object
	 * @return the manager
	 */
	public ProviderManager getProviderManager() { return m_ProviderManager; }

	/**
	 * get the OFS message listener object
	 * @return the listener
	 */
	public OFSListener getOFSListener() { return m_OFSListener; }

	/**
	 * get the PKS message listener object
	 * @return the listener
	 */
	public PKSListener getPKSListener() { return m_PKSListener; }

	/**
	 * get the web message listener object
	 * @return the listener
	 */
	public WebListener getWebListener() { return m_WebListener; }

	/**
	 * get the SVX listener object
	 * @return the listener
	 */
	public SVXListener getSVXListener() { return m_SVXListener; }

	/**
	 * get the SVM listener object
	 * @return the listener
	 */
	public SVMListener getSVMListener() { return m_SVMListener; }

	/**
	 * get the UI screen/tab manager for the wallet tab
	 * @return the tab manager for the plugin's Wallet tab
	 */
	public WTabManager getWalletTabManager() {
		return m_WalletTabManager;
	}

	/**
	 * get the UI screen/tab manager for the SVX tab
	 * @return the tab manager for the plugin's SVX tab
	 */
	public XTabManager getExchangeTabManager() {
		return m_ExchangeTabManager;
	}

	/**
	 * get the UI screen/tab manager for the SVM tab
	 * @return the tab manager for the plugin's SVM tab
	 */
	public MTabManager getMarketplaceTabManager() {
		return m_MarketplaceTabManager;
	}

	/**
	 * switch the tabbed pane to select the designated Component
	 * @param comp the Component (one of the TabManager subclasses) to switch to
	 */
	public void setSelectedTab(Component comp) {
		if (comp == null) {
			return;
		}
		Workspace workspace = SparkManager.getWorkspace();
		SparkTabbedPane tabbedPane = workspace.getWorkspacePane();
		int tIdx = tabbedPane.indexOfComponent(comp);
		if (tIdx == -1) {
			Log.error("No tab for this component");
			return;
		}
		tabbedPane.setSelectedIndex(tIdx);
	}

	/**
	 * get the SDS listener
	 * @return the listener, of type {@link SDSListener}
	 */
	public SDSListener getSDSListener() { return m_SDSListener; }

	/**
	 * get the DHT listener
	 * @return the listener, of type {@link DHTListener}
	 */
	public DHTListener getDHTListener() { return m_DHTListener; }

	/**
	 * get the spent tokens cache
	 * @return the cache, of type {@link SpentTokenCache}
	 */
	public SpentTokenCache getTokenCache() { return m_SpentCache; }

	/**
	 * obtain the fiat asset pricing module
	 * @return the module
	 */
	public AssetPriceCache getPricingModule() { return m_PricingModule; }

	/**
	 * get the user preferences
	 * @return the object containing user preference values
	 */
	public VSPreferences getUserPreferences() { return m_UserPrefs; }

	/**
	 * get the default message wait timeout
	 * @return the timeout, in msec
	 */
	public long getTimeout() {
		long tout = 60000L;
		if (m_UserPrefs != null) {
			tout = m_UserPrefs.getTimeout() * 1000L;
		}
		return tout;
	}

	/**
	 * get the OFS connection
	 * @return the connection, or null if none exists
	 */
	public XMPPConnection getOFSConnection() {
		return SparkManager.getConnection();
	}

	/**
	 * method to obtain the log directory
	 * @return the directory where log files etc. are created
	 */
	public File getLogDir() {
		return Spark.getLogDirectory();
	}

	/**
	 * method to obtain the user's home Spark directory
	 * @return the path to the directory where Spark-related files are stored
	 */
	public String getSparkUserHome() {
		return Spark.getSparkUserHome();
	}

	/**
	 * return test network status
	 * @return true if we're running on a test network
	 */
	public boolean isTestNetwork() { return m_TestNet; }

	/**
	 * return the number of seconds since we were initialized by Spark
	 * @return time since init in seconds, or zero if not yet initialized
	 */
	public int getTimeSinceInit() {
		if (m_InitTimestamp == null) {
			return 0;
		}
		Date now = new Date();
		int duration = (int) ((now.getTime() - m_InitTimestamp.getTime()) / 1000);
		return duration;
	}

	/**
	 * obtain the current SVM plugin configuration
	 * @return the config for the currently connected (or last connected) SVM
	 */
	public SVMPlugin getSVMPlugin() { return m_SVMPlugin; }

	/**
	 * obtain the standard web message filter
	 * @return the filter which looks for website Message broadcasts
	 */
	public PacketFilter getWebMsgFilter() { return m_WebMsgFilter; }

	/**
	 * obtain the standard SVX message filter
	 * @return the filter which looks for unsolicited SVX Message broadcasts
	 */
	public PacketFilter getSVXMsgFilter() { return m_SVXMsgFilter; }

	/**
	 * obtain the standard SVM message filter
	 * @return the filter which looks for unsolicited Message broadcasts
	 */
	public PacketFilter getSVMMsgFilter() { return m_SVMMsgFilter; }

	/**
	 * helper method to find the address of the DHT and connect to it.
	 * (tries multiple paths in sequence if they exist)
	 * @return a live connection to the DHT, or null if something failed
	 */
	public XMPPConnection openDHT() {
		// see if we already have one
		XMPPConnection retConn = m_DHTListener.getDHTConnection();
		if (retConn != null)  {
			return retConn;
		}
		ArrayList<InetSocketAddress> dhtList = m_LoginSecrets.getDHTNodes();
		if (dhtList.isEmpty()) {
			Log.error("No DHTs recorded!");
			m_WalletTabManager.showError("No DHT set!");
			return retConn;
		}

		// reduce connect times (on connect, Smack seems to use 3x reply time)
		int orgTimeout = SmackConfiguration.getPacketReplyTimeout();
		SmackConfiguration.setPacketReplyTimeout(10000);

		// take the list in order, favoring a path with the same domain as the OFS
		String dhtDomain = m_LoginSecrets.getDHTDomain();
		Iterator<InetSocketAddress> dhti = dhtList.iterator();
		InetSocketAddress usedAddr = null;
		while (dhti.hasNext()) {
			InetSocketAddress dhtAddr = dhti.next();

			// make a connection to the DHT Openfire server
			ConnectionConfiguration cc
				= new ConnectionConfiguration(dhtAddr.getHostString(),
														dhtAddr.getPort(),
														dhtDomain);
			cc.setSecurityMode(ConnectionConfiguration.SecurityMode.required);
			cc.setCompressionEnabled(true);
			cc.setSelfSignedCertificateEnabled(true);
			XMPPConnection dhtConn = new XMPPConnection(cc);

			// if we're not on the first iteration, try a pagekite 80 ping first
			boolean pinged = false;
			if (dhtList.indexOf(dhtAddr) != 0) {
				if (!port80KitePing(dhtAddr.getHostString())) {
					Log.error("Ping " + dhtAddr.getHostString() + ":80 failed");
					// presumably, connection also won't work
					continue;
				}
				pinged = true;
			}
			try {
				dhtConn.connect();
			}
			catch (XMPPException xe) {
				Log.error("Could not connect to DHT " + dhtDomain + " at "
							+ dhtAddr.getHostString() + ":" + dhtAddr.getPort());

				// attempt retry after pinging pagekite 80 port
				if (!pinged && port80KitePing(dhtAddr.getHostString())) {
					try {
						dhtConn.connect();
					}
					catch (XMPPException xe2) {
						Log.error("Also failed after :80 ping");
						continue;
					}
				}
				else {
					continue;
				}
			}
			// perform anon login
			try {
				dhtConn.loginAnonymously();
				retConn = dhtConn;
				Log.debug("Successful connect to DHT " + dhtDomain + " at path "
							+ dhtAddr.getHostString() + ":" + dhtAddr.getPort());
				usedAddr = dhtAddr;
				break;
			}
			catch (XMPPException xe) {
				Log.error("Could not login anon to DHT " + dhtDomain + " at path "
							+ dhtAddr.getHostString() + ":" + dhtAddr.getPort(), xe);
				dhtConn.disconnect();
			}
		}
		SmackConfiguration.setPacketReplyTimeout(orgTimeout);
		if (retConn == null) {
			m_WalletTabManager.showError("DHT connect failure");
			return retConn;
		}

		// record connection with the listener so it can be closed later
		m_DHTListener.setDHTConnection(retConn);

		// if needed, resort the DHT and SDS list to the connected hostname is 1st
		String connHost = retConn.getHost();
		if (dhtList.indexOf(usedAddr) != 0) {
			reorderGateways(dhtList, connHost);
			// do the same for SDS if we don't have a connection yet
			if (m_SDSListener.getSDSConnection() == null) {
				reorderGateways(m_LoginSecrets.getSDSGateways(), connHost);
			}
		}
		return retConn;
	}
	
	/**
	 * helper method to find the address of the SDS and connect to it.
	 * (tries multiple paths in sequence if they exist)
	 * @return a live connection to the SDS, or null if something failed
	 */
	public XMPPConnection openSDS() {
		// see if we already have one
		XMPPConnection retConn = m_SDSListener.getSDSConnection();
		if (retConn != null)  {
			return retConn;
		}
		ArrayList<InetSocketAddress> sdsList = m_LoginSecrets.getSDSGateways();
		if (sdsList.isEmpty()) {
			Log.error("No SDSs recorded!");
			m_WalletTabManager.showError("No SDS set!");
			return retConn;
		}

		// reduce connect times (on connect, Smack seems to use 3x reply time)
		int orgTimeout = SmackConfiguration.getPacketReplyTimeout();
		SmackConfiguration.setPacketReplyTimeout(10000);

		// take the list in order, favoring a path with the same domain as the OFS
		String sdsDomain = m_LoginSecrets.getSDSDomain();
		Iterator<InetSocketAddress> sdsi = sdsList.iterator();
		InetSocketAddress usedAddr = null;
		while (sdsi.hasNext()) {
			InetSocketAddress sdsAddr = sdsi.next();

			// make a connection to the SDS Openfire server
			ConnectionConfiguration cc
				= new ConnectionConfiguration(sdsAddr.getHostString(),
														sdsAddr.getPort(),
														sdsDomain);
			cc.setSecurityMode(ConnectionConfiguration.SecurityMode.required);
			cc.setCompressionEnabled(true);
			cc.setSelfSignedCertificateEnabled(true);
			XMPPConnection sdsConn = new XMPPConnection(cc);

			// if we're not on the first iteration, try a pagekite 80 ping first
			boolean pinged = false;
			if (sdsList.indexOf(sdsAddr) != 0) {
				if (!port80KitePing(sdsAddr.getHostString())) {
					Log.error("Ping " + sdsAddr.getHostString() + ":80 failed");
					// presumably, connection also won't work
					continue;
				}
				pinged = true;
			}
			try {
				sdsConn.connect();
			}
			catch (XMPPException xe) {
				Log.error("Could not connect to SDS " + sdsDomain + " at "
							+ sdsAddr.getHostString() + ":" + sdsAddr.getPort());

				// attempt retry after pinging pagekite 80 port
				if (!pinged && port80KitePing(sdsAddr.getHostString())) {
					try {
						sdsConn.connect();
					}
					catch (XMPPException xe2) {
						Log.error("Also failed after :80 ping");
						continue;
					}
				}
				else {
					continue;
				}
			}
			// perform anon login
			try {
				sdsConn.loginAnonymously();
				retConn = sdsConn;
				Log.debug("Successful connect to SDS " + sdsDomain + " at path "
							+ sdsAddr.getHostString() + ":" + sdsAddr.getPort());
				usedAddr = sdsAddr;
				break;
			}
			catch (XMPPException xe) {
				Log.error("Could not login anon to SDS " + sdsDomain + " at path "
							+ sdsAddr.getHostString() + ":" + sdsAddr.getPort(), xe);
				sdsConn.disconnect();
			}
		}
		SmackConfiguration.setPacketReplyTimeout(orgTimeout);
		if (retConn == null) {
			m_WalletTabManager.showError("SDS connect failure");
			return retConn;
		}

		// record connection with the listener so it can be closed later
		m_SDSListener.setSDSConnection(retConn);

		// if needed, resort the SDS and DHT list to the connected hostname is 1st
		String connHost = retConn.getHost();
		if (sdsList.indexOf(usedAddr) != 0) {
			reorderGateways(sdsList, connHost);
			// do the same for DHT if we don't have a connection yet
			if (m_DHTListener.getDHTConnection() == null) {
				reorderGateways(m_LoginSecrets.getDHTNodes(), connHost);
			}
		}
		return retConn;
	}

	/**
	 * utility method to log out of a wallet (assumes login process completed)
	 */
	public void doLogout() {
		SparkTabbedPane tPane = SparkManager.getWorkspace().getWorkspacePane();
		tPane.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		m_StateData.setLoggedIn(false);
		m_SpentCache.purgeSpentTokens();
		m_DHTListener.closeDHTConnection();
		m_DHTListener.getDHTrecs().clear();
		m_SDSListener.closeSDSConnection();
		m_PKSListener.closePKSConnection();
		tPane.setCursor(null);

		// remove any popup windows we might have
		for (JFrame popup : m_PopupList) {
			popup.dispose();
		}
		m_PopupList.clear();

		// reset states on marketplace and exchange tabs
		m_ExchangeTabManager.queueTransition(
												new Integer(XTabManager.M_LOGOUT_EXCHANGE));
		m_ExchangeTabManager.makeTransition();
		m_MarketplaceTabManager.queueTransition(
												new Integer(MTabManager.M_LOGOUT_MARKET));
		m_MarketplaceTabManager.makeTransition();
		resetSecrets();
		// forget the VPs asset/unit/token configuration
		VPAssetConfig.m_AssetConfig.clear();

		// set the main tab state to wallet tab manager
		setSelectedTab(m_WalletTabManager);
		System.gc();
	}

	/**
	 * register a new popup window, for auto-closure during logout
	 * @param win the popup window to be recorded
	 */
	public void registerPopup(JFrame win) {
		if (win != null) {
			m_PopupList.add(win);
		}
	}

	/**
	 * remove a popup window, in the event it was already disposed of
	 * @param win the popup window to be deregistered
	 */
	public void deregisterPopup(JFrame win) {
		if (win != null) {
			m_PopupList.remove(win);
		}
	}

	/**
	 * set test network status
	 * @param stat true if we're running on a test network
	 */
	public void setTestNetwork(boolean stat) { m_TestNet = stat; }

	/**
	 * helper method to get status bar, only if we have one
	 * @return the status bar, or null if we're not in a state where one exists
	 */
	public JPanel getStatusBar() {
		if (m_WalletTabManager != null) {
			HomePane hp = m_WalletTabManager.getHomePane();
			if (hp != null) {
				return hp.getStatusBar();
			}
		}
		return null;
	}

	/**
	 * set the default status bar contents, based on connection status
	 */
	public void setDefaultStatusBar() {
		String connectTxt = null;
		int strut = 0;
		final XMPPConnection conn = getOFSConnection();
		if (conn != null && conn.isConnected()) {
			connectTxt = " Connected to OFS gateway - Ready";
			strut = 260;
		}
		else {
			connectTxt = " Not connected";
			strut = 400;
		}
		final JLabel conLab = new JLabel(connectTxt);
		conLab.setFont(m_OFSConnectListener.M_LabFont);
		final Component strutBox = Box.createHorizontalStrut(strut);
		// get the status bar from the HomePane, and put this label in it
		final JPanel statBar = getStatusBar();
		if (statBar != null) {
			statBar.removeAll();
			statBar.add(conLab);
			statBar.add(strutBox);
			statBar.invalidate();
			statBar.revalidate();
			statBar.repaint();
		}
	}

	/**
	 * method to attempt a connect to a PageKite'd XMPP host prior to connection
	 * @param xmppHost the hostname to connect to
	 * @return true on success
	 */
	public boolean port80KitePing(String xmppHost) {
		/* The following is a hack for supporting OFS hosts which are PageKites.
		 * A PageKite requires an open HTTP connection from the same incoming IP
		 * in order to proxy an encrypted connection across a raw port.  We
		 * therefore open '/' on the host first, before initiating XMPP connect.
		 * However, in the event that the OFShost isn't a PageKite and doesn't
		 * have anything listening on port 80, we will not let a failure abort
		 * our connect attempt.
		 */
		boolean ret = false;
		try {
			URL pkURL = new URL("http", xmppHost, 80, "/");
			URLConnection pkConn = pkURL.openConnection();
			HttpURLConnection preConn = (HttpURLConnection) pkConn;
			preConn.setRequestMethod("GET");
			preConn.setAllowUserInteraction(false);
			preConn.setDoOutput(true);
			preConn.setConnectTimeout(10000);
			preConn.setReadTimeout(10000);
			preConn.connect();
			Object junk = preConn.getContent();
			preConn.disconnect();
			ret = true;
		}
		catch (MalformedURLException mue) {
			Log.error("Could not construct URL for pre-connect step", mue);
		}
		catch (ProtocolException pe) {
			Log.error("Could not connect to URL for pre-connect step", pe);
		}
		catch (IOException ioe) {
			Log.error("Could not connect to URL for pre-connect step");
		}
		return ret;
	}

	/**
	 * method to reorder the gateways on a list, based on which should be first
	 * @param gtwyList the list of gateways to be sorted
	 * @param firstHost the hostname which should become first on the list
	 */
	public void reorderGateways(ArrayList<InetSocketAddress> gtwyList,
										 String firstHost)
	{
		if (gtwyList == null || gtwyList.isEmpty()) {
			return;
		}
		
		// find the indicated host
		InetSocketAddress firstEntry = null;
		for (InetSocketAddress isa : gtwyList) {
			if (isa.getHostString().equals(firstHost)) {
				firstEntry = isa;
				break;
			}
		}
		if (firstEntry != null) {
			// remove from list
			gtwyList.remove(firstEntry);
			// add it back at index 0
			gtwyList.add(0, firstEntry);
		}
		else {
			Log.error("Host " + firstHost + " not found on gtwyList");
		}
	}

	/**
	 * reset the SVM listener
	 * @param newListener the replacement (override) listener to use
	 */
	public void setSVMListener(SVMListener newListener) {
		final XMPPConnection conn = getOFSConnection();
		// remove any packet listener which went to the old one
		if (conn != null && m_SVMListener != null) {
			conn.removePacketListener(m_SVMListener);
		}

		// if null, go back to the default
		if (newListener == null) {
			m_SVMListener = new SVMListener(this);
		}
		else {
			// use what was passed
			m_SVMListener = newListener;
		}

		// add packet listener pointing at the new one
		if (conn != null) {
			conn.addPacketListener(m_SVMListener, m_SVMMsgFilter);
		}
	}

	/**
	 * reset the SVM tab manager
	 * @param tabManager the new tab manager to set
	 */
	public void setSVMTabManager(MTabManager tabManager) {
		if (tabManager != null) {
			Workspace workspace = SparkManager.getWorkspace();
			SparkTabbedPane tabbedPane = workspace.getWorkspacePane();
			// remove the old one
			int tIdx = tabbedPane.indexOfComponent(m_MarketplaceTabManager);
			if (tIdx == -1) {
				Log.error("No tab for old Marketplace tab manager");
				return;
			}
			tabbedPane.removeTabAt(tIdx);
			// now add the new one
			m_MarketplaceTabManager = tabManager;
			tabbedPane.addTab("Marketplaces", M_SVIcon.getIcon(),
									m_MarketplaceTabManager,
									"SilentVault Market operations");
			// select it
			tIdx = tabbedPane.indexOfComponent(m_MarketplaceTabManager);
			tabbedPane.setSelectedIndex(tIdx);
		}
	}

	/**
	 * set the SVM plugin configuration
	 * @param svmPlug the new plugin to use
	 */
	public void setSVMPlugin(SVMPlugin svmPlug) {
		if (svmPlug != null) {
			m_SVMPlugin = svmPlug;
		}
	}

	// END methods
}
