/*
 * last modified---
 * 	08-06-14/jet/ Spark plugin version
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	implement a XML payment confirmation object
 *
 * notes---
 * 	This class is closely related to the corresponding class
 * 	org.vouchersafe.publisher.XMLConfirmPayment.
 */

package org.silentvault.spark;

import org.dom4j.DocumentException;
import org.dom4j.Element;

import org.jivesoftware.smack.util.StringUtils;

import org.jivesoftware.spark.util.log.Log;

import java.util.Vector;
import java.util.Locale;
import java.security.PublicKey;
import java.security.PrivateKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.SignatureException;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.ParseException;


/**
 * Class to instantiate a payment confirmation as an XML object.  The class
 * knows how to store, parse, and emit confirmation details in XML.  It is
 * used when a voucher payment has been effected, and the VP needs to transmit
 * the payment confirmation details back to the payer (via the OFS).  This could
 * be done via a simple reply (REP_payment_sent), but in case that reply cannot
 * be sent, it is stored on the DHT for the payer to pick up subsequently.
 *
 * @author Kevin Wilkerson
 */
public final class XMLConfirmPayment implements Cloneable {
	// BEGIN data members
	
	/**
	 * names of all the elements that are part of a payment confirmation
	 */
	private static Vector<String>	M_ConfirmElements = getConfirmElements();

	/**
	 * the quantity paid to the payee, in metal weight
	 */
	private Double						m_Quantity;

	/**
	 * the units of the quantity (will always be voucher base units of the asset)
	 */
	private String						m_Units;

	/**
	 * the backing asset of the payment
	 */
	private String						m_Asset;

	/**
	 * the Issuer whose vouchers are involved
	 */
	private String						m_Issuer;

	/**
	 * the voucher safe number of the payee (the counterparty)
	 */
	private String						m_PayeeVS;

	/**
	 * the serial number of the payment voucher issued to the payee
	 */
	private String						m_VoucherSerial;

	/**
	 * the TTL expiration time (deadline for payee pickup), in msec
	 */
	private long						m_TTL;

	/**
	 * the voucher issued to the payer for their change. (it is conceivable that
	 * this will be blank, if the amount of the payment exactly equalled the
	 * the input voucher(s))
	 */
	private XMLVoucher				m_ChangeVoucher;

	/**
	 * the private key signature of the VP on the above elements
	 */
	private String						m_Signature;

	/**
	 * whether the payment has been fully parsed/initialized
	 */
	private boolean					m_Initialized;

	/**
	 * the id at which this record is stored on the DHT, if applicable
	 */
	private Integer					m_DHTid;

	// END data members

	// BEGIN methods

	/**
	 * initialization method to set the names of all elements in a payment
	 * @return a Vector of all possible element names
	 */
	private static Vector<String> getConfirmElements() {
		Vector<String> vElts = new Vector<String>(10);
		vElts.add(new String("confirmDetails"));
		vElts.add(new String("quantity"));
		vElts.add(new String("units"));
		vElts.add(new String("asset"));
		vElts.add(new String("issuer"));
		vElts.add(new String("payee_VS"));
		vElts.add(new String("voucher_serial"));
		vElts.add(new String("ttl"));
		vElts.add(new String("voucher"));
		vElts.add(new String("signature"));
		return vElts;
	}

	/**
	 * default constructor
	 */
	public XMLConfirmPayment() {
		m_Quantity = new Double(0.0);
		m_PayeeVS = m_Signature = m_Units = m_Asset = m_VoucherSerial = "";
		m_Issuer = "";
	}


	// get accessor methods
	/**
	 * get method for payment quantity (metal weight)
	 * @return the quantity of the payment made
	 */
	public double getQuantity() { return m_Quantity.doubleValue(); }

	/**
	 * get method for payment units
	 * @return the units of the quantity
	 */
	public String getUnits() { return m_Units; }

	/**
	 * get method for the backing asset of the payment
	 * @return the asset type
	 */
	public String getAsset() { return m_Asset; }

	/**
	 * get method for voucher Issuer
	 * @return the Issuer
	 */
	public String getIssuer() { return m_Issuer; }

	/**
	 * get method for payee VS#
	 * @return the VS#, an alpha-numeric string
	 */
	public String getPayee() { return m_PayeeVS; }

	/**
	 * get method for the serial number of the voucher issued to the payee
	 * @return the serial number, an alphanumeric string
	 */
	public String getVoucherSerial() { return m_VoucherSerial; }

	/**
	 * get method for the TTL (expiration)
	 * @return the time to live, in milliseconds
	 */
	public long getTTL() { return m_TTL; }

	/**
	 * get method for the payment change voucher
	 * @return the change voucher which was issued to the payer, or null if
	 * there wasn't one
	 */
	public XMLVoucher getVoucher() { return m_ChangeVoucher; }

	/**
	 * get method for VP's private key signature on the data
	 * @return the signature, base64-encoded
	 */
	public String getSignature() { return m_Signature; }

	/**
	 * obtain the initialization status of the payment
	 * @return true if all fields are set, false otherwise
	 */
	public boolean isInit() { return m_Initialized; }

	/**
	 * obtain the id of the record where this item is stored on the DHT
	 * @return the id, or null if not applicable or unknown
	 */
	public Integer getDHTid() { return m_DHTid; }

	/**
	 * get method for the display units which should be shown on various screens
	 * @return the units, which may be different from m_Units
	 */
	public String getDisplayUnits() {
		return XMLVoucher.getDisplayUnits(m_Asset);
	}


	/**
	 * method to construct a pending payment from an input XML representation
	 * (normally called by a VSC after retrieving/decrypting one off the DHT)
	 * @param xmlRep the XML representation to be parsed
	 * @return the payment object
	 * @throws DocumentException when the input won't parse
	 */
	public static XMLConfirmPayment getConfirmRep(String xmlRep)
		throws DocumentException
	{
		// parse the input string to see if it's legal XML
		Element pmtElt = EncodingUtils.buildElement(xmlRep);
		if (pmtElt == null) {
			throw new DocumentException("payment confirmation parse error");
		}
		return XMLConfirmPayment.getConfirmRep(pmtElt);
	}

	/**
	 * method to construct a pending payment from a dom4j Element representation
	 * @param pmtElt the Element representing the payment
	 * @return the payment object
	 * @throws DocumentException when the input is incomplete or improper
	 */
	public static XMLConfirmPayment getConfirmRep(Element pmtElt)
		throws DocumentException
	{
		if (pmtElt == null) {
			throw new DocumentException("Null input");
		}

		// prepare output
		XMLConfirmPayment cpmt = new XMLConfirmPayment();

		// must have <quantity/>
		Element quanElt = pmtElt.element("quantity");
		if (quanElt == null) {
			throw new DocumentException("Missing <quantity/> element in "
												+ pmtElt.asXML());
		}
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		Number quan = null;
		try {
			quan = df.parse(quanElt.getText());
			cpmt.m_Quantity = new Double(quan.doubleValue());
		}
		catch (ParseException pe) {
			throw new DocumentException("Invalid <quantity/>", pe);
		}

		// must have <units/>
		Element unitElt = pmtElt.element("units");
		if (unitElt == null) {
			throw new DocumentException("Missing <units/> element in "
												+ pmtElt.asXML());
		}
		cpmt.m_Units = unitElt.getTextTrim();

		// must have <asset/>
		Element assetElt = pmtElt.element("asset");
		if (assetElt == null) {
			throw new DocumentException("Missing <asset/> element in "
												+ pmtElt.asXML());
		}
		cpmt.m_Asset = assetElt.getTextTrim();

		// must have <issuer/>
		Element issElt = pmtElt.element("issuer");
		if (issElt == null) {
			throw new DocumentException("Missing <issuer/> element in "
												+ pmtElt.asXML());
		}
		cpmt.m_Issuer = issElt.getTextTrim();

		// must have a <payee_VS/>
		Element payeeElt = pmtElt.element("payee_VS");
		if (payeeElt == null) {
			throw new DocumentException("Missing <payee_VS/> element in "
												+ pmtElt.asXML());
		}
		cpmt.m_PayeeVS = payeeElt.getTextTrim();

		// must have <voucher_serial/>
		Element serialElt = pmtElt.element("voucher_serial");
		if (serialElt == null) {
			throw new DocumentException("Missing <voucher_serial/> element in "
												+ pmtElt.asXML());
		}
		cpmt.m_VoucherSerial = serialElt.getTextTrim();

		// must have <ttl/>
		Element ttlElt = pmtElt.element("ttl");
		if (ttlElt == null) {
			throw new DocumentException("Missing <ttl/> element in "
												+ pmtElt.asXML());
		}
		Long ttl = null;
		try {
			ttl = Long.parseLong(ttlElt.getText());
			if (ttl.longValue() >= 0L) {
				cpmt.m_TTL = ttl.longValue();
			}
			else {
				Log.error("Bad TTL value in XMLConfirmPayment: \"" + pmtElt.asXML()
							+ "\"");
				throw new DocumentException("Illegal or missing TTL");
			}
		}
		catch (NumberFormatException nfe) {
			throw new DocumentException("Bad TTL specified", nfe);
		}

		// must have a <voucher/>
		Element vouchElt = pmtElt.element("voucher");
		if (vouchElt == null) {
			throw new DocumentException("Missing <voucher/> element in "
												+ pmtElt.asXML());
		}
		// NB: this can be blank, if no change voucher existed
		if (!vouchElt.asXML().equals("<voucher/>")) {
			cpmt.m_ChangeVoucher = XMLVoucher.getVoucherRep(vouchElt);
		}

		// must have <signature/>
		Element sigElt = pmtElt.element("signature");
		if (sigElt == null) {
			throw new DocumentException("Missing <signature/> element in "
												+ pmtElt.asXML());
		}
		cpmt.m_Signature = sigElt.getTextTrim();
		if (cpmt.m_Signature.isEmpty()) {
			throw new DocumentException("Empty <signature/> element in "
												+ pmtElt.asXML());
		}
		// NB: we don't attempt to validate the sig here (see ::validateSig())

		cpmt.m_Initialized = true;
		return cpmt;
	}

	/**
	 * private utility method to build the data to be signed/validated
	 * (used for both signing and validation to guarantee consistency)
	 * @return the XML details to be signed or validated
	 */
	private String buildSignedDetails() {
		// must be initialized first
		if (!m_Initialized) {
			Log.error("Cannot build signature details on uninitialized details");
			return "";
		}

		// build string, equal to all data fields
		StringBuilder sigData = new StringBuilder(512);
		sigData.append("<quantity>" + m_Quantity + "</quantity>");
		sigData.append("<units>" + m_Units + "</units>");
		sigData.append("<asset>" + m_Asset + "</asset>");
		sigData.append("<issuer>" + m_Issuer + "</issuer>");
		sigData.append("<payee_VS>" + m_PayeeVS + "</payee_VS>");
		sigData.append("<voucher_serial>" + m_VoucherSerial);
		sigData.append("</voucher_serial>");
		sigData.append("<ttl>" + m_TTL + "</ttl>");
		if (m_ChangeVoucher != null) {
			sigData.append(m_ChangeVoucher.toXML());
		}
		else {
			sigData.append("<voucher/>");
		}
		return sigData.toString();
	}

	/**
	 * method to validate the signature on the data in the object
	 * @param vpPub the public key of the VP, used to validate signature
	 * @return true if all is well, false if a problem is encountered
	 */
	public boolean validateSig(PublicKey vpPub) {

		// get data to validate
		String sigData = buildSignedDetails();
		Signature dSig = null;
		boolean ok = false;
		try {
			dSig = Signature.getInstance("SHA1withRSA");
			dSig.initVerify(vpPub);
			dSig.update(sigData.getBytes());
			ok = dSig.verify(StringUtils.decodeBase64(m_Signature));
		}
		catch (NoSuchAlgorithmException nsae) {
			Log.error("No such sig verify algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid pubkey for signature validation", ike);
		}
		catch (SignatureException se) {
			Log.error("Unable to validate details signature", se);
		}
		if (!ok) {
			Log.error("Could not verify signature on payment");
			return false;
		}
		return true;
	}

	/**
	 * method to emit the payment confirmation's values as XML
	 * @return valid XML representation of confirmation (assuming all values
	 * set), or the null string if not
	 */
	public String toXML() {
		if (!m_Initialized) {
			Log.error("Payment confirmation is not properly initialized!");
			return "";
		}

		StringBuilder xml = new StringBuilder(1024);
		xml.append("<confirmDetails>");
		xml.append(buildSignedDetails());
		xml.append("<signature>" + m_Signature + "</signature>");
		xml.append("</confirmDetails>");
		return xml.toString();
	}


	// set methods
	/**
	 * set the change voucher the payment, which goes in the payer's safe
	 * @param voucher the voucher (must have valid signature)
	 * @param vpKey the public key of the VP, to validate the voucher signature
	 * @return true on success, false if voucher doesn't check out
	 */
	public boolean setChangeVoucher(XMLVoucher voucher, PublicKey vpKey) {
		if (voucher == null || vpKey == null) {
			return false;
		}
		if (voucher.validateSig(vpKey)) {
			m_ChangeVoucher = voucher;
			return true;
		}
		Log.error("Bad voucher signature; cannot use in payment confirmation");
		return false;
	}

	/**
	 * method to set the payee voucher safe number
	 * @param vsnum the VS# receiving the payment
	 */
	public void setPayee(String vsnum) {
		if (vsnum != null && !vsnum.isEmpty()) {
			m_PayeeVS = new String(vsnum);
		}
	}

	/**
	 * method to set the TTL (time to live)
	 * @param ttl the time in seconds since the epoch when payment expires
	 */
	public void setTTL(long ttl) {
		m_TTL = ttl;
	}

	/**
	 * method to set the payment quantity
	 * @param quan the quantity (metal weight)
	 */
	public void setQuantity(double quan) {
		if (quan > 0.0) {
			m_Quantity = new Double(quan);
		}
	}

	/**
	 * method to set the payment units
	 * @param units the units of the payment
	 */
	public void setUnits(String units) {
		if (units != null && !units.isEmpty()) {
			m_Units = new String(units);
		}
	}

	/**
	 * method to set the backing asset
	 * @param asset the asset type backing the payment
	 */
	public void setAsset(String asset) {
		if (asset != null && !asset.isEmpty()) {
			m_Asset = new String(asset);
		}
	}

	/**
	 * method to set the Issuer
	 * @param issuer the Issuer who holds the backing assets
	 */
	public void setIssuer(String issuer) {
		if (issuer != null && !issuer.isEmpty()) {
			m_Issuer = new String(issuer);
		}
	}

	/**
	 * method to set the voucher serial number (of the payee's voucher)
	 * @param serial the payment voucher serial number
	 */
	public void setVoucherSerial(String serial) {
		if (serial != null && !serial.isEmpty()) {
			m_VoucherSerial = new String(serial);
		}
	}

	/**
	 * method to finalize initialzation -- used if manually building a payment
	 * using the constructor, as opposed to starting with an XML representation
	 */
	public void setInit() { m_Initialized = true; }

	/**
	 * method to set the DHT record id
	 * @param id the record id at which this item is stored on the DHT
	 */
	public void setDHTid(Integer id) {
		if (id != null) {
			m_DHTid = id;
		}
	}


	/**
	 * method to encrypt the object's XML representation, e.g. to place it into
	 * a message to the DHT.  This method encrypts the whole object, but changes
	 * none of its fields.
	 * @param pkey the public key to use to do the encryption (the payee's)
	 * @return the encrypted string, base64-encoded, or the empty string on error
	 */
	public String getEncryptedBase64(PublicKey pkey) {
		if (pkey == null) {
			return "";
		}
		
		// get the XML representation -- will come back "" if not initialized
		String xmlRep = toXML();
		if (xmlRep.isEmpty()) {
			return xmlRep;
		}

		// perform the encryption and base64 encoding -- will return null on error
		String encRep = EncodingUtils.makeBase64PubkeyEncStr(xmlRep, pkey);
		if (encRep == null) {
			return "";
		}
		return encRep;
	}

	/**
	 * method to support the Cloneable interface
	 * @return a new payment object equal to this one
	 */
	@Override
	public Object clone() {
		try {
			XMLConfirmPayment nPayment = (XMLConfirmPayment) super.clone();
			nPayment.m_Quantity = new Double(m_Quantity.doubleValue());
			nPayment.m_Units = new String(m_Units);
			nPayment.m_Asset = new String(m_Asset);
			nPayment.m_Issuer = new String(m_Issuer);
			nPayment.m_PayeeVS = new String(m_PayeeVS);
			nPayment.m_VoucherSerial = new String(m_VoucherSerial);
			if (m_ChangeVoucher != null) {
				nPayment.m_ChangeVoucher = (XMLVoucher) m_ChangeVoucher.clone();
			}
			nPayment.m_Signature = new String(m_Signature);
			return nPayment;
		}
		catch (CloneNotSupportedException cnse) {
			throw new InternalError(cnse.toString());
		}
	}

	/**
	 * method to override Object::equals()
	 * @param obj the object to compare with
	 * @return true if this object is the same, false if not
	 */
	@Override
	public boolean equals(Object obj) {
		// if it's not a payment confirmation, cannot be equal
		if (!(obj instanceof XMLConfirmPayment)) {
			return false;
		}
		XMLConfirmPayment pmt = (XMLConfirmPayment) obj;

		// a confirm is equal to this one if its payee, quantity, units
		// and TTL are the same

		if (m_Quantity.compareTo(pmt.getQuantity()) != 0) {
			return false;
		}
		if (!m_PayeeVS.equals(pmt.getPayee())) {
			return false;
		}
		if (!m_Units.equals(pmt.getUnits())) {
			return false;
		}
		if (!m_Issuer.equals(pmt.getIssuer())) {
			return false;
		}
		if (m_TTL != pmt.getTTL()) {
			return false;
		}
		return true;
	}

	/**
	 * method to override Object::hashCode()
	 * @return the hash code
	 */
	@Override
	public int hashCode() {
		int hc = 0;

		// sum the hash codes for the elements referenced by equals() above
		hc += m_Quantity.hashCode();
		hc += m_PayeeVS.hashCode();
		hc += m_Units.hashCode();
		hc += m_Issuer.hashCode();
		Long ttl = new Long(m_TTL);
		hc += ttl.hashCode();
		return hc;
	}

	// END methods
}
