/*
 * last modified---
 * 	08-31-18/kbw/ add compareTo() for possible future use
 * 	02-18-17/kbw/ allow multiple values for tokens in same asset/units, so
 * 					  long as signature check passes
 * 	07-11-15/jet/ use VPAssetConfig methods in preference to local VectorS
 * 	08-06-14/jet/ Spark plugin version
 * 	03-12-14/jet/ add get25TokensWorth()
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	implement a XML token object
 *
 * notes---
 * 	If this class is changed, org.vouchersafe.publisher.XMLToken must also
 * 	be changed.  This class also appears in other Openfire plugins.
 */

package org.silentvault.spark;

import org.dom4j.DocumentException;
import org.dom4j.Element;

import org.jivesoftware.smack.util.StringUtils;

import org.jivesoftware.spark.util.log.Log;

import java.util.Vector;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.Locale;
import java.security.PublicKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.SignatureException;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.ParseException;


/**
 * Class to instantiate a token as an XML object.  The class knows how to
 * store, parse, and emit a token in XML.  This same class is utilized in
 * both Openfire plugins and the VSC.
 *
 * @author Kevin Wilkerson
 */
public final class XMLToken implements Cloneable, Comparable {
	// BEGIN data members
	/**
	 * precision to use on token quantity
	 */
	public static final int		M_Precision = 4;
	
	/**
	 * names of all the elements that are part of a token
	 */
	private static Vector<String>	m_TokenElements = getTokenElements();

	/**
	 * the quantity of the underlying asset represented by the token
	 */
	private Double						m_Quantity;

	/**
	 * the measurement units of the token
	 */
	private String						m_Units;

	/**
	 * the underlying asset backing the token
	 */
	private String						m_Asset;

	/**
	 * the Issuer of this token, custodian of the backing asset
	 */
	private String						m_Issuer;

	/**
	 * the token serial number
	 */
	private String						m_SerialNumber;

	/**
	 * the signature of the voucher/token publisher on the above detail fields
	 */
	private String						m_Signature;

	/**
	 * whether token has been fully parsed/initialized
	 */
	private boolean					m_Initialized;

	// END data members

	// BEGIN methods

	/**
	 * initialization method to set the names of all elements in a token
	 * @return a Vector of all possible element names
	 */
	private static Vector<String> getTokenElements() {
		Vector<String> tElts = new Vector<String>(8);
		tElts.add(new String("token"));
		tElts.add(new String("details"));
		tElts.add(new String("quantity"));
		tElts.add(new String("units"));
		tElts.add(new String("asset"));
		tElts.add(new String("issuer")); 
		tElts.add(new String("serial"));
		tElts.add(new String("signature"));
		return tElts;
	}

	/**
	 * basic constructor
	 */
	public XMLToken() {
		m_Quantity = new Double(0.0);
		m_Units = m_Asset = m_Issuer = m_SerialNumber = m_Signature = "";
	}

	// get accessor methods
	/**
	 * get method for token quantity value
	 * @return the quantity, as a double precision object
	 */
	public double getQuantity() { return m_Quantity.doubleValue(); }

	/**
	 * get method for token units
	 * @return the units in which token is denominated
	 */
	public String getUnits() { return m_Units; }

	/**
	 * get method for token asset type
	 * @return the type of asset which token represents
	 */
	public String getAsset() { return m_Asset; }

	/**
	 * get method for the value of one token in a particular asset type
	 * @param asset the base asset the token represents
	 * @param units the units of this asset in which the token is denominated
	 * @return the value of one token of that type, zero if not found
	 */
	public static double getValueOfOneToken(String asset, String units) {
		// use equivalent method in VPAssetConfig
		return VPAssetConfig.getValueOfOneToken(asset, units);
	}

	/**
	 * get method for the value of 25 tokens worth of a particular asset
	 * @param asset the base asset the token represents
	 * @return the value of 25 tokens of that type in non-base units, zero if
	 * not found
	 */
	public static double get25TokensWorth(String asset) {
		// use equivalent method in VPAssetConfig
		return VPAssetConfig.get25TokensWorth(asset);
	}

	/**
	 * get method for token's ultimate Issuer
	 * @return the name of the Issuer
	 */
	public String getIssuer() { return m_Issuer; }

	/**
	 * get method for token's serial number
	 * @return the serial number, an alpha-numeric string
	 */
	public String getSerialNumber() { return m_SerialNumber; }

	/**
	 * get method for token signature data
	 * @return the signature, base64-encoded
	 */
	public String getSignature() { return m_Signature; }

	/**
	 * obtain the initialization status of the token
	 * @return true if all fields are set, false otherwise
	 */
	public boolean isInit() { return m_Initialized; }

	/**
	 * method to construct a token from an input XML representation
	 * @param xmlRep the XML representation to be parsed
	 * @param lenient whether to allow parsing without unit and value checks
	 * @return the token object
	 * @throws DocumentException when the input won't parse
	 */
	public static XMLToken getTokenRep(String xmlRep, boolean lenient)
		throws DocumentException
	{
		// parse the input string to see if it's legal XML
		Element tokElt = EncodingUtils.buildElement(xmlRep);
		if (tokElt == null) {
			throw new DocumentException("token parse error");
		}
		return XMLToken.getTokenRep(tokElt, lenient);
	}

	/**
	 * method to construct a token from a dom4j Element representation
	 * @param tokElt the Element representing the token
	 * @param lenient whether to allow parsing without unit and value checks
	 * @return the token object
	 * @throws DocumentException when the input is incomplete or improper
	 */
	public static XMLToken getTokenRep(Element tokElt, boolean lenient)
		throws DocumentException
	{
		// prepare output
		XMLToken token = new XMLToken();

		// Make sure each individual field looks sensible.  Under the root
		// element, we should have <details/> and <signature/>.

		Element detailsElt = tokElt.element("details");
		if (detailsElt == null) {
			throw new DocumentException("Missing <details/> element in "
												+ tokElt.asXML());
		}
		Element sigElt = tokElt.element("signature");
		if (sigElt == null) {
			throw new DocumentException("Missing <signature/> element in "
												+ tokElt.asXML());
		}
		String sig = sigElt.getTextTrim();
		if (sig.isEmpty()) {
			throw new DocumentException("Empty <signature/> element in "
												+ tokElt.asXML());
		}
		token.m_Signature = sig;
		// NB: we don't attempt to validate the signature here.

		// the remaining elements will all be inside the <details/> block
		Iterator eltIt = detailsElt.elementIterator();
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		while (eltIt.hasNext()) {

			// make sure this element is one of those we are expecting
			Element thisElt = (Element) eltIt.next();
			String eltName = thisElt.getName().toLowerCase();
			if (!m_TokenElements.contains(eltName)) {
				throw new DocumentException("Illegal token element: " + eltName);
			}

			// check for each item it can be and handle appropriately
			if (eltName.equals("quantity")) {
				Double quan = null;
				try {
					Number val = df.parse(thisElt.getText());
					quan = new Double(val.doubleValue());
				}
				catch (ParseException pe) {
					Log.error("Token quantity did not parse", pe);
					throw new DocumentException("Unparseable token quantity, "
														+ thisElt.getText());
				}
				if (quan != null) {
					token.m_Quantity = quan;
				}
			}
			else if (eltName.equals("units")) {
				String unitNm = thisElt.getTextTrim().toUpperCase();
				if (!lenient) {
					ArrayList<String> allTokUnits = VPAssetConfig.getAllTokenUnits();
					if (allTokUnits.contains(unitNm)) {
						token.m_Units = unitNm;
					}
					else {
						throw new DocumentException("Illegal token units value: "
															+ unitNm);
					}
				}
				else {
					token.m_Units = unitNm;
				}
			} 
			else if (eltName.equals("asset")) {
				String assetNm = thisElt.getTextTrim().toUpperCase();
				if (!lenient) {
					ArrayList<String> allAssets = VPAssetConfig.getAllAssets();
					if (allAssets.contains(assetNm)) {
						token.m_Asset = assetNm;
					}
					else {
						throw new DocumentException("Illegal token asset value: "
															+ assetNm);
					}
				}
				else {
					token.m_Asset = assetNm;
				}
			}
			else if (eltName.equals("issuer")) {
				String issuerNm = thisElt.getTextTrim().toUpperCase();
				token.m_Issuer = issuerNm;
			}
			else if (eltName.equals("serial")) {
				String serialNum = thisElt.getTextTrim();
				if (!serialNum.isEmpty()) {
					token.m_SerialNumber = serialNum;
				}
				else {
					throw new DocumentException("Empty token serial number");
				}
			}
			else {
				// shouldn't happen, but throw exception anyway
				throw new DocumentException("Unrecognized token element, "
													+ eltName);
			}
		}

		// see if the token looks complete now
		if (token.m_Units.isEmpty()) {
			throw new DocumentException("Missing token units");
		}
		if (token.m_Asset.isEmpty()) {
			throw new DocumentException("Missing token asset type");
		}
		if (token.m_Issuer.isEmpty()) {
			throw new DocumentException("Missing token issuer");
		}
		if (token.m_SerialNumber.isEmpty()) {
			throw new DocumentException("Missing token serial number");
		}

		// check that value and asset/units are consistent with each other
		double val = token.m_Quantity.doubleValue();
		if (val == 0.0) {
			throw new DocumentException("Missing token quantity");
		}
		double def = getValueOfOneToken(token.m_Asset, token.m_Units);
		if (val != def) {
			/* This can happen if we don't have any definitions yet and are
			 * parsing the spentTokens.xml file from a previous session.
			if (!VPAssetConfig.m_AssetConfig.isEmpty()) {
				Log.warning("Token value " + df.format(val)
								+ " inconsistent with definition of "
								+ def + " for " + token.m_Asset + ":" + token.m_Units);
			}
			 */
		}
		token.m_Initialized = true;
		return token;
	}

	/**
	 * method to emit the token's values as XML
	 * @return valid XML representation of token (assuming all values set),
	 * or the null string if not
	 */
	public String toXML() {
		if (!m_Initialized) {
			Log.error("Token is not properly initialized!");
			return "";
		}

		StringBuilder xml = new StringBuilder(256);
		xml.append("<token>");
		xml.append("<details>");
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		df.setMaximumFractionDigits(M_Precision);
		df.setMinimumFractionDigits(M_Precision);
		df.setGroupingSize(0);
		xml.append("<quantity>" + df.format(m_Quantity) + "</quantity>");
		xml.append("<units>" + m_Units + "</units>");
		xml.append("<asset>" + m_Asset + "</asset>");
		xml.append("<issuer>" + m_Issuer + "</issuer>");
		xml.append("<serial>" + m_SerialNumber + "</serial>");
		xml.append("</details>");
		xml.append("<signature>" + m_Signature + "</signature>");
		xml.append("</token>");
		return xml.toString();
	}

	/**
	 * method to set the serial number -- used for building a token manually
	 * @param snum the serial number to set
	 */
	public void setSerialNumber(String snum) {
		if (snum != null && !snum.isEmpty()) {
			m_SerialNumber = new String(snum);
		}
	}

	/**
	 * method to set the backing asset -- used for building a token manually
	 * @param asset the asset type backing the token
	 */
	public void setAsset(String asset) {
		if (asset != null && !asset.isEmpty()) {
			m_Asset = new String(asset);
		}
	}

	/**
	 * method to set the Issuer -- used for building a token manually
	 * @param issuer the Issuer of the vouchers backing the token
	 */
	public void setIssuer(String issuer) {
		if (issuer != null && !issuer.isEmpty()) {
			m_Issuer = new String(issuer);
		}
	}

	/**
	 * method to finalize initialzation -- used if manually building a token
	 * using the constructor, as opposed to starting with the XML code
	 */
	public void setInit() { m_Initialized = true; }

	/**
	 * method to support the Cloneable interface
	 * @return a new token object equal to this one
	 */
	@Override
	public Object clone() {
		try {
			XMLToken nToken = (XMLToken) super.clone();
			nToken.m_Quantity = new Double(m_Quantity.doubleValue());
			nToken.m_Units = new String(m_Units);
			nToken.m_Asset = new String(m_Asset);
			nToken.m_Issuer = new String(m_Issuer);
			nToken.m_SerialNumber = new String(m_SerialNumber);
			nToken.m_Signature = new String(m_Signature);
			return nToken;
		}
		catch (CloneNotSupportedException cnse) {
			throw new InternalError(cnse.toString());
		}
	}

	/**
	 * private utility method to build the data to be signed or validated
	 * @param oldStyle build signature data using original method, which does not
	 * use formatting (required to validate tokens generated that way)
	 * @return the XML details
	 */
	private String buildSignedDetails(boolean oldStyle) {
		// build string, equal to the <details/> block
		StringBuilder sigData = new StringBuilder(256);
		sigData.append("<details>");
		if (oldStyle) {
			// NB: this emits scientific notation, e.g. "5.0E-4", which differs
			// 	 from the human-readable value written by toXML().
			sigData.append("<quantity>" + m_Quantity + "</quantity>");
		}
		else {
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
			df.setMaximumFractionDigits(M_Precision);
			df.setMinimumFractionDigits(M_Precision);
			df.setGroupingSize(0);
			sigData.append("<quantity>" + df.format(m_Quantity.doubleValue()));
			sigData.append("</quantity>");
		}
		sigData.append("<units>" + m_Units + "</units>");
		sigData.append("<asset>" + m_Asset + "</asset>");
		sigData.append("<issuer>" + m_Issuer + "</issuer>");
		sigData.append("<serial>" + m_SerialNumber + "</serial>");
		sigData.append("</details>");
		return sigData.toString();
	}

	/**
	 * method to validate the signature on the token, given a public key
	 * which matches the private key used to make the signature
	 * @param pubKey the public key required to validate the signature
	 * @return true if the signature validates, false otherwise
	 */
	public boolean validateSig(PublicKey pubKey) {
		if (!m_Initialized) {
			Log.error("Token is not properly initialized, cannot validate");
			return false;
		}

		Signature vSig = null;
		byte[] sig = StringUtils.decodeBase64(m_Signature);
		try {
			vSig = Signature.getInstance("SHA1withRSA");
			vSig.initVerify(pubKey);

			// first, try without old style token signing
			vSig.update(buildSignedDetails(false).getBytes());
			if (!vSig.verify(sig)) {
				// try again using the old style
				vSig = Signature.getInstance("SHA1withRSA");
				vSig.initVerify(pubKey);
				vSig.update(buildSignedDetails(true).getBytes());
				return vSig.verify(sig);
			}
			return true;
		}
		catch (NoSuchAlgorithmException nsae) {
			Log.error("No such sig verify algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid pubkey for signature validation", ike);
		}
		catch (SignatureException se) {
			Log.error("Unable to validate token signature", se);
		}
		return false;
	}

	/**
	 * method to override Object::equals()
	 * @param obj the object to compare with
	 * @return true if this object is the same, false if not
	 */
	@Override
	public boolean equals(Object obj) {
		// if it's not a token, cannot be equal
		if (!(obj instanceof XMLToken)) {
			return false;
		}
		XMLToken token = (XMLToken) obj;

		// a token is equal to this one if its serial number, units, amount,
		// asset, and issuer are the same

		if (!m_SerialNumber.equals(token.getSerialNumber())) {
			return false;
		}
		if (!m_Issuer.equals(token.getIssuer())) {
			return false;
		}
		if (!m_Units.equals(token.getUnits())) {
			return false;
		}
		if (!m_Asset.equals(token.getAsset())) {
			return false;
		}
		if (m_Quantity.compareTo(token.getQuantity()) != 0) {
			return false;
		}
		return true;
	}

	/**
	 * method to override Object::hashCode()
	 * @return the hash code
	 */
	@Override
	public int hashCode() {
		int hc = 0;

		// sum the hash codes for the elements referenced by equals() above
		hc += m_SerialNumber.hashCode();
		hc += m_Issuer.hashCode();
		hc += m_Units.hashCode();
		hc += m_Asset.hashCode();
		hc += m_Quantity.hashCode();
		return hc;
	}

	/**
	 * method to override Object::compareTo()
	 * @param tok the other token
	 * @return -1, 0, or 1 as this token is less than, equal to, or greater than
	 * the specified token
	 */
	@Override
	public int compareTo(Object tok) {
		// if it's not a token, cannot be compared
		if (!(tok instanceof XMLToken)) {
			return -1;
		}
		XMLToken token = (XMLToken) tok;

		// to be consistent with equals, must also take issuers into account
		if (m_Issuer.equals(token.getIssuer())) {
			// order by serial number
			return m_SerialNumber.compareTo(token.getSerialNumber());
		}
		// order by Issuer name
		return m_Issuer.compareTo(token.getIssuer());
	}

	// END methods
}
