/*
 * last modified---
 * 	01-10-18/kbw/ make baseUnits() public
 * 	07-11-15/jet/ use VPAssetConfig methods in preference to local VectorS
 * 	08-06-14/jet/ Spark plugin version
 * 	03-18-14/jet/ add j6RetroSigCheck() for Java-7 support
 * 	03-14-14/jet/ for PXGOLD asset: units GRAMS, display units GAU
 * 	09-10-13/jet/ include SLC, SSV, SBC; remove SB-S? and SL-S0
 * 	08-13-13/jet/ add SilentLitecoin SL- asset types
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	implement a XML voucher object
 *
 * notes---
 * 	If this class is changed, org.vouchersafe.publisher.XMLVoucher must also
 * 	be changed.
 */

package org.silentvault.spark;

import org.dom4j.DocumentException;
import org.dom4j.Element;

import org.jivesoftware.smack.util.StringUtils;

import org.jivesoftware.spark.util.log.Log;

import java.util.Vector;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Date;
import java.util.Locale;
import java.util.Calendar;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.security.PublicKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.SignatureException;


/**
 * Class to instantiate a voucher as an XML object.  The class knows how to
 * store, parse, and emit a voucher in XML.  This same class is utilized in
 * both the Openfire VP plugin and VS client.
 *
 * @author Kevin Wilkerson
 */
public final strictfp class XMLVoucher implements Cloneable, Comparable {
	// BEGIN data members
	
	/**
	 * constant for conversion between grams and troy ounces
	 */
	public static final double		M_GramsPerOz = 31.1034768;

	/**
	 * voucher date format
	 */
	public static final String		M_DateFmt = "yyyy-MMM-dd HH:mm:ss";

	/**
	 * precision to use on voucher quantity
	 */
	public static final int			M_Precision = 4;

	/**
	 * names of all the elements that are part of a voucher
	 */
	private static Vector<String>	m_VoucherElements = getVoucherElements();

	/**
	 * the quantity of the underlying asset represented by the voucher
	 */
	private Double						m_Quantity;

	/**
	 * the measurement units of the voucher
	 */
	private String						m_Units;

	/**
	 * the underlying asset backing the voucher
	 */
	private String						m_Asset;

	/**
	 * the Issuer of this voucher, custodian of the backing asset
	 */
	private String						m_Issuer;

	/**
	 * the expiry date of this voucher
	 */
	private Date						m_Expiry;

	/**
	 * the voucher generation (number of times voucher has been re-issued)
	 */
	private Integer					m_Generation;

	/**
	 * the voucher serial number
	 */
	private String						m_SerialNumber;

	/**
	 * optional field: growth rate accrued by the underlying asset.
	 * (Example: interest accrued if voucher represented a bond.)
	 */
	private Double						m_Growth;

	/**
	 * optional field (required if m_Growth present): units for growth rate.
	 * (Examples: bpd, bpm, bpy, basis points per day/month/year respectively.)
	 */
	private String						m_GrowthUnits;

	/**
	 * optional field: cost accrued against the underlying asset.
	 * (Example: storage fee.)
	 */
	private Double						m_Cost;

	/**
	 * optional field (required if m_Cost present): units for accrued cost.
	 * (Examples: pcy, pcm for percent per year or per month respectively.)
	 */
	private String						m_CostUnits;

	/**
	 * optional field (required if m_Growth or m_Cost present): number of days
	 * for which this type of voucher is issued.  From this, along with the
	 * current time and expiry date (m_Expiry), the NPV of the voucher can be
	 * calculated.  (See getNPV().)
	 */
	private Integer					m_ValidDays;

	/**
	 * the signature of the voucher publisher on the above detail fields
	 */
	private String						m_Signature;

	/**
	 * whether voucher has been fully parsed/initialized
	 */
	private boolean					m_Initialized;

	// END data members

	// BEGIN methods

	/**
	 * initialization method to set the names of all elements in a voucher
	 * @return a Vector of all possible element names
	 */
	private static Vector<String> getVoucherElements() {
		Vector<String> vElts = new Vector<String>(15);
		vElts.add(new String("voucher"));
		vElts.add(new String("details"));
		vElts.add(new String("quantity"));
		vElts.add(new String("units"));
		vElts.add(new String("asset"));
		vElts.add(new String("issuer")); 
		vElts.add(new String("expiry"));
		vElts.add(new String("generation"));
		vElts.add(new String("serial"));
		vElts.add(new String("growth"));
		vElts.add(new String("gunits"));
		vElts.add(new String("cost"));
		vElts.add(new String("cunits"));
		vElts.add(new String("vdays"));
		vElts.add(new String("signature"));
		return vElts;
	}

	/**
	 * nullary constructor
	 */
	public XMLVoucher() {
		m_Quantity = new Double(0.0);
		m_Growth = new Double(0.0);
		m_Cost = new Double(0.0);
		m_Units = m_Asset = m_Issuer = m_SerialNumber = m_Signature = "";
		m_GrowthUnits = m_CostUnits = "";
		m_Generation = new Integer(0);
		m_ValidDays = new Integer(0);
		m_Expiry = new Date();
	}

	// get accessor methods
	/**
	 * get method for voucher quantity value
	 * @return the quantity, expressed in getUnits()
	 */
	public double getQuantity() { return m_Quantity.doubleValue(); }

	/**
	 * get method for voucher quantity in base units, depending on asset type
	 * @return the quantity, expressed in baseUnits() for the asset
	 */
	public strictfp double baseUnitValue() {
		double val = m_Quantity.doubleValue();
		/* Obtain the base unit conversion for our asset and unit.  If == 1.0,
		 * we are in base units, and can just use the quantity.  Otherwise we
		 * have to do a conversion.
		 */
		double convRatio = VPAssetConfig.getBaseUnitConversion(m_Asset, m_Units);
		if (convRatio != 1.0) {
			val /= convRatio;
		}
		return val;
	}

	/**
	 * get method for voucher units
	 * @return the units in which voucher is denominated
	 */
	public String getUnits() { return new String(m_Units); }

	/**
	 * get method for base or primary units
	 * @return the primary units defined for the given asset type
	 */
	public String baseUnits() {
		return VPAssetConfig.getBaseUnits(m_Asset);
	}

	/**
	 * get method for the display units which should be shown on various screens.
	 * (this is also typically the units used to look up relative asset prices)
	 * @param asset the asset type for which we want the display units
	 * @return the units, which may be different from m_Units
	 */
	public static String getDisplayUnits(String asset) {
		// NB: display units are always associated with base units
		String bUnits = VPAssetConfig.getBaseUnits(asset);
		return VPAssetConfig.getDisplayUnits(asset, bUnits);
	}

	/**
	 * get method for voucher asset type
	 * @return the type of asset which voucher represents
	 */
	public String getAsset() { return new String(m_Asset); }

	/**
	 * get method for voucher's Issuer
	 * @return the name of the Issuer
	 */
	public String getIssuer() { return new String(m_Issuer); }

	/**
	 * get method for voucher's serial number
	 * @return the serial number, an alpha-numeric string
	 */
	public String getSerialNumber() { return new String(m_SerialNumber); }

	/**
	 * get method for voucher growth rate
	 * @return the growth accrual rate, which may be zero
	 */
	public double getGrowth() { return m_Growth.doubleValue(); }

	/**
	 * get method for units of voucher growth rate
	 * @return the growth units, or the empty string if growth not defined
	 */
	public String getGrowthUnits() { return new String(m_GrowthUnits); }

	/**
	 * get method for voucher cost accrual rate
	 * @return the cost accrual rate, which may be zero
	 */
	public double getCost() { return m_Cost.doubleValue(); }

	/**
	 * get method for units of voucher cost rate
	 * @return the cost units, or the empty string if cost not defined
	 */
	public String getCostUnits() { return new String(m_CostUnits); }

	/**
	 * get method for validity days of this voucher
	 * @return the number of days, which will be zero unless a growth rate, a
	 * cost accrual rate, or both are defined
	 */
	public int getValidDays() { return m_ValidDays.intValue(); }

	/**
	 * get method for voucher signature data
	 * @return the signature, base64-encoded
	 */
	public String getSignature() { return new String(m_Signature); }

	/**
	 * get method for voucher expiration date
	 * @return the date
	 */
	public Date getExpiration() { return new Date(m_Expiry.getTime()); }

	/**
	 * get method for the generation of the voucher
	 * @return the number of times voucher has been circulated
	 */
	public int getGeneration() { return m_Generation.intValue(); }

	/**
	 * obtain the initialization status of the voucher
	 * @return true if all fields are set, false otherwise
	 */
	public boolean isInit() { return m_Initialized; }

	/**
	 * utility method to examine whether voucher is expired
	 * @return true if expiration date has passed, false otherwise
	 */
	public boolean isExpired() {
		Date now = new Date();
		// offset to GMT and DST
		Calendar cal = Calendar.getInstance();
		int gmtOffsetMillis = cal.get(Calendar.ZONE_OFFSET);
		int dstOffsetMillis = cal.get(Calendar.DST_OFFSET);
		Date trueExp
			= new Date(m_Expiry.getTime() + gmtOffsetMillis + dstOffsetMillis);
		return now.after(trueExp);
	}

	/**
	 * utility method to determine the number of days old a voucher is
	 * @return voucher age in days, or 0 if m_ValidDays is not defined
	 */
	private int daysOld() {
		if (m_ValidDays.intValue() == 0) {
			// cannot compute this without a reference point
			return 0;
		}

		// subtract validity days from the expiration
		Calendar origin = Calendar.getInstance();
		int gmtOffsetMillis = origin.get(Calendar.ZONE_OFFSET);
		int dstOffsetMillis = origin.get(Calendar.DST_OFFSET);
		Date trueExp
			= new Date(m_Expiry.getTime() + gmtOffsetMillis + dstOffsetMillis);
		origin.setTime(trueExp);
		origin.add(Calendar.DAY_OF_YEAR, -m_ValidDays.intValue());
		Date issued = origin.getTime();

		// compare issue date with current time
		Date now = new Date();
		long age = now.getTime() - issued.getTime();
		int days = (int) (age / javax.management.timer.Timer.ONE_DAY);
		return days;
	}

	/**
	 * method to construct a voucher from an input XML representation
	 * @param xmlRep the XML representation to be parsed
	 * @return the voucher object
	 * @throws DocumentException when the input won't parse
	 */
	public static XMLVoucher getVoucherRep(String xmlRep)
		throws DocumentException
	{
		// parse the input string to see if it's legal XML
		Element vouchElt = EncodingUtils.buildElement(xmlRep);
		if (vouchElt == null) {
			throw new DocumentException("voucher parse error");
		}
		return XMLVoucher.getVoucherRep(vouchElt);
	}

	/**
	 * method to construct a voucher from a dom4j Element representation
	 * @param vouchElt the Element representing the voucher
	 * @return the voucher object
	 * @throws DocumentException when the input is incomplete or improper
	 */
	public static XMLVoucher getVoucherRep(Element vouchElt)
		throws DocumentException
	{
		// prepare output
		XMLVoucher voucher = new XMLVoucher();

		// Make sure each individual field looks sensible.  Under the root
		// element, we should have <details/> and <signature/>.

		Element detailsElt = vouchElt.element("details");
		if (detailsElt == null) {
			throw new DocumentException("Missing <details/> element in "
												+ vouchElt.asXML());
		}
		Element sigElt = vouchElt.element("signature");
		if (sigElt == null) {
			throw new DocumentException("Missing <signature/> element in "
												+ vouchElt.asXML());
		}
		String sig = sigElt.getTextTrim();
		if (sig.isEmpty()) {
			throw new DocumentException("Blank <signature/> element");
		}
		// NB: we don't attempt to validate the signature here
		voucher.m_Signature = new String(sig);

		// the remaining elements will all be inside the <details/> block
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		Iterator eltIt = detailsElt.elementIterator();
		while (eltIt.hasNext()) {

			// make sure this element is one of those we are expecting
			Element thisElt = (Element) eltIt.next();
			String eltName = thisElt.getName().toLowerCase();
			if (!m_VoucherElements.contains(eltName)) {
				throw new DocumentException("Illegal voucher element: " + eltName);
			}

			// check for each item it can be and handle appropriately
			if (eltName.equals("quantity")) {
				Number quan = null;
				try {
					quan = df.parse(thisElt.getText());
				}
				catch (ParseException pe) {
					Log.error("Voucher quantity parse error", pe);
				}
				if (quan != null && quan.doubleValue() > 0.0) {
					voucher.m_Quantity = new Double(quan.doubleValue());
				}
				else {
					throw new DocumentException("Illegal quantity value: "
														+ thisElt.getText());
				}
			}
			else if (eltName.equals("units")) {
				String unitNm = thisElt.getTextTrim().toUpperCase();
				ArrayList<String> allUnits = VPAssetConfig.getAllUnits();
				if (allUnits.contains(unitNm)) {
					voucher.m_Units = unitNm;
				}
				else {
					throw new DocumentException("Illegal units value: " + unitNm);
				}
			} 
			else if (eltName.equals("asset")) {
				String assetNm = thisElt.getTextTrim().toUpperCase();
				voucher.m_Asset = assetNm;
			}
			else if (eltName.equals("issuer")) {
				String issuerNm = thisElt.getTextTrim().toUpperCase();
				voucher.m_Issuer = issuerNm;
			}
			else if (eltName.equals("expiry")) {
				String expDtstr = thisElt.getText();
				SimpleDateFormat sdf = new SimpleDateFormat(M_DateFmt, Locale.US);
				sdf.setLenient(true);
				try {
					voucher.m_Expiry = sdf.parse(expDtstr);
				}
				catch (ParseException pe) {
					throw new DocumentException("Bad date format: " + expDtstr, pe);
				}
			}
			else if (eltName.equals("generation")) {
				Integer gen = Integer.parseInt(thisElt.getText());
				if (gen != null && gen.intValue() > 0) {
					voucher.m_Generation = gen;
				}
				else {
					throw new DocumentException("Illegal generation value: "
														+ thisElt.getText());
				}
			}
			else if (eltName.equals("serial")) {
				String serialNum = thisElt.getTextTrim();
				if (!serialNum.isEmpty()) {
					voucher.m_SerialNumber = serialNum;
				}
				else {
					throw new DocumentException("Empty serial number");
				}
			}
			else if (eltName.equals("growth")) {
				Double growth = Double.parseDouble(thisElt.getText());
				if (growth != null && growth.doubleValue() > 0.0) {
					voucher.m_Growth = growth;
				}
				else {
					throw new DocumentException("Illegal growth value: "
														+ thisElt.getText());
				}
			}
			else if (eltName.equals("gunits")) {
				String growthUnits = thisElt.getTextTrim();
				if (!growthUnits.isEmpty()) {
					voucher.m_GrowthUnits = growthUnits;
				}
				else {
					throw new DocumentException("Empty growth units");
				}
			}
			else if (eltName.equals("cost")) {
				Double cost = Double.parseDouble(thisElt.getText());
				if (cost != null && cost.doubleValue() > 0.0) {
					voucher.m_Cost = cost;
				}
				else {
					throw new DocumentException("Illegal cost value: "
														+ thisElt.getText());
				}
			}
			else if (eltName.equals("cunits")) {
				String costUnits = thisElt.getTextTrim();
				if (!costUnits.isEmpty()) {
					voucher.m_CostUnits = costUnits;
				}
				else {
					throw new DocumentException("Empty cost units");
				}
			}
			else if (eltName.equals("vdays")) {
				Integer days = Integer.parseInt(thisElt.getText());
				if (days != null && days.intValue() > 0) {
					voucher.m_ValidDays = days;
				}
				else {
					throw new DocumentException("Illegal valid days value: "
														+ thisElt.getText());
				}
			}
			else {
				// shouldn't happen, but throw exception anyway
				throw new DocumentException("Unrecognized voucher element, "
													+ eltName);
			}
		}

		// see if the voucher looks complete now
		if (voucher.m_Quantity.doubleValue() <= 0.0) {
			throw new DocumentException("Missing quantity");
		}
		if (voucher.m_Units.isEmpty()) {
			throw new DocumentException("Missing units");
		}
		if (voucher.m_Asset.isEmpty()) {
			throw new DocumentException("Missing asset type");
		}
		if (voucher.m_Issuer.isEmpty()) {
			throw new DocumentException("Missing issuer");
		}
		if (voucher.m_SerialNumber.isEmpty()) {
			throw new DocumentException("Missing serial number");
		}
		if (voucher.m_Generation.intValue() <= 0) {
			throw new DocumentException("Missing generation value");
		}
		// if we have a growth field, we need days and growth units also
		boolean needDays = false;
		if (voucher.m_Growth.doubleValue() > 0.0) {
			if (voucher.m_GrowthUnits.isEmpty()) {
				throw new DocumentException("Missing units for growth quanta");
			}
			needDays = true;
		}
		// ditto for a cost field
		if (voucher.m_Cost.doubleValue() > 0.0) {
			if (voucher.m_CostUnits.isEmpty()) {
				throw new DocumentException("Missing units for cost quanta");
			}
			needDays = true;
		}
		if (needDays && voucher.m_ValidDays.intValue() == 0) {
			throw new DocumentException("Missing valid days");
		}
		voucher.m_Initialized = true;
		return voucher;
	}

	/**
	 * method to emit the voucher's values as XML
	 * @param fname optional filename to emit along with voucher
	 * @return valid XML representation of voucher (assuming all values set),
	 * or the null string if not
	 */
	public String toXML(String fname) {
		if (!m_Initialized) {
			Log.error("Voucher is not properly initialized!");
			return "";
		}

		StringBuilder xml = new StringBuilder(1024);
		if (fname != null && !fname.isEmpty()) {
			xml.append("<voucher filename=\"" + fname + "\">");
		}
		else {
			xml.append("<voucher>");
		}
		xml.append("<details>");
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		df.setMaximumFractionDigits(M_Precision);
		df.setMinimumFractionDigits(M_Precision);
		df.setGroupingSize(0);
		xml.append("<quantity>" + df.format(m_Quantity) + "</quantity>");
		xml.append("<units>" + m_Units + "</units>");
		xml.append("<asset>" + m_Asset + "</asset>");
		xml.append("<issuer>" + m_Issuer + "</issuer>");
		SimpleDateFormat sdf = new SimpleDateFormat(M_DateFmt, Locale.US);
		xml.append("<expiry>" + sdf.format(m_Expiry) + "</expiry>");
		xml.append("<generation>" + m_Generation + "</generation>");
		xml.append("<serial>" + m_SerialNumber + "</serial>");
		df.setMinimumFractionDigits(0);
		if (m_Growth.doubleValue() > 0.0) {
			xml.append("<growth>" + df.format(m_Growth) + "</growth>");
		}
		if (!m_GrowthUnits.isEmpty()) {
			xml.append("<gunits>" + m_GrowthUnits + "</gunits>");
		}
		if (m_Cost.doubleValue() > 0.0) {
			xml.append("<cost>" + df.format(m_Cost) + "</cost>");
		}
		if (!m_CostUnits.isEmpty()) {
			xml.append("<cunits>" + m_CostUnits + "</cunits>");
		}
		if (m_ValidDays.intValue() > 0) {
			xml.append("<vdays>" + m_ValidDays + "</vdays>");
		}
		xml.append("</details>");
		xml.append("<signature>" + m_Signature + "</signature>");
		xml.append("</voucher>");
		return xml.toString();
	}

	/**
	 * method to invoke previous method with null filename argument
	 * @return valid XML representation of the voucher (if initialized)
	 */
	public String toXML() { return toXML(null); }

	/**
	 * method to support the Cloneable interface
	 * @return a new voucher object equal to this one
	 */
	@Override
	public Object clone() {
		try {
			XMLVoucher nVouch = (XMLVoucher) super.clone();
			nVouch.m_Expiry = (Date) m_Expiry.clone();
			nVouch.m_Quantity = new Double(m_Quantity.doubleValue());
			nVouch.m_Generation = new Integer(m_Generation.intValue());
			nVouch.m_Units = new String(m_Units);
			nVouch.m_Asset = new String(m_Asset);
			nVouch.m_Issuer = new String(m_Issuer);
			nVouch.m_SerialNumber = new String(m_SerialNumber);
			nVouch.m_Growth = new Double(m_Growth.doubleValue());
			nVouch.m_GrowthUnits = new String(m_GrowthUnits);
			nVouch.m_Cost = new Double(m_Cost.doubleValue());
			nVouch.m_CostUnits = new String(m_CostUnits);
			nVouch.m_ValidDays = new Integer(m_ValidDays.intValue());
			nVouch.m_Signature = new String(m_Signature);
			return nVouch;
		}
		catch (CloneNotSupportedException cnse) {
			throw new InternalError(cnse.toString());
		}
	}

	/**
	 * private utility method to build the data to be validated
	 * @return the XML details to be signature validated
	 */
	private String buildSignedDetails() {
		// build string, equal to the <details/> block
		StringBuilder sigData = new StringBuilder(512);
		sigData.append("<details>");
		// NB: quantity may appear differently from human-readable form
		sigData.append("<quantity>" + m_Quantity + "</quantity>");
		sigData.append("<units>" + m_Units + "</units>");
		sigData.append("<asset>" + m_Asset + "</asset>");
		sigData.append("<issuer>" + m_Issuer + "</issuer>");
		SimpleDateFormat sdf = new SimpleDateFormat(M_DateFmt, Locale.US);
		sigData.append("<expiry>" + sdf.format(m_Expiry) + "</expiry>");
		sigData.append("<generation>" + m_Generation + "</generation>");
		sigData.append("<serial>" + m_SerialNumber + "</serial>");
		// NB: growth may appear differently from human-readable form
		if (m_Growth.doubleValue() > 0.0) {
			sigData.append("<growth>" + m_Growth + "</growth>");
		}
		if (!m_GrowthUnits.isEmpty()) {
			sigData.append("<gunits>" + m_GrowthUnits + "</gunits>");
		}
		// NB: cost may appear differently from human-readable form
		if (m_Cost.doubleValue() > 0.0) {
			sigData.append("<cost>" + m_Cost + "</cost>");
		}
		if (!m_CostUnits.isEmpty()) {
			sigData.append("<cunits>" + m_CostUnits + "</cunits>");
		}
		if (m_ValidDays.intValue() > 0) {
			sigData.append("<vdays>" + m_ValidDays + "</vdays>");
		}
		sigData.append("</details>");
		return sigData.toString();
	}

	/**
	 * method to validate the signature on the voucher, given a public key
	 * which matches the private key used to make the signature
	 * @param pubKey the public key required to validate the signature
	 * @return true if the signature validates, false otherwise
	 */
	public boolean validateSig(PublicKey pubKey) {
		if (!m_Initialized) {
			Log.error("Voucher is not properly initialized, cannot validate");
			return false;
		}

		Signature vSig = null;
		try {
			vSig = Signature.getInstance("SHA1withRSA");
			vSig.initVerify(pubKey);
			vSig.update(buildSignedDetails().getBytes());
			boolean j7Result= vSig.verify(StringUtils.decodeBase64(m_Signature));
			if (!j7Result) {
				// before failing, check for old Java-6 signature
				return j6RetroSigCheck(pubKey);
			}
			return j7Result;
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("No such sig algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid pubkey for signature validation", ike);
		}
		catch (SignatureException se) {
			Log.error("Unable to validate voucher signature", se);
		}
		return false;
	}

	/**
	 * Method to validate the signature on the voucher, given a public key
	 * which matches the private key used to make the signature.  This method
	 * explicitly sets 4-digit precision, which is necessary to duplicate the
	 * signature created by Java-6, which had a bug causing quantities such as
	 * "0.005" to be emitted as "0.0050".
	 * @param pubKey the public key required to validate the signature
	 * @return true if the signature validates, false otherwise
	 */
	public boolean j6RetroSigCheck(PublicKey pubKey) {
		if (!m_Initialized) {
			Log.error("Voucher is not properly initialized, cannot validate");
			return false;
		}

		// build string, equal to the <details/> block
		StringBuilder sigData = new StringBuilder(512);
		sigData.append("<details>");
		// explicitly use a 4-digit precision for the quantity
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		df.setMaximumFractionDigits(M_Precision);
		df.setMinimumFractionDigits(M_Precision);
		df.setGroupingSize(0);
		sigData.append("<quantity>" + df.format(m_Quantity) + "</quantity>");
		sigData.append("<units>" + m_Units + "</units>");
		sigData.append("<asset>" + m_Asset + "</asset>");
		sigData.append("<issuer>" + m_Issuer + "</issuer>");
		SimpleDateFormat sdf = new SimpleDateFormat(M_DateFmt, Locale.US);
		sigData.append("<expiry>" + sdf.format(m_Expiry) + "</expiry>");
		sigData.append("<generation>" + m_Generation + "</generation>");
		sigData.append("<serial>" + m_SerialNumber + "</serial>");
		// NB: growth may appear differently from human-readable form
		if (m_Growth.doubleValue() > 0.0) {
			sigData.append("<growth>" + m_Growth + "</growth>");
		}
		if (!m_GrowthUnits.isEmpty()) {
			sigData.append("<gunits>" + m_GrowthUnits + "</gunits>");
		}
		// NB: cost may appear differently from human-readable form
		if (m_Cost.doubleValue() > 0.0) {
			sigData.append("<cost>" + m_Cost + "</cost>");
		}
		if (!m_CostUnits.isEmpty()) {
			sigData.append("<cunits>" + m_CostUnits + "</cunits>");
		}
		if (m_ValidDays.intValue() > 0) {
			sigData.append("<vdays>" + m_ValidDays + "</vdays>");
		}
		sigData.append("</details>");
		Signature vSig = null;
		try {
			vSig = Signature.getInstance("SHA1withRSA");
			vSig.initVerify(pubKey);
			vSig.update(sigData.toString().getBytes());
			return vSig.verify(StringUtils.decodeBase64(m_Signature));
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("No such sig algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid pubkey for signature validation", ike);
		}
		catch (SignatureException se) {
			Log.error("Unable to validate voucher signature", se);
		}
		return false;
	}

	/**
	 * method to compute the net present value (NPV) of the voucher, given
	 * the effect of its age and any listed growth or cost values.
	 * @return the net present value, in base units
	 */
	public strictfp double getNPV() {
		// need a duration and either a growth or a cost
		double gRate = m_Growth.doubleValue();
		double cRate = m_Cost.doubleValue();
		double baseValue = baseUnitValue();
		if (m_ValidDays.intValue() == 0 || (gRate == 0.0 && cRate == 0.0)) {
			// just return the static (original) value
			return baseValue;
		}

		// if voucher is not at least one day old, nothing to compute
		int vAge = daysOld();
		if (vAge <= 0) {
			return baseValue;
		}

		// Compute the value of the growth to date.  Assumed to be compounding.
		double growth = 0.0;
		double npv = baseValue;
		if (gRate > 0.0) {
			// examine the units, which will be in bpd (basis points/day), bpm, bpy
			if (m_GrowthUnits.isEmpty()) {
				Log.error("Missing units for NPV growth calculation");
				return baseValue;
			}
			double factor = 1.0;
			int days = Math.min(vAge, m_ValidDays.intValue());
			// apply growth without exceeding number of days to maturity
			if (m_GrowthUnits.equals("bpd")) {
				factor = Math.pow((1.0 + (gRate / 10000.0)), (double) days);
			}
			else if (m_GrowthUnits.equals("bpm")) {
				factor
					= Math.pow((1.0 + ((gRate / 10000.0) / 30.0)), (double) days);
			}
			else if (m_GrowthUnits.equals("bpy")) {
				factor
					= Math.pow((1.0 + ((gRate / 10000.0) / 365.0)), (double) days);
			}
			else {
				Log.error("Unknown units for NPV growth calculation, "
							+ m_GrowthUnits);
				return baseValue;
			}
			growth = baseValue * (factor - 1.0);
			npv += growth;
		}

		// Compute the value of the cost to date, pro-rata, non-compounding.
		double costs = 0.0;
		double topUp = 0.0;
		double storage = 0.0;
		if (cRate > 0.0) {
			/* examine the units, which will normally be in pcm or pcy, for percent
			 * per month or per year
			 */
			boolean monthly = false;
			if (m_CostUnits.isEmpty()) {
				Log.error("Missing units for NPV cost calculation");
				return baseValue;
			}
			if (m_CostUnits.equals("pcm")) {
				monthly = true;
			}
			else if (m_CostUnits.equals("pcy")) {
				monthly = false;
			}
			else {
				Log.error("Unknown units for NPV cost calculation, " + m_CostUnits);
				return baseValue;
			}

			// pro-rate to age of voucher, without a maximum # of days
			double perDiem = m_Cost.doubleValue() / 100.0;
			if (monthly) {
				// rate is stated monthly
				perDiem /= 30.0;
			}
			else {
				// rate is stated yearly
				perDiem /= 365.0;
			}

			/* If this asset type is of the SS derived series, then we need
			 * to cover a "top-up" of 30% of the total amount including growth.
			 * We then apply the storage fees to the 80% portion which is backed.
			 *
			 * THIS IS TESTING CODE FOR TRIAL ASSET TYPES AND SAFE TO REMOVE.
			 * NORMAL PROCESSING IS IN THE else CLAUSE.
			 */
			if (m_Asset.startsWith("SS-")
				 && !(m_Asset.endsWith("-S0") || m_Asset.endsWith("-T0")))
			{
				double reserve = npv * 0.3;
				double backed = npv * 0.8;

				// pro-rate the reserve top-up to the age of the voucher
				int days = Math.min(vAge, m_ValidDays.intValue());
				double proRate = (double) days / m_ValidDays.doubleValue();
				topUp = reserve * proRate;

				// now compute charges on the backed portion (80% of total)
				storage = backed * perDiem * (double) vAge;
				costs = topUp + storage;
			}
			else {
				// just do a simple interest calculation on the whole balance
				double finance = npv * perDiem * (double) vAge;
				costs = finance;
				//Log.debug("Hit simple finance cost case, charge = " + finance);
			}
			npv -= costs;
		}
	/*
		Log.debug("Calculated NPV: " + npv + " for " + m_Asset + " voucher, age "
					+ vAge + "d, base = " + baseValue + ", growth = " + growth
					+ ", topOwed = " + topUp + ", storage = " + storage);
	 */
		return npv;
	}

	/**
	 * method to override Object::equals()
	 * @param obj the object to compare with
	 * @return true if this object is the same, false if not
	 */
	@Override
	public boolean equals(Object obj) {
		// if it's not a voucher, cannot be equal
		if (!(obj instanceof XMLVoucher)) {
			return false;
		}
		XMLVoucher vouch = (XMLVoucher) obj;

		// a voucher is equal to this one if its serial number, amount, units,
		// asset, and issuer are the same (or signature; but we don't rely on
		// that in case we're comparing two as yet-unsigned vouchers)

		if (!m_SerialNumber.equals(vouch.m_SerialNumber)) {
			return false;
		}
		if (m_Quantity.compareTo(vouch.m_Quantity) != 0) {
			return false;
		}
		if (!m_Units.equals(vouch.m_Units)) {
			return false;
		}
		if (!m_Asset.equals(vouch.m_Asset)) {
			return false;
		}
		if (!m_Issuer.equals(vouch.m_Issuer)) {
			return false;
		}
		return true;
	}

	/**
	 * method to override Object::hashCode()
	 * @return the hash code
	 */
	@Override
	public int hashCode() {
		int hc = 0;

		// sum the hash codes for the members referenced by equals() above
		hc += m_SerialNumber.hashCode();
		hc += m_Quantity.hashCode();
		hc += m_Units.hashCode();
		hc += m_Asset.hashCode();
		hc += m_Issuer.hashCode();
		return hc;
	}

	/**
	 * method to implement Comparable, based on voucher expiration dates
	 * @param voucher the voucher object to compare with this one for order
	 * @return 1 if voucher expires before this one, 0 if the expiration is
	 * the same, and -1 if voucher expires after this one
	 * NB: this implementation is "not consistent with equals" (above)
	 */
	@Override
	public int compareTo(Object voucher) {
		if (!(voucher instanceof XMLVoucher)) {
			// no basis for comparison
			return 0;
		}
		XMLVoucher vouch = (XMLVoucher) voucher;
		int ret = 0;
		Date vouchExp = vouch.getExpiration();
		if (vouchExp.before(m_Expiry)) {
			ret = 1;
		}
		else if (vouchExp.after(m_Expiry)) {
			ret = -1;
		}
		return ret;
	}

	// END methods
}
