/*
 * last modified---
 * 	02-13-18/kbw/ correct premium calculation
 * 	08-05-14/jet/ Spark plugin version
 * 	07-09-13/jet/ new
 *
 * purpose---
 * 	provide a cell renderer for all columns in the offer table display
 */

package org.silentvault.spark.ui;

import javax.swing.border.EmptyBorder;
import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.JCheckBox;
import javax.swing.JRadioButton;
import javax.swing.SwingConstants;
import javax.swing.UIManager;
import javax.swing.border.Border;
import javax.swing.table.TableCellRenderer;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.Dimension;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.XMLVoucher;
import org.silentvault.spark.TradeOffer;
import org.silentvault.spark.WalletClient;

import java.util.Date;
import java.util.Calendar;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import javax.management.timer.Timer;


/**
 * This class is used as the DefaultRenderer for all offer table columns.
 * Nearly expired offers have their expiration date drawn in warning colors.
 * Tooltips are set appropriately to describe the data.
 *
 * @author Justin Turrell
 */
public final class OfferRenderer extends JLabel implements TableCellRenderer {
	// data members
	/**
	 * the border to use for the cell
	 */
	private Border					m_Border = null;

	/**
	 * a helpful date formatter
	 */
	private SimpleDateFormat	m_DateFormatter = null;

	/**
	 * the offer table model in support of which we render the table
	 */
	private OfferTableModel		m_OfferModel;

	/**
	 * the application or plugin of which we're a part
	 */
	private WalletClient			m_Plugin;

	// methods
	/**
	 * constructor
	 * @param offerModel the offer table model we are supporting
	 * @param plugin the plugin or app that we're part of
	 */
	public OfferRenderer(OfferTableModel offerModel, WalletClient plugin) {
		super();
		m_OfferModel = offerModel;
		m_Plugin = plugin;
		setOpaque(true);			// to make background show up
		m_Border = new EmptyBorder(5, 3, 5, 3);
		m_DateFormatter = new SimpleDateFormat("dd-MMM-yy HH:mm");
	}

	/**
	 * method to implement the TableCellRenderer interface
	 * @param offerTable the offer table we're rendering
	 * @param colValue the value of the particular column (String, Date, etc.)
	 * @param isSelected whether this cell (offer) is currently selected
	 * @param hasFocus whether this cell has the focus (we ignore this)
	 * @param row the table row (i.e. which offer in the model)
	 * @param column the table column (we ignore this)
	 */
	public Component getTableCellRendererComponent(	JTable offerTable,
																	Object colValue,
																	boolean isSelected,
																	boolean hasFocus,
																	int row,
																	int column )
	{
		// obtain the offer data, allowing for possibility table is sorted
		int realRow = offerTable.convertRowIndexToModel(row);
		TradeOffer offer = m_OfferModel.getOfferAtRow(realRow);
		if (offer == null) {
			return null;
		}
		if (colValue == null) {
			return null;
		}
		boolean isFilled = offer.getStatus().equals(TradeOffer.M_StatusFilled);

		// set defaults
		setFont(m_Plugin.getExchangeTabManager().getHomePane().M_TableFont);
		setForeground(UIManager.getColor("Table.foreground"));
		setHorizontalAlignment(SwingConstants.TRAILING);

		// use default background color (off-white) if row is even
		Color offWhite = UIManager.getColor("Table.alternateRowColor");
		if (row % 2 != 0) {
			setBackground(Color.WHITE);
		}
		else {
			setBackground(offWhite);
		}

		// override these defaults if the offer is expired or close to it
		Calendar cal = Calendar.getInstance();
		int offGMTmillis = cal.get(Calendar.ZONE_OFFSET);
		int offDSTmillis = cal.get(Calendar.DST_OFFSET);
		Date now = new Date();
		Date tenDaysHence = new Date(now.getTime() + (Timer.ONE_DAY * 10L));
		Date threeDaysHence = new Date(now.getTime() + (Timer.ONE_DAY * 3L));

		/* Examine column type:
		 * JToggleButton - it's a select control
		 * Integer - it's an offer ID number
		 * Double - it's an offer quantity
		 * Float - it's a price ratio
		 * Date - it's an expiration date
		 */
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setGroupingUsed(false);
		if (colValue instanceof Date) {
			// this column is a offer expiration date
			Date expDate = (Date) colValue;
			setText(m_DateFormatter.format(expDate));

			// adjust for timezone before checking remaining durations
			long secs = (expDate.getTime() - now.getTime()) / Timer.ONE_SECOND;
			long days = secs / 86400L;
			Date expDate2
				= new Date(expDate.getTime() + offGMTmillis + offDSTmillis);

			// filled offers cannot expire
			if (isFilled) {
				setToolTipText("A filled offer cannot expire");
				setBackground(Color.LIGHT_GRAY);
			}
			else if (expDate2.before(now)) {
				setToolTipText("This offer has expired!");
				// set ugly colors
				setBackground(Color.RED);
				setForeground(Color.DARK_GRAY);
			}
			else if (expDate2.before(threeDaysHence)) {
				setBackground(new Color(0xFEA98D));
				setToolTipText(m_DateFormatter.format(expDate2)
									+ " (in only " + days + " days!)");
				// use bold italic
				setFont(new Font(Font.SANS_SERIF, Font.BOLD | Font.ITALIC, 13));
			}
			else if (expDate2.before(tenDaysHence)) {
				setBackground(new Color(0xFEE26F));
				setToolTipText(m_DateFormatter.format(expDate2)
									+ " (in only " + days + " days)");
				// use italic
				setFont(new Font(Font.SANS_SERIF, Font.ITALIC, 13));
			}
			else {
				// offer has more than ten days, draw plain and show days
				setToolTipText(m_DateFormatter.format(expDate2)
									+ " (in " + days + " days)");
			}
		}
		else if (colValue instanceof Float) {
			// ratio of OFFERED:WANTED
			Float val = (Float) colValue;
			df.setMaximumFractionDigits(2);
			df.setMinimumFractionDigits(2);
			setText(df.format(val.floatValue()));

			// show premium or discount percentage in tooltip
			float diff = (1.0F - val.floatValue()) * 100.0F;
			if (diff > 0.0F) {
				if (isFilled) {
					setToolTipText("You have received a premium of "
										+ df.format(diff) + " percent vs. "
										+ offer.getExAsset());
				}
				else {
					setToolTipText("Trader is demanding a premium of "
										+ df.format(diff) + " percent");
				}
			}
			else if (diff < 0.0F) {
				if (isFilled) {
					setToolTipText("You have received a discount of "
										+ df.format(-diff) + " percent vs. "
										+ offer.getExAsset());
				}
				else {
					setToolTipText("Trader is offering a discount of "
										+ df.format(-diff) + " percent");
				}
			}
			else {
				if (isFilled) {
					setToolTipText("You exchanged at par for " + offer.getExAsset());
				}
				else {
					setToolTipText("Trader is offering an exchange at par");
				}
			}
		}
		else if (colValue instanceof Double) {
			// amount of offer
			Double val = (Double) colValue;
			df.setMaximumFractionDigits(XMLVoucher.M_Precision);
			df.setMinimumFractionDigits(XMLVoucher.M_Precision);
			setText(df.format(val.doubleValue()));
			String units = XMLVoucher.getDisplayUnits(offer.getOffered());
			if (isFilled) {
				units = XMLVoucher.getDisplayUnits(offer.getExAsset());
				setToolTipText("You have exchanged "
									+ df.format(offer.getExQuantity()) + " " + units
									+ " of " + offer.getExAsset() + " for this");
			}
			else if (offer.getStatus().equals(TradeOffer.M_StatusFunded)) {
				setToolTipText(units + " of " + offer.getOffered()
									+ " held in escrow");
			}
			else {
				setToolTipText(units + " of " + offer.getOffered()
									+ " trader is offering");
			}
			// use bold
			setFont(new Font(Font.SANS_SERIF, Font.BOLD, 13));
		}
		else if (colValue instanceof JCheckBox) {
			// unless we're in multi-select mode, return null here
			JCheckBox cb = (JCheckBox) colValue;
			if (m_OfferModel.getSelectMode() == OfferTableModel.M_SelectMulti) {
				cb.setToolTipText("<html>Click here to select this offer<br/>"
										+ "in your list of selected offers</html>");
				cb.setText("");
				cb.setOpaque(true);
				cb.setHorizontalAlignment(SwingConstants.CENTER);
				if (row % 2 != 0) {
					cb.setBackground(Color.WHITE);
				}
				else {
					cb.setBackground(offWhite);
				}
				return cb;
			}
			return null;
		}
		else if (colValue instanceof JRadioButton) {
			// unless we're in single-select mode, return null here
			JRadioButton rb = (JRadioButton) colValue;
			if (m_OfferModel.getSelectMode() == OfferTableModel.M_SelectSingle) {
				rb.setToolTipText("Click here to select this offer");
				rb.setText("");
				rb.setOpaque(true);
				rb.setHorizontalAlignment(SwingConstants.CENTER);
				if (row % 2 != 0) {
					rb.setBackground(Color.WHITE);
				}
				else {
					rb.setBackground(offWhite);
				}
				return rb;
			}
			return null;
		}
		else if (colValue instanceof Integer) {
			Integer oId = (Integer) colValue;
			setText(oId.toString());
			setToolTipText("The ID of the offer");
		}
		else if (colValue instanceof String) {
			// this is either Offered, Wanted, or Status
			String sVal = (String) colValue;
			setText(sVal);
			switch (column) {
				case 2:
					setToolTipText("Asset offered");
					break;
				case 4:
					setToolTipText("Asset(s) accepted: "
										+ sVal.replace(':', ','));
					break;
				case 6:
					// explain the meaning of each valid status value
					if (sVal.equals(TradeOffer.M_StatusOpen)) {
						setToolTipText("You have not yet funded escrow");
					}
					else if (sVal.equals(TradeOffer.M_StatusFunded)) {
						// NB: this one could be in either MyOffers or Search results
						setToolTipText("Escrow funded, awaiting acceptance");
					}
					else if (sVal.equals(TradeOffer.M_StatusFilled)) {
						setToolTipText("Accepted by another user, payment sent");
					}
					else if (sVal.equals(TradeOffer.M_StatusExpired)) {
						setToolTipText("Expired, your escrow payment refunded");
					}
					else {
						setToolTipText("unexpected status");
					}
					break;
				default:
					setToolTipText("undefined column");
					break;
			}
		}
		else {
			Log.error("Unexpected column data type in offer renderer, "
						+ colValue.getClass());
			setText("column type error");
		}

		// fix up the border to impose inset requirements
		setBorder(m_Border);

		return this;
	}

	// end methods
}
