/*
 * last modified---
 * 	08-05-14/jet/ Spark plugin version
 * 	07-09-13/jet/ new
 *
 * purpose---
 * 	provide a cell editor for the selection column in an offer table
 */

package org.silentvault.spark.ui;

import javax.swing.JCheckBox;
import javax.swing.JToggleButton;
import javax.swing.JTable;
import javax.swing.table.TableCellEditor;
import javax.swing.AbstractCellEditor;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.TradeOffer;
import org.silentvault.spark.WalletClient;
import org.silentvault.spark.ui.svx.*;

import java.util.EventObject;


/**
 * This class is used as a specialty editor for offer table cells which are
 * in the 0th (Select) column of the table.  It is necessary because we need
 * both radio button as well as checkbox support (and DefaultCellEditor only
 * supports the latter).
 *
 * @author Justin Turrell
 */
public final class OfferSelector extends AbstractCellEditor
		implements TableCellEditor, ActionListener {
	// data members
	/**
	 * the editor activation command
	 */
	protected static final String	SELECT = "select";

	/**
	 * the editor component
	 */
	private JToggleButton		m_CellEditor;

	/**
	 * the value (checkbox or radio button) in the table
	 */
	private JToggleButton		m_CellValue;

	/**
	 * the application or plugin of which we're a part
	 */
	private WalletClient			m_Plugin;

	/**
	 * the offer table model in support of which we edit the cells
	 */
	private OfferTableModel		m_OfferModel;


	// methods
	/**
	 * constructor
	 * @param plugin the plugin or app that we're part of
	 * @param offerModel the offer table model we are supporting
	 */
	public OfferSelector(WalletClient plugin, OfferTableModel offerModel) {
		super();
		m_Plugin = plugin;
		m_OfferModel = offerModel;

		// configure the buttons
		m_CellEditor = new JToggleButton();
		m_CellEditor.setVisible(false);
		m_CellEditor.setActionCommand(SELECT);
		m_CellEditor.addActionListener(this);
		m_CellEditor.setBorderPainted(false);
	}


	// method required to implement ActionListener
	public void actionPerformed(ActionEvent e) {
		if (SELECT.equals(e.getActionCommand())) {
			if (m_CellValue == null) {
				Log.error("Null cell value on click to start edit");
			}
			else {
				// only allow selection if corresponding cell value button enabled
				if (m_CellValue.isEnabled()) {
					m_CellEditor.setSelected(!m_CellValue.isSelected());
				}
			}

			// tell JTable to switch back to renderer for this cell
			fireEditingStopped();
		}
		else {
			Log.error("Unexpected action on select button");
		}
	}


	// method required to implement CellEditor
	/**
	 * obtain the value contained in the editor
	 * @return the value
	 */
	public Object getCellEditorValue() {
		// propogate the selected status if the corresponding button is enabled
		if (m_CellValue.isEnabled()) {
			m_CellValue.setSelected(m_CellEditor.isSelected());
		}

		// redraw the table to force repainting any now de-selected button or box
		// (the buttons are in a ButtonGroup, so side-effects are desired)

		XTabManager tm = m_Plugin.getExchangeTabManager();
		final JTable ot = tm.getHomePane().getOfferTable();

		/*
		 * There are certain panes which might want to know about the change of
		 * selection.  Notify them.
		 */
		int mode = m_OfferModel.getSelectMode();
		if (mode == OfferTableModel.M_SelectMulti) {
			Log.error("Unexpected select mode, multi");
		}
		else if (mode == OfferTableModel.M_SelectSingle) {
			// pass the event to whichever screen is up
			int curScreen = tm.getCurrentState();
			if (curScreen == XTabManager.M_GO_OFFERS) {
				OffersPane op = tm.getOffersPane();
				op.offerSelectChanged();
			}
			else if (curScreen == XTabManager.M_GO_SEARCH) {
				SearchPane sp = tm.getSearchPane();
				sp.offerSelectChanged();
			}
			else {
				Log.error("Unexpected screen up (" + curScreen
							+ ") on offer select");
			}
		}

		/* redraw the table to force repainting any now de-selected button
		 * (the buttons are in a ButtonGroup, so side-effects are desired)
		 */
		ot.revalidate();
		ot.repaint();
		return m_CellEditor;
	}


	// method required to implement TableCellEditor
	/**
	 * obtain the cell editor Component
	 * @param offerTable the offer table we're rendering
	 * @param colValue the value of the particular column we are editing
	 * (NB: this value is itself a JCheckBox or JRadioButton)
	 * @param isSelected whether this cell (offer) is currently selected
	 * @param row the table row (i.e. which offer in the model, ignored)
	 * @param column the table column, which we validate
	 */
	public Component getTableCellEditorComponent(JTable offerTable,
																Object colValue,
																boolean isSelected,
																int row,
																int column)
	{
		if (column == 0) {
			m_CellValue = (JToggleButton) colValue;
		}
		else {
			Log.error("getTableCellEditorComponent() invoked on column " + column);
			return null;
		}
		m_CellEditor.setSelected(false);
		return m_CellEditor;
	}

	// optional method extending AbstractCellEditor
	/**
	 * see whether the cell can be edited
	 * @param anEvent the event which starts cell editing
	 * @return always true unless we're in no-select mode
	 */
	@Override
	public boolean isCellEditable(EventObject anEvent) {
		// check with the model
		if (m_OfferModel.getSelectMode() == OfferTableModel.M_SelectNone) {
			return false;
		}
		return true;
	}

	// optional method extending CellEditor
	/**
	 * state whether cell should be selected when edited
	 * @return always false
	 */
	@Override
	public boolean shouldSelectCell(EventObject anEvent) { return false; }

	// end methods
}
