/*
 * last modified---
 * 	08-05-14/jet/ Spark plugin version
 * 	03-20-14/jet/ reword Auto-connect button text
 * 	06-07-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide a UI portmanteau class for wallet preferences
 */

package org.silentvault.spark.ui;

import javax.swing.*;
import javax.swing.border.EmptyBorder;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.GridLayout;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.WalletClient;
import org.silentvault.spark.AssetPriceCache;
import org.silentvault.spark.ui.wallet.WTabManager;

import java.util.Set;
import java.util.Vector;


/**
 * This class works with VSPreferences to provide Swing UI elements and storage
 * for preference values associated with the wallet application.
 *
 * @author Kevin Wilkerson
 */
public class PreferencesUI extends JPanel {
	// BEGIN data members
	/**
	 * the application to which we belong
	 */
	private WalletClient	m_Plugin;

	/**
	 * the number of tokens to purchase in a lot (can be overridden)
	 */
	private JTextField	m_TokenLotSize;

	/**
	 * the preferred fiat currency code for pricing voucher value
	 */
	private JComboBox<String>	m_ValueCurrency;

	/**
	 * the length of time we should wait on responses from partners (in secs)
	 */
	private JTextField	m_Timeout;

	/**
	 * the length of time we should wait before inactivity logout (in minutes)
	 */
	private JTextField	m_Inactivity;

	/**
	 * whether or not we should show safe keys along with other confirmation
	 * data at safe creation
	 */
	private JRadioButton	m_ShowKeysDisplay;

	/**
	 * the domain of the last publisher to which we connected
	 */
	private JTextField	m_PrefPublisher;

	/**
	 * whether or not we should perform housekeeping automatically
	 */
	private JRadioButton	m_AutoHousekeeping;

	// END data members

	// BEGIN methods
	/**
	 * create default panel (constructor)
	 * @param plugin the plugin which owns us
	 */
	public PreferencesUI(WalletClient plugin) {
		super();
		m_Plugin = plugin;
		m_ShowKeysDisplay = new JRadioButton();
		m_PrefPublisher = new JTextField();
		m_PrefPublisher.setDropTarget(null);
		setLayout(new GridLayout());
		buildUI();
	}

	/**
	 * create UI elements and resources
	 */
	public void buildUI() {
		removeAll();
		if (m_TokenLotSize == null) {
			m_TokenLotSize = new JTextField(5);
			m_TokenLotSize.setDropTarget(null);
			m_TokenLotSize.setMinimumSize(new Dimension(30, 20));
		}

		// obtain the set of possible fiat currencies
		AssetPriceCache priceCache = m_Plugin.getPricingModule();
		Set<String> fiatCurrCodes = priceCache.getSupportedCurrencies();
		Vector<String> fiats = new Vector<String>(fiatCurrCodes.size() + 1);
		int idx = 0;
		for (String fCode : fiatCurrCodes) {
			String item = fCode + " - " + priceCache.getCurrencyName(fCode);
			fiats.add(item);
		}

		// append grams of gold and make it default valuation
		fiats.add("GAU - grams of gold");
		if (m_ValueCurrency == null || m_ValueCurrency.getItemCount() == 1) {
			m_ValueCurrency = new JComboBox<String>(fiats);
			//this line of code now causes a NPE:
			//m_ValueCurrency.setSelectedIndex(m_ValueCurrency.getItemCount()-1);
		}

		if (m_Timeout == null) {
			m_Timeout = new JTextField(5);
			m_Timeout.setDropTarget(null);
			m_Timeout.setMinimumSize(new Dimension(20, 20));
		}
		if (m_Inactivity == null) {
			m_Inactivity = new JTextField(5);
			m_Inactivity.setDropTarget(null);
			m_Inactivity.setMinimumSize(new Dimension(20, 20));
		}

		// add preference UI widgets, with labels
		setBorder(BorderFactory.createTitledBorder(
													"Preferences for Your Wallet Client"));
		JPanel contents = new JPanel(new GridBagLayout());
		contents.setBorder(null);
		EmptyBorder eb = new EmptyBorder(4, 8, 4, 8);
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.anchor = GridBagConstraints.CENTER;
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		contents.add(Box.createVerticalStrut(15), gbc);

		// token lot size
		JTextArea tokLab = new JTextArea("Enter the lot size in which you wish "
												+ "to purchase tokens.  (10-200)", 1, 35);
		tokLab.setLineWrap(false);
		tokLab.setEditable(false);
		tokLab.setBorder(eb);
		tokLab.setBackground(m_Plugin.m_LabelBackColor);
		tokLab.setFocusable(false);
		m_TokenLotSize.setToolTipText("the number of tokens you want to "
												+ "buy at one time");
		contents.add(tokLab, gbc);
		contents.add(Box.createVerticalStrut(5), gbc);
		contents.add(m_TokenLotSize, gbc);
		contents.add(Box.createVerticalStrut(10), gbc);

		// fiat display currency
		String currTxt = "Select the currency in which you prefer to see voucher "
								+ "values represented.  All vouchers are actually "
								+ "denominated in asset quantity, so this setting is "
								+ "merely for display purposes.";
		JTextArea currLab = new JTextArea(currTxt, 4, 35);
		currLab.setLineWrap(true);
		currLab.setWrapStyleWord(true);
		currLab.setEditable(false);
		currLab.setBorder(eb);
		currLab.setBackground(m_Plugin.m_LabelBackColor);
		currLab.setFocusable(false);
		contents.add(currLab, gbc);
		m_ValueCurrency.setToolTipText("pick a currency from the drop-down list");
		contents.add(Box.createVerticalStrut(5), gbc);
		contents.add(m_ValueCurrency, gbc);
		contents.add(Box.createVerticalStrut(10), gbc);

		// message reply timeout
		String timeTxt = "Enter the number of seconds which your client will "
								+ "wait for a response before timing out.  Your "
								+ "location and the quality of your internet "
								+ "connection may affect this.  For most users, the "
								+ "one minute default should be acceptable.  (10-120)";
		JTextArea timeLab = new JTextArea(timeTxt, 4, 35);
		timeLab.setLineWrap(true);
		timeLab.setWrapStyleWord(true);
		timeLab.setEditable(false);
		timeLab.setBorder(eb);
		timeLab.setBackground(m_Plugin.m_LabelBackColor);
		timeLab.setFocusable(false);
		contents.add(timeLab, gbc);
		m_Timeout.setToolTipText("seconds before a server request times out");
		contents.add(Box.createVerticalStrut(5), gbc);
		contents.add(m_Timeout, gbc);
		contents.add(Box.createVerticalStrut(10), gbc);

		// inactivity timeout
		String inacTxt = "Enter the number of minutes which your client will "
								+ "wait without message activity before automatically "
								+ "logging out of your wallet. (5-60)";
		JTextArea inactLab = new JTextArea(inacTxt, 3, 35);
		inactLab.setLineWrap(true);
		inactLab.setWrapStyleWord(true);
		inactLab.setEditable(false);
		inactLab.setBorder(eb);
		inactLab.setBackground(m_Plugin.m_LabelBackColor);
		inactLab.setFocusable(false);
		contents.add(inactLab, gbc);
		m_Inactivity.setToolTipText("<html>minutes before your wallet<br/>"
											+ "times out without user input</html>");
		contents.add(Box.createVerticalStrut(5), gbc);
		contents.add(m_Inactivity, gbc);
		contents.add(Box.createVerticalStrut(10), gbc);

		// automate housekeeping
		String autoTxt = "Check the button to auto-process incoming items found "
								+ "on the DHT at login or when Refresh is clicked.  "
								+ "Uncheck to process Refresh items manually.";
		JTextArea autoLab = new JTextArea(autoTxt, 3, 35);
		autoLab.setLineWrap(true);
		autoLab.setWrapStyleWord(true);
		autoLab.setEditable(false);
		autoLab.setBorder(eb);
		autoLab.setBackground(m_Plugin.m_LabelBackColor);
		autoLab.setFocusable(false);
		m_AutoHousekeeping
			= new JRadioButton("Automatic processing of Refresh items");
		m_AutoHousekeeping.setToolTipText("disable to process incoming items "
													+ "manually");
		contents.add(autoLab, gbc);
		contents.add(Box.createVerticalStrut(5), gbc);
		contents.add(m_AutoHousekeeping, gbc);
		contents.add(Box.createVerticalStrut(15), gbc);

		// put contents in a scrolling window in case it's too large
		JScrollPane conScroll = new JScrollPane(contents);
		conScroll.setBorder(null);
		add(conScroll);
		revalidate();
	}

	/**
	 * Method to update the set of possible currencies.
	 * (Called by AssetPriceCache every time a price fetch is performed.)
	 */
	public void updateCurrencyList() {
		// obtain the set of possible fiat currencies
		AssetPriceCache priceCache = m_Plugin.getPricingModule();
		Set<String> fiatCurrCodes = priceCache.getSupportedCurrencies();
		Vector<String> fiats = new Vector<String>(fiatCurrCodes.size() + 1);
		int idx = 0;
		for (String fCode : fiatCurrCodes) {
			String item = fCode + " - " + priceCache.getCurrencyName(fCode);
			fiats.add(item);
		}

		// append grams of gold
		fiats.add("GAU - grams of gold");

		VSPreferences prefs = m_Plugin.getUserPreferences();
		String curSelected = prefs.getValueCurrency();
		m_ValueCurrency.removeAllItems();
		for (String currency: fiats) {
			m_ValueCurrency.addItem(currency);
			if (currency.equals(curSelected)) {
				m_ValueCurrency.setSelectedItem(currency);
			}
		}
		revalidate();
	}

	/**
	 * provide read access to token lot size
	 * @return the token lot size value
	 */
	public String getTokenLotSize() { return m_TokenLotSize.getText(); }

	/**
	 * provide read access to value currency setting
	 * @return the currently set item
	 */
	public String getValueCurrency() {
		String sel = (String) m_ValueCurrency.getSelectedItem();
		return sel;
	}

	/**
	 * provide read access to reply timeout setting
	 * @return the reply timeout
	 */
	public String getTimeout() { return m_Timeout.getText(); }

	/**
	 * provide read access to inactivity setting
	 * @return the inactivity timeout
	 */
	public String getInactivity() { return m_Inactivity.getText(); }

	/**
	 * provide read access to show keys setting
	 * @return true if we should display a newly created safe's keys, else false
	 */
	public boolean getSafeKeysDisp() { return m_ShowKeysDisplay.isSelected(); }

	/**
	 * provide read access to token lot size
	 * @return the token lot size value
	 */
	public String getPrefPublisher() { return m_PrefPublisher.getText(); }

	/**
	 * provide read access to auto-housekeeping setting
	 * @return true if we should automate processing, false otherwise
	 */
	public boolean getAutoHousekeeping() {
		return m_AutoHousekeeping.isSelected();
	}


	/**
	 * write access to token lot size
	 * @param lotSize the lot size to set
	 */
	public void setTokenLotSize(String lotSize) {
		// make sure it's a number from 10-200 (in case file was edited by hand)
		Integer lot = 50;
		try {
			lot = Integer.parseInt(lotSize);
		}
		catch (NumberFormatException nfe) {
			Log.error("Invalid token lot size, " + lotSize);
		}
		int tokens = lot.intValue();
		if (tokens < 10 || tokens > 200) {
			Log.error("Invalid token lot size, " + tokens
						+ ", must be between 10 and 200");
			lotSize = "50";
		}
		m_TokenLotSize.setText(lotSize);
	}

	/**
	 * write access to value currency setting
	 * @param cCode the currency code to set as default
	 */
	public void setValueCurrency(String cCode) {
		int cnt = m_ValueCurrency.getItemCount();
		boolean set = false;
		if (cnt > 1) {
			for (int iii = 0; iii < cnt; iii++) {
				String item = m_ValueCurrency.getItemAt(iii);
				if (item.startsWith(cCode)) {
					m_ValueCurrency.setSelectedIndex(iii);
					set = true;
					break;
				}
			}
		}
		if (!set && cnt > 1) {
			// default to GAU if cCode was not found
			m_ValueCurrency.setSelectedIndex(cnt-1);
		}
	}

	/**
	 * write access to timeout setting
	 * @param timeout the number of seconds to wait before timeouts
	 */
	public void setTimeout(String timeout) {
		// make sure it's a number from 10-120 (in case file was edited by hand)
		Integer tout = 60;
		try {
			tout = Integer.parseInt(timeout);
		}
		catch (NumberFormatException nfe) {
			Log.error("Invalid timeout value, " + timeout);
		}
		int seconds = tout.intValue();
		if (seconds < 10 || seconds > 120) {
			Log.error("Invalid timeout value, " + seconds
						+ ", must be between 10 and 120 seconds");
			timeout = "60";
		}
		m_Timeout.setText(timeout);
	}

	/**
	 * write access to inactivity setting
	 * @param inactive the number of minutes to wait before auto-logout
	 */
	public void setInactivity(String inactive) {
		// make sure it's a number from 5-60 (in case file was edited by hand)
		Integer inact = 20;
		try {
			inact = Integer.parseInt(inactive);
		}
		catch (NumberFormatException nfe) {
			Log.error("Invalid inactivity value, " + inactive);
		}
		int minutes = inact.intValue();
		if (minutes < 5 || minutes > 60) {
			Log.error("Invalid inactivity value, " + minutes
						+ ", must be between 5 and 60 minutes");
			inactive = "20";
		}
		m_Inactivity.setText(inactive);
	}

	/**
	 * write access to show keys setting
	 * @param display true if we should display newly created keys, else false
	 */
	public void setSafeKeysDisp(String display) {
		Boolean disp = Boolean.parseBoolean(display);
		m_ShowKeysDisplay.setSelected(disp.booleanValue());
	}

	/**
	 * write access to last (preferred) publisher
	 * @param pub the publisher we most recently logged into
	 */
	public void setPrefPublisher(String pub) { m_PrefPublisher.setText(pub); }

	/**
	 * write access to housekeeping automation setting
	 * @param automate true if we should automate, false otherwise
	 */
	public void setAutoHousekeeping(String automate) {
		Boolean autoH = Boolean.parseBoolean(automate);
		m_AutoHousekeeping.setSelected(autoH.booleanValue());
	}

	// END methods
}
