/*
 * last modified---
 * 	07-15-15/jet/ use lookup units for relative prices
 * 	08-05-14/jet/ Spark plugin version
 * 	03-28-14/jet/ javadoc fix
 * 	06-07-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide a cell renderer for all columns in the receipt table display
 */

package org.silentvault.spark.ui;

import javax.swing.border.EmptyBorder;
import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.JCheckBox;
import javax.swing.JRadioButton;
import javax.swing.SwingConstants;
import javax.swing.UIManager;
import javax.swing.border.Border;
import javax.swing.table.TableCellRenderer;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.Dimension;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.wallet.*;

import java.util.Date;
import java.util.TreeMap;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;


/**
 * This class is used as the DefaultRenderer for all receipt table columns.
 *
 * @author Kevin Wilkerson
 */
public final class ReceiptRenderer extends JLabel implements TableCellRenderer {
	// data members
	/**
	 * the border to use for the cell
	 */
	private Border					m_Border = null;

	/**
	 * a helpful date formatter
	 */
	private SimpleDateFormat	m_DateFormatter = null;

	/**
	 * the receipt table model in support of which we render the table
	 */
	private ReceiptTableModel	m_ReceiptModel;

	/**
	 * the application or plugin of which we're a part
	 */
	private WalletClient			m_Plugin;


	// methods
	/**
	 * constructor
	 * @param receiptModel the receipt table model we are supporting
	 * @param plugin the plugin or app that we're part of
	 */
	public ReceiptRenderer(ReceiptTableModel receiptModel, WalletClient plugin) {
		super();
		m_ReceiptModel = receiptModel;
		m_Plugin = plugin;
		setOpaque(true);			// to make background show up
		m_Border = new EmptyBorder(5, 3, 5, 3);
		m_DateFormatter = new SimpleDateFormat(XMLVoucher.M_DateFmt);
	}

	/**
	 * method to implement the TableCellRenderer interface
	 * @param receiptTable the receipt table we're rendering (ignored)
	 * @param colValue the value of the particular column (String, Date, etc.)
	 * @param isSelected whether this cell (receipt) is currently selected
	 * @param hasFocus whether this cell has the focus (we ignore this)
	 * @param row the table row (i.e. which voucher in the model)
	 * @param column the table column
	 * @return the rendered Component to be drawn in the table cell
	 */
	public Component getTableCellRendererComponent(	JTable receiptTable,
																	Object colValue,
																	boolean isSelected,
																	boolean hasFocus,
																	int row,
																	int column )
	{
		// obtain the receipt data, allowing for possibility table is sorted
		int realRow = receiptTable.convertRowIndexToModel(row);
		XMLReceipt receipt = m_ReceiptModel.getReceiptAtRow(realRow);
		if (receipt == null) {
			return null;
		}
		if (colValue == null) {
			return null;
		}

		// set defaults
		setFont(m_Plugin.getWalletTabManager().getHomePane().M_TableFont);
		setForeground(UIManager.getColor("Table.foreground"));
		setHorizontalAlignment(SwingConstants.TRAILING);

		// use default background color (off-white) if row is even
		Color offWhite = UIManager.getColor("Table.alternateRowColor");
		if (row % 2 != 0) {
			setBackground(Color.WHITE);
		}
		else {
			setBackground(offWhite);
		}

		/* Examine column type:
		 * JRadioButton - it's a select control
		 * Double - it's a voucher units value
		 * Date - it's an expiration date
		 * String - it's either a type or a VS# or an amount
		 */
		if (colValue instanceof Date) {
			// this column is a receipt issue date
			Date dt = (Date) colValue;
			setText(m_DateFormatter.format(dt));
			setToolTipText("The date and time the receipt was issued");
		}
		else if (colValue instanceof JRadioButton) {
			// select column, return the button
			JRadioButton rb = (JRadioButton) colValue;
			rb.setToolTipText("Click here to select this receipt");
			rb.setOpaque(true);
			rb.setHorizontalAlignment(SwingConstants.CENTER);
			if (row % 2 != 0) {
				rb.setBackground(Color.WHITE);
			}
			else {
				rb.setBackground(offWhite);
			}
			return rb;
		}
		else if (colValue instanceof String) {
			// this could be either a type or a voucher safe number
			String cVal = (String) colValue;
			if (column == 1) {
				// incoming or outgoing, depending on payee
				setText(cVal);
				setHorizontalAlignment(SwingConstants.CENTER);
				if (cVal.equalsIgnoreCase("In")) {
					setToolTipText("<html>This receipt was issued to you<br/> by "
										+ "the indicated wallet ID.</html>");
				}
				else {
					setToolTipText("<html>This receipt was issued by you<br/> to "
										+ "the indicated wallet ID.</html>");
				}
			}
			else if (column == 2) {
				// wallet ID
				setText(cVal);
				setToolTipText("<html>The wallet ID which was the<br/>"
									+ "counterparty to the transaction represented<br/>"
									+ "by the receipt.</html>");
			}
			else {
				Log.error("String not expected for receipt column " + column);
				setText("column value error");
			}
		}
		else if (colValue instanceof Double) {
			// the amount of the receipt in base units
			Double val = (Double) colValue;
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
			df.setGroupingUsed(false);
			df.setMaximumFractionDigits(XMLReceipt.M_Precision);
			df.setMinimumFractionDigits(XMLReceipt.M_Precision);
			setText(df.format(val.doubleValue()));

			// use bold
			setFont(new Font(Font.SANS_SERIF, Font.BOLD, 13));

			// the tooltip text here must give the additional receipt details
			setToolTipText(buildReceiptDetails(receipt));
		}
		else {
			Log.error("Unexpected column type in receipt renderer, "
						+ colValue.getClass());
			setText("column type error");
		}

		// fix up the border to impose inset requirements, unless a select box
		setBorder(m_Border);

		return this;
	}

	/**
	 * helper method to build the remaining receipt details as a tool tip string
	 * @return the string containing the details
	 */
	private String buildReceiptDetails(XMLReceipt rct) {
		if (rct == null || !rct.isInit()) {
			return null;
		}
		StringBuilder details = new StringBuilder(1024);

		// prepend the equivalent amount
		String prefCurr = m_Plugin.getUserPreferences().getValueCurrency();
		AssetPriceCache apc = m_Plugin.getPricingModule();
		String dispUnits = rct.getDisplayUnits();
		String lookUnits = VPAssetConfig.getPriceLookupUnits(rct.getAsset(),
																			  rct.getUnits());
		Double assetPrice = apc.getRelativePrice(lookUnits, prefCurr);
		if (assetPrice == null) {
			assetPrice = new Double(1.0);
			prefCurr = dispUnits;
		}
		double value = assetPrice.doubleValue() * rct.getQuantity();

		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setGroupingUsed(false);
		df.setMaximumFractionDigits(XMLReceipt.M_Precision);
		df.setMinimumFractionDigits(2);
		details.append("<html>Voucher value: " + df.format(rct.getQuantity()));
		details.append(" " + rct.getUnits() + " of ");
		details.append(rct.getAsset() + ".<br/>");
		df.setMaximumFractionDigits(2);
		details.append("Equivalent market value: " + df.format(value));
		details.append(" " + prefCurr + ".<br/>");

		// also prepend the voucher serial number
		details.append("Voucher serial number: " + rct.getVoucherSerial());
		details.append(".<br/>");

		// now add all baggage fields
		TreeMap<String, String> baggage = rct.getBaggageList();
		if (baggage.isEmpty()) {
			details.append("No details for this receipt.");
		}
		else {
			HousekeepingPane hkp
				= m_Plugin.getWalletTabManager().getHousekeepingPane();
			details.append(hkp.buildBaggageTable(baggage, true));
			// NB: we don't show signatures to avoid a flurry of key fetches
		}
		details.append("</html>");
		return details.toString();
	}

	// end methods
}
