/*
 * last modified---
 * 	08-05-14/jet/ Spark plugin version
 * 	06-07-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide a cell editor for the selection column in a receipt table
 */

package org.silentvault.spark.ui;

import javax.swing.JRadioButton;
import javax.swing.JTable;
import javax.swing.table.TableCellEditor;
import javax.swing.AbstractCellEditor;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.XMLReceipt;
import org.silentvault.spark.WalletClient;
import org.silentvault.spark.ui.wallet.WTabManager;

import java.util.EventObject;


/**
 * This class is used as a specialty editor for receipt table cells which are
 * in the 0th (Select) column of the table.  It is necessary because we need
 * JRadioButton support (and DefaultCellEditor only supports the JCheckBox).
 *
 * @author Kevin Wilkerson
 */
public final class ReceiptSelector extends AbstractCellEditor
		implements TableCellEditor, ActionListener {
	// data members
	/**
	 * the editor activation command
	 */
	protected static final String	SELECT = "select";

	/**
	 * the editor component
	 */
	private JRadioButton			m_CellEditor;

	/**
	 * the value (radio button) in the table
	 */
	private JRadioButton			m_CellValue;

	/**
	 * the application or plugin of which we're a part
	 */
	private WalletClient			m_Plugin;

	/**
	 * the receipt table model in support of which we edit the cells
	 */
	private ReceiptTableModel	m_ReceiptModel;


	// methods
	/**
	 * constructor
	 * @param plugin the plugin or app that we're part of
	 * @param receiptModel the receipt table model we are supporting
	 */
	public ReceiptSelector(WalletClient plugin, ReceiptTableModel receiptModel) {
		super();
		m_Plugin = plugin;
		m_ReceiptModel = receiptModel;

		// configure the buttons
		m_CellEditor = new JRadioButton();
		m_CellEditor.setVisible(false);
		m_CellEditor.setActionCommand(SELECT);
		m_CellEditor.addActionListener(this);
		m_CellEditor.setBorderPainted(false);
	}


	// method required to implement ActionListener
	public void actionPerformed(ActionEvent e) {
		if (SELECT.equals(e.getActionCommand())) {
			if (m_CellValue == null) {
				Log.error("Null cell value on click to start edit");
			}
			else {
				m_CellEditor.setSelected(!m_CellValue.isSelected());
			}

			// tell JTable to switch back to renderer for this cell
			fireEditingStopped();
		}
		else {
			Log.error("Unexpected action on select button");
		}
	}


	// method required to implement CellEditor
	/**
	 * obtain the value contained in the editor
	 * @return the value
	 */
	public Object getCellEditorValue() {
		// propogate the selected status
		m_CellValue.setSelected(m_CellEditor.isSelected());

		// redraw the table to force repainting any now de-selected button or box
		// (the buttons are in a ButtonGroup, so side-effects are desired)

		WTabManager tm = m_Plugin.getWalletTabManager();
		JTable rt = tm.getReceiptsPane().getReceiptTable();
		Graphics tgr = rt.getGraphics();
		if (tgr != null) {
			rt.paint(tgr);
		}

		return m_CellEditor;
	}


	// method required to implement TableCellEditor
	/**
	 * obtain the cell editor Component
	 * @param receiptTable the receipt table we're rendering
	 * @param colValue the value of the particular column we are editing
	 * (NB: this value is itself a JRadioButton)
	 * @param isSelected whether this cell (receipt) is currently selected
	 * @param row the table row (i.e. which receipt in the model, ignored)
	 * @param column the table column, which we validate
	 */
	public Component getTableCellEditorComponent(JTable receiptTable,
																Object colValue,
																boolean isSelected,
																int row,
																int column)
	{
		if (column == 0) {
			m_CellValue = (JRadioButton) colValue;
		}
		else {
			Log.error("getTableCellEditorComponent() invoked on column " + column);
			return null;
		}
		m_CellEditor.setSelected(false);
		return m_CellEditor;
	}

	// optional method extending AbstractCellEditor
	/**
	 * see whether the cell can be edited
	 * @param anEvent the event which starts cell editing
	 * @return always true
	 */
	@Override
	public boolean isCellEditable(EventObject anEvent) { return true; }

	// optional method extending CellEditor
	/**
	 * state whether cell should be selected when edited
	 * @return always false
	 */
	@Override
	public boolean shouldSelectCell(EventObject anEvent) { return false; }

	// end methods
}
