/*
 * last modified---
 * 	08-15-18/kbw/ add removeReceipts()
 * 	08-05-14/jet/ Spark plugin version
 * 	06-07-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide a TableModel for the receipt JTable display in the WorkPane
 */

package org.silentvault.spark.ui;

import javax.swing.table.AbstractTableModel;
import javax.swing.JRadioButton;
import javax.swing.ButtonGroup;
import javax.swing.event.TableModelEvent;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;

import java.util.Hashtable;
import java.util.TreeMap;
import java.util.Map;
import java.util.ArrayList;
import java.util.Date;
import java.security.PublicKey;


/**
 * Extends AbstractTableModel to provide a table model which can be used to
 * display a series of decoded receipts in a scrolling window.
 *
 * @author Kevin Wilkerson
 */
public final class ReceiptTableModel extends AbstractTableModel {
	// BEGIN data members
	/**
	 * the column names
	 */
	private final String[]	M_ColumnNames = { "",
															"Type",
															"Wallet ID",
															"Value",
															"Timestamp" };

	/**
	 * the list of receipts we're representing (our data), mapped to their
	 * filenames on the SDS
	 */
	private Hashtable<XMLReceipt, String>	m_ReceiptList;

	/**
	 * list of receipts, without filenames
	 */
	private ArrayList<XMLReceipt>	m_Receipts;

	/**
	 * the owning plugin/application
	 */
	private WalletClient		m_Plugin;

	/**
	 * the public key (the payee's) we need to use to validate a given signature
	 */
	private PublicKey			m_SigKey;

	/**
	 * the list of radio buttons which correspond to the receipts and are shown
	 * in the first column
	 */
	private Hashtable<XMLReceipt, JRadioButton>	m_SelectButtons;

	/**
	 * the button group for the radio buttons, to ensure only one at a time
	 * can be checked
	 */
	private ButtonGroup		m_ButtonGroup;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param receipts the list of receipts which we are to represent
	 * @param plugin the application that owns our display
	 */
	public ReceiptTableModel(Hashtable<XMLReceipt, String> receipts,
									 WalletClient plugin)
	{
		super();
		m_ReceiptList = receipts;
		m_Plugin = plugin;
		m_SelectButtons = new Hashtable<XMLReceipt, JRadioButton>();
		m_ButtonGroup = new ButtonGroup();

		// make a button for each receipt
		for (XMLReceipt rrr : m_ReceiptList.keySet()) {
			JRadioButton rb = new JRadioButton();
			rb.setRolloverEnabled(true);
			rb.setEnabled(true);
			m_SelectButtons.put(rrr, rb);
			m_ButtonGroup.add(rb);
		}
	}

	// methods required to implement AbstractTableModel
	/**
	 * get the row count
	 * @return the count (number of receipts)
	 */
	public int getRowCount() { return m_ReceiptList.size(); }

	/**
	 * get the column count
	 * @return the count
	 */
	public int getColumnCount() { return M_ColumnNames.length; }

	/**
	 * get the name of a given column
	 * @param col the column number
	 * @return the name
	 */
	public String getColumnName(int col) {
		if (col < 0 || col > M_ColumnNames.length) {
			return null;
		}
		return M_ColumnNames[col];
	}

	/**
	 * get the value of a particular cell
	 * @param row the row (receipt) number
	 * @param col the column number
	 * @return the item at that row,col cell
	 */
	public Object getValueAt(int row, int col) {
		if (row >= m_ReceiptList.size() || row < 0 || col < 0
			 || col >= M_ColumnNames.length)
		{
			return null;
		}

		// build list out of hashtable key set
		m_Receipts = new ArrayList<XMLReceipt>(m_ReceiptList.keySet());
		XMLReceipt receipt = m_Receipts.get(row);
		VsSecrets sekrits = m_Plugin.getLoginSecrets();
		boolean incoming = !receipt.getPayee().equals(sekrits.getVSnumber());
		Object ret = null;
		switch (col) {
			case 0:	// radio button
				ret = m_SelectButtons.get(receipt);
				break;

			case 1:	// type of receipt, incoming or outgoing
				if (incoming) {
					ret = "In";
				}
				else {
					ret = "Out";
				}
				break;

			case 2:	// counterparty wallet
				String vs = null;
				if (incoming) {
					vs = new String(receipt.getPayee());
				}
				else {
					vs = new String(receipt.getPayer());
				}
				// strip @publisher
				int atIdx = vs.indexOf('@');
				if (atIdx != -1) {
					vs = vs.substring(0, atIdx);
				}
				ret = vs;
				break;

			case 3:	// amount
				Double value = new Double(receipt.getQuantity());
				ret = value;
				break;

			case 4:	// timestamp
				ret = new Date(receipt.getTimestamp() * 1000L);
				break;

			default:	// impossible
				Log.error("Impossible column number, " + col);
				break;
		}
		return ret;
	}

	/**
	 * method to determine the default renderer to use for a column
	 * @param col the column whose Class we are interested in
	 * @return the Class of object stored in that column (see getValueAt())
	 */
	public Class getColumnClass(int col) {
		Object val = getValueAt(0, col);
		if (val != null) {
			return val.getClass();
		}
		String dummy = "";
		return dummy.getClass();
	}


	/**
	 * get the receipt list
	 * @return a copy of the list (to avoid side-effects)
	 */
	public Hashtable<XMLReceipt, String> getReceiptList() {
		return new Hashtable<XMLReceipt, String>(m_ReceiptList);
	}

	/**
	 * get the receipt at a specific row
	 * @param row the row whose receipt we're interested in
	 * @return the receipt
	 */
	public XMLReceipt getReceiptAtRow(int row) {
		if (row >= m_ReceiptList.size() || row < 0) {
			return null;
		}

		// build list out of hashtable key set
		m_Receipts = new ArrayList<XMLReceipt>(m_ReceiptList.keySet());
		return m_Receipts.get(row);
	}

	/**
	 * return the selected receipt
	 * @return the receipt currently selected, or null if none is selected
	 */
	public XMLReceipt getSelectedReceipt() {
		for (XMLReceipt receipt : m_ReceiptList.keySet()) {
			JRadioButton rad = m_SelectButtons.get(receipt);
			if (rad != null && rad.isSelected()) {
				return receipt;
			}
		}
		return null;
	}

	/**
	 * add a receipt to the list
	 * @param receipt the receipt
	 * @param the filename where it's stored on the SDS
	 */
	public void addReceipt(XMLReceipt receipt, String file) {
		if (receipt != null && file != null) {
			String newFile = m_ReceiptList.put(receipt, file);
			if (newFile == null) {
				// this was added
				JRadioButton rb = new JRadioButton();
				rb.setEnabled(true);
				rb.setRolloverEnabled(true);
				m_SelectButtons.put(receipt, rb);
				m_ButtonGroup.add(rb);

				// notify listeners
				fireTableDataChanged();
			}
		}
	}

	/**
	 * remove a receipt from the list
	 * @param receipt the receipt to remove
	 */
	public void removeReceipt(XMLReceipt receipt) {
		if (receipt != null) {
			m_ReceiptList.remove(receipt);
			JRadioButton rb = m_SelectButtons.get(receipt);
			m_SelectButtons.remove(receipt);
			m_ButtonGroup.remove(rb);

			// notify listeners
			fireTableDataChanged();
		}
	}

	/**
	 * bulk receipt removal
	 * @param receipts the receipts to be removed
	 */
	public void removeReceipts(ArrayList<XMLReceipt> receipts) {
		if (receipts != null && !receipts.isEmpty()) {
			m_Receipts.removeAll(receipts);
			for (XMLReceipt rct : receipts) {
				JRadioButton rb = m_SelectButtons.get(rct);
				m_SelectButtons.remove(rct);
				m_ButtonGroup.remove(rb);
			}

			// notify listeners once
			fireTableDataChanged();
		}
	}

	/**
	 * non-destructively merge a list (map) of receipts into the list
	 * @param receiptList the receipts we need to merge
	 */
	public void mergeReceipts(Map<XMLReceipt, String> receiptList) {
		if (receiptList == null || receiptList.isEmpty()) {
			fireTableDataChanged();
			return;
		}
		m_ReceiptList.putAll(receiptList);

		// rebuild select controls
		for (XMLReceipt rrr : m_ReceiptList.keySet()) {
			if (m_SelectButtons.get(rrr) == null) {
				JRadioButton rb = new JRadioButton();
				rb.setEnabled(true);
				rb.setRolloverEnabled(true);
				m_SelectButtons.put(rrr, rb);
				m_ButtonGroup.add(rb);
			}
		}

		// notify listeners
		fireTableDataChanged();
	}

	/**
	 * deselect everything (clear all radio buttons)
	 */
	public void deselectAll() {
		for (XMLReceipt receipt : m_SelectButtons.keySet()) {
			JRadioButton rad = m_SelectButtons.get(receipt);
			if (rad != null) {
				rad.setSelected(false);
			}
		}
		m_ButtonGroup.clearSelection();
	}

	/**
	 * finalize object when garbage-collected, for data security
	 */
	@Override
	protected void finalize() throws Throwable {
		try {
			// nix sensitive data if there is any
			if (m_ReceiptList != null && !m_ReceiptList.isEmpty()) {
				m_ReceiptList.clear();
			}
			if (m_Receipts != null && !m_Receipts.isEmpty()) {
				m_Receipts.clear();
			}
		} finally {
			super.finalize();
		}
	}

	// END methods
}
