/*
 * last modified---
 * 	04-09-15/jet/ break long info and warning messages up into multiple lines
 * 	08-19-14/jet/ back merge changes from SV Spark client plugin
 * 	08-02-14/jet/ version for SV Spark plugin
 * 	06-05-13/jet/ new base class for all three tab managers
 *
 *	purpose---
 *		provide a base class of a manager for the contents of a tabbed pane tab
 */

package org.silentvault.spark.ui;

import javax.swing.JPanel;
import javax.swing.JOptionPane;
import javax.swing.JLabel;
import javax.swing.ImageIcon;
import java.awt.Dimension;
import java.awt.GridLayout;

import org.jivesoftware.MainWindow;
import org.jivesoftware.spark.SparkManager;
import org.jivesoftware.spark.Workspace;
import org.jivesoftware.spark.component.tabbedPane.*;
import org.jivesoftware.resource.SparkRes;
import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.WalletClient;

import java.util.concurrent.LinkedBlockingQueue;
import java.net.URL;


/**
 * This class is used as a common template for the three derived classes that
 * act as the manager for each tab: Wallet, Exchanges, and Marketplaces.
 * This class manages the state transition queue and provides utility functions
 * for message popups.
 *
 * @author Justin Turrell
 */
public abstract class TabManager extends JPanel {
	// BEGIN data members
	/**
	 * the plugin to which we belong
	 */
	protected WalletClient		m_Plugin;

	/**
	 * the current state value
	 */
	protected int					m_State;

	/**
	 * the actual info icon, inside a label
	 */
	public final JLabel			M_InfoIcon;

	/**
	 * the queue of state transitions which we are asked to perform
	 */
	private LinkedBlockingQueue<Integer>	m_StateQueue;

	/**
	 * whether or not we've completed login tasks for the tab (true if so)
	 */
	protected boolean				m_LoginInit;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the containing plugin
	 */
	public TabManager(WalletClient plugin) {
		super();
		m_Plugin = plugin;
		m_StateQueue = new LinkedBlockingQueue<Integer>(5);

		// to start, add the login pane to it
		setLayout(new GridLayout());
		setBorder(null);
		setPreferredSize(new Dimension(520, 475));

		// set up info icon, which must be per-tab
		URL imgURL = m_Plugin.getClass().getResource(m_Plugin.M_InfoIconName);
		ImageIcon infoIcon = null;
		if (imgURL == null) {
			Log.error("Could not find " + m_Plugin.M_InfoIconName);
		}
		else {
			infoIcon = new ImageIcon(imgURL, "info icon");
		}
		M_InfoIcon = new JLabel(infoIcon);
	}

	/**
	 * state transition method, called by threads that remove states from queue
	 * @param state the state to transition to
	 */
	protected abstract void doState(int state);

	/**
	 * obtain login init state
	 * @return true if we're processing initial login, false otherwise
	 */
	public boolean getLoginInitState() { return m_LoginInit; }

	/**
	 * query the current state (w/out reference to anything that might be queued)
	 * @return the state value now in effect
	 */
	public int getCurrentState() { return m_State; }

	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 * @param cause the cause of the error (if known)
	 * @param action the suggested action for the user to take next
	 */
	public void showError(String errMsg, String cause, String action) {
		StringBuilder msg = new StringBuilder(256);
		msg.append("Error: " + errMsg + "\n");
		if (!cause.isEmpty()) {
			msg.append("Cause: " + cause + "\n");
		}
		if (!action.isEmpty()) {
			msg.append("Suggested action: " + action + "\n");
		}
		JOptionPane.showMessageDialog(this, msg.toString(),
												"Error encountered",
												JOptionPane.ERROR_MESSAGE);
	}

	/**
	 * error display, 2 parameter version
	 * @param errMsg the message to display
	 * @param cause the cause of the error (if known)
	 */
	public void showError(String errMsg, String cause) {
		showError(errMsg, cause, "");
	}

	/**
	 * error display, 1 parameter version
	 * @param errMsg the message to display
	 */
	public void showError(String errMsg) {
		showError(errMsg, "", "");
	}

	/**
	 * display an informational message
	 * @param msg the message to display
	 */
	public void showInfo(String msg) {
		if (msg == null || msg.isEmpty()) {
			return;
		}

		// split a long message into separate lines
		final int lnMax = 50;
		StringBuilder msb = new StringBuilder(msg.length() * 2);
		if (msg.length() > lnMax) {
			int lnCnt = 0;
			String[] tokens = msg.split("\\s");
			for (int ttt = 0; ttt < tokens.length; ttt++) {
				int tokLen = tokens[ttt].length();
				if (lnCnt + tokLen <= lnMax) {
					msb.append(tokens[ttt] + " ");
					lnCnt += (tokLen + 1);
				}
				else {
					// place on new line
					msb.append("\n" + tokens[ttt] + " ");
					lnCnt = tokLen + 1;
				}
			}
		}
		else {
			msb.append(msg);
		}
		JOptionPane.showMessageDialog(this, msb, "Information",
												JOptionPane.INFORMATION_MESSAGE);
	}

	/**
	 * display a warning message
	 * @param msg the message to display
	 */
	public void showWarning(String msg) {
		if (msg == null || msg.isEmpty()) {
			return;
		}

		// split a long message into separate lines
		final int lnMax = 50;
		StringBuilder msb = new StringBuilder(msg.length() * 2);
		if (msg.length() > lnMax) {
			int lnCnt = 0;
			String[] tokens = msg.split("\\s");
			for (int ttt = 0; ttt < tokens.length; ttt++) {
				int tokLen = tokens[ttt].length();
				if (lnCnt + tokLen <= lnMax) {
					msb.append(tokens[ttt] + " ");
					lnCnt += (tokLen + 1);
				}
				else {
					// place on new line
					msb.append("\n" + tokens[ttt] + " ");
					lnCnt = tokLen + 1;
				}
			}
		}
		else {
			msb.append(msg);
		}
		JOptionPane.showMessageDialog(this, msb, "Warning",
												JOptionPane.WARNING_MESSAGE);
	}

	/**
	 * method to add a state transition to the queue (will block on full queue).
	 * NB: this method is called by event handling threads
	 * @param state the state transition to add
	 */
	public void queueTransition(Integer state) {
		if (state == null || state.intValue() == 0) {
			Log.error("Invalid state queue request");
			return;
		}
		try {
			m_StateQueue.put(state);
		}
		catch (InterruptedException ie) {
			Log.error("Got interrupt queueing state transition to " + state, ie);
		}
	}

	/**
	 * method to process the state transition at the head of the queue.
	 * NB: this method is called by sub-threads spawned by event handling threads
	 * NB: it is possible that the sub-thread might pull out and process a
	 * 	 different state than what was queued by its parent.  This is not a
	 * 	 problem so long as the number of created threads always matches the
	 * 	 number of queued states.
	 */
	public synchronized void makeTransition() {
		Integer qHead = m_StateQueue.poll();
		if (qHead == null) {
			Log.error("Empty TabManager state transition queue");
		}
		else {
			doState(qHead.intValue());
		}
	}

	/**
	 * method to cause repetition of login tasks
	 */
	protected void resetFirstLogin() { m_LoginInit = false; }

	/**
	 * method to deallocate all sub-panes (called by WalletClient destructor)
	 */
	public abstract void deallocatePanes();

	// END methods
}
