/*
 * last modified---
 * 	08-05-14/jet/ Spark plugin version
 * 	06-07-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide a cell renderer for validity columns in the token table display
 */

package org.silentvault.spark.ui;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.SwingConstants;
import javax.swing.border.Border;
import javax.swing.table.TableCellRenderer;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.XMLToken;

import java.text.NumberFormat;
import java.text.DecimalFormat;


/**
 * This class is used as the DefaultRenderer for token table columns of one
 * types: Boolean, representing the signature validity of the token.
 * Invalid tokens have this column drawn in a warning color.
 *
 * @author Kevin Wilkerson
 */
public class TokenRenderer extends JLabel implements TableCellRenderer {
	// data members
	/**
	 * the border to use when the cell is not selected
	 */
	private Border					m_UnselectedBorder = null;

	/**
	 * the border to use when the sell is selected
	 */
	private Border					m_SelectedBorder = null;

	/**
	 * whether or not to use a border at all
	 */
	private boolean				m_Bordered = true;


	// methods
	/**
	 * constructor
	 * @param bordered whether or not this cell should be drawn with a border
	 */
	public TokenRenderer(boolean bordered) {
		super();
		m_Bordered = bordered;
		setOpaque(true);			// to make background show up
	}

	/**
	 * method to implement the TableCellRenderer interface
	 * @param tokenTable the token table we're rendering
	 * @param colValue the value of the particular column (Boolean)
	 * @param isSelected whether this cell (token) is currently selected
	 * @param hasFocus whether this cell has the focus (we ignore this)
	 * @param row the table row (i.e. which token, ignored)
	 * @param column the table column (we ignore this)
	 */
	public Component getTableCellRendererComponent(	JTable tokenTable,
																	Object colValue,
																	boolean isSelected,
																	boolean hasFocus,
																	int row,
																	int column )
	{
		// allow for possibility table is sorted
		int realRow = tokenTable.convertRowIndexToModel(row);

		// see which case we're doing
		if (colValue instanceof Boolean) {
			// signature validity
			Boolean valid = (Boolean) colValue;
			if (valid.booleanValue()) {
				// it's valid, use green
				setBackground(Color.GREEN);
				setText("Yes");
				setToolTipText("<html>This token bears a valid digital signature"
									+ "<br/>from the voucher publisher</html>");
			}
			else {
				// it's not valid, red flag it
				setBackground(Color.RED);
				setText("No");
				setToolTipText("This token's signature could not be validated!");
			}
			setForeground(Color.BLACK);
			setHorizontalAlignment(SwingConstants.CENTER);
		}
		else if (colValue instanceof Double) {
			Double quan = (Double) colValue;
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
			df.setGroupingUsed(false);
			df.setMaximumFractionDigits(XMLToken.M_Precision);
			df.setMinimumFractionDigits(XMLToken.M_Precision);
			setText(df.format(quan.doubleValue()));
			setFont(new Font(Font.SANS_SERIF, Font.PLAIN, 14));

			// get the token in question and see if it's spent or commmitted to be
			TokenTableModel ttm = (TokenTableModel) tokenTable.getModel();
			if (ttm.isAvailable(realRow)) {
				setToolTipText("This token is available to pay for operations");
				setBackground(Color.WHITE);
			}
			else {
				setBackground(Color.DARK_GRAY);
				setToolTipText("<html>This token has been spent, or is committed"
									+ "<br/>to be spent on a pending operation</html>");
			}
			setHorizontalAlignment(SwingConstants.LEFT);
		}
		else {
			Log.error("Unexpected column type in token renderer, "
						+ colValue.getClass());
		}

		// fix up the border as needed, using matte borders
		if (m_Bordered) {
			if (isSelected) {
				if (m_SelectedBorder == null) {
					m_SelectedBorder = BorderFactory.createMatteBorder(2, 5, 2, 5,
														tokenTable.getSelectionBackground());
				}
				setBorder(m_SelectedBorder);
			}
			else {
				if (m_UnselectedBorder == null) {
					m_UnselectedBorder = BorderFactory.createMatteBorder(2, 5, 2, 5,
																	tokenTable.getBackground());
				}
				setBorder(m_UnselectedBorder);
			}
		}
		return this;
	}
}
