/*
 * last modified---
 * 	08-04-14/jet/ version for Spark plugin
 * 	06-07-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide a class to manage SilentVault plugin preferences
 */

package org.silentvault.spark.ui;

import org.jivesoftware.spark.preference.Preference;
import org.jivesoftware.spark.util.log.Log;
import org.jivesoftware.resource.SparkRes;
import org.jivesoftware.Spark;

import org.silentvault.spark.WalletClient;
import org.silentvault.spark.ui.wallet.TokensPane;

import javax.swing.*;
import java.awt.Color;

import java.io.*;
import java.util.Properties;


/**
 * This class implements the Preference interface to supply an object which
 * can be registered with the Spark Preferences Manager.  It utilizes the UI
 * class PreferencesUI to contain the data values.  The persistent storage is
 * provided by the file ~/.silentvault/vouchersafe.properties.
 *
 * @author Kevin Wilkerson
 */
public class VSPreferences implements Preference {
	// BEGIN data members
	/**
	 * name of properties file in which preferences are saved
	 */
	private static final String	M_PrefFile = "silentvaultspark.properties";

	// key values for preferences stored as Properties
	/**
	 * default lot size for token purchases
	 */
	private static final String	M_TokLotKey = "tokenLotSize";
	/**
	 * fiat currency display preference
	 */
	private static final String	M_ValCurrKey = "valueCurrency";
	/**
	 * timeout for replies from the OFS
	 */
	private static final String	M_TimeoutKey = "replyTimeout";
	/**
	 * timeout after which, with no activity, session will auto-terminate
	 */
	private static final String	M_InactivityKey = "sessionTimeout";
	/**
	 * whether or not to display the safe's keys upon creation confirmation
	 */
	private static final String	M_SafeKeys = "showSafeKeys";
	/**
	 * the last publisher selected at login
	 */
	private static final String	M_PrefPub = "prefPublisher";
	/**
	 * whether or not to do automated housekeeping
	 */
	private static final String	M_AutoKey = "autoHousekeeping";
	

	/**
	 * the plugin which owns us
	 */
	private WalletClient			m_Plugin;

	/**
	 * the UI element
	 */
	private PreferencesUI		m_PrefUI;

	/**
	 * the actual file in which preferences are saved
	 */
	private File					m_PrefFile;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin our owning plugin
	 */
	public VSPreferences(WalletClient plugin) {
		m_Plugin = plugin;
		m_PrefUI = new PreferencesUI(m_Plugin);

		// make sure the save file exists
		m_PrefFile = new File(m_Plugin.getLogDir().getParentFile(), M_PrefFile);
		if (!m_PrefFile.exists()) {
			try {
				if (m_PrefFile.createNewFile()) {
					Log.debug("Created preferences save file " + m_PrefFile);
				}
				else {
					Log.error("Could not create preferences save file "
								+ m_PrefFile);
				}
			}
			catch (IOException ioe) {
				Log.error("Error creating preferences save file", ioe);
			}
		}
		else {
			//Log.debug("Using existing preferences save file " + m_PrefFile);
		}
	}

	// methods required to implement the Preference interface
	/**
	 * obtain the title of these preferences
	 * @return the title
	 */
	public String getTitle() { return "SilentVault Preferences"; }

	/**
	 * obtain the associated icon
	 * @return the icon
	 */
	public Icon getIcon() {
		return m_Plugin.M_SVIcon.getIcon();
	}

	/**
	 * obtain the generic tooltip
	 * @return the tooltip
	 */
	public String getTooltip() {
		return "Enter your SilentVault preference settings";
	}

	/**
	 * obtain the name of the preferences list
	 * @return the name
	 */
	public String getListName() {
		return "SilentVault";
	}

	/**
	 * obtain the namespace of the preferences list
	 * @return the namespace
	 */
	public String getNamespace() {
		return "SV-Preferences";
	}

	/**
	 * obtain the supporting GUI component
	 * @return the UI object
	 */
	public JComponent getGUI() { return m_PrefUI; }

	/**
	 * load in the current values from persistent (file) storage
	 */
	public void load() {
		FileInputStream fis = null;
		Properties fileProps = new Properties();
		try {
			fis = new FileInputStream(m_PrefFile);
			fileProps.load(fis);
			fis.close();
		}
		catch (FileNotFoundException fnfe) {
			Log.error("Could not find preferences file, " + m_PrefFile, fnfe);
			return;
		}
		catch (IOException ioe) {
			Log.error("Could not read preferences file", ioe);
			return;
		}
		m_PrefUI.setTokenLotSize(fileProps.getProperty(M_TokLotKey, "50"));
		m_PrefUI.setValueCurrency(fileProps.getProperty(M_ValCurrKey, "GAU"));
		m_PrefUI.setTimeout(fileProps.getProperty(M_TimeoutKey, "60"));
		m_PrefUI.setInactivity(fileProps.getProperty(M_InactivityKey, "20"));
		m_PrefUI.setSafeKeysDisp(fileProps.getProperty(M_SafeKeys, "false"));
		m_PrefUI.setPrefPublisher(fileProps.getProperty(M_PrefPub,
																		"@voucher-safe.org"));
		m_PrefUI.setAutoHousekeeping(fileProps.getProperty(M_AutoKey, "true"));
	}

	/**
	 * save the current values to persistent (file) storage
	 */
	public void commit() {
		// build properties hashtable with current values set
		Properties saveProps = new Properties();
		Integer tokLot = new Integer(getTokenLotSize());
		saveProps.setProperty(M_TokLotKey, tokLot.toString());
		saveProps.setProperty(M_ValCurrKey, getValueCurrency());
		Long tmOut = new Long(getTimeout());
		saveProps.setProperty(M_TimeoutKey, tmOut.toString());
		Integer inact = new Integer(getInactivity());
		saveProps.setProperty(M_InactivityKey, inact.toString());
		Boolean showKeys = new Boolean(getSafeKeysDisp());
		saveProps.setProperty(M_SafeKeys, showKeys.toString());
		saveProps.setProperty(M_PrefPub, getPrefPublisher());
		Boolean autoH = new Boolean(getAutoHousekeeping());
		saveProps.setProperty(M_AutoKey, autoH.toString());

		// save to file
		FileOutputStream fos = null;
		try {
			fos = new FileOutputStream(m_PrefFile);
			saveProps.store(fos, "SilentVault Plugin preferences settings");
			fos.close();
		}
		catch (FileNotFoundException fnfe) {
			Log.error("Could not write to preferences file, " + m_PrefFile, fnfe);
		}
		catch (IOException ioe) {
			Log.error("Could not write preferences file", ioe);
		}
	}

	/**
	 * data validator - stub
	 * @return true iff valid
	 */
	public boolean isDataValid() {
		return true;
	}

	/**
	 * obtain context-specific error message - stub
	 * @return the message
	 */
	public String getErrorMessage() {
		return null;
	}

	/**
	 * obtain associated data model - stub
	 * @return the data
	 */
	public Object getData() {
		return null;
	}

	/**
	 * handle shutdown - stub
	 */
	public void shutdown() {
		// do nothing
	}

	// get methods to obtain current preference values
	/**
	 * obtain the token lot size
	 * @return the number of tokens user prefers to purchase at a time
	 */
	public int getTokenLotSize() {
		String tokLot = m_PrefUI.getTokenLotSize();
		if (tokLot.isEmpty()) {
			return 50;
		}
		Integer lotSz = null;
		try {
			lotSz = Integer.parseInt(tokLot);
		}
		catch (NumberFormatException nfe) {
			Log.error("Invalid format for token lot size, " + tokLot);
			lotSz = new Integer(50);
			m_PrefUI.setTokenLotSize(lotSz.toString());
		}
		// make sure it's between 10 and 200 tokens
		int tokens = lotSz.intValue();
		if (tokens < TokensPane.M_MinTokens || tokens > TokensPane.M_MaxTokens) {
			Log.error("Invalid token lot size, " + tokens
						+ ", must be between " + TokensPane.M_MinTokens + " and "
						+ TokensPane.M_MaxTokens);
			lotSz = new Integer(50);
			m_PrefUI.setTokenLotSize(lotSz.toString());
		}
		return lotSz.intValue();
	}

	/**
	 * obtain value currency code
	 * @return the currency code the user wants things priced in
	 */
	public String getValueCurrency() {
		String codeVal = m_PrefUI.getValueCurrency();
		if (codeVal.isEmpty()) {
			return "GAU";
		}
		int descIdx = codeVal.indexOf(" - ");
		return codeVal.substring(0, descIdx);
	}

	/**
	 * obtain default timeout
	 * @return the timeout value, in seconds
	 */
	public long getTimeout() {
		String tOut = m_PrefUI.getTimeout();
		if (tOut.isEmpty()) {
			return 60L;
		}
		Long timeout = null;
		try {
			timeout = Long.parseLong(tOut);
		}
		catch (NumberFormatException nfe) {
			Log.error("Invalid format for timeout value, " + tOut);
			timeout = new Long(60L);
			m_PrefUI.setTimeout(timeout.toString());
		}
		// make sure it's between 10 and 120 seconds
		long seconds = timeout.longValue();
		if (seconds < 10 || seconds > 120) {
			Log.error("Invalid timeout value, " + seconds
						+ ", must be between 10 and 120 seconds");
			timeout = new Long(60L);
			m_PrefUI.setTimeout(timeout.toString());
		}
		return timeout.longValue();
	}

	/**
	 * obtain default inactivity timeout
	 * @return the timeout value, in minutes
	 */
	public int getInactivity() {
		String inac = m_PrefUI.getInactivity();
		if (inac.isEmpty()) {
			return 20;
		}
		Integer inactivity = null;
		try {
			inactivity = Integer.parseInt(inac);
		}
		catch (NumberFormatException nfe) {
			Log.error("Invalid format for inactivity value, " + inac);
			inactivity = new Integer(20);
			m_PrefUI.setInactivity(inactivity.toString());
		}
		// make sure it's between 5 and 60 minutes
		int minutes = inactivity.intValue();
		if (minutes < 5 || minutes > 60) {
			Log.error("Invalid inactivity value, " + minutes
						+ ", must be between 5 and 60 minutes");
			inactivity = new Integer(20);
			m_PrefUI.setInactivity(inactivity.toString());
		}
		return inactivity.intValue();
	}

	/**
	 * obtain the default generated safe keys display value
	 * @return true if we should show keys with other confirmation, else false
	 */
	public boolean getSafeKeysDisp() {
		boolean display = m_PrefUI.getSafeKeysDisp();
		return display;
	}

	/**
	 * obtain the user's preferred (last connected publisher)
	 * @return the publisher domain
	 */
	public String getPrefPublisher() { return m_PrefUI.getPrefPublisher(); }

	/**
	 * obtain the default auto-housekeeping value
	 * @return true if housekeeping should be done automatically, else false
	 */
	public boolean getAutoHousekeeping() {
		boolean autoH = m_PrefUI.getAutoHousekeeping();
		return autoH;
	}

	// END methods
}
