/*
 * last modified---
 * 	02-23-18/kbw/ use grouping size 3
 * 	01-24-18/kbw/ display actual voucher amount and units in tooltip, not base
 * 	08-05-14/jet/ Spark plugin version
 * 	03-28-14/jet/ align row colors in Show Vouchers display
 * 	06-07-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide a cell renderer for all columns in the voucher table display
 */

package org.silentvault.spark.ui;

import javax.swing.border.EmptyBorder;
import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.JCheckBox;
import javax.swing.JRadioButton;
import javax.swing.SwingConstants;
import javax.swing.UIManager;
import javax.swing.border.Border;
import javax.swing.table.TableCellRenderer;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.Dimension;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.XMLVoucher;
import org.silentvault.spark.WalletClient;

import java.util.Date;
import java.util.Calendar;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import javax.management.timer.Timer;


/**
 * This class is used as the DefaultRenderer for all voucher table columns.
 * Invalid, expired, or nearly expired vouchers have their expiration date
 * drawn in warning colors.
 *
 * @author Kevin Wilkerson
 */
public final class VoucherRenderer extends JLabel implements TableCellRenderer {
	// data members
	/**
	 * the border to use for the cell
	 */
	private Border					m_Border = null;

	/**
	 * a helpful date formatter
	 */
	private SimpleDateFormat	m_DateFormatter = null;

	/**
	 * the voucher table model in support of which we render the table
	 */
	private VoucherTableModel	m_VoucherModel;

	/**
	 * the application or plugin of which we're a part
	 */
	private WalletClient			m_Plugin;

	// methods
	/**
	 * constructor
	 * @param vouchModel the voucher table model we are supporting
	 * @param plugin the plugin or app that we're part of
	 */
	public VoucherRenderer(VoucherTableModel vouchModel, WalletClient plugin) {
		super();
		m_VoucherModel = vouchModel;
		m_Plugin = plugin;
		setOpaque(true);			// to make background show up
		m_Border = new EmptyBorder(5, 3, 5, 3);
		m_DateFormatter = new SimpleDateFormat(XMLVoucher.M_DateFmt);
	}

	/**
	 * method to implement the TableCellRenderer interface
	 * @param vouchTable the voucher table we're rendering
	 * @param colValue the value of the particular column (String, Date, etc.)
	 * @param isSelected whether this cell (voucher) is currently selected
	 * @param hasFocus whether this cell has the focus (we ignore this)
	 * @param row the table row (i.e. which voucher in the model)
	 * @param column the table column (we ignore this)
	 * @return the rendered Component to be displayed in the cell
	 */
	public Component getTableCellRendererComponent(	JTable vouchTable,
																	Object colValue,
																	boolean isSelected,
																	boolean hasFocus,
																	int row,
																	int column )
	{
		// obtain the voucher data, allowing for possibility table is sorted
		int realRow = vouchTable.convertRowIndexToModel(row);
		XMLVoucher voucher = m_VoucherModel.getVoucherAtRow(realRow);
		if (voucher == null) {
			return null;
		}
		if (colValue == null) {
			return null;
		}

		// check signature validity
		boolean vouchValid = voucher.validateSig(m_VoucherModel.getSigKey());

		// set defaults
		setFont(m_Plugin.getWalletTabManager().getHomePane().M_TableFont);
		setForeground(UIManager.getColor("Table.foreground"));
		setHorizontalAlignment(SwingConstants.TRAILING);

		// use default background color (off-white) if row is even
		Color offWhite = UIManager.getColor("Table.alternateRowColor");
		if (row % 2 != 1) {
			setBackground(Color.WHITE);
		}
		else {
			setBackground(offWhite);
		}

		// override these defaults if the voucher is expired or close to it
		Date voucherExp = voucher.getExpiration();
		Calendar cal = Calendar.getInstance();
		int offGMTmillis = cal.get(Calendar.ZONE_OFFSET);
		int offDSTmillis = cal.get(Calendar.DST_OFFSET);
		// adjust for timezone
		Date expDate2
			= new Date(voucherExp.getTime() + offGMTmillis + offDSTmillis);
		Date now = new Date();
		Date thirtyDaysHence = new Date(now.getTime() + (Timer.ONE_DAY * 30L));
		Date tenDaysHence = new Date(now.getTime() + (Timer.ONE_DAY * 10L));
		if (expDate2.before(now)) {
			// voucher has expired, draw in ugly gray foreground, red background
			setBackground(Color.RED);
			setForeground(Color.DARK_GRAY);
		}
		else if (expDate2.before(tenDaysHence)) {
			// voucher expires in under ten days, draw in orangeish background
			setBackground(new Color(0xFEA98D));
		}
		else if (expDate2.before(thirtyDaysHence)) {
			// voucher expires in under thirty days, draw in yellow background
			setBackground(new Color(0xFEE26F));
		}

		/* Examine column type:
		 * JToggleButton - it's a select control (col 0)
		 * Double - it's a asset quantity value (col 1) -or-
		 * Double - it's a fiat equivalent value (col 2)
		 * Date - it's an expiration date (col 3)
		 */
		if (colValue instanceof Date) {
			// this column is a voucher expiration date (== voucherExp)
			Date expDate = (Date) colValue;
			long secs = (expDate.getTime() - now.getTime()) / 1000L;
			long days = secs / 86400L;
			expDate2 = new Date(expDate.getTime() + offGMTmillis + offDSTmillis);
			if (expDate2.before(now)) {
				setToolTipText("This voucher has expired!");
			}
			else if (expDate2.before(tenDaysHence)) {
				setToolTipText("<html>This voucher expires in " + days + " days."
									+ "<br/>Please revalidate it to prevent loss of "
									+ "value.</html>");
				// use bold italic
				setFont(new Font(Font.SANS_SERIF, Font.BOLD | Font.ITALIC, 13));
			}
			else if (expDate2.before(thirtyDaysHence)) {
				setToolTipText("<html>This voucher expires in " + days + " days."
									+ "<br/>Please revalidate it to prevent loss of "
									+ "value.</html>");
				// use italic
				setFont(new Font(Font.SANS_SERIF, Font.ITALIC, 13));
			}
			else {
				// voucher has more than thirty days, draw plain and show days
				setToolTipText("This voucher expires in " + days + " days.");
			}
			setText(m_DateFormatter.format(expDate));
		}
		else if (colValue instanceof Double && column == 1) {
			// amount of voucher in base units
			Double val = (Double) colValue;
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
			df.setGroupingSize(3);
			df.setMaximumFractionDigits(XMLVoucher.M_Precision);
			df.setMinimumFractionDigits(1);
			setText(df.format(val.doubleValue()));

			// use bold
			setFont(new Font(Font.SANS_SERIF, Font.BOLD, 13));

			// the tooltip text here must give the additional voucher details
			StringBuilder ttb = new StringBuilder(256);
			ttb.append("<html>");
			ttb.append("Amount: " + df.format(voucher.getQuantity()) + " ");
			ttb.append(voucher.getUnits() + " of ");
			ttb.append(voucher.getAsset() + ".<br/>");
			ttb.append("Serial number: " + voucher.getSerialNumber() + ".<br/>");
			ttb.append("Generation since minting: " + voucher.getGeneration());
			ttb.append(".<br/>");
			if (vouchValid) {
				ttb.append("Signature is valid.");
			}
			else {
				setBackground(Color.RED);
				ttb.append("Signature is invalid!");
			}
			ttb.append("</html>");
			setToolTipText(ttb.toString());
		}
		else if (colValue instanceof Double && column == 2) {
			// amount of voucher in fiat equivalent
			Double val = (Double) colValue;
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
			df.setMaximumFractionDigits(2);
			df.setMinimumFractionDigits(2);
			df.setGroupingSize(3);
			setText(df.format(val.doubleValue()));

			// the tooltip text here must give value qualification message
			String currPref = m_Plugin.getUserPreferences().getValueCurrency();
			StringBuilder ttb = new StringBuilder(256);
			ttb.append("<html>");
			ttb.append("Equivalent: " + df.format(val.doubleValue()) + " ");
			ttb.append(currPref + ".<br/>");
			ttb.append("Note: this is an estimated value<br/>");
			ttb.append("based on available exchange rates.<br/>");
			ttb.append("The true value of your voucher is<br/>");
			ttb.append("reflected in the Value column.");
			ttb.append("</html>");
			setToolTipText(ttb.toString());

			// use italic
			Font itFont = new Font(Font.SANS_SERIF, Font.ITALIC, 13);
			setFont(itFont);
		}
		else if (colValue instanceof JCheckBox) {
			// unless we're in multi-select mode, return null here
			JCheckBox cb = (JCheckBox) colValue;
			if (m_VoucherModel.getSelectMode() == VoucherTableModel.M_SelectMulti)
			{
				cb.setToolTipText("<html>Click here to include this voucher<br/>"
										+ "in your list of selected vouchers</html>");
				cb.setText(" ");
				cb.setOpaque(true);
				cb.setHorizontalAlignment(SwingConstants.CENTER);
				if (row % 2 != 1) {
					cb.setBackground(Color.WHITE);
				}
				else {
					cb.setBackground(offWhite);
				}
				return cb;
			}
			return null;
		}
		else if (colValue instanceof JRadioButton) {
			// unless we're in single-select mode, return null here
			JRadioButton rb = (JRadioButton) colValue;
			if (m_VoucherModel.getSelectMode() == VoucherTableModel.M_SelectSingle)
			{
				rb.setToolTipText("Click here to select this voucher");
				rb.setText(" ");
				rb.setOpaque(true);
				rb.setHorizontalAlignment(SwingConstants.CENTER);
				if (row % 2 != 1) {
					rb.setBackground(Color.WHITE);
				}
				else {
					rb.setBackground(offWhite);
				}
				return rb;
			}
			return null;
		}
		else {
			Log.error("Unexpected column type in voucher renderer, "
						+ colValue.getClass());
			setText("column type error");
		}

		// fix up the border to impose inset requirements, unless a select box
		setBorder(m_Border);

		return this;
	}

	// end methods
}
