/*
 * last modified---
 * 	08-24-16/jet/ handle pluginVersion
 * 	07-09-15/jet/ support SVM plugins as independent within Spark
 * 	02-18-15/jet/ flesh out messaging for login, create, and logout
 * 	01-15-15/jet/ replace Connect button with Login and Create; remove blurb
 * 	10-02-14/jet/ SSM -> SVM
 * 	08-04-14/jet/ version for Spark plugin
 * 	09-12-13/jet/ reword placeholder language
 * 	07-04-13/jet/ new
 *
 *	purpose---
 *		provide the contents of the SVM connect/login screen
 */

package org.silentvault.spark.ui.svm;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.util.StringUtils;

import javax.swing.*;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.text.Document;
import javax.swing.text.JTextComponent;
import javax.swing.text.BadLocationException;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import org.jivesoftware.spark.util.log.Log;
import org.jivesoftware.spark.plugin.Plugin;
import org.jivesoftware.spark.plugin.PublicPlugin;
import org.jivesoftware.spark.PluginManager;

import org.silentvault.spark.*;

import java.util.List;
import java.util.Vector;
import java.security.PublicKey;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.net.HttpURLConnection;


/**
 * This class extends JPanel to provide a screen for selecting the SVM to
 * login to.  It also maintains the list of available SVMs, which is
 * populated after OFS connection by a discovery request.
 *
 * @author Justin Turrell
 */
public final class LoginPane extends JPanel implements ActionListener {
	// BEGIN data members
	/**
	 * wait cursor
	 */
	private final Cursor		M_WaitCursor
										= Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR);
	
	/**
	 * the client application to which we belong
	 */
	private WalletClient		m_Plugin;

	/**
	 * the tab manager which creates and manages us
	 */
	private MTabManager		m_TabManager;

	/**
	 * the session ID value set by our SVM server, post-login or create
	 */
	private String				m_SessionId;

	/**
	 * the challenge from the SVM (encrypted and base64-encoded)
	 */
	private String				m_Challenge;

	/**
	 * prompt for specifying the SVM Id
	 */
	private JComboBox<String>	m_SVMselector;

	/**
	 * the PIN entered by the user
	 */
	private String				m_PIN;

	/**
	 * the prompt for entering the PIN
	 */
	private JLabel				m_PINprompt;

	/**
	 * the UI object for the PIN
	 */
	private JPasswordField	m_PINUI;
	
	/**
	 * whether a PIN has been entered yet
	 */
	private boolean			m_HavePIN;

	/**
	 * the action for handling login attempts
	 */
	private AbstractAction	m_LoginAction;

	/**
	 * the login button
	 */
	private JButton			m_LoginButton;

	/**
	 * the action for handling create attempts
	 */
	private AbstractAction	m_CreateAction;

	/**
	 * the create button
	 */
	private JButton			m_CreateButton;

	/**
	 * the scroll area used to ensure that this pane always displays everything
	 * regardless of window size
	 */
	private JScrollPane		m_ScrollPane;

	/**
	 * all discovered SVMs
	 */
	private Vector<String>	m_SVMIdList;

	/**
	 * track whether an undo is in progress
	 */
	private boolean		m_DoingUndo;

	/**
	 * inner class to provide a DocumentListener for the m_PINUI field
	 */
	final class PINDocumentListener implements DocumentListener {
		/**
		 * handle insert notifications
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			Document doc = de.getDocument();
			String curTxt = "";

			// see if value parses
			if (parseValue()) {
				m_HavePIN = true;
			}
			else {
				m_HavePIN = false;

				// undo this edit
				int changeAt = de.getOffset();
				try {
					curTxt = doc.getText(0, doc.getLength());
					String modTxt = curTxt.substring(0, changeAt)
										+ curTxt.substring(changeAt+1);
					queueTextMod(m_PINUI, modTxt);
				}
				catch (BadLocationException ble) {
					Log.error("Cannot undo bad PIN edit", ble);
				}
			}
			updateConnectButtonStatus();
		}

		/**
		 * handle removal notifications
		 * @param de the deletion event
		 */
		public void removeUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			// see if value parses
			m_HavePIN = parseValue();
			updateConnectButtonStatus();
		}

		/**
		 * handle change notifications (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JPasswordField");
		}

		/**
		 * helper method to parse the current value of m_PINUI as a valid PIN
		 */
		private boolean parseValue() {
			// examine what's in the PIN entry field
			char[] pin = m_PINUI.getPassword();
			String enteredPIN = new String(pin).trim();
			if (enteredPIN.isEmpty()) {
				return false;
			}

			// format allows l to specify long, but we don't
			if (enteredPIN.indexOf('l') != -1 || enteredPIN.indexOf('L') != -1) {
				return false;
			}

			// disallow +-
			if (enteredPIN.startsWith("-") || enteredPIN.startsWith("+")) {
				return false;
			}

			// attempt to parse, very strictly
			int vvv = 0;
			try {
				Integer val = Integer.parseInt(enteredPIN);
				vvv = val.intValue();
			}
			catch (NumberFormatException nfe) {
				return false;
			}

			// NB: length should be 5 or more digits, but we allow leading zeros
			if (vvv > 99999) {
				return false;
			}
			return true;
		}
	}

	/**
	 * document listener for the m_PINUI entry field
	 */
	private PINDocumentListener	m_PINListener;

	/**
	 * inner class to modify a text field from within its document listener,
	 * without triggering a recursive modification exception
	 */
	class TextComponentMod implements Runnable {
		/**
		 * the text field being modified
		 */
		JTextComponent	m_ModComponent;

		/**
		 * the new replacement contents to be set
		 */
		String			m_NewContents;

		/**
		 * constructor
		 * @param txtComp the text component to update
		 * @param newText the replacement text
		 */
		public TextComponentMod(JTextComponent txtComp, String newText) {
			m_ModComponent = txtComp;
			m_NewContents = new String(newText);
		}

		/**
		 * execute main task in background
		 */
		public void run() {
			try { Thread.sleep(25); }
			catch (InterruptedException ie) { /* ignore */ }
			m_ModComponent.setText(m_NewContents);
			m_DoingUndo = false;
			m_ModComponent.setCaretPosition(m_NewContents.length());
			updateConnectButtonStatus();
		}
	}

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param client the owning application
	 * @param mgr the tab manager which creates us
	 */
	@SuppressWarnings("unchecked")
	public LoginPane(WalletClient client, MTabManager mgr) {
		super();
		m_Plugin = client;
		m_TabManager = mgr;
		m_SVMIdList = new Vector<String>();
		// NB: will be replaced in recordSVMList()
		m_SVMselector = new JComboBox<String>();
		m_PIN = m_SessionId = m_Challenge = "";

		// use grid layout
		setLayout(new GridLayout());
		setBorder(null);

		// define the action for performing a login
		m_LoginAction = new AbstractAction("Login") {
			public void actionPerformed(ActionEvent ae) {
				if (m_SVMselector.getItemCount() == 0) {
					m_TabManager.showError("cannot login to SVM",
													"no available SVM found",
													"try a different OFS gateway");
					return;
				}

				// disallow any SVM which does not match our logged-in VP domain
				VsSecrets secrets = m_Plugin.getLoginSecrets();
				String vsNum = secrets.getVSnumber();
				int atIdx = vsNum.indexOf('@');
				String domain = vsNum.substring(atIdx);
				String selSVM = (String) m_SVMselector.getSelectedItem();
				if (!selSVM.contains(domain)) {
					atIdx = selSVM.indexOf('@');
					String svmDom = selSVM.substring(atIdx);
					m_TabManager.showError("cannot connect to SVM",
												"this SVM serves another network",
												"log into a wallet " + svmDom);
					return;
				}

				// check that we have a PIN entered which looks valid
				if (!m_HavePIN) {
					m_TabManager.showError("cannot login to SVM " + selSVM,
												"no PIN entered",
												"please enter your PIN first");
					return;
													
				}

				// send login to server
				m_SVMselector.setEnabled(false);
				setCursor(M_WaitCursor);
				sendLogin1(selSVM);
				m_SVMselector.setEnabled(true);
				setCursor(null);
			}
		};
		// define the corresponding button
		m_LoginButton = new JButton(m_LoginAction);
		m_LoginButton.setFont(m_TabManager.M_ButtonFont);
		m_LoginButton.setDefaultCapable(true);

		// define the action for creating an account
		m_CreateAction = new AbstractAction("Create") {
			public void actionPerformed(ActionEvent ae) {
				if (m_SVMselector.getItemCount() == 0) {
					m_TabManager.showError("cannot create account at SVM",
													"no available SVM found",
													"try a different OFS gateway");
					return;
				}

				// disallow any SVM which does not match our logged-in VP domain
				VsSecrets secrets = m_Plugin.getLoginSecrets();
				String vsNum = secrets.getVSnumber();
				int atIdx = vsNum.indexOf('@');
				String domain = vsNum.substring(atIdx);
				String selSVM = (String) m_SVMselector.getSelectedItem();
				if (!selSVM.contains(domain)) {
					atIdx = selSVM.indexOf('@');
					String svmDom = selSVM.substring(atIdx);
					m_TabManager.showError("cannot connect to SVM",
												"this SVM serves another network",
												"log into a wallet " + svmDom);
					return;
				}

				// check that we have a PIN entered which looks valid
				if (!m_HavePIN) {
					m_TabManager.showError("cannot create account at SVM " + selSVM,
												"no PIN entered",
												"please enter your PIN first");
					return;
													
				}
				m_SVMselector.setEnabled(false);
				setCursor(M_WaitCursor);
				sendCreate1(selSVM);
				m_SVMselector.setEnabled(true);
				setCursor(null);
			}
		};

		// define the corresponding button
		m_CreateButton = new JButton(m_CreateAction);
		m_CreateButton.setFont(m_TabManager.M_ButtonFont);

		// define UI elements for PIN entry
		m_PINprompt = new JLabel("Enter or select your PIN:");
		m_PINprompt.setFont(new Font(Font.SANS_SERIF, Font.BOLD, 13));
		m_PINListener = new PINDocumentListener();
		m_PINUI = new JPasswordField(5);
		m_PINUI.setDropTarget(null);
		m_PINUI.setHorizontalAlignment(JTextField.LEFT);
		m_PINUI.setToolTipText("Your PIN is 1 to 5 digits");
		m_PINUI.setMaximumSize(new Dimension(50, 30));
		m_PINUI.getDocument().addDocumentListener(m_PINListener);
		m_PINUI.addActionListener(this);
		m_PINprompt.setLabelFor(m_PINUI);
	}

	/**
	 * prepare and show the display
	 */
	public void prepDisplay() {
		removeAll();
		updateConnectButtonStatus();

		// create panel for SVM drop-down selector, PIN entry, and the buttons
		JPanel connPane = new JPanel();
		connPane.setBorder(null);
		GridBagLayout gbl = new GridBagLayout();
		connPane.setLayout(gbl);
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.gridheight = 11;
		gbc.anchor = GridBagConstraints.CENTER;
		connPane.add(Box.createVerticalGlue(), gbc);
		StringBuilder adTxt = new StringBuilder(256);
		adTxt.append("\nSelect a market from the drop-down list.\n\n");
		adTxt.append("To log into it, supply your PIN and click Login.\n\n");
		adTxt.append("If you do not yet have an account at the market,\nenter ");
		adTxt.append("a PIN and click Create.\n");
		JTextArea adBlurb = new JTextArea(adTxt.toString(), 6, 40);
		adBlurb.setDropTarget(null);
		adBlurb.setEditable(false);
		connPane.add(adBlurb, gbc);
		connPane.add(Box.createVerticalStrut(20), gbc);
		connPane.add(m_SVMselector, gbc);
		connPane.add(Box.createVerticalStrut(20), gbc);

		// PIN entry group
		connPane.add(m_PINprompt, gbc);
		connPane.add(Box.createVerticalStrut(10), gbc);
		connPane.add(m_PINUI, gbc);

		// login and create buttons
		connPane.add(Box.createVerticalStrut(20), gbc);
		JPanel buttonPane = new JPanel();
		buttonPane.setLayout(new BoxLayout(buttonPane, BoxLayout.X_AXIS));
		buttonPane.setBorder(null);
		buttonPane.add(Box.createHorizontalStrut(15));
		buttonPane.add(m_LoginButton);
		buttonPane.add(Box.createHorizontalStrut(90));
		buttonPane.add(m_CreateButton);
		connPane.add(buttonPane, gbc);
		connPane.add(Box.createVerticalGlue(), gbc);
		connPane.setPreferredSize(new Dimension(480, 270));

		// put everything in a border, inside the scroll area
		setBorder(BorderFactory.createTitledBorder(
														"Connect to SilentVault Market"));
		m_ScrollPane = new JScrollPane(connPane);
		m_ScrollPane.setBorder(null);
		add(m_ScrollPane);
		JRootPane jrp = getRootPane();
		if (jrp != null) {
			jrp.setDefaultButton(m_LoginButton);
		}
		m_SVMselector.requestFocusInWindow();
		revalidate();
	}

	/**
	 * record the list of SVMIds discovered
	 * @param svmList the IDs of the SVMen we have available
	 */
	public void recordSVMList(List<String> svmList) {
		if (svmList != null) {
			m_SVMIdList.clear();
			m_SVMIdList.addAll(svmList);
			m_SVMselector = new JComboBox<String>(m_SVMIdList);
			m_SVMselector.setFocusable(false);
			m_SVMselector.setToolTipText("<html>Select the SilentVault Market"
											+ "<br/>to which you wish to connect</html>");
		}

		// send the SVM tab to the connect screen
		m_TabManager.queueTransition(new Integer(MTabManager.M_LOGIN));
		Runnable goLogin = new Runnable() {
			public void run() {
				m_TabManager.makeTransition();
			}
		};
		new Thread(goLogin).start();
	}

	/**
	 * get SVM list (used for passing this data to subclass
	 * @return the list
	 */
	public List<String> getSVMList() { return m_SVMIdList; }

	/**
	 * utility method to update the status of the Login button
	 */
	private void updateConnectButtonStatus() {
		// activate button only if we're logged into a wallet and have a PIN
		if (m_Plugin.getVscState().isLoggedIn()) {
			if (m_HavePIN) {
				m_PIN = new String(m_PINUI.getPassword());
				m_LoginButton.setEnabled(true);
				m_LoginButton.setToolTipText(
										"Click here to login to the selected market");
				m_CreateButton.setEnabled(true);
				m_CreateButton.setToolTipText(
							"Click here to create an account at the selected market");
			}
			else {
				m_LoginButton.setEnabled(false);
				m_LoginButton.setToolTipText("Please enter your PIN");
				m_CreateButton.setEnabled(false);
				m_CreateButton.setToolTipText("Please enter your PIN");
			}
		}
		else {
			m_PIN = "";
			m_LoginButton.setEnabled(false);
			m_LoginButton.setToolTipText("You must log into your wallet first");
			m_CreateButton.setEnabled(false);
			m_CreateButton.setToolTipText("You must log into your wallet first");
		}
	}

	/**
	 * queue up a modification to a text component from within its document
	 * listener
	 * @param txtComp the text component to update
	 * @param newText the replacement text
	 */
	private void queueTextMod(JTextComponent txtComp, String newText) {
		if (txtComp == null || newText == null) {
			return;
		}

		// do not allow simultaneous queued mods
		if (m_DoingUndo) {
			return;
		}

		// create and start background modification object
		TextComponentMod modifier = new TextComponentMod(txtComp, newText);
		m_DoingUndo = true;
		new Thread(modifier).start();
	}

	// implement ActionListener
	/**
	 * method to process the event when a PIN character is chosen, or enter is
	 * pressed in a text entry field
	 * @param ae the event
	 */
	public void actionPerformed(ActionEvent ae) {
		Object src = ae.getSource();
		if (src instanceof JComboBox) {
			updateConnectButtonStatus();
		}
		else if (src instanceof JTextField) {
			JTextField input = (JTextField) src;
			// move to the next focus item
			input.transferFocus();
		}
	}

	/**
	 * record and process a create stage one reply from the SVM
	 * @param cre1 the reply message, or error
	 */
	@SuppressWarnings("unchecked")
	public void recordCreate1(SVMBlock cre1) {
		if (cre1 == null
			 || !cre1.getOpcode().equalsIgnoreCase("REP_create_acct1"))
		{
			Log.error("recordCreate1(): not a create account ph1 reply message");
			return;
		}
		final String creErr = "cannot create account at SVM";

		// check for error
		if (cre1.isError()) {
			int eCode = cre1.getErrCode();
			String remedy = "try again later";
			if (eCode == HttpURLConnection.HTTP_CONFLICT) {
				remedy = "use Login with your PIN";
			}
			m_TabManager.showError(creErr, cre1.getErrMsg(), remedy);
			return;
		}

		m_Challenge = cre1.getChallenge();
		if (m_Challenge.isEmpty()) {
			Log.error("Missing challenge from SVM on create step 1");
			m_TabManager.showError(creErr,
											"SVM did not provide encrypted challenge");
			return;
		}

		/* evaluate pluginClass and make sure we support it.
		 * if we see "org.silentvault.client.foo" interpret it as
		 * "org.silentvault.spark.foo"
		 */
		PluginManager plugMgr = PluginManager.getInstance();
		ClassLoader plugLoader = plugMgr.getPluginClassLoader();
		String plugClass = cre1.getPluginClass().replace("client", "spark");
		Class pClass = null;
		try {
			pClass = plugLoader.loadClass(plugClass);
		}
		catch (ClassNotFoundException cnfe) {
			Log.error("Could not load child plugin class, "
						+ plugClass, cnfe);
			m_TabManager.showError(creErr,
										"plugin class " + plugClass + " not found",
										"obtain this plugin from the repository");
			return;
		}

		// get this plugin from the manager
		Plugin childPlugin = plugMgr.getPlugin(pClass);
		if (childPlugin == null) {
			Log.error("Could not find child plugin, " + plugClass);
			m_TabManager.showError(creErr,
										"plugin class " + plugClass + " not found",
										"obtain this plugin from the repository");
			return;
		}

		// This plugin should also be a SVMPlugin.  If so, use it.
		SVMPlugin svmPlug = null;
		if (childPlugin instanceof SVMPlugin) {
			svmPlug = (SVMPlugin) childPlugin;
		}
		if (svmPlug != null) {
			m_Plugin.setSVMPlugin(svmPlug);
		}
		else {
			Log.error("child plugin " + plugClass + " is not a SVMPlugin");
		}

		// check plugin version, if one was supplied
		String plugVersion = cre1.getPluginVersion();
		if (!plugVersion.isEmpty()) {
			List<PublicPlugin> pubPlugins = plugMgr.getPublicPlugins();
			for (PublicPlugin pubPlug : pubPlugins) {
				if (pubPlug.getPluginClass().equals(plugClass)) {
					// compare versions
					if (pubPlug.getVersion().compareTo(plugVersion) < 0) {
						Log.error("Version " + pubPlug.getVersion() + " of "
									+ plugClass + " insufficient, need "
									+ plugVersion + " or later");
						m_TabManager.showError(creErr,
													"plugin class " + plugClass + " version "
													+ plugVersion
													+ " is required; you have version "
													+ pubPlug.getVersion(),
												"update this plugin from the repository");
						return;
					}
				}
			}
		}

		// move on to stage 2
		sendCreate2((String) m_SVMselector.getSelectedItem());
	}

	/**
	 * record and process a create stage two reply from the SVM
	 * @param cre2 the reply message, or error
	 */
	public void recordCreate2(SVMBlock cre2) {
		if (cre2 == null
			 || !cre2.getOpcode().equalsIgnoreCase("REP_acct_created"))
		{
			Log.error("recordCreate2(): not a create account ph2 reply message");
			return;
		}

		// check for error
		final String creErr = "cannot create account at SVM";
		if (cre2.isError()) {
			String remedy = "try again later";
			int eCode = cre2.getErrCode();
			if (eCode == HttpURLConnection.HTTP_NOT_ACCEPTABLE) {
				remedy = "check your PIN number";
			}
			m_TabManager.showError(creErr, cre2.getErrMsg(), remedy);
			return;
		}

		// check we got sent a session Id
		setSessionId(cre2.getSessionId());
		if (m_SessionId.isEmpty()) {
			Log.error("Missing sessionId on acct create phase 2 reply");
			m_TabManager.showError(creErr, "SVM did not provide a session Id");
			return;
		}

		// switch tab manager and reply listener to point to plugin's versions
		SVMPlugin plug = m_Plugin.getSVMPlugin();
		final MTabManager mtm = plug.getTabManager();
		m_Plugin.setSVMTabManager(mtm);
		m_Plugin.setSVMListener(plug.getSVMListener());
		// copy the SVM list and session Id from the generic LoginPane (here)
		LoginPane lp = mtm.getLoginPane();
		lp.recordSVMList(m_SVMIdList);
		lp.setSessionId(m_SessionId);

		// now transition to home state
		mtm.queueTransition(new Integer(MTabManager.M_HOME_INIT));
		Runnable goHome = new Runnable() {
			public void run() {
				mtm.makeTransition();
			}
		};
		SwingUtilities.invokeLater(goHome);
		// NB: on home init, queries will be made as to holdings, balances, etc.
		m_SessionId = m_PIN = m_Challenge = "";
		m_PINUI.setText("");
		m_HavePIN = false;
	}

	/**
	 * record and process a login stage two reply from the SVM
	 * @param log1 the reply message, or error
	 */
	@SuppressWarnings("unchecked")
	public void recordLogin1(SVMBlock log1) {
		if (log1 == null
			 || !log1.getOpcode().equalsIgnoreCase("REP_login_phase1"))
		{
			Log.error("recordLogin1(): not a login phase1 reply message");
			return;
		}

		// check for error
		final String logErr = "cannot log in to account at SVM";
		if (log1.isError()) {
			String remedy = "try again later";
			int eCode = log1.getErrCode();
			if (eCode == HttpURLConnection.HTTP_NOT_FOUND) {
				remedy = "check your PIN and try again, or use Create";
			}
			m_TabManager.showError(logErr, log1.getErrMsg(), remedy);
			return;
		}

		m_Challenge = log1.getChallenge();
		if (m_Challenge.isEmpty()) {
			Log.error("Missing challenge from SVM on login step 1");
			m_TabManager.showError(logErr,
											"SVM did not provide encrypted challenge");
			return;
		}

		/* evaluate pluginClass.
		 * if we see "org.silentvault.client.foo" interpret it as
		 * "org.silentvault.spark.foo"
		 */
		PluginManager plugMgr = PluginManager.getInstance();
		ClassLoader plugLoader = plugMgr.getPluginClassLoader();
		String plugClass = log1.getPluginClass().replace("client", "spark");
		Class pClass = null;
		try {
			pClass = plugLoader.loadClass(plugClass);
		}
		catch (ClassNotFoundException cnfe) {
			Log.error("Could not load child plugin class, "
						+ plugClass, cnfe);
			m_TabManager.showError(logErr,
										"plugin class " + plugClass + " not found",
										"obtain a this plugin from the repository");
			return;
		}

		// get this plugin from the manager
		Plugin childPlugin = plugMgr.getPlugin(pClass);
		if (childPlugin == null) {
			Log.error("Could not find child plugin, " + plugClass);
			m_TabManager.showError(logErr,
										"plugin class " + plugClass + " not found",
										"obtain this plugin from the repository");
			return;
		}

		// This plugin should also be a SVMPlugin.  If so, use it.
		SVMPlugin svmPlug = null;
		if (childPlugin instanceof SVMPlugin) {
			svmPlug = (SVMPlugin) childPlugin;
		}
		if (svmPlug != null) {
			m_Plugin.setSVMPlugin(svmPlug);
		}
		else {
			Log.error("child plugin " + plugClass + " is not a SVMPlugin");
		}

		// check plugin version, if one was supplied
		String plugVersion = log1.getPluginVersion();
		if (!plugVersion.isEmpty()) {
			List<PublicPlugin> pubPlugins = plugMgr.getPublicPlugins();
			for (PublicPlugin pubPlug : pubPlugins) {
				if (pubPlug.getPluginClass().equals(plugClass)) {
					// compare versions
					if (pubPlug.getVersion().compareTo(plugVersion) < 0) {
						Log.error("Version " + pubPlug.getVersion() + " of "
									+ plugClass + " insufficient, need "
									+ plugVersion + " or later");
						m_TabManager.showError(logErr,
													"plugin class " + plugClass + " version "
													+ plugVersion
													+ " is required; you have version "
													+ pubPlug.getVersion(),
												"update this plugin from the repository");
						return;
					}
				}
			}
		}

		// move on to stage 2
		sendLogin2((String) m_SVMselector.getSelectedItem());
	}

	/**
	 * record and process a login stage two reply from the SVM
	 * @param log2 the reply message, or error
	 */
	public void recordLogin2(SVMBlock log2) {
		if (log2 == null
			 || !log2.getOpcode().equalsIgnoreCase("REP_logged_in"))
		{
			Log.error("recordLogin2(): not a login phase2 reply message");
			return;
		}

		// check for error
		final String logErr = "cannot log in to SVM";
		if (log2.isError()) {
			int eCode = log2.getErrCode();
			String remedy = "try again later";
			if (eCode == HttpURLConnection.HTTP_NOT_ACCEPTABLE) {
				remedy = "check your PIN and try again";
			}
			m_TabManager.showError(logErr, log2.getErrMsg(), remedy);
			return;
		}

		// record session Id
		setSessionId(log2.getSessionId());
		if (m_SessionId.isEmpty()) {
			Log.error("Missing sessionId on login phase 2 reply");
			m_TabManager.showError(logErr,
										"SVM did not provide a session Id");
		}

		// switch tab manager and reply listener to point to plugin's versions
		SVMPlugin plug = m_Plugin.getSVMPlugin();
		final MTabManager mtm = plug.getTabManager();
		m_Plugin.setSVMTabManager(mtm);
		m_Plugin.setSVMListener(plug.getSVMListener());
		// copy the SVM list and session Id from the generic LoginPane (here)
		LoginPane lp = mtm.getLoginPane();
		lp.recordSVMList(m_SVMIdList);
		lp.setSessionId(m_SessionId);

		// now transition to home state
		mtm.queueTransition(new Integer(MTabManager.M_HOME_INIT));
		Runnable goHome = new Runnable() {
			public void run() {
				mtm.makeTransition();
			}
		};
		SwingUtilities.invokeLater(goHome);
		// NB: on home init, further queries will be made as to account data
		m_SessionId = m_PIN = m_Challenge = "";
		m_PINUI.setText("");
		m_HavePIN = false;
	}

	/**
	 * record and process a logout reply from the SVM
	 * @param logout the reply message, or error
	 */
	public void recordLogout(SVMBlock logout) {
		if (logout == null
			 || !logout.getOpcode().equalsIgnoreCase("REP_logged_out"))
		{
			Log.error("recordLogout(): not a logout reply message");
			return;
		}

		// check for error (about the only way we can fail is not logged in)
		if (logout.isError()) {
			m_TabManager.showError("error logging out from SVM",
											logout.getErrMsg());
			// go to login pane nevertheless
			m_TabManager.queueTransition(new Integer(MTabManager.M_LOGIN));
			m_TabManager.makeTransition();
			return;
		}

		// reset plugin and listener to default
		m_Plugin.setSVMPlugin(new SVMPlugin(m_Plugin));
		m_Plugin.setSVMListener(new SVMListener(m_Plugin));
		final MTabManager mtm = new MTabManager(m_Plugin);
		LoginPane lp = mtm.getLoginPane();
		// forward our list of available SVMs to new generic LoginPane
		lp.recordSVMList(m_SVMIdList);

		// go to login screen in baseclass tab manager
		mtm.queueTransition(new Integer(MTabManager.M_LOGIN));
		Runnable doLogout = new Runnable() {
			public void run() {
				// now switch the manager to the baseclass and go to login there
				m_Plugin.setSVMTabManager(mtm);
				mtm.makeTransition();

				// forget any connected SVM
				VsSecrets secrets = m_Plugin.getLoginSecrets();
				secrets.setSVMId("");
				secrets.setSVMKey(null);
			}
		};
		SwingUtilities.invokeLater(doLogout);
		m_SessionId = m_PIN = "";
		m_PINUI.setText("");
		m_HavePIN = false;
	}

	/**
	 * utility method to send the OFS-SVM a request to create an account, phase 1
	 * @param svmId the Id of the SVM at which we want to create an account
	 */
	private void sendCreate1(String svmId) {
		if (svmId == null || svmId.isEmpty()) {
			Log.error("missing SVM Id");
			return;
		}

		XMPPConnection conn = m_Plugin.getOFSConnection();
		if (conn == null || !conn.isConnected()) {
			Log.error("Cannot create account @SVM without OFS gateway connection");
			m_TabManager.showError("cannot create account at SVM",
											"no OFS gateway connection",
											"connect to OFS first");
			return;
		}

		// obtain the pubkey of the selected SVM if missing or different
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		if (!svmId.equals(secrets.getSVMId()) || secrets.getSVMKey() == null) {
			PublicKey svmKey = m_Plugin.getPKSListener().obtainVPpubkey(svmId);
			if (svmKey == null) {
				Log.error("Could not obtain pubkey for SVM Id " + svmId);
				m_TabManager.showError("cannot connect to SVM " + svmId,
											"SVM pubkey not found on key server");
				return;
			}
			secrets.setSVMId(svmId);
			secrets.setSVMKey(svmKey);
		}

		// prepare account create phase1 message
		SVMClientBlock createBlock = new SVMClientBlock();
		createBlock.setOpcode("REQ_create_acct1");
		createBlock.setWalletId(secrets.getVSnumber());
		// set the PIN (document listener guarantees us it is a valid int)
		createBlock.setPIN(m_PIN);

		// send to server; answer will go to recordCreate1() synchronously
		SVMListener svmListener = m_Plugin.getSVMListener();
		if (!svmListener.sendRequest(createBlock)) {
			Log.error("Account create1 request failed to SVM Id " + svmId);
			m_TabManager.showError("cannot create account at SVM " + svmId,
										"create account failed at phase 1");
		}
	}

	/**
	 * utility method to send the OFS-SVM a request to create an account, phase 2
	 * @param svmId the Id of the SVM at which we want to create an account
	 */
	private void sendCreate2(String svmId) {
		if (svmId == null || svmId.isEmpty()) {
			Log.error("missing SVM Id");
			return;
		}
		final String creErr = "cannot create account at SVM";

		XMPPConnection conn = m_Plugin.getOFSConnection();
		if (conn == null || !conn.isConnected()) {
			Log.error("Cannot create account @SVM without OFS gateway connection");
			m_TabManager.showError("cannot create account at SVM",
											"no OFS gateway connection",
											"connect to OFS first");
			return;
		}

		// obtain the pubkey of the selected SVM if missing or different
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		if (!svmId.equals(secrets.getSVMId()) || secrets.getSVMKey() == null) {
			PublicKey svmKey = m_Plugin.getPKSListener().obtainVPpubkey(svmId);
			if (svmKey == null) {
				Log.error("Could not obtain pubkey for SVM Id " + svmId);
				m_TabManager.showError("cannot connect to SVM " + svmId,
											"SVM pubkey not found on key server");
				return;
			}
			secrets.setSVMId(svmId);
			secrets.setSVMKey(svmKey);
		}

		// prepare account create phase1 message
		SVMClientBlock createBlock = new SVMClientBlock();
		createBlock.setOpcode("REQ_create_acct2");

		// build the hash of wallet Id + PIN
		MessageDigest md = null;
		try {
			md = MessageDigest.getInstance("SHA-512");
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("Cannot find SHA-512 hash algorithm", nsae);
			m_TabManager.showError(creErr, "can't construct SHA512 hash",
											nsae.getMessage());
			return;
		}
		md.update(secrets.getVSnumber().getBytes());
		md.update(m_PIN.getBytes());
		byte[] acctHash = md.digest();
		createBlock.setAcctHash(StringUtils.encodeBase64(acctHash, false));

		// supply the JID
		createBlock.setJID(conn.getUser());

		// decrypt the challenge and return it
		String decChallenge
			= EncodingUtils.getStrFromBase64PubkeyEnc(m_Challenge,
																	secrets.getPrivKey());
		if (decChallenge == null) {
			Log.error("Unable to decrypt acct create challenge with privkey");
			m_TabManager.showError(creErr,
											"failed decrypting challenge from server");
			return;
		}
		// this should be b64 data
		createBlock.setChallengeReply(decChallenge);

		// send to server; answer will go to recordCreate2() synchronously
		SVMListener svmListener = m_Plugin.getSVMListener();
		if (!svmListener.sendRequest(createBlock)) {
			Log.error("Account create1 request failed to SVM Id " + svmId);
			m_TabManager.showError(creErr + " " + svmId,
											"create account failed at phase 2");
		}
	}

	/**
	 * utility method to send the OFS-SVM a request to log into an account,
	 * phase 1
	 * @param svmId the Id of the SVM at which we want to log in
	 */
	private void sendLogin1(String svmId) {
		if (svmId == null || svmId.isEmpty()) {
			Log.error("missing SVM Id");
			return;
		}

		XMPPConnection conn = m_Plugin.getOFSConnection();
		if (conn == null || !conn.isConnected()) {
			Log.error("Cannot log into acct @SVM without OFS gateway connection");
			m_TabManager.showError("cannot log in to account at SVM",
											"no OFS gateway connection",
											"connect to OFS first");
			return;
		}

		// obtain the pubkey of the selected SVM if missing or different
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		if (!svmId.equals(secrets.getSVMId()) || secrets.getSVMKey() == null) {
			PublicKey svmKey = m_Plugin.getPKSListener().obtainVPpubkey(svmId);
			if (svmKey == null) {
				Log.error("Could not obtain pubkey for SVM Id " + svmId);
				m_TabManager.showError("cannot connect to SVM " + svmId,
											"SVM pubkey not found on key server");
				return;
			}
			secrets.setSVMId(svmId);
			secrets.setSVMKey(svmKey);
		}

		// prepare account login phase1 message
		SVMClientBlock loginBlock = new SVMClientBlock();
		loginBlock.setOpcode("REQ_login_phase1");

		// build the hash of wallet Id + PIN
		MessageDigest md = null;
		try {
			md = MessageDigest.getInstance("SHA-512");
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("Cannot find SHA-512 hash algorithm", nsae);
			m_TabManager.showError("cannot log in to SVM",
											"can't construct SHA512 hash",
											nsae.getMessage());
			return;
		}
		md.update(secrets.getVSnumber().getBytes());
		md.update(m_PIN.getBytes());
		byte[] acctHash = md.digest();
		loginBlock.setAcctHash(StringUtils.encodeBase64(acctHash, false));

		// send to server; answer will go to recordCreate1() synchronously
		SVMListener svmListener = m_Plugin.getSVMListener();
		if (!svmListener.sendRequest(loginBlock)) {
			Log.error("Account login1 request failed to SVM Id " + svmId);
			m_TabManager.showError("cannot log in to account at SVM " + svmId,
										"login failed at phase 1");
		}
	}

	/**
	 * utility method to send the OFS-SVM a request to log into an account,
	 * phase 2
	 * @param svmId the Id of the SVM at which we want to log in
	 */
	private void sendLogin2(String svmId) {
		if (svmId == null || svmId.isEmpty()) {
			Log.error("missing SVM Id");
			return;
		}

		XMPPConnection conn = m_Plugin.getOFSConnection();
		if (conn == null || !conn.isConnected()) {
			Log.error("Cannot log into acct @SVM without OFS gateway connection");
			m_TabManager.showError("cannot log in to account at SVM",
											"no OFS gateway connection",
											"connect to OFS first");
			return;
		}

		// obtain the pubkey of the selected SVM if missing or different
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		if (!svmId.equals(secrets.getSVMId()) || secrets.getSVMKey() == null) {
			PublicKey svmKey = m_Plugin.getPKSListener().obtainVPpubkey(svmId);
			if (svmKey == null) {
				Log.error("Could not obtain pubkey for SVM Id " + svmId);
				m_TabManager.showError("cannot connect to SVM " + svmId,
											"SVM pubkey not found on key server");
				return;
			}
			secrets.setSVMId(svmId);
			secrets.setSVMKey(svmKey);
		}

		// prepare account create phase1 message
		SVMClientBlock loginBlock = new SVMClientBlock();
		loginBlock.setOpcode("REQ_login_phase2");

		// build the hash of wallet Id + PIN
		MessageDigest md = null;
		try {
			md = MessageDigest.getInstance("SHA-512");
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("Cannot find SHA-512 hash algorithm", nsae);
			m_TabManager.showError("cannot login at SVM",
											"can't construct SHA512 hash",
											nsae.getMessage());
			return;
		}
		md.update(secrets.getVSnumber().getBytes());
		md.update(m_PIN.getBytes());
		byte[] acctHash = md.digest();
		loginBlock.setAcctHash(StringUtils.encodeBase64(acctHash, false));

		// supply the JID
		loginBlock.setJID(conn.getUser());

		// decode the challenge and return it
		String decChallenge
			= EncodingUtils.getStrFromBase64PubkeyEnc(m_Challenge,
																	secrets.getPrivKey());
		if (decChallenge == null) {
			Log.error("Unable to decrypt acct create challenge with privkey");
			m_TabManager.showError("cannot log into account at SVM",
										"failed decrypting challenge from server");
			return;
		}
		// this should be b64 data
		String chalRep = decChallenge;
		loginBlock.setChallengeReply(chalRep);

		// send to server; answer will go to recordCreate1() synchronously
		SVMListener svmListener = m_Plugin.getSVMListener();
		if (!svmListener.sendRequest(loginBlock)) {
			Log.error("Account login2 request failed to SVM Id " + svmId);
			m_TabManager.showError("cannot log into account at SVM " + svmId,
										"login failed at phase 2");
		}
	}

	/**
	 * utility method to send the OFS-SVM a request to log out of an account
	 * @param svmId the Id of the SVM from which we want to logout
	 */
	public void sendLogout(String svmId) {
		if (svmId == null || svmId.isEmpty()) {
			// probably means we logged out already; go back to starting status
			m_TabManager.queueTransition(new Integer(MTabManager.M_LOGIN));
			m_TabManager.makeTransition();
			return;
		}

		XMPPConnection conn = m_Plugin.getOFSConnection();
		// this doesn't really make sense:
		if (conn == null || !conn.isConnected()) {
			Log.error("Cannot log out of SVM without OFS gateway connection");
			m_TabManager.showError("cannot log out of SVM",
											"no OFS gateway connection",
											"connect to OFS first");
			return;
		}

		// obtain the pubkey of the selected SVM if missing or different
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		if (!svmId.equals(secrets.getSVMId()) || secrets.getSVMKey() == null) {
			PublicKey svmKey = m_Plugin.getPKSListener().obtainVPpubkey(svmId);
			// again this doesn't make much sense if we were already logged into it
			if (svmKey == null) {
				Log.error("Could not obtain pubkey for SVM Id " + svmId);
				m_TabManager.showError("cannot log out from SVM " + svmId,
											"SVM pubkey not found on key server");
				return;
			}
			secrets.setSVMId(svmId);
			secrets.setSVMKey(svmKey);
		}

		// prepare account logout message
		SVMClientBlock logoutBlock = new SVMClientBlock();
		logoutBlock.setOpcode("REQ_logout");
		logoutBlock.setSessionId(m_SessionId);

		// send to server; answer will go to recordLogout() synchronously
		SVMListener svmListener = m_Plugin.getSVMListener();
		if (!svmListener.sendRequest(logoutBlock)) {
			Log.error("Account logout request failed to SVM Id " + svmId);
			m_TabManager.showError("cannot log out of account at SVM " + svmId,
											"failure sending logout message");
		}
	}

	/**
	 * obtain the session Id assigned by our SVM
	 * @return the Id
	 */
	public String getSessionId() { return m_SessionId; }

	/**
	 * set the session Id
	 * @param sessId the Id to set
	 */
	public void setSessionId(String sessId) {
		if (sessId != null) {
			m_SessionId = new String(sessId);
		}
	}

	// END methods
}
