/*
 * last modified---
 * 	10-02-14/jet/ SSM -> SVM (cosmetic)
 * 	08-04-14/jet/ version for Spark plugin
 * 	06-11-13/jet/ new
 *
 * purpose---
 * 	provide a UI panel for displaying the SVM screen tab once a connection
 * 	to a SVM has been established
 */

package org.silentvault.spark.ui.svm;

import javax.swing.*;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Component;
import java.awt.Dimension;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;
import org.silentvault.spark.ui.wallet.HomePane;


/**
 * Extends JPanel to provide a screen for displaying the configuration of a
 * connected SVM, and presenting interface elements to examine those contents.
 * The actual UI controls and data displayed vary with the WorkPane subclass.
 * This class also manages configuration data obtained from the SVM during the
 * connection / discovery phase.
 *
 * This class is generic, and will be overriden by another class defined in
 * each SVMPlugin instance, for the specific SVM.
 *
 * @author Justin Turrell
 */
public class MHomePane extends JPanel {
	// BEGIN data members
	/**
	 * the application to which we belong
	 */
	protected WalletClient			m_Plugin;

	/**
	 * the tab manager which creates and manages us
	 */
	protected MTabManager			m_TabManager;

	/**
	 * the current title bar text
	 */
	protected String					m_TitlebarText;

	/**
	 * the work pane, which encloses an object {@link WorkPane} whose actual
	 * contents will vary based on its subclass
	 */
	protected JPanel					m_WorkPane;

	/**
	 * the tool bar container for menu items
	 */
	protected JPanel					m_WorkBar;

	/**
	 * the complete Home pane, placed into the scroll area (m_ScrollPane)
	 */
	protected JPanel					m_HomePane;

	/**
	 * the scroll area used to ensure that this pane always displays everything
	 * regardless of window size
	 */
	protected JScrollPane			m_ScrollPane;

	/**
	 * the wallet tab HomePane, from which we use font and color constants,
	 * voucher and token tables, etc.
	 */
	protected HomePane				m_WalletHome;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param client the application which owns us
	 * @param mgr the tab manager which created us
	 */
	public MHomePane(WalletClient client, MTabManager mgr) {
		super();
		m_Plugin = client;
		m_TabManager = mgr;
		// NB: this code assumes the wallet tab manager is initialized first
		m_WalletHome = m_Plugin.getWalletTabManager().getHomePane();

		// use grid layout and set border
		setLayout(new GridLayout());

		// create toolbar, for containing menu button actions
		m_WorkBar = new JPanel();
		m_WorkBar.setLayout(new BoxLayout(m_WorkBar, BoxLayout.X_AXIS));
		m_WorkBar.setBorder(null);
		m_WorkBar.setVisible(false);
		m_WorkBar.setPreferredSize(new Dimension(460, 25));

		// create default work pane, to be filled with WorkPane subclasses
		m_WorkPane = new JPanel();
		m_WorkPane.setLayout(new GridLayout());

		// set up main scroll pane containing the objects created above
		buildHomePane();
		m_ScrollPane = new JScrollPane(m_HomePane);
		m_ScrollPane.setBorder(null);
		add(m_ScrollPane);
	}

	/**
	 * helper method to fully initialize the main home pane (which goes in the
	 * scroll pane which we actually display)
	 */
	protected void buildHomePane() {
		GridBagLayout gbl = new GridBagLayout();
		m_HomePane = new JPanel(gbl);
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.weighty = 1.0;
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.gridheight = 4;
		gbc.anchor = GridBagConstraints.CENTER;
		gbc.fill = GridBagConstraints.NONE;
		m_HomePane.add(Box.createVerticalGlue(), gbc);
		m_HomePane.add(m_WorkBar, gbc);
		m_HomePane.add(m_WorkPane, gbc);
		m_HomePane.add(Box.createVerticalGlue(), gbc);
		m_HomePane.setBorder(null);
	}

	/**
	 * update the title bar text to reflect the wallet we're logged into
	 */
	public void updateTitleText() {
		m_TitlebarText = "Wallet ID: "
								+ m_Plugin.getLoginSecrets().getVSnumber() + " ";
		setBorder(BorderFactory.createTitledBorder(m_TitlebarText));
		revalidate();
	}


	// get methods
	/**
	 * method to obtain the WorkPane sub-element
	 * @return the component, which is enclosed in the JPanel
	 */
	public WorkPane getWorkPane() {
		if (m_WorkPane == null) {
			return null;
		}
		// there should be exactly one component inside this panel
		Component comp = null;
		try {
			comp = m_WorkPane.getComponent(0);
		}
		catch (ArrayIndexOutOfBoundsException aioobe) {
			Log.error("Missing WorkPane subclass in enclosing JPanel!", aioobe);
		}
		return (WorkPane) comp;
	}

	/**
	 * obtain menu toolbar
	 * @return the toolbar panel
	 */
	public JPanel getMenuBar() { return m_WorkBar; }

	/**
	 * obtain the wallet HomePane
	 * @return the home pane of the wallet tab
	 */
	public HomePane getWalletHome() { return m_WalletHome; }


	// set methods
	/**
	 * set the contents of the work pane (called by XTabManager)
	 * @param workPane a subclass object of WorkPane
	 */
	public void setWorkPane(WorkPane workPane) {
		if (workPane != null) {
			workPane.prepDisplay();
			m_WorkPane.removeAll();
			m_WorkPane.add(workPane);
		}

		/* certain objects in our home pane might have been changed by the call
		 * to the subclass's prepDisplay()
		 */
		m_HomePane.revalidate();
		m_ScrollPane.setViewportView(m_HomePane);
	}

	/**
	 * set the wallet HomePane
	 * @param walletHome the new home pane for the wallet tab
	 */
	public void setWalletHome(HomePane walletHome) {
		if (walletHome != null) {
			m_WalletHome = walletHome;
		}
	}

	// END methods
}
