/*
 * last modified---
 * 	05-17-18/kbw/ use SV tutorials, as voucher-safe.com is gone
 * 	03-17-15/jet/ add m_EscrowHandler, getEscrowHandler()
 * 	01-15-16/jet/ allow extension
 * 	10-02-14/jet/ SSM -> SVM
 * 	08-04-14/jet/ version for Spark plugin
 * 	06-05-13/jet/ new
 *
 *	purpose---
 *		provide a manager for the contents of the marketplace tab
 */

package org.silentvault.spark.ui.svm;

import javax.swing.*;
import java.awt.Cursor;
import java.awt.Color;
import java.awt.Font;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Desktop;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.net.URI;
import java.net.URISyntaxException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.io.IOException;


/**
 * This class is used to manage the graphical contents of the marketplace tab.
 * It handles creating and destroying
 * the embedded pane subclasses and manages state transitions between them.
 * It also provides utility functions such as error popups and wait cursors.
 *
 * This class should be overridden for client-side support for a given SVM.
 *
 * @author Justin Turrell
 */
public class MTabManager extends TabManager {
	// BEGIN data members
	// possible states we can do
	/**
	 * select a SVM to login to
	 */
	public static final int	M_LOGIN = 1;
	/**
	 * move to home state, following connect and login
	 */
	public static final int	M_HOME_INIT = 2;
	/**
	 * return to the main menu screen
	 */
	public static final int	M_GO_MAIN_MENU = 3;
	/**
	 * display tutorials in a popup window
	 */
	public static final int	M_TUTORIALS = 4;
	/**
	 * log out of this marketplace
	 */
	public static final int	M_LOGOUT_MARKET = 5;

	/**
	 * standard font for buttons in thist tab
	 */
	public static final Font	M_ButtonFont = new Font(Font.SANS_SERIF,
																		Font.BOLD, 12);

	// display panes we control (more can be defined in subclasses)
	/**
	 * the Home pane, used after connect and disco have occurred
	 */
	protected MHomePane		m_HomePane;

	/**
	 * the login/connect pane
	 */
	protected LoginPane		m_LoginPane;

	/**
	 * the sub-pane which is presently handling escrow payments, and wishes to
	 * receive a call on its confirmEscrowFunding() method from the WTabManager
	 * upon confirmation of a payment made to an escrow wallet
	 */
	protected WorkPane		m_EscrowHandler;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the containing plugin
	 */
	public MTabManager(WalletClient plugin) {
		super(plugin);
		m_HomePane = new MHomePane(m_Plugin, this);
		m_LoginPane = new LoginPane(m_Plugin, this);
	}

	/**
	 * state transition method, called by threads that remove states from queue
	 * @param state the state to transition to
	 */
	public void doState(int state) {
		//Log.debug("MTabManager state transition to " + state);
		switch (state) {
			case M_LOGIN:
				removeAll();
				add(m_LoginPane);
				m_LoginPane.prepDisplay();
				break;

			case M_HOME_INIT:
				// get rid of login pane and tab containing it
	 			removeAll();
				// make a default home pane and recreate the tab using it
				m_HomePane.updateTitleText();
				add(m_HomePane);

				// prepare all the WorkPane subclass objects (TBD)
				//m_MenuPane = new MenuPane(m_Plugin, m_HomePane, this);
				//m_HomePane.setWorkPane(m_MenuPane);
				break;

			case M_GO_MAIN_MENU:
				//m_HomePane.setWorkPane(m_MenuPane);
				break;

			case M_TUTORIALS:
				// get the desktop browser for this system and try to open the URI
				URI tutURI = null;
				boolean needManual = false;
				try {
					// TBD: SVM-related tutorials
					tutURI = new URI("https://silentvault.com");
				}
				catch (URISyntaxException use) {
					Log.error("Bad URI for tutorials");
					needManual = true;
				}
				Desktop browser = null;
				try {
					browser = Desktop.getDesktop();
				}
				catch (UnsupportedOperationException uoe) {
					needManual = true;
				}
				finally {
					if (!needManual) {
						if (!browser.isSupported(Desktop.Action.BROWSE)) {
							needManual = true;
						}
					}
				}
				if (needManual) {
					// just put up the URL in a popup window with instructions
					JFrame tutPop = new JFrame("SVM Tutorials");
					tutPop.setDefaultCloseOperation(
															WindowConstants.DISPOSE_ON_CLOSE);
					JEditorPane tutViewer = new JEditorPane("text/html", null);
					tutViewer.setBackground(Color.LIGHT_GRAY);
					StringBuilder content = new StringBuilder(512);
					content.append("<html><body bgcolor=\"silver\">");
					content.append("<center><br/><br/><p>");
					content.append("For tutorials help, paste the following URL ");
					content.append("into your browser:</p><br/><br/>");
					content.append("https://silentvault.com<br/>");
					content.append("</center></body></html>");
					tutViewer.setText(content.toString());
					tutPop.add(tutViewer);
					tutPop.setSize(new Dimension(400, 200));
					tutPop.setLocationRelativeTo(this);
					tutPop.setVisible(true);
					m_Plugin.registerPopup(tutPop);
				}
				else {
					try {
						// actually invoke the default browser
						browser.browse(tutURI);
					}
					catch (IOException ioe) {
						Log.error("Could not invoke browser on tutorials page", ioe);
						showError("Could not display tutorials",
									"Error opening default browser",
									"Point your browser to: https://silentvault.com");
					}
				}
				break;

			case M_LOGOUT_MARKET:
				// user was logged out, go back to login pane
				Runnable goLogin = new Runnable() {
					public void run() {
						doState(M_LOGIN);
					}
				};
				SwingUtilities.invokeLater(goLogin);
				// NB: if we were really logged in, a subclass would have got this
				break;

			// any more cases would be handled in a subclass

			default:
				Log.error("Unsupported SVM UI state " + state);
				break;
		}
		if (state != M_TUTORIALS) {
			m_State = state;
			revalidate();
			repaint();
		}
	}


	// get methods
	/**
	 * obtain the login pane
	 * @return the pane used for logins
	 */
	public LoginPane getLoginPane() { return m_LoginPane; }

	/**
	 * obtain the home pane
	 * @return the pane used for home safe display
	 */
	public MHomePane getHomePane() { return m_HomePane; }

	/**
	 * obtain the work pane (a component in the home pane which contains the
	 * WorkPane subclass)
	 * @return the pane used for work area display
	 */
	public WorkPane getWorkPane() {
		if (m_HomePane == null) {
			return null;
		}
		return m_HomePane.getWorkPane();
	}

	/**
	 * obtain the pane designated to receive confirmEscrowFunding() calls
	 * @return the pane which wishes to receive these notification (may be null)
	 */
	public WorkPane getEscrowHandler() { return m_EscrowHandler; }


	/**
	 * method to deallocate all sub-panes (called by WalletClient destructor)
	 */
	public void deallocatePanes() {
		if (m_LoginPane != null) {
			m_LoginPane.removeAll();
			m_LoginPane = null;
		}
		if (m_HomePane != null) {
			m_HomePane.removeAll();
			m_HomePane = null;
		}
	}

	// END methods
}
