/*
 * last modified---
 * 	02-09-18/kbw/ change 24 hours to 7 days; add units to Amount label
 * 	03-08-17/jet/ correct parsing of a leading decimal point in amount field
 * 	12-03-16/jet/ improve prompts for external wallet addresses, to reduce
 * 					  confusion with wallet Ids
 * 	07-15-15/jet/ use lookup units for relative prices
 * 	08-06-14/jet/ Spark plugin version
 * 	08-27-13/jet/ add refund address element to form (m_RefundUI)
 * 	08-09-13/jet/ new
 *
 * purpose---
 * 	provide a UI screen to enter orders to buy vouchers for external payments
 */

package org.silentvault.spark.ui.svx;

import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import java.awt.event.*;
import java.awt.Font;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Component;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;
import org.silentvault.spark.ui.wallet.HomePane;
import org.silentvault.spark.ui.wallet.WTabManager;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Date;
import java.util.Calendar;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.ParseException;


/**
 * Extends WorkPane to provide a screen for entering buy orders, sending
 * them to the SVX, and then funding them using an external currency payment.
 *
 * @author Justin Turrell
 */
public final class BuyPane extends WorkPane {
	// BEGIN data members
	/**
	 * action to initiate an entry of the indicated buy offer
	 */
	private AbstractAction			m_EnterAction;

	/**
	 * button to go with enter action
	 */
	private JButton					m_EnterButton;

	/**
	 * action to acknowledge order resolution language and reset
	 */
	private AbstractAction			m_AckAction;

	/**
	 * button to go with acknowledge action
	 */
	private JButton					m_AckButton;

	/**
	 * mode toggle: if true, display order entry form; if false show confirmation
	 */
	private boolean					m_EnterMode;

	/**
	 * control to select asset being purchased by user
	 */
	private JComboBox<String>		m_OrderedUI;

	/**
	 * control to select external payment mechanism selected by user
	 */
	private JComboBox<String>		m_ExtPaymentUI;

	/**
	 * box for entering desired quantity
	 */
	private JTextField				m_AmountUI;

	/**
	 * label describing m_AmountUI
	 */
	private JLabel						m_AmountLab;

	/**
	 * box for entering external wallet address for refunds
	 */
	private JTextField				m_RefundUI;

	/**
	 * value specified via m_AmountUI
	 */
	private double						m_Amount;

	/**
	 * calculated commission amount
	 */
	private double						m_Commission;

	/**
	 * label describing commission rate (adjusted by m_OrderedUI)
	 */
	private JLabel						m_CommRate;

	/**
	 * total payment to be made to fund order (includes commission)
	 */
	private double						m_PaymentAmount;

	/**
	 * label describing the total payment amount (m_PaymentAmount)
	 */
	private JLabel						m_TotalOwed;

	/**
	 * panel to contain order entry form
	 */
	private JPanel						m_FormPane;

	/**
	 * order ID entered and assigned by SVX
	 */
	private int							m_OrderId;

	/**
	 * external wallet identifier where SVX wants payment
	 */
	private String						m_ExtWalletId;

	/**
	 * formatter for amounts
	 */
	private DecimalFormat	m_AmtFormatter;

	/**
	 * inner class to provide a DocumentListener for the amount field
	 */
	final class AmtDocumentListener implements DocumentListener {
		/**
		 * handle insert notifications
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			Document doc = de.getDocument();
			String curTxt = "";

			// interpret a '.' or ',' as 0 if it's the first character added
			int curLen = doc.getLength();
			if (curLen == 1) {
				try {
					curTxt = doc.getText(0, 1);
				}
				catch (BadLocationException ble) {
					Log.error("Bad insert doc location", ble);
				}
				if (curTxt.equals(".")) {
					// force this to "0."
					queueTextMod(m_AmountUI, "0.");
					return;
				}
				// NB: for internationalization
				if (curTxt.equals(",")) {
					// force this to "0,"
					queueTextMod(m_AmountUI, "0,");
					return;
				}
			}

			// disallow letters etc. altogether
			int changeAt = de.getOffset();
			try {
				curTxt = doc.getText(0, doc.getLength());
			}
			catch (BadLocationException ble) {
				Log.error("Cannot undo bad value edit", ble);
				return;
			}
			char chg = curTxt.charAt(changeAt);
			if (!Character.isDigit(chg) && chg != '.' && chg != ',') {
				// undo this edit
				String modTxt = curTxt.substring(0, changeAt)
									+ curTxt.substring(changeAt+1);
				queueTextMod(m_AmountUI, modTxt);
				return;
			}

			// otherwise just parse
			if (parseValue()) {
				updateValueDisplays();
			}
			else {
				// undo this edit
				String modTxt = curTxt.substring(0, changeAt)
									+ curTxt.substring(changeAt+1);
				queueTextMod(m_AmountUI, modTxt);
			}
		}

		/**
		 * handle removal notifications
		 * @param de the deletion event
		 */
		public void removeUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			// if the only thing left is a '.' or ',', force it to a 0
			Document doc = de.getDocument();
			int curLen = doc.getLength();
			if (curLen == 1) {
				String curTxt = "";
				try {
					curTxt = doc.getText(0, 1);
				}
				catch (BadLocationException ble) {
					Log.error("Bad remove doc location", ble);
				}
				if (curTxt.equals(".")) {
					// force this to "0."
					queueTextMod(m_AmountUI, "0.");
					return;
				}
				// NB: for internationalization
				if (curTxt.equals(",")) {
					// force this to "0,"
					queueTextMod(m_AmountUI, "0,");
					return;
				}
			}

			// otherwise just parse
			if (parseValue()) {
				updateValueDisplays();
			}
		}

		/**
		 * handle change notifications (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JTextField");
		}

		/**
		 * helper method to parse the current value of m_AmountUI
		 *	@return true on success, false on failure
		 */
		private boolean parseValue() {
			// examine what's in the value entry field
			String enteredVal = m_AmountUI.getText();
			if (enteredVal.isEmpty() || enteredVal.equals("0.00")
				 || enteredVal.equals("0") || enteredVal.equals("0,00"))
			{
				return true;
			}

			// attempt to parse
			double vvv = 0.0;
			try {
				Number val = m_AmtFormatter.parse(enteredVal);
				vvv = val.doubleValue();
			}
			catch (ParseException pe) {
				return false;
			}
			if (vvv < 0.0) {
				return false;
			}
			return true;
		}
	}

	/**
	 * document listener for dynamic updates of m_AmountUI
	 */
	private AmtDocumentListener	m_AmountListener;

	/**
	 * whether an undo mod is ongoing
	 */
	private boolean					m_DoingUndo;

	/**
	 * inner class to modify a text component from within its document listener,
	 * without triggering a recursive modification exception
	 */
	class TextComponentMod implements Runnable {
		/**
		 * the text component to be modified
		 */
		JTextComponent			m_ModComponent;

		/**
		 * its new contents
		 */
		String					m_NewContents;

		/**
		 * constructor
		 * @param txtComp the text component to update
		 * @param newText the replacement text
		 */
		public TextComponentMod(JTextComponent txtComp, String newText) {
			m_ModComponent = txtComp;
			m_NewContents = new String(newText);
		}

		/**
		 * execute main task in background
		 */
		public void run() {
			try { Thread.sleep(25); }
			catch (InterruptedException ie) { /* ignore */ }
			m_ModComponent.setText(m_NewContents);
			m_DoingUndo = false;
			m_ModComponent.setCaretPosition(m_NewContents.length());

			// update value displays
			updateValueDisplays();
		}
	}

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the XHomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public BuyPane(WalletClient plugin, XHomePane home, XTabManager mgr) {
		super(plugin, home, mgr);
		m_EnterMode = true;
		m_ExtWalletId = "";
		m_AmtFormatter = new DecimalFormat("#,###,##0.0###");

		// define Enter action
		m_EnterAction = new AbstractAction("Enter Purchase") {
			public void actionPerformed(ActionEvent ae) {
				// error check specified inputs
				String ordered = (String) m_OrderedUI.getSelectedItem();
				if (ordered == null || ordered.isEmpty()) {
					showError("no asset to be purchased selected",
								"select a voucher asset type from the drop-down list");
					return;
				}
				String amtText = m_AmountUI.getText();
				SVXAsset offConf = m_ParentHome.getSVXConfig().get(ordered);
				if (m_Amount < offConf.m_MinSale) {
					showError("illegal offer amount",
								"enter " + offConf.m_MinSale + " or more");
					return;
				}

				// recompute commission and total
				m_Commission = (offConf.m_SaleCommission / 100.0) * m_Amount;
				AssetPriceCache apc = m_Plugin.getPricingModule();
				String lUnits = VPAssetConfig.getBasePriceLookupUnits(ordered);
				String asPayment = (String) m_ExtPaymentUI.getSelectedItem();
				Double relPrice = apc.getRelativePrice(lUnits, asPayment);
				if (relPrice == null) {
					relPrice = new Double(1.0);
				}
				m_PaymentAmount
					= (m_Amount + m_Commission) * relPrice.doubleValue();

				// make sure they've specified a refund address
				String refHash = m_RefundUI.getText();
				if (refHash.isEmpty()) {
					showError("missing refund address",
								"enter your " + asPayment + " wallet hash");
					return;
				}

				// send REQ_voucher_purchase to SVX
				sendEnterPurchase();
			}
		};
		// define associated button
		final Font buttonFont = m_TabManager.M_ButtonFont;
		m_EnterButton = new JButton(m_EnterAction);
		m_EnterButton.setFont(buttonFont);
		m_EnterButton.setToolTipText("Submit the purchase order");

		// define order acknowledgement action
		m_AckAction = new AbstractAction("Acknowledged") {
			public void actionPerformed(ActionEvent ae) {
				// reset for next order entry
				m_EnterMode = true;
				m_OrderId = 0;
				m_Amount = m_Commission = m_PaymentAmount = 0.0;
				m_AmountUI.setText("");
				m_AmountLab.setText("Amount: ");
				m_EnterButton.setEnabled(true);
				m_TabManager.queueTransition(new Integer(XTabManager.M_GO_BUY));
				m_TabManager.makeTransition();
			}
		};
		// define associated button
		m_AckButton = new JButton(m_AckAction);
		m_AckButton.setFont(buttonFont);
		m_AckButton.setToolTipText("Return to purchase order entry screen");

		// populate button bar (replaced in prepDisplay())
		m_ButtonPane.removeAll();
		m_ButtonPane.add(Box.createHorizontalStrut(150));		// spacer
		m_ButtonPane.add(m_EnterButton);
		m_ButtonPane.add(Box.createHorizontalStrut(25));		// spacer
		m_ButtonPane.add(m_AckButton);

		// create controls for order entry form
		m_FormPane = new JPanel(new GridBagLayout());
		m_FormPane.setBorder(null);
		m_GBC.fill = GridBagConstraints.NONE;
		m_GBC.anchor = GridBagConstraints.NORTH;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_GBC.gridheight = 16;
		m_FormPane.add(Box.createVerticalGlue(), m_GBC);
		final Color labColor = UIManager.getColor("standardGray");
		HomePane whp = m_ParentHome.getWalletHome();
		JLabel headLab = new JLabel("Enter an order to purchase a voucher as "
											+ "follows:", SwingConstants.LEFT);
		headLab.setFont(whp.M_LabelFont);
		headLab.setForeground(labColor);
		m_FormPane.add(headLab, m_GBC);
		m_FormPane.add(Box.createVerticalStrut(25), m_GBC);

		// build list of sold assets selectable
		Hashtable<String, SVXAsset> svxConfig = m_ParentHome.getSVXConfig();
		Vector<String> trAssets = new Vector<String>(svxConfig.size());
		for (String asset : svxConfig.keySet()) {
			SVXAsset aConf = svxConfig.get(asset);
			if (aConf.m_MinSale > 0.0) {
				trAssets.add(aConf.m_AssetName);
			}
		}

		// based on the list, create label and drop-down for ordered assets
		final Dimension rowSize = new Dimension(450, 25);
		final Dimension boxSz = new Dimension(50, 25);
		JPanel offPane = new JPanel(new GridLayout(1, 3));
		offPane.setBorder(null);
		offPane.setPreferredSize(rowSize);
		// ordered asset list: label
		JLabel ordLab = new JLabel("Buying: ", SwingConstants.CENTER);
		ordLab.setFont(whp.M_LabelFont);
		ordLab.setForeground(labColor);
		ordLab.setToolTipText("Voucher asset you wish to purchase");
		offPane.add(ordLab);
		m_OrderedUI = new JComboBox<String>(trAssets);
		m_OrderedUI.addActionListener(this);
		m_OrderedUI.setPreferredSize(boxSz);
		m_OrderedUI.setToolTipText("Select the asset you wish to purchase");
		ordLab.setLabelFor(m_OrderedUI);
		offPane.add(m_OrderedUI);
		offPane.add(Box.createHorizontalStrut(15));
		m_FormPane.add(offPane, m_GBC);
		m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

		/* init controls for accepted external currencies; note this is populated
		 * dynamically in the listener based on the selected asset being purchased
		 */
		JPanel extPayPane = new JPanel(new GridLayout(1, 3));
		extPayPane.setBorder(null);
		extPayPane.setPreferredSize(rowSize);
		JLabel payLab = new JLabel("Pay with: ", SwingConstants.CENTER);
		payLab.setFont(whp.M_LabelFont);
		payLab.setForeground(labColor);
		payLab.setToolTipText("External currency you want to pay with");
		extPayPane.add(payLab);
		m_ExtPaymentUI = new JComboBox<String>();
		m_ExtPaymentUI.setPreferredSize(boxSz);
		m_ExtPaymentUI.setToolTipText("Select the currency you wish to use to "
												+ "pay for your order");
		m_ExtPaymentUI.addActionListener(this);
		payLab.setLabelFor(m_ExtPaymentUI);
		extPayPane.add(m_ExtPaymentUI);
		extPayPane.add(Box.createHorizontalStrut(15));
		m_FormPane.add(extPayPane, m_GBC);
		m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

		// add control for entering the amount on order
		JPanel amtPane = new JPanel(new GridLayout(1, 3));
		amtPane.setPreferredSize(rowSize);
		m_AmountUI = new JTextField(6);
		m_AmountUI.setDropTarget(null);
		String selAsset = (String) m_OrderedUI.getSelectedItem();
		String dUnits = XMLVoucher.getDisplayUnits(selAsset);
		m_AmountUI.setToolTipText("Enter the amount in " + dUnits + " of " 
											+ selAsset + " you wish to purchase");
		m_AmountUI.setPreferredSize(boxSz);
		m_AmountUI.setText("0.0");
		m_AmountListener = new AmtDocumentListener();
		m_AmountUI.getDocument().addDocumentListener(m_AmountListener);

		m_AmountLab = new JLabel("Amount: ", SwingConstants.CENTER);
		m_AmountLab.setFont(whp.M_LabelFont);
		m_AmountLab.setForeground(labColor);
		m_AmountLab.setLabelFor(m_AmountUI);
		m_AmountLab.setToolTipText("Quantity you wish to buy");
		amtPane.add(m_AmountLab);
		amtPane.add(m_AmountUI);
		amtPane.add(Box.createHorizontalStrut(15));
		m_FormPane.add(amtPane, m_GBC);
		m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

		// add labels to show commission which SVX will charge
		JPanel commPane = new JPanel(new GridLayout(1, 3));
		commPane.setBorder(null);
		commPane.setPreferredSize(rowSize);
		JLabel comLab = new JLabel("Commission: ", SwingConstants.CENTER);
		comLab.setFont(whp.M_LabelFont);
		comLab.setForeground(labColor);
		comLab.setToolTipText("<html>The sale commission charged by the SVX<br/>"
									+ "on orders for your selected asset</html>");
		commPane.add(comLab);
		m_CommRate = new JLabel("0.0", SwingConstants.LEFT);
		m_CommRate.setFont(whp.M_LabelFont);
		m_CommRate.setForeground(labColor);
		m_CommRate.setToolTipText("The commission amount");
		// NB: m_CommRate will be updated with changes in m_OrderedUI
		commPane.add(m_CommRate);
		commPane.add(Box.createHorizontalStrut(15));
		m_FormPane.add(commPane, m_GBC);
		m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

		// add labels to show the calculated external payment amount
		JPanel totPane = new JPanel(new GridLayout(1, 3));
		totPane.setBorder(null);
		totPane.setPreferredSize(rowSize);
		JLabel costLab = new JLabel("Total cost: ", SwingConstants.CENTER);
		costLab.setFont(whp.M_LabelFont);
		costLab.setForeground(labColor);
		costLab.setToolTipText("The total cost of your order");
		totPane.add(costLab);
		// NB: m_TotalOwed will be updated with changes in m_OrderedUI
		m_TotalOwed = new JLabel("0.0", SwingConstants.LEFT);
		m_TotalOwed.setFont(whp.M_LabelFont);
		m_TotalOwed.setForeground(labColor);
		m_TotalOwed.setToolTipText("<html>The total cost of your order in<br/>"
											+ "your selected external currency</html>");
		totPane.add(m_TotalOwed);
		totPane.add(Box.createHorizontalStrut(15));
		m_FormPane.add(totPane, m_GBC);
		m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

		// refund entry
		JPanel refPane = new JPanel();
		BoxLayout rbl = new BoxLayout(refPane, BoxLayout.X_AXIS);
		refPane.setLayout(rbl);
		refPane.setBorder(null);
		refPane.setPreferredSize(rowSize);
		JLabel refLab = new JLabel("Refund hash: ", SwingConstants.CENTER);
		refLab.setFont(whp.M_LabelFont);
		refLab.setForeground(labColor);
		refLab.setToolTipText("<html>The external wallet address where you "
									+ "want any refund sent.<br/>A refund would be sent "
									+ "if you did not pick up your voucher<br/>within 7 "
									+ "days, or if a technical problem occurred.<br/>"
									+ "<br/>(Because our customers might be using a "
									+ "third-party wallet,<br/>we cannot simply reverse "
									+ "a spend to issue a refund.)<br/></html>");
		// add control for entering refund address
		m_RefundUI = new JTextField(40);
		m_RefundUI.setDropTarget(null);
		m_RefundUI.setToolTipText("<html>The <i>external</i> wallet hash where "
										+ "you'd want a refund to be sent.<br/>Note this "
										+ "is <b>not</b> your SilentVault wallet ID."
										+ "</html>");
		m_RefundUI.setPreferredSize(new Dimension(150, 25));
		refLab.setLabelFor(m_RefundUI);
		refPane.add(refLab);
		refPane.add(Box.createHorizontalStrut(30));
		refPane.add(m_RefundUI);
		m_FormPane.add(refPane, m_GBC);
		m_FormPane.add(Box.createVerticalGlue(), m_GBC);

		// set the user instructions
		m_UserInstructs = "<html>Enter your purchase order below.<br/><br/>Note "
								+ "that you must fulfill the indicated<br/>external "
								+ "payment, and this must be<br/>confirmed, before "
								+ "your voucher will be<br/>issued and sent to your "
								+ "wallet.<br/><br/>"
								+ "Unpaid purchase orders expire 7 days<br/>after "
								+ "entry.</html>";
		m_WorkArea.setBorder(null);
	}

	/**
	 * helper method to queue up a modification to a text component from within a
	 * document listener
	 * @param txtComp the text component to update
	 * @param newText the replacement text
	 */
	private void queueTextMod(JTextComponent txtComp, String newText) {
		if (txtComp == null || newText == null) {
			return;
		}

		// do not allow simultaneous queued mods
		if (m_DoingUndo) {
			return;
		}

		// create and start background modification object
		TextComponentMod modifier = new TextComponentMod(txtComp, newText);
		m_DoingUndo = true;
		SwingUtilities.invokeLater(modifier);
	}

	/**
	 * perform a (blocking) query to the SVX entering the order in the database
	 */
	private void sendEnterPurchase() {
		// send a REQ_voucher_purchase to register the purchase order at the SVX
		SVXListener svxIface = m_Plugin.getSVXListener();
		SVXClientBlock mkOffer = new SVXClientBlock();
		mkOffer.setOpcode("REQ_voucher_purchase");
		mkOffer.setWalletId(m_Plugin.getLoginSecrets().getVSnumber());

		// set purchase data based on user entries
		String ordered = (String) m_OrderedUI.getSelectedItem();
		mkOffer.setAsset(ordered);
		mkOffer.setQuantity(m_Amount);
		String asPayment = (String) m_ExtPaymentUI.getSelectedItem();
		mkOffer.setCurrency(asPayment);
		mkOffer.setExtWallet(m_RefundUI.getText());

		// send message
		setCursor(M_WaitCursor);
		svxIface.sendRequest(mkOffer);
		setCursor(null);
		// NB: reply comes synchronously to processPurchaseTerms() below
	}

	/**
	 * record the reply to our query of extant offers
	 * @param reply the reply received from the SVX
	 */
	public void processPurchaseTerms(SVXBlock reply) {
		if (reply == null) {
			Log.error("Missing SVX offer search reply");
			return;
		}
		if (!reply.getOpcode().equalsIgnoreCase("REP_purchase_terms")) {
			Log.error("Unexpected reply opcode, " + reply.getOpcode());
		}
		if (reply.isError()) {
			m_TabManager.showError("SVX purchase entry failed", reply.getErrMsg());
			return;
		}

		// record values returned from SVX
		m_OrderId = reply.getOrderId();
		m_ExtWalletId = reply.getExtWallet();

		// check echoes
		if (reply.getQuantity() != m_Amount) {
			m_Amount = reply.getQuantity();
		}
		if (m_Commission != reply.getCommission()) {
			m_Commission = reply.getCommission();
		}
		// round to 4 digits before comparing
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(0);
		double expected = m_PaymentAmount;
		String amtStr = df.format(m_PaymentAmount);
		try {
			Number quan = df.parse(amtStr);
			expected = quan.doubleValue();
		}
		catch (ParseException pe) {
			Log.error("Error parsing 4-digit total, " + amtStr, pe);
		}
		if (expected != reply.getTotal()) {
			// popup?
		/*
			Log.warning("Total payment amount changed, " + reply.getTotal()
							+ " vs. " + expected + " expected");
		 */
			m_PaymentAmount = reply.getTotal();
		}

		// go to confirmation
		m_EnterMode = false;
		Runnable goConf = new Runnable() {
			public void run() {
				m_TabManager.queueTransition(new Integer(XTabManager.M_GO_BUY));
				m_TabManager.makeTransition();
			}
		};
		SwingUtilities.invokeLater(goConf);
	}

	// extend WorkPane
	/**
	 * generate display for this sub-class
	 */
	public void prepDisplay() {
		// set screen label
		buildScreenLabel("Enter a voucher asset purchase order");

		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(3);
		HomePane whp = m_ParentHome.getWalletHome();

		if (m_EnterMode) {
			// populate button bar
			m_ButtonPane.removeAll();
			m_ButtonPane.add(Box.createHorizontalStrut(185));		// spacer
			m_ButtonPane.add(m_EnterButton);
			m_AckButton.setVisible(false);
			m_EnterButton.setVisible(true);
			m_EnterButton.setToolTipText("Submit the purchase order");

			// now take account of XMenuPane's drop-down sell asset selection
			XMenuPane xmp = m_TabManager.getMenuPane();
			String asset = xmp.getSellAsset();
			// NB: does nothing if this asset is not in our allowed list
			m_OrderedUI.setSelectedItem(asset);
			String dUnits = XMLVoucher.getDisplayUnits(asset);

			// based on the selected asset, obtain list of external payment types
			Hashtable<String, SVXAsset> svxConfig = m_ParentHome.getSVXConfig();
			SVXAsset aConf = svxConfig.get(asset);
			String[] accepted = aConf.m_SaleCurrencies.split(":");
			m_ExtPaymentUI.removeActionListener(this);
			m_ExtPaymentUI.removeAllItems();
			for (int aaa = 0; aaa < accepted.length; aaa++) {
				m_ExtPaymentUI.addItem(accepted[aaa]);
			}
			String asPayment = (String) m_ExtPaymentUI.getSelectedItem();
			m_ExtPaymentUI.addActionListener(this);

			// compute commission based on selections
			m_Commission = (aConf.m_SaleCommission / 100.0) * m_Amount;
			m_CommRate.setText(df.format(m_Commission));
			m_CommRate.setToolTipText("The commission amount "
												+ "(" + df.format(aConf.m_SaleCommission)
												+ " %)");
			AssetPriceCache apc = m_Plugin.getPricingModule();
			String lUnits = VPAssetConfig.getBasePriceLookupUnits(asset);
			Double relPrice = apc.getRelativePrice(lUnits, asPayment);
			if (relPrice == null) {
				relPrice = new Double(1.0);
			}
			m_PaymentAmount = (m_Amount + m_Commission) * relPrice.doubleValue();
			df.setMaximumFractionDigits(XMLVoucher.M_Precision * 2);
			m_TotalOwed.setText(df.format(m_PaymentAmount) + " " + asPayment);
			m_TotalOwed.setToolTipText("<html>The total cost of your order in "
												+ asPayment + " at "
												+ df.format(relPrice.doubleValue())
												+ " per " + dUnits);

			// place entire quasi-table UI into the main scroll area
			m_WorkArea.setViewportView(m_FormPane);
		}
		else {
			// populate button bar
			m_ButtonPane.removeAll();
			m_ButtonPane.add(Box.createHorizontalStrut(185));		// spacer
			m_ButtonPane.add(m_AckButton);
			m_AckButton.setVisible(true);
			m_EnterButton.setVisible(false);

			// display language about offer handling, place in m_WorkArea
			JPanel confPane = new JPanel(new GridBagLayout());
			confPane.setBorder(null);
			confPane.setPreferredSize(new Dimension(470, 250));
			m_GBC.gridheight = 5;
			confPane.add(Box.createVerticalGlue(), m_GBC);
			JLabel confLab = new JLabel("Purchase Order Now Entered",
												SwingConstants.CENTER);
			final Color labColor = UIManager.getColor("standardGray");
			confLab.setFont(whp.M_LabelFont);
			confLab.setForeground(labColor);
			confPane.add(confLab, m_GBC);
			confPane.add(Box.createVerticalStrut(5), m_GBC);
			StringBuilder confTxt = new StringBuilder(512);
			String vehicle = (String) m_ExtPaymentUI.getSelectedItem();
			confTxt.append("Thank you for your purchase order, ID # " + m_OrderId);
			confTxt.append(".\n\nPlease make an external payment of ");
			df.setMaximumFractionDigits(XMLVoucher.M_Precision * 2);
			confTxt.append(df.format(m_PaymentAmount) + " " + vehicle);
			confTxt.append(" to this address:\n\n" + m_ExtWalletId + "\n\n");
			confTxt.append("You must make your payment within 7 days.\n\n");
			confTxt.append("After your payment has been confirmed, your voucher ");
			confTxt.append("will be spent\nto your wallet.  To pick it up, use ");
			confTxt.append("the Refesh button in the Wallet tab.\n\n");
			confTxt.append("You must pick up your voucher within 7 days from ");
			confTxt.append("its issuance.\n");

			JTextArea confBlurb = new JTextArea(confTxt.toString(), 13, 40);
			confBlurb.setDropTarget(null);
			confBlurb.setEditable(false);
			confPane.add(confBlurb, m_GBC);
			confPane.add(Box.createVerticalGlue(), m_GBC);
			m_WorkArea.setViewportView(confPane);

			if (m_Plugin.isTestNetwork()) {
			final StringBuilder popTxt = new StringBuilder(128);
				popTxt.append("Because this is a test environment, you do not ");
				popTxt.append("need\nto make an actual bitcoin/litecoin payment.");
				popTxt.append("\n\nYour voucher will be issued automatically, ");
				popTxt.append("and should arrive momentarily.\n");
				popTxt.append("Go to the Overview (main) page in ");
				popTxt.append("the Wallet tab and click Refresh to pick it up.\n");
				Runnable notice = new Runnable() {
					public void run() {
						JOptionPane.showMessageDialog(m_TabManager, popTxt.toString(),
																"Voucher Purchase Completed",
															JOptionPane.INFORMATION_MESSAGE);
					}
				};
				SwingUtilities.invokeLater(notice);
			}
		}
		revalidate();
	}

	/**
	 * update the text of the labels which show values dependent on inputs
	 */
	private void updateValueDisplays() {
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(3);
		String txtAmt = m_AmountUI.getText();
		if (txtAmt.isEmpty()) {
			txtAmt = "0.0";
		}

		double vvv = 0.0;
		try {
			Number val = m_AmtFormatter.parse(txtAmt);
			vvv = val.doubleValue();
		}
		catch (ParseException pe) {
			Log.error("Pre-parsed value did not parse, " + txtAmt, pe);
			return;
		}
		m_Amount = vvv;
		String trAsset = (String) m_OrderedUI.getSelectedItem();
		if (trAsset == null || trAsset.isEmpty()) {
			return;
		}
		SVXAsset aConf = m_ParentHome.getSVXConfig().get(trAsset);
		String dUnits = XMLVoucher.getDisplayUnits(aConf.m_AssetName);
		String asPayment = (String) m_ExtPaymentUI.getSelectedItem();
		AssetPriceCache apc = m_Plugin.getPricingModule();
		String lUnits = VPAssetConfig.getBasePriceLookupUnits(aConf.m_AssetName);
		Double relPrice = apc.getRelativePrice(lUnits, asPayment);
		if (relPrice == null) {
			relPrice = new Double(1.0);
		}
		m_Commission = (aConf.m_SaleCommission / 100.0) * m_Amount;
		m_CommRate.setText(df.format(m_Commission));
		m_CommRate.setToolTipText("The commission amount ("
										+ df.format(aConf.m_SaleCommission) + " %)");
		m_PaymentAmount = (m_Amount + m_Commission) * relPrice.doubleValue();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision * 2);
		m_TotalOwed.setText(df.format(m_PaymentAmount) + " " + asPayment);
		m_TotalOwed.setToolTipText("<html>The total cost of your order in "
											+ asPayment + " at "
											+ df.format(relPrice.doubleValue()) + " per "
											+ dUnits);
	}

	// implement ActionListener
	/**
	 * method to update displayed commission rate and max amount tooltips
	 * when an offered asset is selected
	 */
	@Override
	public void actionPerformed(ActionEvent ae) {
		// when m_OrderedUI changes, update m_CommRate appropriately
		Object evSrc = ae.getSource();
		if (evSrc instanceof JComboBox) {
			JComboBox src = (JComboBox) evSrc;
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
			df.setMaximumFractionDigits(XMLVoucher.M_Precision);
			df.setMinimumFractionDigits(XMLVoucher.M_Precision);
			df.setGroupingSize(3);
			Hashtable<String, SVXAsset> svxConfig = m_ParentHome.getSVXConfig();
			String trAsset = (String) m_OrderedUI.getSelectedItem();
			if (trAsset == null || trAsset.isEmpty()) {
				return;
			}
			SVXAsset aConf = svxConfig.get(trAsset);
			if (aConf == null) {
				Log.error("Config for purchased asset " + trAsset + " not found");
				return;
			}
			String dUnits = XMLVoucher.getDisplayUnits(trAsset);
			m_AmountUI.setToolTipText("Enter the amount in " + dUnits + " of " 
											+ trAsset + " you wish to purchase");
			m_AmountLab.setText("Amount in " + dUnits + ": ");
			if (src == m_OrderedUI) {
				// recompute accepted payment assets
				String[] accepted = aConf.m_SaleCurrencies.split(":");
				// prevent recursion
				m_ExtPaymentUI.removeActionListener(this);
				m_ExtPaymentUI.removeAllItems();
				for (int aaa = 0; aaa < accepted.length; aaa++) {
					m_ExtPaymentUI.addItem(accepted[aaa]);
				}
				m_ExtPaymentUI.addActionListener(this);
			}
			String asPayment = (String) m_ExtPaymentUI.getSelectedItem();
			AssetPriceCache apc = m_Plugin.getPricingModule();
			String lUnits = VPAssetConfig.getBasePriceLookupUnits(trAsset);
			Double relPrice = apc.getRelativePrice(lUnits, asPayment);
			if (relPrice == null) {
				relPrice = new Double(1.0);
			}

			// recompute commission based on selection and amount
			m_Commission = (aConf.m_SaleCommission / 100.0) * m_Amount;
			m_CommRate.setText(df.format(m_Commission));
			m_CommRate.setToolTipText("The commission amount ("
											+ df.format(aConf.m_SaleCommission) + " %)");
			m_PaymentAmount = (m_Amount + m_Commission) * relPrice.doubleValue();
			df.setMaximumFractionDigits(XMLVoucher.M_Precision * 2);
			m_TotalOwed.setText(df.format(m_PaymentAmount) + " " + asPayment);
			m_TotalOwed.setToolTipText("<html>The total cost of your order in "
												+ asPayment + " at "
												+ df.format(relPrice.doubleValue())
												+ " per " + dUnits);
		}
	}

	// END methods
}
