/*
 * last modified---
 * 	08-06-14/jet/ Spark plugin version
 * 	06-13-13/jet/ new
 *
 *	purpose---
 *		provide the contents of the SVX connect screen
 */

package org.silentvault.spark.ui.svx;

import javax.swing.*;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;

import java.util.List;
import java.util.Vector;
import java.security.PublicKey;


/**
 * This class extends JPanel to provide a screen for selecting the SVX to
 * connect to.  It also maintains the list of available SVXen, which is
 * populated after OFS connection by a discovery request.  Once a connection
 * is made, a REQ_discover_config is automatically sent to the SVX.
 *
 * @author Justin Turrell
 */
public final class ConnectPane extends JPanel {
	// BEGIN data members
	/**
	 * wait cursor
	 */
	private final Cursor			M_WaitCursor
										= Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR);
	
	/**
	 * the client application to which we belong
	 */
	private WalletClient			m_Plugin;

	/**
	 * the tab manager which creates and manages us
	 */
	private XTabManager			m_TabManager;

	/**
	 * prompt for specifying the SVX Id
	 */
	private JComboBox<String>	m_SVXselector;

	/**
	 * the connect button
	 */
	private JButton				m_ConnectButton;

	/**
	 * the action for handling connection attempts
	 */
	private AbstractAction		m_ConnectAction;

	/**
	 * the scroll area used to ensure that this pane always displays everything
	 * regardless of window size
	 */
	private JScrollPane			m_ScrollPane;

	/**
	 * all discovered SVXen
	 */
	private Vector<String>		m_SVXIdList;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param client the owning application
	 * @param mgr the tab manager which creates us
	 */
	public ConnectPane(WalletClient client, XTabManager mgr) {
		super();
		m_Plugin = client;
		m_TabManager = mgr;
		m_SVXIdList = new Vector<String>();
		// NB: will be replaced in recordSVXList()
		m_SVXselector = new JComboBox<String>();

		// use grid layout
		setLayout(new GridLayout());
		setBorder(null);

		// define the action for performing a connect
		m_ConnectAction = new AbstractAction("Connect to SVX") {
			public void actionPerformed(ActionEvent ae) {
				if (m_SVXselector.getItemCount() == 0) {
					m_TabManager.showError("cannot connect to SVX",
												"no available SVX found",
												"try a different gateway");
				}
				else {
					// disallow any SVX which does not match our logged-in VP domain
					VsSecrets secrets = m_Plugin.getLoginSecrets();
					String vsNum = secrets.getVSnumber();
					int atIdx = vsNum.indexOf('@');
					String domain = vsNum.substring(atIdx);
					String selSVX = (String) m_SVXselector.getSelectedItem();
					if (!selSVX.contains(domain)) {
						atIdx = selSVX.indexOf('@');
						String svxDom = selSVX.substring(atIdx);
						m_TabManager.showError("cannot connect to SVX",
													"this SVX serves another network",
													"log into a wallet " + svxDom);
						return;
					}

					setCursor(M_WaitCursor);
					sendGetConfig();
					setCursor(null);
				}
			}
		};
		// define the corresponding button
		m_ConnectButton = new JButton(m_ConnectAction);
		m_ConnectButton.setFont(m_TabManager.M_ButtonFont);
		m_ConnectButton.setDefaultCapable(true);
	}

	/**
	 * prepare and show the display
	 */
	public void prepDisplay() {
		removeAll();

		// activate button only if we're logged into a wallet
		if (m_Plugin.getVscState().isLoggedIn()) {
			m_ConnectButton.setEnabled(true);
			m_ConnectButton.setToolTipText(
										"Click here to connect to the selected exchange");
		}
		else {
			m_ConnectButton.setEnabled(false);
			m_ConnectButton.setToolTipText("You must log into your wallet first");
		}

		// create panel for SVX drop-down selector and the button
		JPanel connPane = new JPanel();
		connPane.setBorder(null);
		GridBagLayout gbl = new GridBagLayout();
		connPane.setLayout(gbl);
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.gridheight = 5;
		gbc.anchor = GridBagConstraints.CENTER;
		connPane.add(Box.createVerticalGlue(), gbc);
		connPane.add(m_SVXselector, gbc);
		connPane.add(Box.createVerticalStrut(25), gbc);
		connPane.add(m_ConnectButton, gbc);
		connPane.add(Box.createVerticalGlue(), gbc);
		connPane.setPreferredSize(new Dimension(480, 100));

		// put everything in a border, inside the scroll area
		setBorder(BorderFactory.createTitledBorder(
														"Connect to SilentVault Exchange"));
		m_ScrollPane = new JScrollPane(connPane);
		m_ScrollPane.setBorder(null);
		add(m_ScrollPane);
		JRootPane jrp = getRootPane();
		if (jrp != null) {
			jrp.setDefaultButton(m_ConnectButton);
		}
		m_SVXselector.requestFocusInWindow();
		revalidate();
	}

	/**
	 * record the list of SVXIds discovered
	 * @param svxList the IDs of the SVXen we have available
	 */
	public void recordSVXList(List<String> svxList) {
		if (svxList != null) {
			m_SVXIdList.clear();
			m_SVXIdList.addAll(svxList);
			m_SVXselector = new JComboBox<String>(m_SVXIdList);
			m_SVXselector.setFocusable(false);
			m_SVXselector.setToolTipText("<html>Select the SilentVault Exchange"
											+ "<br/>to which you wish to connect</html>");
		}

		// send the SVX tab to the connect screen
		m_TabManager.queueTransition(new Integer(XTabManager.M_CONNECT));
		Runnable goConn = new Runnable() {
			public void run() {
				m_TabManager.makeTransition();
			}
		};
		SwingUtilities.invokeLater(goConn);
	}

	/**
	 * utility method to send the OFS-SVX a request for the config data of our
	 * selected SVX
	 */
	private void sendGetConfig() {
		XMPPConnection conn = m_Plugin.getOFSConnection();
		if (conn == null || !conn.isConnected()) {
			Log.error("Cannot connect to SVX without OFS gateway connection");
			m_TabManager.showError("cannot connect to SVX",
										"no OFS gateway connection",
										"connect to OFS first");
			return;
		}

		// obtain the pubkey of the selected SVX if missing or different
		String svxId = (String) m_SVXselector.getSelectedItem();
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		if (!svxId.equals(secrets.getSVXId()) || secrets.getSVXKey() == null) {
			PublicKey svxKey = m_Plugin.getPKSListener().obtainVPpubkey(svxId);
			if (svxKey == null) {
				Log.error("Could not obtain pubkey for SVX Id " + svxId);
				m_TabManager.showError("cannot connect to SVX",
											"SVX pubkey not found on key server");
				return;
			}
			secrets.setSVXId(svxId);
			secrets.setSVXKey(svxKey);
		}

		// prepare discovery message
		SVXClientBlock discBlock = new SVXClientBlock();
		discBlock.setOpcode("REQ_discover_config");
		discBlock.setWalletId(secrets.getVSnumber());
		discBlock.setJID(conn.getUser());

		// send to server; answer will go to processConfigReply() synchronously
		SVXListener svxListener = m_Plugin.getSVXListener();
		if (!svxListener.sendRequest(discBlock)) {
			Log.error("Disco request failed to SVX Id " + svxId);
			m_TabManager.showError("cannot connect to SVX",
										"discovery request failed");
		}
	}

	/**
	 * record the results of SVX config discovery
	 * @param discoRep the message we got in response
	 */
	public void processConfigReply(SVXBlock svxRep) {
		if (svxRep == null) {
			Log.error("Missing SVX config reply");
			return;
		}
		if (!svxRep.getOpcode().equalsIgnoreCase("REP_config")) {
			Log.error("Unexpected reply opcode, " + svxRep.getOpcode());
		}
		if (svxRep.isError()) {
			m_TabManager.showError("SVX config discovery failed",
											svxRep.getErrMsg());
			return;
		}

		// pass to HomePane, which stores this data
		XHomePane homePane = m_TabManager.getHomePane();
		homePane.recordConfig(svxRep);
	}

	// END methods
}
