/*
 * last modified---
 * 	08-06-14/jet/ Spark plugin version
 * 	03-19-14/jet/ font spacing fixes for OpenJDK-7
 * 	09-11-13/jet/ present single revised popup on escrow funding success
 * 	07-26-13/jet/ allow expired offers to be eligible for pickup also
 * 	07-09-13/jet/ new
 *
 * purpose---
 * 	provide a UI screen to show the offers related to the user's wallet
 */

package org.silentvault.spark.ui.svx;

import javax.swing.*;
import java.awt.event.*;
import java.awt.Font;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;
import org.silentvault.spark.ui.wallet.HomePane;
import org.silentvault.spark.ui.wallet.WTabManager;
import org.silentvault.spark.ui.wallet.PaymentPane;
import org.silentvault.spark.ui.wallet.ConfirmPane;

import java.util.ArrayList;
import java.util.Date;
import java.util.Calendar;
import javax.management.timer.Timer;


/**
 * Extends WorkPane to provide a screen for showing open/funded/filled offers
 * belonging to our wallet.
 *
 * @author Justin Turrell
 */
public final class OffersPane extends WorkPane {
	// BEGIN data members
	/**
	 * action to refresh offer list by re-querying SVX for offers for this wallet
	 */
	private AbstractAction			m_RefreshAction;
	
	/**
	 * button to go with refresh action
	 */
	private JButton					m_RefreshButton;

	/**
	 * action to initiate a fund of a selected open offer
	 */
	private AbstractAction			m_FundAction;

	/**
	 * button to go with fund action
	 */
	private JButton					m_FundButton;

	/**
	 * action to initiate a cancellation of a selected funded offer
	 */
	private AbstractAction			m_CancelAction;

	/**
	 * button to go with cancel action
	 */
	private JButton					m_CancelButton;

	/**
	 * action to initiate a pickup of a selected filled offer
	 */
	private AbstractAction			m_PickupAction;

	/**
	 * button to go with pickup action
	 */
	private JButton					m_PickupButton;
	
	/**
	 * flag to indicate (to WTabManager) that we're making an escrow payment
	 */
	private boolean					m_DoingEscrow;

	/**
	 * currently selected trade offer, if any
	 */
	private TradeOffer				m_SelectedOffer;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the XHomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public OffersPane(WalletClient plugin, XHomePane home, XTabManager mgr) {
		super(plugin, home, mgr);

		// define Refresh action
		m_RefreshAction = new AbstractAction("Refresh") {
			public void actionPerformed(ActionEvent ae) {
				// get fresh list of offer status
				queryOffers();
				// reset buttons
				m_FundButton.setEnabled(false);
				m_CancelButton.setEnabled(false);
				m_PickupButton.setEnabled(false);
				// rebuild table
				prepDisplay();
			}
		};
		// define associated button
		final Font buttonFont = m_TabManager.M_ButtonFont;
		m_RefreshButton = new JButton(m_RefreshAction);
		m_RefreshButton.setFont(buttonFont);
		m_RefreshButton.setToolTipText("Refresh offer status");

		// define Fund action
		m_FundAction = new AbstractAction("Fund") {
			public void actionPerformed(ActionEvent ae) {
				// obtain selected offer and double-check it is 'open'
				OfferTableModel otm = m_ParentHome.getOfferModel();
				ArrayList<TradeOffer> selOffers = otm.getSelectedOffers();
				if (selOffers.isEmpty()) {
					showError("cannot fund offer", "no offer selected",
								"select an offer with open status to be funded");
					return;
				}
				m_SelectedOffer = selOffers.get(0);
				if (!m_SelectedOffer.getStatus().equals(TradeOffer.M_StatusOpen)) {
					showError("cannot fund offer", "offer must have open status");
					return;
				}

				// build a (decoded) WebMessage object to pass data to PaymentPane
				WebMessage webPass = new WebMessage();
				webPass.setAmount(m_SelectedOffer.getQuantity());
				webPass.setVSnumber(m_SelectedOffer.getEscrowWallet());
				String asset = m_SelectedOffer.getOffered();
				// determine issuer for this asset by examining a voucher
				HomePane whp = m_ParentHome.getWalletHome();
				VoucherTableModel vtm = whp.getVoucherModel();
				XMLVoucher firstV = vtm.getFirstAssetVoucher(asset);
				if (firstV == null) {
					// should not happen, since we vetted asset balance already
					showError("cannot fund escrow for offer",
								"no vouchers found for asset " + asset);
					Log.error("No vouchers found to fund offer "
								+ m_SelectedOffer.getOfferId() + " with " + asset);
					return;
				}
				vtm.selectIssuer(firstV.getIssuer());
				webPass.setIssuer(firstV.getIssuer());
				webPass.setCurrency(XMLVoucher.getDisplayUnits(asset));
				webPass.setBaggageFields("offerId");
				webPass.addOrderRecord("offerId",
								new Integer(m_SelectedOffer.getOfferId()).toString());
				webPass.setDecoded(true);
				m_Plugin.getWebListener().addWebMessage(webPass);
				WTabManager wtm = m_Plugin.getWalletTabManager();
				PaymentPane pp = wtm.getPaymentPane();
				pp.setSelectedSiteMsg(webPass);

				// disable button to prevent multiple entry
				m_FundButton.setEnabled(false);
				m_FundButton.setToolTipText("Please make your escrow "
													+ "payment in the Wallet tab.");

				// send user to payment screen in WTabManager
				m_DoingEscrow = true;
				wtm.queueTransition(new Integer(WTabManager.M_PAYMENTS1));
				m_Plugin.setSelectedTab(wtm);
				wtm.makeTransition();
			}
		};
		// define associated button
		m_FundButton = new JButton(m_FundAction);
		m_FundButton.setFont(buttonFont);
		m_FundButton.setToolTipText("Fund escrow for the selected open offer");
		m_FundButton.setEnabled(false);

		// define Cancel action
		m_CancelAction = new AbstractAction("Cancel") {
			public void actionPerformed(ActionEvent ae) {
				// obtain selected offer
				OfferTableModel otm = m_ParentHome.getOfferModel();
				ArrayList<TradeOffer> selOffers = otm.getSelectedOffers();
				if (selOffers.isEmpty()) {
					showError("cannot cancel offer", "no offer selected",
						"select an offer with open or funded status to be canceled");
					return;
				}
				m_SelectedOffer = selOffers.get(0);

				// double-check offer status, must be 'open' or 'funded'
				String stat = m_SelectedOffer.getStatus();
				if (!stat.equals(TradeOffer.M_StatusOpen)
					 && !stat.equals(TradeOffer.M_StatusFunded))
				{
					showError("cannot cancel offer",
								"offer must have open or funded status");
					return;
				}

				// send REQ_cancel_offer request
				sendCancelOffer();
			}
		};
		// define associated button
		m_CancelButton = new JButton(m_CancelAction);
		m_CancelButton.setFont(buttonFont);
		m_CancelButton.setEnabled(false);
		m_CancelButton.setToolTipText("Cancel the selected offer");

		// define Pickup action
		m_PickupAction = new AbstractAction("Pickup") {
			public void actionPerformed(ActionEvent ae) {
				// obtain the selected offer
				OfferTableModel otm = m_ParentHome.getOfferModel();
				ArrayList<TradeOffer> selOffers = otm.getSelectedOffers();
				if (selOffers.isEmpty()) {
					showError("cannot pickup offer", "no offer selected",
								"select an offer with filled status to pick up");
					return;
				}
				m_SelectedOffer = selOffers.get(0);

				// double-check offer status and exchange timestamp
				String stat = m_SelectedOffer.getStatus();
				if (!stat.equals(TradeOffer.M_StatusFilled)
					 && !stat.equals(TradeOffer.M_StatusExpired))
				{
					showError("cannot pickup offer", "offer has improper status",
						"select an offer with filled or expired status to pick up");
					return;
				}

				// make sure the escrow payout was at least a week in the past
				Date payout = m_SelectedOffer.getExTimestamp();
				Calendar cal = Calendar.getInstance();
				int offGMTmillis = cal.get(Calendar.ZONE_OFFSET);
				int offDSTmillis = cal.get(Calendar.DST_OFFSET);
				Date now = new Date();
				payout = new Date(payout.getTime() + offGMTmillis + offDSTmillis);
				Date weekAgo = new Date(now.getTime() - Timer.ONE_WEEK);
				if (!payout.before(weekAgo)) {
					showError("cannot pickup offer",
								"escrow payout was less than 7 days ago",
							"look for an incoming voucher payment not yet picked up");
					return;
				}

				// send REQ_offer_pickup
				sendPickupOffer();
			}
		};
		// define associated button
		m_PickupButton = new JButton(m_PickupAction);
		m_PickupButton.setFont(buttonFont);
		m_PickupButton.setToolTipText("Pick up escrow from the selected offer");
		m_PickupButton.setEnabled(false);

		// populate button bar
		m_ButtonPane.removeAll();
		m_ButtonPane.add(Box.createHorizontalStrut(10));		// spacer
		m_ButtonPane.add(m_FundButton);
		m_ButtonPane.add(Box.createHorizontalStrut(60));		// spacer
		m_ButtonPane.add(m_PickupButton);
		m_ButtonPane.add(Box.createHorizontalStrut(60));		// spacer
		m_ButtonPane.add(m_CancelButton);
		m_ButtonPane.add(Box.createHorizontalStrut(50));		// spacer
		m_ButtonPane.add(m_RefreshButton);

		// set the user instructions
		m_UserInstructs = "Select an offer from the table.<br/>Depending upon "
								+ "its status, action buttons may be available.";
		m_WorkArea.setBorder(null);
	}

	/**
	 * perform a (blocking) query to the SVX obtaining all extant offers
	 */
	public void queryOffers() {
		// send a REQ_offer_status to fetch all extant offers for the wallet
		SVXListener svxIface = m_Plugin.getSVXListener();
		SVXClientBlock getOffers = new SVXClientBlock();
		getOffers.setOpcode("REQ_offer_status");
		getOffers.setWalletId(m_Plugin.getLoginSecrets().getVSnumber());
		setCursor(M_WaitCursor);
		svxIface.sendRequest(getOffers);
		setCursor(null);
		// NB: reply comes synchronously to processOffersReply() below
	}

	/**
	 * record the reply to our query of extant offers
	 * @param reply the reply received from the SVX
	 */
	public void processOffersReply(SVXBlock reply) {
		if (reply == null) {
			Log.error("Missing SVX config reply");
			return;
		}
		if (!reply.getOpcode().equalsIgnoreCase("REP_offer_status")) {
			Log.error("Unexpected reply opcode, " + reply.getOpcode());
		}
		if (reply.isError()) {
			showError("SVX offer fetch failed", reply.getErrMsg());
			return;
		}

		// pass to the data model, which also updates the table on screen
		OfferTableModel otm = m_ParentHome.getOfferModel();
		ArrayList<SVXBlock.OfferStatus> repOffers = reply.getWalletOffers();
		if (repOffers.isEmpty()) {
			otm.mergeOffers(null);
			return;
		}
		ArrayList<TradeOffer> offers
			= new ArrayList<TradeOffer>(repOffers.size());
		for (SVXBlock.OfferStatus off : repOffers) {
			TradeOffer offer = new TradeOffer();
			offer.setOfferId(off.m_OfferId);
			offer.setStatus(off.m_Status);
			offer.setOffered(off.m_Offered.m_EscrowAsset);
			offer.setQuantity(off.m_Offered.m_Quantity);
			offer.setWanted(off.m_Offered.m_WantedAssets);
			offer.setPrice(off.m_Offered.m_Price);
			offer.setEscrowWallet(off.m_EscrowWallet);
			offer.setExpiration(off.m_Offered.m_Expiry);
			offer.setExAsset(off.m_ExAsset);
			offer.setExQuantity(off.m_ExQuantity);
			offer.setExTimestamp(off.m_ExTimestamp);
			offers.add(offer);
		}
		otm.mergeOffers(offers);
	}

	/**
	 * query whether we're doing an escrow payment or not
	 * @return true if in the middle of conducting a payment to escrow
	 */
	public boolean doingEscrow() { return m_DoingEscrow; }

	/**
	 * make funding confirmation request
	 */
	public void confirmEscrowFunding() {
		if (m_SelectedOffer == null) {
			Log.error("No offer selected");
			return;
		}

		// send a REQ_fund_offer to ask SVX to pickup the escrow payment
		SVXListener svxIface = m_Plugin.getSVXListener();
		SVXClientBlock fundOffer = new SVXClientBlock();
		fundOffer.setOpcode("REQ_fund_offer");
		fundOffer.setWalletId(m_Plugin.getLoginSecrets().getVSnumber());
		fundOffer.setOfferId(m_SelectedOffer.getOfferId());
		fundOffer.setAsset(m_SelectedOffer.getOffered());

		// return focus to exchange tab
		m_Plugin.setSelectedTab(m_TabManager);

		// put up info saying what we are doing (on SVX tab)
		WTabManager wtm = m_Plugin.getWalletTabManager();
		PaymentPane pp = wtm.getPaymentPane();
		final StringBuilder popTxt = new StringBuilder(256);
		// obtain last confirmation recorded by ConfirmPane off the DHT
		XMLConfirmPayment conf = pp.getLatestConfirmation();
		if (conf != null) {
			/* we deferred showing payment success, and so must recapitulate
			 * what the ConfirmPane would have displayed as popup text.
			 */
			popTxt.append("Your payment of " + conf.getQuantity());
			popTxt.append(" " + conf.getUnits() + " of " + conf.getAsset());
			if (!conf.getAsset().equals(conf.getIssuer())) {
				popTxt.append(" at " + conf.getIssuer());
			}
			popTxt.append("\nwas sent to ");
			popTxt.append(conf.getPayee() + " successfully.");
			// show that change voucher was received if there was one
			if (conf.getVoucher() != null) {
				popTxt.append("\nYou have received a voucher back as change.");
			}

		}
		else {
			Log.error("Unexpected: no XMLConfirmPayment found in PaymentPane");
			// show something generic
			popTxt.append("Your escrow payment to fund offer ID ");
			popTxt.append(fundOffer.getOfferId() + " succeeded.");
		}

		// add generic indication that we're now telling the SVX to pick it up
		popTxt.append("\n\nThe SVX is now processing the escrow payment.  ");
		popTxt.append("This may take a few moments.\nPlease wait for the ");
		popTxt.append("confirmation screen, read it carefully, and acknowledge.");

		// for safety, do in separate thread while allowing message to proceed
		Runnable notice = new Runnable() {
			public void run() {
				JOptionPane.showMessageDialog(m_TabManager, popTxt.toString(),
														"Escrow Payment Completed",
														JOptionPane.INFORMATION_MESSAGE);
			}
		};
		SwingUtilities.invokeLater(notice);

		// reset ConfirmPane to default behavior
		ConfirmPane cp = wtm.getConfirmPane();
		cp.setDeferPopup(false);

		// send message to SVX
		setCursor(M_WaitCursor);
		svxIface.sendRequest(fundOffer);
		setCursor(null);
		// NB: reply comes synchronously to processFundingReply() below
	}

	/**
	 * record the results of an offer funding attempt (post-payment)
	 * @param reply the reply received from the SVX
	 */
	public void processFundingReply(SVXBlock reply) {
		if (reply == null) {
			Log.error("Missing SVX offer funding reply");
			return;
		}
		if (!reply.getOpcode().equalsIgnoreCase("REP_offer_funded")) {
			Log.error("Unexpected reply opcode, " + reply.getOpcode());
		}
		m_DoingEscrow = false;
		OfferTableModel otm = m_ParentHome.getOfferModel();
		if (reply.isError()) {
			showError("SVX offer funding failed", reply.getErrMsg());
			// reset
			m_FundButton.setEnabled(true);
			m_FundButton.setToolTipText("Fund escrow for the selected open offer");
			otm.deselectOffer(m_SelectedOffer);
			m_SelectedOffer = null;
			return;
		}

		// update offer status on success
		if (m_SelectedOffer != null) {
			// double-check SVX seems to be referring to the same offer
			boolean same = true;
			if (m_SelectedOffer.getOfferId() != reply.getOfferId()) {
				Log.error("OP: funding oId inconsistent, "
							+ m_SelectedOffer.getOfferId() + " vs. "
							+ reply.getOfferId());
				same = false;
			}
			// check against offered type
			if (!m_SelectedOffer.getOffered().equals(reply.getAsset())) {
				Log.error("OP: funding asset inconsistent, "
							+ m_SelectedOffer.getOffered() + " vs. "
							+ reply.getAsset());
				same = false;
			}
			if (m_SelectedOffer.getQuantity() != reply.getQuantity()) {
				Log.error("OP: funding amount inconsistent, "
							+ m_SelectedOffer.getQuantity() + " vs. "
							+ reply.getQuantity());
				same = false;
			}
			if (same) {
				// update status on offer
				m_SelectedOffer.setStatus(TradeOffer.M_StatusFunded);
			}
			otm.deselectOffer(m_SelectedOffer);
			m_SelectedOffer = null;
			// fix up button states and tooltips
			offerSelectChanged();
		}

		// go back to redisplay offers, in separate thread since we are an event
		Runnable goBack = new Runnable() {
			public void run() {
				m_TabManager.queueTransition(new Integer(XTabManager.M_GO_OFFERS));
				m_TabManager.makeTransition();
			}
		};
		SwingUtilities.invokeLater(goBack);
	}

	/**
	 * make offer cancellation request
	 */
	public void sendCancelOffer() {
		if (m_SelectedOffer == null) {
			Log.error("No offer selected");
			return;
		}
		// send a REQ_cancel_offer to cancel the offer at the SVX
		SVXListener svxIface = m_Plugin.getSVXListener();
		SVXClientBlock cancOffer = new SVXClientBlock();
		cancOffer.setOpcode("REQ_cancel_offer");
		cancOffer.setWalletId(m_Plugin.getLoginSecrets().getVSnumber());
		cancOffer.setOfferId(m_SelectedOffer.getOfferId());

		// send message
		setCursor(M_WaitCursor);
		svxIface.sendRequest(cancOffer);
		setCursor(null);
		// NB: reply comes synchronously to processCancelReply() below
	}

	/**
	 * record the results of an offer cancelation attempt
	 * @param reply the reply received from the SVX
	 */
	public void processCancelReply(SVXBlock reply) {
		if (reply == null) {
			Log.error("Missing SVX offer cancel reply");
			return;
		}
		if (!reply.getOpcode().equalsIgnoreCase("REP_offer_canceled")) {
			Log.error("Unexpected reply opcode, " + reply.getOpcode());
		}
		OfferTableModel otm = m_ParentHome.getOfferModel();
		if (reply.isError()) {
			showError("SVX offer cancelation failed", reply.getErrMsg());
			// reset
			m_CancelButton.setEnabled(true);
			if (m_SelectedOffer != null) {
				if (m_SelectedOffer.getStatus().equals(TradeOffer.M_StatusFunded)) {
					m_CancelButton.setToolTipText("Withdraw escrow for the selected "
															+ "funded offer");
				}
				else {
					m_CancelButton.setToolTipText("Cancel the selected open offer");
				}
				m_SelectedOffer = null;
			}
			else {
				m_CancelButton.setToolTipText("Cancel the selected offer");
			}
			return;
		}

		// remove offer on success (it will now be marked canceled)
		if (m_SelectedOffer != null) {
			// verify offer is the same as expected, and show any voucher amount
			if (reply.getOfferId() != m_SelectedOffer.getOfferId()) {
				Log.error("Canceled oId inconsistent, "
							+ m_SelectedOffer.getOfferId() + " vs. "
							+ reply.getOfferId());
			}
			double vouchQuan = reply.getQuantity();
			final WTabManager wtm = m_Plugin.getWalletTabManager();
			if (vouchQuan > 0.0) {
				// show user a notification that they got an escrow refund
				wtm.showInfo("An escrow refund of " + vouchQuan + " "
								+ m_SelectedOffer.getOffered()
								+ " has been sent to your wallet.\n\n"
								+ "We will now use Refresh in the Wallet tab to pick "
								+ "it up.");
			}

			// regardless, need to remove selected offer from the table
			ArrayList<TradeOffer> rms = new ArrayList<TradeOffer>(1);
			rms.add(m_SelectedOffer);
			otm.removeOffers(rms);
			m_SelectedOffer = null;
			// fix up button status and tooltips
			offerSelectChanged();

			// like the popup says, we need to trigger the pickup via housekeeping
			Runnable goHKP = new Runnable() {
				public void run() {
					m_Plugin.setSelectedTab(wtm);
					wtm.queryDHTrecords(false);
				}
			};
			new Thread(goHKP).start();
		}

		// go back to redisplay offers, in separate thread since we are an event
		Runnable goBack = new Runnable() {
			public void run() {
				m_TabManager.queueTransition(new Integer(XTabManager.M_GO_OFFERS));
				m_TabManager.makeTransition();
			}
		};
		SwingUtilities.invokeLater(goBack);
	}

	/**
	 * make offer pickup request
	 */
	public void sendPickupOffer() {
		if (m_SelectedOffer == null) {
			Log.error("No offer selected");
			return;
		}
		// send a REQ_offer_pickup to pickup the offer at the SVX
		SVXListener svxIface = m_Plugin.getSVXListener();
		SVXClientBlock pickupOffer = new SVXClientBlock();
		pickupOffer.setOpcode("REQ_offer_pickup");
		pickupOffer.setWalletId(m_Plugin.getLoginSecrets().getVSnumber());
		pickupOffer.setOfferId(m_SelectedOffer.getOfferId());

		// send message
		setCursor(M_WaitCursor);
		svxIface.sendRequest(pickupOffer);
		setCursor(null);
		// NB: reply comes synchronously to processPickupReply() below
	}

	/**
	 * record the results of an offer pickup attempt
	 * @param reply the reply received from the SVX
	 */
	public void processPickupReply(SVXBlock reply) {
		if (reply == null) {
			Log.error("Missing SVX offer pickup reply");
			return;
		}
		if (!reply.getOpcode().equalsIgnoreCase("REP_offer_pickup")) {
			Log.error("Unexpected reply opcode, " + reply.getOpcode());
		}
		OfferTableModel otm = m_ParentHome.getOfferModel();
		if (reply.isError()) {
			showError("SVX offer pickup failed", reply.getErrMsg());
			// reset
			m_PickupButton.setEnabled(true);
			m_PickupButton.setToolTipText("Pick up escrow from the selected "
													+ "offer");
			otm.deselectOffer(m_SelectedOffer);
			m_SelectedOffer = null;
			return;
		}

		/* remove offer on success (it will not be marked closed until our
		 * receipt is seen by the SVX during its next housekeeping)
		 */
		if (m_SelectedOffer != null) {
			// verify offer is the same as expected, and show any voucher amount
			if (reply.getOfferId() != m_SelectedOffer.getOfferId()) {
				Log.error("Picked up oId inconsistent, "
							+ m_SelectedOffer.getOfferId() + " vs. "
							+ reply.getOfferId());
			}
			double vouchQuan = reply.getQuantity();
			final WTabManager wtm = m_Plugin.getWalletTabManager();
			if (vouchQuan > 0.0) {
				// show user a notification that they got an escrow refund
				wtm.showInfo("An escrow payment of " + vouchQuan
								+ m_SelectedOffer.getExAsset()
								+ " has been sent to your wallet.\n\n"
								+ "We will now use Refresh in the Wallet tab to "
								+ "pick it up.");
			}

			// need to remove selected offer from the table
			ArrayList<TradeOffer> rms = new ArrayList<TradeOffer>(1);
			rms.add(m_SelectedOffer);
			otm.removeOffers(rms);
			m_SelectedOffer = null;
			// fix up button status and tooltips
			offerSelectChanged();

			// like the popup says, we need to trigger the pickup via housekeeping
			Runnable goHKP = new Runnable() {
				public void run() {
					m_Plugin.setSelectedTab(wtm);
					wtm.queryDHTrecords(false);
				}
			};
			new Thread(goHKP).start();
		}
		else {
			// go back to redisplay offers, in separate thread since we're an event
			Runnable goBack = new Runnable() {
				public void run() {
					m_TabManager.queueTransition(
														new Integer(XTabManager.M_GO_OFFERS));
					m_TabManager.makeTransition();
				}
			};
			SwingUtilities.invokeLater(goBack);
		}
	}

	// extend WorkPane
	/**
	 * generate display for this sub-class
	 */
	public void prepDisplay() {
		// set screen label
		buildScreenLabel("Trade offers outstanding for your wallet");

		// place the offer list into work area, set for radio select mode
		OfferTableModel otm = m_ParentHome.getOfferModel();
		otm.deselectAll();
		otm.setSelectMode(OfferTableModel.M_SelectSingle);
		JTable vt = m_ParentHome.getOfferTable();
		m_WorkArea.setViewportView(vt);
		revalidate();
	}

	/**
	 * method to receive notification of a change in offer selection
	 */
	public void offerSelectChanged() {
		// set button status according to state of selected offer
		OfferTableModel otm = m_ParentHome.getOfferModel();
		if (otm.getSelectMode() != OfferTableModel.M_SelectSingle) {
			// not meant for us, do nothing
			return;
		}

		ArrayList<TradeOffer> selOffers = otm.getSelectedOffers();
		if (selOffers.isEmpty()) {
			// nothing is selected, so disable all buttons
			m_FundButton.setEnabled(false);
			m_CancelButton.setEnabled(false);
			m_PickupButton.setEnabled(false);
			return;
		}
		if (selOffers.size() > 1) {
			Log.error("Unexpected number of offer selections, "
							+ selOffers.size());
			return;
		}

		// examine the status of the offer to determine which buttons to enable
		TradeOffer offer = selOffers.get(0);
		String stat = offer.getStatus();
		Calendar cal = Calendar.getInstance();
		int offGMTmillis = cal.get(Calendar.ZONE_OFFSET);
		int offDSTmillis = cal.get(Calendar.DST_OFFSET);
		Date now = new Date();
		if (stat.equals(TradeOffer.M_StatusOpen)) {
			// an open offer may be funded or canceled
			m_FundButton.setEnabled(true);
			m_CancelButton.setEnabled(true);
			m_CancelButton.setToolTipText("Cancel the selected open offer");
			m_PickupButton.setEnabled(false);
		}
		else if (stat.equals(TradeOffer.M_StatusFunded)) {
			// a funded offer may be canceled, and escrow retrieved, UNLESS expired
			if (offer.getExpiration().after(now)) {
				m_CancelButton.setEnabled(true);
				m_CancelButton.setToolTipText("Cancel and withdraw escrow for the "
														+ "selected funded offer");
			}
			else {
				// we have to let the SVX send us the refund at next HKP interval
				m_CancelButton.setEnabled(false);
				m_CancelButton.setToolTipText("Cancel the selected funded offer");
			}
			m_FundButton.setEnabled(false);
			m_PickupButton.setEnabled(false);
		}
		else if (stat.equals(TradeOffer.M_StatusFilled)) {
			m_CancelButton.setEnabled(false);
			m_FundButton.setEnabled(false);

			/* Examine the fill date.  If this is at least one week ago, this
			 * offer can be picked up manually.  (Otherwise, we expect automatic
			 * pickup, but it could be the user has turned off housekeeping, or
			 * the SVX just hasn't seen the user's receipt yet in its own HKP.)
			 */
			Date filled = offer.getExTimestamp();
			filled = new Date(filled.getTime() + offGMTmillis + offDSTmillis);
			Date weekAgo = new Date(now.getTime() - Timer.ONE_WEEK);
			if (filled.before(weekAgo)) {
				m_PickupButton.setEnabled(true);
			}
			else {
				m_PickupButton.setEnabled(false);
			}
		}
		else if (stat.equals(TradeOffer.M_StatusExpired)) {
			/* Expired offers are returned to us by the SVX iff the escrow refund
			 * we were issued was sent more than 1 week ago.  In this case, as
			 * with a filled offer, we can allow an attempt at pickup.
			 */
			Date refunded = offer.getExTimestamp();
			refunded = new Date(refunded.getTime() + offGMTmillis + offDSTmillis);
			Date weekAgo = new Date(now.getTime() - Timer.ONE_WEEK);
			if (refunded.before(weekAgo)) {
				m_PickupButton.setEnabled(true);
			}
			else {
				m_PickupButton.setEnabled(false);
			}
		}
	}

	// END methods
}
