/*
 * last modified---
 * 	07-15-15/jet/ adjust formatting, for scroll bars when lots of assets
 * 	08-06-14/jet/ Spark plugin version
 * 	03-19-14/jet/ font spacing fixes for OpenJDK-7
 * 	11-09-13/jet/ specify price ratio as par, or percentage discount/premium
 * 	09-12-13/jet/ permit searches passing 0.0 as price, to disable criteria;
 * 					  always go to search criteria screen when button clicked
 * 	07-10-13/jet/ new
 *
 * purpose---
 * 	provide a UI screen to search and display available offers on the SVX
 */

package org.silentvault.spark.ui.svx;

import javax.swing.*;
import java.awt.event.*;
import java.awt.Font;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;
import org.silentvault.spark.ui.wallet.HomePane;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Date;


/**
 * Extends WorkPane to provide a screen for searching funded offers and
 * displaying them, along with controls to accept an offer.
 *
 * @author Justin Turrell
 */
public final class SearchPane extends WorkPane implements ItemListener {
	// BEGIN data members
	/**
	 * action to return to search screen
	 */
	private AbstractAction			m_ReturnAction;
	
	/**
	 * button to go with return action
	 */
	private JButton					m_ReturnButton;

	/**
	 * action to initiate an acceptance of the selected funded offer
	 */
	private AbstractAction			m_AcceptAction;

	/**
	 * button to go with accept action
	 */
	private JButton					m_AcceptButton;

	/**
	 * action to initiate a search for offers based on selected criteria
	 */
	private AbstractAction			m_SearchAction;

	/**
	 * button to go with pickup action
	 */
	private JButton					m_SearchButton;
	
	/**
	 * mode toggle: if true, display search selectors; if false display results
	 */
	private boolean					m_ShowResults;

	/**
	 * list of controls to specify offered assets
	 */
	private ArrayList<JCheckBox>	m_OfferedBoxes;

	/**
	 * list of controls to specify wanted assets
	 */
	private ArrayList<JCheckBox>	m_WantedBoxes;

	/**
	 * box for entering price percentage (for premium/discount)
	 */
	private JTextField				m_PercentUI;

	/**
	 * value specified via m_PercentUI
	 */
	private float						m_Ratio;

	/**
	 * radio button for selecting par (equal) price ratio offers
	 */
	private JRadioButton				m_EqualRatio;

	/**
	 * radio button for selecting premium price ratio offers
	 */
	private JRadioButton				m_PremiumRatio;

	/**
	 * radio button for selecting discount price ratio offers
	 */
	private JRadioButton				m_DiscountRatio;

	/**
	 * radio button for de-selecting price ratio as a criteria
	 */
	private JRadioButton				m_IgnoreRatio;

	/**
	 * button group for price ratio selection
	 */
	private ButtonGroup				m_RatioGroup;

	/**
	 * box for entering minimum quantity
	 */
	private JTextField				m_MinimumUI;

	/**
	 * value specified via m_MinimumUI
	 */
	private double						m_Minimum;

	/**
	 * box for entering maximum quantity
	 */
	private JTextField				m_MaximumUI;

	/**
	 * value specified via m_MaximumUI
	 */
	private double						m_Maximum;

	/**
	 * panel to contain search input UI
	 */
	private JPanel						m_CriteriaPane;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the XHomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public SearchPane(WalletClient plugin, XHomePane home, XTabManager mgr) {
		super(plugin, home, mgr);
		m_Ratio = 0.0F;

		// define Return action
		m_ReturnAction = new AbstractAction("New Search") {
			public void actionPerformed(ActionEvent ae) {
				m_ShowResults = false;
				m_TabManager.queueTransition(new Integer(XTabManager.M_GO_SEARCH));
				m_TabManager.makeTransition();
			}
		};
		// define associated button (only shown on results screen)
		final Font buttonFont = m_TabManager.M_ButtonFont;
		m_ReturnButton = new JButton(m_ReturnAction);
		m_ReturnButton.setFont(buttonFont);
		m_ReturnButton.setToolTipText("Make another search");
		m_ReturnButton.setEnabled(true);

		// define Accept action
		m_AcceptAction = new AbstractAction("Accept") {
			public void actionPerformed(ActionEvent ae) {
				// make sure we have an offer selected
				OfferTableModel otm = m_ParentHome.getOfferModel();
				ArrayList<TradeOffer> selOffers = otm.getSelectedOffers();
				if (selOffers.isEmpty()) {
					showError("cannot accept offer", "none selected",
								"select an offer to accept");
					return;
				}
				TradeOffer selected = selOffers.get(0);

				// go to AcceptPane and tell it which offer
				AcceptPane ap = m_TabManager.getAcceptPane();
				ap.setOffer(selected);
				m_TabManager.queueTransition(new Integer(XTabManager.M_GO_ACCEPT));
				m_TabManager.makeTransition();
			}
		};
		// define associated button
		m_AcceptButton = new JButton(m_AcceptAction);
		m_AcceptButton.setFont(buttonFont);
		m_AcceptButton.setToolTipText("Accept the selected funded offer");
		m_AcceptButton.setEnabled(false);

		// define Search action
		m_SearchAction = new AbstractAction("Search") {
			public void actionPerformed(ActionEvent ae) {
				// error-check values in buttons and m_PercentUI
				if (m_EqualRatio.isSelected()) {
					m_Ratio = 1.0F;
				}
				else if (m_IgnoreRatio.isSelected()) {
					m_Ratio = 0.0F;
				}
				else if (m_PremiumRatio.isSelected()) {
				String ratioTxt = m_PercentUI.getText();
					if (ratioTxt.isEmpty()) {
						m_Ratio = 0.0F;
					}
					else {
						// convert to ratio
						try {
							Float pct = Float.parseFloat(ratioTxt);
							float prem = pct.floatValue();
							if (prem < 0.0F) {
								showError("illegal premium value",
											"enter a positive percentage");
								return;
							}
							// apply translation to premium ratio
							m_Ratio = 1.0F / (1.0F + (prem / 100.0F));
						}
						catch (NumberFormatException nfe) {
							showError("illegal premium value",
										"enter a positive percentage");
							return;
						}
					}
				}
				else if (m_DiscountRatio.isSelected()) {
					String ratioTxt = m_PercentUI.getText();
					if (ratioTxt.isEmpty()) {
						m_Ratio = 1.0F;
					}
					else {
						// convert to ratio
						try {
							Float pct = Float.parseFloat(ratioTxt);
							float disc = pct.floatValue();
							if (disc <= 0.0F) {
								showError("illegal discount value",
											"enter a positive percentage");
								return;
							}
							// apply translation to discount ratio
							m_Ratio = 1.0F + (disc / 100.0F);
						}
						catch (NumberFormatException nfe) {
							showError("illegal discount value",
										"enter a positive percentage");
							return;
						}
					}
				}

				// error-check values in m_MinimumUI and m_MaximumUI
				String minTxt = m_MinimumUI.getText();
				if (minTxt.isEmpty()) {
					m_Minimum = 0.0;
				}
				else {
					try {
						Double min = Double.parseDouble(minTxt);
						if (min.doubleValue() <= 0.0) {
							showError("illegal minimum value",
										"enter a positive number");
							return;
						}
						m_Minimum = min.doubleValue();
					}
					catch (NumberFormatException nfe) {
						showError("illegal minimum value", "enter a positive number");
						return;
					}
				}
				String maxTxt = m_MaximumUI.getText();
				if (maxTxt.isEmpty()) {
					m_Maximum = 0.0;
				}
				else {
					try {
						Double max = Double.parseDouble(maxTxt);
						if (max.doubleValue() <= 0.0) {
							showError("illegal maximum value",
										"enter a positive number");
							return;
						}
						m_Maximum = max.doubleValue();
					}
					catch (NumberFormatException nfe) {
						showError("illegal maximum value", "enter a positive number");
						return;
					}
				}
				if (m_Minimum > m_Maximum && m_Maximum > 0.0) {
					showError("minimum offer amount cannot exceed maximum",
								"enter a smaller min or larger max");
					return;
				}

				// send the search request to the SVX
				queryOffers();
				m_ShowResults = true;
				prepDisplay();
			}
		};
		// define associated button
		m_SearchButton = new JButton(m_SearchAction);
		m_SearchButton.setFont(buttonFont);
		m_SearchButton.setToolTipText("Search for offers based on your selection "
												+ "criteria");
		m_SearchButton.setEnabled(true);

		// populate button bar
		m_ButtonPane.removeAll();
		m_ButtonPane.add(Box.createHorizontalStrut(15));		// spacer
		m_ButtonPane.add(m_SearchButton);
		m_ButtonPane.add(Box.createHorizontalStrut(105));		// spacer
		m_ButtonPane.add(m_AcceptButton);
		m_ButtonPane.add(Box.createHorizontalStrut(85));		// spacer
		m_ButtonPane.add(m_ReturnButton);

		// create controls for selection criteria
		m_CriteriaPane = new JPanel(new GridBagLayout());
		m_CriteriaPane.setBorder(null);
		m_GBC.fill = GridBagConstraints.NONE;
		m_GBC.anchor = GridBagConstraints.NORTH;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		final Color labColor = UIManager.getColor("standardGray");
		HomePane whp = m_ParentHome.getWalletHome();
		JLabel headLab = new JLabel("Restrict matches by these criteria:",
											SwingConstants.LEFT);
		headLab.setFont(whp.M_LabelFont);
		headLab.setForeground(labColor);

		// build list of trade assets selectable
		Hashtable<String, SVXAsset> svxConfig = m_ParentHome.getSVXConfig();
		ArrayList<String> trAssets = new ArrayList<String>(svxConfig.size());
		for (String asset : svxConfig.keySet()) {
			SVXAsset aConf = svxConfig.get(asset);
			if (aConf.m_MinExchange > 0.0) {
				trAssets.add(aConf.m_AssetName);
			}
		}

		// based on the list, create label and checkboxes for offered assets
		final Dimension rowSize = new Dimension(470, 25);
		m_OfferedBoxes = new ArrayList<JCheckBox>(trAssets.size());
		// use max of 3 checkboxes per row
		final int maxBoxes = 3;
		int rowCnt = (trAssets.size() / maxBoxes) + 1;
		m_GBC.gridheight = 9 + (2 * rowCnt);
		m_CriteriaPane.add(Box.createVerticalGlue(), m_GBC);
		m_CriteriaPane.add(headLab, m_GBC);
		m_CriteriaPane.add(Box.createVerticalStrut(10), m_GBC);
		int aItem = 0;
		for (int row = 0; row < rowCnt; row++) {
			int rStart = aItem;
			int rEnd = Math.min(rStart + maxBoxes, trAssets.size());
			int rCnt = rEnd - rStart;
			JPanel offPane = new JPanel(new GridLayout(1, rCnt));
			offPane.setBorder(null);
			offPane.setPreferredSize(rowSize);
			if (row == 0) {
				// on the first row, include the label
				JLabel offLab = new JLabel("Offered assets: ", SwingConstants.LEFT);
				offLab.setFont(whp.M_LabelFont);
				offLab.setForeground(labColor);
				offPane.add(offLab);
				// do one less checkbox
				rEnd--;
			}
			int onRow = 0;
			for (int aaa = aItem; aaa < rEnd; aaa++) {
				String asset = trAssets.get(aaa);
				JCheckBox oBox = new JCheckBox(asset);
				// these boxes must have select listeners to adjust m_PercentUI
				oBox.addItemListener(this);
				oBox.setToolTipText("Check to find trades offering " + asset);
				m_OfferedBoxes.add(oBox);
				offPane.add(oBox);
				onRow++;
			}
			aItem = rEnd;
			// pad with struts if row isn't full, to retain alignment
			if (row != 0) {
				int boxesShort = maxBoxes - onRow;
				for (int sss = 0; sss < boxesShort; sss++) {
					offPane.add(Box.createHorizontalStrut(30));
				}
			}
			m_CriteriaPane.add(offPane, m_GBC);
		}
		m_CriteriaPane.add(Box.createVerticalStrut(5), m_GBC);

		// based on the list, create label and checkboxes for wanted assets
		m_WantedBoxes = new ArrayList<JCheckBox>(trAssets.size());
		// again, partition into rows
		aItem = 0;
		for (int row = 0; row < rowCnt; row++) {
			int rStart = aItem;
			int rEnd = Math.min(rStart + maxBoxes, trAssets.size());
			int rCnt = rEnd - rStart;
			JPanel wantPane = new JPanel(new GridLayout(1, rCnt));
			wantPane.setBorder(null);
			wantPane.setPreferredSize(rowSize);
			if (row == 0) {
				// on the first row, add the label
				JLabel wantLab = new JLabel("Wanted assets: ", SwingConstants.LEFT);
				wantLab.setFont(whp.M_LabelFont);
				wantLab.setForeground(labColor);
				wantPane.add(wantLab);
				// do one less checkbox
				rEnd--;
			}
			int onRow = 0;
			for (int aaa = aItem; aaa < rEnd; aaa++) {
				String asset = trAssets.get(aaa);
				JCheckBox wBox = new JCheckBox(asset);
				// these boxes must have select listeners to adjust m_PercentUI
				wBox.addItemListener(this);
				wBox.setToolTipText("Check to find trades accepting " + asset);
				m_WantedBoxes.add(wBox);
				wantPane.add(wBox);
				onRow++;
			}
			aItem = rEnd;
			// pad with struts if row isn't full, to retain alignment
			if (row != 0) {
				int boxesShort = maxBoxes - onRow;
				for (int sss = 0; sss < boxesShort; sss++) {
					wantPane.add(Box.createHorizontalStrut(30));
				}
			}
			m_CriteriaPane.add(wantPane, m_GBC);
		}
		m_CriteriaPane.add(Box.createVerticalStrut(5), m_GBC);

		// add control for entering ratio (price)
		JPanel ratioPane = new JPanel(new GridLayout(1, 6));
		ratioPane.setBorder(null);
		ratioPane.setPreferredSize(rowSize);
		m_IgnoreRatio = new JRadioButton("Ignore", true);
		m_IgnoreRatio.setToolTipText("Find all matching trades");
		m_EqualRatio = new JRadioButton("Equal", false);
		m_EqualRatio.setToolTipText("Find trades at par (1:1) or better");
		m_DiscountRatio = new JRadioButton("Disc.", false);
		m_DiscountRatio.setToolTipText("Find trades offered at a percentage "
												+ "discount");
		m_PremiumRatio = new JRadioButton("Prem.", false);
		m_PremiumRatio.setToolTipText("Include trades requiring a premium "
												+ "percentage");
		m_RatioGroup = new ButtonGroup();
		m_RatioGroup.add(m_IgnoreRatio);
		m_RatioGroup.add(m_EqualRatio);
		m_RatioGroup.add(m_DiscountRatio);
		m_RatioGroup.add(m_PremiumRatio);
		ratioPane.setPreferredSize(rowSize);
		m_PercentUI = new JTextField("", 2);
		m_PercentUI.setDropTarget(null);
		String ratioTip = "<html>This ratio is between Offered and Wanted assets."
								+ "<br/><br/>"
								+ "To ignore this criterion and return all otherwise "
								+ "matching trades, use the Ignore button.<br/><br/>"
								+ "To restrict your search to trades specifying equal "
								+ "exchange or better, use the Equal button.<br/><br/>"
								+ "To restrict your search to trades offering a given "
								+ "discount (more of Offered than Wanted),<br/>"
								+ "click the Discount button and enter a percentage "
								+ "for the minimum discount.<br/><br/>"
								+ "To include in your search trades demanding a "
								+ "premium (less of Offered than Wanted),<br/>click "
								+ "the Premium button an enter a percentage for "
								+ "the maximum premium.<br/><br/>"
								+ "Note that if you check "
								+ "multiple boxes for Offered or Wanted assets, you "
								+ "must use Ignore.<br/></html>";
		m_PercentUI.setToolTipText(ratioTip);
		JLabel ratioLab = new JLabel("Ratio:", SwingConstants.CENTER);
		ratioLab.setToolTipText("To specify ratio of Offered to Wanted");
		ratioLab.setFont(whp.M_LabelFont);
		ratioLab.setForeground(labColor);
		ratioPane.add(ratioLab);
		ratioPane.add(m_IgnoreRatio);
		ratioPane.add(m_EqualRatio);
		ratioPane.add(m_DiscountRatio);
		ratioPane.add(m_PremiumRatio);
		ratioPane.add(m_PercentUI);
		m_CriteriaPane.add(ratioPane, m_GBC);
		m_CriteriaPane.add(Box.createVerticalStrut(10), m_GBC);

		// add controls for specifying minimum and/or maximum quantities
		JPanel quanPane = new JPanel(new GridLayout(1, 4));
		quanPane.setPreferredSize(rowSize);
		m_MinimumUI = new JTextField(5);
		m_MinimumUI.setDropTarget(null);
		m_MinimumUI.setToolTipText("Specify minimum offer amount");
		m_MaximumUI = new JTextField(5);
		m_MaximumUI.setDropTarget(null);
		m_MaximumUI.setToolTipText("Specify maximum offer amount");
		JLabel qLab1 = new JLabel("Quantity:  min. ", SwingConstants.CENTER);
		qLab1.setFont(whp.M_LabelFont);
		qLab1.setForeground(labColor);
		qLab1.setLabelFor(m_MinimumUI);
		quanPane.add(qLab1);
		quanPane.add(m_MinimumUI);
		JLabel qLab2 = new JLabel(" max. ", SwingConstants.CENTER);
		qLab2.setFont(whp.M_LabelFont);
		qLab2.setForeground(labColor);
		qLab2.setLabelFor(m_MaximumUI);
		quanPane.add(qLab2);
		quanPane.add(m_MaximumUI);
		m_CriteriaPane.add(quanPane, m_GBC);
		m_CriteriaPane.add(Box.createVerticalGlue(), m_GBC);

		// set the user instructions
		m_UserInstructs = "Specify search criteria using the controls below.<br/>"
								+ "<br/>Note that you can specify a desired premium or "
								+ "discount ratio for only one asset at a time.";
		m_WorkArea.setBorder(null);
	}

	/**
	 * perform a (blocking) query to the SVX obtaining all extant offers
	 */
	private void queryOffers() {
		// send a REQ_get_offers to fetch all matching offers from other users
		SVXListener svxIface = m_Plugin.getSVXListener();
		SVXClientBlock getOffers = new SVXClientBlock();
		getOffers.setOpcode("REQ_get_offers");

		// set search criteria based on values in controls
		StringBuilder offered = new StringBuilder(128);
		boolean first = true;
		for (JCheckBox oBox : m_OfferedBoxes) {
			if (oBox.isSelected()) {
				if (!first) {
					offered.append(":" + oBox.getText());
				}
				else {
					offered.append(oBox.getText());
					first = false;
				}
			}
		}
		getOffers.setOfferedAssets(offered.toString());
		StringBuilder wanted = new StringBuilder(128);
		first = true;
		int wTally = 0;
		for (JCheckBox wBox : m_WantedBoxes) {
			if (wBox.isSelected()) {
				wTally++;
				if (!first) {
					wanted.append(":" + wBox.getText());
				}
				else {
					wanted.append(wBox.getText());
					first = false;
				}
			}
		}
		getOffers.setWantedAssets(wanted.toString());
		if (wTally > 1) {
			// code is supposed to check for this, but...
			m_PercentUI.setText("");
			if (!m_EqualRatio.isSelected()) {
				m_IgnoreRatio.setSelected(true);
				m_Ratio = 0.0F;
			}
			else {
				m_Ratio = 1.0F;
			}
		}
		getOffers.setPrice(m_Ratio);
		if (m_Minimum > 0.0) {
			getOffers.setMinAmount(m_Minimum);
		}
		if (m_Maximum > 0.0) {
			getOffers.setMaxAmount(m_Maximum);
		}

		// send message
		setCursor(M_WaitCursor);
		svxIface.sendRequest(getOffers);
		setCursor(null);
		// NB: reply comes synchronously to processOffersReply() below
	}

	/**
	 * record the reply to our query of extant offers
	 * @param reply the reply received from the SVX
	 */
	public void processOffersReply(SVXBlock reply) {
		if (reply == null) {
			Log.error("Missing SVX config reply");
			return;
		}
		if (!reply.getOpcode().equalsIgnoreCase("REP_offers")) {
			Log.error("Unexpected reply opcode, " + reply.getOpcode());
		}
		if (reply.isError()) {
			m_TabManager.showError("SVX offer search failed", reply.getErrMsg());
			return;
		}

		// pass to the data model, which also updates the table on screen
		OfferTableModel otm = m_ParentHome.getOfferModel();
		ArrayList<SVXBlock.OpenOffer> repOffers = reply.getMatchingOffers();
		if (repOffers.isEmpty()) {
			otm.mergeOffers(null);
			return;
		}
		ArrayList<TradeOffer> offers
			= new ArrayList<TradeOffer>(repOffers.size());
		for (SVXBlock.OpenOffer off : repOffers) {
			TradeOffer offer = new TradeOffer();
			offer.setOfferId(off.m_OfferId);
			// only funded status orders are retrieved
			offer.setStatus(TradeOffer.M_StatusFunded);
			offer.setOffered(off.m_EscrowAsset);
			offer.setQuantity(off.m_Quantity);
			offer.setWanted(off.m_WantedAssets);
			offer.setPrice(off.m_Price);
			// NB: m_EscrowWallet is not returned with REP_offers
			offer.setExpiration(off.m_Expiry);
			offers.add(offer);
		}
		otm.mergeOffers(offers);
	}

	// extend WorkPane
	/**
	 * generate display for this sub-class
	 */
	public void prepDisplay() {
		// set screen label
		buildScreenLabel("Search trade offers entered by other users");

		if (m_ShowResults) {
			// make Return button enabled, others not (Accept set on select)
			m_ReturnButton.setEnabled(true);
			m_AcceptButton.setEnabled(false);
			m_SearchButton.setEnabled(false);

			// place the offer list into work area, set for radio select mode
			OfferTableModel otm = m_ParentHome.getOfferModel();
			otm.deselectAll();
			otm.setSelectMode(OfferTableModel.M_SelectSingle);
			JTable vt = m_ParentHome.getOfferTable();
			m_WorkArea.setViewportView(vt);

			// make sure that next invocation takes us to criteria selection
			m_ShowResults = false;
		}
		else {
			// make Return and Accept buttons disabled, Search enabled
			m_ReturnButton.setEnabled(false);
			m_AcceptButton.setEnabled(false);
			m_SearchButton.setEnabled(true);

			// place entire quasi-table UI into the main scroll area
			m_WorkArea.setViewportView(m_CriteriaPane);
		}
		revalidate();
	}

	/**
	 * method to receive notification of a change in offer selection
	 */
	public void offerSelectChanged() {
		// set button status according to state of selected offer
		OfferTableModel otm = m_ParentHome.getOfferModel();
		if (otm.getSelectMode() != OfferTableModel.M_SelectSingle) {
			// not meant for us, do nothing
			return;
		}

		ArrayList<TradeOffer> selOffers = otm.getSelectedOffers();
		if (selOffers.isEmpty()) {
			m_AcceptButton.setEnabled(false);
			return;
		}
		if (selOffers.size() > 1) {
			Log.error("Unexpected number of offer selections, "
							+ selOffers.size());
			return;
		}
		TradeOffer offer = selOffers.get(0);
		String stat = offer.getStatus();
		if (stat.equals(TradeOffer.M_StatusFunded)) {
			// update button
			m_AcceptButton.setEnabled(true);
		}
		// else: must not be meant for us
	}

	// implement ItemListener
	/**
	 * method to reset m_Ratio to 1.0 if multiple wanted assets are checked
	 */
	public void itemStateChanged(ItemEvent ie) {
		// tally the number of wanted checkboxes that are checked now
		int wchkd = 0;
		for (JCheckBox wBox : m_WantedBoxes) {
			if (wBox.isSelected()) {
				wchkd++;
			}
		}
		if (wchkd > 1) {
			// must override ratio to 0.0
			m_PercentUI.setText("");
			m_PercentUI.setEnabled(false);
			m_IgnoreRatio.setSelected(true);
			m_Ratio = 0.0F;
		}

		// tally the number of offered checkboxes that are checked now
		int ochkd = 0;
		for (JCheckBox wBox : m_OfferedBoxes) {
			if (wBox.isSelected()) {
				ochkd++;
			}
		}
		if (ochkd > 1) {
			// must override ratio to 0.0
			m_PercentUI.setText("");
			m_PercentUI.setEnabled(false);
			m_IgnoreRatio.setSelected(true);
			m_Ratio = 0.0F;
		}
		if (ochkd <= 1 && wchkd <= 1) {
			m_PercentUI.setEnabled(true);
		}
	}

	// END methods
}
