/*
 * last modified---
 * 	02-13-18/kbw/ correct calculation of premium ratio; limit to 4 digits
 * 	07-15-15/jet/ use lookup units for relative prices; adjust formatting
 * 					  to accommodate lots of assets making scroll bars appear
 * 	08-06-14/jet/ Spark plugin version
 * 	03-12-14/jet/ show commission as max of percent calculated or 25 tokens
 * 	11-09-13/jet/ specify price ratio as par, or percentage discount/premium
 * 	09-11-13/jet/ present single revised popup on escrow funding success
 * 	07-12-13/jet/ new
 *
 * purpose---
 * 	provide a UI screen to enter and fund trade offers
 */

package org.silentvault.spark.ui.svx;

import javax.swing.*;
import java.awt.event.*;
import java.awt.Font;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;
import org.silentvault.spark.ui.wallet.HomePane;
import org.silentvault.spark.ui.wallet.WTabManager;
import org.silentvault.spark.ui.wallet.PaymentPane;
import org.silentvault.spark.ui.wallet.ConfirmPane;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Date;
import java.util.Calendar;
import java.text.NumberFormat;
import java.text.DecimalFormat;


/**
 * Extends WorkPane to provide a screen for entering trade offers, sending
 * them to the SVX, and then funding them using a voucher payment.
 *
 * @author Justin Turrell
 */
public final class TradePane extends WorkPane implements ItemListener {
	// BEGIN data members
	/**
	 * action to initiate an entry of the indicated trade offer
	 */
	private AbstractAction			m_EnterAction;

	/**
	 * button to go with enter action
	 */
	private JButton					m_EnterButton;

	/**
	 * action to acknowledge offer resolution language and reset
	 */
	private AbstractAction			m_AckAction;

	/**
	 * button to go with acknowledge action
	 */
	private JButton					m_AckButton;

	/**
	 * mode toggle: if true, display trade entry form; if false show confirmation
	 */
	private boolean					m_EnterMode;

	/**
	 * control to select asset being offered
	 */
	private JComboBox<String>		m_OfferedUI;

	/**
	 * list of controls to specify wanted assets
	 */
	private ArrayList<JCheckBox>	m_WantedBoxes;

	/**
	 * box for entering desired quantity (restricted by voucher total in wallet)
	 */
	private JTextField				m_AmountUI;

	/**
	 * value specified via m_AmountUI
	 */
	private double						m_Amount;

	/**
	 * box for entering price ratio premium percent
	 */
	private JTextField				m_PremiumUI;

	/**
	 * box for entering price ratio discount percent
	 */
	private JTextField				m_DiscountUI;

	/**
	 * radio button for par (equal) price ratio
	 */
	private JRadioButton				m_EqualRatio;

	/**
	 * radio button for premium price ratio (refers to m_PremiumUI)
	 */
	private JRadioButton				m_PremiumRatio;

	/**
	 * radio button for discount price ratio (refers to m_DiscountUI)
	 */
	private JRadioButton				m_DiscountRatio;

	/**
	 * button group for price ratio selection
	 */
	private ButtonGroup				m_RatioGroup;

	/**
	 * value specified via m_RatioGroup and possibly m_PremiumUI or m_DiscountUI,
	 * computed appropriately for server messaging
	 */
	private float						m_Ratio;

	/**
	 * control to select expiration, expressed as days in future
	 */
	private JTextField				m_ExpDaysUI;

	/**
	 * value specified via m_ExpDaysUI
	 */
	private short						m_ExpDays;

	/**
	 * label describing commission rate (adjusted by m_OfferedUI)
	 */
	private JLabel						m_CommRate;

	/**
	 * panel to contain trade entry form
	 */
	private JPanel						m_FormPane;

	/**
	 * offer ID entered and assigned by SVX
	 */
	private int							m_OfferId;

	/**
	 * escrow wallet designated by the SVX for payment
	 */
	private String						m_EscrowWalletId;

	/**
	 * total payment to be made to fund escrow (includes commission)
	 */
	private double						m_EscrowAmount;

	/**
	 * flag to indicate (to WTabManager) that we're making an escrow payment
	 */
	private boolean					m_DoingEscrow;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the XHomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public TradePane(WalletClient plugin, XHomePane home, XTabManager mgr) {
		super(plugin, home, mgr);
		m_Ratio = 1.0F;
		m_EnterMode = true;
		m_ExpDays = 30;
		m_EscrowWalletId = "";

		// define Enter action
		m_EnterAction = new AbstractAction("Enter Offer") {
			public void actionPerformed(ActionEvent ae) {
				DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
				df.setMaximumFractionDigits(XMLVoucher.M_Precision);
				df.setMinimumFractionDigits(XMLVoucher.M_Precision);
				df.setGroupingUsed(false);
				// error-check value in buttons and input for specifying ratio
				if (m_EqualRatio.isSelected()) {
					m_Ratio = 1.0F;
				}
				else if (m_PremiumRatio.isSelected()) {
					String ratioTxt = m_PremiumUI.getText();
					if (ratioTxt.isEmpty()) {
						m_Ratio = 1.0F;
					}
					else {
						// convert to ratio
						try {
							Float pct = Float.parseFloat(ratioTxt);
							float prem = pct.floatValue();
							if (prem <= 0.0F) {
								showError("illegal premium value",
											"enter a positive percentage");
								return;
							}
							// apply translation to premium ratio
							m_Ratio = 1.0F - (prem / 100.0F);
						}
						catch (NumberFormatException nfe) {
							showError("illegal premium value",
										"enter a positive percentage");
							return;
						}
					}
				}
				else if (m_DiscountRatio.isSelected()) {
					String ratioTxt = m_DiscountUI.getText();
					if (ratioTxt.isEmpty()) {
						m_Ratio = 1.0F;
					}
					else {
						// convert to ratio
						try {
							Float pct = Float.parseFloat(ratioTxt);
							float disc = pct.floatValue();
							if (disc <= 0.0F) {
								showError("illegal discount value",
											"enter a positive percentage");
								return;
							}
							// apply translation to discount ratio
							m_Ratio = 1.0F + (disc / 100.0F);
						}
						catch (NumberFormatException nfe) {
							showError("illegal discount value",
										"enter a positive percentage");
							return;
						}
					}
				}
				if (m_Ratio != 1.0F) {
					// limit to 4 digits precision
					try { m_Ratio = Float.parseFloat(df.format(m_Ratio)); }
					catch (NumberFormatException nfe) { /* ignore */ }
				}

				// error check specified amount
				String offered = (String) m_OfferedUI.getSelectedItem();
				String amtText = m_AmountUI.getText();
				SVXAsset offConf = m_ParentHome.getSVXConfig().get(offered);
				if (amtText.isEmpty()) {
					m_Amount = 0.0;
				}
				else {
					try {
						Double amt = Double.parseDouble(amtText);
						m_Amount = amt.doubleValue();
					}
					catch (NumberFormatException nfe) {
						showError("illegal offer amount",
									"enter " + offConf.m_MinExchange + " or more");
						return;
					}
				}
				if (m_Amount < offConf.m_MinExchange) {
					showError("illegal offer amount",
								"enter " + offConf.m_MinExchange + " or more");
					return;
				}
				// make sure user has this much
				HomePane whp = m_ParentHome.getWalletHome();
				VoucherTableModel vtm = whp.getVoucherModel();
				double avail = vtm.getVoucherAssetSum(offered);
				if (m_Amount > avail) {
					showError("illegal offer amount", "you do not have this much "
								+ offered + " in your wallet");
					return;
				}

				// check exp days
				String daysTxt = m_ExpDaysUI.getText();
				if (daysTxt.isEmpty()) {
					m_ExpDays = 30;
				}
				else {
					try {
						Short days = Short.parseShort(daysTxt);
						m_ExpDays = days.shortValue();
					}
					catch (NumberFormatException nfe) {
						showError("illegal expiration days",
									"enter a number from 1 to 60");
						return;
					}
				}
				if (m_ExpDays <= 0 || m_ExpDays > 60) {
					showError("illegal expiration days",
								"enter a number from 1 to 60");
					return;
				}

				// make sure that at least one wanted asset is checked
				boolean checked = false;
				for (JCheckBox wBox : m_WantedBoxes) {
					if (wBox.isSelected()) {
						checked = true;
						break;
					}
				}
				if (!checked) {
					showError("no wanted assets selected",
								"select at least one asset to accept in trade");
					return;
				}

				// send REQ_make_offer to SVX
				sendEnterOffer();
			}
		};
		// define associated button
		final Font buttonFont = m_TabManager.M_ButtonFont;
		m_EnterButton = new JButton(m_EnterAction);
		m_EnterButton.setFont(buttonFont);
		m_EnterButton.setToolTipText("Submit the trade offer");

		// define funding acknowledgement action
		m_AckAction = new AbstractAction("Acknowledged") {
			public void actionPerformed(ActionEvent ae) {
				// reset for next trade entry
				m_EnterMode = true;
				m_OfferId = 0;
				m_EscrowWalletId = "";
				m_EscrowAmount = 0.0;
				m_Amount = 0.0;
				m_AmountUI.setText("");
				m_Ratio = 1.0F;
				m_PremiumUI.setText("");
				m_DiscountUI.setText("");
				m_EqualRatio.setSelected(true);
				m_ExpDaysUI.setText("30");
				m_ExpDays = 30;
				for (JCheckBox wBox : m_WantedBoxes) {
					wBox.setSelected(false);
				}
				m_EnterButton.setEnabled(true);
				m_TabManager.queueTransition(new Integer(XTabManager.M_GO_TRADE));
				m_TabManager.makeTransition();
			}
		};
		// define associated button
		m_AckButton = new JButton(m_AckAction);
		m_AckButton.setFont(buttonFont);
		m_AckButton.setToolTipText("Return to trade entry screen");

		// populate button bar (replaced in prepDisplay())
		m_ButtonPane.removeAll();
		m_ButtonPane.add(Box.createHorizontalStrut(170));		// spacer
		m_ButtonPane.add(m_EnterButton);
		m_ButtonPane.add(Box.createHorizontalStrut(30));		// spacer
		m_ButtonPane.add(m_AckButton);

		// create controls for offer entry form
		m_FormPane = new JPanel(new GridBagLayout());
		m_FormPane.setBorder(null);
		m_GBC.fill = GridBagConstraints.NONE;
		m_GBC.anchor = GridBagConstraints.NORTH;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		final Color labColor = UIManager.getColor("standardGray");
		HomePane whp = m_ParentHome.getWalletHome();
		JLabel headLab = new JLabel("Enter a new trade offer as follows:",
											SwingConstants.LEFT);
		headLab.setFont(whp.M_LabelFont);
		headLab.setForeground(labColor);

		// build list of trade assets selectable
		Hashtable<String, SVXAsset> svxConfig = m_ParentHome.getSVXConfig();
		Vector<String> trAssets = new Vector<String>(svxConfig.size());
		for (String asset : svxConfig.keySet()) {
			SVXAsset aConf = svxConfig.get(asset);
			if (aConf.m_MinExchange > 0.0) {
				trAssets.add(aConf.m_AssetName);
			}
		}
		m_WantedBoxes = new ArrayList<JCheckBox>(trAssets.size());
		final int maxBoxes = 4;
		int rowCnt = (trAssets.size() / maxBoxes) + 1;
		m_GBC.gridheight = 14 + rowCnt;
		m_FormPane.add(Box.createVerticalGlue(), m_GBC);
		m_FormPane.add(headLab, m_GBC);
		m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

		// based on the list, create label and drop-down for offered assets
		final Dimension rowSize = new Dimension(470, 25);
		final Dimension boxSz = new Dimension(50, 25);
		JPanel offPane = new JPanel(new GridLayout(1, 2));
		offPane.setBorder(null);
		offPane.setPreferredSize(rowSize);
		// offered boxes: label
		JLabel offLab = new JLabel("Offered: ", SwingConstants.CENTER);
		offLab.setFont(whp.M_LabelFont);
		offLab.setForeground(labColor);
		offLab.setToolTipText("Asset you wish to trade");
		offPane.add(offLab);
		// NB: prepDisplay() will update the asset list to match wallet contents
		m_OfferedUI = new JComboBox<String>(trAssets);
		m_OfferedUI.addActionListener(this);
		m_OfferedUI.setPreferredSize(boxSz);
		m_OfferedUI.setToolTipText("Select the asset you wish to offer");
		offLab.setLabelFor(m_OfferedUI);
		offPane.add(m_OfferedUI);
		m_FormPane.add(offPane, m_GBC);
		m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

		// add control for entering the amount on offer
		JPanel amtPane = new JPanel(new GridLayout(1, 2));
		amtPane.setPreferredSize(rowSize);
		m_AmountUI = new JTextField(6);
		m_AmountUI.setDropTarget(null);
		m_AmountUI.setToolTipText("Enter amount you wish to offer");
		m_AmountUI.setPreferredSize(boxSz);
		JLabel amtLab = new JLabel("Amount: ", SwingConstants.CENTER);
		amtLab.setFont(whp.M_LabelFont);
		amtLab.setForeground(labColor);
		amtLab.setLabelFor(m_AmountUI);
		amtLab.setToolTipText("Quantity you wish to trade");
		amtPane.add(amtLab);
		amtPane.add(m_AmountUI);
		m_FormPane.add(amtPane, m_GBC);
		m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

		// based on the list, create label and checkboxes for wanted assets
		// (format into 4 boxes per row)
		int aItem = 0;
		for (int row = 0; row < rowCnt; row++) {
			int rStart = aItem;
			int rEnd = Math.min(rStart + maxBoxes, trAssets.size());
			int rCnt = rEnd - rStart;
			JPanel wantPane = new JPanel(new GridLayout(1, rCnt));
			wantPane.setBorder(null);
			wantPane.setPreferredSize(rowSize);
			JLabel wantLab = new JLabel("Wanted: ", SwingConstants.CENTER);
			wantLab.setFont(whp.M_LabelFont);
			wantLab.setForeground(labColor);
			wantLab.setToolTipText("List of assets you will accept in trade");
			wantPane.add(wantLab);
			if (row != 0) {
				// on later rows, make this label invisible (still want placeholder)
				wantLab.setVisible(false);
			}
			int onRow = 0;
			for (int aaa = aItem; aaa < rEnd; aaa++) {
				String asset = trAssets.get(aaa);
				JCheckBox wBox = new JCheckBox(asset);
				// these boxes must have select listeners to adjust radio buttons
				wBox.addItemListener(this);
				wBox.setToolTipText("Check to accept " + asset);
				m_WantedBoxes.add(wBox);
				wantPane.add(wBox);
				onRow++;
			}
			aItem = rEnd;
			// pad with struts if row isn't full, to retain alignment
			int boxesShort = maxBoxes - onRow;
			for (int sss = 0; sss < boxesShort; sss++) {
				wantPane.add(Box.createHorizontalStrut(50));
			}
			m_FormPane.add(wantPane, m_GBC);
		}
		m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

		// add labels to show commission which SVX will charge
		JPanel commPane = new JPanel(new GridLayout(1, 2));
		commPane.setBorder(null);
		commPane.setPreferredSize(rowSize);
		JLabel comLab = new JLabel("SVX commission rate: ",
											SwingConstants.CENTER);
		comLab.setFont(whp.M_LabelFont);
		comLab.setForeground(labColor);
		comLab.setToolTipText("The commission charged by the SVX "
									+ "for trades in your offered asset");
		commPane.add(comLab);
		m_CommRate = new JLabel("0.0 %", SwingConstants.CENTER);
		m_CommRate.setFont(whp.M_LabelFont);
		m_CommRate.setForeground(labColor);
		m_CommRate.setToolTipText("The commission rate");
		// NB: m_CommRate will be updated with changes in m_OfferedUI
		commPane.add(m_CommRate);
		m_FormPane.add(commPane, m_GBC);
		m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

		// add controls for entering ratio (price)
		JPanel ratioPane = new JPanel(new GridLayout(1, 6));
		ratioPane.setBorder(null);
		ratioPane.setPreferredSize(rowSize);
		m_EqualRatio = new JRadioButton("Equal", true);
		m_EqualRatio.setToolTipText("Check to trade at par (1:1)");
		m_PremiumRatio = new JRadioButton("Prem.", false);
		m_PremiumRatio.setToolTipText("Check to trade at a premium, in percent");
		m_DiscountRatio = new JRadioButton("Disc.", false);
		m_DiscountRatio.setToolTipText("Check to trade at a discount, "
												+ "in percent");
		m_RatioGroup = new ButtonGroup();
		m_RatioGroup.add(m_EqualRatio);
		m_RatioGroup.add(m_PremiumRatio);
		m_RatioGroup.add(m_DiscountRatio);
		m_PremiumUI = new JTextField("", 2);
		m_PremiumUI.setDropTarget(null);
		String blurbage = "<br/>Note that if you check multiple boxes "
								+ "for your Wanted assets,<br/>then this value is "
								+ "ignored, and trades will be entered at par (1:1)."
								+ "<br/><br/>"
								+ "All escrow transactions are settled according to "
								+ "the relative market<br/>prices between the two "
								+ "asset types involved in the trade, at the "
								+ "time<br/>your offer is accepted by another party."
								+ "</html>";
		String premTip = "<html>Select the button and enter a premium percentage "
								+ "for your Offered asset.<br/>For example, to require "
								+ "3.5% more of your Wanted asset, enter 3.5."
								+ blurbage;
		m_PremiumUI.setToolTipText(premTip);
		m_DiscountUI = new JTextField("", 2);
		m_DiscountUI.setDropTarget(null);
		String discTip = "<html>Select the button and enter a discount percentage"
								+ " for your Offered asset.<br/>For example, to give "
								+ "3.5% more for your Wanted asset, enter 3.5."
								+ blurbage;
		m_DiscountUI.setToolTipText(discTip);
		JLabel ratioLab = new JLabel("Pricing:", SwingConstants.CENTER);
		ratioLab.setFont(whp.M_LabelFont);
		ratioLab.setForeground(labColor);
		ratioLab.setToolTipText("To set ratio at which Offered is traded "
										+ "for Wanted");
		ratioPane.add(ratioLab);
		ratioPane.add(m_EqualRatio);
		ratioPane.add(m_PremiumRatio);
		ratioPane.add(m_PremiumUI);
		ratioPane.add(m_DiscountRatio);
		ratioPane.add(m_DiscountUI);
		m_FormPane.add(ratioPane, m_GBC);
		m_FormPane.add(Box.createVerticalStrut(5), m_GBC);

		// add controls for entering the expiration date
		JPanel expPane = new JPanel(new GridLayout(1, 2));
		expPane.setBorder(null);
		expPane.setPreferredSize(rowSize);
		JLabel expLab = new JLabel("Days till expiration: ",
											SwingConstants.CENTER);
		expLab.setFont(whp.M_LabelFont);
		expLab.setForeground(labColor);
		expLab.setToolTipText("Number of days till your offer will expire");
		// box to enter number of days in future
		m_ExpDaysUI = new JTextField("30", 3);
		m_ExpDaysUI.setDropTarget(null);
		m_ExpDaysUI.setToolTipText("Enter 1-60 days in future");
		m_ExpDaysUI.setPreferredSize(boxSz);
		expLab.setLabelFor(m_ExpDaysUI);
		expPane.add(expLab);
		expPane.add(m_ExpDaysUI);
		m_FormPane.add(expPane, m_GBC);
		m_FormPane.add(Box.createVerticalGlue(), m_GBC);

		// set the user instructions
		m_UserInstructs = "Enter your trade offer below.  Note that you must "
								+ "fund escrow before your offer will be seen by "
								+ "other users, so you can only trade assets and "
								+ "amounts for which you have vouchers in your wallet.";
		m_WorkArea.setBorder(null);
	}

	/**
	 * perform a (blocking) query to the SVX entering the offer in the database
	 */
	public void sendEnterOffer() {
		// send a REQ_make_offer to register the trade offer at the SVX
		SVXListener svxIface = m_Plugin.getSVXListener();
		SVXClientBlock mkOffer = new SVXClientBlock();
		mkOffer.setOpcode("REQ_make_offer");

		// set search criteria based on values in controls
		String offered = (String) m_OfferedUI.getSelectedItem();
		mkOffer.setAsset(offered);
		mkOffer.setQuantity(m_Amount);
		StringBuilder wanted = new StringBuilder(128);
		boolean first = true;
		int wTally = 0;
		for (JCheckBox wBox : m_WantedBoxes) {
			if (wBox.isSelected()) {
				wTally++;
				if (!first) {
					wanted.append(":" + wBox.getText());
				}
				else {
					wanted.append(wBox.getText());
					first = false;
				}
			}
		}
		mkOffer.setWantedAssets(wanted.toString());
		if (wTally > 1) {
			// code is supposed to check for this, but...
			m_Ratio = 1.0F;
			m_PremiumUI.setText("");
			m_DiscountUI.setText("");
			m_EqualRatio.setSelected(true);
		}
		mkOffer.setPrice(m_Ratio);
		// calculate expiration date at GMT
		Calendar cal = Calendar.getInstance();
		int gmtOffsetMillis = cal.get(Calendar.ZONE_OFFSET);
		int dstOffsetMillis = cal.get(Calendar.DST_OFFSET);
		Date tNow = new Date();
		Date expDate = new Date(tNow.getTime()
									+ (m_ExpDays * javax.management.timer.Timer.ONE_DAY)
										+ gmtOffsetMillis + dstOffsetMillis);
		mkOffer.setExpiration(expDate.getTime());
		mkOffer.setWalletId(m_Plugin.getLoginSecrets().getVSnumber());

		// send message
		setCursor(M_WaitCursor);
		svxIface.sendRequest(mkOffer);
		setCursor(null);
		// NB: reply comes synchronously to processOfferReply() below
	}

	/**
	 * record the reply to our query of extant offers
	 * @param reply the reply received from the SVX
	 */
	public void processOfferReply(SVXBlock reply) {
		if (reply == null) {
			Log.error("Missing SVX offer search reply");
			return;
		}
		if (!reply.getOpcode().equalsIgnoreCase("REP_offer_entered")) {
			Log.error("Unexpected reply opcode, " + reply.getOpcode());
		}
		if (reply.isError()) {
			m_TabManager.showError("SVX offer entry failed", reply.getErrMsg());
			return;
		}

		// record values returned from SVX
		m_OfferId = reply.getOfferId();
		m_EscrowWalletId = reply.getWalletId();
		m_EscrowAmount = reply.getTotal();

		// build a (decoded) WebMessage object to pass data to the PaymentPane
		WebMessage webPass = new WebMessage();
		webPass.setAmount(m_EscrowAmount);
		webPass.setVSnumber(m_EscrowWalletId);
		String asset = (String) m_OfferedUI.getSelectedItem();
		// determine issuer for this asset by examining a voucher of the type
		HomePane whp = m_ParentHome.getWalletHome();
		VoucherTableModel vtm = whp.getVoucherModel();
		XMLVoucher firstV = vtm.getFirstAssetVoucher(asset);
		if (firstV == null) {
			// should not happen, since we vetted asset balance already
			m_TabManager.showError("Cannot fund escrow for offer",
										"no vouchers found for asset " + asset);
			Log.error("No vouchers found to fund offer " + m_OfferId + " with "
						+ asset);
			return;
		}
		vtm.selectIssuer(firstV.getIssuer());
		webPass.setIssuer(firstV.getIssuer());
		webPass.setCurrency(XMLVoucher.getDisplayUnits(asset));
		webPass.setBaggageFields("offerId");
		webPass.addOrderRecord("offerId", new Integer(m_OfferId).toString());
		webPass.setDecoded(true);
		m_Plugin.getWebListener().addWebMessage(webPass);
		final WTabManager wtm = m_Plugin.getWalletTabManager();
		PaymentPane pp = wtm.getPaymentPane();
		pp.setSelectedSiteMsg(webPass);

		// send user to payment screen in WTabManager
		Runnable goPayment = new Runnable() {
			public void run() {
				// disable button to prevent multiple entry
				m_EnterButton.setEnabled(false);
				m_EnterButton.setToolTipText("<html>Your trade has been entered."
													+ "<br/>Please make your escrow "
													+ "payment in the Wallet tab.</html>");
				revalidate();
				wtm.queueTransition(new Integer(WTabManager.M_PAYMENTS1));
				m_Plugin.setSelectedTab(wtm);
				wtm.makeTransition();
			}
		};
		m_DoingEscrow = true;
		SwingUtilities.invokeLater(goPayment);
	}

	/**
	 * query whether we're doing an escrow payment or not
	 * @return true if in the middle of conducting a payment to escrow
	 */
	public boolean doingEscrow() { return m_DoingEscrow; }

	/**
	 * make funding confirmation request
	 */
	public void confirmEscrowFunding() {
		// send a REQ_fund_offer to pickup the escrow payment at the SVX
		SVXListener svxIface = m_Plugin.getSVXListener();
		SVXClientBlock fundOffer = new SVXClientBlock();
		fundOffer.setOpcode("REQ_fund_offer");
		fundOffer.setWalletId(m_Plugin.getLoginSecrets().getVSnumber());
		fundOffer.setOfferId(m_OfferId);
		fundOffer.setAsset((String) m_OfferedUI.getSelectedItem());

		// return focus to exchange tab
		m_Plugin.setSelectedTab(m_TabManager);

		// put up info saying what we are doing (on SVX tab)
		WTabManager wtm = m_Plugin.getWalletTabManager();
		PaymentPane pp = wtm.getPaymentPane();
		final StringBuilder popTxt = new StringBuilder(256);
		// obtain last confirmation recorded by ConfirmPane off the DHT
		XMLConfirmPayment conf = pp.getLatestConfirmation();
		if (conf != null) {
			/* we deferred showing payment success, and so must recapitulate
			 * what the ConfirmPane would have displayed as popup text.
			 */
			popTxt.append("Your payment of " + conf.getQuantity());
			popTxt.append(" " + conf.getUnits() + " of " + conf.getAsset());
			if (!conf.getAsset().equals(conf.getIssuer())) {
				popTxt.append(" at " + conf.getIssuer());
			}
			popTxt.append("\nwas sent to ");
			popTxt.append(conf.getPayee() + " successfully.");
			// show that change voucher was received if there was one
			if (conf.getVoucher() != null) {
				popTxt.append("\nYou have received a voucher back as change.");
			}
		}
		else {
			Log.error("Unexpected: no XMLConfirmPayment found in PaymentPane");
			// show something generic
			popTxt.append("Your escrow payment to fund offer ID ");
			popTxt.append(fundOffer.getOfferId() + " succeeded.");
		}

		// add generic indication that we're now telling the SVX to pick it up
		popTxt.append("\n\nThe SVX is now processing the escrow payment.  ");
		popTxt.append("This may take a few moments.\nPlease wait for the ");
		popTxt.append("confirmation screen, read it carefully, and acknowledge.");

		// for safety, do in separate thread while allowing message to proceed
		Runnable notice = new Runnable() {
			public void run() {
				JOptionPane.showMessageDialog(m_TabManager, popTxt.toString(),
														"Escrow Payment Completed",
														JOptionPane.INFORMATION_MESSAGE);
			}
		};
		SwingUtilities.invokeLater(notice);

		// reset ConfirmPane to default behavior
		ConfirmPane cp = wtm.getConfirmPane();
		cp.setDeferPopup(false);

		// send message to SVX to trigger pickup
		setCursor(M_WaitCursor);
		svxIface.sendRequest(fundOffer);
		setCursor(null);
		// NB: reply comes synchronously to processFundingReply() below
	}

	/**
	 * record the results of an offer funding attempt (post-payment)
	 * @param reply the reply received from the SVX
	 */
	public void processFundingReply(SVXBlock reply) {
		if (reply == null) {
			Log.error("Missing SVX offer entry reply");
			return;
		}
		if (!reply.getOpcode().equalsIgnoreCase("REP_offer_funded")) {
			Log.error("Unexpected reply opcode, " + reply.getOpcode());
		}
		m_DoingEscrow = false;
		if (reply.isError()) {
			m_TabManager.showError("SVX offer funding failed", reply.getErrMsg());
			// reset
			m_AckButton.doClick();
			return;
		}

		// go to confirmation
		m_EnterMode = false;
		Runnable goConf = new Runnable() {
			public void run() {
				m_TabManager.queueTransition(new Integer(XTabManager.M_GO_TRADE));
				m_TabManager.makeTransition();
			}
		};
		SwingUtilities.invokeLater(goConf);
	}

	// extend WorkPane
	/**
	 * generate display for this sub-class
	 */
	public void prepDisplay() {
		// set screen label
		buildScreenLabel("Enter a new trade offer");

		HomePane whp = m_ParentHome.getWalletHome();
		if (m_EnterMode) {
			// abandon any uncompleted escrow payment for a previous offer
			m_DoingEscrow = false;

			// populate button bar
			m_ButtonPane.removeAll();
			m_ButtonPane.add(Box.createHorizontalStrut(195));		// spacer
			m_ButtonPane.add(m_EnterButton);
			m_AckButton.setVisible(false);
			m_EnterButton.setVisible(true);
			m_EnterButton.setToolTipText("Submit the trade offer");

			// rebuild available assets based on wallet contents
			VoucherTableModel vtm = whp.getVoucherModel();
			Hashtable<String, SVXAsset> svxConfig = m_ParentHome.getSVXConfig();
			m_OfferedUI.removeAllItems();
			for (String tAsset : svxConfig.keySet()) {
				double sum = vtm.getVoucherAssetSum(tAsset);
				if (sum > 0.0) {
					m_OfferedUI.addItem(tAsset);
				}
			}

			// now take account of XMenuPane's drop-down trade asset selection
			XMenuPane xmp = m_TabManager.getMenuPane();
			String asset = xmp.getTradeAsset();
			// NB: does nothing if this asset is not in our allowed list
			m_OfferedUI.setSelectedItem(asset);

			// place entire quasi-table UI into the main scroll area
			m_WorkArea.setViewportView(m_FormPane);
		}
		else {
			// populate button bar
			m_ButtonPane.removeAll();
			m_ButtonPane.add(Box.createHorizontalStrut(180));		// spacer
			m_ButtonPane.add(m_AckButton);
			m_AckButton.setVisible(true);
			m_EnterButton.setVisible(false);

			// display language about offer handling, place in m_WorkArea
			JPanel confPane = new JPanel(new GridBagLayout());
			confPane.setBorder(null);
			confPane.setPreferredSize(new Dimension(470, 250));
			m_GBC.gridheight = 5;
			confPane.add(Box.createVerticalGlue(), m_GBC);
			JLabel confLab = new JLabel("Trade Offer Now Funded and Available",
												SwingConstants.CENTER);
			final Color labColor = UIManager.getColor("standardGray");
			confLab.setFont(whp.M_LabelFont);
			confLab.setForeground(labColor);
			confPane.add(confLab, m_GBC);
			confPane.add(Box.createVerticalStrut(5), m_GBC);
			StringBuilder confTxt = new StringBuilder(512);
			confTxt.append("Your trade offer, ID # " + m_OfferId + " has been ");
			confTxt.append("entered at the SVX, and your side\nof the escrow ");
			confTxt.append("has now been funded.\n\n");
			confTxt.append("When another user accepts your trade and also funds ");
			confTxt.append("escrow, you\nwill receive the asset type provided ");
			confTxt.append("by that user, which will be\none of your wanted ");
			confTxt.append("assets: ");
			boolean first = true;
			for (JCheckBox wBox : m_WantedBoxes) {
				if (wBox.isSelected()) {
					if (!first) {
						confTxt.append(", " + wBox.getText());
					}
					else {
						confTxt.append(wBox.getText());
						first = false;
					}
				}
			}
			confTxt.append(".\n\n");
			confTxt.append("The actual amount received will depend on the ");
			confTxt.append("relative market\nprices of the two assets being ");
			confTxt.append("traded, and your discount ratio (" +m_Ratio+ ").\n\n");
			confTxt.append("If your offer is not taken before it expires, it ");
			confTxt.append("will be canceled and\nyour escrow will be refunded ");
			confTxt.append("to your wallet.  You may also cancel\nyour offer ");
			confTxt.append("manually at any time from the My Offers screen.\n");
			JTextArea confBlurb = new JTextArea(confTxt.toString(), 10, 40);
			//confBlurb.setDropTarget(null);
			confBlurb.setEditable(false);
			confPane.add(confBlurb, m_GBC);
			confPane.add(Box.createVerticalGlue(), m_GBC);
			m_WorkArea.setViewportView(confPane);
		}
		revalidate();
	}

	// implement ItemListener
	/**
	 * method to reset m_Ratio to 1.0 if multiple wanted assets are checked
	 */
	public void itemStateChanged(ItemEvent ie) {
		// tally the number of wanted checkboxes that are checked now
		int chkd = 0;
		for (JCheckBox wBox : m_WantedBoxes) {
			if (wBox.isSelected()) {
				chkd++;
			}
		}
		if (chkd > 1) {
			// must override ratio to 1.0
			m_Ratio = 1.0F;
			m_PremiumUI.setText("");
			m_PremiumUI.setEnabled(false);
			m_DiscountUI.setText("");
			m_DiscountUI.setEnabled(false);
			m_EqualRatio.setSelected(true);
		}
		else {
			m_PremiumUI.setEnabled(true);
			m_DiscountUI.setEnabled(true);
		}
	}

	// implement ActionListener
	/**
	 * method to update displayed commission rate and max amount tooltips
	 * when an offered asset is selected
	 */
	@Override
	public void actionPerformed(ActionEvent ae) {
		// when m_OfferedUI changes, update m_CommRate appropriately
		Object evSrc = ae.getSource();
		if (evSrc instanceof JComboBox) {
			Hashtable<String, SVXAsset> svxConfig = m_ParentHome.getSVXConfig();
			String trAsset = (String) m_OfferedUI.getSelectedItem();
			if (trAsset == null || trAsset.isEmpty()) {
				return;
			}
			SVXAsset aConf = svxConfig.get(trAsset);
			if (aConf == null) {
				Log.error("Config for traded asset " + trAsset + " not found");
				return;
			}
			Float comm = new Float(aConf.m_ExchangeCommission);
			double twentyFiveTokens = VPAssetConfig.get25TokensWorth(trAsset);
			m_CommRate.setText(comm + " %");
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
			df.setMaximumFractionDigits(XMLVoucher.M_Precision);
			df.setMinimumFractionDigits(XMLVoucher.M_Precision);
			m_CommRate.setToolTipText("The commission rate for " + trAsset
										+ " (min. " + df.format(twentyFiveTokens) + ")");

			// update tooltip for max amount
			HomePane whp = m_ParentHome.getWalletHome();
			VoucherTableModel vtm = whp.getVoucherModel();
			double avail = vtm.getVoucherAssetSum(trAsset);
			df.setGroupingSize(3);
			m_AmountUI.setToolTipText("You have up to " + df.format(avail) + " "
												+ XMLVoucher.getDisplayUnits(trAsset)
												+ " available");

			// uncheck any Wanted box for same asset
			for (JCheckBox wBox : m_WantedBoxes) {
				String wAsset = wBox.getText();
				if (wAsset.equals(trAsset)) {
					wBox.setSelected(false);
					wBox.setEnabled(false);
					wBox.setToolTipText("Cannot accept " + wAsset + " for itself");
				}
				else {
					wBox.setEnabled(true);

					// also, show current relative price
					AssetPriceCache apc = m_Plugin.getPricingModule();
					String units1 = XMLVoucher.getDisplayUnits(trAsset);
					String units2 = XMLVoucher.getDisplayUnits(wAsset);
					String lUnits1 = VPAssetConfig.getBasePriceLookupUnits(trAsset);
					String lUnits2 = VPAssetConfig.getBasePriceLookupUnits(wAsset);
					Double relPrice = apc.getRelativePrice(lUnits1, lUnits2);
					wBox.setToolTipText("<html>Check to accept " + wAsset + "<br/>"
												+ " (current price = " + df.format(relPrice)
												+ " " + units2 + "/" + units1 + ")</html>");
				}
			}
		}
	}

	// END methods
}
