/*
 * last modified---
 * 	01-22-15/jet/ move menu toolbar definitions to XHomePane
 * 	08-06-14/jet/ Spark plugin version
 * 	06-05-13/jet/ new
 *
 * purpose---
 * 	provide a UI widget to act as the work area for doing user input/output
 */

package org.silentvault.spark.ui.svx;

import javax.swing.*;
import javax.swing.border.EmptyBorder;
import java.awt.Dimension;
import java.awt.Cursor;
import java.awt.Font;
import java.awt.Component;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.GridLayout;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;
import org.silentvault.spark.ui.wallet.HomePane;


/**
 * Extends JPanel to provide a base screen for doing exchange-related I/O.
 * This class will never be used directly, but subclasses of it will be
 * swapped into the XHomePane by the XTabManager, which can be invoked by
 * messaging callbacks.
 *
 * @author Justin Turrell
 */
public abstract class WorkPane extends JPanel implements ActionListener {
	// BEGIN data members
	/**
	 * standard wait (hourglass) cursor
	 */
	protected final Cursor	M_WaitCursor = 
											Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR);

	/**
	 * width of items in this pane
	 */
	private final int			M_WpWidth = 485;

	/**
	 * the plugin/application to which we belong
	 */
	protected WalletClient	m_Plugin;

	/**
	 * the home pane in which we are installed
	 */
	protected XHomePane		m_ParentHome;

	/**
	 * the tab manager which creates and manages us
	 */
	protected XTabManager	m_TabManager;

	/**
	 * a pane to contain all of our UI objects
	 */
	private JPanel				m_WorkPane;

	/**
	 * panel in which toolbar is installed (borrowed from XHomePane)
	 */
	private JPanel				m_WorkBar;

	/**
	 * our standard GridBagConstraints definition
	 */
	protected GridBagConstraints	m_GBC;

	/**
	 * the layout to use for m_WorkArea and m_ButtonPane
	 */
	private GridBagLayout	m_GBLayout;

	/**
	 * the work area, whose contents are set by our subclasses
	 */
	protected JScrollPane	m_WorkArea;

	/**
	 * a label for the screen title
	 */
	private JLabel				m_ScreenTitle;

	/**
	 * a pane to contain the screen title and icon
	 */
	private JPanel				m_ScreenLabel;

	/**
	 * the instructional text (used as a icon tooltip in m_ScreenLabel)
	 */
	protected String			m_UserInstructs;

	/**
	 * a panel in which to install action buttons (used by subclasses)
	 */
	protected JPanel			m_ButtonPane;

	/**
	 * pre-defined border object to supply insets
	 */
	protected EmptyBorder	m_InsetBorder;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the XHomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public WorkPane(WalletClient plugin, XHomePane home, XTabManager mgr) {
		super();
		m_Plugin = plugin;
		m_ParentHome = home;
		m_TabManager = mgr;
		m_WorkBar = m_ParentHome.getMenuBar();
		setLayout(new GridLayout());
		setOpaque(true);

		// establish pane and layout controls
		m_WorkPane = new JPanel();
		m_InsetBorder = new EmptyBorder(4, 8, 4, 8);
		m_WorkPane.setPreferredSize(new Dimension(M_WpWidth, 360));
		m_GBLayout = new GridBagLayout();
		m_WorkPane.setLayout(m_GBLayout);
		m_GBC = new GridBagConstraints();
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_GBC.anchor = GridBagConstraints.CENTER;
		m_GBC.fill = GridBagConstraints.NONE;

		// label containing screen title
		m_ScreenTitle = new JLabel("Default WorkPane");
		m_ScreenTitle.setFont(m_ParentHome.getWalletHome().M_HeaderFont);

		// create empty screen label panel, using blue background
		m_ScreenLabel = new JPanel();
		m_ScreenLabel.setLayout(new BoxLayout(m_ScreenLabel, BoxLayout.X_AXIS));
		m_ScreenLabel.setPreferredSize(new Dimension(M_WpWidth, 23));
		m_ScreenLabel.setBackground(
									m_ParentHome.getWalletHome().M_TitleBackgroundColor);
		EmptyBorder margin = new EmptyBorder(0, 4, 0, 4);
		m_ScreenLabel.setBorder(margin);
		m_ScreenLabel.setVisible(true);

		// the work area will be populated by subclasses via setViewPort()
		m_WorkArea = new JScrollPane();
		m_WorkArea.setPreferredSize(new Dimension(M_WpWidth, 260));
		m_WorkArea.setOpaque(true);

		// create panel in which to install screen-specific buttons
		m_ButtonPane = new JPanel();
		m_ButtonPane.setLayout(new BoxLayout(m_ButtonPane, BoxLayout.X_AXIS));
		m_ButtonPane.setPreferredSize(new Dimension(M_WpWidth, 25));

		// put it all top to bottom in the work pane
		m_GBC.gridheight = 5;
		m_WorkPane.add(m_ScreenLabel, m_GBC);
		m_WorkPane.add(Box.createVerticalStrut(3), m_GBC);
		m_WorkPane.add(m_WorkArea, m_GBC);
		m_WorkPane.add(Box.createVerticalStrut(5), m_GBC);
		m_WorkPane.add(m_ButtonPane, m_GBC);

		// everything is now in the work pane
		add(m_WorkPane);
	}

	/**
	 * method to build and display the screen label pane
	 * @param screenLabel the label text for the screen
	 * NB: implied parameters: m_UserInstructs
	 */
	protected void buildScreenLabel(	String screenLabel) {
		if (screenLabel == null) {
			Log.error("Null screen label");
			return;
		}

		m_ScreenTitle.setText(screenLabel);
		StringBuilder instTip = new StringBuilder(512);
		instTip.append("<html><table border=\"0\" width=\"300\"><tr><td>");
		instTip.append(m_UserInstructs);
		instTip.append("</td></tr></table></html>");
		m_TabManager.M_InfoIcon.setToolTipText(instTip.toString());

		JPanel part1 = new JPanel(new FlowLayout(FlowLayout.LEADING));
		part1.setOpaque(false);
		EmptyBorder eb = new EmptyBorder(0, 0, 4, 0);
		part1.setBorder(eb);

		m_ScreenLabel.removeAll();
		part1.add(m_ScreenTitle);
		part1.add(Box.createHorizontalStrut(5));
		part1.add(m_TabManager.M_InfoIcon);
		m_ScreenLabel.add(part1);
	}

	/**
	 * method to enable/disable all buttons in the m_ButtonPane
	 * @param enabled status to set
	 */
	protected void setButtonState(boolean enabled) {
		Component[] buttons = m_ButtonPane.getComponents();
		for (int iii = 0; iii < buttons.length; iii++) {
			buttons[iii].setEnabled(enabled);
		}
	}

	/**
	 * method required to implement ActionListener (overriden in sub-classes)
	 * @param ae the event from the issuer select list
	 */
	public void actionPerformed(ActionEvent ae) {
		Log.warning("Hit superclass actionPerformed(), should not happen");
	}


	// error display methods
	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 */
	protected void showError(String errMsg) {
		m_ParentHome.setCursor(null);
		m_TabManager.showError(errMsg);
	}

	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 * @param cause the cause of the error
	 */
	protected void showError(String errMsg, String cause) {
		m_ParentHome.setCursor(null);
		m_TabManager.showError(errMsg, cause);
	}

	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 * @param cause the cause of the error
	 * @param action suggested remedial action
	 */
	protected void showError(String errMsg, String cause, String action) {
		m_ParentHome.setCursor(null);
		m_TabManager.showError(errMsg, cause, action);
	}

	/**
	 * method to be overridden by subclasses, to prepare display in whatever
	 * ways might be required (for example, by making the menu bar visible or
	 * setting the contents)
	 */
	abstract void prepDisplay();

	// END methods
}
