/*
 * last modified---
 * 	01-22-15/jet/ move menu toolbar definitions here from WorkPane
 * 	08-06-14/jet/ Spark plugin version
 * 	07-16-13/jet/ add configTableColumns()
 * 	06-11-13/jet/ new
 *
 * purpose---
 * 	provide a UI panel for displaying the SVX screen tab once a connection
 * 	to a SVX has been established
 */

package org.silentvault.spark.ui.svx;

import javax.swing.*;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Font;
import java.awt.event.MouseEvent;
import java.awt.event.ActionEvent;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;
import org.silentvault.spark.ui.wallet.HomePane;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.Date;


/**
 * Extends JPanel to provide a screen for displaying the configuration of a
 * connected SVX, and presenting interface elements to examine those contents.
 * The actual UI controls and data displayed vary with the WorkPane subclass.
 * This class also manages configuration data obtained from the SVX during the
 * connection / discovery phase.
 *
 * @author Justin Turrell
 */
public final class XHomePane extends JPanel {
	// BEGIN data members
	/**
	 * tool tips for offer table columns
	 */
	protected final String[]		M_OfferTooltips =
		{	"Click on a radio button to select an offer",
			"The unique ID of the offer at the SVX",
			"The asset type offered",
			"The quantity offered",
			"The list of assets accepted in trade",
			"The ratio of Offered to Wanted demanded",
			"The status of the offer",
			"The expiration date of the offer" };

	/**
	 * the application to which we belong
	 */
	private WalletClient				m_Plugin;

	/**
	 * the tab manager which creates and manages us
	 */
	private XTabManager				m_TabManager;

	/**
	 * the current title bar text
	 */
	private String						m_TitlebarText;

	/**
	 * the work pane, which encloses an object {@link WorkPane} whose actua
	 * contents will vary based on its subclass
	 */
	private JPanel						m_WorkPane;

	/**
	 * a toolbar for the 6 standard menu buttons
	 */
	private JToolBar					m_MenuToolbar;

	/**
	 * the tool bar container for menu items (m_MenuToolbar)
	 */
	private JPanel						m_WorkBar;

	/**
	 * the complete Home pane, placed into the scroll area (m_ScrollPane)
	 */
	private JPanel						m_HomePane;

	/**
	 * the scroll area used to ensure that this pane always displays everything
	 * regardless of window size
	 */
	private JScrollPane				m_ScrollPane;

	/**
	 * the wallet tab HomePane, from which we use font and color constants,
	 * voucher and token tables, etc.
	 */
	private HomePane					m_WalletHome;

	/**
	 * the table model for the offers table, which backs m_OffersTable
	 */
	private OfferTableModel			m_OffersModel;

	/**
	 * the table which contains the offers shown in OffersPane and SearchPane
	 */
	private JTable						m_OffersTable;

	/**
	 * font for use in table columns
	 */
	public final Font					M_TableFont;

	/**
	 * action to return to the MenuPane
	 */
	private AbstractAction	m_HomeAction;

	/**
	 * action to move to the OffersPane
	 */
	private AbstractAction	m_OffersAction;

	/**
	 * action to move to the BuyPane
	 */
	private AbstractAction	m_BuyAction;

	/**
	 * action to move to the SellPane
	 */
	private AbstractAction	m_SellAction;

	/**
	 * action to move to the TradePane
	 */
	private AbstractAction	m_TradeAction;

	/**
	 * action to move to the SearchPane
	 */
	private AbstractAction	m_SearchAction;

	/**
	 * hash table of asset configuration at this SVX, keyed by asset names
	 */
	private Hashtable<String, SVXAsset>	m_SVXAssetConfig;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param client the application which owns us
	 * @param mgr the tab manager which creates us
	 */
	public XHomePane(WalletClient client, XTabManager mgr) {
		super();
		m_Plugin = client;
		m_TabManager = mgr;
		// NB: this code assumes wallet tab was initialized first
		m_WalletHome = m_Plugin.getWalletTabManager().getHomePane();
		m_SVXAssetConfig = new Hashtable<String, SVXAsset>();

		// use grid layout and set border
		setLayout(new GridLayout());

		// init font values
		M_TableFont = new Font(Font.SANS_SERIF, Font.PLAIN, 13);

		// set up offers table model (empty) and table
		m_OffersModel
			= new OfferTableModel(new ArrayList<TradeOffer>(), m_Plugin);
		m_OffersTable = mkNewOffersTable();

		// create toolbar, for containing menu button actions
		m_WorkBar = new JPanel();
		m_WorkBar.setLayout(new BoxLayout(m_WorkBar, BoxLayout.X_AXIS));
		m_WorkBar.setBorder(null);
		m_WorkBar.setVisible(false);
		m_WorkBar.setPreferredSize(new Dimension(485, 25));

		// create toolbar actions
		setHomeAction();
		setOffersAction();
		setBuyAction();
		setSellAction();
		setTradeAction();
		setSearchAction();

		// populate the toolbar
		final Font buttonFont = new Font(Font.SANS_SERIF, Font.BOLD, 13);

		m_MenuToolbar = new JToolBar(SwingConstants.HORIZONTAL);
		final Dimension sepSz = new Dimension(20, 30);
		m_MenuToolbar.setFloatable(false);
		m_MenuToolbar.setRollover(true);
		m_MenuToolbar.setBorderPainted(false);
		m_MenuToolbar.setOpaque(true);

		// button for return to MainMenu (SVX Home)
		JButton goHome = m_MenuToolbar.add(m_HomeAction);
		goHome.setToolTipText("Return to SVX Homepage");
		goHome.setFont(buttonFont);
		goHome.setRolloverEnabled(true);
		m_MenuToolbar.addSeparator(sepSz);

		// button for transfer to My Offers
		JButton offers = m_MenuToolbar.add(m_OffersAction);
		offers.setToolTipText("View your open trade offers");
		offers.setFont(buttonFont);
		offers.setRolloverEnabled(true);
		m_MenuToolbar.addSeparator(sepSz);

		// button for transfer to Buy
		JButton buy = m_MenuToolbar.add(m_BuyAction);
		buy.setToolTipText("Purchase vouchers from the SVX using an external "
								+ "currency");
		buy.setFont(buttonFont);
		buy.setRolloverEnabled(true);
		m_MenuToolbar.addSeparator(sepSz);

		// button for transfer to Sell
		JButton sell = m_MenuToolbar.add(m_SellAction);
		sell.setToolTipText("Sell vouchers to the SVX for external currency");
		sell.setFont(buttonFont);
		sell.setRolloverEnabled(true);
		m_MenuToolbar.addSeparator(sepSz);

		// button for transfer to Trade
		JButton trade = m_MenuToolbar.add(m_TradeAction);
		trade.setToolTipText("Trade between voucher asset types with other "
									+ "users");
		trade.setFont(buttonFont);
		trade.setRolloverEnabled(true);
		m_MenuToolbar.addSeparator(sepSz);

		// button for transfer to Search
		JButton search = m_MenuToolbar.add(m_SearchAction);
		search.setToolTipText("Search open user trade offers");
		search.setFont(buttonFont);
		search.setRolloverEnabled(true);
		m_MenuToolbar.add(Box.createHorizontalStrut(50));

		// install the 6-button toolbar into the work bar panel
		m_WorkBar.removeAll();
		m_WorkBar.add(Box.createHorizontalGlue());
		m_WorkBar.add(m_MenuToolbar);
		m_WorkBar.add(Box.createHorizontalGlue());
		m_WorkBar.setVisible(true);

		// create default work pane, to be filled with WorkPane subclasses
		m_WorkPane = new JPanel();
		m_WorkPane.setLayout(new GridLayout());

		// set up main scroll pane containing the objects created above
		buildHomePane();
		m_ScrollPane = new JScrollPane(m_HomePane);
		m_ScrollPane.setBorder(null);
		add(m_ScrollPane);
	}

	/**
	 * helper method to fully initialize the main home pane (which goes in the
	 * scroll pane which we actually display)
	 */
	private void buildHomePane() {
		GridBagLayout gbl = new GridBagLayout();
		m_HomePane = new JPanel(gbl);
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.weighty = 1.0;
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.gridheight = 4;
		gbc.anchor = GridBagConstraints.CENTER;
		gbc.fill = GridBagConstraints.NONE;
		m_HomePane.add(Box.createVerticalGlue(), gbc);
		m_HomePane.add(m_WorkBar, gbc);
		m_HomePane.add(m_WorkPane, gbc);
		m_HomePane.add(Box.createVerticalGlue(), gbc);
		m_HomePane.setBorder(null);
	}

	/**
	 * update the title bar text to reflect the wallet we're logged into
	 */
	public void updateTitleText() {
		m_TitlebarText = "Wallet ID: "
								+ m_Plugin.getLoginSecrets().getVSnumber() + " ";
		setBorder(BorderFactory.createTitledBorder(m_TitlebarText));
		revalidate();
	}

	// get methods
	/**
	 * method to obtain the WorkPane sub-element
	 * @return the component, which is enclosed in the JPanel
	 */
	public WorkPane getWorkPane() {
		if (m_WorkPane == null) {
			return null;
		}
		// there should be exactly one component inside this panel
		Component comp = null;
		try {
			comp = m_WorkPane.getComponent(0);
		}
		catch (ArrayIndexOutOfBoundsException aioobe) {
			Log.error("Missing WorkPane subclass in enclosing JPanel!", aioobe);
		}
		return (WorkPane) comp;
	}

	/**
	 * obtain menu toolbar
	 * @return the toolbar panel
	 */
	public JPanel getMenuBar() { return m_WorkBar; }

	/**
	 * obtain the wallet HomePane
	 * @return the home pane of the wallet tab
	 */
	public HomePane getWalletHome() { return m_WalletHome; }


	// set methods
	/**
	 * set the contents of the work pane (called by XTabManager)
	 * @param workPane a subclass object of WorkPane
	 */
	public void setWorkPane(WorkPane workPane) {
		if (workPane != null) {
			m_WorkPane.removeAll();
			m_WorkPane.add(workPane);
		}

		/* certain objects in our home pane might have been changed by the call
		 * to the subclass's prepDisplay()
		 */
		m_HomePane.revalidate();
		m_ScrollPane.setViewportView(m_HomePane);
	}

	/**
	 * reset the wallet HomePane
	 * @param walletHome the home pane in the wallet tab to use
	 */
	public void setWalletHome(HomePane walletHome) {
		if (walletHome != null) {
			m_WalletHome = walletHome;
		}
	}

	/**
	 * record the configuration info for the currently connected SVX
	 * @param svxConf the config message reply
	 */
	public void recordConfig(SVXBlock svxConf) {
		m_SVXAssetConfig.clear();
		if (svxConf == null || !svxConf.getOpcode().equals("REP_config")) {
			return;
		}

		// add all the tradeable assets
		ArrayList<SVXBlock.TradeAsset> tradeable = svxConf.getTradeAssets();
		for (SVXBlock.TradeAsset trAsset : tradeable) {
			SVXAsset asset = m_SVXAssetConfig.get(trAsset.m_AssetName);
			if (asset == null) {
				asset = new SVXAsset(trAsset.m_AssetName);
				m_SVXAssetConfig.put(trAsset.m_AssetName, asset);
			}
			asset.m_MinExchange = trAsset.m_MinTrade;
			asset.m_ExchangeCommission = trAsset.m_Commission;
		}

		// add all the saleable assets
		ArrayList<SVXBlock.SaleAsset> saleable = svxConf.getSaleAssets();
		for (SVXBlock.SaleAsset saleAsset : saleable) {
			SVXAsset asset = m_SVXAssetConfig.get(saleAsset.m_AssetName);
			if (asset == null) {
				asset = new SVXAsset(saleAsset.m_AssetName);
				m_SVXAssetConfig.put(saleAsset.m_AssetName, asset);
			}
			asset.m_MinSale = saleAsset.m_MinTrade;
			asset.m_SaleCurrencies = saleAsset.m_Currencies;
			asset.m_SaleCommission = saleAsset.m_Commission;
		}

		// add all the repurchaseable assets
		ArrayList<SVXBlock.PurchAsset> purchaseable = svxConf.getPurchAssets();
		for (SVXBlock.PurchAsset purchAsset : purchaseable) {
			SVXAsset asset = m_SVXAssetConfig.get(purchAsset.m_AssetName);
			if (asset == null) {
				asset = new SVXAsset(purchAsset.m_AssetName);
				m_SVXAssetConfig.put(purchAsset.m_AssetName, asset);
			}
			asset.m_MinPurchase = purchAsset.m_MinTrade;
			asset.m_PurchCurrencies = purchAsset.m_Currencies;
			asset.m_PurchCommission = purchAsset.m_Commission;
		}

		// tell the MenuPane about the config so controls can be updated
		XMenuPane xmp = m_TabManager.getMenuPane();
		xmp.updateControls(m_SVXAssetConfig);
	}

	/**
	 * utility method to establish the Action which returns to the MenuPane
	 */
	private void setHomeAction() {
		m_HomeAction = new AbstractAction("Home") {
			public void actionPerformed(ActionEvent ae) {
				// switch to MenuPane
				m_TabManager.queueTransition(
													new Integer(XTabManager.M_GO_MAIN_MENU));
				m_TabManager.makeTransition();
			}
		};
	}

	/**
	 * utility method to establish the Action which moves to the OffersPane
	 */
	private void setOffersAction() {
		m_OffersAction = new AbstractAction("My Offers") {
			public void actionPerformed(ActionEvent ae) {
				// switch to OffersPane
				m_TabManager.queueTransition(new Integer(XTabManager.M_GO_OFFERS));
				m_TabManager.makeTransition();
			}
		};
	}

	/**
	 * utility method to establish the Action which moves to the BuyPane
	 */
	private void setBuyAction() {
		m_BuyAction = new AbstractAction("Buy") {
			public void actionPerformed(ActionEvent ae) {
				// switch to BuyPane
				m_TabManager.queueTransition(new Integer(XTabManager.M_GO_BUY));
				m_TabManager.makeTransition();
			}
		};
	}

	/**
	 * utility method to establish the Action which moves to the SellPane
	 */
	private void setSellAction() {
		m_SellAction = new AbstractAction("Sell") {
			public void actionPerformed(ActionEvent ae) {
				// switch to SellPane
				m_TabManager.queueTransition(new Integer(XTabManager.M_GO_SELL));
				m_TabManager.makeTransition();
			}
		};
	}

	/**
	 * utility method to establish the Action which moves to the TradePane
	 */
	private void setTradeAction() {
		m_TradeAction = new AbstractAction("Trade") {
			public void actionPerformed(ActionEvent ae) {
				// switch to OffersPane
				m_TabManager.queueTransition(new Integer(XTabManager.M_GO_TRADE));
				m_TabManager.makeTransition();
			}
		};
	}

	/**
	 * utility method to establish the Action which moves to the SearchPane
	 */
	private void setSearchAction() {
		m_SearchAction = new AbstractAction("Search") {
			public void actionPerformed(ActionEvent ae) {
				// clear any existing offers from the table
				m_OffersModel.mergeOffers(null);

				// switch to OffersPane
				m_TabManager.queueTransition(new Integer(XTabManager.M_GO_SEARCH));
				m_TabManager.makeTransition();
			}
		};
	}

	// get methods
	/**
	 * obtain the asset configuration retrieved from the SVX
	 * @return hash table describing the configuration
	 */
	public Hashtable<String, SVXAsset> getSVXConfig() {
		return m_SVXAssetConfig;
	}

	/**
	 * obtain the table which displays offers
	 * @return the table
	 */
	public JTable getOfferTable() { return m_OffersTable; }

	/**
	 * obtain the data model which underlies offers table
	 * @return the model
	 */
	public OfferTableModel getOfferModel() { return m_OffersModel; }

	/**
	 * helper method to create a JTable (for offers), also providing column
	 * tooltip definitions
	 */
	private JTable mkNewOffersTable() {
		JTable newTab = new JTable(m_OffersModel) {
			// provide table header tool tips
			protected JTableHeader createDefaultTableHeader() {
				return new JTableHeader(columnModel) {
					public String getToolTipText(MouseEvent e) {
						String tip = null;
						Point ppp = e.getPoint();
						int index = columnModel.getColumnIndexAtX(ppp.x);
						int realIndex = columnModel.getColumn(index).getModelIndex();
						return M_OfferTooltips[realIndex];
					}
				};
			}
			public boolean isCellEditable(int row, int col) {
				// objects in the select column can be editable, others are not
				if (col == 0) {
					if (m_OffersModel.getSelectMode()
						 != OfferTableModel.M_SelectNone)
					{
						return true;
					}
				}
				return false;
			}
		};
		newTab.setCellSelectionEnabled(false);
		// this size should be same as WorkPane.m_WorkArea
		newTab.setPreferredScrollableViewportSize(new Dimension(485, 260));
		newTab.setFillsViewportHeight(true);

		// set special renderer for the all columns
		OfferRenderer or = new OfferRenderer(m_OffersModel, m_Plugin);
		newTab.setDefaultRenderer(JCheckBox.class, or);
		newTab.setDefaultRenderer(JRadioButton.class, or);
		newTab.setDefaultRenderer(Integer.class, or);
		newTab.setDefaultRenderer(Float.class, or);
		newTab.setDefaultRenderer(Double.class, or);
		newTab.setDefaultRenderer(Date.class, or);
		newTab.setDefaultRenderer(String.class, or);

		// configure the table columns and renderers
		configTableColumns(newTab);

		// set default sorter
		newTab.setAutoCreateRowSorter(true);
		newTab.setAutoCreateColumnsFromModel(true);

		// set spacing preferences
		newTab.setRowHeight(20);
		newTab.setIntercellSpacing(new Dimension(2, 2));
		newTab.setShowGrid(true);
		newTab.setGridColor(UIManager.getColor("TableHeader.background"));

		// establish table as listener for model
		m_OffersModel.addTableModelListener(newTab);
		return newTab;
	}

	/**
	 * Helper method to configure the columns and renderers of the
	 * offers table.  This needs to be called when the table is created,
	 * and after each change to the table.
	 * @param tab the table to configure (== m_OffersTable)
	 */
	private void configTableColumns(JTable tab) {
		// set special renderer for column headers, to cause right justify
		OffColRenderer ocr = new OffColRenderer();

		// set max width on select column, and set cell editor
		TableColumnModel tcm = tab.getColumnModel();
		TableColumn tc = tcm.getColumn(0);
		tc.setMaxWidth(35);
		OfferSelector os = new OfferSelector(m_Plugin, m_OffersModel);
		tc.setCellEditor(os);
		tc.setHeaderRenderer(ocr);

		// offer ID
		tc = tcm.getColumn(1);
		tc.setHeaderRenderer(ocr);
		tc.setMaxWidth(35);

		// offered asset
		tc = tcm.getColumn(2);
		tc.setHeaderRenderer(ocr);

		// make sure amount column is wide enough to show full amount
		tc = tcm.getColumn(3);
		tc.setMinWidth(70);
		tc.setHeaderRenderer(ocr);

		// wanted asset(s)
		tc = tcm.getColumn(4);
		tc.setHeaderRenderer(ocr);

		// price ratio
		tc = tcm.getColumn(5);
		tc.setMaxWidth(55);
		tc.setHeaderRenderer(ocr);

		// offer status
		tc = tcm.getColumn(6);
		tc.setHeaderRenderer(ocr);

		// make sure expiry column is as wide as we can reasonably get it
		tc = tcm.getColumn(7);
		tc.setMinWidth(125);
		tc.setHeaderRenderer(ocr);
	}

	// END methods
}
