/*
 * last modified---
 * 	08-06-14/jet/ Spark plugin version
 * 	06-14-13/jet/ new
 *
 * purpose---
 * 	provide a UI screen to represent the SVX main menu after connection
 */

package org.silentvault.spark.ui.svx;

import javax.swing.*;
import java.awt.event.*;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Font;
import java.awt.Component;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;
import org.silentvault.spark.ui.wallet.HomePane;

import java.util.Hashtable;
import java.util.Set;
import java.util.Vector;
import java.util.Locale;
import java.text.NumberFormat;
import java.text.DecimalFormat;


/**
 * Extends WorkPane to provide a screen for the SVX tab home page.
 *
 * @author Justin Turrell
 */
public final class XMenuPane extends WorkPane {
	// BEGIN data members
	/**
	 * constant indicating no assets were defined for this category
	 */
	private final String				M_None = "--None--";

	/**
	 * action to disconnect from this SVX
	 */
	private AbstractAction			m_DisconnectAction;
	
	/**
	 * button to go with disconnect action
	 */
	private JButton					m_DisconnButton;

	/**
	 * drop-down list containing names of assets sold at this SVX
	 */
	private JComboBox<String>		m_SoldList;

	/**
	 * dynamic label representing currencies accepted for sell orders
	 */
	private JLabel						m_SellPmtLabel;

	/**
	 * dynamic label representing commission rate for sell orders
	 */
	private JLabel						m_SellCommLabel;

	/**
	 * drop-down list containing names of purchased by this SVX
	 */
	private JComboBox<String>		m_BoughtList;

	/**
	 * dynamic label representing currencies offered on purchase orders
	 */
	private JLabel						m_PurchPmtLabel;

	/**
	 * dynamic label representing commission rate for purchase orders
	 */
	private JLabel						m_PurchCommLabel;

	/**
	 * drop-down list containing names of assets traded at this SVX
	 */
	private JComboBox<String>		m_TradedList;

	/**
	 * dynamic label representing commission rate for exchange orders
	 */
	private JLabel						m_TradeCommLabel;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the XHomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public XMenuPane(WalletClient plugin, XHomePane home, XTabManager mgr) {
		super(plugin, home, mgr);

		// define disconnection action
		m_DisconnectAction = new AbstractAction("Disconnect") {
			public void actionPerformed(ActionEvent ae) {
				// switch back to connect screen via tab manager
				m_TabManager.queueTransition(
												new Integer(XTabManager.M_LOGOUT_EXCHANGE));
				m_TabManager.makeTransition();
			}
		};
		// define matching button
		m_DisconnButton = new JButton(m_DisconnectAction);
		m_DisconnButton.setFont(m_TabManager.M_ButtonFont);
		m_ButtonPane.removeAll();
		m_ButtonPane.add(Box.createHorizontalStrut(195));		// spacer
		m_ButtonPane.add(m_DisconnButton);

		// get the SVX's asset configuration
		Hashtable<String, SVXAsset> svxConfig = m_ParentHome.getSVXConfig();
		Set<String> assets = svxConfig.keySet();
		Vector<String> soldHere = new Vector<String>(assets.size());
		Vector<String> purchHere = new Vector<String>(assets.size());
		Vector<String> tradedHere = new Vector<String>(assets.size());
		for (String asset : assets) {
			SVXAsset aConf = svxConfig.get(asset);
			if (aConf.m_MinExchange > 0.0) {
				tradedHere.add(asset);
			}
			if (aConf.m_MinSale > 0.0) {
				soldHere.add(asset);
			}
			if (aConf.m_MinPurchase > 0.0) {
				purchHere.add(asset);
			}
		}
		if (soldHere.isEmpty()) {
			soldHere.add(M_None);
		}
		if (tradedHere.isEmpty()) {
			tradedHere.add(M_None);
		}
		if (purchHere.isEmpty()) {
			purchHere.add(M_None);
		}
		// define asset selection drop-downs
		Dimension boxSz = new Dimension(50, 25);
		m_SoldList = new JComboBox<String>(soldHere);
		m_SoldList.addActionListener(this);
		m_SoldList.setActionCommand("sold");
		m_SoldList.setPreferredSize(boxSz);
		m_BoughtList = new JComboBox<String>(purchHere);
		m_BoughtList.addActionListener(this);
		m_BoughtList.setActionCommand("bought");
		m_BoughtList.setPreferredSize(boxSz);
		m_TradedList = new JComboBox<String>(tradedHere);
		m_TradedList.addActionListener(this);
		m_TradedList.setActionCommand("traded");
		m_TradedList.setPreferredSize(boxSz);

		// define some labels
		final Color labColor = UIManager.getColor("standardGray");
		m_SellPmtLabel = new JLabel();
		HomePane whp = m_ParentHome.getWalletHome();
		m_SellPmtLabel.setFont(whp.M_LabelFont);
		m_SellPmtLabel.setForeground(labColor);
		m_SellPmtLabel.setHorizontalAlignment(SwingConstants.CENTER);
		m_SellPmtLabel.setToolTipText("External currencies accepted as payment");
		m_SellCommLabel = new JLabel();
		m_SellCommLabel.setFont(whp.M_LabelFont);
		m_SellCommLabel.setForeground(labColor);
		m_SellCommLabel.setHorizontalAlignment(SwingConstants.CENTER);
		m_SellCommLabel.setToolTipText("Sale commission charged by exchange");
		m_PurchPmtLabel = new JLabel();
		m_PurchPmtLabel.setFont(whp.M_LabelFont);
		m_PurchPmtLabel.setForeground(labColor);
		m_PurchPmtLabel.setHorizontalAlignment(SwingConstants.CENTER);
		m_PurchPmtLabel.setToolTipText("External currencies offered as payment");
		m_PurchCommLabel = new JLabel();
		m_PurchCommLabel.setFont(whp.M_LabelFont);
		m_PurchCommLabel.setForeground(labColor);
		m_PurchCommLabel.setHorizontalAlignment(SwingConstants.CENTER);
		m_PurchCommLabel.setToolTipText(
											"Repurchase commission charged by exchange");
		m_TradeCommLabel = new JLabel();
		m_TradeCommLabel.setFont(whp.M_LabelFont);
		m_TradeCommLabel.setForeground(labColor);
		m_TradeCommLabel.setHorizontalAlignment(SwingConstants.CENTER);
		m_TradeCommLabel.setToolTipText("Escrow fee charged by exchange");

		// set the user instructions
		m_UserInstructs = "Select assets and currencies from the drop-down lists "
								+ "to view operations and fees supported at this SVX."
								+ "<br/><br/>To begin an action, click on the rollover "
								+ "menu above.";
		m_WorkArea.setBorder(null);
	}

	// extend WorkPane
	/**
	 * generate display based on selections in the drop-down lists
	 */
	public void prepDisplay() {
		// set screen label
		String svxId = m_Plugin.getLoginSecrets().getSVXId();
		buildScreenLabel("Silent-Vault Exchange (SVX): " + svxId);

		m_DisconnButton.setToolTipText("<html>Click here to disconnect from<br/>"
												+ m_Plugin.getLoginSecrets().getSVXId()
												+ "</html>");

		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		df.setMinimumFractionDigits(2);
		df.setMaximumFractionDigits(4);
		df.setGroupingSize(0);

		// prepare a screen showing buy/sell/trade information in a quasi-table
		Hashtable<String, SVXAsset> svxConfig = m_ParentHome.getSVXConfig();
		JPanel confPane = new JPanel(new GridBagLayout());
		confPane.setBorder(null);
		m_GBC.fill = GridBagConstraints.NONE;
		m_GBC.anchor = GridBagConstraints.NORTH;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_GBC.gridheight = 8;
		confPane.add(Box.createVerticalGlue(), m_GBC);
		final Color labColor = UIManager.getColor("standardGray");
		HomePane whp = m_ParentHome.getWalletHome();
		JLabel headLab = new JLabel("Supported Assets and Rates on This Exchange",
											SwingConstants.CENTER);
		headLab.setFont(whp.M_LabelFont);
		headLab.setForeground(labColor);
		confPane.add(headLab, m_GBC);
		confPane.add(Box.createVerticalStrut(25), m_GBC);

		// sells row
		JPanel sellPane = new JPanel(new GridLayout(1, 4));
		sellPane.setBorder(null);
		final Dimension rowSize = new Dimension(450, 25);
		sellPane.setPreferredSize(rowSize);
		// Sells: label
		JLabel sellLab = new JLabel("Sells:", SwingConstants.CENTER);
		sellLab.setFont(whp.M_LabelFont);
		sellLab.setForeground(labColor);
		sellLab.setLabelFor(m_SoldList);
		sellPane.add(sellLab);
		// drop-down of all voucher assets sold at this SVX
		sellPane.add(m_SoldList);
		// list of external currencies accepted as payment
		String pmtCurrs = "";
		SVXAsset aConf = svxConfig.get((String)m_SoldList.getSelectedItem());
		float comm = 0.0F;
		if (aConf != null) {
			pmtCurrs = aConf.m_SaleCurrencies;
			comm = aConf.m_SaleCommission;
		}
		m_SellPmtLabel.setText(" for " + pmtCurrs);
		sellPane.add(m_SellPmtLabel);
		// label showing commission rate for current drop-down selection
		m_SellCommLabel.setText(" fee: " + df.format(comm) + " %");
		sellPane.add(m_SellCommLabel);
		confPane.add(sellPane, m_GBC);
		confPane.add(Box.createVerticalStrut(25), m_GBC);

		// buys row
		JPanel buyPane = new JPanel(new GridLayout(1, 4));
		buyPane.setBorder(null);
		buyPane.setPreferredSize(rowSize);
		// Buys: label
		JLabel buyLab = new JLabel("Buys:", SwingConstants.CENTER);
		buyLab.setFont(whp.M_LabelFont);
		buyLab.setForeground(labColor);
		buyLab.setLabelFor(m_BoughtList);
		buyPane.add(buyLab);
		// drop-down of all voucher assets repurchased by this SVX
		buyPane.add(m_BoughtList);
		// list of external currencies offered as payment
		aConf = svxConfig.get((String)m_BoughtList.getSelectedItem());
		String purchCurrs = "";
		comm = 0.0F;
		if (aConf != null) {
			purchCurrs = aConf.m_PurchCurrencies;
			comm = aConf.m_PurchCommission;
		}
		m_PurchPmtLabel.setText(" with " + purchCurrs);
		buyPane.add(m_PurchPmtLabel);
		// label showing commission rate for current drop-down selection
		m_PurchCommLabel.setText(" fee: " + df.format(comm) + " %");
		buyPane.add(m_PurchCommLabel);
		confPane.add(buyPane, m_GBC);
		confPane.add(Box.createVerticalStrut(25), m_GBC);

		// trades row
		JPanel tradePane = new JPanel(new GridLayout(1, 4));
		tradePane.setBorder(null);
		tradePane.setPreferredSize(rowSize);
		// Trades: label
		JLabel tradeLab = new JLabel("Trades:", SwingConstants.CENTER);
		tradeLab.setFont(whp.M_LabelFont);
		tradeLab.setForeground(labColor);
		tradeLab.setLabelFor(m_TradedList);
		tradePane.add(tradeLab);
		// drop-down of all voucher assets traded at this SVX
		tradePane.add(m_TradedList);
		tradePane.add(Box.createHorizontalStrut(50));
		aConf = svxConfig.get((String)m_TradedList.getSelectedItem());
		comm = 0.0F;
		if (aConf != null) {
			comm = aConf.m_ExchangeCommission;
		}
		// label showing commission rate for current drop-down selection
		m_TradeCommLabel.setText(" fee: " + df.format(comm) + " %");
		tradePane.add(m_TradeCommLabel);
		confPane.add(tradePane, m_GBC);
		confPane.add(Box.createVerticalGlue(), m_GBC);
		
		// put entire quasi-table into the main scroll area
		m_WorkArea.setViewportView(confPane);
	}

	/**
	 * obtain the currently selected sell asset
	 * @return the asset selected from the m_SoldList drop-down
	 */
	public String getSellAsset() {
		String asset = (String) m_SoldList.getSelectedItem();
		if (asset == null || asset.equals(M_None) || asset.isEmpty()) {
			return "";
		}
		return asset;
	}

	/**
	 * obtain the currently selected repurchase asset
	 * @return the asset selected from the m_BoughtList drop-down
	 */
	public String getBuyAsset() {
		String asset = (String) m_BoughtList.getSelectedItem();
		if (asset == null || asset.equals(M_None) || asset.isEmpty()) {
			return "";
		}
		return asset;
	}

	/**
	 * obtain the currently selected trade asset
	 * @return the asset selected from the m_TradedList drop-down
	 */
	public String getTradeAsset() {
		String asset = (String) m_TradedList.getSelectedItem();
		if (asset == null || asset.equals(M_None) || asset.isEmpty()) {
			return "";
		}
		return asset;
	}

	/**
	 * update the controls based on the recorded config (called by XHomePane)
	 * @param config the new config supplied by caller
	 */
	public void updateControls(Hashtable<String, SVXAsset> svxConfig) {
		Vector<String> soldHere = new Vector<String>(svxConfig.size());
		Vector<String> purchHere = new Vector<String>(svxConfig.size());
		Vector<String> tradedHere = new Vector<String>(svxConfig.size());
		for (SVXAsset aConf : svxConfig.values()) {
			if (aConf.m_MinExchange > 0.0) {
				tradedHere.add(aConf.m_AssetName);
			}
			if (aConf.m_MinSale > 0.0) {
				soldHere.add(aConf.m_AssetName);
			}
			if (aConf.m_MinPurchase > 0.0) {
				purchHere.add(aConf.m_AssetName);
			}
		}
		if (soldHere.isEmpty()) {
			soldHere.add(M_None);
		}
		if (tradedHere.isEmpty()) {
			tradedHere.add(M_None);
		}
		if (purchHere.isEmpty()) {
			purchHere.add(M_None);
		}

		// define asset selection drop-downs
		Dimension boxSz = new Dimension(50, 25);
		m_SoldList = new JComboBox<String>(soldHere);
		m_SoldList.addActionListener(this);
		m_SoldList.setActionCommand("sold");
		m_SoldList.setPreferredSize(boxSz);
		m_SoldList.setToolTipText("Asset types sold by this exchange");
		m_BoughtList = new JComboBox<String>(purchHere);
		m_BoughtList.addActionListener(this);
		m_BoughtList.setActionCommand("bought");
		m_BoughtList.setPreferredSize(boxSz);
		m_BoughtList.setToolTipText("Asset types repurchased by this exchange");
		m_TradedList = new JComboBox<String>(tradedHere);
		m_TradedList.addActionListener(this);
		m_TradedList.setActionCommand("traded");
		m_TradedList.setPreferredSize(boxSz);
		m_TradedList.setToolTipText("Asset types traded on this exchange");
	}

	/**
	 * 'push' the disconnect button programmatically
	 */
	public void pushDisconnect() {
		if (m_DisconnButton != null) {
			m_DisconnButton.doClick();
		}
	}

	// implement ActionListener
	/**
	 * method to process the event when an asset is selected in a drop-down
	 * @param ae the event from one of the asset selection JComboBoxS
	 */
	@Override
	public void actionPerformed(ActionEvent ae) {
		// obtain the source (will be a JComboBox)
		Object src = ae.getSource();
		if (!(src instanceof JComboBox)) {
			Log.error("Non drop-down action source");
			return;
		}
		JComboBox selector = (JComboBox) src;

		// get the currently selected item and asset info
		String asset = (String) selector.getSelectedItem();
		if (asset.equals(M_None) || asset.isEmpty()) {
			return;
		}
		SVXAsset aConf = m_ParentHome.getSVXConfig().get(asset);
		if (aConf == null) {
			Log.error("No such configured asset, " + asset);
			return;
		}
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		df.setMinimumFractionDigits(2);
		df.setMaximumFractionDigits(4);
		df.setGroupingSize(0);

		// get the command associated with this action and update label(s)
		String cmd = ae.getActionCommand();
		if (cmd.equals("sold")) {
			m_SellPmtLabel.setText(" for " + aConf.m_SaleCurrencies);
			m_SellCommLabel.setText(" fee: "
											+ df.format(aConf.m_SaleCommission) + " %");
		}
		else if (cmd.equals("bought")) {
			m_PurchPmtLabel.setText(" with " + aConf.m_PurchCurrencies);
			m_PurchCommLabel.setText(" fee: "
											+ df.format(aConf.m_PurchCommission) + " %");
		}
		else if (cmd.equals("traded")) {
			m_TradeCommLabel.setText(" fee: "
										+ df.format(aConf.m_ExchangeCommission) + " %");
		}
		else {
			Log.error("Unrecognized action command, " + cmd);
		}
		revalidate();
		repaint();
	}

	// END methods
}
