/*
 * last modified---
 * 	05-17-18/kbw/ use SV tutorials, as voucher-safe.com is gone
 * 	08-04-14/jet/ version for Spark plugin
 * 	06-05-13/jet/ new
 *
 *	purpose---
 *		provide a manager for the contents of the exchange tab
 */

package org.silentvault.spark.ui.svx;

import javax.swing.*;
import java.awt.Cursor;
import java.awt.Color;
import java.awt.Font;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Desktop;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.net.URI;
import java.net.URISyntaxException;
import java.io.IOException;


/**
 * This class is used to manage the graphical contents of the exchanges tab.
 * It handles creating and destroying
 * the embedded pane subclasses and manages state transitions between them.
 * It also provides utility functions such as error popups and wait cursors.
 *
 * @author Justin Turrell
 */
public class XTabManager extends TabManager {
	// BEGIN data members
	// possible states we can do
	/**
	 * select a SVX to connect to
	 */
	public static final int	M_CONNECT = 1;
	/**
	 * move to home state, following connect and discovery
	 */
	public static final int	M_HOME_INIT = 2;
	/**
	 * return to the main menu screen
	 */
	public static final int	M_GO_MAIN_MENU = 3;
	/**
	 * display tutorials in a popup window
	 */
	public static final int	M_TUTORIALS = 4;
	/**
	 * log out of this exchange
	 */
	public static final int	M_LOGOUT_EXCHANGE = 5;
	/**
	 * move to MyOffers screen (OffersPane)
	 */
	public static final int	M_GO_OFFERS = 6;
	/**
	 * move to the Buy screen (BuyPane)
	 */
	public static final int	M_GO_BUY = 7;
	/**
	 * move to the Buy screen (SellPane)
	 */
	public static final int	M_GO_SELL = 8;
	/**
	 * move to the Trade screen (TradePane)
	 */
	public static final int	M_GO_TRADE = 9;
	/**
	 * move to the Search screen (SearchPane)
	 */
	public static final int	M_GO_SEARCH = 10;
	/**
	 * move to the order Accept screen (AcceptPane)
	 */
	public static final int M_GO_ACCEPT = 11;

	/**
	 * standard font for buttons in this tab
	 */
	public static final Font	M_ButtonFont = new Font(Font.SANS_SERIF,
																		Font.BOLD, 12);

	// display panes we control
	/**
	 * the connection pane, used to pick a SVX to connect
	 */
	private ConnectPane		m_ConnectPane;

	/**
	 * the Home pane, used after connect has occurred
	 */
	private XHomePane			m_HomePane;

	/**
	 * the MainMenu pane, to show disco info retrieved from the SVX
	 */
	private XMenuPane			m_MenuPane;

	/**
	 * the OffersPane, to show extant offers in table format
	 */
	private OffersPane		m_OffersPane;

	/**
	 * the SearchPane, to select and accept offers
	 */
	private SearchPane		m_SearchPane;
	
	/**
	 * the TradePane, to enter and fund trade offers
	 */
	private TradePane			m_TradePane;

	/**
	 * the AcceptPane, to accept another user's funded trade offer
	 */
	private AcceptPane		m_AcceptPane;

	/**
	 * the BuyPane, to purchase vouchers with external currency payments
	 */
	private BuyPane			m_BuyPane;

	/**
	 * the SellPane, to repurchase vouchers for external currency payments
	 */
	private SellPane			m_SellPane;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the containing plugin
	 */
	public XTabManager(WalletClient plugin) {
		super(plugin);
		m_ConnectPane = new ConnectPane(m_Plugin, this);
		m_HomePane = new XHomePane(m_Plugin, this);
		m_MenuPane = new XMenuPane(m_Plugin, m_HomePane, this);
		m_HomePane.setWorkPane(m_MenuPane);
	}

	/**
	 * state transition method, called by threads that remove states from queue
	 * @param state the state to transition to
	 */
	public void doState(int state) {
		//Log.debug("Hit transition to state " + state);
		switch (state) {
			case M_CONNECT:
				removeAll();
				add(m_ConnectPane);
				m_ConnectPane.prepDisplay();
				break;

			case M_HOME_INIT:
				// replace ConnectPane with HomePane
	 			removeAll();
				add(m_HomePane);

				// now prepare all the WorkPane subclass objects
				m_OffersPane = new OffersPane(m_Plugin, m_HomePane, this);
				m_SearchPane = new SearchPane(m_Plugin, m_HomePane, this);
				m_TradePane = new TradePane(m_Plugin, m_HomePane, this);
				m_AcceptPane = new AcceptPane(m_Plugin, m_HomePane, this);
				m_BuyPane = new BuyPane(m_Plugin, m_HomePane, this);
				m_SellPane = new SellPane(m_Plugin, m_HomePane, this);

				m_HomePane.setWorkPane(m_MenuPane);
				m_HomePane.updateTitleText();
				m_MenuPane.prepDisplay();
				break;

			case M_GO_MAIN_MENU:
				m_HomePane.setWorkPane(m_MenuPane);
				m_MenuPane.prepDisplay();
				break;

			case M_TUTORIALS:
				// get the desktop browser for this system and try to open the URI
				URI tutURI = null;
				boolean needManual = false;
				try {
					// TBD: SVX-related tutorials
					tutURI = new URI("https://silentvault.com");
				}
				catch (URISyntaxException use) {
					Log.error("Bad URI for tutorials");
					needManual = true;
				}
				Desktop browser = null;
				try {
					browser = Desktop.getDesktop();
				}
				catch (UnsupportedOperationException uoe) {
					needManual = true;
				}
				finally {
					if (!needManual) {
						if (!browser.isSupported(Desktop.Action.BROWSE)) {
							needManual = true;
						}
					}
				}
				if (needManual) {
					// just put up the URL in a popup window with instructions
					JFrame tutPop = new JFrame("SVX Tutorials");
					tutPop.setDefaultCloseOperation(
															WindowConstants.DISPOSE_ON_CLOSE);
					JEditorPane tutViewer = new JEditorPane("text/html", null);
					tutViewer.setBackground(Color.LIGHT_GRAY);
					StringBuilder content = new StringBuilder(512);
					content.append("<html><body bgcolor=\"silver\">");
					content.append("<center><br/><br/><p>");
					content.append("For tutorials help, paste the following URL ");
					content.append("into your browser:</p><br/><br/>");
					content.append("https://silentvault.com<br/>");
					content.append("</center></body></html>");
					tutViewer.setText(content.toString());
					tutPop.add(tutViewer);
					tutPop.setSize(new Dimension(400, 200));
					tutPop.setLocationRelativeTo(this);
					tutPop.setVisible(true);
					m_Plugin.registerPopup(tutPop);
				}
				else {
					try {
						// actually invoke the default browser
						browser.browse(tutURI);
					}
					catch (IOException ioe) {
						Log.error("Could not invoke browser on tutorials page", ioe);
						showError("Could not display tutorials",
									"Error opening default browser",
									"Point your browser to: https://silentvault.com");
					}
				}
				break;

			case M_LOGOUT_EXCHANGE:
				// user was auto-logged out, show message screen leading to login
				removeAll();
				add(m_ConnectPane);
				m_HomePane.recordConfig(null);
				m_ConnectPane.prepDisplay();
				break;

			case M_GO_OFFERS:
				// query SVX for wallet offers and display them
				m_HomePane.setWorkPane(m_OffersPane);
				m_OffersPane.queryOffers();
				m_OffersPane.prepDisplay();
				break;

			case M_GO_SEARCH:
				// query SVX for matching offers and display them
				m_HomePane.setWorkPane(m_SearchPane);
				m_SearchPane.prepDisplay();
				break;

			case M_GO_TRADE:
				// enter and fund trade offers
				m_HomePane.setWorkPane(m_TradePane);
				m_TradePane.prepDisplay();
				break;

			case M_GO_ACCEPT:
				// accept an existing offer from another user
				m_HomePane.setWorkPane(m_AcceptPane);
				m_AcceptPane.prepDisplay();
				break;

			case M_GO_BUY:
				// sell the user vouchers for external currency payments
				m_HomePane.setWorkPane(m_BuyPane);
				m_BuyPane.prepDisplay();
				break;

			case M_GO_SELL:
				// buy back the user vouchers with external currency payments
				m_HomePane.setWorkPane(m_SellPane);
				m_SellPane.prepDisplay();
				break;

			default:
				Log.error("Unsupported SVM UI state " + state);
				break;
		}
		if (state != M_TUTORIALS) {
			m_State = state;
			revalidate();
			repaint();
		}
	}


	// get methods
	/**
	 * obtain the connect pane
	 * @return the pane used for connecting to a SVX
	 */
	public ConnectPane getConnectPane() { return m_ConnectPane; }

	/**
	 * obtain the home pane
	 * @return the pane used for home safe display
	 */
	public XHomePane getHomePane() { return m_HomePane; }

	/**
	 * obtain the menu pane
	 * @return the pane used for main menu display
	 */
	public XMenuPane getMenuPane() { return m_MenuPane; }

	/**
	 * obtain the offers pane
	 * @return the pane used for outstanding offers display
	 */
	public OffersPane getOffersPane() { return m_OffersPane; }

	/**
	 * obtain the search pane
	 * @return the pane used for searching and displaying existing offers
	 */
	public SearchPane getSearchPane() { return m_SearchPane; }

	/**
	 * obtain the trade pane
	 * @return the pane used for entering and funding new trade offers
	 */
	public TradePane getTradePane() { return m_TradePane; }

	/**
	 * obtain the accept pane
	 * @return the pane used for accepting an existing trade offer
	 */
	public AcceptPane getAcceptPane() { return m_AcceptPane; }

	/**
	 * obtain the buy pane
	 * @return the pane used for user purchases of vouchers
	 */
	public BuyPane getBuyPane() { return m_BuyPane; }

	/**
	 * obtain the sell pane
	 * @return the pane used for user sales of vouchers
	 */
	public SellPane getSellPane() { return m_SellPane; }

	/**
	 * obtain the work pane (a component in the home pane which contains the
	 * WorkPane subclass)
	 * @return the pane used for work area display
	 */
	public WorkPane getWorkPane() {
		if (m_HomePane == null) {
			return null;
		}
		return m_HomePane.getWorkPane();
	}

	/**
	 * method to deallocate all sub-panes (called by WalletClient destructor)
	 */
	public void deallocatePanes() {
		if (m_ConnectPane != null) {
			m_ConnectPane.removeAll();
			m_ConnectPane = null;
		}
		if (m_HomePane != null) {
			m_HomePane.removeAll();
			m_HomePane = null;
		}
		if (m_MenuPane != null) {
			m_MenuPane.removeAll();
			m_MenuPane = null;
		}
		if (m_OffersPane != null) {
			m_OffersPane.removeAll();
			m_OffersPane = null;
		}
		if (m_SearchPane != null) {
			m_SearchPane.removeAll();
			m_SearchPane = null;
		}
		if (m_TradePane != null) {
			m_TradePane.removeAll();
			m_TradePane = null;
		}
		if (m_AcceptPane != null) {
			m_AcceptPane.removeAll();
			m_AcceptPane = null;
		}
		if (m_BuyPane != null) {
			m_BuyPane.removeAll();
			m_BuyPane = null;
		}
		if (m_SellPane != null) {
			m_SellPane.removeAll();
			m_SellPane = null;
		}
	}

	// END methods
}
