/*
 * last modified---
 * 	08-15-18/kbw/ explicitly cancel PacketCollectors after use
 * 	01-29-18/kbw/ use getBasePriceLookupUnits()
 * 	07-15-15/jet/ use VPAssetConfig.getPriceLookupUnits() to determine values
 * 	08-05-14/jet/ Spark plugin version
 * 	09-11-13/jet/ add m_DeferPopup and setDeferPopup()
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide a screen to handle incoming payment confirmations
 */

package org.silentvault.spark.ui.wallet;

import javax.swing.*;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.GridLayout;
import java.awt.Color;
import java.awt.Font;
import java.awt.Dimension;
import java.awt.event.ActionEvent;

import org.dom4j.Element;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Date;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.security.PublicKey;
import java.net.HttpURLConnection;


/**
 * Extends WorkPane to provide a screen for handling incoming payment
 * confirmations.  These are placed on the DHT by the VP following a successful
 * payment.
 *
 * @author Kevin Wilkerson
 */
public final class ConfirmPane extends WorkPane {
	// BEGIN data members
	/**
	 * action for submitting confirmation
	 */
	private AbstractAction			m_SubmitAction;

	/**
	 * storage for tokens used in an operation storing something on the SDS
	 */
	private ArrayList<XMLToken>   m_PaymentTokens;

	/**
	 * storage for a pending payment record we are placing on the SDS
	 */
	private XMLPayment				m_PaymentRecord;

	/**
	 * storage for a payment confirmation record received from the VP via DHT
	 */
	private XMLConfirmPayment		m_PaymentConf;

	/**
	 * whether we override default behavior and show screen to user.
	 * (used for login processing as opposed to ordinary housekeeping)
	 */
	private boolean					m_ShowDisplay;

	/**
	 * whether we defer the success popup message, and instead record the
	 * confirmation in the PaymentPane for later retrieval
	 */
	private boolean					m_DeferPopup;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the HomePane in which we're installed
	 * @param mgr the tab manager which created us
	 */
	public ConfirmPane(WalletClient plugin, HomePane home, WTabManager mgr) {
		super(plugin, home, mgr);
		m_PaymentTokens = new ArrayList<XMLToken>();

		// set user instructions
		String instTxt = "Your payment was made successfully and you will "
							+ "receive a receipt from the payee once the payment "
							+ "has been picked up.<br/><br/>"
							+ "To complete the payment a pending payment record is "
							+ "stored in your wallet.  This is used to reclaim the "
							+ "payment if it's not picked up in time.";
		m_UserInstructs = instTxt;

		// set token cost
		m_TokenCost.setText("Complete costs 1 token");

		// set action for submitting and storing pending payment record
		setSubmitAction();

		// we don't want a border around the whole thing
		m_WorkArea.setBorder(null);
	}

	/**
	 * method to establish the action which submits and stores a pending payment
	 */
	private void setSubmitAction() {
		m_SubmitAction = new AbstractAction("Complete Payment") {
			public void actionPerformed(ActionEvent actionEvent) {
				/* Perform these steps in order:
				 * 	1. Validate VP's signature on the confirmation record.
				 * 	2. Generate a XMLPayment object and store it on SDS.
				 * 	3. On receipt of reply, mark the DHT record read.
				 * 	4. Update display (also do this on timeout/error).
				 */
				// step 1
				if (m_PaymentConf == null) {
					Log.error("Missing conf proc record");
					showError("cannot process payment confirmation",
								"internal error, no confirm record exists");
					return;
				}

				// validate signature of VP
				VsSecrets sekrits = m_Plugin.getLoginSecrets();
				if (!m_PaymentConf.validateSig(sekrits.getVPKey())) {
					Log.error("Bad signature on payment confirmation object");
					showError("cannot process payment confirmation",
								"publisher signature did not verify",
								"please report to customer service");
					return;
				}

				// step 2
				m_PaymentRecord = new XMLPayment();
				m_PaymentRecord.setQuantity(m_PaymentConf.getQuantity());
				m_PaymentRecord.setUnits(m_PaymentConf.getUnits());
				m_PaymentRecord.setAsset(m_PaymentConf.getAsset());
				m_PaymentRecord.setSerialNumber(m_PaymentConf.getVoucherSerial());
				m_PaymentRecord.setPayee(m_PaymentConf.getPayee());
				m_PaymentRecord.setIssuer(m_PaymentConf.getIssuer());
				m_PaymentRecord.setTTL(m_PaymentConf.getTTL());
				m_PaymentRecord.setInit();

				// obtain any baggage which we might have recorded from a pmt
				PaymentPane pp = m_TabManager.getPaymentPane();
				Hashtable<String, String> bagHash = pp.getPaymentBaggage();
				String bagXML = bagHash.get(m_PaymentConf.getPayee());
				if (bagXML != null && !bagXML.isEmpty()) {
					Element bagElt = EncodingUtils.buildElement(bagXML);
					if (bagElt == null) {
						Log.error("Stored payment baggage did not parse: " + bagXML);
						showError("problem processing payment confirmation",
									"parse error in baggage fields",
									"item will be shown without full baggage info");
					}
					else {
						Iterator fldIt = bagElt.elementIterator();
						while (fldIt.hasNext()) {
							Element fldElt = (Element) fldIt.next();
							String eltNm = fldElt.getName();
							if (!eltNm.equalsIgnoreCase("field_names")
								 && !eltNm.equalsIgnoreCase("signature"))
							{
								m_PaymentRecord.addBaggageField(eltNm,
																		  fldElt.getTextTrim());
							}
						}
					}
					// intentional side-effect back to payment pane
					bagHash.remove(m_PaymentConf.getPayee());
				}

				// need one token for payment
				TokenTableModel ttm = m_ParentHome.getTokenModel();
				VoucherTableModel vtm = m_ParentHome.getVoucherModel();
				m_PaymentTokens
					= ttm.getAvailableTokens(VoucherRequest.M_ReceiptCost,
													 vtm.getSelectedIssuer());
				if (m_PaymentTokens == null) {
					// Record ourselves with tab manager, so we return here after
					// the user has purchased tokens.  Then go purchase tokens.

					m_TabManager.setPendingTokens(m_TabManager.getConfirmPane());
					m_TabManager.queueTransition(
												new Integer(WTabManager.M_PURCHASE_TOKENS));
					m_TabManager.makeTransition();
					return;
				}

				// steps 3 and 4
				sendStorePayment();
			}
		};
	}
	/**
	 * private utility method to request that the SDS store a payment record.
	 * NB: input will be found in m_PaymentRecord
	 * NB: this method will block pending the SDS's reply
	 */
	private void sendStorePayment() {
		if (m_PaymentRecord == null || !m_PaymentRecord.isInit()) {
			Log.error("Missing or uninitialized payment record cannot be stored");
			return;
		}
		final String genErr = "cannot store pending payment";

		VscState state = m_Plugin.getVscState();
		if (!state.isLoggedIn()) {
			Log.error("Cannot store pending payments while not logged in!");
			showError(genErr, "not logged in", "please log in first");
			return;
		}

		// open a connection to the SDS if we don't have one
		XMPPConnection sdsConn = m_Plugin.getSDSListener().getSDSConnection();
		if (sdsConn == null) {
			sdsConn = m_Plugin.openSDS();
		}
		if (sdsConn == null) {
			// can't do anything
			Log.error("Cannot store pending payment without SDS connection!");
			showError(genErr, "SDS unavailable", "try again later");
			return;
		}

		// make encrypted blob out of the record
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		String encPmtRec
			= m_PaymentRecord.getEncryptedBase64(secrets.getPubKey());
		if (encPmtRec.isEmpty()) {
			Log.error("error encrypting XMLPayment record with public key");
			showError(genErr, "encrypt failed");
			return;
		}

		// build message to send
		SDSMessage storePmt = new SDSMessage();
		storePmt.setType(IQ.Type.SET);
		storePmt.setFrom(sdsConn.getUser());
		storePmt.setTo(sdsConn.getServiceName());
		storePmt.setPacketID("pstore_" + state.getNextSDSid());
		storePmt.setOpcode("REQ_store_pending_pmt");
		String rIdx = new String(secrets.getReceiptIndex());
		storePmt.setFolder_index(rIdx);
		String rwCap = new String(secrets.getReceiptRWCap());
		storePmt.setReadwritecap(rwCap);
		storePmt.setPayment_blob(encPmtRec);

		// add one token as payment
		XMLToken pmtTok = m_PaymentTokens.get(0);
		storePmt.addPaymentToken(pmtTok.toXML());
		TokenTableModel ttm = m_ParentHome.getTokenModel();
		ttm.commitTokens(m_PaymentTokens);

		// wait cursor and disable buttons
		m_ParentHome.setCursor(M_WaitCursor);
		setButtonState(false);

		// register synchronous reply collector
		PacketIDFilter pFilter = new PacketIDFilter(storePmt.getPacketID());
		PacketCollector repColl = sdsConn.createPacketCollector(pFilter);

		// send request to the SDS
		sdsConn.sendPacket(storePmt);
		//Log.debug(storePmt.getOpcode() + " sent with index: " + rIdx);

		// record outgoing packet
		state.setLastActivity(storePmt);

		// wait for reply or timeout
		SDSMessage rep = null;
		IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
		m_ParentHome.setCursor(null);
		setButtonState(true);
		if (reply == null) {
			// got timeout; build error
			rep = new SDSMessage();
			rep.setOpcode("REP_payment_stored");
			rep.setFrom(storePmt.getTo());
			rep.setTo(storePmt.getFrom());
			rep.setPacketID(storePmt.getPacketID());
			rep.setType(IQ.Type.ERROR);
			rep.setErrcode(HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
			rep.setErrmsg("timed out");
		}
		else {
			rep = (SDSMessage) reply;
		}
		repColl.cancel();

		// invoke callback with result (calls processPmtStoreReply() below)
		m_Plugin.getSDSListener().processPacket(rep);
	}

	/**
	 * method to process a reply received from SDS to a pending payment store
	 * @param storeRep the reply message from the SDS
	 */
	public void processPmtStoreReply(SDSMessage storeRep) {
		m_ParentHome.setCursor(null);
		if (storeRep == null
			|| !storeRep.getOpcode().equalsIgnoreCase("REP_payment_stored"))
		{
			Log.error("Bad payment store reply received");
			return;
		}
		TokenTableModel ttm = m_ParentHome.getTokenModel();
		HousekeepingPane hkp = m_TabManager.getHousekeepingPane();
		SpentTokenCache stc = m_Plugin.getTokenCache();
		String vsNo = m_Plugin.getLoginSecrets().getVSnumber();

		// see whether this reply is an error
		if (storeRep.getType() == IQ.Type.ERROR) {
			/* If the error here is 402 (payment required) then the implication is
			 * that our token was no good, i.e. we are attempting to double-spend
			 * it somehow and the SDS has rejected it.  In this case we need to
			 * record the token as spent, get another one, and retry the store.
			 */
			if (storeRep.getErrcode() == HttpURLConnection.HTTP_PAYMENT_REQUIRED) {
				// record token as spent, and report to spent cache
				ttm.spendTokens(m_PaymentTokens);
				stc.recordToken(m_PaymentTokens.get(0), vsNo);
				m_PaymentTokens.clear();

				// get another token and recurse, in another thread
				VoucherTableModel vtm = m_ParentHome.getVoucherModel();
				m_PaymentTokens
					= ttm.getAvailableTokens(VoucherRequest.M_ReceiptCost,
													 vtm.getSelectedIssuer());
				Runnable retry = new Runnable() {
					public void run() {
						if (m_PaymentTokens == null || m_PaymentTokens.isEmpty()) {
							// need to go purchase tokens
							m_TabManager.setPendingTokens(
																m_TabManager.getConfirmPane());
							m_TabManager.queueTransition(new Integer(
																WTabManager.M_PURCHASE_TOKENS));
							m_TabManager.makeTransition();
							return;
						}

						// steps 3 and 4
						sendStorePayment();
					}
				};
				new Thread(retry).start();
				return;
			}

			// uncommit the token and display error message
			ttm.uncommitTokens(m_PaymentTokens);
			m_PaymentTokens.clear();

			/* About the only way this can happen is if we lost connectivity to
			 * the SDS.  Suggest manual processing to prevent recursion.
			 */
			if (hkp.automating()) {
				showError("error storing pending payment", "code "
							+ storeRep.getErrcode() + ": " + storeRep.getErrmsg(),
							"switch to manual processing");
			}
			else {
				showError("error storing pending payment", "code "
							+ storeRep.getErrcode() + ": " + storeRep.getErrmsg());
			}
			// either way, housekeeping processing stops here
			return;
		}

		// record token as spent, and report to spent cache
		ttm.spendTokens(m_PaymentTokens);
		stc.recordToken(m_PaymentTokens.get(0), vsNo);
		m_PaymentTokens.clear();

		// tell the payment table about the new data and filename where stored
		if (m_PaymentRecord != null) {
			Hashtable<XMLPayment, String> pmtTabData
				= hkp.getPaymentModel().getPaymentList();
			if (pmtTabData == null) {
				Log.error("payment table model data is null");
			}
			else {
				pmtTabData.put(m_PaymentRecord, storeRep.getFilename());
				m_PaymentRecord = null;
			}
		}
		else {
			Log.error("Null confirm payment object, cannot mark read on DHT");
			return;
		}

		// update the housekeeping pane's list of confirmations to process
		hkp.discardConfirmation(m_PaymentConf);

		if (!m_ShowDisplay) {
			if (!m_DeferPopup) {
				// show a popup to user indicating that payment was successful
				final StringBuilder popTxt = new StringBuilder(256);
				popTxt.append("Your payment of " + m_PaymentConf.getQuantity());
				popTxt.append(" " + m_PaymentConf.getUnits() + " of ");
				popTxt.append(m_PaymentConf.getAsset());
				if (!m_PaymentConf.getAsset().equals(m_PaymentConf.getIssuer())) {
					popTxt.append(" at " + m_PaymentConf.getIssuer());
				}
				popTxt.append("\nwas sent to ");
				popTxt.append(m_PaymentConf.getPayee() + " successfully.");
				// show that change voucher was received if there was one
				if (m_PaymentConf.getVoucher() != null) {
					popTxt.append("\nYou have received a voucher back as change.");
				}
				Runnable notice = new Runnable() {
					public void run() {
						m_TabManager.showInfo(popTxt.toString());
					}
				};
				SwingUtilities.invokeLater(notice);
			}
			else {
				/* This state indicates that we're going to build the popup later,
				 * together with some other information, in one of the ../svx UI
				 * panes.  But in order to do this, we need to record the confirm
				 * record we're working with, in the PaymentPane.
				 */
				PaymentPane pp = m_TabManager.getPaymentPane();
				pp.setLatestConfirmation(m_PaymentConf);
			}
		}
		// else: we constructed the regular screen display in prepDisplay()

		/* Check the XMLConfirmPayment again, and see if it has a DHT record id.
		 * If so, we also need to mark that record read on the DHT to prevent
		 * duplicate processing.
		 */
		Integer dhtId = m_PaymentConf.getDHTid();
		if (dhtId != null) {
			// mark this record read on the DHT
			m_ParentHome.setCursor(M_WaitCursor);
			// NB: reply will send us back to the main menu
			m_TabManager.markReadOnDHT(dhtId, true);
		}
		else {
			Log.error("No DHT id found on confirmation record: "
						+ m_PaymentConf.toXML());
			showError("could not mark payment confirmation record read on DHT",
						"no DHT id found");
		}

		// clear buttons so they don't click on Complete again by mistake
		m_ButtonPane.removeAll();
		m_WorkBar.removeAll();
		m_PaymentConf = null;
	}


	// method required to extend WorkPane
	/**
	 * method to prepare display for this sub-class
	 */
	public void prepDisplay() {
		// make sure we have a confirmation to process
		final HousekeepingPane hkp = m_TabManager.getHousekeepingPane();
		m_PaymentConf = hkp.getConfRecord();
		if (m_PaymentConf == null) {
			showError("cannot process payment confirmation",
						"no confirmation record found on DHT",
						"try again later using the Refresh function");
			return;
		}

		// define sole button, to submit and store pending payment record
		JButton submit = new JButton(m_SubmitAction);
		submit.setToolTipText("Click here to store your pending payment record");
		submit.setFont(m_Plugin.M_ButtonFont);

		// validate signature on confirmation record
		VsSecrets sekrits = m_Plugin.getLoginSecrets();
		boolean validSig = m_PaymentConf.validateSig(sekrits.getVPKey());
		if (!validSig) {
			showError("cannot process payment confirmation",
						"payment confirmation signature by VP is not valid!",
						"please report problem to customer service");
			// allow display without processing
			submit.setEnabled(false);
		}

		// deal with the change voucher, if there is one
		XMLVoucher chnge = m_PaymentConf.getVoucher();
		if (chnge != null) {
			/* tell voucher table model about the new voucher, though we don't
			 * yet know (without querying for a new list) in what filename it's
			 * stored
			 */
			m_ParentHome.addVoucher(chnge, "pending");

			// now to get the actual voucher filename, we query for a refresh
			XMPPConnection sdsConn = m_Plugin.getSDSListener().getSDSConnection();
			m_TabManager.queryVouchers(sdsConn, false);
		}

		// set heading pane
		buildValueHeader();

		if (!m_ShowDisplay) {
			// we do not want to display anything to user; automate processing
			ActionEvent trigger = new ActionEvent(submit, 0, "AUTOCOMPLETE");
			m_SubmitAction.actionPerformed(trigger);
			return;
		}

		// set screen label
		buildScreenLabel("Complete Your Payment");

		// provide action button in the button pane and toolbar
		m_ButtonPane.removeAll();
		m_WorkBar.removeAll();
		m_ButtonPane.add(Box.createHorizontalStrut(300));		// spacer
		m_ButtonPane.add(submit);

		// duplicate it in toolbar
		JButton dupSubmit = new JButton(m_SubmitAction);
		dupSubmit.setToolTipText("Click here to store your pending payment "
										+ "record");
		dupSubmit.setFont(m_Plugin.M_ButtonFont);
		m_WorkBar.add(dupSubmit);

		// prepare a panel to place inside the scroll area
		JPanel confForm = new JPanel(new GridBagLayout());
		m_GBC.fill = GridBagConstraints.NONE;
		m_GBC.anchor = GridBagConstraints.NORTHWEST;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		confForm.add(Box.createVerticalGlue(), m_GBC);
		confForm.add(Box.createVerticalStrut(10), m_GBC);

		// display certain selected confirmation details

		// show the payee, plus pickup days
		JLabel payeeLab = new JLabel("Payment to");
		payeeLab.setFont(m_ParentHome.M_LabelFont);
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		confForm.add(payeeLab, m_GBC);
		Date now = new Date();
		long pickupusec = m_PaymentConf.getTTL() - now.getTime();
		long days = pickupusec / javax.management.timer.Timer.ONE_DAY + 1;
		String payeeVS = m_PaymentConf.getPayee();
		int atIdx = payeeVS.indexOf('@');
		if (atIdx != -1) {
			payeeVS = payeeVS.substring(0, atIdx);
		}
		String payeeDays = " :  " + payeeVS + "  (Pickup " + days + " days)";
		JLabel payee = new JLabel(payeeDays, SwingConstants.TRAILING);
		payee.setFont(m_ParentHome.M_DataFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		confForm.add(payee, m_GBC);
		confForm.add(Box.createVerticalStrut(5), m_GBC);

		// amount in base units
		JLabel amtLab = new JLabel("Payment amount ");
		amtLab.setFont(m_ParentHome.M_LabelFont);
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		confForm.add(amtLab, m_GBC);
		StringBuilder amtStr = new StringBuilder(32);
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(0);
		df.setGroupingSize(3);
		double amt = m_PaymentConf.getQuantity();
		amtStr.append(" = " + df.format(amt) + " ");
		amtStr.append(m_PaymentConf.getDisplayUnits() + " ");
		JLabel amount = new JLabel(amtStr.toString());
		amount.setFont(m_ParentHome.M_DataFont);

		// obtain desired currency from preferences and also show value therein
		String prefCurrency = m_Plugin.getUserPreferences().getValueCurrency();
		String lookUnits
			= VPAssetConfig.getBasePriceLookupUnits(m_PaymentConf.getAsset());
		AssetPriceCache apc = m_Plugin.getPricingModule();
		Double price = apc.getRelativePrice(lookUnits, prefCurrency);
		if (price == null) {
			price = new Double(1.0);
			prefCurrency = m_PaymentConf.getDisplayUnits();
		}
		df.setMaximumFractionDigits(2);
		df.setMinimumFractionDigits(2);
		StringBuilder fiatAmt = new StringBuilder(32);
		fiatAmt.append("( ");
		fiatAmt.append(df.format(amt * price.doubleValue()));
		fiatAmt.append(" " + prefCurrency + " )");
		JLabel fiat = new JLabel(fiatAmt.toString());
		fiat.setFont(m_ParentHome.M_DataItalFont);

		// place these amounts side-by-side into a mini-pane
		JPanel amtPane = new JPanel(new GridLayout(1, 2));
		amtPane.setBorder(null);
		amtPane.add(amount);
		amtPane.add(fiat);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		confForm.add(amtPane, m_GBC);
		confForm.add(Box.createVerticalStrut(15), m_GBC);

		// message label, regardless of how many baggage fields we have
		JLabel msgLab = new JLabel("Message:");
		msgLab.setFont(m_ParentHome.M_LabelFont);
		confForm.add(msgLab, m_GBC);
		confForm.add(Box.createVerticalStrut(5), m_GBC);

		// show the other details (Message or other baggage fields)
		PaymentPane pp = m_TabManager.getPaymentPane();
		Hashtable<String, String> bagHash = pp.getPaymentBaggage();
		String bagXML = bagHash.get(m_PaymentConf.getPayee());
		if (bagXML != null && !bagXML.isEmpty()) {
			Element bagElt = EncodingUtils.buildElement(bagXML);
			if (bagElt == null) {
				Log.error("Stored payment baggage did not parse: "
							+ bagXML);
			}
			else {
				List bagElements = bagElt.elements();
				Iterator fldIt = bagElt.elementIterator();
				while (fldIt.hasNext()) {
					Element fldElt = (Element) fldIt.next();
					String eltNm = fldElt.getName();
					// skip meta fields
					if (eltNm.equalsIgnoreCase("field_names")
						 || eltNm.equalsIgnoreCase("signature"))
					{
						continue;
					}
					else if (!eltNm.equals("Message")) {
						// strip the leading B
						eltNm = eltNm.substring(1);
					}

					// fix any escaped angle brackets
					String eltVal = fldElt.getTextTrim();
					eltVal = eltVal.replaceAll(";lt;", "<");
					eltVal = eltVal.replaceAll(";gt;", ">");

					// special case if we have only one baggage item, = Message
					if (bagElements.size() == 3 && eltNm.equals("Message")) {
						JTextArea msg = new JTextArea(1, 40);
						msg.setEditable(false);
						msg.setLineWrap(true);
						msg.setWrapStyleWord(true);
						msg.setFont(m_ParentHome.M_DataFont);
						msg.setBorder(null);
						msg.setBackground(m_Plugin.m_LabelBackColor);
						msg.setFocusable(false);
						msg.setText(eltVal);
						m_GBC.fill = GridBagConstraints.HORIZONTAL;
						m_GBC.gridwidth = GridBagConstraints.REMAINDER;
						confForm.add(msg, m_GBC);
						// we're done now
						continue;
					}

					JLabel bagLab = new JLabel(eltNm);
					bagLab.setFont(m_ParentHome.M_LabelFont);
					m_GBC.gridwidth = GridBagConstraints.RELATIVE;
					m_GBC.fill = GridBagConstraints.NONE;
					confForm.add(bagLab, m_GBC);

					// show the value in a text area so we can get it all
					JTextArea bagValue = new JTextArea(1, 20);
					bagValue.setEditable(false);
					bagValue.setLineWrap(true);
					bagValue.setWrapStyleWord(true);
					bagValue.setBackground(m_Plugin.m_LabelBackColor);
					bagValue.setFont(m_ParentHome.M_DataFont);
					bagValue.setBorder(null);
					bagValue.setFocusable(false);
					bagValue.setText(":  " + eltVal);
					m_GBC.gridwidth = GridBagConstraints.REMAINDER;
					confForm.add(bagValue, m_GBC);
				}
			}
		}

		// now show system-generated details below, centered in smaller font
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_GBC.fill = GridBagConstraints.NONE;
		confForm.add(Box.createVerticalStrut(50), m_GBC);

		JPanel sysPane = new JPanel(new GridBagLayout());
		sysPane.setBorder(null);

		// serial number of voucher issued to payee
		JLabel serialLab = new JLabel("Serial number of payee voucher:");
		serialLab.setFont(m_ParentHome.M_DetailFont);
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		sysPane.add(serialLab, m_GBC);

		JLabel serialNum = new JLabel(m_PaymentConf.getVoucherSerial());
		serialNum.setFont(m_ParentHome.M_DetailFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		sysPane.add(serialNum, m_GBC);
		sysPane.add(Box.createVerticalStrut(5), m_GBC);

		// show expiration date
		JLabel expLab = new JLabel("Payment available for pickup until:  ");
		expLab.setFont(m_ParentHome.M_DetailFont);
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		sysPane.add(expLab);

		SimpleDateFormat sdf = new SimpleDateFormat(XMLVoucher.M_DateFmt);
		String expStr = sdf.format(new Date(m_PaymentConf.getTTL()));
		JLabel expDate = new JLabel(expStr);
		expDate.setFont(m_ParentHome.M_DetailFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		sysPane.add(expDate, m_GBC);
		sysPane.add(Box.createVerticalStrut(5), m_GBC);

		// show the amount and serial number of change voucher, if there is one
		if (chnge != null) {
			serialLab = new JLabel("Change voucher serial number:");
			serialLab.setFont(m_ParentHome.M_DetailFont);
			m_GBC.gridwidth = GridBagConstraints.RELATIVE;
			sysPane.add(serialLab, m_GBC);

			serialNum = new JLabel(chnge.getSerialNumber());
			serialNum.setFont(m_ParentHome.M_DetailFont);
			m_GBC.gridwidth = GridBagConstraints.REMAINDER;
			sysPane.add(serialNum, m_GBC);
			sysPane.add(Box.createVerticalStrut(5), m_GBC);

			amtLab = new JLabel("Amount of your change:");
			amtLab.setFont(m_ParentHome.M_DetailFont);
			m_GBC.gridwidth = GridBagConstraints.RELATIVE;
			sysPane.add(amtLab, m_GBC);

			amtStr = new StringBuilder(64);
			df.setMaximumFractionDigits(XMLVoucher.M_Precision);
			df.setMinimumFractionDigits(0);
			amtStr.append(df.format(chnge.getQuantity()) + " ");
			amtStr.append(chnge.getUnits() + " of " + chnge.getAsset());
			amount = new JLabel(amtStr.toString());
			amount.setFont(m_ParentHome.M_DetailFont);
			m_GBC.gridwidth = GridBagConstraints.REMAINDER;
			sysPane.add(amount, m_GBC);
		}
		m_GBC.anchor = GridBagConstraints.NORTH;
		m_GBC.weightx = 0.0;
		confForm.add(sysPane, m_GBC);
		confForm.add(Box.createVerticalGlue(), m_GBC);

		// put entire form into the main scroll area
		m_WorkArea.setViewportView(confForm);
	}

	/**
	 * method to set display state
	 * @param override if true, actually show screen to user and prompt for input
	 */
	public void setDisplayOverride(boolean override) {
		m_ShowDisplay = override;
	}

	/**
	 * set the popup deferral state
	 * @param defer if true, do not display success popup to user at this time
	 */
	public void setDeferPopup(boolean defer) {
		m_DeferPopup = defer;
	}

	// END methods
}
