/*
 * last modified---
 * 	08-31-18/kbw/ do not commit tokens stored in the SpentTokenCache
 * 	08-05-14/jet/ Spark plugin version
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide a UI panel for displaying the 'home' of a wallet once the
 * 	wallet has been logged in
 */

package org.silentvault.spark.ui.wallet;

import javax.swing.*;
import java.awt.event.*;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.FlowLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Component;
import java.awt.Color;
import java.awt.Point;
import java.awt.Cursor;
import java.awt.Font;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;
import javax.swing.table.DefaultTableColumnModel;

import org.dom4j.DocumentException;
import org.dom4j.Element;

import org.jivesoftware.spark.util.log.Log;

import org.jivesoftware.smack.packet.IQ;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;


/**
 * Extends JPanel to provide a screen for displaying the contents of a voucher
 * safe and presenting interface elements to manipulate those contents.
 * The actual UI controls and data displayed vary with the WorkPane subclass.
 *
 * @author Kevin Wilkerson
 */
public final class HomePane extends JPanel {
	// BEGIN data members
	/**
	 * tool tips for voucher table columns
	 */
	protected final String[]   M_VoucherTooltips =
		{	"Click here to select this voucher",
			"The quantity of backing asset represented by the voucher",
			"The equivalent value in your chosen currency",
			"The expiration date of the voucher" };

	/**
	 * tool tips for token table columns
	 */
	protected final String[]   M_TokenTooltips =
		{  "The unique serial number of your token",
			"The quantity of backing asset represented by the token",
			"The type of backing asset",
			"The Issuer where the backing assets are stored",
			"Whether the token bears a valid signature from the publisher" };

	/**
	 * the plugin/application to which we belong
	 */
	private WalletClient			m_Plugin;

	/**
	 * the tab manager which creates and manages us
	 */
	private WTabManager			m_TabManager;

	/**
	 * the current title bar text
	 */
	private String					m_TitlebarText;

	/**
	 * the table model for the voucher table, which also tracks vouchers
	 */
	private VoucherTableModel	m_VoucherModel;

	/**
	 * the table which contains the vouchers that are shown in various panes
	 */
	private JTable					m_VoucherTable;

	/**
	 * the table model for the token table, which also tracks tokens
	 */
	private TokenTableModel		m_TokenModel;

	/**
	 * the table which contains the tokens which can be shown in a popup pane
	 */
	private JTable					m_TokenTable;

	/**
	 * the work pane, which encloses an object {@link WorkPane} whose actual
	 * contents will vary based on its subclass
	 */
	private JPanel					m_WorkPane;

	/**
	 * the toolbar container for menu operations
	 */
	private JPanel					m_WorkBar;

	/**
	 * the complete Home Pane, placed into the scroll area (m_ScrollPane)
	 */
	private JPanel					m_HomePane;

	/**
	 * the scroll area used to ensure that this pane always displays everything
	 * regardless of window size
	 */
	private JScrollPane			m_ScrollPane;

	/**
	 * new vouchers that have arrived from the SDS and must be decoded and parsed
	 * before (re)building the voucher table
	 */
	private Hashtable<String, String>	m_IncomingVouchers;

	/**
	 * status/progress bar
	 */
	private JPanel					m_StatusBar;

	/**
	 * standard font for use in titles and headings
	 */
	public final Font				M_HeaderFont;

	/**
	 * standard font for use in table columns
	 */
	public final Font				M_TableFont;

	/**
	 * standard font for use in labels
	 */
	public final Font				M_LabelFont;

	/**
	 * standard font for use in non-table data
	 */
	public final Font				M_DataFont;

	/**
	 * standard font for use with italicized data
	 */
	public final Font				M_DataItalFont;

	/**
	 * standard font for use in non-table titles
	 */
	public final Font				M_TitleFont;

	/**
	 * standard font for use in fine-print details
	 */
	public final Font				M_DetailFont;

	/**
	 * standard color for title field backgrounds
	 */
	public final Color			M_TitleBackgroundColor;

	// END data members
	

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the plugin or application which owns us
	 * @param mgr the tab manager which created us
	 */
	public HomePane(WalletClient plugin, WTabManager mgr) {
		super();
		m_Plugin = plugin;
		m_TabManager = mgr;

		// use grid layout
		setLayout(new GridLayout());

		// set up voucher table model (empty)
		m_VoucherModel
			= new VoucherTableModel(new Hashtable<XMLVoucher, String>(),
											m_Plugin.getLoginSecrets().getVPKey(),
											m_Plugin);
		m_VoucherTable = mkNewVoucherTable();

		// set up token model (empty)
		m_TokenModel = new TokenTableModel(new ArrayList<XMLToken>(), m_Plugin);
		m_TokenTable = mkNewTokenTable();

		// init font values
		M_HeaderFont = new Font(Font.SANS_SERIF, Font.BOLD, 14);
		M_TableFont = new Font(Font.SANS_SERIF, Font.PLAIN, 13);
		M_LabelFont = new Font(Font.SANS_SERIF, Font.BOLD, 13);
		M_DataFont = new Font(Font.SANS_SERIF, Font.PLAIN, 12);
		M_DataItalFont = new Font(Font.SANS_SERIF, Font.ITALIC, 12);
		M_TitleFont = new Font(Font.SANS_SERIF, Font.PLAIN, 16);
		M_DetailFont = new Font(Font.SANS_SERIF, Font.PLAIN, 10);

		// init color value for title backgrounds
		M_TitleBackgroundColor = new Color(0x99CCFF);

		// provide standard dimension for voucher and receipt tables
		UIManager.put("vouchTableDim", new Dimension(450, 240));

		// specify defaults for tool tips
		ToolTipManager ttm = ToolTipManager.sharedInstance();
		ttm.setInitialDelay(200);
		ttm.setDismissDelay(60000);
		ttm.setReshowDelay(1000);

		// create toolbar, for containing menu button actions
		m_WorkBar = new JPanel();
		m_WorkBar.setLayout(new BoxLayout(m_WorkBar, BoxLayout.X_AXIS));
		m_WorkBar.setBorder(null);
		m_WorkBar.setVisible(false);
		m_WorkBar.setPreferredSize(new Dimension(485, 25));

		// create default work pane, to be filled with WorkPane subclasses
		m_WorkPane = new JPanel();
		m_WorkPane.setLayout(new GridLayout());
		
		// create empty status bar panel
		m_StatusBar = new JPanel();
		m_StatusBar.setLayout(new BoxLayout(m_StatusBar, BoxLayout.X_AXIS));
		m_StatusBar.setPreferredSize(new Dimension(485, 20));
		m_StatusBar.setVisible(true);
		m_StatusBar.setBorder(BorderFactory.createLineBorder(
																Color.LIGHT_GRAY.darker(), 1));
		m_StatusBar.setBackground(Color.LIGHT_GRAY);

		// set up main scroll pane containing the objects created above
		buildHomePane();
		m_ScrollPane = new JScrollPane(m_HomePane);
		m_ScrollPane.setBorder(null);
		add(m_ScrollPane);
	}
	
	/**
	 * helper method to fully initialize the main home pane (which goes in the
	 * scroll pane which we actually display)
	 */
	private void buildHomePane() {
		GridBagLayout gbl = new GridBagLayout();
		m_HomePane = new JPanel(gbl);
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.weighty = 1.0;
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.gridheight = 5;
		gbc.anchor = GridBagConstraints.CENTER;
		gbc.fill = GridBagConstraints.NONE;
		m_HomePane.add(Box.createVerticalGlue(), gbc);
		m_HomePane.add(m_WorkBar, gbc);
		m_HomePane.add(m_WorkPane, gbc);
		m_HomePane.add(m_StatusBar, gbc);
		m_HomePane.add(Box.createVerticalGlue(), gbc);
		m_HomePane.setBorder(null);
	}

	/**
	 * helper method to create a JTable (for vouchers), also providing
	 * column tooltip definitions
	 */
	private JTable mkNewVoucherTable() {
		// fix column header for fiat equivalent column
		String prefCurrency = m_Plugin.getUserPreferences().getValueCurrency();
		m_VoucherModel.setEquivHeader(prefCurrency);

		JTable newTab = new JTable(m_VoucherModel) {
			// provide table header tool tips
			protected JTableHeader createDefaultTableHeader() {
				return new JTableHeader(columnModel) {
					public String getToolTipText(MouseEvent e) {
						String tip = null;
						Point ppp = e.getPoint();
						int index = columnModel.getColumnIndexAtX(ppp.x);
						int realIndex = columnModel.getColumn(index).getModelIndex();
						return M_VoucherTooltips[realIndex];
					}
				};
			}
			public boolean isCellEditable(int row, int col) {
				// objects in the select column can be editable, others are not
				if (col == 0) {
					if (m_VoucherModel.getSelectMode()
						 != VoucherTableModel.M_SelectNone)
					{
						return true;
					}
				}
				return false;
			}
		};
		newTab.setCellSelectionEnabled(false);
		// this size should be same as WorkPane.m_WorkArea
		newTab.setPreferredScrollableViewportSize(
													UIManager.getDimension("vouchTableDim"));
		newTab.setFillsViewportHeight(true);

		// set special renderer for the all columns
		VoucherRenderer vr = new VoucherRenderer(m_VoucherModel, m_Plugin);
		newTab.setDefaultRenderer(JCheckBox.class, vr);
		newTab.setDefaultRenderer(JRadioButton.class, vr);
		newTab.setDefaultRenderer(Float.class, vr);
		newTab.setDefaultRenderer(Double.class, vr);
		newTab.setDefaultRenderer(Date.class, vr);

		// configure the table columns and renderers
		configTableColumns(newTab);

		// set default sorter
		newTab.setAutoCreateRowSorter(true);
		newTab.setAutoCreateColumnsFromModel(true);

		// set spacing preferences
		newTab.setRowHeight(20);
		newTab.setIntercellSpacing(new Dimension(2, 2));
		newTab.setShowGrid(true);
		newTab.setGridColor(UIManager.getColor("TableHeader.background"));

		// establish table as listener for model
		m_VoucherModel.addTableModelListener(newTab);
		return newTab;
	}

	/**
	 * helper method to create a JTable (for tokens) while also providing
	 * column tooltip definitions
	 */
	private JTable mkNewTokenTable() {
		JTable newTab = new JTable(m_TokenModel) {
			// provide table header tool tips
			protected JTableHeader createDefaultTableHeader() {
				return new JTableHeader(columnModel) {
					public String getToolTipText(MouseEvent e) {
						String tip = null;
						Point ppp = e.getPoint();
						int index = columnModel.getColumnIndexAtX(ppp.x);
						int realIndex = 0;
						try {
							realIndex = columnModel.getColumn(index).getModelIndex();
						}
						catch (ArrayIndexOutOfBoundsException aioobe) {
							// can happen when cursor goes out of popup window
							return null;
						}
						return M_TokenTooltips[realIndex];
					}
				};
			}
		};
		newTab.setRowSelectionAllowed(true);
		newTab.setColumnSelectionAllowed(false);
		newTab.setSelectionMode(ListSelectionModel.SINGLE_INTERVAL_SELECTION);
		newTab.setSelectionBackground(Color.LIGHT_GRAY);
		newTab.setSelectionForeground(Color.RED);
		// set special renderer for signature validity column
		newTab.setDefaultRenderer(Boolean.class, new TokenRenderer(false));
		newTab.setDefaultRenderer(Double.class, new TokenRenderer(true));
		newTab.setFillsViewportHeight(true);
		// set default sorter
		newTab.setAutoCreateRowSorter(true);
		newTab.setAutoCreateColumnsFromModel(true);
		newTab.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
		newTab.setShowGrid(true);
		newTab.setIntercellSpacing(new Dimension(2, 2));

		// define table as listener for changes
		m_TokenModel.addTableModelListener(newTab);
		return newTab;
	}

	/**
	 * Helper method to configure the columns and renderers of the
	 * voucher table.  This needs to be called when the table is created,
	 * and after each change to the table.
	 * @param tab the table to configure (== m_VoucherTable)
	 */
	public void configTableColumns(JTable tab) {
		// set special renderer for column headers, to cause right justify
		ExpColRenderer ecr = new ExpColRenderer();

		// make sure expiry column is wide enough to show whole date
		TableColumnModel tcm = tab.getColumnModel();
		TableColumn tc = tcm.getColumn(3);
		tc.setMinWidth(135);
		tc.setHeaderRenderer(ecr);

		// make sure equiv column is wide enough to show fiat equivalent
		tc = tcm.getColumn(2);
		tc.setMinWidth(100);
		tc.setHeaderRenderer(ecr);

		// make sure value column is wide enough to show GAU
		tc = tcm.getColumn(1);
		tc.setMinWidth(80);
		tc.setHeaderRenderer(ecr);

		// set max width on select column, and set cell editor
		tc = tcm.getColumn(0);
		tc.setMaxWidth(30);
		tc.setHeaderRenderer(ecr);
		VoucherSelector vs = new VoucherSelector(m_Plugin, m_VoucherModel);
		tc.setCellEditor(vs);
	}

	/**
	 * update the title bar text to reflect the wallet we're logged into
	 */
	public void updateTitleText() {
		m_TitlebarText = "Wallet ID: "
								+ m_Plugin.getLoginSecrets().getVSnumber() + " ";
		setBorder(BorderFactory.createTitledBorder(m_TitlebarText));
		revalidate();
	}


	// get methods
	/**
	 * method to obtain the WorkPane sub-element
	 * @return the component, which is enclosed in the JPanel
	 */
	public WorkPane getWorkPane() { 
		if (m_WorkPane == null) {
			return null;
		}
		// there should be exactly one component inside this panel
		Component comp = null;
		try {
			comp = m_WorkPane.getComponent(0);
		}
		catch (ArrayIndexOutOfBoundsException aioobe) {
			Log.error("Missing WorkPane subclass in enclosing JPanel!", aioobe);
		}
		return (WorkPane) comp;
	}

	/**
	 * method to obtain the object which manages the token table data
	 * @return the table model manager
	 */
	public TokenTableModel getTokenModel() { return m_TokenModel; }

	/**
	 * method to obtain the object which manages the voucher table data
	 * @return the table model manager
	 */
	public VoucherTableModel getVoucherModel() { return m_VoucherModel; }

	/**
	 * obtain the voucher table
	 * @return the table reference
	 */
	public JTable getVoucherTable() { return m_VoucherTable; }

	/**
	 * obtain the token table
	 * @return the table reference
	 */
	public JTable getTokenTable() { return m_TokenTable; }

	/**
	 * obtain menu toolbar
	 * @return the toolbar panel
	 */
	public JPanel getMenuBar() { return m_WorkBar; }

	/**
	 * obtain status bar
	 */
	public JPanel getStatusBar() { return m_StatusBar; }


	// write methods
	/**
	 * set the contents of the work pane (called by WTabManager)
	 * @param workPane a subclass object of WorkPane
	 */
	public void setWorkPane(WorkPane workPane) {
		if (workPane != null) {
			workPane.prepDisplay();
			m_WorkPane.removeAll();
			m_WorkPane.add(workPane);
		}

		// certain objects in our home pane might have been changed by the call
		// to the subclass's prepDisplay()

		m_HomePane.revalidate();
		m_ScrollPane.setViewportView(m_HomePane);
	}

	/**
	 * add a voucher to the voucher list and regen the table
	 * @param voucher the voucher to add
	 * @param file the filename where it's stored
	 */
	public void addVoucher(XMLVoucher voucher, String file) {
		if (voucher == null || !voucher.isInit()
			 || file == null || file.isEmpty())
		{
			return;
		}
		m_VoucherModel.addVoucher(voucher, file);
		configTableColumns(m_VoucherTable);
	}

	/**
	 * remove one or more vouchers from the voucher list and regen table
	 * @param vouchList the list of vouchers we're done with
	 */
	public void spendVouchers(ArrayList<XMLVoucher> vouchList) {
		if (vouchList != null && !vouchList.isEmpty()) {
			m_VoucherModel.removeVouchers(vouchList);
			configTableColumns(m_VoucherTable);
		}
	}

	/**
	 * register the incoming list of encrypted vouchers found on the SDS
	 * @param vouchers the vouchers in a hashtable, mapped against their files
	 */
	public void recordVouchers(Hashtable<String, String> vouchers) {
		// store the vouchers we have received
		m_IncomingVouchers = vouchers;
		setCursor(null);
	}

	/**
	 * register the incoming list of tokens found on the SDS
	 * @param tokens the tokens in the list
	 * @param trueCnt the count of token items returned, which must be
	 * compared against tokens.size() to deduce whether any failed to parse
	 */
	public void recordTokens(ArrayList<XMLToken> tokens, int trueCnt) {
		// merge tokens into the table model and table
		m_TokenModel.mergeTokens(tokens);

		if (trueCnt > tokens.size()) {
			Log.error((trueCnt - tokens.size())
						+ " tokens in wallet did not parse");
			m_TabManager.showError("warning: "
											+ (trueCnt - tokens.size()) + " tokens in "
											+ "wallet would not parse");
		}

		/* Audit the list of tokens returned against our spent tokens cache, so
		 * that we don't attempt to purge tokens which no longer exist.  Note
		 * that we do not do this if there are no tokens.  The assumption here is
		 * that we queryTokens() only after invoking STC.purgeSpentTokens().
		 * (Discrepancies can be introduced via stale tokens lingering in the
		 * spentTokens.xml file flagged for the VS#.)
		 */
		setCursor(null);
		if (tokens.isEmpty()) {
			return;
		}
		SpentTokenCache stc = m_Plugin.getTokenCache();
		String vsNo = m_Plugin.getLoginSecrets().getVSnumber();
		ArrayList<XMLToken> tbPurged = stc.getSpentTokens(vsNo);
		tbPurged.removeAll(tokens);
		if (stc.removeTokens(tbPurged)) {
			Log.debug("Removed " + tbPurged.size()
					+ " tokens from SpentTokenCache which are no longer in folder");
		}
	}

	/**
	 * rebuild the voucher list (called by WTabManager only)
	 */
	public void rebuildVoucherList() {
		if (m_IncomingVouchers != null && !m_IncomingVouchers.isEmpty()) {
			// decode the vouchers we just recorded
			setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			MenuPane mp = m_TabManager.getMenuPane();
			mp.disableMenu();
			VsSecrets secrets = m_Plugin.getLoginSecrets();
			Hashtable<XMLVoucher, String> existingVouchers
				= m_VoucherModel.getVoucherList();
			Hashtable<XMLVoucher, String> newVouchers
				= new Hashtable<XMLVoucher, String>();
			JProgressBar jpb = new JProgressBar();
			jpb.setMinimum(0);
			jpb.setMaximum(m_IncomingVouchers.size());
			jpb.setString("Decoding " + m_IncomingVouchers.size()
								+ " vouchers in wallet, please wait");
			jpb.setStringPainted(true);
			jpb.setBackground(Color.LIGHT_GRAY);
			jpb.setFont(M_DataFont);
			jpb.setOpaque(true);
			m_StatusBar.removeAll();
			m_StatusBar.add(jpb);
			m_StatusBar.revalidate();
			m_StatusBar.repaint();
			int vCnt = 0;
			for (String vouchFile : m_IncomingVouchers.keySet()) {
				jpb.setValue(vCnt+1);

				// see whether we already have this voucher decoded in our list
				if (existingVouchers.containsValue(vouchFile)) {
					// no need to decrypt this file again
					vCnt += 1;
					for (XMLVoucher keyVouch : existingVouchers.keySet()) {
						if (existingVouchers.get(keyVouch).equals(vouchFile)) {
							newVouchers.put(keyVouch, vouchFile);
							break;
						}
					}
					continue;
				}

				String encVoucher = m_IncomingVouchers.get(vouchFile);
				String decVoucher
					= EncodingUtils.getStrFromBase64PubkeyEnc(encVoucher,
																			secrets.getPrivKey());
				if (decVoucher == null || decVoucher.isEmpty()) {
					Log.error("Voucher index " + vCnt + " would not decode with "
								+ "the wallet's private key!");
				}
				else {
					XMLVoucher vouch = null;
					try {
						vouch = XMLVoucher.getVoucherRep(decVoucher);
						vCnt += 1;
						newVouchers.put(vouch, vouchFile);
					}
					catch (DocumentException de) {
						Log.error("Voucher index " + vCnt + " would not parse! "
									+ "XML was: " + decVoucher);
					}
				}
			}
			m_StatusBar.remove(jpb);
			m_Plugin.setDefaultStatusBar();
			mp.enableMenu();
			setCursor(null);

			// update model and table with parsed vouchers
			m_VoucherModel.mergeVouchers(newVouchers);
			configTableColumns(m_VoucherTable);

			// issue error if not all of them worked
			int diff = m_IncomingVouchers.size() - vCnt;
			if (diff > 0) {
				Log.error(diff + " vouchers in wallet did not decrypt, or parse");
				m_TabManager.showError("warning: " + diff + " vouchers in wallet "
												+ "would not decrypt or would not parse");
			}
			m_IncomingVouchers.clear();
		}
		// else: nothing to do
	}

	/**
	 * record the results of a validation operation (validate/merge/split)
	 * @param valRep the message received by the callback in response
	 */
	public void processValidateReply(OFSMessage valRep) {
		if (valRep == null
			 || !valRep.getOpcode().equalsIgnoreCase("REP_validation"))
		{
			Log.error("processValidateReply, bad input message");
			return;
		}

		ArrayList<XMLVoucher> vList = m_VoucherModel.getCommittedVouchers();
		if (valRep.getType() == IQ.Type.ERROR) {
			// decommit the vouchers we previously allocated to this operation
			m_VoucherModel.uncommitVouchers(vList);

			// ditto tokens
			ArrayList<XMLToken> tList = m_TokenModel.getCommittedTokens();
			m_TokenModel.uncommitTokens(tList);
			m_TokenTable.clearSelection();
		}
		else {
			setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

			// spend and decommit all the vouchers we committed to this operation
			m_VoucherModel.removeVouchers(vList);
			configTableColumns(m_VoucherTable);

			// same for tokens, spend all currently committed
			m_TokenModel.spendTokens(null);
			m_TokenTable.clearSelection();

			// now decrypt and parse the <validation/> reply block from the VP
			VsSecrets secrets = m_Plugin.getLoginSecrets();
			final String err = "error processing validate/split/merge reply";
			final String remedy = "use Resync to update voucher/token lists";
			String decVal
				= EncodingUtils.getStrFromBase64PubkeyEnc(valRep.getValidation(),
																		secrets.getPrivKey());
			if (decVal == null || decVal.isEmpty()) {
				Log.error("Unable to decrypt voucher validate/split/merge reply");
				m_TabManager.showError(err,
											"could not decrypt validation block", remedy);
				return;
			}

			Element valElt = EncodingUtils.buildElement(decVal);
			if (valElt == null) {
				Log.error("Unparseable <validation/>");
				m_TabManager.showError(err, "validation block did not parse",
											remedy);
				return;
			}
			Element vListElt = valElt.element("voucherList");
			if (vListElt == null) {
				Log.error("Missing <validation/> voucherList");
				m_TabManager.showError(err, "new voucher list missing", remedy);
				return;
			}
			Iterator vIt = vListElt.elementIterator();
			boolean ok = true;
			while (vIt.hasNext()) {
				Element vElt = (Element) vIt.next();
				String eltName = vElt.getName().toLowerCase();
				if (!eltName.equals("voucher")) {
					ok = false;
					Log.error("Illegal element in <voucherList/> section, "
								+ eltName);
					continue;
				}

				// get the filename from the attribute
				String fname = vElt.attributeValue("filename");
				if (fname == null || fname.isEmpty()) {
					Log.error("Empty filename attribute in <voucherList>");
					ok = false;
					continue;
				}

				// obtain the voucher itself
				XMLVoucher vouch = null;
				try {
					vouch = XMLVoucher.getVoucherRep(vElt);
				}
				catch (DocumentException de) {
					Log.error("Voucher parse error", de);
					ok = false;
					continue;
				}

				// add to the table's backing model
				addVoucher(vouch, fname);
			}
			setCursor(null);

			if (!ok) {
				m_TabManager.showError(err, "parse error on one or more vouchers",
											remedy);
			}
			if (valRep.syncPending()) {
				m_TabManager.showError("warning: synchronization pending",
								"the voucher validate/split/merge succeeded, but the "
								+ "publisher\nhas queued one or more operations due to "
								+ "a connectivity loss at the SDS.",
								"wait a few minutes and use Resync before performing "
								+ "further operations");
			}
		}
	}

	// END methods
}
