/*
 * last modified---
 * 	08-15-18/kbw/ explicitly cancel PacketCollectors after use
 * 	01-29-18/kbw/ use getBasePriceLookupUnits()
 * 	11-24-16/jet/ remove distinction of new mintage; always send receipts
 * 	07-15-15/jet/ use VPAssetConfig.getPriceLookupUnits() to compute value
 * 	08-05-14/jet/ Spark plugin version
 * 	03-19-14/jet/ improve spacing with OpenJDK-7 fonts
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide a UI widget to handle processing of incoming voucher payments
 */

package org.silentvault.spark.ui.wallet;

import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;

import org.dom4j.Element;
import org.dom4j.DocumentException;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.util.StringUtils;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;

import java.util.ArrayList;
import java.util.TreeMap;
import java.util.Hashtable;
import java.util.HashMap;
import java.util.Date;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.security.*;
import java.net.HttpURLConnection;


/**
 * Extends WorkPane to provide a screen for handling incoming payments.
 *
 * @author Kevin Wilkerson
 */
public final class IncomingPane extends WorkPane {
	// BEGIN data members
	/**
	 * maximum returned message length
	 */
	private final int					M_MaxMessage = 256;

	/**
	 * storage for token used in storing a receipt on the DHT or SDS
	 */
	private ArrayList<XMLToken>   m_PaymentTokens;

	/**
	 * storage for the pending payment record we are processing
	 */
	private XMLPendingPayment		m_IncPayment;

	/**
	 * storage for the receipt back to the payer, to be stored on the DHT
	 */
	private XMLReceipt				m_PayerReceipt;

	/**
	 * storage for our copy of the receipt, to be stored on the SDS
	 */
	private XMLReceipt				m_ReceiptCopy;

	/**
	 * UI element to hold payer receipt message, if any
	 */
	private JTextArea					m_ReceiptMsg;

	/**
	 * the computed cost of picking up the indicated payment
	 */
	private int							m_TokCost;

	/**
	 * the actual cost of picking up the indicated payment
	 */
	private int							m_TokensSpent;

	/**
	 * action to pick up the payment
	 */
	private AbstractAction			m_PickupAction;

	/**
	 * the action to skip over this item and move to the next
	 */
	private AbstractAction			m_SkipAction;

	/**
	 * flag to record when we are in the middle of buying tokens to store
	 * receipts, and should not re-display pickup screen
	 */
	private boolean					m_SuppressDisplay;

	/**
	 * flag to indicate whether we need to re-sync token list before a pickup
	 */
	private boolean					m_NeedTokenSync;

	/**
	 * inner class to provide a DocumentListener for the Message entry box
	 */
	final class MessageDocumentListener implements DocumentListener {
		/**
		 * handle inserts
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			Document doc = de.getDocument();
			int remaining = M_MaxMessage - doc.getLength();
			Integer rem = new Integer(remaining);
			m_CharCount.setText(rem.toString());
			if (remaining < 0) {
				// silently throw away the excess text
				String maxText = "";
				try {
					maxText = doc.getText(0, M_MaxMessage);
				}
				catch (BadLocationException ble) {
					Log.error("Error stripping excess text from message box", ble);
					return;
				}
				queueTextMod(maxText);
			}
		}

		/**
		 * handle deletions
		 * @param de the removal event
		 */
		public void removeUpdate(DocumentEvent de) {
			Document doc = de.getDocument();
			int remaining = M_MaxMessage - doc.getLength();
			Integer rem = new Integer(remaining);
			m_CharCount.setText(rem.toString());
		}

		/**
		 * handle changes (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JTextArea");
		}
	}

	/**
	 * document listener for dynamic updates of m_MessageField
	 */
	private MessageDocumentListener  m_MessageListener;

	/**
	 * document length counter which shows how many Message chars are left
	 */
	private JLabel							m_CharCount;

	/**
	 * inner class to modify a text component from within its document listener,
	 * without triggering a recursive modification exception
	 */
	class TextAreaMod implements Runnable {
		String		m_NewContents;

		/**
		 * constructor
		 * @param newText the replacement text
		 */
		public TextAreaMod(String newText) {
			m_NewContents = new String(newText);
		}

		/**
		 * execute main task in background
		 */
		public void run() {
			try { Thread.sleep(25); }
			catch (InterruptedException ie) { /* ignore */ }
			m_ReceiptMsg.setText(m_NewContents);
			m_DoingUndo = false;
			m_ReceiptMsg.setCaretPosition(m_NewContents.length());
		}
	}

	/**
	 * track whether an undo mod is in progress
	 */
	private boolean						m_DoingUndo;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the HomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public IncomingPane(WalletClient plugin, HomePane home, WTabManager mgr) {
		super(plugin, home, mgr);
		m_PaymentTokens = new ArrayList<XMLToken>();

		// init UI object for optional message back to payer
		m_ReceiptMsg = new JTextArea(2, 36);
		m_ReceiptMsg.setToolTipText("Enter any message for payer here");
		m_ReceiptMsg.setLineWrap(true);
		m_ReceiptMsg.setWrapStyleWord(true);
		m_ReceiptMsg.setBorder(m_InsetBorder);
		m_MessageListener = new MessageDocumentListener();
		m_ReceiptMsg.getDocument().addDocumentListener(m_MessageListener);

		Integer mlen = new Integer(M_MaxMessage);
		m_CharCount = new JLabel(mlen.toString());
		m_CharCount.setHorizontalAlignment(SwingConstants.CENTER);
		m_CharCount.setFont(m_ParentHome.M_DataFont);
		m_CharCount.setToolTipText("The number of message characters remaining");

		// define process incoming payment action
		setPickupAction();

		// define skip to next action
		final HousekeepingPane hkp = m_TabManager.getHousekeepingPane();
		m_SkipAction = new AbstractAction("Do this one later") {
			public void actionPerformed(ActionEvent actionEvent) {
				hkp.processNextItem(-1);
			}
		};

		// we want no border around the whole thing
		m_WorkArea.setBorder(null);
	}

	/**
	 * utility method to define the action to process and pickup the payment
	 */
	private void setPickupAction() {
		m_PickupAction = new AbstractAction("Pick up payment") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_IncPayment == null) {
					Log.error("Missing pending payment record");
					showError("cannot process pending payment",
								"internal error, no pending payment record found");
					return;
				}

				/* Perform these steps in order:
				 * 	1. Validate the signatures on the payment and voucher.
				 * 	1b. Ensure we have sufficient tokens available.
				 *		2. Generate a REQ_pickup_payment of the voucher.
				 *		3. Place copy of receipt into our safe on the SDS.
				 *		4. On receipt of reply, create receipt and send to DHT.
				 *		5. On successful upload, mark our DHT record read.
				 */
				// step 1
				PublicKey vpKey = m_Plugin.getLoginSecrets().getVPKey();

				// validate signature of VP
				if (!m_IncPayment.validateSig(vpKey)) {
					Log.error("Bad signature on pending payment object");
					showError("cannot process pending payment",
								"publisher signature did not verify");
					return;
				}

				// validate sig of VP on new voucher
				XMLVoucher incVouch = m_IncPayment.getVoucher();
				if (!incVouch.validateSig(vpKey)) {
					Log.error("Bad signature on pending payment voucher");
					showError("cannot process pending payment",
								"publisher signature on voucher did not verify");
					return;
				}

				// step 1b - make sure we have enough tokens
				if (m_TokCost > 0) {
					// NB: this step should have been taken care of in prepDisplay()
					TokenTableModel ttm = m_ParentHome.getTokenModel();
					VoucherTableModel vtm = m_ParentHome.getVoucherModel();
					m_PaymentTokens = ttm.getAvailableTokens(m_TokCost,
																		vtm.getSelectedIssuer());
					if (m_PaymentTokens == null) {
						// Record ourselves with tab manager, so we return here after
						// the user has purchased tokens.  Then go purchase tokens.

						m_TabManager.setPendingTokens(m_TabManager.getIncomingPane());
						m_TabManager.queueTransition(
												new Integer(WTabManager.M_PURCHASE_TOKENS));
						m_TabManager.makeTransition();
						return;
					}

					// if we have more than the tokens required for a validation (4),
					// remove the excess from m_PaymentTokens

					int diff
						= m_PaymentTokens.size() - VoucherRequest.M_ValidateCost;
					for (int ttt = 0; ttt < diff; ttt++) {
						m_PaymentTokens.remove(ttt);
					}
				}

				// if we have a receipt message, make sure it's not too long
				String receiptMsg = m_ReceiptMsg.getText().trim();
				if (receiptMsg.length() > M_MaxMessage) {
					showError("cannot process pending payment",
								"receipt message is too long",
								"please enter a payer message of " + M_MaxMessage
								+ " chars or less");
					return;
				}

				// step 2 (leading to 3, 4, 5)
				sendPickupPayment(incVouch);
			}
		};
	}


	// method required to extend WorkPane
	/**
	 * method to prepare display for this sub-class
	 */
	public void prepDisplay() {
		/* If we had to go buy tokens to store a receipt (which happens if we're
		 * an empty safe picking up our first voucher), then we'll return here
		 * via WTabManager.  But in this case we don't want to display our pickup
		 * info again, just go make another attempt to store the receipt(s).
		 */
		if (m_SuppressDisplay) {
			sendReceiptToSDS();		// (synchronous)

			// if we built one, also send the payer receipt to the DHT
			if (m_PayerReceipt != null) {
				sendReceiptToDHT();
				// NB: step 5 follows on upload receipt reply
			}
			else {
				// perform step 5
				markPendPayReadOnDHT();
			}

			// reset flag
			m_SuppressDisplay = false;
			return;
		}

		// obtain the item from housekeeping
		final HousekeepingPane hkp = m_TabManager.getHousekeepingPane();
		m_IncPayment = m_TabManager.getHousekeepingPane().getPendPayRecord();
		if (m_IncPayment == null) {
			Log.error("Missing pending payment record, cannot process");
			showError("cannot do refresh item", "incoming payment not found");
			return;
		}
		final boolean automating = hkp.automating();

		// set heading pane
		buildValueHeader();

		// add the return to overview button as the sole contents of toolbar
		m_WorkBar.removeAll();
		if (!automating) {
			m_WorkBar.add(m_MainButton);
			m_WorkBar.setVisible(true);
		}

		// determine the payer and issuer
		XMLVoucher voucher = m_IncPayment.getVoucher();
		String issuer = voucher.getIssuer();
		String payer = m_IncPayment.getPayer();

		// we'll need to pay tokens only if our safe is presently non-empty of
		// vouchers from the same Issuer as the incoming payment voucher

		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		vtm.selectIssuer(issuer);
		m_TokCost = m_TokensSpent = 0;
		boolean useTokens = true;
		if (vtm.getRowCount() > 0) {
			Hashtable<XMLVoucher, String> issVouchers = vtm.getIssuerVouchers();
			// check for vouchers from this Issuer
			useTokens = !issVouchers.isEmpty();
		}
		else {
			// no vouchers of any kind
			useTokens = false;
		}
		if (useTokens) {
			// base cost just to pick up the payment, plus 2 tokens for receipts
			m_TokCost = VoucherRequest.M_ReceiveCost;
		}

		// set instructions and token cost accordingly
		String usrTxt = null;
		if (useTokens) {
			// normal cost includes two tokens for receipt storage
			m_TokensSpent = m_TokCost;
			usrTxt = "The payment described has been made to you by "
						+ "another wallet. You must pick up the payment "
						+ "before it expires, or it will revert back to the "
						+ "payer.  Picking up the payment will validate the "
						+ "voucher and store it in your wallet.  It will also "
						+ "generate a receipt and send it back to the payer.  "
						+ "A copy of the receipt will be stored in your "
						+ "wallet for your records.  You may delete this receipt "
						+ "later if you wish.";
		}
		else {
			// pickup cost is zero but we still need 2 tokens for payer and
			// payee receipts
			m_TokensSpent = 2;
			usrTxt = "The payment described has been made to you by "
						+ "another wallet. You must pick up the payment "
						+ "before it expires, or it will revert back to the "
						+ "payer.  Picking up the payment will validate the "
						+ "voucher and store it in your wallet.  It will also "
						+ "generate a receipt and send it back to the payer.";
		}
		m_UserInstructs = usrTxt;
		String tTxt = "Pick up costs " + m_TokCost + " tokens";
		m_TokenCost.setText(tTxt);

		// set screen label
		buildScreenLabel("Incoming Payment");

		// verify sufficient tokens are present
		if (m_TokCost > 0) {
			// obtain latest list in case we had a timeout and tokens were used
			if (m_NeedTokenSync) {
				m_TabManager.queryTokens(null, true);
				m_NeedTokenSync = false;
			}
			TokenTableModel ttm = m_ParentHome.getTokenModel();
			m_PaymentTokens
				= ttm.getAvailableTokens(m_TokCost, vtm.getSelectedIssuer());
			if (m_PaymentTokens == null) {
				// Record ourselves with tab manager, so we return here after
				// the user has purchased tokens.  Then go purchase tokens.

				m_TabManager.setPendingTokens(this);
				m_TabManager.queueTransition(
												new Integer(WTabManager.M_PURCHASE_TOKENS));
				// avoid recursion
				Runnable getToks = new Runnable() {
					public void run() {
						m_TabManager.makeTransition();
					}
				};
				SwingUtilities.invokeLater(getToks);
				return;
			}
		}

		// provide action buttons in the button pane
		m_ButtonPane.removeAll();
		final JButton pickButton = new JButton(m_PickupAction);

		// if automating, just push our own button
		if (automating) {
			// do this in another thread to avoid recursion
			Runnable doPickup = new Runnable() {
				public void run() {
					m_WorkArea.setViewportView(null);
					ActionEvent trigger
						= new ActionEvent(pickButton, 0, "AUTOPICKUP");
					m_PickupAction.actionPerformed(trigger);
				}
			};
			SwingUtilities.invokeLater(doPickup);
			return;
		}

		// first, the button to skip over this item
		JButton skipButton = new JButton(m_SkipAction);
		skipButton.setToolTipText("Click here to skip to the next refresh item");
		skipButton.setFont(m_Plugin.M_ButtonFont);
		m_ButtonPane.add(Box.createHorizontalStrut(20));	 // spacer
		m_ButtonPane.add(skipButton);
		m_ButtonPane.add(Box.createHorizontalStrut(150));	 // spacer

		// second, the button to pickup this payment
		pickButton.setToolTipText("Click here to pick up this payment");
		pickButton.setFont(m_Plugin.M_ButtonFont);
		m_ButtonPane.add(pickButton);

		// place the details about the payment into a panel
		JPanel payDets = new JPanel(new GridBagLayout());
		payDets.setBorder(null);
		m_GBC.fill = GridBagConstraints.NONE;
		m_GBC.anchor = GridBagConstraints.NORTHWEST;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		payDets.add(Box.createVerticalGlue(), m_GBC);
		payDets.add(Box.createVerticalStrut(10), m_GBC);
		final Color labColor = UIManager.getColor("standardGray");

		// determine whether signature is valid
		VsSecrets sekrits = m_Plugin.getLoginSecrets();
		boolean validSig = m_IncPayment.validateSig(sekrits.getVPKey());
		if (!validSig) {
			showError("cannot pick up payment",
						"payment signature by VP is not valid!");
			pickButton.setEnabled(false);
		}

		// show the payer
		JLabel payerLab = new JLabel("This voucher is from:");
		payerLab.setFont(m_ParentHome.M_LabelFont);
		payerLab.setForeground(labColor);
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		payDets.add(payerLab, m_GBC);
		JLabel payerFld = new JLabel(payer, SwingConstants.TRAILING);
		payerFld.setFont(m_ParentHome.M_DataFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		payDets.add(payerFld, m_GBC);
		payDets.add(Box.createVerticalStrut(5), m_GBC);

		// display certain selected voucher details

		// amount, including units and asset
		StringBuilder amtStr = new StringBuilder(32);
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(0);
		df.setGroupingSize(3);
		double vAmt = voucher.getNPV();
		amtStr.append(df.format(vAmt) + " ");
		String dUnits = VPAssetConfig.getDisplayUnits(voucher.getAsset(),
																	 voucher.baseUnits());
		amtStr.append(dUnits + " of " + voucher.getAsset());
		JLabel amtLab = new JLabel("Amount:");
		amtLab.setFont(m_ParentHome.M_LabelFont);
		amtLab.setForeground(labColor);
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		payDets.add(amtLab, m_GBC);
		JLabel amount = new JLabel(amtStr.toString(), SwingConstants.TRAILING);
		amount.setFont(m_ParentHome.M_DataFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		payDets.add(amount, m_GBC);

		// obtain desired currency from preferences and also show value therein
		String prefCurrency = m_Plugin.getUserPreferences().getValueCurrency();
		AssetPriceCache apc = m_Plugin.getPricingModule();
		String lUnits = VPAssetConfig.getBasePriceLookupUnits(voucher.getAsset());
		Double price = apc.getRelativePrice(lUnits, prefCurrency);
		if (price == null) {
			price = new Double(1.0);
			prefCurrency = VPAssetConfig.getDisplayUnits(voucher.getAsset(),
																		voucher.getUnits());
		}
		df.setMaximumFractionDigits(2);
		df.setMinimumFractionDigits(2);
		StringBuilder fiatAmt = new StringBuilder(32);
		fiatAmt.append("( ");
		fiatAmt.append(df.format(vAmt * price.doubleValue()));
		fiatAmt.append(" " + prefCurrency + " )");
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		payDets.add(Box.createHorizontalStrut(50), m_GBC);
		JLabel fiat = new JLabel(fiatAmt.toString(), SwingConstants.TRAILING);
		fiat.setFont(m_ParentHome.M_DataItalFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		payDets.add(fiat, m_GBC);
		payDets.add(Box.createVerticalStrut(5), m_GBC);

		// add baggage fields if there are any
		TreeMap<String, String> bagFields = m_IncPayment.getBaggageList();
		if (!bagFields.isEmpty()) {
			for (String bagField : bagFields.keySet()) {
				// skip meta-field
				if (bagField.equals("field_names")) {
					continue;
				}
				String bagFld = bagField;
				String bagVal = bagFields.get(bagField);

				// parse off the leading "B" that was added to the field name,
				// unless the field is "Message"

				JLabel bagLabel = null;
				if (bagFld.equals("Message")) {
					bagLabel = new JLabel(bagFld + ": ");
				}
				else {
					bagLabel = new JLabel(bagFld.substring(1) + ": ");
				}
				bagLabel.setFont(m_ParentHome.M_LabelFont);
				bagLabel.setForeground(labColor);
				m_GBC.gridwidth = GridBagConstraints.RELATIVE;
				payDets.add(bagLabel, m_GBC);

				// undo anti-XML confusion replacements
				bagVal = bagVal.replaceAll(";lt;", "<");
				bagVal = bagVal.replaceAll(";gt;", ">");

				// show the value in a text area so we can get it all
				JTextArea bagValue = new JTextArea(1, 20);
				bagValue.setEditable(false);
				bagValue.setLineWrap(true);
				bagValue.setWrapStyleWord(true);
				bagValue.setBackground(m_Plugin.m_LabelBackColor);
				bagValue.setFont(m_ParentHome.M_DataFont);
				bagValue.setBorder(null);
				bagValue.setFocusable(false);
				bagValue.setText(bagVal);
				m_GBC.gridwidth = GridBagConstraints.REMAINDER;
				payDets.add(bagValue, m_GBC);
			}
			payDets.add(Box.createVerticalStrut(10), m_GBC);
		}
		payDets.add(Box.createVerticalStrut(10), m_GBC);

		// serial number
		JLabel serialLab = new JLabel("Voucher serial number:");
		serialLab.setFont(m_ParentHome.M_DetailFont);
		serialLab.setForeground(labColor);
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		payDets.add(serialLab, m_GBC);
		JLabel serialNum
			= new JLabel(voucher.getSerialNumber(), SwingConstants.TRAILING);
		serialNum.setFont(m_ParentHome.M_DetailFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		payDets.add(serialNum, m_GBC);
		payDets.add(Box.createVerticalStrut(5), m_GBC);

		// show expiration date
		JLabel expLab = new JLabel("You must pick up this payment before:  ");
		expLab.setFont(m_ParentHome.M_DetailFont);
		expLab.setForeground(labColor);
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		payDets.add(expLab, m_GBC);
		SimpleDateFormat sdf = new SimpleDateFormat(XMLVoucher.M_DateFmt);
		String expStr = sdf.format(new Date(m_IncPayment.getTTL()));
		JLabel expDate = new JLabel(expStr, SwingConstants.TRAILING);
		expDate.setFont(m_ParentHome.M_DetailFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		payDets.add(expDate, m_GBC);
		payDets.add(Box.createVerticalStrut(5), m_GBC);

		// obtain optional receipt message (in addition to baggage fields)
		m_DoingUndo = false;
		JLabel recMsgLab = new JLabel("Optional - enter message "
											+ "to include with payer's receipt:");
		recMsgLab.setLabelFor(m_ReceiptMsg);
		recMsgLab.setForeground(labColor);
		JPanel labelPane = new JPanel();
		labelPane.setLayout(new BoxLayout(labelPane, BoxLayout.X_AXIS));
		labelPane.add(recMsgLab);
		labelPane.add(Box.createHorizontalStrut(20));

		// reset text and char count
		m_ReceiptMsg.setText("");
		Integer mlen = new Integer(M_MaxMessage);
		m_CharCount.setText(mlen.toString());
		labelPane.add(m_CharCount);

		// put actual message entry into scroll area
		JPanel msgPane = new JPanel(new GridBagLayout());
		msgPane.add(labelPane, m_GBC);
		JScrollPane msgScroll = new JScrollPane(m_ReceiptMsg);
		msgPane.add(msgScroll, m_GBC);
		payDets.add(msgPane, m_GBC);
		payDets.add(Box.createVerticalGlue(), m_GBC);

		// install everything into the main WorkPane scroll area
		m_WorkArea.setViewportView(payDets);
	}


	/**
	 * private utility method to construct and send a payment pickup request
	 * @param vouch the voucher we're going to pick up and validate
	 * NB: this method will block pending the reply from the OFS
	 */
	private void sendPickupPayment(XMLVoucher vouch) {
		if (vouch == null || !vouch.isInit()) {
			Log.error("null input pickup voucher");
			return;
		}
		final String genericErr = "could not pick up payment";
		VscState state = m_Plugin.getVscState();
		if (!state.isLoggedIn()) {
			Log.error("Cannot pickup payments while not logged in!");
			showError(genericErr, "not logged in", "log in first");
			return;
		}
		VsSecrets secData = m_Plugin.getLoginSecrets();

		// build <voucherRequest/>
		VoucherRequest vReq = new VoucherRequest();
		vReq.setAction("validate");
		vReq.setSigningVS(secData.getVSnumber());
		vReq.setFolderHash(new String(secData.getVoucherIndex()));
		vReq.setFolderCap(secData.getVouchRWCap());
		HashMap<String, XMLVoucher> inpVouchers = vReq.getVouchers();
		inpVouchers.put("pending", vouch);
		vReq.setPayerVS(m_IncPayment.getPayer());
		vReq.setInitialized();

		// sign the voucher request with our private key
		if (!vReq.signDetails(secData.getPrivKey())) {
			Log.error("Unable to sign VR with VS privkey");
			showError(genericErr, "error signing pickup request");
			return;
		}

		// encrypt the details to the VP's key so the OFS can't read them
		if (!vReq.encryptDetails(secData.getVPKey(), false)) {
			Log.error("Unable to encrypt VR with VP pubkey");
			showError(genericErr, "error encrypting pickup request for VP");
			return;
		}

		// see whether we're using tokens (m_PaymentTokens will be empty if not)
		if (m_PaymentTokens != null && !m_PaymentTokens.isEmpty()) {
			// add the tokens for the validate, and commit them
			for (XMLToken tok : m_PaymentTokens) {
				vReq.addPaymentToken(tok);
			}
			m_ParentHome.getTokenModel().commitTokens(m_PaymentTokens);
		}

		// build message to send
		final XMPPConnection conn = m_Plugin.getOFSConnection();
		OFSMessage pickup = new OFSMessage();
		pickup.setType(IQ.Type.SET);
		pickup.setFrom(conn.getUser());
		pickup.setTo(conn.getServiceName());
		pickup.setPacketID("pickup_" + state.getNextOFSid());
		pickup.setOpcode("REQ_pickup_payment");
		pickup.setVoucher_publisher(secData.getPublisher());
		pickup.setVouchReq(vReq);

		// register a synchronous collector for the reply
		PacketIDFilter pFilter = new PacketIDFilter(pickup.getPacketID());
		PacketCollector repColl = conn.createPacketCollector(pFilter);

		// wait cursor and grey out buttons
		m_ParentHome.setCursor(M_WaitCursor);
		setButtonState(false);

		// send it to the server
		conn.sendPacket(pickup);

		// record outgoing packet
		state.setLastActivity(pickup);

		// wait for reply or timeout
		OFSMessage rep = null;
		IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
		m_ParentHome.setCursor(null);
		setButtonState(true);
		if (reply == null) {
			// got timeout; build error
			rep = new OFSMessage();
			rep.setOpcode("REP_payment_confirmed");
			rep.setFrom(pickup.getTo());
			rep.setTo(pickup.getFrom());
			rep.setType(IQ.Type.ERROR);
			rep.setPacketID(pickup.getPacketID());
			rep.setErrcode(HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
			rep.setErrmsg("timed out");
		}
		else {
			rep = (OFSMessage) reply;
		}
		repColl.cancel();

		// invoke callback with result (calls processPickupReply() below)
		m_Plugin.getOFSListener().processPacket(rep);
	}

	/**
	 * method to process a reply received from OFS to a payment pickup
	 * @param pickupRep the reply message from the OFS
	 */
	public void processPickupReply(OFSMessage pickupRep) {
		if (pickupRep == null
			|| !pickupRep.getOpcode().equalsIgnoreCase("REP_payment_confirmed"))
		{
			Log.error("Bad payment pickup reply received");
			return;
		}
		final String err = "could not process payment pickup";

		// see whether this reply is an error
		HousekeepingPane hkp = m_TabManager.getHousekeepingPane();
		if (pickupRep.getType() == IQ.Type.ERROR) {
			m_ParentHome.getTokenModel().uncommitTokens(m_PaymentTokens);
			m_PaymentTokens.clear();
			if (hkp.automating()) {
				// go to manual processing, to avoid endless recursion
				hkp.manualOverride();
			}
			int code = pickupRep.getErrcode();
			if (code == HttpURLConnection.HTTP_GATEWAY_TIMEOUT) {
				showError(err, "timeout " + code + ": " + pickupRep.getErrmsg(),
							"Wait a few minutes and try again.  If it fails, "
							+ "your original payment pickup may have completed.");
				m_NeedTokenSync = true;

			}
			else if (code == HttpURLConnection.HTTP_NOT_FOUND) {
				showError(err, "code " + code + ": " + pickupRep.getErrmsg(),
							"this usually indicates that you picked up this "
							+ "payment previously");
			}
			else {
				showError(err, "code " + code + ": " + pickupRep.getErrmsg());
			}

			/* If it's a 404, remove this item from the list so we don't repeat
			 * doing it endlessly until it vanishes off the DHT.
			 */
			if (code == HttpURLConnection.HTTP_NOT_FOUND) {
				markPendPayReadOnDHT();
			}
			return;
		}

		// decode/decrypt the <confirmation/>
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		final String remedy = "use Resync button to rebuild your voucher list";
		String decConf
			= EncodingUtils.getStrFromBase64PubkeyEnc(pickupRep.getConfirmation(),
																	secrets.getPrivKey());
		if (decConf == null || decConf.isEmpty()) {
			Log.error("Unable to decode <confirmation/> of "
						+ pickupRep.getOpcode());
			showError(err, "could not decode/decrypt confirmation details",
						remedy);
			m_PaymentTokens.clear();
			return;
		}

		// parse confirmation
		Element confElt = EncodingUtils.buildElement(decConf);
		if (confElt == null) {
			Log.error("Unable to parse <confirmation/> of "
						+ pickupRep.getOpcode());
			showError(err, "could not parse confirmation details", remedy);
			m_PaymentTokens.clear();
			return;
		}
		XMLVoucher newVouch = null;
		String vouchFile = null;
		try {
			// get new voucher
			Element vouchElt = confElt.element("voucher");
			if (vouchElt == null) {
				throw new DocumentException("Missing new voucher");
			}
			newVouch = XMLVoucher.getVoucherRep(vouchElt);

			// obtain the voucher's storage filename
			Element fileElt = confElt.element("voucherFile");
			if (fileElt == null) {
				throw new DocumentException("Missing new voucher filename");
			}
			vouchFile = fileElt.getTextTrim();

			// get VP signature and validate it
			Element sigElt = confElt.element("signature");
			if (sigElt == null) {
				throw new DocumentException("Missing VP signature");
			}
			String b64Sig = sigElt.getTextTrim();
			int sigIdx = decConf.lastIndexOf("<signature");
			String sigData = decConf.substring(0, sigIdx);
			Signature vpSig = null;
			try {
				vpSig = Signature.getInstance("SHA1withRSA");
				vpSig.initVerify(secrets.getVPKey());
				vpSig.update(sigData.getBytes());
				if (!vpSig.verify(StringUtils.decodeBase64(b64Sig))) {
					throw new DocumentException("Invalid VP signature on reply");
				}
			}
			catch (NoSuchAlgorithmException nsae) {
				// impossible
				throw new DocumentException("No such sig algorithm", nsae);
			}
			catch (InvalidKeyException ike) {
				throw new DocumentException("Invalid VP pubkey", ike);
			}
			catch (SignatureException se) {
				throw new DocumentException("Unable to validate payment pickup "
													+ "reply signature", se);
			}
		}
		catch (DocumentException de) {
			Log.error("Payment pickup confirmation parse error", de);
			showError(err, de.getMessage(), remedy);
			m_PaymentTokens.clear();
			return;
		}

		// add the new voucher to the voucher table model
		if (newVouch != null) {
			m_ParentHome.addVoucher(newVouch, vouchFile);
		}

		// record the payment tokens, if any, as spent
		if (!m_PaymentTokens.isEmpty()) {
			m_ParentHome.getTokenModel().spendTokens(m_PaymentTokens);
			m_PaymentTokens.clear();
		}

		// display warning if we have sync pending
		if (pickupRep.syncPending()) {
			showError("warning: folder updates were queued", "SDS timeout",
						"wait a few minutes and use Resync to see folder updates");
		}

		// step 3 - upload copy of receipt to our folder on the SDS
		m_ReceiptCopy = new XMLReceipt();
		m_ReceiptCopy.setQuantity(newVouch.getQuantity());
		m_ReceiptCopy.setUnits(newVouch.getUnits());
		m_ReceiptCopy.setAsset(newVouch.getAsset());
		m_ReceiptCopy.setPayee(secrets.getVSnumber());
		m_ReceiptCopy.setPayer(m_IncPayment.getPayer());
		m_ReceiptCopy.setVoucherSerial(
											m_IncPayment.getVoucher().getSerialNumber());
		Date now = new Date();
		m_ReceiptCopy.setTimestamp(now.getTime() / 1000L);		// seconds
		TreeMap<String, String> pmtBaggage = m_IncPayment.getBaggageList();
		if (!pmtBaggage.isEmpty()) {
			for (String key : pmtBaggage.keySet()) {
				String bagValue = pmtBaggage.get(key);
				if (key.equals("Message")) {
					// to make sure the original Message is not overwritten by
					// a Message which we entered, munge the field name

					m_ReceiptCopy.addBaggageField("BMessage", bagValue);
				}
				else if (!key.equals("field_names")) {
					m_ReceiptCopy.addBaggageField(key, bagValue);
				}
			}
		}

		// see whether we have an additional message to go with receipt
		String receiptMsg = m_ReceiptMsg.getText().trim();
		if (!receiptMsg.isEmpty()) {
			// add it to the baggage
			m_ReceiptCopy.addBaggageField("Message", receiptMsg);
		}
		m_ReceiptCopy.setInit();

		// now sign receipt
		if (!m_ReceiptCopy.signDetails(secrets.getPrivKey())) {
			Log.error("error signing receipt with private key");
			showError("could not generate receipt copy",
						"could not affix signature");
			return;
		}

		// step 4 -- generate receipt for payer and upload to DHT
		m_PayerReceipt = new XMLReceipt();
		m_PayerReceipt.setQuantity(newVouch.getQuantity());
		m_PayerReceipt.setUnits(newVouch.getUnits());
		m_PayerReceipt.setAsset(newVouch.getAsset());
		m_PayerReceipt.setPayee(secrets.getVSnumber());
		m_PayerReceipt.setPayer(m_IncPayment.getPayer());
		m_PayerReceipt.setVoucherSerial(
											m_IncPayment.getVoucher().getSerialNumber());
		m_PayerReceipt.setTimestamp(now.getTime() / 1000L);		// seconds
		if (!pmtBaggage.isEmpty()) {
			for (String key : pmtBaggage.keySet()) {
				String val = pmtBaggage.get(key);
				if (key.equals("Message")) {
					// to make sure the original Message is not overwritten by
					// a Message which we entered, munge the field name

					m_PayerReceipt.addBaggageField("BMessage", val);
				}
				else if (!key.equals("field_names")) {
					m_PayerReceipt.addBaggageField(key, val);
				}
			}
		}

		// see whether we have an additional message to go with receipt
		if (!receiptMsg.isEmpty()) {
			// add it to the baggage
			m_PayerReceipt.addBaggageField("Message", receiptMsg);
		}
		m_PayerReceipt.setInit();

		// now sign receipt
		if (!m_PayerReceipt.signDetails(secrets.getPrivKey())) {
			Log.error("error signing receipt with private key");
			showError("could not generate payer receipt",
						"could not affix signature");
			return;
		}

		// send our receipt copy to the SDS (synchronous)
		sendReceiptToSDS();

		// and send payer receipt to the DHT (asynch)
		if (!m_SuppressDisplay) {
			sendReceiptToDHT();
		}
	}

	/**
	 * private utility method to request that the SDS store a receipt record.
	 * implied param: m_ReceiptCopy contains the receipt to be stored
	 * NB: this method will block pending the SDS's reply
	 */
	private void sendReceiptToSDS() {
		if (m_ReceiptCopy == null || !m_ReceiptCopy.isInit()) {
			Log.error("Missing or uninitialized receipt record cannot be stored");
			return;
		}
		final String genErr = "cannot store receipt";

		VscState state = m_Plugin.getVscState();
		if (!state.isLoggedIn()) {
			Log.error("Cannot store receipts while not logged in!");
			showError(genErr, "not logged in", "please log in first");
			return;
		}

		// open a connection to the SDS if we don't have one
		XMPPConnection sdsConn = m_Plugin.getSDSListener().getSDSConnection();
		if (sdsConn == null) {
			sdsConn = m_Plugin.openSDS();
		}
		if (sdsConn == null) {
			// can't do anything
			Log.error("Cannot store receipt without SDS connection!");
			showError(genErr, "SDS unavailable", "try again later");
			return;
		}

		// make encrypted blob out of the record
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		String encRctRec = m_ReceiptCopy.getEncryptedBase64(secrets.getPubKey());
		if (encRctRec.isEmpty()) {
			Log.error("error encrypting XMLReceipt record with public key");
			showError(genErr, "encrypt failed");
			return;
		}

		// build message to send
		SDSMessage storeRct = new SDSMessage();
		storeRct.setType(IQ.Type.SET);
		storeRct.setFrom(sdsConn.getUser());
		storeRct.setTo(sdsConn.getServiceName());
		storeRct.setPacketID("rstore_" + state.getNextSDSid());
		storeRct.setOpcode("REQ_store_receipt");
		String rIdx = new String(secrets.getReceiptIndex());
		storeRct.setFolder_index(rIdx);
		String rwCap = new String(secrets.getReceiptRWCap());
		storeRct.setReadwritecap(rwCap);
		storeRct.setReceipt_blob(encRctRec);

		// add one token for payment
		TokenTableModel ttm = m_ParentHome.getTokenModel();
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		m_PaymentTokens = ttm.getAvailableTokens(VoucherRequest.M_ReceiptCost,
															  vtm.getSelectedIssuer());
		if (m_PaymentTokens == null || m_PaymentTokens.isEmpty()) {
			// vector to buy tokens, since we must have at least 1 voucher by now
			m_SuppressDisplay = true;
			m_TabManager.setPendingTokens(this);
			m_TabManager.queueTransition(
												new Integer(WTabManager.M_PURCHASE_TOKENS));
			m_TabManager.makeTransition();
			return;
		}
		storeRct.addPaymentToken(m_PaymentTokens.get(0).toXML());
		ttm.commitTokens(m_PaymentTokens);

		// wait cursor and grey out buttons
		m_ParentHome.setCursor(M_WaitCursor);
		setButtonState(false);

		// register synchronous reply collector
		PacketIDFilter pFilter = new PacketIDFilter(storeRct.getPacketID());
		PacketCollector repColl = sdsConn.createPacketCollector(pFilter);

		// send request to the SDS
		sdsConn.sendPacket(storeRct);
		//Log.debug(storeRct.getOpcode() + " sent with index: " + rIdx);

		// record outgoing packet
		state.setLastActivity(storeRct);

		// wait for reply or timeout
		SDSMessage rep = null;
		IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
		m_ParentHome.setCursor(null);
		setButtonState(true);
		if (reply == null) {
			// got timeout; build error
			rep = new SDSMessage();
			rep.setOpcode("REP_receipt_stored");
			rep.setFrom(storeRct.getTo());
			rep.setTo(storeRct.getFrom());
			rep.setPacketID(storeRct.getPacketID());
			rep.setType(IQ.Type.ERROR);
			rep.setErrcode(java.net.HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
			rep.setErrmsg("timed out");
		}
		else {
			rep = (SDSMessage) reply;
		}
		repColl.cancel();

		// process the result
		processRctStoreReply(rep);
	}

	/**
	 * process a reply received from SDS to a receipt store
	 * @param storeRep the reply message from the SDS
	 */
	private void processRctStoreReply(SDSMessage storeRep) {
		if (storeRep == null
			|| !storeRep.getOpcode().equalsIgnoreCase("REP_receipt_stored"))
		{
			Log.error("Bad receipt store reply received");
			return;
		}
		TokenTableModel ttm = m_ParentHome.getTokenModel();
		SpentTokenCache stc = m_Plugin.getTokenCache();
		String vsNo = m_Plugin.getLoginSecrets().getVSnumber();
		HousekeepingPane hkp = m_TabManager.getHousekeepingPane();

		// see whether this reply is an error
		if (storeRep.getType() == IQ.Type.ERROR) {
			/* If the error here is 402 (payment required) then the implication is
			 * that our token was no good, i.e. we are attempting to double-spend
			 * it somehow and the SDS has rejected it.  In this case we need to
			 * record the token as spent, get another one, and retry the store.
			 */
			if (storeRep.getErrcode() == HttpURLConnection.HTTP_PAYMENT_REQUIRED) {
				// purge the bad token and record with cache
				ttm.spendTokens(m_PaymentTokens);
				stc.recordToken(m_PaymentTokens.get(0), vsNo);
				m_PaymentTokens.clear();

				// recurse, in the same thread to preserve synchronicity
				sendReceiptToSDS();
				return;
			}

			// uncommit the token and display error message
			ttm.uncommitTokens(m_PaymentTokens);
			m_PaymentTokens.clear();
			showError("error storing receipt", "code " + storeRep.getErrcode()
						+ ": " + storeRep.getErrmsg());
			if (hkp.automating()) {
				/* About the only way this can happen is if we lost connectivity to
				 * the SDS.  Suggest going to manual processing as a precaution.
				 */
				showError("error storing receipt", "code " + storeRep.getErrcode()
							+ ": " + storeRep.getErrmsg(),
							"switch to manual processing of housekeeping items");
			}
			else {
				showError("error storing receipt", "code " + storeRep.getErrcode()
							+ ": " + storeRep.getErrmsg());
			}
			// either way, housekeeping stops here
			return;
		}

		// record token as spent, and report to spent cache
		ttm.spendTokens(m_PaymentTokens);
		stc.recordToken(m_PaymentTokens.get(0), vsNo);
		m_PaymentTokens.clear();

		// tell the receipt table about the new data and filename where stored
		if (m_ReceiptCopy != null) {
			hkp.getReceiptModel().addReceipt(m_ReceiptCopy,
														storeRep.getFilename());
		}
		m_ReceiptCopy = null;
	}

	/**
	 * private utility method to upload a receipt to the DHT.
	 * implied param: m_PayerReceipt contains the payer's receipt
	 */
	private void sendReceiptToDHT() {
		if (m_PayerReceipt == null || !m_PayerReceipt.isInit()) {
			Log.error("cannot send null or uninitialized receipt to DHT");
			return;
		}

		final String err = "cannot upload payer receipt";
		VscState state = m_Plugin.getVscState();
		if (!state.isLoggedIn()) {
			Log.error("Cannot upload receipts while not logged in!");
			showError(err, "not logged in", "please log in first");
			return;
		}

		// obtain the payer's public key from the PKS (blocking call)
		PublicKey payerKey
			= m_Plugin.getPKSListener().obtainVSpubkey(m_IncPayment.getPayer());
		if (payerKey == null) {
			showError(err, "could not obtain payer pubkey from PKS");
			return;
		}

		// open a connection to the DHT if we don't have one
		XMPPConnection dhtConn = m_Plugin.getDHTListener().getDHTConnection();
		if (dhtConn == null) {
			dhtConn = m_Plugin.openDHT();
		}
		if (dhtConn == null) {
			// can't do anything
			Log.error("Cannot upload receipts without DHT connection!");
			showError(err, "DHT unavailable", "try again later");
			return;
		}

		// encrypt the receipt for the payer
		String encReceipt = m_PayerReceipt.getEncryptedBase64(payerKey);
		if (encReceipt == null || encReceipt.isEmpty()) {
			Log.error("Unable to encrypt payer receipt");
			showError(err, "could not encrypt receipt for payer");
			return;
		}

		// compute hash of payer VS#
		MessageDigest md = null;
		try {
			md = MessageDigest.getInstance("SHA-1");
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("Can't build SHA-1 hash", nsae);
		}
		md.update(m_IncPayment.getPayer().getBytes());
		String hashKey = StringUtils.encodeBase64(md.digest(), false);

		// build message to send
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		DHTMessage uplRec = new DHTMessage();
		uplRec.setType(IQ.Type.SET);
		uplRec.setFrom(dhtConn.getUser());
		uplRec.setTo(dhtConn.getServiceName());
		uplRec.setPacketID("store_" + state.getNextDHTid());
		uplRec.setOpcode("REQ_store_receipt");
		uplRec.setReceipt_blob(encReceipt);
		uplRec.setRecipient_hash(hashKey);

		// add one token for payment
		TokenTableModel ttm = m_ParentHome.getTokenModel();
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		m_PaymentTokens = ttm.getAvailableTokens(VoucherRequest.M_ReceiptCost,
															  vtm.getSelectedIssuer());
		if (m_PaymentTokens == null || m_PaymentTokens.isEmpty()) {
			// This should never happen, because we check tokens beforehand.
			// However it could occur if we had the DHT reject some of our tokens.

			Log.error("No token available to store payer receipt record");
			showError(err, "no token available", "purchase tokens first");
			return;
		}
		uplRec.setToken(m_PaymentTokens.get(0));
		ttm.commitTokens(m_PaymentTokens);

		// wait cursor and grey out buttons
		m_ParentHome.setCursor(M_WaitCursor);
		setButtonState(false);

		// register async reply listener
		PacketIDFilter pFilter = new PacketIDFilter(uplRec.getPacketID());
		m_Plugin.getDHTListener().addFilter(uplRec.getPacketID(), pFilter);

		// send request to the DHT
		dhtConn.sendPacket(uplRec);

		// record outgoing packet
		state.setLastActivity(uplRec);

		// async reply arrives at recordReceiptUpload() below
	}

	/**
	 * handle the results of a receipt upload to the DHT
	 * @param dhtRep the message received by the DHT listener
	 */
	public void recordReceiptUpload(DHTMessage dhtRep) {
		if (dhtRep == null
			 || !dhtRep.getOpcode().equalsIgnoreCase("REP_receipt_stored"))
		{
			Log.error("null or improper opcode on DHT receipt upload reply");
			return;
		}
		setButtonState(true);
		TokenTableModel ttm = m_ParentHome.getTokenModel();
		SpentTokenCache stc = m_Plugin.getTokenCache();
		String vsNo = m_Plugin.getLoginSecrets().getVSnumber();

		// if this message is an error, just put up a message
		if (dhtRep.getType() == IQ.Type.ERROR) {
			/* If the error here is 402 (payment required) then the implication is
			 * that our token was no good, i.e. we are attempting to double-spend
			 * it somehow and the DHT has rejected it.  In this case we need to
			 * record the token as spent, get another one, and retry the store.
			 */
			if (dhtRep.getErrcode() == HttpURLConnection.HTTP_PAYMENT_REQUIRED) {
				// purge token and notify cache
				ttm.spendTokens(m_PaymentTokens);
				stc.recordToken(m_PaymentTokens.get(0), vsNo);
				m_PaymentTokens.clear();

				// recurse, in a different thread to avoid blocking an event thread
				Runnable retry = new Runnable() {
					public void run() {
						// NB: this can fail if we run out of tokens
						sendReceiptToDHT();
					}
				};
				new Thread(retry).start();
				return;
			}

			// uncommit the token and display error message
			ttm.uncommitTokens(m_PaymentTokens);
			m_PaymentTokens.clear();

			/* About the only way this can happen is if we lost connectivity to
			 * the DHT.  Suggest going to manual housekeeping.
			 */
			HousekeepingPane hkp = m_TabManager.getHousekeepingPane();
			if (hkp.automating()) {
				showError("unable to upload payer receipt to DHT",
							"code " + dhtRep.getErrcode() + ": " + dhtRep.getErrmsg(),
							"switch to manual processing of housekeeping items");
			}
			else {
				showError("unable to upload payer receipt to DHT",
							"code " + dhtRep.getErrcode() + ": " + dhtRep.getErrmsg());
			}

			// either way, housekeeping stops here
			return;
		}

		// record tokens as spent, and report to cache
		ttm.spendTokens(m_PaymentTokens);
		stc.recordToken(m_PaymentTokens.get(0), vsNo);
		m_PaymentTokens.clear();

		// go on to step 5
		markPendPayReadOnDHT();
		m_PayerReceipt = null;
	}

	/**
	 * utility method to perform step 5, marking the pending payment record
	 * read on the DHT.
	 * implied param: m_IncPayment contains the pendpay record to mark
	 */
	private void markPendPayReadOnDHT() {
		// tell housekeeping we're done with this pending payment record
		HousekeepingPane hkp = m_TabManager.getHousekeepingPane();
		hkp.discardPayment(m_IncPayment, m_TokensSpent);

		// clear so they don't click on Submit again by mistake
		m_ButtonPane.removeAll();

		Integer dhtId = m_IncPayment.getDHTid();
		// step 5 - mark the original pending payment record record read on DHT
		if (dhtId != null) {
			// mark this record read on the DHT
			m_ParentHome.setCursor(M_WaitCursor);
			m_TabManager.markReadOnDHT(dhtId, false);
			// NB: on reply to record mark, we'll go to housekeeping for next item
		}
		else {
			Log.error("No DHT id on pending payment record: "
						+ m_IncPayment.toXML());
			showError("could not mark pending payment record read on DHT",
						"no DHT id found");
		}
		m_IncPayment = null;
	}

	/**
	 * method to queue up a modification to the message text area from within
	 * its document listener
	 * @param newText the replacement text
	 */
	private void queueTextMod(String newText) {
		if (newText == null) {
			return;
		}

		// do not allow simultaneous queued mods
		if (m_DoingUndo) {
			return;
		}

		// create and start background modification object
		TextAreaMod modifier = new TextAreaMod(newText);
		m_DoingUndo = true;
		SwingUtilities.invokeLater(modifier);
	}

	// END methods
}
