/*
 * last modified---
 * 	11-23-16/jet/ add Resync button to rollover menu
 * 	07-15-15/jet/ use lookup units for relative prices
 * 	08-05-14/jet/ Spark plugin version
 * 	03-19-14/jet/ font spacing fixes for OpenJDK-7
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide a UI widget to represent the main menu (Overview) after login
 */

package org.silentvault.spark.ui.wallet;

import javax.swing.*;
import java.awt.event.*;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Font;
import java.awt.Component;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.filter.PacketIDFilter;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;

import java.util.Set;
import java.util.Hashtable;
import java.util.TreeSet;
import java.util.List;
import java.util.Date;
import java.util.Calendar;
import java.text.NumberFormat;
import java.text.DecimalFormat;


/**
 * Extends WorkPane to provide a screen for the safe's main menu (Overview).
 *
 * @author Kevin Wilkerson
 */
public final class MenuPane extends WorkPane {
	// BEGIN data members
	/**
	 * action to ping DHT for new activity
	 */
	private AbstractAction	m_ActivityAction;

	/**
	 * action to initiate a payment-related operation (new, outgoing)
	 */
	private AbstractAction	m_PaymentsAction;

	/**
	 * action to initiate a receipt-related operation (based on listed receipts)
	 */
	private AbstractAction	m_ReceiptsAction;

	/**
	 * action to display the voucher table for the selected Issuer in a popup
	 */
	private AbstractAction	m_ShowVouchersAction;

	/**
	 * action to resync wallet contents with the SDS
	 */
	private AbstractAction	m_ResyncAction;

	/**
	 * inner class for handling popup window closures
	 */
	private class PopupCloser extends WindowAdapter {
		public void windowClosing(WindowEvent we) {
			JFrame popup = (JFrame) we.getWindow();
			m_Plugin.deregisterPopup(popup);
			m_VoucherPopup = null;
		}
	}

	/**
	 * voucher popup display window
	 */
	private JFrame				m_VoucherPopup;

	/**
	 * menu toolbar
	 */
	private JToolBar			m_MenuToolbar;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the HomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public MenuPane(WalletClient plugin, HomePane home, WTabManager mgr) {
		super(plugin, home, mgr);

		// create toolbar actions
		setActivityAction();
		setPaymentsAction();
		setReceiptsAction();
		setShowVouchersAction();
		setResyncAction();

		// populate the toolbar and heading panel, and make visible
		final Font buttonFont = new Font(Font.SANS_SERIF, Font.BOLD, 13);
		JButton payments = new JButton(m_PaymentsAction);
		payments.setToolTipText("Make a payment to another wallet");
		payments.setFont(buttonFont);
		payments.setRolloverEnabled(true);

		// button to download/refresh receipts
		JButton receipts = new JButton(m_ReceiptsAction);
		receipts.setToolTipText("View receipts (payment history)");
		receipts.setFont(buttonFont);
		receipts.setRolloverEnabled(true);

		// button to query for new message activity
		JButton activity = new JButton(m_ActivityAction);
		activity.setToolTipText("Check for new incoming items");
		activity.setFont(buttonFont);
		activity.setRolloverEnabled(true);

		// button to resync with the SDS
		JButton resync = new JButton(m_ResyncAction);
		resync.setToolTipText("Resynchronize wallet with network");
		resync.setFont(buttonFont);
		resync.setRolloverEnabled(true);

		m_MenuToolbar = new JToolBar(SwingConstants.HORIZONTAL);
		final Dimension sepSz = new Dimension(20, 30);
		m_MenuToolbar.setFloatable(false);
		m_MenuToolbar.setRollover(true);
		m_MenuToolbar.setBorderPainted(false);
		m_MenuToolbar.setOpaque(true);
		m_MenuToolbar.add(payments);
		m_MenuToolbar.addSeparator(sepSz);
		m_MenuToolbar.add(receipts);
		m_MenuToolbar.addSeparator(sepSz);
		m_MenuToolbar.add(activity);
		m_MenuToolbar.addSeparator(sepSz);
		m_MenuToolbar.add(resync);
		m_MenuToolbar.add(Box.createHorizontalStrut(200));

		// set the user instructions
		m_UserInstructs = "Select an Issuer and click the Show Vouchers button "
									+ "below, to see the additional details of "
									+ "the vouchers from that Issuer.<br/><br/>"
									+ "To begin an action, click on the rollover menu "
									+ "above.";
		m_WorkArea.setBorder(null);
	}

	/**
	 * utility method to establish the Action which can handle activity updates
	 */
	private void setActivityAction() {
		m_ActivityAction = new AbstractAction("Refresh") {
			public void actionPerformed(ActionEvent actionEvent) {
				closeVoucherPopup();

				// set wait cursor
				m_ParentHome.setCursor(M_WaitCursor);

				// ask Housekeeping to query the DHT for an updated records list
				HousekeepingPane hkp = m_TabManager.getHousekeepingPane();
				hkp.refreshDHT(false);
				hkp.initProcessing();

				// NB: when the DHT's reply is received we'll switch to the
				// 	 HousekeepingPane
			}
		};
	}

	/**
	 * utility method to establish the Action which can initiate a payment
	 * related operation
	 */
	private void setPaymentsAction() {
		m_PaymentsAction = new AbstractAction("Payment") {
			public void actionPerformed(ActionEvent actionEvent) {
				closeVoucherPopup();

				// switch to payments screen
				m_TabManager.queueTransition(new Integer(WTabManager.M_PAYMENTS1));
				m_TabManager.makeTransition();
			}
		};
	}

	/**
	 * utility method to establish the Action which can initiate a receipts
	 * related operation
	 */
	private void setReceiptsAction() {
		m_ReceiptsAction = new AbstractAction("Receipts") {
			public void actionPerformed(ActionEvent actionEvent) {
				closeVoucherPopup();

				// set wait cursor
				m_ParentHome.setCursor(M_WaitCursor);

				// request receipt records from the SDS -- callback will populate
				// the table using a call to updateReceiptRecords()
				queryReceipts();

				// switch to receipt listing screen
				m_TabManager.queueTransition(
													new Integer(WTabManager.M_RECEIPT_LIST));
				m_TabManager.makeTransition();
			}
		};
	}

	/**
	 * utility method to resynchronize wallet contents (vouchers, tokens, and
	 * pending payments) with the SDS
	 */
	private void setResyncAction() {
		m_ResyncAction = new AbstractAction("Resync") {
			public void actionPerformed(ActionEvent actionEvent) {
				closeVoucherPopup();

				// first, sync up any tokens we've spent
				SpentTokenCache stc = m_Plugin.getTokenCache();
				stc.purgeSpentTokens();

				// next, query all safe contents
				if (!m_TabManager.querySafeContents(true)) {
					Log.error("Error doing resync with SDS (vouchers, tokens)");
				}
				else {
					// next, query payment records
					HousekeepingPane hkp = m_TabManager.getHousekeepingPane();
					m_ParentHome.setCursor(M_WaitCursor);
					SDSMessage pmtRep = hkp.queryPayments(true);
					m_ParentHome.setCursor(null);
					if (pmtRep != null) {
						hkp.recordPayments(pmtRep.getPayment_list());
						hkp.updatePaymentRecords();
						m_TabManager.showInfo("Wallet contents resynchronized with "
													+ "network");
					}
					else {
						Log.error("Error doing resync with SDS (pend payments)");
					}
				}
			}
		};
	}

	/**
	 * utility method to define the action which can pop up the voucher table
	 */
	private void setShowVouchersAction() {
		m_ShowVouchersAction = new AbstractAction("Show Vouchers") {
			public void actionPerformed(ActionEvent actionEvent) {
				// don't create a second one
				if (m_VoucherPopup != null) {
					return;
				}

				// create a popup window which displays the voucher table
				VoucherTableModel vtm = m_ParentHome.getVoucherModel();
				m_VoucherPopup = new JFrame(vtm.getSelectedIssuer() + " Vouchers");
				m_VoucherPopup.setDefaultCloseOperation(
															WindowConstants.DISPOSE_ON_CLOSE);
				m_VoucherPopup.setSize(new Dimension(450, 300));
				m_VoucherPopup.addWindowListener(new PopupCloser());
				JScrollPane sp = new JScrollPane(m_ParentHome.getVoucherTable());
				sp.setBorder(null);
				m_VoucherPopup.setContentPane(sp);
				m_VoucherPopup.setLocationRelativeTo(m_TabManager);
				m_VoucherPopup.setVisible(true);

				// register this popup with client so it can be closed on logout
				m_Plugin.registerPopup(m_VoucherPopup);
			}
		};
	}

	/**
	 * private utility method to request a list of receipts from the SDS
	 * (reply callback will invoke updateReceiptRecords())
	 */
	private void queryReceipts() {
		VscState state = m_Plugin.getVscState();
		if (!state.isLoggedIn()) {
			Log.error("Cannot list receipts while not logged in!");
			showError("cannot list receipts", "not logged in", "log in first");
			return;
		}

		// open a connection to the SDS if we don't have one
		XMPPConnection sdsConn = m_Plugin.getSDSListener().getSDSConnection();
		if (sdsConn == null) {
			sdsConn = m_Plugin.openSDS();
		}
		if (sdsConn == null) {
			// can't do anything
			Log.error("Cannot list receipts without SDS connection!");
			showError("cannot list receipts", "SDS unavailable",
						"try again later");
			return;
		}

		// build message to send
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		SDSMessage getReceipts = new SDSMessage();
		getReceipts.setType(IQ.Type.GET);
		getReceipts.setFrom(sdsConn.getUser());
		getReceipts.setTo(sdsConn.getServiceName());
		getReceipts.setPacketID("rlist_" + state.getNextSDSid());
		getReceipts.setOpcode("REQ_list_receipts");
		String rIdx = new String(secrets.getReceiptIndex());
		getReceipts.setFolder_index(rIdx); 
		String rCap = new String(secrets.getReceiptRWCap());
		getReceipts.setReadcap(rCap);

		// register async reply listener
		PacketIDFilter pFilter = new PacketIDFilter(getReceipts.getPacketID());
		m_Plugin.getSDSListener().addFilter(getReceipts.getPacketID(), pFilter);

		// send request to the SDS
		sdsConn.sendPacket(getReceipts); 
		//Log.debug(getReceipts.getOpcode() + " sent with index: " + rIdx);

		// record outgoing packet
		state.setLastActivity(getReceipts);
	}

	// method required to extend WorkPane
	/**
	 * method to prepare display for this sub-class (called by HomePane)
	 */
	public void prepDisplay() {
		// install the menu toolbar
		m_WorkBar.removeAll();
		m_WorkBar.add(Box.createHorizontalGlue());
		m_WorkBar.add(m_MenuToolbar);
		m_WorkBar.add(Box.createHorizontalGlue());
		m_WorkBar.setVisible(true);

		// add heading
		buildValueHeader();

		// set screen label
		buildScreenLabel("Wallet Overview");

		// provide Show Vouchers button in the button pane
		m_ButtonPane.removeAll();
		JButton showButton = new JButton(m_ShowVouchersAction);
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		showButton.setToolTipText("<html>Click here to show all<br/>vouchers "
											+ "from selected Issuer</html>");
		showButton.setFont(m_Plugin.M_ButtonFont);
		m_ButtonPane.add(Box.createHorizontalStrut(310));		// spacer
		m_ButtonPane.add(showButton);

		// get a list of all Issuers for which we have vouchers
		vtm.setSelectMode(VoucherTableModel.M_SelectNone);
		vtm.deselectAll();
		TreeSet<String> issuers = new TreeSet<String>();
		Hashtable<XMLVoucher, String> vouchList = vtm.getVoucherList();
		for (XMLVoucher vouch : vouchList.keySet()) {
			issuers.add(vouch.getIssuer());
		}
		// in case they have tokens but no vouchers, also include token Issuers
		TokenTableModel ttm = m_ParentHome.getTokenModel();
		List<XMLToken> allToks = ttm.getAllTokens();
		for (XMLToken tok : allToks) {
			issuers.add(tok.getIssuer());
		}

		// prepare a screen for showing per-Issuer information in a quasi-table
		JPanel sumPane = new JPanel(new GridBagLayout());
		sumPane.setBorder(null);
		m_GBC.fill = GridBagConstraints.NONE;
		m_GBC.anchor = GridBagConstraints.NORTH;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_GBC.gridheight = (issuers.size() * 2) + 5;
		sumPane.add(Box.createVerticalGlue(), m_GBC);
		sumPane.add(Box.createVerticalStrut(20), m_GBC);
		final Color labColor = UIManager.getColor("standardGray");

		// title row
		JPanel titlePane = new JPanel(new GridLayout(1, 6));
		titlePane.setBorder(null);
		final Dimension rowSize = new Dimension(470, 15);
		titlePane.setPreferredSize(rowSize);
		JLabel issLab = new JLabel("Issuer", SwingConstants.CENTER);
		issLab.setFont(m_ParentHome.M_LabelFont);
		issLab.setForeground(labColor);
		issLab.setToolTipText("The voucher Issuer");
		titlePane.add(issLab);

		JLabel vCntLab = new JLabel("Vouchers", SwingConstants.CENTER);
		vCntLab.setFont(m_ParentHome.M_LabelFont);
		vCntLab.setForeground(labColor);
		vCntLab.setToolTipText("The number of vouchers from this Issuer");
		titlePane.add(vCntLab);

		JLabel valLab = new JLabel("Value", SwingConstants.CENTER);
		valLab.setFont(m_ParentHome.M_LabelFont);
		valLab.setForeground(labColor);
		valLab.setToolTipText("The total value of the vouchers, in asset units");
		titlePane.add(valLab);

		JLabel equivLab = new JLabel("Equivalent", SwingConstants.CENTER);
		equivLab.setFont(m_ParentHome.M_LabelFont);
		equivLab.setForeground(labColor);
		equivLab.setToolTipText("The equivalent in your preferred currency");
		titlePane.add(equivLab);

		JLabel expLab = new JLabel("Exp", SwingConstants.CENTER);
		expLab.setFont(m_ParentHome.M_LabelFont);
		expLab.setForeground(labColor);
		expLab.setToolTipText("The closest voucher expiration, in days");
		titlePane.add(expLab);

		JLabel tokLab = new JLabel("Tokens", SwingConstants.CENTER);
		tokLab.setFont(m_ParentHome.M_LabelFont);
		tokLab.setForeground(labColor);
		tokLab.setToolTipText("The number of tokens from this Issuer");
		titlePane.add(tokLab);
		sumPane.add(titlePane, m_GBC);
		sumPane.add(Box.createVerticalStrut(10), m_GBC);

		// handle special case of safe with no vouchers or tokens
		if (issuers.isEmpty()) {
			String emptyTxt = null;
			// if we don't have VS contents yet, do not display scary message
			if (m_TabManager != null && m_TabManager.getLoginInitState()) {
				emptyTxt = "Downloading wallet contents, please wait";
			}
			else {
				emptyTxt = "Your wallet is presently empty";
			}
			JLabel emptyLab = new JLabel(emptyTxt);
			emptyLab.setFont(m_ParentHome.M_HeaderFont);
			sumPane.add(Box.createVerticalStrut(10), m_GBC);
			sumPane.add(emptyLab, m_GBC);
			sumPane.add(Box.createVerticalStrut(150), m_GBC);
			m_WorkArea.setViewportView(sumPane);
			showButton.setEnabled(false);
			showButton.setToolTipText("There are no vouchers to display");
			return;
		}

		// remember the currently selected Issuer so we can reset it
		String selectedIss = vtm.getSelectedIssuer();
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setGroupingSize(0);
		String prefCurrency = null;
		for (String iss : issuers) {
			prefCurrency = m_Plugin.getUserPreferences().getValueCurrency();
			JPanel rowPane = new JPanel(new GridLayout(1, 6));
			rowPane.setBorder(null);
			rowPane.setPreferredSize(rowSize);

			JLabel iLab = new JLabel(iss, SwingConstants.CENTER);
			iLab.setFont(m_ParentHome.M_DataFont);
			iLab.setToolTipText("Issuer of vouchers and tokens");
			rowPane.add(iLab);

			vtm.selectIssuer(iss);
			Hashtable<XMLVoucher, String> issVouchers = vtm.getIssuerVouchers();
			Integer vCnt = new Integer(issVouchers.size());
			String asset = "";
			if (!issVouchers.isEmpty()) {
				// get the first voucher in the list
				XMLVoucher first = vtm.getFirstVoucher();
				asset = first.getAsset();
			}
			JLabel vLab = new JLabel(vCnt.toString(), SwingConstants.CENTER);
			vLab.setFont(m_ParentHome.M_DataFont);
			vLab.setToolTipText("Number of vouchers from " + iss);
			rowPane.add(vLab);

			// this total will be in base units for the asset
			double val = vtm.getVoucherSum();
			df.setMaximumFractionDigits(XMLVoucher.M_Precision);
			df.setMinimumFractionDigits(1);
			JLabel gauLab = new JLabel(df.format(val), SwingConstants.CENTER);
			gauLab.setFont(m_ParentHome.M_DataFont);
			String dispUnits = "asset units";
			if (!asset.isEmpty()) {
				dispUnits = XMLVoucher.getDisplayUnits(asset);
			}
			gauLab.setToolTipText("Voucher total in " + dispUnits);
			rowPane.add(gauLab);

			AssetPriceCache apc = m_Plugin.getPricingModule();
			Double assetPrice = null;
			if (vCnt.intValue() > 0) {
				String lookUnits = VPAssetConfig.getBasePriceLookupUnits(asset);
				assetPrice = apc.getRelativePrice(lookUnits, prefCurrency);
			}
			if (assetPrice == null) {
				assetPrice = new Double(1.0);
				prefCurrency = dispUnits;
			}
			df.setMaximumFractionDigits(2);
			df.setMinimumFractionDigits(2);
			JLabel eqLab = new JLabel(df.format(val * assetPrice.doubleValue()),
												SwingConstants.CENTER);
			eqLab.setFont(m_ParentHome.M_DataItalFont);
			eqLab.setToolTipText("Voucher total in " + prefCurrency);
			rowPane.add(eqLab);

			// determine the earliest expiration day of any voucher in the set
			Calendar cal = Calendar.getInstance();
			int offGMTmillis = cal.get(Calendar.ZONE_OFFSET);
			int offDSTmillis = cal.get(Calendar.DST_OFFSET);
			Date now = new Date();
			Date earliest = null;
			for (XMLVoucher vouch : issVouchers.keySet()) {
				Date voucherExp = vouch.getExpiration();
				// adjust for timezone
				Date vExp
					= new Date(voucherExp.getTime() + offGMTmillis + offDSTmillis);
				if (earliest == null || earliest.after(vExp)) {
					earliest = voucherExp;
				}
			}
			JLabel exLab = new JLabel();
			exLab.setHorizontalAlignment(SwingConstants.CENTER);
			exLab.setFont(m_ParentHome.M_DataFont);
			if (earliest == null) {
				// must have been no vouchers
				exLab.setText("N/A");
			}
			else {
				long secs = (earliest.getTime() - now.getTime()) / 1000L;
				Long days = new Long(secs / 86400L);
				Date thirtyDaysHence = new Date(now.getTime() + 2592000000L);
				Date tenDaysHence = new Date(now.getTime() + 864000000L);
				exLab.setText(days.toString());
				if (earliest.before(now)) {
					exLab.setBackground(Color.RED);
					exLab.setForeground(Color.DARK_GRAY);
					exLab.setOpaque(true);
					exLab.setToolTipText("You have an expired voucher!");
				}
				else if (earliest.before(tenDaysHence)) {
					exLab.setBackground(new Color(0xFEA98D));
					exLab.setOpaque(true);
					exLab.setToolTipText("<html>You have a voucher expiring in less"
												+ "<br/>than ten days, please revalidate "
												+ "now</html>");
				}
				else if (earliest.before(thirtyDaysHence)) {
					exLab.setBackground(new Color(0xFEE26F));
					exLab.setOpaque(true);
					exLab.setToolTipText("<html>You have a voucher expiring in less"
												+ "<br/>than thirty days, please revalidate"
												+ " now</html>");
				}
				else {
					exLab.setToolTipText("Days to your earliest voucher expiration");
				}
			}
			rowPane.add(exLab);

			Integer tokCnt = new Integer(ttm.getAvailableCount(iss));
			JLabel tLab = new JLabel(tokCnt.toString(), SwingConstants.CENTER);
			tLab.setFont(m_ParentHome.M_DataFont);
			tLab.setToolTipText("Number of tokens from " + iss);
			rowPane.add(tLab);

			sumPane.add(rowPane, m_GBC);
			sumPane.add(Box.createVerticalStrut(10), m_GBC);
		}
		vtm.selectIssuer(selectedIss);

		// provide some trailing space if there's less than 8 Issuers listed
		int rowSpace = 15 * (8 - issuers.size());
		sumPane.add(Box.createVerticalStrut(rowSpace), m_GBC);
		sumPane.add(Box.createVerticalGlue(), m_GBC);
		m_WorkArea.setViewportView(sumPane);
	}

	/**
	 * close the voucher table popup if it exists
	 */
	public void closeVoucherPopup() {
		// close any popup
		if (m_VoucherPopup != null) {
			m_Plugin.deregisterPopup(m_VoucherPopup);
			m_VoucherPopup.dispose();
			m_VoucherPopup = null;
		}
	}

	/**
	 * deactivate the menu toolbar items
	 */
	public void disableMenu() {
		int cnt = m_MenuToolbar.getComponentCount();
		for (int iii = 0; iii < cnt; iii++) {
			Component comp = m_MenuToolbar.getComponentAtIndex(iii);
			if (comp instanceof JToolBar.Separator) {
				// skip over it
				continue;
			}
			comp.setEnabled(false);
		}
	}

	/**
	 * (re)activate the menu toolbar items
	 */
	public void enableMenu() {
		int cnt = m_MenuToolbar.getComponentCount();
		for (int iii = 0; iii < cnt; iii++) {
			Component comp = m_MenuToolbar.getComponentAtIndex(iii);
			if (comp instanceof JToolBar.Separator) {
				// skip over it
				continue;
			}
			comp.setEnabled(true);
		}
	}

	// END methods
}
