/*
 * last modified---
 * 	08-15-18/kbw/ explicitly cancel PacketCollectors after use
 * 	08-05-14/jet/ Spark plugin version
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide a UI widget to handle performing merges on vouchers
 */

package org.silentvault.spark.ui.wallet;

import javax.swing.*;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.event.ActionEvent;
import java.awt.Color;
import java.awt.Font;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.filter.PacketIDFilter;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;

import java.util.List;
import java.util.Hashtable;
import java.util.ArrayList;
import java.util.HashMap;
import java.net.HttpURLConnection;


/**
 * Extends WorkPane to provide a screen for doing voucher merge operations.
 *
 * @author Kevin Wilkerson
 */
public final class MergePane extends WorkPane {
	// BEGIN data members
	/**
	 * action to merge 2 or more vouchers
	 */
	private AbstractAction	m_MergeAction;

	/**
	 * action button to merge vouchers
	 */
	private JButton			m_MergeButton;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the HomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public MergePane(WalletClient plugin, HomePane home, WTabManager mgr) {
		super(plugin, home, mgr);
		m_HeadingPane.setLayout(new GridLayout(1, 3));

		// provide user instructions
		String mergeTxt = "Check the boxes for at least two vouchers (max 25).  "
								 + "The selected vouchers will be added together to "
								 + "form one new voucher.<br/><br/>"
								 + "Note: Merging is not usually necessary.  Vouchers "
								 + "are merged automatically when multiple vouchers "
								 + "are selected to make a payment.";
		m_UserInstructs = mergeTxt;

		// set token cost
		m_TokenCost.setText("Merge uses " + VoucherRequest.M_MergeCost
								+ " tokens");

		// define button actions
		setMergeAction();
		m_MergeButton = new JButton(m_MergeAction);
		m_MergeButton.setFont(m_Plugin.M_ButtonFont);
	}

	/**
	 * utility method to establish the Action which can process a click on the
	 * Merge voucher ops button.
	 * NB: this method will block waiting on the OFS's reply
	 */
	private void setMergeAction() {
		m_MergeAction = new AbstractAction("Merge Now") {
			public void actionPerformed(ActionEvent actionEvent) {
				// we need at more than one row selected in the voucher table
				VoucherTableModel vtm = m_ParentHome.getVoucherModel();
				ArrayList<XMLVoucher> selVouchers = vtm.getSelectedVouchers();
				if (selVouchers.size() <= 1) {
					showError("no vouchers for merge", "less than two selected",
								"please select two or more vouchers to merge");
					return;
				}

				// ensure we have enough tokens
				TokenTableModel ttm = m_ParentHome.getTokenModel();
				ArrayList<XMLToken> safeTokens = ttm.getAvailableTokens(
																	VoucherRequest.M_MergeCost,
																	vtm.getSelectedIssuer());
				if (safeTokens == null) {
					// Record ourselves with tab manager, so we return here after
					// the user has purchased tokens.  Then go purchase tokens.

					m_TabManager.setPendingTokens(m_TabManager.getMergePane());
					m_TabManager.queueTransition(
												new Integer(WTabManager.M_PURCHASE_TOKENS));
					m_TabManager.makeTransition();
					return;
				}

				// send the message to the OFS
				final XMPPConnection conn = m_Plugin.getOFSConnection();
				VscState stData = m_Plugin.getVscState();
				VsSecrets secData = m_Plugin.getLoginSecrets();

				// build message to send
				OFSMessage merge = new OFSMessage();
				merge.setType(IQ.Type.SET);
				merge.setFrom(conn.getUser());
				merge.setTo(conn.getServiceName());
				merge.setPacketID("valid_" + stData.getNextOFSid());
				merge.setOpcode("REQ_revalidate");
				merge.setVoucher_publisher(secData.getPublisher());

				// build voucher request and add it to message
				VoucherRequest vReq = new VoucherRequest();
				vReq.setAction("merge");
				vReq.setSigningVS(secData.getVSnumber());
				vReq.setFolderHash(new String(secData.getVoucherIndex()));
				vReq.setFolderCap(secData.getVouchRWCap());
				merge.setVouchReq(vReq);

				// obtain the selected voucher(s) and add them to voucherRequest
				Hashtable<XMLVoucher, String> vouchers = vtm.getVoucherList();
				HashMap<String, XMLVoucher> inpVouchers = vReq.getVouchers();
				for (XMLVoucher voucher : selVouchers) {
					String sdsFile = vouchers.get(voucher);
					inpVouchers.put(sdsFile, voucher);
				}
				vReq.setInitialized();
				// NB: the <output/> value is implicitly the sum of the inputs

				// supply payment tokens (merge costs 12)
				for (int iii = 0; iii < VoucherRequest.M_MergeCost; iii++) {
					vReq.addPaymentToken(safeTokens.get(iii));
				}

				// sign the voucher request with our private key
				if (!vReq.signDetails(secData.getPrivKey())) {
					Log.error("Unable to sign VR with VS privkey");
					showError("could not sign voucherRequest!");
					return;
				}

				// encrypt the details to the VP's key so the OFS can't read them
				if (!vReq.encryptDetails(secData.getVPKey(), true)) {
					Log.error("Unable to encrypt VR with VP pubkey");
					showError("could not encrypt voucherRequest for VP!");
					return;
				}

				// mark these vouchers and tokens as committed to an operation
				vtm.commitVouchers(selVouchers);
				ttm.commitTokens(safeTokens);
				vtm.deselectAll();

				// register a collector for the reply
				PacketIDFilter pFilter = new PacketIDFilter(merge.getPacketID());
				PacketCollector repColl = conn.createPacketCollector(pFilter);

				// send it to the server
				m_ParentHome.setCursor(M_WaitCursor);
				setButtonState(false);
				conn.sendPacket(merge);

				// record outgoing packet
				stData.setLastActivity(merge);

				// wait for reply or timeout
				OFSMessage rep = null;
				IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
				m_ParentHome.setCursor(null);
				setButtonState(true);
				if (reply == null) {
					// got timeout; build error
					rep = new OFSMessage();
					rep.setOpcode("REP_validation");
					rep.setFrom(merge.getTo());
					rep.setTo(merge.getFrom());
					rep.setPacketID(merge.getPacketID());
					rep.setType(IQ.Type.ERROR);
					rep.setErrcode(HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
					rep.setErrmsg("timed out");
				}
				else {
					rep = (OFSMessage) reply;
				}
				repColl.cancel();

				// invoke callback with result
				m_Plugin.getOFSListener().processPacket(rep);
			}
		};
	}

	// method required to extend WorkPane
	/**
	 * method to prepare display for this sub-class
	 */
	public void prepDisplay() {
		// set heading pane
		buildValueHeader();

		// add the return to overview button as the sole contents of toolbar
		m_WorkBar.removeAll();
		m_WorkBar.add(m_MainButton);
		m_WorkBar.setVisible(true);

		// set screen label
		buildScreenLabel("Merge Vouchers");

		// provide action buttons in the button pane
		m_ButtonPane.removeAll();

		// first, the button to validate the selected voucher
		m_MergeButton.setToolTipText("Select two or more vouchers to merge");
		m_MergeButton.setEnabled(false);
		m_ButtonPane.add(Box.createHorizontalStrut(335));	// spacer
		m_ButtonPane.add(m_MergeButton);

		// place the voucher table into work area, set for checkbox select mode
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		vtm.deselectAll();
		vtm.setSelectMode(VoucherTableModel.M_SelectMulti);
		JTable vt = m_ParentHome.getVoucherTable();
		m_WorkArea.setViewportView(vt);
	}

	/**
	 * method to receive notification of a change in voucher selection
	 */
	public void voucherSelectChanged() {
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		ArrayList<XMLVoucher> selVouchers = vtm.getSelectedVouchers();
		if (vtm.getSelectMode() != VoucherTableModel.M_SelectMulti) {
			// not meant for us; do nothing
			return;
		}

		// impose count limit because of processing impact on VP
		int selCount = selVouchers.size();
		if (selCount > 25) {
			showError("too many vouchers selected for merge", "limit is 25",
						"perform multiple merge operations");
			XMLVoucher unchecked = selVouchers.remove(selCount - 1);
			vtm.deselectVoucher(unchecked);
			return;
		}

		if (selCount > 1) {
			// make the button active and change tooltip
			m_MergeButton.setToolTipText("Click here to merge the selected "
													+ "vouchers");
			m_MergeButton.setEnabled(true);
		}
		else {
			m_MergeButton.setToolTipText("Select two or more vouchers to merge");
			m_MergeButton.setEnabled(false);
		}
	}

	// END methods
}
