/*
 * last modified---
 * 	08-15-18/kbw/ explicitly cancel PacketCollectors after use
 * 	01-03-18/kbw/ ensure that payment amount states base units
 * 	11-24-16/jet/ add code to prevent spending all vouchers from an Issuer
 * 					  and leaving too few tokens to receive a payment or receipt;
 * 					  make voucher amount total copyable by using a JTextField
 * 	08-22-15/jet/ fix race condition by not setting m_IssuerSel to issuer of
 * 					  decoded WebMessage; instead set VoucherTableModel to issuer
 * 	07-15-15/jet/ broaden currency code selector to allow for asset:unit size;
 * 					  use VPAssetConfig.getPriceLookupUnits() to determine values
 * 	08-05-14/jet/ Spark plugin version
 * 	03-28-14/jet/ handle selected Issuer changes during a payment
 * 	03-19-14/jet/ font spacing fixes for OpenJDK-7
 * 	09-11-13/jet/ add m_LatestConfirmation and its methods; revise the look
 * 					  of the first payment screen when doing a SVX/SVM payment
 * 	07-17-13/jet/ add setSelectedSiteMsg()
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide a UI widget to handle making voucher payments
 */

package org.silentvault.spark.ui.wallet;

import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import javax.swing.border.EmptyBorder;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.event.*;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.util.StringUtils;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;

import java.util.ArrayList;
import java.util.Set;
import java.util.Vector;
import java.util.Hashtable;
import java.util.HashMap;
import java.util.TreeSet;
import java.util.TreeMap;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.ParseException;
import java.security.*;
import java.net.HttpURLConnection;


/**
 * Extends WorkPane to provide a screen for making voucher payments.
 *
 * @author Kevin Wilkerson
 */
public final class PaymentPane extends WorkPane {
	// BEGIN data members
	/**
	 * number of supported (regular) baggage fields
	 */
	private final int					M_RegBagFields = 5;
	
	/**
	 * max length of baggage field names
	 */
	private final int					M_MaxFieldLen = 14;

	/**
	 * max length of baggage field data
	 */
	private final int					M_MaxBagLen = 256;

	/**
	 * most lenient format for amounts (used for voucher totals)
	 */
	private final DecimalFormat	M_AmtFormatter;

	/**
	 * the phase we're in (1 = first, 2 = second)
	 */
	private int							m_Phase;

	/**
	 * modal controlling advanced (versus simple) message mode
	 */
	private boolean					m_Advanced;

	/**
	 * list of messages, if any, describing order details from vendor websites
	 */
	private ArrayList<WebMessage>	m_SiteMessages;

	/**
	 * the one to use to initialize order
	 */
	private WebMessage				m_SelectedSiteMsg;

	/**
	 * action to return to menu
	 */
	private AbstractAction			m_MenuAction;

	/**
	 * action to abort payment
	 */
	private AbstractAction			m_AbortAction;

	/**
	 * action to continue to phase 2
	 */
	private AbstractAction			m_ContinueAction;

	/**
	 * action to submit payment
	 */
	private AbstractAction			m_SubmitAction;

	/**
	 * action to attempt decode of a website order message
	 */
	private AbstractAction			m_DecodeWebOrderAction;

	/**
	 * action to expand/contract message fields (toggle advanced mode)
	 */
	private AbstractAction			m_ExpandAction;

	/**
	 * action to go back from screen 2 to screen 1
	 */
	private AbstractAction			m_BackAction;

	/**
	 * action to select vouchers manually
	 */
	private AbstractAction			m_SelectAction;

	/**
	 * UI element to hold amount of a payment (in base units)
	 */
	private JTextField				m_PaymentAmount;

	/**
	 * UI element to hold payee for a payment (set from m_PrevPayeeSelect)
	 */
	private JTextField				m_PaymentPayee;

	/**
	 * UI element to pick from among previous payees
	 */
	private JComboBox<String>		m_PrevPayeeSelect;

	/**
	 * UI element to hold amount of a payment (in fiat currency)
	 */
	private JTextField				m_ValueEntry;

	/**
	 * UI element to hold drop-down list of supported fiat currencies
	 */
	private JComboBox<String>		m_FiatSelection;

	/**
	 * UI element to hold baggage field names
	 */
	private JTextField[]				m_BaggageFields;

	/**
	 * UI element to hold baggage field values
	 */
	private JTextField[]				m_BaggageValues;

	/**
	 * UI element to hold time-to-live days
	 */
	private JComboBox<String>		m_TTLdays;

	/**
	 * UI element to hold drop-down list of website order messages
	 */
	private JComboBox<String>		m_WebsiteMessages;

	/**
	 * UI element to hold decryption code for a web message
	 */
	private JTextField				m_WebDecodeKey;

	/**
	 * UI element to hold total selected voucher value
	 */
	private JLabel						m_TotalVoucherValue;

	/**
	 * UI element for GAU (base units) amount label preface
	 */
	final private JLabel				M_GAUpreface;

	/**
	 * UI element for GAU (base units) amount label suffix
	 */
	final private JLabel				M_GAUsuffix;

	/**
	 * UI element for message field entry (non-advanced mode)
	 */
	private JTextArea					m_MessageField;

	/**
	 * UI element for message fields, in either simple or advanced mode
	 */
	private JPanel						m_MessagePane;

	/**
	 * UI element for the expand/contract action
	 */
	private JCheckBox					m_ExpandButton;

	/**
	 * button for backing up one screen
	 */
	private JButton					m_BackButton;

	/**
	 * button for continuing to next screen
	 */
	private JButton					m_ContButton;

	/**
	 * abort button
	 */
	private JButton					m_AbortButton;

	/**
	 * button for submitting payment
	 */
	private JButton					m_SubmitButton;

	/**
	 * button for manual voucher selection
	 */
	private JButton					m_SelectButton;

	/**
	 * storage for tokens used in a send payment operation
	 */
	private ArrayList<XMLToken>	m_PaymentTokens;

	/**
	 * temp storage for baggage fields, in plain text
	 */
	private TreeMap<String, String> 	m_BagFields;

	/**
	 * the allowed pickup time, in usec
	 */
	private long						m_TTL;

	/**
	 * the amount of the payment, in base units
	 */
	private double						m_Amount;

	/**
	 * the VS payee, after processing
	 */
	private String						m_PayeeVS;

	/**
	 * the owner of the m_PayeeKey (last time it was fetched).  If this value
	 * == m_PayeeVS, then we do not need to re-fetch the pubkey.
	 */
	private String						m_VSKeyFetched;

	/**
	 * the public key of the payee
	 */
	private PublicKey					m_PayeeKey;

	/**
	 * Hash table in which to store payment baggage fields in clear XML.
	 * This is used so that we can perpetuate the same baggage field values
	 * sent to the payee in our stored pending payment record, once we pull the
	 * confirmation record off the DHT.  Because the key to this table is the
	 * payee VS#, making multiple payments to the same payee in succession
	 * without picking up and processing the confirmation records, will result
	 * in overwriting the baggage field values with those of the latest payment.
	 * (This is meant to be impossible, barring queued events at the VP.)
	 */
	private Hashtable<String, String>	m_PaymentBaggage;

	/**
	 * inner class to provide a DocumentListener for the amount field
	 */
	final class PmtDocumentListener implements DocumentListener {
		/**
		 * handle insert notifications
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			Document doc = de.getDocument();
			String curTxt = "";

			// interpret a '.' or ',' as 0 if it's the first character added
			int curLen = doc.getLength();
			if (curLen == 1) {
				try {
					curTxt = doc.getText(0, 1);
				}
				catch (BadLocationException ble) {
					Log.error("Bad insert doc location", ble);
				}
				if (curTxt.equals(".")) {
					// force this to "0."
					queueTextMod(m_ValueEntry, "0.");
					return;
				}
				// NB: for internationalization
				if (curTxt.equals(",")) {
					// force this to "0,"
					queueTextMod(m_ValueEntry, "0,");
					return;
				}
			}

			// disallow letters etc. altogether
			int changeAt = de.getOffset();
			try {
				curTxt = doc.getText(0, doc.getLength());
			}
			catch (BadLocationException ble) {
				Log.error("Cannot undo bad value edit", ble);
				return;
			}
			char chg = curTxt.charAt(changeAt);
			if (!Character.isDigit(chg) && chg != '.' && chg != ',') {
				// undo this edit
				String modTxt = curTxt.substring(0, changeAt)
									+ curTxt.substring(changeAt+1);
				queueTextMod(m_ValueEntry, modTxt);
				return;
			}

			// otherwise just parse
			if (parseValue()) {
				updateGAUdisplay(true);
			}
			else {
				// undo this edit
				String modTxt = curTxt.substring(0, changeAt)
									+ curTxt.substring(changeAt+1);
				queueTextMod(m_ValueEntry, modTxt);
			}
		}

		/**
		 * handle removal notifications
		 * @param de the deletion event
		 */
		public void removeUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			// if the only thing left is a '.' or ',', force it to a 0
			Document doc = de.getDocument();
			int curLen = doc.getLength();
			if (curLen == 1) {
				String curTxt = "";
				try {
					curTxt = doc.getText(0, 1);
				}
				catch (BadLocationException ble) {
					Log.error("Bad remove doc location", ble);
				}
				if (curTxt.equals(".")) {
					// force this to "0."
					queueTextMod(m_ValueEntry, "0.");
					return;
				}
				// NB: for internationalization
				if (curTxt.equals(",")) {
					// force this to "0,"
					queueTextMod(m_ValueEntry, "0,");
					return;
				}
			}

			// otherwise just parse
			if (parseValue()) {
				updateGAUdisplay(true);
			}
		}

		/**
		 * handle change notifications (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JTextField");
		}

		/**
		 * helper method to parse the current value of m_ValueEntry
		 * @return true on success, false on failure
		 */
		private boolean parseValue() {
			// examine what's in the value entry field
			String enteredVal = m_ValueEntry.getText();
			if (enteredVal.isEmpty() || enteredVal.equals("0.00")
				 || enteredVal.equals("0") || enteredVal.equals("0,00"))
			{
				return true;
			}
			
			// attempt to parse
			double vvv = 0.0;
			try {
				Number val = M_AmtFormatter.parse(enteredVal);
				vvv = val.doubleValue();
			}
			catch (ParseException pe) {
				return false;
			}
			if (vvv < 0.0) {
				return false;
			}
			return true;
		}
	}

	/**
	 * document listener for dynamic updates of m_ValueEntry
	 */
	private PmtDocumentListener	m_PaymentListener;

	/**
	 * inner class to provide a DocumentListener for the payee field
	 */
	final class PayeeDocumentListener implements DocumentListener {
		/**
		 * handle insert notifications
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			Document doc = de.getDocument();
			int curLen = doc.getLength();
			String curTxt = null;
			try {
				curTxt = doc.getText(0, curLen);
			}
			catch (BadLocationException ble) { return; }

			// min length is 3 chars + "-nnn", and we need a '-' in the string
			m_HavePayee = (curLen >= 7 && curTxt.indexOf('-') != -1);
			updateGAUdisplay(false);
		}

		/**
		 * handle removal notifications
		 * @param de the deletion event
		 */
		public void removeUpdate(DocumentEvent de) {
			Document doc = de.getDocument();
			int curLen = doc.getLength();
			String curTxt = null;
			try {
				curTxt = doc.getText(0, curLen);
			}
			catch (BadLocationException ble) { return; }

			// min length is 3 chars + "-nnn", and we need a '-' in the string
			m_HavePayee = (curLen >= 7 && curTxt.indexOf('-') != -1);
			updateGAUdisplay(false);
		}

		/**
		 * handle change notifications (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JTextField");
		}
	}

	/**
	 * document listener for dynamic updates of m_PaymentPayee
	 */
	private PayeeDocumentListener	m_PayeeListener;

	/**
	 * whether or not we've seen what might be a valid payee VS#
	 */
	private boolean					m_HavePayee;

	/**
	 * inner class to provide a DocumentListener for the Message entry box
	 * and other baggage field values
	 */
	final class MessageDocumentListener implements DocumentListener {
		JTextComponent		m_TextComp;
		int					m_MaxLen;

		/**
		 * constructor
		 * @param txtComp the text component for which we're the doc listener
		 * @param maxLen the max char length we wish to permit in the component
		 */
		public MessageDocumentListener(JTextComponent txtComp, int maxLen) {
			m_TextComp = txtComp;
			m_MaxLen = maxLen;
		}

		/**
		 * handle inserts
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			Document doc = de.getDocument();
			int remaining = m_MaxLen - doc.getLength();
			Integer rem = new Integer(remaining);
			m_CharCount.setText(rem.toString());
			if (remaining < 0) {
				// silently throw away the excess text
				String maxText = "";
				try {
					maxText = doc.getText(0, m_MaxLen);
				}
				catch (BadLocationException ble) {
					Log.error("Error stripping excess text from message box", ble);
					return;
				}
				queueTextMod(m_TextComp, maxText);
			}
		}

		/**
		 * handle deletions
		 * @param de the removal event
		 */
		public void removeUpdate(DocumentEvent de) {
			Document doc = de.getDocument();
			int remaining = m_MaxLen - doc.getLength();
			Integer rem = new Integer(remaining);
			m_CharCount.setText(rem.toString());
		}

		/**
		 * handle changes (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JTextArea");
		}
	}

	/**
	 * document listener for dynamic updates of m_MessageField
	 */
	private MessageDocumentListener	m_MessageListener;

	/**
	 * document length counter which shows how many chars are left
	 */
	private JLabel						m_CharCount;

	/**
	 * inner class to modify a text component from within its document listener,
	 * without triggering a recursive modification exception
	 */
	class TextComponentMod implements Runnable {
		/**
		 * the text component to be modified
		 */
		JTextComponent		m_ModComponent;

		/**
		 * its new contents
		 */
		String				m_NewContents;

		/**
		 * constructor
		 * @param txtComp the text component to update
		 * @param newText the replacement text
		 */
		public TextComponentMod(JTextComponent txtComp, String newText) {
			m_ModComponent = txtComp;
			m_NewContents = new String(newText);
		}

		/**
		 * execute main task in background
		 */
		public void run() {
			try { Thread.sleep(25); }
			catch (InterruptedException ie) { /* ignore */ }
			m_ModComponent.setText(m_NewContents);
			m_DoingUndo = false;
			m_ModComponent.setCaretPosition(m_NewContents.length());

			// if we just undid something on the amount entry, update value display
			if (m_ModComponent == m_ValueEntry) {
				updateGAUdisplay(true);
			}
		}
	}

	/**
	 * track whether an undo mod is in progress
	 */
	private boolean			m_DoingUndo;

	/**
	 * flag indicating that we need a manual token resync before next attempt
	 */
	private boolean			m_NeedTokenSync;

	/**
	 * inner class for handling popup window closures
	 */
	private class PopupCloser extends WindowAdapter {
		public void windowClosing(WindowEvent we) {
			JFrame popup = (JFrame) we.getWindow();
			m_Plugin.deregisterPopup(popup);
			m_VoucherPopup = null;
		}
	}

	/**
	 * the popup window for selecting vouchers manually
	 */
	private JFrame				m_VoucherPopup;

	/**
	 * the asset type of vouchers being worked with for the selected Issuer
	 */
	private String				m_VoucherAssetType;

	/**
	 * whether or not we appear to be working on a SVX/SVM payment
	 */
	private boolean			m_InternalTabPmt;

	/**
	 * the most recently received incoming payment confirmation
	 */
	private XMLConfirmPayment	m_LatestConfirmation;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the HomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public PaymentPane(WalletClient plugin, HomePane home, WTabManager mgr) {
		super(plugin, home, mgr);
		m_PaymentTokens = new ArrayList<XMLToken>();
		m_BagFields = new TreeMap<String, String>();
		m_PaymentBaggage = new Hashtable<String, String>();
		m_PayeeVS = m_VSKeyFetched = m_VoucherAssetType = "";

		// set the default phase
		m_Phase = 1;

		// define formatter for payment amounts
		M_AmtFormatter = new DecimalFormat("#,###,##0.0###");

		// obtain any messages from website message listener
		m_SiteMessages = m_Plugin.getWebListener().getWebsiteMessages();

		// set token cost
		m_TokenCost.setText("Payment uses " + VoucherRequest.M_PaymentCost
									+ " tokens");

		// set abort action
		m_AbortAction = new AbstractAction("Abort Payment") {
			public void actionPerformed(ActionEvent actionEvent) {
				// clear existing data entry fields
				resetForNextPayment();
				m_PaymentBaggage.clear();
				if (m_SelectedSiteMsg == null) {
					/* NB: this probably means we're aborting something we couldn't
					 * decode properly.  So just blindly remove the first one in
					 * the list.
					 */
					if (!m_SiteMessages.isEmpty()) {
						m_SiteMessages.remove(0);
					}
				}

				// dispose of any voucher select popup window
				if (m_VoucherPopup != null) {
					m_Plugin.deregisterPopup(m_VoucherPopup);
					m_VoucherPopup.dispose();
					m_VoucherPopup = null;
				}

				// switch back to main menu screen
				m_TabManager.queueTransition(
													new Integer(WTabManager.M_GO_MAIN_MENU));
				m_TabManager.makeTransition();
			}
		};
		m_AbortButton = new JButton(m_AbortAction);
		m_AbortButton.setToolTipText("Click here to abort making payment");
		m_AbortButton.setFont(m_Plugin.M_ButtonFont);
		m_AbortButton.setFocusable(false);

		// set continue action (transition from phase 1 to phase 2), define button
		setContinueAction();
		m_ContButton = new JButton(m_ContinueAction);
		m_ContButton.setFont(m_Plugin.M_ButtonFont);
		m_ContButton.setEnabled(false);
		m_ContButton.setDefaultCapable(true);

		// set submit payment action, define button
		setSubmitAction();
		m_SubmitButton = new JButton(m_SubmitAction);
		m_SubmitButton.setFont(m_Plugin.M_ButtonFont);
		m_SubmitButton.setEnabled(false);
		m_SubmitButton.setDefaultCapable(true);

		// set manual voucher selection action and define button
		setSelectAction();
		m_SelectButton = new JButton(m_SelectAction);
		m_SelectButton.setFont(m_Plugin.M_ButtonFont);
		m_SelectButton.setEnabled(false);
		m_SelectButton.setToolTipText("<html>Click here to override<br/>"
												+ "automatic voucher selection</html>");
		m_VoucherPopup = null;

		// set website order message decode action
		setWebOrderAction();

		// set mode expansion action (NB: no longer used)
		m_ExpandAction = new AbstractAction("Advanced message fields") {
			public void actionPerformed(ActionEvent actionEvent) {
				m_Advanced = !m_Advanced;

				// reset the char count to first displayed item
				int len = 0;
				if (m_Advanced) {
					len = M_MaxFieldLen - m_BaggageFields[0].getText().length();
				}
				else {
					len = M_MaxBagLen - m_MessageField.getText().length();
				}
				Integer leng = new Integer(len);
				m_CharCount.setText(leng.toString());
				buildMessagePane();
				m_MessagePane.revalidate();
			}
		};
		m_ExpandButton = new JCheckBox(m_ExpandAction);
		m_ExpandButton.setToolTipText("Click here to toggle advanced field mode");
		m_ExpandButton.setHorizontalTextPosition(SwingConstants.LEADING);
		m_ExpandButton.setHorizontalAlignment(SwingConstants.TRAILING);
		m_ExpandButton.setFocusable(false);

		// set action to return to first screen from second
		m_BackAction = new AbstractAction("Back") {
			public void actionPerformed(ActionEvent actionEvent) {
				m_Phase = 1;
				if (!m_InternalTabPmt) {
					m_IssuerSel.setEnabled(true);
				}
				m_PrevPayeeSelect.setSelectedItem(m_PaymentPayee.getText());
				prepDisplay();
			}
		};
		m_BackButton = new JButton(m_BackAction);
		m_BackButton.setFont(m_Plugin.M_ButtonFont);
		m_BackButton.setToolTipText("Return to the previous screen");
		m_BackButton.setFocusable(false);

		// establish some other UI objects

		// field to show the total value in asset units on the Confirm screen
		m_TotalVoucherValue = new JLabel();
		m_TotalVoucherValue.setBorder(m_InsetBorder);
		m_TotalVoucherValue.setFont(m_ParentHome.M_LabelFont);
		m_TotalVoucherValue.setOpaque(true);
		m_TotalVoucherValue.setToolTipText("total asset value of vouchers "
													+ "selected, including any change back");

		// this text field is only used to supply a document listener for the
		// payee field, because JComboBox doesn't support them

		m_PayeeListener = new PayeeDocumentListener();
		m_PaymentPayee = new JTextField(25);
		m_PaymentPayee.setDropTarget(null);
		m_PaymentPayee.getDocument().addDocumentListener(m_PayeeListener);

		m_PaymentListener = new PmtDocumentListener();
		m_ValueEntry = new JTextField(12);
		m_ValueEntry.setDropTarget(null);
		m_ValueEntry.setHorizontalAlignment(JTextField.RIGHT);
		m_ValueEntry.setText("0");
		m_ValueEntry.setToolTipText("Enter a positive decimal number");
		m_ValueEntry.getDocument().addDocumentListener(m_PaymentListener);
		m_ValueEntry.addActionListener(this);

		// we will override the "GAU" units stuff in prepDisplay() as needed
		Font gauFont = new Font(Font.SANS_SERIF, Font.BOLD, 12);
		M_GAUpreface = new JLabel(" = ");
		M_GAUpreface.setFont(gauFont);
		M_GAUpreface.setHorizontalAlignment(SwingConstants.TRAILING);
		M_GAUpreface.setToolTipText("payment equivalent amount in grams of gold");

		M_GAUsuffix = new JLabel(" GAU (grams of gold)");
		M_GAUsuffix.setFont(gauFont);
		M_GAUsuffix.setToolTipText("payment weight in GAU");
		M_GAUsuffix.setHorizontalAlignment(SwingConstants.TRAILING);

		// payment amount in asset units, shown on entry screen
		m_PaymentAmount = new JTextField(12);
		m_PaymentAmount.setDropTarget(null);
		m_PaymentAmount.setPreferredSize(new Dimension(100, 25));
		m_PaymentAmount.setFont(gauFont);
		m_PaymentAmount.setBackground(m_Plugin.m_LabelBackColor);
		m_PaymentAmount.setHorizontalAlignment(JTextField.RIGHT);
		m_PaymentAmount.setEditable(false);
		m_PaymentAmount.setOpaque(true);

		final String[] ttlVals = { "7", "6", "5", "4", "3", "2", "1" };
		m_TTLdays = new JComboBox<String>(ttlVals);
		m_TTLdays.setToolTipText("Select number of days from list");
		m_TTLdays.setPreferredSize(new Dimension(50, 25));

		m_MessageField = new JTextArea(3, 36);
		m_MessageField.setToolTipText("Enter any message to the payee here");
		m_MessageField.setBorder(m_InsetBorder);
		m_MessageField.setLineWrap(true);
		m_MessageField.setWrapStyleWord(true);
		m_MessageListener
			= new MessageDocumentListener(m_MessageField, M_MaxBagLen);
		m_MessageField.getDocument().addDocumentListener(m_MessageListener);

		m_MessagePane = new JPanel(new GridBagLayout());
		m_MessagePane.setBorder(null);

		// set up default baggage fields
		m_BaggageFields = new JTextField[M_RegBagFields];
		m_BaggageValues = new JTextField[M_RegBagFields];
		for (int iii = 0; iii < M_RegBagFields; iii++) {
			m_BaggageFields[iii] = new JTextField(10);
			m_BaggageFields[iii].setDropTarget(null);
			m_BaggageFields[iii].setToolTipText("<html>enter a field name here"
															+ "<br/>(max " + M_MaxFieldLen
															+ " chars)</html>");
			MessageDocumentListener listen
				= new MessageDocumentListener(m_BaggageFields[iii], M_MaxFieldLen);
			m_BaggageFields[iii].getDocument().addDocumentListener(listen);
			m_BaggageFields[iii].addActionListener(this);
			m_BaggageValues[iii] = new JTextField(25);
			m_BaggageValues[iii].setDropTarget(null);
			m_BaggageValues[iii].setToolTipText("<html>enter the corresponding "
															+ "value here<br/>(max "
															+ M_MaxBagLen + " chars)</html>");
			listen
				= new MessageDocumentListener(m_BaggageValues[iii], M_MaxBagLen);
			m_BaggageValues[iii].getDocument().addDocumentListener(listen);
			m_BaggageValues[iii].addActionListener(this);
		}

		// set up character count display
		Integer ilen = new Integer(M_MaxBagLen);
		m_CharCount = new JLabel(ilen.toString());
		m_CharCount.setHorizontalAlignment(SwingConstants.CENTER);
		m_CharCount.setFont(m_ParentHome.M_DataFont);
		m_CharCount.setToolTipText("The number of field characters remaining");
	}

	/**
	 * utility method to establish the action to transition between phases
	 */
	private void setContinueAction() {
		m_ContinueAction = new AbstractAction("Continue Payment") {
			public void actionPerformed(ActionEvent actionEvent) {
				/* validation algorithm - make sure we have these things:
				 * 	1. a payee (will be in m_PaymentPayee)
				 * 	2. a payment amount (will be in m_PaymentAmount)
				 * 	3. a TTL value (will be in m_TTLdays)
				 * 	4. valid values (in m_BaggageValues[]) for any named baggage
				 * 		fields (in m_BaggageFields[])
				 */
				final String genericErr = "cannot process payment";

				// step 1 - validate payee
				String payeeVS = m_PaymentPayee.getText().toLowerCase();
				if (payeeVS.isEmpty()) {
					showError(genericErr, "no payee wallet specified",
								"enter your desired payee");
					return;
				}

				// silently peel off the @publisher if it was given
				VsSecrets sekrits = m_Plugin.getLoginSecrets();
				String vp = sekrits.getPublisher();
				int atIdx = vp.lastIndexOf('@');
				vp = vp.substring(atIdx);
				if (payeeVS.endsWith(vp)) {
					int vpIdx = payeeVS.indexOf(vp);
					payeeVS = payeeVS.substring(0, vpIdx);
				}
				m_PayeeVS = payeeVS.concat(vp);
				
				// obtain the payee VS's public key if we need it
				if (!m_VSKeyFetched.equals(m_PayeeVS) || m_PayeeKey == null) {
					m_PayeeKey = m_Plugin.getPKSListener().obtainVSpubkey(m_PayeeVS);
					if (m_PayeeKey == null) {
						showError(genericErr,
									"could not obtain payee wallet's public key",
									"check validity of payee, or try again later");
						return;
					}
					m_VSKeyFetched = new String(m_PayeeVS);
				}
				// else: we already have the key for this VS

				// step 2 - validate amount, parsing strictly
				String amtStr = m_PaymentAmount.getText();
				if (amtStr.isEmpty()) {
					showError(genericErr, "no payment amount entered",
								"enter payment amount");
					return;
				}
				Number amount = null;
				try {
					amount = M_AmtFormatter.parse(amtStr);
				}
				catch (ParseException pe) {
					Log.error("Error formatting payment amount", pe);
					showError(genericErr, "improper amount entered",
								"enter positive decimal payment amount");
					return;
				}
				if (amount.doubleValue() <= 0.0) {
					showError(genericErr, "illegal amount entered",
								"enter positive decimal payment amount");
					return;
				}
				m_Amount = amount.doubleValue();

				// step 3 - compute TTL
				String selStr = (String) m_TTLdays.getSelectedItem();
				Integer days = Integer.parseInt(selStr.trim());
				m_TTL = days.longValue() * 86400000L;

				// step 4 - make sure any baggage field with a name has a value, and
				// count how many we have and what the field names are

				m_BagFields.clear();
				if (m_Advanced) {
					for (int iii = 0; iii < m_BaggageFields.length; iii++) {
						String field = m_BaggageFields[iii].getText();
						if (!field.isEmpty()) {
							// only allow alphanumeric chars and _ in field names
							String alteredFld = field.replaceAll("[^a-zA-Z0-9_]", "");
							if (alteredFld.isEmpty()) {
								if (m_SelectedSiteMsg != null) {
									showError(genericErr,
												"improper baggage field name: " + field,
												"notify vendor website administrator");
								}
								else {
									showError(genericErr,
												"improper baggage field name: " + field,
											"please enter only alphanumeric characters");
								}
								return;
							}

							// additionally, ensure that a field starts with a letter,
							// to avoid misinterpretation by the parser as a markup

							String fieldName = "B" + alteredFld;
							String fldVal = m_BaggageValues[iii].getText();
							if (fldVal.length() > M_MaxBagLen) {
								showError(genericErr, field + " value is too long",
											"enter a value of " + M_MaxBagLen
											+ " chars or less");
								return;
							}
							if (fldVal.isEmpty()) {
								showError(genericErr, "no value entered for " + field,
											"enter a value for this baggage field");
								return;
							}

							// replace any newlines with a space
							String val = fldVal.replace('\n', ' ');
							m_BagFields.put(fieldName, val);
						}
					}
				}
				else {
					// all we have is "Message", if one was entered
					String msgTxt = m_MessageField.getText();
					if (!msgTxt.isEmpty()) {
						if (msgTxt.length() > M_MaxBagLen) {
							showError(genericErr, "message text is too long",
										"enter a message of " + M_MaxBagLen
										+ " chars or less");
							return;
						}

						// replace any newlines with a space
						String val = msgTxt.replace('\n', ' ');
						m_BagFields.put("Message", val);
					}
				}

				// auto-select vouchers
				VoucherTableModel vtm = m_ParentHome.getVoucherModel();
				vtm.setSelectMode(VoucherTableModel.M_SelectMulti);
				ArrayList<XMLVoucher> selVouchers
					= vtm.autoSelectVouchers(m_Amount);
				if (selVouchers == null) {
					showError(genericErr,
								"your wallet contains insufficient voucher value",
								"please reduce the amount of your payment");
					return;
				}
				vtm.selectVouchers(selVouchers);

				// move to next phase
				m_Phase = 2;
				prepDisplay();
			}
		};
	}

	/**
	 * utility method to establish the action to submit the payment
	 */
	private void setSubmitAction() {
		m_SubmitAction = new AbstractAction("Confirm Payment") {
			public void actionPerformed(ActionEvent actionEvent) {
				final String genErr = "cannot perform payment";
				/* validation algorithm:
				 * step 1 - ensure we have sufficient voucher value selected
				 * step 2 - determine type of payment and validate token sufficiency
				 * step 3 - prepare and sign baggage fields, if any
				 * step 4 - prepare and sign payment message
				 * step 5 - send payment
				 */
				// step 1

				// confirm auto-selected vouchers
				String vVal = m_TotalVoucherValue.getText();
				int unitsIdx = vVal.indexOf('[');
				Double vouchTotal = null;
				try {
					Number amt = M_AmtFormatter.parse(vVal.substring(0, unitsIdx));
					vouchTotal = new Double(amt.doubleValue());
				}
				catch (ParseException pe) {
					Log.error("Error parsing value amount", pe);
					return;
				}
				if (vouchTotal.doubleValue() < m_Amount) {
					showError(genErr, "insufficient voucher value selected",
								"please select more voucher value");
					return;
				}

				/* step 2
				 * If there is one voucher and its value is exactly equal to the
				 * amount of the payment, then this is a validate type payment.
				 * If there is one voucher whose value is greater than the amount
				 * of the payment, it is a split.  Otherwise it is a merge.
				 */
				VoucherTableModel vtm = m_ParentHome.getVoucherModel();
				ArrayList<XMLVoucher> selVouchers = vtm.getSelectedVouchers();
				String spendTyp = null;
				int tokCost = 1;		// for pending payment record
				if (m_Amount == vouchTotal.doubleValue()) {
					if (selVouchers.size() == 1) {
						spendTyp = "validate";
						tokCost += VoucherRequest.M_ValidateCost;
					}
					else {
						spendTyp = "merge";
						tokCost += VoucherRequest.M_MergeCost;
					}
				}
				else if (selVouchers.size() == 1) {
					spendTyp = "split";
					tokCost += VoucherRequest.M_SplitCost;
				}
				else {
					spendTyp = "merge";
					tokCost += VoucherRequest.M_BothCost;
				}

				/* If we're spending all of our value from this Issuer, then we
				 * impose an additional requirement that we must leave at least
				 * enough tokens in the wallet to receive at least one payment.
				 * Note that if the user is attempting to spend all value out of
				 * the wallet, that this may compel them to reduce the amount of
				 * their payment by the cost of the lot of tokens purchased.
				 */
				int totTokCost = tokCost;
				double availVal = vtm.getVoucherSum();
				XMLVoucher firstV = selVouchers.get(0);
				// NB: 10 tokens is the minimum lot size, so we need this much value
				double tenTokens
					= XMLToken.getValueOfOneToken(firstV.getAsset(), null)
						* TokensPane.M_MinTokens;
				if (availVal <= (m_Amount + tenTokens)) {
					// ensure we'll have at least 6 tokens left after the payment
					totTokCost += VoucherRequest.M_ReceiveCost;
				}

				// see if we need to resync our tokens before selecting them
				if (m_NeedTokenSync) {
					m_TabManager.queryTokens(null, true);
					m_NeedTokenSync = false;
				}
				TokenTableModel ttm = m_ParentHome.getTokenModel();
				m_PaymentTokens
					= ttm.getAvailableTokens(totTokCost, vtm.getSelectedIssuer());
				if (m_PaymentTokens == null) {
					// make sure we blast the list of any previously committed tokens
					ttm.spendTokens(null);

					/* also, because our token purchase may force us to reduce the
					 * amount of our spend, reset to phase 1 screen
					 */
					m_Phase = 1;

					/* Record ourselves with tab manager, so we return here after
					 * the user has purchased tokens.  Then go purchase tokens.
					 */
					m_TabManager.setPendingTokens(m_TabManager.getPaymentPane());
					m_TabManager.queueTransition(
												new Integer(WTabManager.M_PURCHASE_TOKENS));
					m_TabManager.makeTransition();
					return;
				}

				// step 3
				// set up signature and init with our private key
				VsSecrets secrets = m_Plugin.getLoginSecrets();
				Signature dSig = null;
				byte[] sig = null;
				boolean ok = false;
				try {
					dSig = Signature.getInstance("SHA1withRSA");
					dSig.initSign(secrets.getPrivKey());
					ok = true;
				}
				catch (NoSuchAlgorithmException nsae) {
					// impossible
					Log.error("No such signature algorithm", nsae);
				}
				catch (InvalidKeyException ike) {
					Log.error("Invalid privkey for details signature", ike);
				}
				if (!ok) {
					showError(genErr, "internal error signing with private key");
					return;
				}

				// build the <payDetails/> element
				StringBuilder payDets = new StringBuilder(2048);
				payDets.append("<payer_VS>" + secrets.getVSnumber());
				payDets.append("</payer_VS>");
				payDets.append("<payee_VS>" + m_PayeeVS + "</payee_VS>");
				payDets.append("<ttl>" + m_TTL + "</ttl>");
				if (m_BagFields.isEmpty()) {
					payDets.append("<baggage_fields></baggage_fields>");
				}
				else {
					// build baggage fields section, and encrypt with pubkey of
					// payee VS

					StringBuilder baggage = new StringBuilder(1024);
					baggage.append("<field_names>");
					boolean firstFld = true;
					for (String field : m_BagFields.keySet()) {
						if (!firstFld) {
							baggage.append(":");
						}
						baggage.append(field);
						firstFld = false;
					}
					baggage.append("</field_names>");

					for (String field : m_BagFields.keySet()) {
						String fldVal = m_BagFields.get(field);
						// fix embedded newlines or markup chars that confuse parser
						fldVal = fldVal.replace('\n', ' ');
						fldVal = fldVal.replaceAll("<", ";lt;");
						fldVal = fldVal.replaceAll(">", ";gt;");
						baggage.append("<" + field + ">");
						baggage.append(fldVal);
						baggage.append("</" + field + ">");
					}

					// add our private key signature
					String signedBaggage = baggage.toString();
					try {
						dSig.update(signedBaggage.getBytes());
						sig = dSig.sign();
					}
					catch (SignatureException se) {
						Log.error("Unable to create baggage fields signature", se);
						showError(genErr, "error signing baggage field details");
						return;
					}
					baggage.append("<signature>");
					baggage.append(StringUtils.encodeBase64(sig, false));
					baggage.append("</signature>");

					// encrypt the entire baggage section with it
					String encBaggage = EncodingUtils.makeBase64PubkeyEncStr(
																				baggage.toString(),
																				m_PayeeKey);
					if (encBaggage == null || encBaggage.isEmpty()) {
						Log.error("could not encrypt baggage fields for payee "
									+ m_PayeeVS);
						showError(genErr,
									"unable to encrypt baggage fields to payee");
						return;
					}
					payDets.append("<baggage_fields>" + encBaggage);
					payDets.append("</baggage_fields>");

					// place the clear text baggage XML into the hash table for later
					// retrieval, indexed by the payee VS#

					String clearBaggage = "<baggage_fields>" + baggage.toString()
												+ "</baggage_fields>";
					m_PaymentBaggage.put(m_PayeeVS, clearBaggage);
				}

				// now sign the details with our private key
				String signedDets = payDets.toString();
				ok = false;
				try {
					dSig.update(signedDets.getBytes());
					sig = dSig.sign();
					ok = true;
				}
				catch (SignatureException se) {
					Log.error("Unable to create payment details signature", se);
				}
				if (!ok) {
					showError(genErr, "error signing payment details");
					return;
				}
				payDets.append("<signature>");
				payDets.append(StringUtils.encodeBase64(sig, false));
				payDets.append("</signature>");

				// now encrypt the whole block with VP's public key
				String payDetails = "<payDetails>" + payDets.toString()
											+ "</payDetails>";
				String encPay = EncodingUtils.makeBase64PubkeyEncStr(payDetails,
																			  secrets.getVPKey());
				if (encPay == null || encPay.isEmpty()) {
					showError(genErr, "error encrypting payment details for VP");
					return;
				}

				// step 4
				// construct signed <voucherRequest/> object
				VoucherRequest vReq = new VoucherRequest();
				vReq.setAction(spendTyp);
				vReq.setSigningVS(secrets.getVSnumber());
				vReq.setFolderHash(new String(secrets.getVoucherIndex()));
				vReq.setFolderCap(secrets.getVouchRWCap());

				// add selected input vouchers
				Hashtable<XMLVoucher, String> vouchers = vtm.getVoucherList();
				HashMap<String, XMLVoucher> inpVouchers = vReq.getVouchers();
				String baseUnits = "";
				for (XMLVoucher voucher : selVouchers) {
					String storeFile = vouchers.get(voucher);
					inpVouchers.put(storeFile, voucher);
					if (baseUnits.isEmpty()) {
						baseUnits = voucher.baseUnits();
					}
				}

				// add payment amount in base units; the change is implicit
				vReq.addOutputValue(m_Amount, baseUnits);
				vReq.setInitialized();

				// add our signature and encrypt to VP
				if (!vReq.signDetails(secrets.getPrivKey())) {
					showError(genErr, "error signing voucher request");
					return;
				}
				if (!vReq.encryptDetails(secrets.getVPKey(), false)) {
					showError(genErr, "error encrypting voucher request");
					return;
				}

				/* If we have more than the tokens required for the actual payment
				 * cost, remove the excess.  (This happens because we don't spend
				 * the 1 token for storing the pendpay record in the VoucherRequest
				 * object for the spend.  It can also occur if we're buffering up an
				 * extra 6 tokens for receiving another payment after spending all
				 * of our vouchers from this Issuer.)
				 */
				int diff = m_PaymentTokens.size() - (tokCost-1);
				for (int ttt = 0; ttt < diff; ttt++) {
					m_PaymentTokens.remove(ttt);
				}

				// add the indicated number of tokens
				for (XMLToken tok : m_PaymentTokens) {
					vReq.addPaymentToken(tok);
				}
				ttm.commitTokens(m_PaymentTokens);
				
				// commit the selected vouchers to the operation
				vtm.commitVouchers(selVouchers);

				// dispose of any voucher select popup window
				if (m_VoucherPopup != null) {
					m_Plugin.deregisterPopup(m_VoucherPopup);
					m_VoucherPopup.dispose();
					m_VoucherPopup = null;
				}

				// step 5 - now send the actual request to the OFS
				sendMakePayment(encPay, vReq);
			}
		};
	}

	/**
	 * utility method to establish the action to decode a website order message
	 */
	private void setWebOrderAction() {
		m_DecodeWebOrderAction = new AbstractAction("Decode Web Order Info") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_SiteMessages.isEmpty()) {
					Log.error("Decode action without messages to decode!");
					return;
				}

				// get selected item in list
				int sel = m_WebsiteMessages.getSelectedIndex();
				if (sel == -1) {
					showError("cannot decode web message", "none selected",
								"please select a website message to decode");
					return;
				}
				WebMessage selMsg = m_SiteMessages.get(sel);

				// obtain decryption key entered
				String decKey = m_WebDecodeKey.getText();
				if (decKey == null || decKey.isEmpty()) {
					showError("cannot decode web message", "missing decrypt key",
								"please enter the decryption code from the website");
					return;
				}

				// decrypt
				if (!selMsg.decodeInfo(decKey)) {
					showError("cannot decode web message",
								"incorrect key or bad message format",
								"please try again");
					return;
				}
				m_SelectedSiteMsg = selMsg;

				// we can only make payment with vouchers from the specified Issuer
				String issuer = selMsg.getIssuer();
				VoucherTableModel vtm = m_ParentHome.getVoucherModel();
				if (!issuer.isEmpty()) {
					int issCnt = m_IssuerSel.getItemCount();
					boolean found = false;
					for (int iii = 0; iii < issCnt; iii++) {
						Object selObj = m_IssuerSel.getItemAt(iii);
						String iss = (String) selObj;
						if (iss.equals(issuer)) {
							// tell the VoucherTableModel to use this issuer
							vtm.selectIssuer(issuer);
							found = true;
							break;
						}
					}
					if (!found) {
						showError("cannot make payment for this website order",
									"you do not have any vouchers from Issuer " + issuer,
									"abort this payment, and obtain appropriate "
									+ "vouchers,\n or make other payment arrangements "
									+ "with the vendor");
						resetForNextPayment();
						prepDisplay();
						return;
					}
				}

				// reinvoke order preparation (this will call buildValueHeader())
				m_Phase = 1;
				prepDisplay();
			}
		};
	}

	/**
	 * utility method to define the action to select vouchers manually in popup
	 */
	private void setSelectAction() {
		m_SelectAction = new AbstractAction("Select Manually") {
			public void actionPerformed(ActionEvent actionEvent) {
				// don't create a second one
				if (m_VoucherPopup != null) {
					return;
				}

				// create a popup window which displays the voucher table
				VoucherTableModel vtm = m_ParentHome.getVoucherModel();
				vtm.setSelectMode(VoucherTableModel.M_SelectMulti);
				m_VoucherPopup = new JFrame("Payment Vouchers");
				m_VoucherPopup.setDefaultCloseOperation(
															WindowConstants.DISPOSE_ON_CLOSE);
				m_VoucherPopup.setSize(new Dimension(450, 300));
				m_VoucherPopup.addWindowListener(new PopupCloser());
				JTable vt = m_ParentHome.getVoucherTable();
				JScrollPane sp = new JScrollPane(m_ParentHome.getVoucherTable());
				sp.setBorder(null);
				m_VoucherPopup.setContentPane(sp);
				m_VoucherPopup.setLocationRelativeTo(m_TabManager);
				m_VoucherPopup.setVisible(true);

				// register this popup with client so it can be closed on logout
				m_Plugin.registerPopup(m_VoucherPopup);
			}
		};
	}

	// method required to extend WorkPane
	/**
	 * prepare display for this sub-class (in one of two modes)
	 */
	public void prepDisplay() {
		// use standard WorkPane header
		buildValueHeader();

		// set button to match modal
		m_ExpandButton.setSelected(m_Advanced);

		VoucherTableModel vtm = m_TabManager.getHomePane().getVoucherModel();
		String issuer = vtm.getSelectedIssuer();
		XMLVoucher vvv = vtm.getFirstVoucher();
		if (vvv != null) {
			// NB: should have at least one voucher, or we wouldn't be here
			m_VoucherAssetType = vvv.getAsset();
		}

		// set the user instructions, based on phase
		StringBuilder uTxt = new StringBuilder(256);
		if (m_Phase == 1) {
			uTxt.append("Select the voucher type (" + issuer + ") and enter the ");
			uTxt.append("wallet ID you wish to send it to.<br/><br/>");
			uTxt.append("Enter the payment amount.  ");
			uTxt.append("If the payee doesn't pick up payment within the ");
			uTxt.append("set time, you can reclaim it.");
		}
		else {
			uTxt.append("Select one or more vouchers to cover the payment value.");
			uTxt.append("  (Hint: use your oldest vouchers first.)<br/><br/>");
			uTxt.append("If you select a voucher(s) with greater value than the ");
			uTxt.append("payment, you will receive a new voucher as \"change\".");
		}
		m_UserInstructs = uTxt.toString();

		// set screen label, with text based on phase
		if (m_Phase == 1) {
			buildScreenLabel("Make a Payment");
		}
		else {
			buildScreenLabel("Confirm Your Payment");
		}

		// provide action buttons in the button pane
		m_WorkBar.removeAll();
		m_ButtonPane.removeAll();

		// set buttons for each phase
		JRootPane jrp = m_ParentHome.getRootPane();
		if (m_Phase == 1) {
			// add the return to overview button as the sole contents of toolbar
			m_WorkBar.add(m_MainButton);

			// add the button to continue to phase 2
			String curTxtAmt = m_PaymentAmount.getText();
			if (m_HavePayee) {
				if (curTxtAmt.isEmpty() || curTxtAmt.equals("0.00")
					 || curTxtAmt.equals("0,00") || curTxtAmt.equals("0"))
				{
					m_ContButton.setToolTipText("Enter a payment amount");
					m_ContButton.setEnabled(false);
				}
				else {
					m_ContButton.setToolTipText("Click here to continue payment");
					m_ContButton.setEnabled(true);
					// in case issuer select changed, rebuild asset units equivalence
					updateGAUdisplay(true);
				}
			}
			else {
				m_ContButton.setEnabled(false);
				if (curTxtAmt.isEmpty() || curTxtAmt.equals("0.00")
					 || curTxtAmt.equals("0,00") || curTxtAmt.equals("0"))
				{
					m_ContButton.setToolTipText("Enter a payee and a payment "
														+ "amount");
				}
				else {
					m_ContButton.setToolTipText("Enter a payee");
					// in case issuer select changed, rebuild asset units equivalence
					updateGAUdisplay(true);
				}
			}
			m_ButtonPane.add(Box.createHorizontalStrut(290));	// spacer
			m_ButtonPane.add(m_ContButton);
			if (jrp != null) {
				jrp.setDefaultButton(m_ContButton);
			}
		}
		else {
			// add the abort button as the sole contents of the toolbar
			m_WorkBar.add(m_AbortButton);

			// first, the button to return to the previous (phase 1) screen
			m_ButtonPane.add(Box.createHorizontalStrut(25));	// spacer
			m_ButtonPane.add(m_BackButton);
			m_ButtonPane.add(Box.createHorizontalStrut(230));	// spacer

			// second, the button to submit the payment
			m_SubmitButton.setToolTipText("Select voucher(s) to cover payment");
			m_SubmitButton.setEnabled(false);
			m_ButtonPane.add(m_SubmitButton);
			if (jrp != null) {
				jrp.setDefaultButton(m_SubmitButton);
			}
		}
		m_WorkBar.setVisible(false);
		m_WorkBar.setVisible(true);
		m_ButtonPane.setVisible(false);
		m_ButtonPane.setVisible(true);

		// prepare a form to place inside the scroll area
		JPanel pmtForm = new JPanel(new GridBagLayout());
		pmtForm.setBorder(null);
		pmtForm.setPreferredSize(UIManager.getDimension("vouchTableDim"));
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_GBC.anchor = GridBagConstraints.NORTHWEST;
		m_GBC.fill = GridBagConstraints.NONE;

		// display website message decode panel if we have any on hand
		boolean webDecodeOnly = false;
		if (m_Phase == 1 && !m_SiteMessages.isEmpty()
		 	 && (m_SelectedSiteMsg == null || !m_SelectedSiteMsg.isDecoded()))
		{
			// build a panel for selecting and decoding these messages
			JPanel webOrderPane = new JPanel();
			webOrderPane.setLayout(new BoxLayout(webOrderPane, BoxLayout.Y_AXIS));
			webOrderPane.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createTitledBorder("Select Website Order Details"),
											BorderFactory.createEmptyBorder(5, 5, 5, 5)));

			final String insTxt =
				"Select message, enter secret from website, and click on Decode."
				+ "\n\nOnce decoded, the order details will populate the Payment "
				+ "form, and grey out the unmodifiable fields.";
			JTextArea ordPrompt = new JTextArea(insTxt, 5, 35);
			ordPrompt.setLineWrap(true);
			ordPrompt.setWrapStyleWord(true);
			ordPrompt.setEditable(false);
			ordPrompt.setBackground(m_Plugin.m_LabelBackColor);
			ordPrompt.setFocusable(false);

			// see if this is our first trip through here
			if (m_WebsiteMessages == null || m_SelectedSiteMsg == null) {
				String[] siteSubjects = new String[m_SiteMessages.size()];
				int idx = 0;
				for (WebMessage wmsg : m_SiteMessages) {
					siteSubjects[idx++] = wmsg.getSubject();
				}
				m_WebsiteMessages = new JComboBox<String>(siteSubjects);
			}
			if (m_WebDecodeKey == null) {
				m_WebDecodeKey = new JTextField(16);
				m_WebDecodeKey.setDropTarget(null);
				m_WebDecodeKey.setToolTipText("<html>Enter the decryption code<br/>"
													+ "displayed on the vendor's website"
													+ "</html>");
			}
			else {
				// clear decryption key
				m_WebDecodeKey.setText("");
			}
			webOrderPane.add(ordPrompt);
			webOrderPane.add(Box.createVerticalStrut(5));
			webOrderPane.add(m_WebsiteMessages);
			webOrderPane.add(Box.createVerticalStrut(5));
			webOrderPane.add(m_WebDecodeKey);
			webOrderPane.add(Box.createVerticalStrut(5));

			// add the decode button
			JButton decode = new JButton(m_DecodeWebOrderAction);
			decode.setToolTipText("<html>Click here to decrypt the website<br/>"
										+ "message using the decryption code</html>");
			decode.setFont(m_Plugin.M_ButtonFont);
			webOrderPane.add(decode);
			pmtForm.add(webOrderPane, m_GBC);
			pmtForm.add(Box.createVerticalStrut(10), m_GBC);
			webDecodeOnly = true;

			// add abort button to the top button bar
			m_WorkBar.add(m_AbortButton);
		}

		// see whether we have any website messages to pre-fill order with
		boolean useWeb = false;
		if (m_SelectedSiteMsg != null && m_SelectedSiteMsg.isDecoded()) {
			useWeb = true;
			m_Advanced = true;
			m_IssuerSel.setEnabled(false);
			m_ExpandButton.setSelected(true);
			m_ExpandButton.setEnabled(false);
		}
		else {
			m_ExpandButton.setSelected(false);
		}

		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		VsSecrets sekrits = m_Plugin.getLoginSecrets();
		AssetPriceCache apc = m_Plugin.getPricingModule();
		Set<String> fiatCurrCodes = apc.getSupportedCurrencies();
		if (m_Phase == 1 && !webDecodeOnly) {
			// present editable drop-down list of previous payees, if any
			Vector<String> prevPayees = getPreviousPayees();
			if (prevPayees == null) {
				m_PrevPayeeSelect = new JComboBox<String>();
				m_PrevPayeeSelect.setToolTipText("<html>Download your receipts "
															+ "first<br/>to "
															+ "populate this list</html>");
			}
			else {
				m_PrevPayeeSelect = new JComboBox<String>(prevPayees);
				m_PrevPayeeSelect.setToolTipText("Select a previous payee, or enter"
															+ " a new one");
			}

			// if we already have a selected payee, add that and select it
			String curPayee = m_PaymentPayee.getText();
			m_PrevPayeeSelect.insertItemAt(curPayee, 0);
			m_PrevPayeeSelect.setSelectedIndex(0);
			m_PrevPayeeSelect.setPreferredSize(new Dimension(287, 25));
			m_PrevPayeeSelect.addActionListener(this);
			m_PrevPayeeSelect.setEditable(true);

			/* In the case of a payment triggered by another tab, we simply want
			 * to indicate that a system SVX/SVM wallet is being paid, without
			 * showing a regular payment screen at all.  Thus we construct the
			 * form one way in this case, generally without user inputs, and in
			 * the regular manner otherwise.
			 */
			JLabel payeeLab = null;
			if (m_InternalTabPmt) {
				// add label showing why we're making a payment
				JLabel preface = new JLabel("To fund your order you must make the "
													+ "following payment -");
				pmtForm.add(preface, m_GBC);
				pmtForm.add(Box.createVerticalStrut(10), m_GBC);
				payeeLab = new JLabel("The escrow wallet ID assigned to your "
											+ "order is: ");
			}
			else {
				payeeLab = new JLabel("Enter the wallet ID you wish to pay: ");
			}
			payeeLab.setLabelFor(m_PrevPayeeSelect);
			pmtForm.add(payeeLab, m_GBC);

			// override payee if a merchant web order, or SVX/SVM order
			if (useWeb) {
				// m_InternalTabPmt == true means SVX/SVM order; otherwise website
				String webPayee = m_SelectedSiteMsg.getVSnumber();
				int atIdx = webPayee.indexOf('@');
				if (atIdx != -1) {
					webPayee = webPayee.substring(0, atIdx);
				}
				m_PaymentPayee.getDocument().removeDocumentListener(
																					m_PayeeListener);
				m_PaymentPayee.setText(webPayee);
				m_PaymentPayee.getDocument().addDocumentListener(m_PayeeListener);
				m_HavePayee = true;

				// like wise make the payee selector a list of length 1
				m_PrevPayeeSelect.removeAllItems();
				m_PrevPayeeSelect.insertItemAt(webPayee, 0);
				m_PrevPayeeSelect.setSelectedIndex(0);
				m_PrevPayeeSelect.setEditable(false);
				if (!m_InternalTabPmt) {
					m_PrevPayeeSelect.setToolTipText("This payee is from your "
																+ "website order");
				}
				else {
					m_PrevPayeeSelect.setToolTipText("This payee is from your "
																+ "SVX/SVM order");
				}
			}
			JPanel promptPane = new JPanel();
			promptPane.setLayout(new BoxLayout(promptPane, BoxLayout.X_AXIS));
			promptPane.add(m_PrevPayeeSelect);
			String vp = sekrits.getPublisher();
			int atIdx = vp.indexOf('@');
			JLabel pubLab = new JLabel(vp.substring(atIdx));
			pubLab.setForeground(UIManager.getColor("standardGray"));
			promptPane.add(pubLab);
			if (useWeb) {
				m_GBC.fill = GridBagConstraints.HORIZONTAL;
			}
			pmtForm.add(promptPane, m_GBC);
			pmtForm.add(Box.createVerticalStrut(5), m_GBC);

			// check whether we need to disable issuer selection
			if (m_InternalTabPmt) {
				// freeze issuer selection to what was set by our caller
				m_IssuerSel.setEnabled(false);
			}

			/* For the amount, show a label/text field and a drop down list of fiat
			 * currencies.  Using this will fill in the actual gram weight field.
			 */
			JLabel valLab = null;
			if (m_InternalTabPmt) {
				valLab = new JLabel("The amount owed (including fees) to fund your "
										+ "order is: ");
			}
			else {
				valLab = new JLabel("Enter payment amount and pick currency "
											+ "code from list:");
			}
			valLab.setLabelFor(m_ValueEntry);
			pmtForm.add(valLab, m_GBC);

			JPanel valuePane = new JPanel();
			valuePane.setLayout(new BoxLayout(valuePane, BoxLayout.X_AXIS));
			if (useWeb) {
				df.setMaximumFractionDigits(XMLVoucher.M_Precision);
				df.setMinimumFractionDigits(2);
				df.setGroupingUsed(false);
				m_ValueEntry.getDocument().removeDocumentListener(
																				m_PaymentListener);
				m_ValueEntry.setText(df.format(m_SelectedSiteMsg.getAmount()));
				m_ValueEntry.getDocument().addDocumentListener(m_PaymentListener);
				m_ValueEntry.setEditable(false);
				if (!m_InternalTabPmt) {
					m_ValueEntry.setToolTipText("This amount comes from your "
														+ "website order");
				}
				else {
					m_ValueEntry.setToolTipText("This amount comes from your "
														+ "SVX/SVM order");
				}
			}
			else {
				m_ValueEntry.setEditable(true);
			}
			m_DoingUndo = false;
			valuePane.add(m_ValueEntry);

			// add drop-down list of all currencies for which we have prices + GAU
			String[] fiats = new String[fiatCurrCodes.size() + 1];
			int idx = 0;
			for (String fCode : fiatCurrCodes) {
				fiats[idx++] = new String(fCode);
			}
			fiats[idx++] = "GAU";
			if (m_FiatSelection == null) {
				// init this object
				m_FiatSelection = new JComboBox<String>(fiats);
				m_FiatSelection.setToolTipText("Select desired currency code");
				m_FiatSelection.setPreferredSize(new Dimension(120, 25));

				// make the default whatever the user has set in preferences
				String currDflt = m_Plugin.getUserPreferences().getValueCurrency();
				m_FiatSelection.setSelectedItem(currDflt);
			}
			if (useWeb) {
				// make sure the order specified a legal currency
				String orderCur = m_SelectedSiteMsg.getCurrency();
				if (orderCur.equalsIgnoreCase("GAU")
					 || fiatCurrCodes.contains(orderCur))
				{
					// use what the order specifies
					m_FiatSelection.setSelectedItem(orderCur);
					m_FiatSelection.setEnabled(false);
					if (!m_InternalTabPmt) {
						m_FiatSelection.setToolTipText("This currency comes from "
																+ "your website order");
					}
					else {
						m_FiatSelection.setToolTipText("This currency comes from "
																+ "your SVX/SVM order");
					}
				}
				else {
					Log.error("Website order specified unrecognized currency code, "
								+ orderCur);
					showError("Cannot pre-process website order",
								"unknown currency code in order details: " + orderCur,
								"report to vendor website administrator");
					// prevent repetition of error
					resetForNextPayment();
					return;
				}
			}
			else {
				m_FiatSelection.addActionListener(this);
			}
			valuePane.add(m_FiatSelection);

			// actual value entry, set by m_ValueEntry and m_FiatSelection above
			if (useWeb) {
				String selectedCurr = (String) m_FiatSelection.getSelectedItem();
				double ratio = 1.0;
				Double price = apc.getCurrPrice(selectedCurr);
				if (price != null) {
					ratio = price.doubleValue();
				}
				else {
					// fallback
					selectedCurr = "GAU";
					m_FiatSelection.setSelectedItem(selectedCurr);
				}
				Double weight = new Double(m_SelectedSiteMsg.getAmount() / ratio);
				m_PaymentAmount.setText(df.format(weight.doubleValue()));
				updateGAUdisplay(true);
			}

			// adjust display labels based on the asset type we're working with
			String dispUnits = XMLVoucher.getDisplayUnits(m_VoucherAssetType);
			M_GAUpreface.setToolTipText("payment amount in " + dispUnits);
			M_GAUsuffix.setText(" " + dispUnits);
			M_GAUsuffix.setToolTipText("payment amount in " + dispUnits);
			m_PaymentAmount.setToolTipText("payment amount in " + dispUnits);

			valuePane.add(M_GAUpreface);
			valuePane.add(m_PaymentAmount);
			valuePane.add(M_GAUsuffix);
			pmtForm.add(valuePane, m_GBC);
			pmtForm.add(Box.createVerticalStrut(10), m_GBC);

			/* Ask for the TTL.  For simplicity, pick 1-7 days from drop down list,
			 * 7 being the default.  (Note SVX/SVM orders default to 1 day.)
			 */
			JPanel ttlPane = new JPanel();
			ttlPane.setLayout(new BoxLayout(ttlPane, BoxLayout.X_AXIS));
			JLabel ttlLab
				= new JLabel("Select time for payee to pick up this payment: ");
			ttlLab.setLabelFor(m_TTLdays);
			ttlPane.add(ttlLab);
			ttlPane.add(m_TTLdays);
			JLabel dayLab = new JLabel(" days.");
			ttlPane.add(dayLab);
			if (useWeb) {
				m_GBC.fill = GridBagConstraints.NONE;
			}
			pmtForm.add(ttlPane, m_GBC);
			pmtForm.add(Box.createVerticalStrut(10), m_GBC);

			/* allocate baggage fields, using from order otherwise allowing
			 * for up to 5
			 */
			int numBagFields = M_RegBagFields;
			Hashtable<String, String> orderBaggage = null;
			if (useWeb) {
				// recreate baggage fields
				orderBaggage = m_SelectedSiteMsg.getOrderData();
				numBagFields = orderBaggage.size();
				m_BaggageFields = new JTextField[numBagFields];
				m_BaggageValues = new JTextField[numBagFields];
				for (int iii = 0; iii < numBagFields; iii++) {
					m_BaggageFields[iii] = new JTextField(10);
					m_BaggageFields[iii].setDropTarget(null);
					m_BaggageFields[iii].setToolTipText("enter a field name here");
					m_BaggageValues[iii] = new JTextField(25);
					m_BaggageValues[iii].setDropTarget(null);
					m_BaggageValues[iii].setToolTipText("<html>enter the "
																	+ "corresponding value here"
																	+ "<br/>(max " + M_MaxBagLen
																	+ " chars)</html>");
				}

				// preload form with baggage from order, not editable
				idx = 0;
				for (String bagFld : orderBaggage.keySet()) {
					// ignore any empty fields
					String webValue = orderBaggage.get(bagFld);
					if (webValue.isEmpty()) {
						continue;
					}
					m_BaggageFields[idx].setText(StringUtils.escapeForXML(bagFld));
					m_BaggageFields[idx].setEditable(false);
					String src = "website";
					if (m_InternalTabPmt) {
						src = "SVX/SVM";
					}
					m_BaggageFields[idx].setToolTipText("this field comes from your "
																	+ src + " order");
					m_BaggageValues[idx].setText(webValue);
					m_BaggageValues[idx].setEditable(false);
					m_BaggageValues[idx].setToolTipText("this value comes from your "
																	+ src + " order");
					idx++;
				}
			}

			// add message/baggage fields, depending on advanced mode
			buildMessagePane();
			pmtForm.add(m_MessagePane, m_GBC);
		}
		else if (m_Phase == 2) {		// phase 2
			voucherSelectChanged();

			// prepare and add the payment confirmation summary
			JPanel sumPane = new JPanel(new GridBagLayout());
			sumPane.setBorder(m_InsetBorder);
			sumPane.setPreferredSize(new Dimension(450, 235));
			m_GBC.anchor = GridBagConstraints.NORTHWEST;
			m_GBC.fill = GridBagConstraints.NONE;
			m_GBC.gridwidth = GridBagConstraints.REMAINDER;
			sumPane.add(Box.createVerticalGlue(), m_GBC);
			sumPane.add(Box.createVerticalStrut(20), m_GBC);

			// payee, plus pickup days
			String payee = m_PaymentPayee.getText();
			JLabel pLab = new JLabel("Payment to");
			pLab.setFont(m_ParentHome.M_LabelFont);
			m_GBC.gridwidth = GridBagConstraints.RELATIVE;
			sumPane.add(pLab, m_GBC);
			String pickupDays
				= "  (Pickup " + (String) m_TTLdays.getSelectedItem() + " days)";
			JLabel payeeLab = new JLabel("  :  " + payee + pickupDays,
													SwingConstants.TRAILING);
			payeeLab.setFont(m_ParentHome.M_DataFont);
			payeeLab.setToolTipText("wallet ID of recipient");
			m_GBC.gridwidth = GridBagConstraints.REMAINDER;
			sumPane.add(payeeLab, m_GBC);
			sumPane.add(Box.createVerticalStrut(5), m_GBC);

			// amount
			JLabel amtLab = new JLabel("Payment amount");
			amtLab.setFont(m_ParentHome.M_LabelFont);
			m_GBC.gridwidth = GridBagConstraints.RELATIVE;
			m_GBC.weightx = 0.0;
			sumPane.add(amtLab, m_GBC);

			double amt = 0.0;
			Number enteredAmt = null;
			try {
				Number pmtAmt = M_AmtFormatter.parse(m_PaymentAmount.getText());
				amt = pmtAmt.doubleValue();
				enteredAmt = M_AmtFormatter.parse(m_ValueEntry.getText());
			}
			catch (ParseException pe) {
				enteredAmt = new Double(0.0);
				Log.error("Amount parse error", pe);
			}
			df.setMaximumFractionDigits(XMLVoucher.M_Precision);
			df.setMinimumFractionDigits(0);
			df.setGroupingUsed(false);
			ArrayList<XMLVoucher> pmtVouchers = vtm.getSelectedVouchers();
			if (pmtVouchers.isEmpty()) {
				ArrayList<XMLVoucher> selVouchers
					= vtm.autoSelectVouchers(m_Amount);
				if (selVouchers == null) {
					showError("cannot process payment",
								"your wallet contains insufficient voucher value",
								"please reduce the amount of your payment");
					return;
				}
				vtm.selectVouchers(selVouchers);
				pmtVouchers = selVouchers;
			}
			XMLVoucher first = pmtVouchers.get(0);
			String dispUnits = XMLVoucher.getDisplayUnits(first.getAsset());
			JLabel gauLab
				= new JLabel(" = " + df.format(amt) + " " + dispUnits + "  ");
			gauLab.setFont(m_ParentHome.M_DataFont);
			gauLab.setToolTipText("voucher asset units");
			df.setMaximumFractionDigits(2);
			df.setMinimumFractionDigits(2);
			JLabel fiatLab
				= new JLabel("( " + df.format(enteredAmt.doubleValue()) + " "
									+ (String) m_FiatSelection.getSelectedItem() + " )");
			fiatLab.setFont(m_ParentHome.M_DataItalFont);
			fiatLab.setToolTipText("payment units");

			// place these amounts side-by-side into a mini-pane
			JPanel amtPane = new JPanel(new GridLayout(1, 2));
			amtPane.setBorder(null);
			amtPane.add(gauLab);
			amtPane.add(fiatLab);
			m_GBC.gridwidth = GridBagConstraints.REMAINDER;
			m_GBC.weightx = 1.0;
			sumPane.add(amtPane, m_GBC);
			sumPane.add(Box.createVerticalStrut(15), m_GBC);

			// notice about auto-voucher selection, and button for manual override
			JLabel mselLab = new JLabel("Vouchers have been auto-selected. ");
			mselLab.setFont(m_ParentHome.M_LabelFont);
			FlowLayout fLay = new FlowLayout(FlowLayout.LEADING);
			fLay.setAlignOnBaseline(true);
			JPanel selPane = new JPanel(fLay);
			selPane.setBorder(null);
			selPane.add(mselLab);
			m_SelectButton.setEnabled(true);
			selPane.add(m_SelectButton);
			sumPane.add(selPane, m_GBC);
			m_GBC.weightx = 0.0;
			sumPane.add(Box.createVerticalStrut(5), m_GBC);

			// message label, regardless of advanced setting
			JLabel msgLab = new JLabel("Message:");
			msgLab.setFont(m_ParentHome.M_LabelFont);
			sumPane.add(msgLab, m_GBC);
			sumPane.add(Box.createVerticalStrut(5), m_GBC);

			if (m_Advanced) {
				// baggage, if any
				if (!m_BagFields.isEmpty()) {
					for (String fld : m_BagFields.keySet()) {
						String fldNm = fld.substring(1);
						JLabel fldLab = new JLabel(fldNm);
						fldLab.setFont(m_ParentHome.M_DataFont);
						m_GBC.gridwidth = GridBagConstraints.RELATIVE;
						m_GBC.fill = GridBagConstraints.NONE;
						sumPane.add(fldLab, m_GBC);

						String bagVal = m_BagFields.get(fld);
						JLabel valLab = new JLabel(bagVal);
						valLab.setFont(m_ParentHome.M_DataFont);
						if (bagVal.length() > 35) {
							// add the whole thing onto the tooltip so it can be seen
							StringBuilder tipTxt = new StringBuilder(256);
							tipTxt.append("<html><table border=\"0\" width=\"200\">");
							tipTxt.append("<tr><td>" + bagVal + "</td></tr>");
							tipTxt.append("</table></html>");
							valLab.setToolTipText(tipTxt.toString());
						}
						m_GBC.gridwidth = GridBagConstraints.REMAINDER;
						m_GBC.fill = GridBagConstraints.HORIZONTAL;
						sumPane.add(valLab, m_GBC);
						sumPane.add(Box.createVerticalStrut(5), m_GBC);
					}
				}
			}
			else {
				// message, if there is one
				JTextArea msg = new JTextArea(1, 70);
				msg.setEditable(false);
				msg.setLineWrap(true);
				msg.setWrapStyleWord(true);
				msg.setFont(m_ParentHome.M_DataFont);
				msg.setBorder(null);
				msg.setBackground(m_Plugin.m_LabelBackColor);
				msg.setFocusable(false);
				msg.setText(m_BagFields.get("Message"));
				m_GBC.fill = GridBagConstraints.HORIZONTAL;
				sumPane.add(msg, m_GBC);
			}
			m_GBC.fill = GridBagConstraints.NONE;
			sumPane.add(Box.createVerticalStrut(10), m_GBC);

			// supply the total vouchers selected
			JLabel selLab = new JLabel("Vouchers selected   =");
			selLab.setFont(m_ParentHome.M_LabelFont);
			selLab.setBorder(new EmptyBorder(4, 0, 4, 8));
			m_GBC.gridwidth = GridBagConstraints.RELATIVE;
			sumPane.add(selLab, m_GBC);
			m_GBC.gridwidth = GridBagConstraints.REMAINDER;
			sumPane.add(m_TotalVoucherValue, m_GBC);
			sumPane.add(Box.createVerticalStrut(20), m_GBC);

			pmtForm.add(sumPane, m_GBC);

			// disable the Issuer select control, as we can't change this now
			m_IssuerSel.setEnabled(false);
		}
		pmtForm.add(Box.createVerticalGlue(), m_GBC);

		// put entire form into the main scroll area
		m_WorkArea.setBorder(null);
		m_WorkArea.setViewportView(pmtForm);
		m_Plugin.setDefaultStatusBar();
	}

	/**
	 * utility method to build the message/baggage field area
	 */
	private void buildMessagePane() {
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.anchor = GridBagConstraints.NORTHWEST;
		gbc.fill = GridBagConstraints.NONE;
		m_MessagePane.removeAll();
		final Color labColor = UIManager.getColor("standardGray");

		if (!m_Advanced) {
			// emit Message prompt, with the char count and expand button
			JPanel labelPane = new JPanel();
			labelPane.setLayout(new BoxLayout(labelPane, BoxLayout.X_AXIS));
			JLabel msgLab = new JLabel("Message");
			msgLab.setForeground(labColor);
			labelPane.add(msgLab);
			labelPane.add(Box.createHorizontalStrut(30));
			labelPane.add(m_CharCount);
			//labelPane.add(Box.createHorizontalStrut(125));
			//labelPane.add(m_ExpandButton);
			labelPane.add(Box.createHorizontalStrut(175));

			// add this pane together with the message field itself
			gbc.gridheight = 3;
			gbc.gridwidth = GridBagConstraints.REMAINDER;
			m_MessagePane.add(labelPane, gbc);
			JScrollPane msgScroll = new JScrollPane(m_MessageField);
			m_MessagePane.add(msgScroll, gbc);
		}
		else {
			// NB: the only way we get to this case is on a website/SVX|SVM order.

			int numBags = m_BaggageFields.length;
			gbc.gridheight = numBags + 2;

			// heading, only if we're a plain website order
			if (!m_InternalTabPmt) {
				JLabel fldLab = new JLabel("Field");
				fldLab.setForeground(labColor);
				gbc.gridwidth = GridBagConstraints.RELATIVE;
				gbc.fill = GridBagConstraints.NONE;
				gbc.weightx = 0.0;
				m_MessagePane.add(fldLab, gbc);
				JLabel valLab = new JLabel("Value");
				valLab.setForeground(labColor);
				gbc.fill = GridBagConstraints.HORIZONTAL;
				gbc.gridwidth = GridBagConstraints.REMAINDER;
				gbc.weightx = 1.0;
				m_MessagePane.add(valLab, gbc);
			}
			else {
				JLabel orderLab = new JLabel("The order you are funding by making "
													+ "this payment is:");
				orderLab.setForeground(labColor);
				gbc.fill = GridBagConstraints.HORIZONTAL;
				gbc.gridwidth = GridBagConstraints.REMAINDER;
				gbc.weightx = 1.0;
				m_MessagePane.add(orderLab, gbc);
			}

			// the fields themselves
			for (int iii = 0; iii < numBags; iii++) {
				gbc.fill = GridBagConstraints.NONE;
				gbc.gridwidth = GridBagConstraints.RELATIVE;
				gbc.weightx = 0.0;
				m_MessagePane.add(m_BaggageFields[iii], gbc);
				gbc.fill = GridBagConstraints.HORIZONTAL;
				gbc.gridwidth = GridBagConstraints.REMAINDER;
				gbc.weightx = 1.0;
				m_MessagePane.add(m_BaggageValues[iii], gbc);
			}
		}
		m_MessagePane.add(Box.createHorizontalStrut(100), gbc);
	}

	/**
	 * method to receive notification of a change in voucher selection
	 */
	public void voucherSelectChanged() {
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		ArrayList<XMLVoucher> selVouchers = vtm.getSelectedVouchers();
		if (selVouchers.isEmpty()
			 || vtm.getSelectMode() != VoucherTableModel.M_SelectMulti)
		{
			// not meant for us; do nothing
			return;
		}

		// total up selected vouchers, in asset units
		double tot = 0.0;
		String units = "";
		for (XMLVoucher vouch : selVouchers) {
			tot += vouch.getNPV();
			if (units.isEmpty()) {
				units = XMLVoucher.getDisplayUnits(vouch.getAsset());
			}
		}

		// display in total selected field, in correct color/font
		Number pmtAmt = null;
		double amt = 0.0;
		try {
			pmtAmt = M_AmtFormatter.parse(m_PaymentAmount.getText());
			amt = pmtAmt.doubleValue();
		}
		catch (ParseException pe) { /* ignore */ }
		if (tot >= amt) {
			m_TotalVoucherValue.setForeground(Color.BLACK);
			m_SubmitButton.setEnabled(true);
			m_SubmitButton.setToolTipText("Click here to submit your payment");
		}
		else {
			m_TotalVoucherValue.setForeground(Color.RED);
			m_SubmitButton.setEnabled(false);
			m_SubmitButton.setToolTipText("Select voucher(s) to cover payment");
		}
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(3);
		String valText = "";
		if (tot > amt) {
			// indicate that change will be received
			valText = df.format(tot) + " [" + units + "] (with change)";
		}
		else {
			valText = df.format(tot) + " [" + units + "]";
		}
		m_TotalVoucherValue.setText(valText);
	}

	/**
	 * private utility method to request that the OFS/VP send a payment.
	 * NB: this method will block pending the OFS's reply
	 * @param payDetails the <payDetails/> block, signed and encrypted
	 * @param vReq the <voucherRequest/> block, signed and encrypted, with tokens
	 */
	private void sendMakePayment(String payDetails, VoucherRequest vReq) {
		if (payDetails == null || payDetails.isEmpty() || vReq == null
			 || !vReq.isInit())
		{
			return;
		}
		
		VscState state = m_Plugin.getVscState();
		if (!state.isLoggedIn()) {
			Log.error("Cannot send payments while not logged in!");
			showError("cannot send payment", "not logged in", "log in first");
			return;
		}

		// wait cursor and grey out buttons
		m_ParentHome.setCursor(M_WaitCursor);
		setButtonState(false);

		// build message to send
		final XMPPConnection conn = m_Plugin.getOFSConnection();
		OFSMessage pmt = new OFSMessage();
		pmt.setType(IQ.Type.SET);
		pmt.setFrom(conn.getUser());
		pmt.setTo(conn.getServiceName());
		pmt.setPacketID("pmt_" + state.getNextOFSid());
		pmt.setOpcode("REQ_send_payment");
		pmt.setVoucher_publisher(m_Plugin.getLoginSecrets().getPublisher());
		pmt.setPayDetails(payDetails);
		pmt.setVouchReq(vReq);

		// register a synchronous collector for the reply
		PacketIDFilter pFilter = new PacketIDFilter(pmt.getPacketID());
		PacketCollector repColl = conn.createPacketCollector(pFilter);

		// send it to the server
		conn.sendPacket(pmt);
		//Log.debug(pmt.getOpcode() + " sent to OFS");

		// record outgoing packet
		state.setLastActivity(pmt);

		// wait for reply or timeout
		OFSMessage rep = (OFSMessage) repColl.nextResult(m_Plugin.getTimeout());
		m_ParentHome.setCursor(null);
		setButtonState(true);
		if (rep == null) {
			// got timeout; build error
			rep = new OFSMessage();
			rep.setOpcode("REP_payment_sent");
			rep.setFrom(pmt.getTo());
			rep.setTo(pmt.getFrom());
			rep.setPacketID(pmt.getPacketID());
			rep.setType(IQ.Type.ERROR);
			rep.setErrcode(java.net.HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
			rep.setErrmsg("timed out");
		}
		repColl.cancel();

		// invoke callback with result (calls processPaymentReply() below)
		m_Plugin.getOFSListener().processPacket(rep);
	}

	/**
	 * method to process a reply received from OFS to a payment sent
	 * @param pmtRep the reply message from the OFS
	 */
	public void processPaymentReply(OFSMessage pmtRep) {
		if (pmtRep == null
			 || !pmtRep.getOpcode().equalsIgnoreCase("REP_payment_sent"))
		{
			Log.error("Bad payment reply received");
			return;
		}
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		ArrayList<XMLVoucher> usedVouchers = vtm.getCommittedVouchers();
		TokenTableModel ttm = m_ParentHome.getTokenModel();

		// see whether this reply is an error
		if (pmtRep.getType() == IQ.Type.ERROR) {
			int code = pmtRep.getErrcode();
			// Uncommit vouchers and tokens and display error message.  Note there
			// is a chance tokens were actually spent.
			ttm.uncommitTokens(m_PaymentTokens);
			m_PaymentTokens.clear();
			vtm.uncommitVouchers(usedVouchers);
			switch (code) {
				case HttpURLConnection.HTTP_UNAVAILABLE:
				case HttpURLConnection.HTTP_GATEWAY_TIMEOUT:
					showError("timeout processing payment",
								"code " + code + ": " + pmtRep.getErrmsg(),
								"wait a few minutes, then look in Housekeeping "
								+ "(Refresh) for a confirmation");
					m_NeedTokenSync = true;
					break;

				case HttpURLConnection.HTTP_INTERNAL_ERROR:
					showError("internal error processing payment",
								"code " + code + ": " + pmtRep.getErrmsg(),
								"Wait a few minutes, then look in Housekeeping "
								+ "(Refresh) for a confirmation.\nIf none found, try "
								+ "again. Use the Resync button to synchronize your "
								+ "vouchers and tokens.");
					m_NeedTokenSync = true;
					break;

				default:
					showError("error processing payment",
								"code " + code + ": " + pmtRep.getErrmsg(),
								"wait a few minutes, use Resync and Refresh, and "
								+ "try again");
					break;
			}
			return;
		}

		// record tokens as spent
		ttm.spendTokens(m_PaymentTokens);
		m_PaymentTokens.clear();

		// record vouchers as used up
		m_ParentHome.spendVouchers(usedVouchers);

		/* If we're a payment initiated from another tab (SVX or SVM tabs), then
		 * we need to tell the ConfirmPane not to show the popup, but instead
		 * record back to us (via setLatestConfirmation()) the confirmation
		 * received.  This will allow us to supply it to our invoker (e.g.
		 * AcceptPane), which then will construct the popup properly, combined
		 * with some other information of its own.
		 */
		if (m_InternalTabPmt) {
			m_LatestConfirmation = null;
			ConfirmPane cp = m_TabManager.getConfirmPane();
			cp.setDeferPopup(true);
		}

		// clear existing data entry fields
		resetForNextPayment();

		// just so they don't click on the payment again by mistake
		m_ButtonPane.removeAll();

		// display warning if we have sync pending
		if (pmtRep.syncPending()) {
			showError("warning: payment succeeded but folder updates were queued",
						"internal server timeout",
						"wait a few minutes and click Refresh to complete "
						+ "your payment");
		}
		else {
			// set wait cursor and pull the confirmation record off the DHT
			m_ParentHome.setCursor(M_WaitCursor);

			// ask Housekeeping to query the DHT for an updated records list
			HousekeepingPane hkp = m_TabManager.getHousekeepingPane();
			hkp.refreshDHT(true);

			// NB: on reply we will go to screen 3 (ConfirmPane)
		}
	}

	/**
	 * method to make stored baggage accessible (to the ConfirmPane)
	 * @return the baggage hash table (actual table; side-effects intended)
	 */
	public Hashtable<String, String> getPaymentBaggage() {
		return m_PaymentBaggage;
	}

	/**
	 * method to obtain the most recently received confirmation from the DHT.
	 * (called by classes in ../svx to construct aggregated success popup msg)
	 */
	public XMLConfirmPayment getLatestConfirmation() {
		return m_LatestConfirmation;
	}

	/**
	 * method to set the just-processed confirmation to a payment we made.
	 * (called by ConfirmPane for a payment related to a SVX/SVM order)
	 * @param conf the confirmation record
	 */
	public void setLatestConfirmation(XMLConfirmPayment conf) {
		m_LatestConfirmation = conf;
	}

	/**
	 * method required to implement ActionListener
	 * @param ae the event from the user input object
	 */
	@Override
	public void actionPerformed(ActionEvent ae) {
		Object src = ae.getSource();
		if (src == m_IssuerSel) {
			super.actionPerformed(ae);
			prepDisplay();
			updateGAUdisplay(true);
			return;
		}
		if (src == m_FiatSelection) {
			// because the currency might have changed, this counts as a new amount
			updateGAUdisplay(true);
		}
		else if (src == m_PrevPayeeSelect) {
			// this will trigger the document listener processing
			String payee = (String) m_PrevPayeeSelect.getSelectedItem();
			m_PaymentPayee.setText(payee);
		}
		else if (src instanceof JTextField) {
			// just move the focus to the next item
			JTextField input = (JTextField) src;
			input.transferFocus();
		}
		else {
			Log.error("Unexpected source for combo box action");
		}
	}

	/**
	 * helper method to update the GAU amount (base units) when input changes
	 * @param amtChanged whether the amount changed
	 */
	private void updateGAUdisplay(boolean amtChanged) {
		if (amtChanged) {
			// look up current price for selected fiat currency in module
			String selectedCurr = (String) m_FiatSelection.getSelectedItem();
			AssetPriceCache apc = m_Plugin.getPricingModule();
			double ratio = 1.0;
			String dispUnits = XMLVoucher.getDisplayUnits(m_VoucherAssetType);
			String lookUnits
				= VPAssetConfig.getBasePriceLookupUnits(m_VoucherAssetType);
			Double price = apc.getRelativePrice(lookUnits, selectedCurr);
			if (price != null) {
				ratio = price.doubleValue();
			}

			// convert to voucher base units, parsing leniently
			String enteredVal = m_ValueEntry.getText();
			double vvv = 0.0;
			if (!enteredVal.isEmpty()) {
				try {
					Number val = M_AmtFormatter.parse(enteredVal);
					vvv = val.doubleValue();
				}
				catch (ParseException pe) {
					// this should not happen if PmtDocumentListener did its job
					Log.error("Could not parse pmt amount value", pe);
					return;
				}
			}
			double baseAmt = vvv / ratio;

			// test minimum amount (this test is redundant with the M_AmtFormatter)
			double minimum = 1.0 / (10 ^ XMLVoucher.M_Precision);
			if (baseAmt < minimum) {
				if (amtChanged) {
					m_ContButton.setEnabled(false);
					m_ContButton.setToolTipText("Enter a payment amount more than "
														+ minimum + " " + dispUnits);
				}
			}

			// update display with valid value
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
			df.setMaximumFractionDigits(XMLVoucher.M_Precision);
			df.setMinimumFractionDigits(XMLVoucher.M_Precision);
			df.setGroupingUsed(false);
			m_PaymentAmount.setText(df.format(baseAmt));
		}

		// if we also have a payee, activate button to continue
		String curAmtTxt = m_PaymentAmount.getText();
		if (!curAmtTxt.isEmpty() && !curAmtTxt.equals("0.00")
			 && !curAmtTxt.equals("0,00") && !curAmtTxt.equals("0"))
		{
			if (m_HavePayee) {
				m_ContButton.setToolTipText("Click here to continue payment");
				m_ContButton.setEnabled(true);
			}
			else {
				m_ContButton.setToolTipText("Enter a payee for your payment");
				m_ContButton.setEnabled(false);
			}
		}
		else {
			if (m_HavePayee) {
				m_ContButton.setToolTipText("Enter an amount for your payment");
				m_ContButton.setEnabled(false);
			}
			else {
				m_ContButton.setToolTipText("Enter a payee and a payment amount");
				m_ContButton.setEnabled(false);
			}
		}
	}

	/**
	 * helper method to build the list of previous payees, from receipts
	 * @return the list of voucher safes from which receipts have been received,
	 * or null if there are none
	 */
	private Vector<String> getPreviousPayees() {
		ReceiptTableModel rtm
			= m_TabManager.getHousekeepingPane().getReceiptModel();
		if (rtm.getRowCount() == 0) {
			return null;
		}

		// establish publisher
		VsSecrets sekrits = m_Plugin.getLoginSecrets();
		String vp = sekrits.getPublisher();
		int atIdx = vp.lastIndexOf('@');
		vp = vp.substring(atIdx);

		// build a list of all payees from receipts where payee is us
		Hashtable<XMLReceipt, String> receipts = rtm.getReceiptList();
		TreeSet<String> payees = new TreeSet<String>();
		for (XMLReceipt receipt : receipts.keySet()) {
			String payeeVS = receipt.getPayee();
			if (!payeeVS.equals(sekrits.getVSnumber())) {
				int vpIdx = payeeVS.indexOf(vp);
				payeeVS = payeeVS.substring(0, vpIdx);
				payees.add(payeeVS);
			}
		}
		if (payees.isEmpty()) {
			// didn't find any
			return null;
		}
		return new Vector<String>(payees);
	}

	/**
	 * helper method to reset fields for the next payment
	 */
	private void resetForNextPayment() {
		// reset for next payment
		m_Phase = 1;
		m_IssuerSel.setEnabled(true);
		m_Advanced = false;
		m_ValueEntry.getDocument().removeDocumentListener(m_PaymentListener);
		m_ValueEntry.setText("0");
		m_ValueEntry.getDocument().addDocumentListener(m_PaymentListener);
		m_ValueEntry.setToolTipText("Enter a positive decimal number");
		if (m_FiatSelection != null) {
			m_FiatSelection.setToolTipText("Select desired currency code");
			m_FiatSelection.setEnabled(true);
		}
		m_MessageField.setText("");
		for (int iii = 0; iii < m_BaggageFields.length; iii++) {
			m_BaggageFields[iii].setText("");
		}
		for (int jjj = 0; jjj < m_BaggageValues.length; jjj++) {
			m_BaggageValues[jjj].setText("");
		}
		m_PaymentAmount.setText("");
		m_BagFields.clear();
		if (m_WebDecodeKey != null) {
			m_WebDecodeKey.setText("");
		}
		m_TotalVoucherValue.setText("");
		m_PaymentPayee.getDocument().removeDocumentListener(m_PayeeListener);
		m_PaymentPayee.setText("");
		m_PaymentPayee.getDocument().addDocumentListener(m_PayeeListener);

		// discard any web order we were working on
		if (m_SelectedSiteMsg != null && m_SelectedSiteMsg.isDecoded()) {
			m_Plugin.getWebListener().purgeWebMessage(m_SelectedSiteMsg);
			m_SelectedSiteMsg = null;
		}
		if (m_WebsiteMessages != null) {
			m_WebsiteMessages.removeAllItems();
			m_WebsiteMessages = null;
		}
		m_InternalTabPmt = false;
	}

	/**
	 * method to queue up a modification to a text component from within a
	 * document listener
	 * @param txtComp the text component to update
	 * @param newText the replacement text
	 */
	private void queueTextMod(JTextComponent txtComp, String newText) {
		if (txtComp == null || newText == null) {
			return;
		}

		// do not allow simultaneous queued mods
		if (m_DoingUndo) {
			return;
		}

		// create and start background modification object
		TextComponentMod modifier = new TextComponentMod(txtComp, newText);
		m_DoingUndo = true;
		SwingUtilities.invokeLater(modifier);
	}

	/**
	 * method to programmatically set the web message being worked on.
	 * (used with payments originated from the SVX or SVM tabs)
	 * @param webMsg the message to set (should be in WebListener's list)
	 */
	public void setSelectedSiteMsg(WebMessage webMsg) {
		if (webMsg != null && webMsg.isDecoded()) {
			m_SelectedSiteMsg = webMsg;
			// override default number of days also
			m_TTLdays.setSelectedItem("1");
			m_InternalTabPmt = true;
		}
	}
	
	// END methods
}
