/*
 * last modified---
 * 	05-09-18/kbw/ for better graphics performance when deleting non-sequential
 * 					  receipts (i.e. all Out or all In), temp remove the listener
 * 	08-05-14/jet/ Spark plugin version
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide a UI widget to handle display and manipulation of stored receipts
 */

package org.silentvault.spark.ui.wallet;

import javax.swing.*;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Component;
import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.awt.print.PrinterException;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Clipboard;
import java.awt.Toolkit;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;
import javax.swing.table.DefaultTableColumnModel;

import org.dom4j.DocumentException;

import org.jivesoftware.resource.SparkRes;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.filter.PacketIDFilter;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.TreeMap;
import java.util.Date;
import java.text.MessageFormat;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.FileNotFoundException;


/**
 * Extends WorkPane to provide a screen for listing and manipulating stored
 * existing receipts.
 *
 * @author Kevin Wilkerson
 */
public final class ReceiptsPane extends WorkPane {
	// BEGIN data members
	/**
	 * tool tips for receipt table columns
	 */
	protected final String[]	M_ReceiptToolTips =
	{	"Click here to select this receipt",
		"Type of receipt",
		"Counterparty wallet ID",
		"The value in voucher units",
		"Date/time receipt was issued"	};

	/**
	 * the table model for the receipt records table
	 */
	private ReceiptTableModel	m_ReceiptTableModel;

	/**
	 * the display table for receipts
	 */
	private JTable					m_ReceiptTable;

	/**
	 * incoming receipts, not yet decoded
	 */
	private Hashtable<String, String>	m_IncomingReceipts;

	/**
	 * action for deleting a single selected receipt
	 */
	private AbstractAction		m_PurgeAction;

	/**
	 * action for displaying the XML of a selected receipt
	 */
	private AbstractAction		m_XMLAction;

	/**
	 * action for printing a receipt's XML
	 */
	private AbstractAction		m_PrintAction;

	/**
	 * action for copying receipt XML to clipboard
	 */
	private AbstractAction		m_ClipboardAction;

	/**
	 * action for initiating download of receipts
	 */
	private AbstractAction		m_InitDownloadAction;

	/**
	 * action for performing downloading of receipts
	 */
	private AbstractAction		m_DoDownloadAction;

	/**
	 * action for initiating a mass purge of receipts
	 */
	private AbstractAction		m_InitPurgeAction;

	/**
	 * action for purging a group of receipts
	 */
	private AbstractAction		m_MassPurgeAction;

	/**
	 * action for selecting the file in which to save the receipt data
	 */
	private AbstractAction		m_SelectFileAction;

	/**
	 * UI element for XML receipt download mode
	 */
	private JRadioButton			m_ModeXML;

	/**
	 * UI element for CSV receipt download mode
	 */
	private JRadioButton			m_ModeCSV;

	/**
	 * the button group for the download selectors
	 */
	private ButtonGroup			m_ModeGroup;

	/**
	 * label for download data mode selection
	 */
	private JLabel					m_ModeLabel;

	/**
	 * the currently selected receipt
	 */
	private XMLReceipt			m_SelectedReceipt;

	/**
	 * UI element to display receipt XML in popup JFrame
	 */
	private JTextArea				m_XML;

	/**
	 * UI element for specifying receipt type Incoming
	 */
	private JRadioButton			m_TypeIncoming;

	/**
	 * UI element for specifying receipt type outgoing
	 */
	private JRadioButton			m_TypeOutgoing;

	/**
	 * UI element for specifying both types of receipts
	 */
	private JRadioButton			m_TypeBoth;

	/**
	 * label for receipt type selection
	 */
	private JLabel					m_TypeLabel;

	/**
	 * button group for the receipt type selectors
	 */
	private ButtonGroup			m_TypeGroup;

	/**
	 * UI element for displaying or entering the currently selected save file
	 */
	private JLabel					m_DownloadSaveFile;

	/**
	 * UI element for graphically selecting a save file
	 */
	private JFileChooser			m_SaveFilePicker;

	/**
	 * label for download save file
	 */
	private JLabel					m_SaveFileLabel;

	/**
	 * popup frame for receipt downloads
	 */
	private JFrame					m_PopDownload;

	/**
	 * popup frame for receipt mass purge
	 */
	private JFrame					m_PopPurge;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the HomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public ReceiptsPane(WalletClient plugin, HomePane home, WTabManager mgr) {
		super(plugin, home, mgr);
		// take receipt table model from Housekeeping pane so we use the same one
		m_ReceiptTableModel
			= m_TabManager.getHousekeepingPane().getReceiptModel();
		m_ReceiptTable = mkNewReceiptTable();
		m_IncomingReceipts = new Hashtable<String, String>();

		// set user instructional text
		String usrTxt = "An incoming receipt is from someone you paid.  An "
							+ "outgoing receipt was issued to someone who paid you."
							+ "<br/><br/>Click \"Show XML\" to see the raw signed "
							+ "receipt message.  This XML message can be used to "
							+ "prove to a third party that a payment has been made. "
							+ "The signature can be verified at "
							+ "https://vouchi.com/validate.cfm";
		m_UserInstructs = usrTxt;

		// define download save file elements
		m_DownloadSaveFile = new JLabel();
		m_DownloadSaveFile.setFont(m_ParentHome.M_DataFont);
		m_DownloadSaveFile.setForeground(UIManager.getColor("standardGray"));
		m_DownloadSaveFile.setBorder(BorderFactory.createLineBorder(Color.BLACK));
		m_DownloadSaveFile.setToolTipText("path to selected save file");
		m_SaveFilePicker = new JFileChooser();
		m_SaveFilePicker.setMultiSelectionEnabled(false);
		m_SaveFileLabel = new JLabel("Save download to:");
		m_SaveFileLabel.setFont(m_ParentHome.M_DataFont);
		m_SaveFileLabel.setForeground(UIManager.getColor("standardGray"));
		m_SaveFileLabel.setHorizontalAlignment(SwingConstants.LEFT);
		m_SelectFileAction = new AbstractAction("Browse") {
			public void actionPerformed(ActionEvent actionEvent) {
				int rC = m_SaveFilePicker.showSaveDialog(m_ParentHome);
				if (rC == JFileChooser.APPROVE_OPTION) {
					File saveFile = m_SaveFilePicker.getSelectedFile();
					try {
						m_DownloadSaveFile.setText(saveFile.getCanonicalPath());
					}
					catch (IOException ioe) { /* ignore */ }
				}
				else {
					m_DownloadSaveFile.setText("");
				}
				m_PopDownload.toFront();
			}
		};

		// define download mode selection buttons
		m_ModeXML = new JRadioButton("XML Format", true);
		m_ModeXML.setToolTipText("actual XML representation, digitally signed");
		m_ModeCSV = new JRadioButton("CSV Format", false);
		m_ModeCSV.setToolTipText("comma separated values");
		m_ModeGroup = new ButtonGroup();
		m_ModeGroup.add(m_ModeXML);
		m_ModeGroup.add(m_ModeCSV);
		m_ModeLabel = new JLabel("Data Save Mode:");
		m_ModeLabel.setFont(m_ParentHome.M_DataFont);
		m_ModeLabel.setForeground(UIManager.getColor("standardGray"));
		m_ModeLabel.setHorizontalAlignment(SwingConstants.LEFT);

		// define receipt type selection buttons
		m_TypeIncoming = new JRadioButton("From others", false);
		m_TypeIncoming.setToolTipText("Incoming receipts");
		m_TypeOutgoing = new JRadioButton("Issued to others", true);
		m_TypeOutgoing.setToolTipText("Outgoing receipts");
		m_TypeBoth = new JRadioButton("Both", false);
		m_TypeBoth.setToolTipText("All receipts");
		m_TypeGroup = new ButtonGroup();
		m_TypeGroup.add(m_TypeIncoming);
		m_TypeGroup.add(m_TypeOutgoing);
		m_TypeGroup.add(m_TypeBoth);
		m_TypeLabel = new JLabel("Specify Receipt Type:");
		m_TypeLabel.setFont(m_ParentHome.M_DataFont);
		m_TypeLabel.setForeground(UIManager.getColor("standardGray"));
		m_TypeLabel.setHorizontalAlignment(SwingConstants.LEFT);

		// define button actions
		setPurgeAction();
		setXMLAction();
		setPrintAction();
		setClipboardAction();
		setDoDownloadAction();
		setInitDownloadAction();
		setMassPurgeAction();
		setInitPurgeAction();

		m_HeadingPane.setLayout(new GridLayout(1, 3));
	}

	/**
	 * helper method to create a JTable (for receipts), also providing column
	 * tooltip definitions
	 */
	private JTable mkNewReceiptTable() {
		JTable newTab = new JTable(m_ReceiptTableModel) {
			// provide table header tool tips
			protected JTableHeader createDefaultTableHeader() {
				return new JTableHeader(columnModel) {
					public String getToolTipText(MouseEvent me) {
						String tip = null;
						Point ppp = me.getPoint();
						int index = columnModel.getColumnIndexAtX(ppp.x);
						int realIndex = columnModel.getColumn(index).getModelIndex();
						return M_ReceiptToolTips[realIndex];
					}
				};
			}
			public boolean isCellEditable(int row, int col) {
				// objects in the select column can be editable, others are not
				if (col == 0) {
					return true;
				}
				return false;
			}
		};
		newTab.setCellSelectionEnabled(false);
		// this size should be same as WorkPane.m_WorkArea
		newTab.setPreferredScrollableViewportSize(
													UIManager.getDimension("vouchTableDim"));
		newTab.setFillsViewportHeight(true);

		// set special renderer for all columns
		ReceiptRenderer rr = new ReceiptRenderer(m_ReceiptTableModel, m_Plugin);
		newTab.setDefaultRenderer(JRadioButton.class, rr);
		newTab.setDefaultRenderer(String.class, rr);
		newTab.setDefaultRenderer(Double.class, rr);
		newTab.setDefaultRenderer(Date.class, rr);

		// configure the table columns and renderers
		configTableColumns(newTab);

		// set default sorter
		newTab.setAutoCreateRowSorter(true);
		newTab.setAutoCreateColumnsFromModel(true);

		// set spacing preferences
		newTab.setRowHeight(20);
		newTab.setIntercellSpacing(new Dimension(2, 2));
		newTab.setShowGrid(true);
		newTab.setGridColor(UIManager.getColor("TableHeader.background"));

		// establish table as listener for model
		m_ReceiptTableModel.addTableModelListener(newTab);
		return newTab;
	}

	/**
	 * helper method to define the button action to purge a selected receipt
	 */
	private void setPurgeAction() {
		m_PurgeAction = new AbstractAction("Delete selected") {
			public void actionPerformed(ActionEvent actionEvent) {
				m_SelectedReceipt = m_ReceiptTableModel.getSelectedReceipt();
				if (m_SelectedReceipt == null) {
					showError("cannot delete receipt", "none selected",
								"select a receipt to be deleted");
					return;
				}

				// put up a confirm dialog to make sure the user is serious
				String confTxt = "Are you sure you want to delete this receipt?\n"
									+ "If you delete a receipt it can't be undone, "
									+ "it's gone forever.";
				int conf = JOptionPane.showConfirmDialog(
																m_TabManager.getReceiptsPane(),
										  						confTxt,
										  						"Confirm Deletion",
																JOptionPane.OK_CANCEL_OPTION,
																JOptionPane.QUESTION_MESSAGE);
				if (conf == 0) {
					// get the associated filename and send the message to the SDS
					ArrayList<String> delFiles = new ArrayList<String>(1);
					Hashtable<XMLReceipt, String> allReceipts
						= m_ReceiptTableModel.getReceiptList();
					String file = allReceipts.get(m_SelectedReceipt);
					if (file == null || file.isEmpty()) {
						Log.error("filename not found for receipt: "
									+ m_SelectedReceipt.toXML());
						showError("cannot delete receipt", "filename not found");
						return;
					}
					delFiles.add(file);
					sendDeleteReceipts(delFiles);
				}
			}
		};
	}

	/**
	 * helper method to define the button action to show XML for a receipt
	 */
	private void setXMLAction() {
		m_XMLAction = new AbstractAction("Show XML") {
			public void actionPerformed(ActionEvent actionEvent) {
				m_SelectedReceipt = m_ReceiptTableModel.getSelectedReceipt();
				if (m_SelectedReceipt == null) {
					showError("cannot show receipt XML", "none selected",
								"select a single receipt to display");
					return;
				}

				// create a frame and put the XML inside it
				JFrame popup
					= new JFrame("XML Receipt from " + m_SelectedReceipt.getPayee());
				popup.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
				m_XML = new JTextArea(m_SelectedReceipt.toXML(), 10, 40);
				m_XML.setEditable(false);
				m_XML.setLineWrap(true);
				m_XML.setWrapStyleWord(false);
				m_XML.setFont(m_ParentHome.M_DataFont);
				popup.add(m_XML, BorderLayout.CENTER);
				popup.setSize(new Dimension(400, 300));
				popup.setLocationRelativeTo(m_TabManager);
				m_Plugin.registerPopup(popup);

				// supply buttons
				JPanel buttonPane = new JPanel();
				buttonPane.setLayout(new BoxLayout(buttonPane, BoxLayout.X_AXIS));
				buttonPane.setOpaque(true);
				buttonPane.add(Box.createHorizontalStrut(30));		// spacer
				JButton printButton = new JButton(m_PrintAction);
				printButton.setToolTipText("Click here to print this receipt XML");
				printButton.setFont(m_Plugin.M_ButtonFont);
				buttonPane.add(printButton);
				buttonPane.add(Box.createHorizontalStrut(140));		// spacer

				JButton copyButton = new JButton(m_ClipboardAction);
				copyButton.setToolTipText("Click here to copy XML text to the "
													+ "clipboard");
				copyButton.setFont(m_Plugin.M_ButtonFont);
				buttonPane.add(copyButton);
				popup.add(buttonPane, BorderLayout.SOUTH);
				popup.setVisible(true);
			}
		};
	}

	/**
	 * helper method to define the button action to print a receipt's XML
	 */
	private void setPrintAction() {
		m_PrintAction = new AbstractAction("Print") {
			public void actionPerformed(ActionEvent actionEvent) {
				m_SelectedReceipt = m_ReceiptTableModel.getSelectedReceipt();
				if (m_XML == null || m_SelectedReceipt == null) {
					showError("cannot print receipt", "none selected",
								"please select a receipt");
					return;
				}
				MessageFormat head = new MessageFormat("Receipt from "
																+ m_SelectedReceipt.getPayee());
				boolean printed = true;
				String err = "printing canceled";
				try {
					printed = m_XML.print(head, new MessageFormat("Page - {0}"),
												 true, null, null, false);
				}
				catch (PrinterException pe) {
					Log.error("error printing receipt XML", pe);
					printed = false;
					err = pe.getMessage();
				}
				catch (SecurityException se) {
					Log.error("error printing receipt XML", se);
					printed = false;
					err = se.getMessage();
				}
				if (!printed) {
					showError("unable to print receipt", err);
				}
			}
		};
	}

	/**
	 * helper method to define the button action for copy of XML to clipboard
	 */
	private void setClipboardAction() {
		m_ClipboardAction = new AbstractAction("Copy to clipboard") {
			public void actionPerformed(ActionEvent actionEvent) {
				if (m_XML == null) {
					showError("cannot copy receipt to clipboard", "none selected",
								"please select a receipt");
					return;
				}
				StringSelection ss = new StringSelection(m_XML.getText());
				Toolkit tk = Toolkit.getDefaultToolkit();
				Clipboard sysClip = null;
				sysClip = tk.getSystemClipboard();
				sysClip.setContents(ss, ss);
			}
		};
	}

	/**
	 * helper method to define the button action to initiate a receipt download
	 */
	private void setInitDownloadAction() {
		m_InitDownloadAction = new AbstractAction("Download") {
			public void actionPerformed(ActionEvent actionEvent) {
				// create a frame and put the selection screen inside it
				m_PopDownload = new JFrame("Download Receipt Data");
				m_PopDownload.setDefaultCloseOperation(
															WindowConstants.DISPOSE_ON_CLOSE);
				m_PopDownload.setSize(new Dimension(350, 300));
				m_PopDownload.setLocationRelativeTo(m_TabManager);
				m_Plugin.registerPopup(m_PopDownload);
				JPanel popPane = new JPanel();
				popPane.setLayout(new BoxLayout(popPane, BoxLayout.Y_AXIS));

				// specify receipt type
				popPane.add(Box.createVerticalStrut(20));
				JPanel labPane = new JPanel();
				labPane.setLayout(new BoxLayout(labPane, BoxLayout.X_AXIS));
				labPane.add(m_TypeLabel);
				labPane.add(Box.createHorizontalStrut(175));
				popPane.add(labPane);
				popPane.add(Box.createVerticalStrut(5));
				JPanel typePane = new JPanel();
				typePane.setLayout(new BoxLayout(typePane, BoxLayout.X_AXIS));
				typePane.setOpaque(true);
				typePane.add(m_TypeIncoming);
				typePane.add(Box.createHorizontalStrut(20));
				typePane.add(m_TypeOutgoing);
				typePane.add(Box.createHorizontalStrut(20));
				typePane.add(m_TypeBoth);
				popPane.add(typePane);
				popPane.add(Box.createVerticalStrut(20));

				// specify download mode
				JPanel labPane1 = new JPanel();
				labPane1.setLayout(new BoxLayout(labPane1, BoxLayout.X_AXIS));
				labPane1.add(m_ModeLabel);
				labPane1.add(Box.createHorizontalStrut(100));
				popPane.add(labPane1);
				popPane.add(Box.createVerticalStrut(5));
				JPanel modePane = new JPanel();
				modePane.setLayout(new BoxLayout(modePane, BoxLayout.X_AXIS));
				modePane.setOpaque(true);
				modePane.add(m_ModeXML);
				modePane.add(Box.createHorizontalStrut(20));
				modePane.add(m_ModeCSV);
				popPane.add(modePane);
				popPane.add(Box.createVerticalStrut(20));

				// specify save file destination
				JPanel labPane2 = new JPanel();
				labPane2.setLayout(new BoxLayout(labPane2, BoxLayout.X_AXIS));
				labPane2.add(m_SaveFileLabel);
				labPane2.add(Box.createHorizontalStrut(170));
				popPane.add(labPane2);
				popPane.add(Box.createVerticalStrut(5));
				JPanel savePane = new JPanel();
				savePane.setLayout(new BoxLayout(savePane, BoxLayout.X_AXIS));
				savePane.setOpaque(true);
				savePane.add(m_DownloadSaveFile);
				savePane.add(Box.createHorizontalStrut(20));
				JButton browse = new JButton(m_SelectFileAction);
				browse.setFont(m_Plugin.M_ButtonFont);
				browse.setToolTipText("Click here to select a save file");
				savePane.add(browse);
				popPane.add(savePane);
				popPane.add(Box.createVerticalStrut(20));

				// supply cancel and download buttons
				JPanel buttonPane = new JPanel();
				buttonPane.setLayout(new BoxLayout(buttonPane, BoxLayout.X_AXIS));
				AbstractAction cancelAction = new AbstractAction("Cancel") {
					public void actionPerformed(ActionEvent ae) {
						m_Plugin.deregisterPopup(m_PopDownload);
						m_PopDownload.dispose();
					}
				};
				JButton cancel = new JButton(cancelAction);
				cancel.setFont(m_Plugin.M_ButtonFont);
				buttonPane.add(cancel);
				buttonPane.add(Box.createHorizontalStrut(55));
				JButton download = new JButton(m_DoDownloadAction);
				download.setToolTipText("Click here to download and save "
												+ "receipt data");
				download.setFont(m_Plugin.M_ButtonFont);
				buttonPane.add(download);
				popPane.add(buttonPane);
				popPane.add(Box.createVerticalStrut(20));
				m_PopDownload.add(popPane);
				m_PopDownload.setVisible(true);
			}
		};
	}

	/**
	 * helper method to define the button action to initiate a receipt purge
	 */
	private void setInitPurgeAction() {
		m_InitPurgeAction = new AbstractAction("Purge") {
			public void actionPerformed(ActionEvent actionEvent) {
				// create a frame and put the selection screen inside it
				m_PopPurge = new JFrame("Purge Receipt Data");
				m_PopPurge.setDefaultCloseOperation(
															WindowConstants.DISPOSE_ON_CLOSE);
				m_PopPurge.setSize(new Dimension(350, 150));
				m_PopPurge.setLocationRelativeTo(m_TabManager);
				m_Plugin.registerPopup(m_PopPurge);
				JPanel popPane = new JPanel();
				popPane.setLayout(new BoxLayout(popPane, BoxLayout.Y_AXIS));

				// specify receipt type
				popPane.add(Box.createVerticalStrut(20));
				JPanel labPane = new JPanel();
				labPane.setLayout(new BoxLayout(labPane, BoxLayout.X_AXIS));
				labPane.add(m_TypeLabel);
				labPane.add(Box.createHorizontalStrut(175));
				popPane.add(labPane);
				popPane.add(Box.createVerticalStrut(5));
				JPanel typePane = new JPanel();
				typePane.setLayout(new BoxLayout(typePane, BoxLayout.X_AXIS));
				typePane.setOpaque(true);
				typePane.add(m_TypeIncoming);
				typePane.add(Box.createHorizontalStrut(20));
				typePane.add(m_TypeOutgoing);
				typePane.add(Box.createHorizontalStrut(20));
				typePane.add(m_TypeBoth);
				popPane.add(typePane);
				popPane.add(Box.createVerticalStrut(20));

				// supply cancel and purge buttons
				JButton purge = new JButton(m_MassPurgeAction);
				purge.setToolTipText("Click here to purge receipt data");
				purge.setFont(m_Plugin.M_ButtonFont);
				JPanel buttonPane = new JPanel();
				buttonPane.setLayout(new BoxLayout(buttonPane, BoxLayout.X_AXIS));
				AbstractAction cancelAction = new AbstractAction("Cancel") {
					public void actionPerformed(ActionEvent ae) {
						m_Plugin.deregisterPopup(m_PopPurge);
						m_PopPurge.dispose();
					}
				};
				JButton cancel = new JButton(cancelAction);
				cancel.setFont(m_Plugin.M_ButtonFont);
				buttonPane.add(cancel);
				buttonPane.add(Box.createHorizontalStrut(55));
				buttonPane.add(purge);
				popPane.add(buttonPane);
				popPane.add(Box.createVerticalStrut(20));
				m_PopPurge.add(popPane);
				m_PopPurge.setVisible(true);
			}
		};
	}

	/**
	 * helper method to define the button action to perform a receipt download
	 */
	private void setDoDownloadAction() {
		m_DoDownloadAction = new AbstractAction("Download") {
			public void actionPerformed(ActionEvent actionEvent) {
				int numReceipts = m_ReceiptTableModel.getRowCount();
				// no point if we have no receipts
				if (numReceipts == 0) {
					showError("cannot download receipt data", "no receipts exist");
					m_Plugin.deregisterPopup(m_PopDownload);
					return;
				}

				// make sure we have a save file selected
				File saveFile = m_SaveFilePicker.getSelectedFile();
				if (saveFile == null) {
					showError("cannot download receipt data", "no file selected",
								"please select a save file");
					m_PopDownload.toFront();
					return;
				}
				if (saveFile.exists()) {
					saveFile.delete();
				}

				// make sure we can open it
				try {
					saveFile.createNewFile();
					FileWriter fw = new FileWriter(saveFile);
					// output a column header in CSV mode
					final String hd = "amount,units,asset,payerVS,payeeVS,voucher "
											+ "serial,timestamp,\"baggagefield:value\"\n";
					if (m_ModeCSV.isSelected()) {
						fw.write(hd, 0, hd.length());
					}

					// loop through all the receipt records in table display order
					VsSecrets sekrits = m_Plugin.getLoginSecrets();
					for (int iii = 0; iii < numReceipts; iii++) {
						int row = m_ReceiptTable.convertRowIndexToModel(iii);
						XMLReceipt rct = m_ReceiptTableModel.getReceiptAtRow(row);
						if (rct == null) {
							continue;
						}
						boolean incoming
							= !rct.getPayee().equals(sekrits.getVSnumber());

						// see if we should skip over this receipt
						if (!m_TypeBoth.isSelected()) {
							if (m_TypeIncoming.isSelected() && !incoming) {
								continue;
							}
							else if (m_TypeOutgoing.isSelected() && incoming) {
								continue;
							}
						}

						// use correct output format based on setting
						String line = null;
						if (m_ModeXML.isSelected()) {
							line = rct.toXML() + "\n";
						}
						else {
							line = rct.emitAsCSV() + "\n";
						}
						fw.write(line, 0, line.length());
					}
					fw.flush();
					fw.close();
				}
				catch (FileNotFoundException fnfe) {
					showError("cannot download receipt data",
						"could not open indicated save file, " + saveFile.getName(),
								"check directory permissions");
					return;
				}
				catch (IOException ioe) {
					Log.error("Could not write downloaded receipt data to "
								+ saveFile.getName(), ioe);
					showError("cannot download receipt data",
								"error writing to save file",
								"check file and directory permissions");
					return;
				}
				finally {
					m_Plugin.deregisterPopup(m_PopDownload);
					m_PopDownload.dispose();
				}
			}
		};
	}

	/**
	 * helper method to define the button action to perform a purge of receipts
	 */
	private void setMassPurgeAction() {
		m_MassPurgeAction = new AbstractAction("Delete Receipts") {
			public void actionPerformed(ActionEvent actionEvent) {
				int numReceipts = m_ReceiptTableModel.getRowCount();
				// no point if we have no receipts
				if (numReceipts == 0) {
					showError("cannot purge receipt data", "no receipts exist");
					m_Plugin.deregisterPopup(m_PopPurge);
					return;
				}

				// loop through all receipts, winnowing out undesired ones
				VsSecrets sekrits = m_Plugin.getLoginSecrets();
				ArrayList<XMLReceipt> delReceipts
					= new ArrayList<XMLReceipt>(numReceipts);
				for (int iii = 0; iii < numReceipts; iii++) {
					XMLReceipt rct = m_ReceiptTableModel.getReceiptAtRow(iii);
					if (rct == null) {
						continue;
					}
					boolean incoming = !rct.getPayee().equals(sekrits.getVSnumber());

					// see if we should skip over this receipt
					if (!m_TypeBoth.isSelected()) {
						if (m_TypeIncoming.isSelected() && !incoming) {
							continue;
						}
						else if (m_TypeOutgoing.isSelected() && incoming) {
							continue;
						}
					}
					delReceipts.add(rct);
				}
				if (delReceipts.isEmpty()) {
					showError("cannot purge receipt data",
								"no receipts match selection criteria",
								"check incoming/outgoing setting");
					m_PopPurge.toFront();
					return;
				}

				// put up a confirm dialog to make sure the user is serious
				String confTxt = "Are you sure you want to delete these "
									+ delReceipts.size() + " receipts?\n"
									+ "If you delete receipts it can't be undone, "
									+ "they're gone forever.";
				int conf = JOptionPane.showConfirmDialog(
																m_TabManager.getReceiptsPane(),
										  						confTxt,
										  						"Confirm Deletion",
																JOptionPane.OK_CANCEL_OPTION,
																JOptionPane.QUESTION_MESSAGE);
				if (conf == 0) {
					// get the associated filenames and send the message to the SDS
					ArrayList<String> delFiles
						= new ArrayList<String>(delReceipts.size());
					Hashtable<XMLReceipt, String> allReceipts
						= m_ReceiptTableModel.getReceiptList();
					for (XMLReceipt delRct : delReceipts) {
						String file = allReceipts.get(delRct);
						if (file == null || file.isEmpty()) {
							Log.error("filename not found for receipt: "
										+ delRct.toXML());
						}
						else {
							delFiles.add(file);
						}
					}
					sendDeleteReceipts(delFiles);
				}
				m_Plugin.deregisterPopup(m_PopPurge);
				m_PopPurge.dispose();
			}
		};
	}

	/**
	 * Helper method to configure the columns and renderers of the receipts
	 * table.  This needs to be called when the table is created, and after
	 * each change to the table.
	 * @param tab the table to configure (== m_ReceiptTable)
	 */
	private void configTableColumns(JTable tab) {
		// set special renderer for column headers, to cause right justify
		ReceiptHeaderRenderer rhr = new ReceiptHeaderRenderer();

		// make sure timestamp column is wide enough to show the whole date
		TableColumnModel tcm = tab.getColumnModel();
		TableColumn tc = tcm.getColumn(4);
		tc.setMinWidth(135);
		tc.setHeaderRenderer(rhr);

		// make sure amount column is wide enough to show GAU
		tc = tcm.getColumn(3);
		tc.setMinWidth(80);
		tc.setHeaderRenderer(rhr);

		// make sure counterparty column is wide enough to hold VS#
		tc = tcm.getColumn(2);
		tc.setMinWidth(85);
		tc.setHeaderRenderer(rhr);

		// make sure type column is wide enough to hold incoming/outgoing
		tc = tcm.getColumn(1);
		tc.setMaxWidth(45);
		tc.setHeaderRenderer(rhr);

		// set max width on select column, and set cell editor
		tc = tcm.getColumn(0);
		tc.setMaxWidth(30);
		tc.setHeaderRenderer(rhr);
		ReceiptSelector rs = new ReceiptSelector(m_Plugin, m_ReceiptTableModel);
		tc.setCellEditor(rs);
	}

	/**
	 * record receipts found on the SDS
	 * @param receipts the table of encrypted receipts found, keyed by filenames
	 */
	public void recordReceipts(Hashtable<String, String> receipts) {
		if (receipts == null || receipts.isEmpty()) {
			// nothing to do
			return;
		}
		m_IncomingReceipts.clear();
		m_IncomingReceipts.putAll(receipts);
		m_ParentHome.setCursor(null);
	}

	/**
	 * rebuild the receipts list (called by WTabManager only)
	 */
	public void rebuildReceiptList() {
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		Hashtable<XMLReceipt, String> existingReceipts
			= m_ReceiptTableModel.getReceiptList();
		Hashtable<XMLReceipt, String> newReceipts
			= new Hashtable<XMLReceipt, String>();

		// Examine all records returned.  Use a progress bar as we decrypt each.
		JProgressBar decodeProg = new JProgressBar();
		decodeProg.setString("Decoding " + m_IncomingReceipts.size()
									+ " receipt records in wallet, please wait");
		decodeProg.setStringPainted(true);
		decodeProg.setMinimum(0);
		decodeProg.setMaximum(m_IncomingReceipts.size());
		decodeProg.setOpaque(true);
		decodeProg.setBackground(Color.LIGHT_GRAY);
		decodeProg.setFont(m_ParentHome.M_DataFont);
		MenuPane mp = m_TabManager.getMenuPane();
		if (!m_IncomingReceipts.isEmpty()) {
			m_ParentHome.setCursor(M_WaitCursor);
			JPanel statBar = m_Plugin.getStatusBar();
			statBar.removeAll();
			statBar.add(decodeProg);
			statBar.invalidate();
			statBar.revalidate();
			statBar.repaint();
			mp.disableMenu();
		}
		int rCnt = 0;
		for (String fname : m_IncomingReceipts.keySet()) {
			decodeProg.setValue(rCnt+1);

			// see whether we already have this receipt decoded in our list
			if (existingReceipts.containsValue(fname)) {
				// no need to decrypt this file again
				rCnt += 1;
				continue;
			}

			String encRec = m_IncomingReceipts.get(fname);
			XMLReceipt receipt = null;

			// attempt decrypt
			String decRec = EncodingUtils.getStrFromBase64PubkeyEnc(encRec,
																			secrets.getPrivKey());
			if (decRec == null || decRec.isEmpty()) {
				Log.error("Could not decrypt a receipt, filename " + fname);
			}
			else {
				try {
					receipt = XMLReceipt.getReceiptRep(decRec);
					rCnt += 1;
					newReceipts.put(receipt, fname);
				}
				catch (DocumentException de) {
					Log.error("Unable to parse receipt", de);
				}
			}
		}
		if (!m_IncomingReceipts.isEmpty()) {
			m_Plugin.setDefaultStatusBar();
			mp.enableMenu();
			m_ParentHome.setCursor(null);

			// rebuild configuration of table columns, since data changed
			m_ReceiptTableModel.mergeReceipts(newReceipts);
		}

		// double-check that all worked
		int diff = m_IncomingReceipts.size() - rCnt;
		if (diff > 0) {
			Log.error("Error parsing " + diff + " receipts");
			showError("warning: " + diff + " receipts "
						+ "would not decrypt or did not parse");
		}
		m_IncomingReceipts.clear();
	}

	// method required to extend WorkPane
	/**
	 * method to prepare display for this sub-class
	 */
	public void prepDisplay() {
		// set our heading pane
		buildValueHeader();

		// add the return to overview button as the sole contents of toolbar
		m_WorkBar.removeAll();
		m_WorkBar.add(m_MainButton);
		m_WorkBar.setVisible(true);

		// set screen label
		buildScreenLabel("My Receipts");

		// provide action buttons in the button pane
		m_ButtonPane.removeAll();

		// first, button to delete a receipt
		JButton delButton = new JButton(m_PurgeAction);
		delButton.setToolTipText("Click here to delete the selected receipt");
		delButton.setFont(m_Plugin.M_ButtonFont);
		m_ButtonPane.add(delButton);
		m_ButtonPane.add(Box.createHorizontalStrut(20));	 // spacer

		// second, button to show a receipt's raw XML
		JButton xmlButton = new JButton(m_XMLAction);
		xmlButton.setToolTipText("<html>Click here to display the raw<br/>"
										+ "XML for the selected receipt</html>");
		xmlButton.setFont(m_Plugin.M_ButtonFont);
		m_ButtonPane.add(xmlButton);
		m_ButtonPane.add(Box.createHorizontalStrut(20));	 // spacer

		// third, button to download bulk receipt data
		JButton dnldButton = new JButton(m_InitDownloadAction);
		dnldButton.setToolTipText("Click here to download receipt data");
		dnldButton.setFont(m_Plugin.M_ButtonFont);
		//dnldButton.setIcon(SparkRes.getImageIcon(SparkRes.SAVE_AS_16x16));
		m_ButtonPane.add(dnldButton);
		m_ButtonPane.add(Box.createHorizontalStrut(20));	 // spacer

		// fourth, button to perform mass purge of receipt data
		JButton purgeButton = new JButton(m_InitPurgeAction);
		purgeButton.setToolTipText("Click here to purge receipts");
		purgeButton.setFont(m_Plugin.M_ButtonFont);
		m_ButtonPane.add(purgeButton);

		// place the details about the receipts into a panel
		JPanel rctDets = new JPanel(new GridBagLayout());
		m_GBC.fill = GridBagConstraints.HORIZONTAL;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_GBC.anchor = GridBagConstraints.CENTER;

		// get present receipts list
		Hashtable<XMLReceipt, String> receipts
			= m_ReceiptTableModel.getReceiptList();

		// see if we have no receipts
		m_SelectedReceipt = null;
		if (receipts.isEmpty()) {
			// no receipts - just say so
			JLabel noneLab
				= new JLabel("There are no receipts stored in your wallet.");
			noneLab.setFont(m_ParentHome.M_DataFont);
			noneLab.setForeground(UIManager.getColor("standardGray"));
			noneLab.setHorizontalAlignment(SwingConstants.CENTER);
			rctDets.add(noneLab, m_GBC);

			// install everything into the main WorkPane scroll area
			m_WorkArea.setViewportView(rctDets);
		}
		else {
			// otherwise, just display the table
			m_WorkArea.setViewportView(m_ReceiptTable);
		}
	}

	/**
	 * private utility method to tell the SDS to delete some receipts.
	 * @param delFiles list of receipt filenames to be deleted
	 */
	private void sendDeleteReceipts(ArrayList<String> delFiles) {
		if (delFiles == null || delFiles.isEmpty()) {
			Log.error("Cannot delete no receipts!");
			return;
		}
		VscState state = m_Plugin.getVscState();
		if (!state.isLoggedIn()) {
			Log.error("Cannot delete receipts while not logged in!");
			showError("cannot delete receipts", "not logged in", "log in first");
			return;
		}

		// open a connection to the SDS if we don't have one
		XMPPConnection sdsConn = m_Plugin.getSDSListener().getSDSConnection();
		if (sdsConn == null) {
			sdsConn = m_Plugin.openSDS();
		}
		if (sdsConn == null) {
			// can't do anything
			Log.error("Cannot delete receipts without SDS connection!");
			showError("cannot delete receipts", "SDS unavailable",
						"try again later");
			return;
		}

		// build message to send
		m_ParentHome.setCursor(M_WaitCursor);
		setButtonState(false);
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		SDSMessage delReceipts = new SDSMessage();
		delReceipts.setType(IQ.Type.SET);
		delReceipts.setFrom(sdsConn.getUser());
		delReceipts.setTo(sdsConn.getServiceName());
		delReceipts.setPacketID("rdel_" + state.getNextSDSid());
		delReceipts.setOpcode("REQ_delete_receipts");
		String rIdx = new String(secrets.getReceiptIndex());
		delReceipts.setFolder_index(rIdx);
		String rwCap = new String(secrets.getReceiptRWCap());
		delReceipts.setReadwritecap(rwCap);
		for (String dfile : delFiles) {
			delReceipts.addReceiptToDelList(dfile);
		}

		// register async reply listener
		PacketIDFilter pFilter = new PacketIDFilter(delReceipts.getPacketID());
		m_Plugin.getSDSListener().addFilter(delReceipts.getPacketID(), pFilter);

		// send request to the SDS
		sdsConn.sendPacket(delReceipts);

		// record outgoing packet
		state.setLastActivity(delReceipts);
	}

	/**
	 * process a confirmation list of deleted receipts, by removing from table
	 * @param delFiles a list of the files confirmed purged off the SDS
	 */
	public void confirmReceiptDeletions(ArrayList<String> delFiles) {
		m_ParentHome.setCursor(null);
		setButtonState(true);
		if (delFiles == null || delFiles.isEmpty()) {
			Log.error("No receipts were confirmed deleted");
			showError("error purging receipts", "no receipts confirmed deleted",
						"try clicking on Receipts again to update the list");
			return;
		}

		// iterate through receipts in table, seeing if each was on the list
		Hashtable<XMLReceipt, String> allReceipts
			= m_ReceiptTableModel.getReceiptList();
		// for performance, disable listener during removals
		m_ReceiptTableModel.removeTableModelListener(m_ReceiptTable);
		for (XMLReceipt rect : allReceipts.keySet()) {
			String file = allReceipts.get(rect);
			if (delFiles.contains(file)) {
				m_ReceiptTableModel.removeReceipt(rect);
			}
		}
		m_ReceiptTableModel.addTableModelListener(m_ReceiptTable);

		// NB: SDSListener callback will now tell WTabManager to redisplay us
	}

	/**
	 * obtain the receipts display table
	 * @return the table
	 */
	public JTable getReceiptTable() { return m_ReceiptTable; }

	// END methods
}
