/*
 * last modified---
 * 	08-15-18/kbw/ explicitly cancel PacketCollectors after use
 * 	08-05-14/jet/ Spark plugin version
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide a screen for recovering a voucher safe wallet's login data
 */

package org.silentvault.spark.ui.wallet;

import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.Font;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Color;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;

import org.dom4j.Element;
import org.dom4j.DocumentException;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.util.StringUtils;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;

import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.List;
import java.util.ArrayList;
import java.util.Vector;


/**
 * Extends JPanel to provide a screen for recovering an existing wallet.
 * This operation has these phases:
 * 1) Initiate by requesting a list of supported publishers from the OFS.
 *	2) Select a publisher (this step can be obviated if there is only one VP).
 *	3) Obtain from the OFS the list of 5 recovery questions used by that VP.
 *	4) Prompt the user for their VS# and recovery answers.
 *	5) Ask the OFS to look up the recovery details on the SDS.
 *	6) Display results to user, and transit to login pane if they got it right.
 *		the recovery questions.
 *	Thus only step 4, and perhaps step 2, require user input.
 *
 * @author Kevin Wilkerson
 */
public final class RecoveryPane extends JPanel implements ActionListener {
	// BEGIN data members
	/**
	 * the font to use for prompts (should be same as in LoginPane)
	 */
	private final Font		M_PromptFont;

	/**
	 * the plugin/application to which we belong
	 */
	private WalletClient		m_Plugin;

	/**
	 * the tab manager which creates and manages us
	 */
	private WTabManager		m_TabManager;

	/**
	 * the display list of available voucher publishers
	 */
	private ArrayList<String>	m_PubList;

	/**
	 * the actual list of full ids of possible voucher publishers
	 */
	private List<String>		m_AllPublishers;

	/**
	 * the UI element for publisher selection
	 */
	private JComboBox<String>	m_PubSelect;

	/**
	 * the recovery blob, encrypted and base64-encoded
	 */
	private String				m_RecoveryBlob;

	/**
	 * the OFS's reply to query for recovery questions
	 */
	private OFSMessage		m_RecoveryQues;

	/**
	 * the hash of the VS#+answers
	 */
	private String				m_RecoveryHash;

	/**
	 * the VS# being recovered
	 */
	private String				m_RecoverySafe;

	/**
	 * the current title bar text
	 */
	private String				m_TitlebarText;

	/**
	 * the UI object for entry of the recovered VS number
	 */
	private JTextField		m_VSnumUI;

	/**
	 * the first UI object for the answers to the recovery questions
	 */
	private JTextField		m_Answer1UI;
	/**
	 * the second UI object for the answers to the recovery questions
	 */
	private JTextField		m_Answer2UI;
	/**
	 * the third UI object for the answers to the recovery questions
	 */
	private JTextField		m_Answer3UI;
	/**
	 * the fourth UI object for the answers to the recovery questions
	 */
	private JTextField		m_Answer4UI;
	/**
	 * the fifth UI object for the answers to the recovery questions
	 */
	private JTextField		m_Answer5UI;

	/**
	 * the submit button
	 */
	private JButton			m_SubmitButton;

	/**
	 * the action for handling submission of VP selection
	 */
	private Action				m_Phase1Submit;

	/**
	 * the action for handling submission of QnA data
	 */
	private Action				m_Phase2Submit;

	/**
	 * the scroll area used to ensure that this tab always displays everything
	 * regardless of window size
	 */
	private JScrollPane		m_ScrollPane;

	/**
	 * inner class to provide a DocumentListener for the answer text entry fields
	 */
	final class AnswerDocumentListener implements DocumentListener {
		/**
		 * handle insert notifications
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			Document doc = de.getDocument();
			int curLen = doc.getLength();
			if (curLen > 0) {
				updateSubmitButtonStatus();
			}
		}

		/**
		 * handle removal notifications
		 * @param de the deletion event
		 */
		public void removeUpdate(DocumentEvent de) {
			Document doc = de.getDocument();
			int curLen = doc.getLength();
			if (curLen == 0) {
				updateSubmitButtonStatus();
			}
		}

		/**
		 * handle change notifications (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JTextField");
		}
	}

	/**
	 * the DocumentListener used for all answer entry fields
	 */
	private AnswerDocumentListener	m_AnswerListener;

	/**
	 * whether a VS# has been entered yet
	 */
	private boolean			m_HaveVSnumber;

	/**
	 * inner class to provide a DocumentListener for the m_VSnumUI text field
	 */
	final class SafeDocumentListener implements DocumentListener {
		/**
		 * handle insert notifications
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			Document doc = de.getDocument();
			int curLen = doc.getLength();
			String curTxt = "";
			try {
				curTxt = doc.getText(0, curLen);
			}
			catch (BadLocationException ble) { return; }

			// min length is 3 chars + "-nnn", and we need a '-' in the string
			if (curLen >= 7 && curTxt.indexOf('-') != -1) {
				m_HaveVSnumber = true;
			}
			else {
				m_HaveVSnumber = false;
			}
			updateSubmitButtonStatus();
		}

		/**
		 * handle removal notifications
		 * @param de the deletion event
		 */
		public void removeUpdate(DocumentEvent de) {
			Document doc = de.getDocument();
			int curLen = doc.getLength();
			String curTxt = "";
			try {
				curTxt = doc.getText(0, curLen);
			}
			catch (BadLocationException ble) { return; }

			// min length is 3 chars + "-nnn", and we need a '-' in the string
			if (curLen >= 7 && curTxt.indexOf('-') != -1) {
				m_HaveVSnumber = true;
			}
			else {
				m_HaveVSnumber = false;
			}
			updateSubmitButtonStatus();
		}

		/**
		 * handle change notifications (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JTextField");
		}
	}

	/**
	 * document listener for dynamic updates of m_VSnumUI
	 */
	private SafeDocumentListener	m_VSnumberListener;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param mgr the tab manager which creates us
	 */
	public RecoveryPane(WalletClient plugin, WTabManager mgr) {
		super();
		m_Plugin = plugin;
		m_TabManager = mgr;
		m_TitlebarText = m_RecoveryHash = m_RecoveryBlob = m_RecoverySafe = "";
		M_PromptFont = new Font(Font.SANS_SERIF, Font.BOLD, 13);
		m_AnswerListener = new AnswerDocumentListener();
		m_VSnumberListener = new SafeDocumentListener();
		m_PubList = new ArrayList<String>();

		// use grid layout
		setLayout(new GridLayout(1, 1));
	}

	/**
	 * Initialize for phase 1, publisher selection.  The first time we are
	 * called, we'll do nothing but send REQ_create_safe to the OFS.  On the
	 * REP_publisher_list reply, the callback will call us again.  We'll then
	 * proceed to prompt the user for their publisher selection, unless only
	 * one publisher is listed, in which case we'll skip that.
	 */
	public void initPhase1() {
		// see whether we have a publisher list yet
		if (m_PubList == null || m_PubList.isEmpty()) {
			// we do not, request the list (waits for synchronous reply)
			setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			sendCreateSafe();
			return;
		}

		// check the magnitude of the list
		if (m_PubList.size() == 1) {
			// just assume the only one given is automatically selected
			m_Plugin.getLoginSecrets().setPublisher(m_AllPublishers.get(0));
			m_PubSelect = new JComboBox<String>(new Vector<String>(m_PubList));
			m_PubSelect.setSelectedIndex(0);
			sendGetQuestions();
			return;
		}

		// set up to prompt for publisher selection
		m_TitlebarText = "Wallet Recovery (Step 1)";
		if (m_ScrollPane != null) {
			m_ScrollPane.removeAll();
		}
		removeAll();

		// publisher select label and tool
		JLabel pubPrompt = new JLabel("Select the publisher where your "
												+ "wallet was created",
												SwingConstants.CENTER);
		pubPrompt.setFont(M_PromptFont);
		m_PubSelect = new JComboBox<String>(new Vector<String>(m_PubList));
		m_PubSelect.setToolTipText("Pick desired publisher from list");
		m_PubSelect.addActionListener(this);

		// define button to abort out and return to login screen
		AbstractAction retAction = new AbstractAction("Abort") {
			public void actionPerformed(ActionEvent actionEvent) {
				m_ScrollPane.removeAll();
				removeAll();
				m_TabManager.queueTransition(new Integer(WTabManager.M_LOGIN1));
				m_TabManager.makeTransition();
			}
		};
		JButton retButton = new JButton(retAction);
		retButton.setToolTipText("<html>Click here to abort recovery<br/>and "
										+ "return to login screen</html>");
		retButton.setFont(m_Plugin.M_ButtonFont);

		// define the action and button for handling submission of publisher pick
		m_Phase1Submit = new AbstractAction("Select") {
			public void actionPerformed(ActionEvent actionEvent) {
				// retrieve publisher pick
				String pub = (String) m_PubSelect.getSelectedItem();
				if (pub == null || pub.isEmpty()) {
					showError("missing voucher publisher", "none selected",
								"please pick a voucher publisher");
					return;
				}
				// get the full publisher id which corresponds to the pick
				for (String vPub : m_AllPublishers) {
					if (vPub.endsWith(pub)) {
						m_Plugin.getLoginSecrets().setPublisher(vPub);
						break;
					}
				}

				// request this publisher's recovery questions of the OFS
				sendGetQuestions();
			}
		};
		m_SubmitButton = new JButton(m_Phase1Submit);
		m_SubmitButton.setToolTipText("<html>Select your publisher from<br/>"
												+ "the list and click here</html>");
		m_SubmitButton.setFont(m_Plugin.M_ButtonFont);
		m_SubmitButton.setDefaultCapable(true);

		// put everything in a grid bag with border, inside the scroll area
		setBorder(BorderFactory.createTitledBorder(m_TitlebarText));
		JPanel selectPane = new JPanel();
		GridBagLayout gbl = new GridBagLayout();
		GridBagConstraints gbc = new GridBagConstraints();
		selectPane.setLayout(gbl);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.anchor = GridBagConstraints.NORTH;
		gbc.fill = GridBagConstraints.NONE;
		selectPane.add(Box.createVerticalGlue(), gbc);
		selectPane.add(Box.createVerticalStrut(10), gbc);
		selectPane.add(pubPrompt, gbc);
		selectPane.add(Box.createVerticalStrut(5), gbc);
		selectPane.add(m_PubSelect, gbc);
		selectPane.add(Box.createVerticalStrut(50), gbc);

		// make a borderless panel for the last two buttons, one beside the other
		JPanel buttonPane = new JPanel();
		buttonPane.setLayout(new BoxLayout(buttonPane, BoxLayout.X_AXIS));
		buttonPane.setPreferredSize(new Dimension(450, 30));
		buttonPane.setBorder(null);
		buttonPane.add(Box.createHorizontalGlue());
		buttonPane.add(retButton);
		buttonPane.add(Box.createHorizontalStrut(130));  // spacer
		buttonPane.add(m_SubmitButton);
		buttonPane.add(Box.createHorizontalGlue());
		selectPane.add(buttonPane, gbc);
		selectPane.add(Box.createVerticalStrut(100), gbc);
		selectPane.add(Box.createVerticalGlue(), gbc);
		m_ScrollPane = new JScrollPane(selectPane);
		m_ScrollPane.setBorder(null);
		add(m_ScrollPane);
		JRootPane jrp = getRootPane();
		if (jrp != null) {
			jrp.setDefaultButton(m_SubmitButton);
		}
	}

	/**
	 * initialize for Phase 2, entry of VS#, recovery answers
	 */
	public void initPhase2() {
		setCursor(null);
		removeAll();
		if (m_RecoveryQues == null) {
			Log.error("Null recovery questions reply, cannot init recovery");
			return;
		}

		// make sure we have valid-looking recovery questions
		final String initErr = "cannot recover wallet";
		if (m_RecoveryQues.getRecovery_ques_1().isEmpty()) {
			Log.error("First recovery question is missing!");
			showError(initErr, "first recovery question is missing");
			return;
		}
		if (m_RecoveryQues.getRecovery_ques_2().isEmpty()) {
			Log.error("Second recovery question is missing!");
			showError(initErr, "second recovery question is missing");
			return;
		}
		if (m_RecoveryQues.getRecovery_ques_3().isEmpty()) {
			Log.error("Third recovery question is missing!");
			showError(initErr, "third recovery question is missing");
			return;
		}
		if (m_RecoveryQues.getRecovery_ques_4().isEmpty()) {
			Log.error("Fourth recovery question is missing!");
			showError(initErr, "fourth recovery question is missing");
			return;
		}
		if (m_RecoveryQues.getRecovery_ques_5().isEmpty()) {
			Log.error("Fifth recovery question is missing!");
			showError(initErr, "fifth recovery question is missing");
			return;
		}

		// Ask the user to enter a VS#, and answers to the recovery questions.
		m_TitlebarText = "Wallet Recovery (Step 2)";
		if (m_ScrollPane != null) {
			m_ScrollPane.removeAll();
		}

		// layout setup
		GridBagLayout gbl = new GridBagLayout();
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.anchor = GridBagConstraints.NORTH;
		gbc.fill = GridBagConstraints.NONE;
		gbc.gridwidth = GridBagConstraints.REMAINDER;

		// obtain the VS# first
		JLabel vsPrompt = new JLabel("Enter the wallet ID to be recovered",
												SwingConstants.CENTER);
		vsPrompt.setFont(M_PromptFont);
		JPanel vsGrp = new JPanel();
		vsGrp.setLayout(new BoxLayout(vsGrp, BoxLayout.X_AXIS));
		vsGrp.setBorder(null);
		m_VSnumUI = new JTextField(20);
		m_VSnumUI.setDropTarget(null);
		m_VSnumUI.setToolTipText("<html>A VS number resembles an email<br/>"
										+ "address, and is case-insensitive</html>");
		m_VSnumUI.getDocument().addDocumentListener(m_VSnumberListener);
		m_VSnumUI.addActionListener(this);
		vsGrp.add(m_VSnumUI);
		String pub = m_Plugin.getLoginSecrets().getPublisher();
		int atIdx = pub.indexOf('@');
		JLabel suffixLab = new JLabel(pub.substring(atIdx));
		suffixLab.setLabelFor(m_VSnumUI);
		suffixLab.setForeground(Color.LIGHT_GRAY.darker());
		vsGrp.add(suffixLab);

		// panel to get answers to recovery questions
		JPanel recovGrp = new JPanel();
		recovGrp.setLayout(gbl);
		String recovDesc = "Please enter your answers to the following five "
								+ "standard questions.\n\nYour answers must match "
								+ "the answers you gave when you created your "
								+ "wallet, including punctuation and spaces.\n\n"
								+ "(However, case is not significant.)\n";
		JTextArea recovExpl = new JTextArea(recovDesc, 5, 40);
		recovExpl.setEditable(false);
		recovExpl.setLineWrap(true);
		recovExpl.setWrapStyleWord(true);
		recovExpl.setBackground(m_Plugin.m_LabelBackColor);
		recovExpl.setBorder(null);
		recovExpl.setFocusable(false);
		lastItemConstraints(gbc);
		recovGrp.add(recovExpl, gbc);
		recovGrp.add(Box.createVerticalStrut(20), gbc);

		JLabel q1Lab = new JLabel(m_RecoveryQues.getRecovery_ques_1() + ": ",
											SwingConstants.RIGHT);
		q1Lab.setFont(M_PromptFont);
		m_Answer1UI = new JTextField(15);
		m_Answer1UI.setDropTarget(null);
		final String qTTtxt = "Enter your answer to this question";
		m_Answer1UI.setToolTipText(qTTtxt);
		m_Answer1UI.getDocument().addDocumentListener(m_AnswerListener);
		m_Answer1UI.addActionListener(this);
		q1Lab.setLabelFor(m_Answer1UI);
		firstItemConstraints(gbc);
		recovGrp.add(q1Lab, gbc);
		lastItemConstraints(gbc);
		recovGrp.add(m_Answer1UI, gbc);

		JLabel q2Lab = new JLabel(m_RecoveryQues.getRecovery_ques_2() + ": ",
											SwingConstants.RIGHT);
		q2Lab.setFont(M_PromptFont);
		m_Answer2UI = new JTextField(15);
		m_Answer2UI.setDropTarget(null);
		m_Answer2UI.setToolTipText(qTTtxt);
		m_Answer2UI.getDocument().addDocumentListener(m_AnswerListener);
		m_Answer2UI.addActionListener(this);
		q2Lab.setLabelFor(m_Answer2UI);
		firstItemConstraints(gbc);
		recovGrp.add(q2Lab, gbc);
		lastItemConstraints(gbc);
		recovGrp.add(m_Answer2UI, gbc);

		JLabel q3Lab = new JLabel(m_RecoveryQues.getRecovery_ques_3() + ": ",
											SwingConstants.RIGHT);
		q3Lab.setFont(M_PromptFont);
		m_Answer3UI = new JTextField(15);
		m_Answer3UI.setDropTarget(null);
		m_Answer3UI.setToolTipText(qTTtxt);
		m_Answer3UI.getDocument().addDocumentListener(m_AnswerListener);
		m_Answer3UI.addActionListener(this);
		q3Lab.setLabelFor(m_Answer3UI);
		firstItemConstraints(gbc);
		recovGrp.add(q3Lab, gbc);
		lastItemConstraints(gbc);
		recovGrp.add(m_Answer3UI, gbc);

		JLabel q4Lab = new JLabel(m_RecoveryQues.getRecovery_ques_4() + ": ",
											SwingConstants.RIGHT);
		q4Lab.setFont(M_PromptFont);
		m_Answer4UI = new JTextField(15);
		m_Answer4UI.setDropTarget(null);
		m_Answer4UI.setToolTipText(qTTtxt);
		m_Answer4UI.getDocument().addDocumentListener(m_AnswerListener);
		m_Answer4UI.addActionListener(this);
		q4Lab.setLabelFor(m_Answer4UI);
		firstItemConstraints(gbc);
		recovGrp.add(q4Lab, gbc);
		lastItemConstraints(gbc);
		recovGrp.add(m_Answer4UI, gbc);

		JLabel q5Lab = new JLabel(m_RecoveryQues.getRecovery_ques_5() + ": ",
											SwingConstants.RIGHT);
		q5Lab.setFont(M_PromptFont);
		m_Answer5UI = new JTextField(15);
		m_Answer5UI.setDropTarget(null);
		m_Answer5UI.setToolTipText(qTTtxt);
		m_Answer5UI.getDocument().addDocumentListener(m_AnswerListener);
		m_Answer5UI.addActionListener(this);
		q5Lab.setLabelFor(m_Answer5UI);
		firstItemConstraints(gbc);
		recovGrp.add(q5Lab, gbc);
		lastItemConstraints(gbc);
		recovGrp.add(m_Answer5UI, gbc);

		// define submit button and its action
		setPhase2Action();
		m_SubmitButton = new JButton(m_Phase2Submit);
		m_SubmitButton.setFont(m_Plugin.M_ButtonFont);
		m_SubmitButton.setDefaultCapable(true);
		m_HaveVSnumber = false;
		updateSubmitButtonStatus();

		// define button to abort out and return to login screen
		AbstractAction retAction = new AbstractAction("Abort") {
			public void actionPerformed(ActionEvent actionEvent) {
				m_ScrollPane.removeAll();
				removeAll();
				m_TabManager.queueTransition(new Integer(WTabManager.M_LOGIN1));
				m_TabManager.makeTransition();
			}
		};
		JButton retButton = new JButton(retAction);
		retButton.setToolTipText("<html>Click here to abort recovery<br/>"
										+ "and return to login screen</html>");
		retButton.setFont(m_Plugin.M_ButtonFont);
		retButton.setFocusable(false);

		// make a borderless panel for the last two buttons, one beside the other
		JPanel buttonPane = new JPanel();
		buttonPane.setLayout(new BoxLayout(buttonPane, BoxLayout.X_AXIS));
		buttonPane.setPreferredSize(new Dimension(450, 30));
		buttonPane.setBorder(null);
		buttonPane.add(Box.createHorizontalGlue());
		buttonPane.add(retButton);
		buttonPane.add(Box.createHorizontalStrut(180));  // spacer
		buttonPane.add(m_SubmitButton);
		buttonPane.add(Box.createHorizontalGlue());

		// put everything in a grid bag with a border, inside scroll area
		setBorder(BorderFactory.createTitledBorder(m_TitlebarText));
		JPanel recovPane = new JPanel();
		recovPane.setLayout(gbl);
		recovPane.setBorder(null);
		recovPane.add(Box.createVerticalGlue(), gbc);
		recovPane.add(vsPrompt, gbc);
		recovPane.add(Box.createVerticalStrut(5), gbc);
		recovPane.add(vsGrp, gbc);
		recovPane.add(Box.createVerticalStrut(20), gbc);
		recovPane.add(recovGrp, gbc);
		recovPane.add(Box.createVerticalStrut(10), gbc);
		recovPane.add(buttonPane, gbc);
		recovPane.add(Box.createVerticalGlue(), gbc);
		m_ScrollPane = new JScrollPane(recovPane);
		m_ScrollPane.setBorder(null);
		add(m_ScrollPane);
		JRootPane jrp = getRootPane();
		if (jrp != null) {
			jrp.setDefaultButton(m_SubmitButton);
		}
		m_VSnumUI.requestFocusInWindow();
	}

	/**
	 * initialize for Phase 3, confirmation of recovered safe information
	 */
	public void initPhase3() {
		if (m_RecoveryBlob == null || m_RecoveryBlob.isEmpty()) {
			Log.error("No recovery data, cannot provide user confirmation");
			return;
		}

		// decrypt the returned recovery blob with the recovery hash
		final String err = "could not process recovery details";
		String recData
			= EncodingUtils.getStrFromBase64DESEnc(m_RecoveryBlob, m_RecoveryHash);
		if (recData == null || recData.isEmpty()) {
			Log.error("Recovery blob decrypt failed");
			showError(err, "decrypt failed");
			return;
		}

		/* now parse it -- should look like:
		 * <VS_number/>
		 * <voucher_publisher/>
		 * <pin/>
		 * <long-phrase/>
		 * <passphrase/>
		 */
		Element recElt = EncodingUtils.buildElement("<recovery_details>"
																	+ recData
																	+ "</recovery_details>");
		if (recElt == null) {
			Log.error("Recovery data failed to parse");
			showError(err, "parse failed");
			return;
		}
		String vsNum = null;
		String pub = null;
		String pin = null;
		String lphrase = null;
		String pphrase = null;
		try {
			Element vsElt = recElt.element("VS_number");
			if (vsElt == null) {
				throw new DocumentException("missing wallet ID");
			}
			vsNum = vsElt.getTextTrim();
			// make sure it matches the one we're recovering
			if (!vsNum.equals(m_RecoverySafe)) {
				throw new DocumentException("VS# " + vsNum + " does not match!");
			}

			Element pubElt = recElt.element("voucher_publisher");
			if (pubElt == null) {
				throw new DocumentException("missing voucher publisher");
			}
			pub = pubElt.getTextTrim();
			// make sure it matches the one we're recovering
			String vp = (String) m_PubSelect.getSelectedItem();
			// get the full publisher id which corresponds to the pick
			for (String vPub : m_AllPublishers) {
				if (vPub.endsWith(vp)) {
					vp = vPub;
					break;
				}
			}
			if (!pub.equals(vp)) {
				throw new DocumentException("VP " + pub + " does not match!");
			}

			Element pinElt = recElt.element("pin");
			if (pinElt == null) {
				throw new DocumentException("missing PIN");
			}
			pin = pinElt.getText();

			Element lpElt = recElt.element("long-phrase");
			if (lpElt == null) {
				throw new DocumentException("missing long phrase");
			}
			lphrase = lpElt.getTextTrim();

			Element ppElt = recElt.element("passphrase");
			if (ppElt == null) {
				throw new DocumentException("missing passphrase");
			}
			pphrase = ppElt.getTextTrim();
		}
		catch (DocumentException de) {
			Log.error("recovery data parse error", de);
			showError(err, de.getMessage());
			return;
		}

		// display confirmation, vector to login pane on user accept
		m_TitlebarText = "Wallet Recovery (Success)";
		if (m_ScrollPane != null) {
			m_ScrollPane.removeAll();
		}
		removeAll();
		setBorder(BorderFactory.createTitledBorder(m_TitlebarText));
		setCursor(null);

		// panel and layout setup
		GridBagLayout gbl = new GridBagLayout();
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.anchor = GridBagConstraints.NORTH;
		JPanel confPane = new JPanel(gbl);
		confPane.setBorder(null);
		lastItemConstraints(gbc);
		confPane.add(Box.createVerticalGlue(), gbc);
		confPane.add(Box.createVerticalStrut(25), gbc);

		// show header
		JLabel header
			= new JLabel("Your wallet's login data", SwingConstants.CENTER);
		header.setFont(M_PromptFont);
		confPane.add(header, gbc);
		confPane.add(Box.createVerticalStrut(20), gbc);

		// recapitulate everything (all login data) on a single text area, so
		// that it can be grabbed using copy and paste

		StringBuilder loginTxt = new StringBuilder(1024);
		JTextArea loginData = new JTextArea(7, 40);
		loginData.setEditable(false);
		loginData.setLineWrap(true);
		loginData.setWrapStyleWord(true);
		loginData.setBackground(m_Plugin.m_LabelBackColor);

		// build data
		loginTxt.append("Your wallet ID is:  " + vsNum + "\n\n");
		loginTxt.append("Your PIN is:\t\t   " + pin + "\n\n");

		loginTxt.append("Your long-phrase is:\t   " + lphrase + "\n\n");
		loginTxt.append("Your passphrase is:\t   " + pphrase + "\n");
		loginData.setText(loginTxt.toString());
		confPane.add(loginData, gbc);
		confPane.add(Box.createVerticalStrut(50), gbc);

		JLabel instructs = new JLabel("Please record this data securely, "
												+ "and keep it strictly private.",
												SwingConstants.CENTER);
		instructs.setFont(M_PromptFont);

		// define submit button and its action
		AbstractAction confirmed = new AbstractAction("Acknowledged") {
			public void actionPerformed(ActionEvent actionEvent) {
				// go to login screen
				m_ScrollPane.removeAll();
				removeAll();
				m_TabManager.queueTransition(new Integer(WTabManager.M_LOGIN1));
				m_TabManager.makeTransition();
			}
		};
		m_SubmitButton = new JButton(confirmed);
		m_SubmitButton.setToolTipText("Click here to proceed to login screen");
		m_SubmitButton.setFont(m_Plugin.M_ButtonFont);
		m_SubmitButton.setDefaultCapable(true);

		// put everything in inside scroll area
		gbc.fill = GridBagConstraints.NONE;
		confPane.add(instructs, gbc);
		confPane.add(Box.createVerticalStrut(40), gbc);
		confPane.add(m_SubmitButton, gbc);
		confPane.add(Box.createVerticalStrut(10), gbc);
		confPane.add(Box.createVerticalGlue(), gbc);
		m_ScrollPane = new JScrollPane(confPane);
		m_ScrollPane.setBorder(null);
		add(m_ScrollPane);
		JRootPane jrp = getRootPane();
		if (jrp != null) {
			jrp.setDefaultButton(m_SubmitButton);
		}
	}

	/**
	 * private utility method to prepare GridBagConstraints for use with the
	 * first item on a line (such as a prompt/label)
	 */
	private void firstItemConstraints(GridBagConstraints gbc) {
		if (gbc == null) {
			return;
		}
		gbc.gridwidth = GridBagConstraints.RELATIVE;
		gbc.fill = GridBagConstraints.NONE;
	}

	/**
	 * private utility method to prepare GridBagConstraints for use with the
	 * last item on a line (such as an entry field)
	 */
	private void lastItemConstraints(GridBagConstraints gbc) {
		if (gbc == null) {
			return;
		}
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.fill = GridBagConstraints.HORIZONTAL;
	}

	/**
	 * method to define the action for submit button in Phase2 context
	 */
	private void setPhase2Action() {
		m_Phase2Submit = new AbstractAction("Submit") {
			public void actionPerformed(ActionEvent actionEvent) {
				// get and validate the VS# entered, and compare to confirmation
				String vsStr = m_VSnumUI.getText();
				if (vsStr.isEmpty()) {
					showError("missing wallet ID",
								"none entered", "please enter your wallet ID");
					return;
				}

				// strip @domain from entry if they entered it manually themselves
				int atIdx = vsStr.indexOf('@');
				if (atIdx != -1) {
					vsStr = vsStr.substring(0, atIdx);
				}

				// now add @domain
				String pub = m_Plugin.getLoginSecrets().getPublisher();
				atIdx = pub.indexOf('@');
				vsStr = vsStr + pub.substring(atIdx);
				if (vsStr.indexOf('@') == -1 || vsStr.indexOf('.') == -1) {
					showError("missing wallet ID",
								"wallet ID entered was improper",
								"please try again");
					return;
				}
				m_RecoverySafe = vsStr;

				// get the answer to the first recovery question
				final String notEmp = "answer cannot be empty";
				String ans1 = m_Answer1UI.getText();
				if (ans1 == null || ans1.isEmpty()) {
					showError("unsuitable answer to first recovery question",
								notEmp, "please re-enter your answer");
					return;
				}

				// get the answer to the second recovery question
				String ans2 = m_Answer2UI.getText();
				if (ans2 == null || ans2.isEmpty()) {
					showError("unsuitable answer to second recovery question",
								notEmp, "please re-enter your answer");
					return;
				}

				// get the answer to the third recovery question
				String ans3 = m_Answer3UI.getText();
				if (ans3 == null || ans3.isEmpty()) {
					showError("unsuitable answer to third recovery question",
								notEmp, "please re-enter your answer");
					return;
				}

				// get the answer to the fourth recovery question
				String ans4 = m_Answer4UI.getText();
				if (ans4 == null || ans4.isEmpty()) {
					showError("unsuitable answer to fourth recovery question",
								notEmp, "please re-enter your answer");
					return;
				}

				// get the answer to the fifth recovery question
				String ans5 = m_Answer5UI.getText();
				if (ans5 == null || ans5.isEmpty()) {
					showError("unsuitable answer to fifth recovery question",
								notEmp, "please re-enter your answer");
					return;
				}

				// compute the recovery hash
				computeHash();

				// set wait cursor
				setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

				// everything looks okay; send recovery lookup to the OFS
				sendLookupRecovery();
			}
		};
	}


	// set methods
	/**
	 * set the publisher list
	 * @param publishers the list of available voucher publishers
	 */
	public boolean setPubList(List<String> publishers) {
		if (publishers == null || publishers.isEmpty()) {
			return false;
		}

		m_AllPublishers = publishers;
		m_PubList.clear();
		for (String vPub : publishers) {
			// for display, strip off userid, i.e. everything before the '@'
			int atIdx = vPub.indexOf('@');
			if (atIdx != -1) {
				vPub = vPub.substring(atIdx);
			}
			m_PubList.add(vPub);
		}
		return true;
	}

	/**
	 * set the recovery data reply message (invoked by OFS callback)
	 * @param safeData the message giving the encoded safe recovery data
	 */
	public void setRecoveryBlob(OFSMessage safeData) {
		if (safeData != null && safeData.getOpcode().equals("REP_recovery_blob"))
		{
			m_RecoveryBlob = safeData.getRecovery_blob();
		}
	}

	/**
	 * set the recovery questions reply message (invoked by OFS callback)
	 * @param quesRep the reply message giving the 5 recovery questions
	 */
	public void setRecoveryQues(OFSMessage quesRep) {
		if (quesRep != null && quesRep.getOpcode().equals("REP_questions")) {
			m_RecoveryQues = quesRep;
		}
	}


	/**
	 * send a REQ_create_safe to the OFS, and block on result
	 */
	private void sendCreateSafe() {
		final XMPPConnection conn = m_Plugin.getOFSConnection();
		VscState stData = m_Plugin.getVscState();

		// build message to send
		OFSMessage creS = new OFSMessage();
		creS.setType(IQ.Type.GET);
		creS.setFrom(conn.getUser());
		creS.setTo(conn.getServiceName());
		creS.setPacketID("create_" + stData.getNextOFSid());
		creS.setOpcode("REQ_create_safe");

		// register a synchronous collector for the reply
		PacketIDFilter pFilter = new PacketIDFilter(creS.getPacketID());
		PacketCollector repColl = conn.createPacketCollector(pFilter);

		// send it to the server
		conn.sendPacket(creS);

		// record outgoing packet
		stData.setLastActivity(creS);

		// wait for reply or timeout
		OFSMessage rep = (OFSMessage) repColl.nextResult(m_Plugin.getTimeout());
		setCursor(null);
		repColl.cancel();
		if (rep == null) {
			// got timeout
			showError("cannot recover wallet data",
						"timeout asking for publisher list",
						"please try again later");
			return;
		}
		stData.setLastActivity(rep);
		if (setPubList(rep.getVoucherPublishers())) {
			initPhase1();
		}
		else {
			showError("cannot recover wallet data", "null publisher list");
			Log.error("null publisher list");
		}
	}

	/**
	 * send a REQ_recovery_ques to the OFS
	 */
	private void sendGetQuestions() {
		final XMPPConnection conn = m_Plugin.getOFSConnection();
		VscState stData = m_Plugin.getVscState();

		// build message to send
		OFSMessage getQues = new OFSMessage();
		getQues.setType(IQ.Type.GET);
		getQues.setFrom(conn.getUser());
		getQues.setTo(conn.getServiceName());
		getQues.setPacketID("ques_" + stData.getNextOFSid());
		getQues.setOpcode("REQ_recovery_ques");
		getQues.setVoucher_publisher(m_Plugin.getLoginSecrets().getPublisher());

		// register a listener for the reply
		PacketIDFilter pFilter = new PacketIDFilter(getQues.getPacketID());
		m_Plugin.getOFSListener().addFilter(getQues.getPacketID(), pFilter);

		// send it to the server
		conn.sendPacket(getQues);

		// record outgoing packet
		stData.setLastActivity(getQues);
	}
	
	/**
	 * send a REQ_lookup_recovery to the OFS
	 */
	private void sendLookupRecovery() {
		if (m_RecoveryHash.isEmpty()) {
			Log.error("Cannot lookup recovery blob without hash index!");
			return;
		}
		final XMPPConnection conn = m_Plugin.getOFSConnection();
		VscState stData = m_Plugin.getVscState();

		// build message to send
		OFSMessage lookupRec = new OFSMessage();
		lookupRec.setType(IQ.Type.GET);
		lookupRec.setFrom(conn.getUser());
		lookupRec.setTo(conn.getServiceName());
		lookupRec.setPacketID("recov_" + stData.getNextOFSid());
		lookupRec.setOpcode("REQ_lookup_recovery");

		// compute the hash of the recovery encryption hash
		MessageDigest md = null;
		try {
			md = MessageDigest.getInstance("SHA-512");
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("Cannot find SHA-512 hash algorithm", nsae);
			showError("can't construct SHA512 hash", nsae.getMessage());
			return;
		}
		byte[] recovIdx = md.digest(StringUtils.decodeBase64(m_RecoveryHash));
		lookupRec.setRecovery_index(StringUtils.encodeBase64(recovIdx, false));
		lookupRec.setVoucher_publisher(m_Plugin.getLoginSecrets().getPublisher());

		// register a listener for the reply
		PacketIDFilter pFilter = new PacketIDFilter(lookupRec.getPacketID());
		m_Plugin.getOFSListener().addFilter(lookupRec.getPacketID(), pFilter);

		// send it to the server
		conn.sendPacket(lookupRec);

		// record outgoing packet
		stData.setLastActivity(lookupRec);
	}

	/**
	 * utility method to compute the SHA512 hash of the VS# + five recovery
	 * question answers
	 */
	private void computeHash() {
		StringBuilder rh = new StringBuilder(1024);
		rh.append(m_RecoverySafe + ":");
		rh.append(m_Answer1UI.getText().toLowerCase() + ":");
		rh.append(m_Answer2UI.getText().toLowerCase() + ":");
		rh.append(m_Answer3UI.getText().toLowerCase() + ":");
		rh.append(m_Answer4UI.getText().toLowerCase() + ":");
		rh.append(m_Answer5UI.getText().toLowerCase());
		MessageDigest md = null;
		try {
			md = MessageDigest.getInstance("SHA-512");
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("Cannot find SHA-512 hash algorithm", nsae);
			showError("can't construct SHA512 hash", nsae.getMessage());
			return;
		}
		byte[] recovHash = md.digest(rh.toString().getBytes());
		m_RecoveryHash = StringUtils.encodeBase64(recovHash, false);
		return;
	}

	/**
	 * utility method to update the status of the phase 2 Submit button
	 */
	private void updateSubmitButtonStatus() {
		String a1 = m_Answer1UI.getText();
		String a2 = m_Answer2UI.getText();
		String a3 = m_Answer3UI.getText();
		String a4 = m_Answer4UI.getText();
		String a5 = m_Answer5UI.getText();

		boolean questionsAnswered = !a1.isEmpty() && !a2.isEmpty()
												&& !a3.isEmpty() && !a4.isEmpty()
												&& !a5.isEmpty();
		if (m_HaveVSnumber) {
			if (questionsAnswered) {
				m_SubmitButton.setEnabled(true);
				m_SubmitButton.setToolTipText("<html>Click here to submit your VS#"
														+ "<br/>and answers for recovery"
														+ "</html>");
			}
			else {
				m_SubmitButton.setEnabled(false);
				m_SubmitButton.setToolTipText("Enter all of your five answers");
			}
		}
		else {
			if (questionsAnswered) {
				m_SubmitButton.setEnabled(false);
				m_SubmitButton.setToolTipText("Enter your wallet ID");
			}
			else {
				m_SubmitButton.setEnabled(false);
				m_SubmitButton.setToolTipText("<html>Enter your wallet ID"
														+ "<br/>and answers to all questions"
														+ "</html>");
			}
		}
	}

	// implement ActionListener
	/**
	 * utility method to implement ActionListener for text field components
	 * @param ae the event from one of the JTextField objects
	 */
	public void actionPerformed(ActionEvent ae) {
		Object src = ae.getSource();
		if (src instanceof JTextField) {
			JTextField input = (JTextField) src;
			// move to the next focus item
			input.transferFocus();
		}
		else if (src instanceof JComboBox) {
			JComboBox box = (JComboBox) src;
			// move to the next focus item
			box.transferFocus();
		}
	}


	// error display methods
	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 */
	private void showError(String errMsg) {
		m_TabManager.showError(errMsg);
	}

	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 * @param cause the cause of the error
	 */
	private void showError(String errMsg, String cause) {
		m_TabManager.showError(errMsg, cause);
	}

	/**
	 * display an error message to user in a popup window
	 * @param errMsg the message to display
	 * @param cause the cause of the error
	 * @param action suggested remedial action
	 */
	private void showError(String errMsg, String cause, String action) {
		m_TabManager.showError(errMsg, cause, action);
	}

	// END methods
}
