/*
 * last modified---
 * 	08-15-18/kbw/ explicitly cancel PacketCollectors after use
 * 	02-09-18/kbw/ buildSplitEntry() only after m_VoucherAssetType is set
 * 	01-03-18/kbw/ ensure all amounts are listed in base units
 * 	08-05-14/jet/ Spark plugin version
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide a UI widget to handle performing voucher splits
 */

package org.silentvault.spark.ui.wallet;

import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.Font;
import java.awt.Color;
import java.awt.Dimension;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.HashMap;
import java.net.HttpURLConnection;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.ParseException;


/**
 * Extends WorkPane to provide a screen for doing splits on vouchers.
 *
 * @author Kevin Wilkerson
 */
public final class SplitPane extends WorkPane {
	// BEGIN data members
	/**
	 * the maximum number of split outputs which may be done in one operation
	 */
	private static final int	M_MaxSplitOutputs = 2;

	/**
	 * action to split a voucher
	 */
	private AbstractAction		m_SplitAction;

	/**
	 * UI elements for user input of split values
	 */
	private JTextField[]			m_SplitInputs;

	/**
	 * UI element to display remainder
	 */
	private JTextField			m_Remainder;

	/**
	 * user-entered values for split; the last elt is the remainder
	 */
	private double[]				m_SplitValues;

	/**
	 * button to process the split
	 */
	private JButton				m_SplitButton;

	/**
	 * the panel containing the voucher split value entry form
	 */
	private JPanel					m_SplitEntry;

	/**
	 * inner class to provide a DocumentListener for the JTextFieldS
	 */
	final class SplitDocumentListener implements DocumentListener {
		JTextField			m_TextComp;

		/**
		 * constructor
		 * @param txtComp the text field for which we're the doc listener
		 */
		public SplitDocumentListener(JTextField txtComp) {
			m_TextComp = txtComp;
		}

		/**
		 * handle insert notifications
		 * @param de the insertion event
		 */
		public void insertUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			Document doc = de.getDocument();
			String curTxt = "";

			// interpret a '.' as '0.' if it's the first character added
			int curLen = doc.getLength();
			if (curLen == 1) {
				try {
					curTxt = doc.getText(0, 1);
				}
				catch (BadLocationException ble) {
					Log.error("Bad insert doc location", ble);
				}
				if (curTxt.equals(".")) {
					// force this to "0."
					queueTextMod(m_TextComp, "0.");
					return;
				}
			}

			// disallow letters etc. altogether
			int changeAt = de.getOffset();
			try {
				curTxt = doc.getText(0, doc.getLength());
			}
			catch (BadLocationException ble) {
				Log.error("Cannot undo bad value edit", ble);
				return;
			}
			char chg = curTxt.charAt(changeAt);
			if (!Character.isDigit(chg) && chg != '.' && chg != ',') {
				// undo this edit
				String modTxt = curTxt.substring(0, changeAt)
									+ curTxt.substring(changeAt+1);
				queueTextMod(m_TextComp, modTxt);
				return;
			}

			// otherwise just parse
			if (parseValue()) {
				recomputeRemainder(m_TextComp);
			}
			else {
				// undo this edit
				String modTxt = curTxt.substring(0, changeAt)
									+ curTxt.substring(changeAt+1);
				queueTextMod(m_TextComp, modTxt);
			}
		}

		/**
		 * handle removal notifications
		 * @param de the deletion event
		 */
		public void removeUpdate(DocumentEvent de) {
			// don't do any processing while we're undoing an edit
			if (m_DoingUndo) {
				return;
			}

			// if the only thing left is a '.', force it to a '0.'
			Document doc = de.getDocument();
			int curLen = doc.getLength();
			if (curLen == 1) {
				String curTxt = "";
				try {
					curTxt = doc.getText(0, 1);
				}
				catch (BadLocationException ble) {
					Log.error("Bad remove doc location", ble);
				}
				if (curTxt.equals(".")) {
					// force this to "0."
					queueTextMod(m_TextComp, "0.");
					return;
				}
			}

			// otherwise just parse
			if (parseValue()) {
				recomputeRemainder(m_TextComp);
			}
		}

		/**
		 * handle change notifications (unused)
		 * @param de the change event
		 */
		public void changedUpdate(DocumentEvent de) {
			// plain text components don't fire these events
			Log.debug("Got unexpected changedUpdate event from JTextField");
		}

		/**
		 * helper method to parse the current value of m_TextComp
		 * @return true on success, else false
		 */
		private boolean parseValue() {
			// examine what's in the value entry field
			String enteredVal = m_TextComp.getText();
			if (enteredVal.isEmpty() || enteredVal.equals("0.00")) {
				return true;
			}

			// attempt to parse, localized
			double vvv = 0.0;
			DecimalFormat df = new DecimalFormat("0.0");
			try {
				Number val = df.parse(enteredVal);
				vvv = val.doubleValue();
			}
			catch (ParseException pe) {
				return false;
			}
			if (vvv < 0.0) {
				return false;
			}
			return true;
		}
	}

	/**
	 * inner class to modify a text field from within its document listener,
	 * without triggering a recursive modification exception
	 */
	class TextFieldMod implements Runnable {
		/**
		 * the text component being modified
		 */
		JTextField			m_ModComponent;

		/**
		 * the new contents of the component
		 */
		String				m_NewContents;

		/**
		 * constructor
		 * @param txtField the text component to update
		 * @param newText the replacement text
		 */
		public TextFieldMod(JTextField txtField, String newText) {
			m_ModComponent = txtField;
			m_NewContents = new String(newText);
		}

		/**
		 * execute main task in background
		 */
		public void run() {
			try { Thread.sleep(25); }
			catch (InterruptedException ie) { /* ignore */ }
			m_ModComponent.setText(m_NewContents);
			m_DoingUndo = false;
			m_ModComponent.setCaretPosition(m_NewContents.length());
			recomputeRemainder(m_ModComponent);
		}
	}

	/**
	 * track whether an undo mod is in progress
	 */
	private boolean				m_DoingUndo;

	/**
	 * the asset type of vouchers being worked with for the selected Issuer
	 */
	private String					m_VoucherAssetType;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the HomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public SplitPane(WalletClient plugin, HomePane home, WTabManager mgr) {
		super(plugin, home, mgr);
		m_HeadingPane.setLayout(new GridLayout(1, 3));

		m_SplitInputs = new JTextField[M_MaxSplitOutputs];
		for (int iii = 0; iii < m_SplitInputs.length; iii++) {
			m_SplitInputs[iii] = new JTextField(8);
			m_SplitInputs[iii].setDropTarget(null);
			m_SplitInputs[iii].setHorizontalAlignment(JTextField.TRAILING);
			m_SplitInputs[iii].setFont(home.M_DataFont);
			m_SplitInputs[iii].addActionListener(this);
		}
		m_SplitValues = new double[M_MaxSplitOutputs+1];
		m_SplitEntry = new JPanel();
		m_SplitEntry.setOpaque(true);
		m_SplitEntry.setBorder(null);
		m_SplitEntry.setPreferredSize(new Dimension(420, 90));

		// provide user textual instructions
		String splitTxt = "Select a voucher then enter split amount(s).  The "
								+ "remainder is calculated automatically to make the "
								+ "final voucher in the split.<br/><br/>"
								+ "Note: splitting is not usually necessary.  Vouchers "
								+ "are split automatically during payments where the "								  + "voucher has a greater value than the payment.";
		m_UserInstructs = splitTxt;
		m_VoucherAssetType = "";

		// set token cost
		m_TokenCost.setText("Split uses " + VoucherRequest.M_SplitCost
								+ " tokens");

		// define button actions
		setSplitAction();
		m_SplitButton = new JButton(m_SplitAction);
		m_SplitButton.setToolTipText("<html>Click here to split the selected<br/>"
												+ "voucher as indicated</html>");
		m_SplitButton.setFont(m_Plugin.M_ButtonFont);

		// define remainder display field
		m_Remainder = new JTextField(8);
		m_Remainder.setDropTarget(null);
		m_Remainder.setEditable(false);
		m_Remainder.setOpaque(false);
		m_Remainder.setHorizontalAlignment(JTextField.RIGHT);
		m_Remainder.setFont(home.M_LabelFont);
		m_Remainder.setBorder(m_InsetBorder);
		m_Remainder.setBackground(m_Plugin.m_LabelBackColor);
		m_Remainder.setFocusable(false);

		// we don't want a border around the whole thing
		m_WorkArea.setBorder(null);
	}

	/**
	 * utility method to establish the Action which can process a click on the
	 * Split button
	 */
	private void setSplitAction() {
		m_SplitAction = new AbstractAction("Split now") {
			public void actionPerformed(ActionEvent actionEvent) {
				// we need exactly one row selected in the voucher table
				VoucherTableModel vtm = m_ParentHome.getVoucherModel();
				ArrayList<XMLVoucher> selVouchers = vtm.getSelectedVouchers();
				if (selVouchers.size() < 1) {
					showError("cannot perform split", "no vouchers selected",
								"please select a voucher to split");
					return;
				}
				// this case should never occur, but...
				if (selVouchers.size() > 1) {
					showError("too many vouchers to split", "more than one selected",
								"please select exactly one voucher to split");
					return;
				}
				int tokCost = VoucherRequest.M_SplitCost;

				// ensure we have enough tokens
				TokenTableModel ttm = m_ParentHome.getTokenModel();
				ArrayList<XMLToken> safeTokens = ttm.getAvailableTokens(tokCost,
																		vtm.getSelectedIssuer());
				if (safeTokens == null) {
					// Record ourselves with tab manager, so we return here after
					// the user has purchased tokens.  Then go purchase tokens.

					m_TabManager.setPendingTokens(m_TabManager.getSplitPane());
					m_TabManager.queueTransition(
												new Integer(WTabManager.M_PURCHASE_TOKENS));
					m_TabManager.makeTransition();
					return;
				}

				// make sure that the output values don't exceed the input
				XMLVoucher vouch = selVouchers.get(0);
				double outTotGr = 0.0;
				for (int iii = 0; iii < M_MaxSplitOutputs; iii++) {
					outTotGr += m_SplitValues[iii];
				}
				// add remainder, if it's positive
				double rem = m_SplitValues[M_MaxSplitOutputs];
				if (rem > 0.0) {
					outTotGr += rem;
				}
				if (outTotGr > vouch.getNPV()) {
					showError("could not execute split",
								"output value exceeds input",
								"adjust output values or select a larger voucher");
					return;
				}
				if (outTotGr <= 0.0 || outTotGr == rem) {
					showError("could not execute split",
								"no split values specified",
								"please enter at least one split value");
					return;
				}

				// prepare the message to the OFS
				final XMPPConnection conn = m_Plugin.getOFSConnection();
				VscState stData = m_Plugin.getVscState();
				VsSecrets secData = m_Plugin.getLoginSecrets();
				OFSMessage split = new OFSMessage();
				split.setType(IQ.Type.SET);
				split.setFrom(conn.getUser());
				split.setTo(conn.getServiceName());
				split.setPacketID("valid_" + stData.getNextOFSid());
				split.setOpcode("REQ_revalidate");
				split.setVoucher_publisher(secData.getPublisher());

				// build voucher request and add it to message
				VoucherRequest vReq = new VoucherRequest();
				vReq.setAction("split");
				vReq.setSigningVS(secData.getVSnumber());
				vReq.setFolderHash(new String(secData.getVoucherIndex()));
				vReq.setFolderCap(secData.getVouchRWCap());
				// set output values, leaving remainder implicit
				for (int iii = 0; iii < M_MaxSplitOutputs; iii++) {
					if (m_SplitValues[iii] > 0.0) {
						vReq.addOutputValue(m_SplitValues[iii], vouch.baseUnits());
					}
				}
				split.setVouchReq(vReq);

				// add the selected voucher to the voucherRequest
				Hashtable<XMLVoucher, String> vouchers = vtm.getVoucherList();
				HashMap<String, XMLVoucher> inpVouchers = vReq.getVouchers();
				String sdsFile = vouchers.get(vouch);
				inpVouchers.put(sdsFile, vouch);
				vReq.setInitialized();

				// supply payment tokens
				for (int iii = 0; iii < tokCost; iii++) {
					vReq.addPaymentToken(safeTokens.get(iii));
				}

				// sign the voucher request with our private key
				if (!vReq.signDetails(secData.getPrivKey())) {
					Log.error("Unable to sign split VR with VS privkey");
					showError("could not sign voucherRequest!");
					return;
				}

				// encrypt the details to the VP's key so the OFS can't read them
				if (!vReq.encryptDetails(secData.getVPKey(), true)) {
					Log.error("Unable to encrypt split VR with VP pubkey");
					showError("could not encrypt voucherRequest for VP!");
					return;
				}

				// flag these vouchers and tokens as committed to an operation
				vtm.commitVouchers(selVouchers);
				ttm.commitTokens(safeTokens);

				// register a collector for the reply
				PacketIDFilter pFilter = new PacketIDFilter(split.getPacketID());
				PacketCollector repColl = conn.createPacketCollector(pFilter);

				// send the actual split message
				m_ParentHome.setCursor(M_WaitCursor);
				setButtonState(false);
				conn.sendPacket(split);

				// record outgoing packet
				stData.setLastActivity(split);

				// wait for reply or timeout
				OFSMessage rep = null;
				IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
				m_ParentHome.setCursor(null);
				setButtonState(true);
				if (reply == null) {
					// got timeout; build error
					rep = new OFSMessage();
					rep.setOpcode("REP_validation");
					rep.setFrom(split.getTo());
					rep.setTo(split.getFrom());
					rep.setPacketID(split.getPacketID());
					rep.setType(IQ.Type.ERROR);
					rep.setErrcode(HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
					rep.setErrmsg("timed out");
				}
				else {
					rep = (OFSMessage) reply;
				}
				repColl.cancel();

				// invoke callback with result
				m_Plugin.getOFSListener().processPacket(rep);
			}
		};
	}


	/**
	 * helper method to recompute and show remainder
	 * @param textField the text object which triggered the computation
	 */
	private void recomputeRemainder(JTextField textField) {
		if (textField == null) {
			return;
		}

		// loop through all split outputs to determine which one this is
		int idx = -1;
		for (int iii = 0; iii < M_MaxSplitOutputs; iii++) {
			if (m_SplitInputs[iii] == textField) {
				idx = iii;
				break;
			}
		}
		if (idx < 0) {
			Log.error("Did not find JTextField in m_SplitInputs");
			return;
		}

		// attempt to parse leniently
		double vvv = 0.0;
		String dispUnits = XMLVoucher.getDisplayUnits(m_VoucherAssetType);
		String vStr = textField.getText();
		DecimalFormat df = new DecimalFormat("#,###,##0.0###");
		if (vStr.isEmpty()) {
			m_SplitValues[idx] = 0.0;
		}
		else {
			try {
				Number val = df.parse(vStr);
				vvv = val.doubleValue();

				// check for positive values below 0.0001 GAU
				if (vvv > 0.0 && vvv < 0.0001) {
					showError("illegal split output value",
								"value in field " + (idx+1) + " is too small",
								"please enter a value at least 0.0001 " + dispUnits);
					return;
				}
				m_SplitValues[idx] = vvv;
			}
			catch (ParseException pe) {
				// this should not happen if SplitDocumentListener did its job
				showError("illegal split output value",
							"non-numeric value \"" + vStr + "\" in field " + (idx+1),
							"please enter a valid positive decimal number");
				return;
			}
		}

		// recompute remainder using this value
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		ArrayList<XMLVoucher> selVouchers = vtm.getSelectedVouchers();
		if (selVouchers.isEmpty()) {
			// NB: this should be impossible at this point
			showError("cannot compute split portions", "no voucher selected",
						"please select an input voucher to split");
			return;
		}
		XMLVoucher vouch = selVouchers.get(0);
		double vAmt = vouch.getNPV();
		double rem = vAmt;
		for (int jjj = 0; jjj < M_MaxSplitOutputs; jjj++) {
			rem -= m_SplitValues[jjj];
		}
		if (rem < 0.0 || rem == vAmt) {
			m_Remainder.setForeground(Color.RED);
			m_SplitButton.setEnabled(false);
			m_SplitButton.setToolTipText("Enter split amounts first");
		}
		else {
			m_Remainder.setForeground(Color.BLACK);

			// input is acceptable, so allow the button to work
			m_SplitButton.setEnabled(true);
			m_SplitButton.setToolTipText("<html>Click here to split the selected"
												+ "<br/>voucher as indicated</html>");
		}

		// set new values and rebuild display
		m_SplitValues[M_MaxSplitOutputs] = rem;
		buildSplitEntry();
		textField.requestFocusInWindow();
	}


	// method required to extend WorkPane
	/**
	 * method to prepare display for this sub-class
	 */
	public void prepDisplay() {
		// set heading pane
		super.buildValueHeader();

		// add the return to overview button as the sole contents of toolbar
		m_WorkBar.removeAll();
		m_WorkBar.add(m_MainButton);
		m_WorkBar.setVisible(true);

		// set screen label
		buildScreenLabel("Split a Voucher");

		// provide action buttons in the button pane
		m_ButtonPane.removeAll();

		// first, the button to perform the voucher split
		m_SplitButton.setEnabled(false);
		m_SplitButton.setToolTipText("Please select a voucher to split");
		m_ButtonPane.add(Box.createHorizontalStrut(350));	// spacer
		m_ButtonPane.add(m_SplitButton);

		// build the initial split value entry form (m_SplitEntry)
		m_DoingUndo = false;
		for (int iii = 0; iii < m_SplitValues.length; iii++) {
			m_SplitValues[iii] = 0.0;
		}
		for (int jjj = 0; jjj < m_SplitInputs.length; jjj++) {
			JTextField txtField = new JTextField(8);
			txtField.setDropTarget(null);
			txtField.setText("");
			txtField.setHorizontalAlignment(JTextField.TRAILING);
			txtField.setFont(m_ParentHome.M_DataFont);
			txtField.addActionListener(this);

			// supply a document listener
			SplitDocumentListener docLis = new SplitDocumentListener(txtField);
			txtField.getDocument().addDocumentListener(docLis);
			txtField.setEnabled(false);
			txtField.setToolTipText("Please select a voucher to split");
			m_SplitInputs[jjj] = txtField;
		}
		m_Remainder.setForeground(Color.BLACK);

		// place the voucher table into a scroll pane, set for radio select mode
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		vtm.deselectAll();
		vtm.setSelectMode(VoucherTableModel.M_SelectSingle);
		JTable vt = m_ParentHome.getVoucherTable();
		JScrollPane vouchSelect = new JScrollPane(vt);
		vouchSelect.setOpaque(true);
		vouchSelect.setPreferredSize(new Dimension(450, 135));

		// note which Issuer is selected, and based on that, what voucher asset is
		// NB: should have at least one voucher, or we wouldn't be here
		XMLVoucher vvv = vtm.getFirstVoucher();
		if (vvv != null) {
			m_VoucherAssetType = vvv.getAsset();
		}
		buildSplitEntry();

		// put this scroll pane into another pane, along with the split entry pane
		JPanel splitPane = new JPanel(new GridBagLayout());
		splitPane.setOpaque(true);
		splitPane.setBorder(null);
		m_GBC.fill = GridBagConstraints.NONE;
		m_GBC.gridheight = 3;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_GBC.anchor = GridBagConstraints.NORTH;
		splitPane.add(vouchSelect, m_GBC);
		splitPane.add(Box.createVerticalStrut(10), m_GBC);
		splitPane.add(m_SplitEntry, m_GBC);

		// put entire form into the main scroll area
		m_WorkArea.setViewportView(splitPane);
	}

	/**
	 * utility method to (re-)build the panel showing the split entry form.
	 * implied param: m_SplitEntry
	 */
	private void buildSplitEntry() {
		m_SplitEntry.setLayout(new GridLayout(3, 3, 10, 5));
		m_SplitEntry.removeAll();

		// for each field, display a label and the field, initialized with any
		// existing value

		String dispUnits = XMLVoucher.getDisplayUnits(m_VoucherAssetType);
		for (int iii = 0; iii < M_MaxSplitOutputs; iii++) {
			// add input field prefaced by its label
			JLabel lab = new JLabel("Amount (" + dispUnits + "): ");
			lab.setFont(m_ParentHome.M_LabelFont);
			lab.setHorizontalAlignment(SwingConstants.TRAILING);
			lab.setLabelFor(m_SplitInputs[iii]);
			m_SplitEntry.add(lab);
			m_SplitEntry.add(m_SplitInputs[iii]);
			m_SplitEntry.add(Box.createHorizontalStrut(20));
		}

		// add remainder
		JLabel rLab = new JLabel("Remainder: ");
		rLab.setFont(m_ParentHome.M_LabelFont);
		rLab.setBorder(m_InsetBorder);
		rLab.setHorizontalAlignment(SwingConstants.TRAILING);
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(3);
		m_Remainder.setText(df.format(m_SplitValues[M_MaxSplitOutputs]));
		rLab.setLabelFor(m_Remainder);
		m_SplitEntry.add(rLab);
		m_SplitEntry.add(m_Remainder);
		JLabel rLab2 = new JLabel(" " + dispUnits);
		rLab2.setFont(m_ParentHome.M_LabelFont);
		rLab2.setBorder(m_InsetBorder);
		m_SplitEntry.add(rLab2);

		// force repaint in case we changed anything
		m_SplitEntry.revalidate();
		m_SplitEntry.repaint();
	}

	/**
	 * method to receive notification of a change in voucher selection
	 */
	public void voucherSelectChanged() {
		// recompute remainder using new selection
		VoucherTableModel vtm = m_ParentHome.getVoucherModel();
		ArrayList<XMLVoucher> selVouchers = vtm.getSelectedVouchers();
		if (!selVouchers.isEmpty()
			 && vtm.getSelectMode() == VoucherTableModel.M_SelectSingle)
		{
			XMLVoucher vouch = selVouchers.get(0);
			m_VoucherAssetType = vouch.getAsset();

			// enable the text entries
			for (int iii = 0; iii < m_SplitInputs.length; iii++) {
				JTextField txtField = m_SplitInputs[iii];
				txtField.setEnabled(true);
				txtField.setToolTipText("Enter a positive decimal number");
			}

			// compute remainder vs. current split output values
			double vAmt = vouch.getNPV();
			double rem = vAmt;
			for (int jjj = 0; jjj < M_MaxSplitOutputs; jjj++) {
				rem -= m_SplitValues[jjj];
			}

			// set new values and rebuild display
			m_SplitValues[M_MaxSplitOutputs] = rem;
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
			df.setMaximumFractionDigits(XMLVoucher.M_Precision);
			df.setMinimumFractionDigits(XMLVoucher.M_Precision);
			df.setGroupingSize(3);
			m_Remainder.setText(df.format(m_SplitValues[M_MaxSplitOutputs]));
			if (rem < 0.0 || rem == vAmt) {
				m_Remainder.setForeground(Color.RED);
				m_SplitButton.setEnabled(false);
				m_SplitButton.setToolTipText("Enter split amounts first");
			}
			else {
				m_Remainder.setForeground(Color.BLACK);
				m_SplitButton.setEnabled(true);
				m_SplitButton.setToolTipText("<html>Click here to split the"
														+ "<br/>selected "
														+ "voucher as indicated</html>");
			}
		}
		// else: not meant for us
	}

	/**
	 * helper method to queue up a modification to a text field from within a
	 * document listener
	 * @param txtField the text field to update
	 * @param newText the replacement text
	 */
	private void queueTextMod(JTextField txtField, String newText) {
		if (txtField == null || newText == null) {
			 return;
		}

		// do not allow simultaneous queued mods
		if (m_DoingUndo) {
			return;
		}
		
		// create and start background modification object
		TextFieldMod modifier = new TextFieldMod(txtField, newText);
		m_DoingUndo = true;
		SwingUtilities.invokeLater(modifier);
	}

	// implement ActionListener
	/**
	 * method to handle ActionEventS from JTextField inputs
	 * @param ae the event from one of the split value inputs
	 */
	@Override
	public void actionPerformed(ActionEvent ae) {
		Object src = ae.getSource();
		if (src == m_IssuerSel) {
			super.actionPerformed(ae);
			return;
		}
		if (src instanceof JTextField) {
			JTextField input = (JTextField) src;
			// move to the next focus item
			input.transferFocus();
		}
	}

	/**
	 * override for WorkPane's method so that we can redraw our split pane
	 * when the selected Issuer changes
	 */
	@Override
	protected void buildValueHeader() {
		prepDisplay();
	}

	// END methods
}
