/*
 * last modified---
 * 	08-15-18/kbw/ explicitly cancel PacketCollectors after use
 * 	01-29-18/kbw/ use getBasePriceLookupUnits()
 * 	07-15-15/jet/ use VPAssetConfig.getPriceLookupUnits() to determine value
 * 	08-05-14/jet/ Spark plugin version
 * 	06-05-13/jet/ SilentVault client version
 *
 * purpose---
 * 	provide a UI widget to handle unclaimed (still outstanding) payments
 */

package org.silentvault.spark.ui.wallet;

import javax.swing.*;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.event.ActionEvent;
import java.awt.Font;
import java.awt.Color;
import java.awt.GridLayout;

import org.dom4j.Element;
import org.dom4j.DocumentException;

import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.IQ;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.util.StringUtils;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;

import java.util.Date;
import java.util.Hashtable;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.security.*;
import java.net.HttpURLConnection;


/**
 * Extends WorkPane to provide a screen for processing unclaimed voucher
 * payments.
 *
 * @author Kevin Wilkerson
 */
public final class UnclaimedPane extends WorkPane {
	// BEGIN data members
	/**
	 * the action to reclaim an unclaimed payment
	 */
	private AbstractAction		m_ReclaimAction;

	/**
	 * associated reclaim button
	 */
	private JButton				m_ReclaimButton;

	/**
	 * the action to skip over this item and move to the next
	 */
	private AbstractAction		m_SkipAction;

	/**
	 * associated skip button
	 */
	private JButton				m_SkipButton;

	/**
	 * the unclaimed payment record we are working with
	 */
	private XMLPayment			m_UnclaimedPmt;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the HomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public UnclaimedPane(WalletClient plugin, HomePane home, WTabManager mgr) {
		super(plugin, home, mgr);

		// provide user textual instructions
		String unclaimedTxt = "This voucher has not been picked up by the "
									+ "person you sent it to.<br/><br/>Reclaim it now "
									+ "to return it to your wallet.";
		m_UserInstructs = unclaimedTxt;

		// set up button actions and associated buttons
		setReclaimAction();
		m_ReclaimButton = new JButton(m_ReclaimAction);
		m_ReclaimButton.setFont(plugin.M_ButtonFont);
		m_SkipAction = new AbstractAction("Do this one later") {
			public void actionPerformed(ActionEvent actionEvent) {
				m_TabManager.getHousekeepingPane().processNextItem(-1);
			}
		};
		m_SkipButton = new JButton(m_SkipAction);
		m_SkipButton.setToolTipText("Click here to skip to the next housekeeping "
											+ "item");
		m_SkipButton.setFont(plugin.M_ButtonFont);
		m_HeadingPane.setLayout(new GridLayout(1, 3));

		// no border around the whole thing
		m_WorkArea.setBorder(null);
	}


	// method required to extend WorkPane
	/**
	 * method to prepare display for this sub-class
	 */
	public void prepDisplay() {
		// set our heading pane
		buildValueHeader();

		// set screen label
		buildScreenLabel("Unclaimed Voucher");

		// get the unclaimed payment we're working with
		HousekeepingPane hkp = m_TabManager.getHousekeepingPane();
		m_UnclaimedPmt = hkp.getPaymentRecord();
		if (m_UnclaimedPmt == null) {
			Log.error("Missing unclaimed payment record, cannot process");
			showError("cannot do reclaim item",
						"unclaimed voucher payment not found");
			return;
		}

		m_WorkBar.removeAll();
		if (hkp.automating()) {
			// push our own button, in another thread to avoid recursion
			Runnable doReclaim = new Runnable() {
				public void run() {
					sendReclaimPayment(m_UnclaimedPmt);
				}
			};
			new Thread(doReclaim).start();
			return;
		}

		// add the return to overview button as the sole contents of toolbar
		m_WorkBar.add(m_MainButton);
		m_WorkBar.setVisible(true);

		// provide action buttons in the button pane
		// first, the button to skip over this item
		m_ButtonPane.removeAll();
		m_ButtonPane.add(m_SkipButton);
		m_ButtonPane.add(Box.createHorizontalStrut(215));	 // spacer

		// next, the button to reclaim this voucher
		// -- disable if it's not reclaimable (though we shouldn't get here if so)

		if (m_UnclaimedPmt.isExpired()) {
			m_ReclaimButton.setEnabled(true);
			m_ReclaimButton.setToolTipText("Click here to reclaim this voucher "
													+ "payment");
		}
		else {
			m_ReclaimButton.setToolTipText("This voucher payment has not yet "
													+ "expired");
			m_ReclaimButton.setEnabled(false);
		}
		m_ButtonPane.add(m_ReclaimButton);

		// place the details about the payment into the work area
		JPanel unclDets = new JPanel(new GridBagLayout());
		m_GBC.fill = GridBagConstraints.NONE;
		m_GBC.anchor = GridBagConstraints.NORTHWEST;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		unclDets.add(Box.createVerticalGlue(), m_GBC);
		final Color labColor = UIManager.getColor("standardGray");

		// payee
		JLabel sentLab = new JLabel("Voucher sent to:  ");
		sentLab.setFont(m_ParentHome.M_LabelFont);
		sentLab.setForeground(labColor);
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		unclDets.add(sentLab, m_GBC);
		JLabel sent
			= new JLabel(m_UnclaimedPmt.getPayee(), SwingConstants.TRAILING);
		sent.setFont(m_ParentHome.M_DataFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		unclDets.add(sent, m_GBC);
		unclDets.add(Box.createVerticalStrut(15), m_GBC);

		// pick up date (past)
		JLabel dateLab = new JLabel("Expired at: ");
		dateLab.setFont(m_ParentHome.M_LabelFont);
		dateLab.setForeground(labColor);
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		unclDets.add(dateLab, m_GBC);
		Date pickupDate = new Date(m_UnclaimedPmt.getTTL());
		SimpleDateFormat sdf = new SimpleDateFormat(XMLVoucher.M_DateFmt);
		JLabel pickup
			= new JLabel(sdf.format(pickupDate), SwingConstants.TRAILING);
		pickup.setFont(m_ParentHome.M_DataFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		unclDets.add(pickup, m_GBC);
		unclDets.add(Box.createVerticalStrut(15), m_GBC);

		// amount in base units
		String dispUnits = m_UnclaimedPmt.getDisplayUnits();
		JLabel amtLab = new JLabel("Amount:");
		amtLab.setFont(m_ParentHome.M_LabelFont);
		amtLab.setForeground(labColor);
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		unclDets.add(amtLab, m_GBC);
		double quan = m_UnclaimedPmt.getQuantity();
		StringBuilder amtTxt = new StringBuilder(32);
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(0);
		df.setGroupingSize(3);
		amtTxt.append(df.format(quan) + " " + dispUnits);
		JLabel amt = new JLabel(amtTxt.toString(), SwingConstants.TRAILING);
		amt.setFont(m_ParentHome.M_DataFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		unclDets.add(amt, m_GBC);
		unclDets.add(Box.createVerticalStrut(5), m_GBC);

		// also show amount in preferred fiat currency
		String prefCurrency = m_Plugin.getUserPreferences().getValueCurrency();
		String lookUnits
			= VPAssetConfig.getBasePriceLookupUnits(m_UnclaimedPmt.getAsset());
		AssetPriceCache apc = m_Plugin.getPricingModule();
		Double price = apc.getRelativePrice(lookUnits, prefCurrency);
		if (price == null) {
			price = new Double(1.0);
			prefCurrency = dispUnits;
		}
		StringBuilder fiatTxt = new StringBuilder(32);
		fiatTxt.append("( ");
		df.setMinimumFractionDigits(2);
		df.setMaximumFractionDigits(2);
		fiatTxt.append(df.format(quan * price.doubleValue()));
		fiatTxt.append(" " + prefCurrency + " )");
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		unclDets.add(Box.createHorizontalStrut(50), m_GBC);
		JLabel fiat = new JLabel(fiatTxt.toString(), SwingConstants.TRAILING);
		fiat.setFont(m_ParentHome.M_DataItalFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		unclDets.add(fiat, m_GBC);
		unclDets.add(Box.createVerticalStrut(15), m_GBC);

		// see what baggage we have
		Hashtable<String, String> pmtBaggage = m_UnclaimedPmt.getBaggageList();
		for (String bagFld : pmtBaggage.keySet()) {
			String bagVal = pmtBaggage.get(bagFld);
			String bagNm = bagFld;
			if (bagFld.equals("Message")) {
				bagNm = "BYour message";
			}

			// show all pairs, knocking first char ('B') off of names
			JLabel bagLab = new JLabel(bagNm.substring(1) + ":");
			bagLab.setFont(m_ParentHome.M_LabelFont);
			bagLab.setForeground(labColor);
			m_GBC.gridwidth = GridBagConstraints.RELATIVE;
			unclDets.add(bagLab, m_GBC);

			// use a text area so we can get it all to fit
			JTextArea bagTxt = new JTextArea(1, 20);
			bagTxt.setText(bagVal);
			bagTxt.setFont(m_ParentHome.M_DataFont);
			bagTxt.setEditable(false);
			bagTxt.setBorder(null);
			bagTxt.setBackground(m_Plugin.m_LabelBackColor);
			bagTxt.setLineWrap(true);
			bagTxt.setWrapStyleWord(true);
			bagTxt.setFocusable(false);
			m_GBC.gridwidth = GridBagConstraints.REMAINDER;
			unclDets.add(bagTxt, m_GBC);
			unclDets.add(Box.createVerticalStrut(15), m_GBC);
		}
		unclDets.add(Box.createVerticalStrut(25), m_GBC);

		// voucher serial number
		JLabel serialLab = new JLabel("Voucher serial number:  ");
		serialLab.setFont(m_ParentHome.M_DetailFont);
		serialLab.setForeground(labColor);
		m_GBC.gridwidth = GridBagConstraints.RELATIVE;
		unclDets.add(serialLab, m_GBC);
		JLabel serial = new JLabel(m_UnclaimedPmt.getSerialNumber(),
											SwingConstants.TRAILING);
		serial.setFont(m_ParentHome.M_DetailFont);
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		unclDets.add(serial, m_GBC);
		unclDets.add(Box.createVerticalGlue(), m_GBC);

		// install everything into the main WorkPane scroll area
		m_WorkArea.setViewportView(unclDets);
	}

	/**
	 * private utility method to establish the Action which can process a
	 * request to reclaim an expired outgoing payment
	 */
	private void setReclaimAction() {
		m_ReclaimAction = new AbstractAction("Reclaim") {
			public void actionPerformed(ActionEvent actionEvent) {
				// prepare and send a message to the OFS to reclaim this payment
				sendReclaimPayment(m_UnclaimedPmt);
			}
		};
	}

	/**
	 * private utility method to request the OFS reclaim an expired payment.
	 * NB: this method will block pending the OFS's reply
	 * @param pmt the payment to reclaim
	 */
	private void sendReclaimPayment(XMLPayment pmt) {
		if (pmt == null || !pmt.isInit()) {
			return;
		}
		// double check it's expired (should be, since we only do those)
		if (!m_UnclaimedPmt.isExpired()) {
			showError("cannot process unclaimed payment record",
						"the selected record has not yet expired");
			return;
		}
		
		VscState state = m_Plugin.getVscState();
		if (!state.isLoggedIn()) {
			Log.error("Cannot reclaim payments while not logged in!");
			showError("cannot reclaim payment", "not logged in", "log in first");
			return;
		}
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		final XMPPConnection conn = m_Plugin.getOFSConnection();

		// construct payDetails
		StringBuilder payDets = new StringBuilder(512);
		payDets.append("<payer_VS>" + secrets.getVSnumber() + "</payer_VS>");
		payDets.append("<payee_VS>" + pmt.getPayee() + "</payee_VS>");
		payDets.append("<voucher_serial>" + pmt.getSerialNumber());
		payDets.append("</voucher_serial>");
		payDets.append("<issuer>" + pmt.getIssuer() + "</issuer>");
		payDets.append("<quantity>" + pmt.getQuantity() + "</quantity>");
		payDets.append("<units>" + pmt.getUnits() + "</units>");
		payDets.append("<asset>" + pmt.getAsset() + "</asset>");

		// sign the details with our private key
		String signedDets = payDets.toString();
		Signature dSig = null;
		byte[] sig = null;
		boolean ok = false;
		try {
			dSig = Signature.getInstance("SHA1withRSA");
			dSig.initSign(secrets.getPrivKey());
			dSig.update(signedDets.getBytes());
			sig = dSig.sign();
			ok = true;
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("No such signature algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid privkey for details signature", ike);
		}
		catch (SignatureException se) {
			Log.error("Unable to create details signature", se);
		}
		if (!ok) {
			Log.error("Could not place signature on <payDetails/> block");
			showError("unable to reclaim payment",
						"error signing request details");
			return;
		}
		payDets.append("<signature>" + StringUtils.encodeBase64(sig, false));
		payDets.append("</signature>");

		// encrypt the whole block with VP's public key
		String payBlock = "<payDetails>" + payDets.toString() + "</payDetails>";
		String encPay = EncodingUtils.makeBase64PubkeyEncStr(payBlock,
																				secrets.getVPKey());
		if (encPay == null || encPay.isEmpty()) {
			showError("unable to reclaim payment", "error encrypting request");
			return;
		}

		// construct signed voucherRequest object
		VoucherRequest vrq = new VoucherRequest();
		vrq.setAction("validate");
		vrq.setSigningVS(secrets.getVSnumber());
		vrq.setFolderHash(new String(secrets.getVoucherIndex()));
		vrq.setFolderCap(secrets.getVouchRWCap());
		vrq.setInitialized();
		if (!vrq.signDetails(secrets.getPrivKey())) {
			showError("unable to reclaim payment",
						"error signing voucher request");
			return;
		}
		if (!vrq.encryptDetails(secrets.getVPKey(), false)) {
			showError("unable to reclaim payment",
						"error encrypting voucher request");
			return;
		}

		// wait cursor and grey out buttons
		m_ParentHome.setCursor(M_WaitCursor);
		setButtonState(false);

		// build message to send
		OFSMessage recPmt = new OFSMessage();
		recPmt.setType(IQ.Type.SET);
		recPmt.setFrom(conn.getUser());
		recPmt.setTo(conn.getServiceName());
		recPmt.setPacketID("reclaim_" + state.getNextOFSid());
		recPmt.setOpcode("REQ_reclaim_payment");
		recPmt.setVoucher_publisher(secrets.getPublisher());
		recPmt.setPayDetails(encPay);
		recPmt.setVouchReq(vrq);

		// register a synchronous collector for the reply
		PacketIDFilter pFilter = new PacketIDFilter(recPmt.getPacketID());
		PacketCollector repColl = conn.createPacketCollector(pFilter);

		// send it to the server
		conn.sendPacket(recPmt);

		// record outgoing packet
		state.setLastActivity(recPmt);

		// wait for reply or timeout
		OFSMessage rep = null;
		IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
		m_ParentHome.setCursor(null);
		setButtonState(true);
		if (reply == null) {
			// got timeout; build error
			rep = new OFSMessage();
			rep.setOpcode("REP_payment_reverted");
			rep.setFrom(recPmt.getTo());
			rep.setTo(recPmt.getFrom());
			rep.setPacketID(recPmt.getPacketID());
			rep.setType(IQ.Type.ERROR);
			rep.setErrcode(HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
			rep.setErrmsg("timed out");
		}
		else {
			rep = (OFSMessage) reply;
		}
		repColl.cancel();

		// invoke callback with result (which will call processReclaimReply())
		m_Plugin.getOFSListener().processPacket(rep);
	}

	/**
	 * process a reply to a payment reclaim operation
	 * @param revert the reply message from the OFS
	 * implied param: payment record will be stored in m_UnclaimedPmt
	 */
	public void processReclaimReply(OFSMessage revert) {
		if (revert == null
			|| !revert.getOpcode().equalsIgnoreCase("REP_payment_reverted"))
		{
			Log.error("null or improper payment reclaim reply");
			return;
		}
		final String err = "cannot reclaim payment";
		final HousekeepingPane hkp = m_TabManager.getHousekeepingPane();

		// obtain the pending payment we're working on
		if (m_UnclaimedPmt == null) {
			Log.error("proc reply cannot find saved XMLPayment record");
			showError(err, "saved payment record not found");
			return;
		}

		// see if reply was an error
		if (revert.getType() == IQ.Type.ERROR) {
			int code = revert.getErrcode();

			// if it's a 404, remove this payment from the list
			if (code == HttpURLConnection.HTTP_NOT_FOUND) {
				if (!hkp.automating()) {
					showError(err, "voucher not found; payee must have picked it "
								+ "up, or you reclaimed it previously",
								"look for a receipt from the payee using Refresh, or "
								+ "in your receipt history");
				}

				/* in case they've waited too long to get their receipt and it's
				 * now expired off DHT, delete this pending payment from their safe
				 */
				m_ParentHome.setCursor(M_WaitCursor);
				hkp.sendDelPendPmt(m_UnclaimedPmt, false);
				m_UnclaimedPmt = null;
				// NB: display will be updated when reply to delete is received
			}
			else {
				// do not delete the item, and to prevent recursion disable auto
				if (hkp.automating()) {
					showError(err, revert.getErrmsg(),
								"switching to manual processing of housekeeping items");
					hkp.manualOverride();
				}
				else {
					showError(err, revert.getErrmsg());
				}
			}
			return;
		}

		// decode/decrypt the <confirmation/>
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		String decConf
			= EncodingUtils.getStrFromBase64PubkeyEnc(revert.getConfirmation(),
																	secrets.getPrivKey());
		if (decConf == null || decConf.isEmpty()) {
			Log.error("Unable to decode <confirmation/> of " + revert.getOpcode());
			showError("error processing payment reclaim reply",
						"could not decode");
			return;
		}

		// parse it
		Element confElt = EncodingUtils.buildElement(decConf);
		if (confElt == null) {
			Log.error("Unable to parse <confirmation/> of " + revert.getOpcode());
			showError("error processing payment reclaim reply",
						"could not parse confirmation details");
			return;
		}

		// get the <voucher/>
		Element vouchElt = confElt.element("voucher");
		if (vouchElt == null) {
			Log.error("Missing <voucher/> element in <confirmation/>");
			showError("error processing payment reclaim reply",
						"missing voucher element");
			return;
		}

		// get voucher's filename
		Element vouchFileElt = confElt.element("voucherFile");
		if (vouchFileElt == null) {
			Log.error("Missing <voucherFile/> element in <confirmation/>");
			showError("error processing payment reclaim reply",
						"missing voucher filename element");
			return;
		}
		String vouchFile = vouchFileElt.getTextTrim();

		// get VP's signature, and verify it
		Element sigElt = confElt.element("signature");
		if (sigElt == null) {
			Log.error("Missing <signature/> element in <confirmation/>");
			showError("error processing payment reclaim reply",
						"missing signature");
			return;
		}
		StringBuilder cSigData = new StringBuilder(1024);
		cSigData.append("<confirmation>");
		cSigData.append(vouchElt.asXML());
		if (vouchFile.isEmpty()) {
			cSigData.append("<voucherFile/>");
		}
		else {
			cSigData.append("<voucherFile>" + vouchFile + "</voucherFile>");
		}
		Signature cSig = null;
		boolean ok = false;
		try {
			cSig = Signature.getInstance("SHA1withRSA");
			cSig.initVerify(secrets.getVPKey());
			cSig.update(cSigData.toString().getBytes());
			ok = cSig.verify(StringUtils.decodeBase64(sigElt.getTextTrim()));
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("No such sig algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid pubkey for signature validation", ike);
		}
		catch (SignatureException se) {
			Log.error("Unable to validate <confirmation/> signature", se);
		}
		if (!ok) {
			Log.error("publisher's sig on reclaim confirmation did not verify");
			showError("error processing payment reclaim reply",
						"publisher's signature on confirmation did not verify");
			return;
		}

		// now parse the voucher and validate its signature
		XMLVoucher newVouch = null;
		try {
			newVouch = XMLVoucher.getVoucherRep(vouchElt);
		}
		catch (DocumentException de) {
			Log.error("Unparsable <voucher/> element in <confirmation/>");
			showError("error processing payment reclaim reply",
						"voucher element did not parse");
			return;
		}

		if (!newVouch.validateSig(secrets.getVPKey())) {
			Log.error("Bad sig on <voucher/> element in <confirmation/>");
			showError("error processing payment reclaim reply",
						"voucher signature did not verify");
			return;
		}

		// show warning message if we're pending
		if (revert.syncPending()) {
			showError("warning: synchronization pending",
						"the payment reclaim succeeded but the publisher has queued "
						+ "storing the new voucher in your wallet",
						"wait a few minutes and then use Resync button to "
						+ "see the new voucher");
		}
		else {
			// tell parent about any new voucher so it can update display
			m_ParentHome.addVoucher(newVouch, vouchFile);
		}

		// delete the pending payment from our safe
		m_ParentHome.setCursor(M_WaitCursor);
		hkp.sendDelPendPmt(m_UnclaimedPmt, false);
		// tell housekeeping this one worked successfully
		hkp.recordReclaimSuccess(m_UnclaimedPmt);
		m_UnclaimedPmt = null;
		// NB: display will be updated when reply to delete is received
	}

	// END methods
}
