/*
 * last modified---
 * 	10-26-16/jet/ add <pendDeploy/> section to REP_account_details
 * 	09-13-16/jet/ add <vfqRatio/> to REP_stats
 * 	09-06-16/jet/ add <vfqTotal/> to REP_betting_ended
 * 	08-06-16/jet/ add REP_account_hist, REP_tranche_hist, other new fields
 * 	04-30-15/jet/ Spark plugin version
 * 	04-14-15/jet/ add <tossSeconds/> to REP_stats
 * 	02-10-15/jet/ add <age/> to REP_order_book orders
 * 	01-16-15/jet/ rename from SAVBlock
 * 	12-16-14/jet/ add <bettorId/> to REP_betting_init
 * 	11-12-14/jet/ put back REP_xfer_confirmed
 * 	11-11-14/jet/ add DiscountPolicy and its usage with REP_account_details
 * 	11-06-14/jet/ new
 *
 * purpose---
 * 	class to instantiate a reply message from the ABC to the VSC
 *
 * notes---
 * 	maintain in parallel with com.silentvault.market.SAVBlock
 *
 * 	these messages are delivered embedded inside a <rep_blob/> element
 * 	within a REP_svm_message message routed from the OFS
 *
 * 	see underlying message definitions in vsc_sav.xml and vsc_ofs-svm.xml
 */

package org.silentvault.spark;

import org.dom4j.Element;

import org.jivesoftware.spark.util.log.Log;

import java.util.Date;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Locale;
import java.text.NumberFormat;
import java.text.DecimalFormat;


/**
 * This class instantiates server replies generated by the ABC to a connected
 * client VSC via the OFS gateway.  It knows how to emit itself as a
 * correctly formatted XML body based on the embedded opcode.  It also
 * deals with validating signatures, and representing error returns.
 *
 * @author Justin Turrell
 */
public final class ABCBlock extends SVMBlock {
	// BEGIN unique data members
	/**
	 * constant indicating that a discount policy applies to purchases
	 */
	public static final String	M_PolicyPurch = "purchase";
	/**
	 * constant indicating that a discount policy applies to sales
	 */
	public static final String	M_PolicySale = "sale";

	/**
	 * bettor ID (public user number on tables or in market, used with
	 * opcodes REP_balances, REP_betting_init, and REP_account_details)
	 */
	private int					m_BettorId;

	/**
	 * total balance of BTC held in account, including all tranches (used
	 * with opcode REP_balances)
	 */
	private double				m_BTCTotal;

	/**
	 * total balance of IB held in all tranches (used with opcode REP_balances)
	 */
	private double				m_IBTotal;

	/**
	 * total balance of VFQ held in account (used with opcodes REP_balances and
	 * REP_betting_ended)
	 */
	private double				m_VFQTotal;

	/**
	 * account balance of BTC not in a tranche (used with opcodes REP_tranches,
	 * REP_betting_ended, and REP_tranche_modified)
	 */
	private double				m_BTCBalance;

	/**
	 * inner class to represent a tranche specification
	 */
	public final class Tranche {
		/**
		 * ID of tranche within account
		 */
		public int		m_TrancheId;

		/**
		 * amount of BTC chips in tranche
		 */
		public double	m_BTC;

		/**
		 * amount of IB in tranche
		 */
		public double	m_IB;

		/**
		 * original BTC funding amount provided by account
		 */
		public double	m_StartBTC;

		/**
		 * sum total of any BTC additions made on top of m_StartBTC
		 */
		public double	m_Additions;

		/**
		 * flag whether or not tranche is presently betting
		 */
		public boolean	m_Active;

		/**
		 * tranche's current Martingale betting iteration
		 */
		public byte		m_Iteration;

		/**
		 * current Martingale betting increment (not the last bet)
		 */
		public double	m_BetIncrement;

		/**
		 * current count of consecutive losses in Martingale cycle
		 */
		public byte		m_Losses;

		/**
		 * ID of open order to buy or sell IB, or zero if none
		 */
		public int		m_OpenOrder;

		/**
		 * flag whether or not tranche is presently receiving extra IB bonusing
		 */
		public boolean	m_Bonus;

		/**
		 * constructor
		 * @param tId tranche ID number
		 */
		public Tranche(int tId) {
			m_TrancheId = tId;
		}

		/**
		 * nullary constructor
		 */
		public Tranche() {
			this(0);
		}
	}

	/**
	 * list of tranches owned by account (used with opcode REP_tranches)
	 */
	private ArrayList<Tranche>	m_Tranches;

	/**
	 * flag indicating whether account presently has at least 1 tranche betting
	 * (used with opcode REP_account_details)
	 */
	private boolean				m_Active;

	/**
	 * initial bankroll in BTC wallet transferred into account (used with
	 * opcodes REP_account_details and REP_betting_init)
	 */
	private double					m_Bankroll;

	/**
	 * subsequent bankroll additions made to account (used with opcode
	 * REP_account_details)
	 */
	private double					m_Additions;

	/**
	 * subsequent bankroll subtractions made from account (used with opcode
	 * REP_account_details)
	 */
	private double					m_Subtractions;

	/**
	 * starting BTC chip total across all tranches (used with opcode
	 * REP_account_details)
	 */
	private double					m_BTCChipStart;

	/**
	 * size (in BTC) configured per tranche (used with opcodes
	 * REP_account_details and REP_details_config)
	 */
	private float					m_TrancheSize;

	/**
	 * the first toss in which account participated in betting, in YYMMDDHHmm
	 * format (used with opcode REP_account_details)
	 */
	private int						m_FirstToss;

	/**
	 * the last or most recent toss in which account participated in betting,
	 * in YYMMDDHHmm format (used with opcode REP_account_details)
	 */
	private int						m_LastToss;

	/**
	 * highest cash-out net value (assuming IB redemption at Issuer) seen
	 * (used with opcode REP_account_details)
	 */
	private double					m_HighValue;

	/**
	 * toss on which m_HighValue occurred, as YYMMDDHHmm
	 * (used with opcode REP_account_details)
	 */
	private int						m_HighToss;

	/**
	 * lowest cash-out net value (assuming IB redemption at Issuer) seen
	 * (used with opcode REP_account_details)
	 */
	private double					m_LowValue;

	/**
	 * toss on which m_LowValue occurred, as YYMMDDHHmm
	 * (used with opcode REP_account_details)
	 */
	private int						m_LowToss;

	/**
	 * number of tosses after which a 'wholeness' check occurs for this account,
	 * overriding the system default (used with opcode REP_account_details), or
	 * representing the system default (when used with opcode REP_stats)
	 */
	private int						m_WholenessTosses;

	/**
	 * the toss on which a 'wholeness' check was last conducted for this account,
	 * in YYMMDDHHmm format (used with opcode REP_account_details)
	 */
	private int						m_LastWhChkToss;

	/**
	 * profit goal (upper stop) set by owner, in percent, or zero if none set
	 * (used with opcodes REP_account_details and REP_details_config)
	 */
	private float					m_ProfitTarget;

	/**
	 * stop loss limit set by owner, in percent, or zero if none set
	 * (used with opcodes REP_account_details and REP_details_config)
	 */
	private float					m_StopLoss;

	/**
	 * owner's selected sale discount policy (string expansion, in this format:
	 * policyID##discount1 (%): tosses to wait :: discount2 (%) : tosses to wait)
	 * (used with opcodes REP_account_details and REP_details_config)
	 */
	private String					m_SalePolicy;

	/**
	 * owner's selected purchase discount policy (string expansion, in format:)
	 * policyID##discount1 (%): tosses to wait :: discount2 (%) : tosses to wait)
	 * (used with opcodes REP_account_details and REP_details_config)
	 */
	private String					m_PurchPolicy;

	/**
	 * owner's selection as to permitting 'bleed-in' of VFQ balance to pay for
	 * IB purchases (used with opcode REP_details_config)
	 */
	private boolean				m_UseVFQ;

	/**
	 * count of accounts current betting (used with opcode REP_stats)
	 */
	private int						m_BettorCount;

	/**
	 * scheduled time of next toss, as Unix timestamp in UTC (used with opcode
	 * REP_stats)
	 */
	private long					m_NextToss;

	/**
	 * number of seconds until next toss occurs (used with opcode REP_stats)
	 */
	private int						m_TossSeconds;

	/**
	 * number of tosses occurring per hour (used with opcode REP_stats)
	 */
	private byte					m_TossRate;

	/**
	 * current ratio of BTC backing IB at Issuer (used with opcode REP_stats)
	 */
	private double					m_IBRatio;

	/**
	 * current ratio of VFQ in relation to IB at Issuer (used with opcode
	 * REP_stats)
	 */
	private double					m_VFQRatio;

	/**
	 * inner class to represent an open order specification.
	 */
	public final class OpenOrder {
		/**
		 * unique order Id in database
		 */
		public int		m_OrderId;

		/**
		 * tranche which owns the order
		 */
		public int		m_TrancheId;

		/**
		 * quantity of IB involved
		 */
		public double	m_IB;

		/**
		 * first toss where this order was available for resolution
		 */
		public int		m_FirstToss;

		/**
		 * age of order, in tosses
		 */
		public byte		m_Age;

		/**
		 * discount policy followed on sales (empty if order is a buy)
		 */
		public String	m_SalePolicy;

		/**
		 * discount policy followed on purchases (empty if order is a sell)
		 */
		public String	m_PurchPolicy;

		/**
		 * constructor
		 * @param oId the order ID number
		 */
		public OpenOrder(int oId) {
			m_OrderId = oId;
			m_SalePolicy = m_PurchPolicy = "";
		}

		/**
		 * nullary constructor
		 */
		public OpenOrder() {
			this(0);
		}
	}

	/**
	 * list of open orders to be returned (used with opcode REP_order_book)
	 */
	private ArrayList<OpenOrder>	m_OpenOrders;

	/**
	 * inner class to represent a discount policy specification
	 */
	public final class DiscountPolicy {
		/**
		 * policy ID in database
		 */
		public short	m_PolicyId;

		/**
		 * whether the policy is for purchases or sales
		 */
		public String	m_Type;

		/**
		 * descriptive name of the policy
		 */
		public String	m_Name;

		/**
		 * policy specification in disc%:tosses::disc%:tosses::... format
		 */
		public String	m_Spec;

		/**
		 * constructor
		 * @param pId the policy ID number
		 */
		public DiscountPolicy(short pId) {
			m_PolicyId = pId;
			m_Type = m_Name = m_Spec = "";
		}

		/**
		 * nullary constructor
		 */
		public DiscountPolicy() {
			this((short)0);
		}
	}

	/**
	 * list of possible discount policies fetched from server (used with opcode
	 * REP_account_details)
	 */
	private ArrayList<DiscountPolicy> m_Policies;

	/**
	 * inner class to represent an account history entry
	 */
	public final class AccountHistory {
		/**
		 * toss on which this history was generated (database index, YYMMDDHHmm)
		 */
		public int		m_Toss;

		/**
		 * total amount of BTC account had (in tranches)
		 */
		public double	m_BtcBal;

		/**
		 * total amount of BTC chips account had (in tranches)
		 */
		public double	m_BtcChipBal;

		/**
		 * total amount of IB account had (in tranches)
		 */
		public double	m_IBBal;

		/**
		 * total amount of VFQ account had (never in tranches)
		 */
		public double	m_VFQBal;

		/**
		 * total additions to the deployed bankroll, made on this particular toss
		 */
		public double	m_Additions;

		/**
		 * total subtractions from the bankroll, made on this particular toss
		 */
		public double	m_Subtractions;

		/**
		 * IB ratio which existed as of this toss (to calculate IB value)
		 */
		public double	m_IBRatio;

		/**
		 * profit goal which existed as of this toss, in percent
		 */
		public float	m_ProfitTarget;

		/**
		 * stop loss level which existed as of this toss, in percent
		 */
		public float	m_StopLoss;

		/**
		 * count of tranches, betting or not, that existed on this toss
		 */
		public int		m_TrancheCount;

		/**
		 * tranche size which owner had set, as of this toss
		 */
		public float	m_TrancheSize;

		/**
		 * constructor
		 * @param toss the toss number
		 */
		public AccountHistory(int toss) {
			m_Toss = toss;
		}

		/**
		 * nullary constructor
		 */
		public AccountHistory() {
			this(0);
		}
	}

	/**
	 * list of account history entries which a given account has (used with
	 * opcode REP_account_hist)
	 */
	private ArrayList<AccountHistory> m_AccountHistory;

	/**
	 * inner class to represent a tranche history entry
	 */
	public final class TrancheHistory {
		/**
		 * ID of tranche within account
		 */
		public int		m_TrancheId;

		/**
		 * toss on which this history was generated (database index, YYMMDDHHmm)
		 */
		public int		m_Toss;

		/**
		 * total amount of free BTC tranche had
		 */
		public double	m_BtcBal;

		/**
		 * total amount of BTC chips tranche had
		 */
		public double	m_BtcChipBal;

		/**
		 * total amount of IB tranche had
		 */
		public double	m_IBBal;

		/**
		 * total additions to the BTC start bankroll, made on this particular toss
		 */
		public double	m_Additions;

		/**
		 * IB ratio which existed as of this toss (to calculate IB value)
		 */
		public double	m_IBRatio;

		/**
		 * flag whether or not tranche was receiving subsidy bonusing on this toss
		 */
		public boolean	m_Bonus;

		/**
		 * flag whether or not tranche was actively betting on this toss
		 */
		public boolean	m_Active;

		/**
		 * constructor
		 * @param tId the tranche Id
		 * @param toss the toss number
		 */
		public TrancheHistory(int tId, int toss) {
			m_TrancheId = tId;
			m_Toss = toss;
		}

		/**
		 * nullary constructor
		 */
		public TrancheHistory() {
			this(0, 0);
		}
	}

	/**
	 * list of tranche history entries which a given tranche has (used with
	 * opcode REP_tranche_history)
	 */
	private ArrayList<TrancheHistory> m_TrancheHistory;

	/**
	 * inner class to represent a queued deployment awaiting suitable IB ratio
	 */
	public final class QueuedDeploy {
		/**
		 * toss on which this entry was queued (YYMMDDHHmm)
		 */
		public int		m_Toss;

		/**
		 * the amount to be deployed when the ratio is satisfied
		 */
		public double	m_Bankroll;

		/**
		 * the required IB reserve ratio which must be attained before processing
		 */
		public float	m_IBRatio;

		/**
		 * constructor
		 * @param toss the toss number on which the queue item was entered
		 */
		public QueuedDeploy(int toss) {
			m_Toss = toss;
		}

		/**
		 * nullary constructor
		 */
		public QueuedDeploy() {
			this(0);
		}
	}

	/**
	 * list of queued deployments to tranches (used with REP_account_details)
	 */
	private ArrayList<QueuedDeploy>	m_QueuedDeployments;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param isErr true if this block represents a returned error message
	 */
	public ABCBlock(boolean isErr) {
		super(isErr);
		m_SalePolicy = m_PurchPolicy = "";
		m_Tranches = new ArrayList<Tranche>(100);
		m_OpenOrders = new ArrayList<OpenOrder>(100);
		m_Policies = new ArrayList<DiscountPolicy>();
		m_AccountHistory = new ArrayList<AccountHistory>(10000);
		m_TrancheHistory = new ArrayList<TrancheHistory>(10000);
		m_QueuedDeployments = new ArrayList<QueuedDeploy>();
	}

	/**
	 * nullary constructor (equivalent to invoking ABCBlock(false))
	 */
	public ABCBlock() {
		this(false);
	}


	// get accessor methods
	/**
	 * obtain the bettor ID
	 * @return the public number assigned to the account on the tables
	 */
	public int getBettorId() { return m_BettorId; }

	/**
	 * obtain the total bitcoin held by the account
	 * @return the amount of BTC owned, including any chips in tranches
	 */
	public double getBTCTotal() { return m_BTCTotal; }

	/**
	 * obtain the total IB held by the account
	 * @return the amount of IB held across all tranches
	 */
	public double getIBTotal() { return m_IBTotal; }

	/**
	 * obtain the total VFQ held by the account
	 * @return the amount of VFQ held (never included in tranches)
	 */
	public double getVFQTotal() { return m_VFQTotal; }

	/**
	 * obtain the total BTC held outside of chips in tranches
	 * @return the amount of "free" BTC available for withdrawal
	 */
	public double getBTCBalance() { return m_BTCBalance; }

	/**
	 * obtain the list of tranches owned by the account
	 * @return the tranche list (may be empty)
	 */
	public ArrayList<Tranche> getTranches() { return m_Tranches; }

	/**
	 * determine whether the account is actively betting
	 * @return true if the account has at least one tranche currently betting
	 */
	public boolean isActive() { return m_Active; }

	/**
	 * obtain the initial bankroll in BTC transferred into the account
	 * @return the total amount of BTC funded by the owning wallet
	 */
	public double getBankroll() { return m_Bankroll; }

	/**
	 * obtain the subsequent bankroll additions in BTC
	 * @return the total additions made on top of the original bankroll
	 */
	public double getAdditions() { return m_Additions; }

	/**
	 * obtain the subsequent bankroll subtractions in BTC
	 * @return the total subtractions made from the original bankroll
	 */
	public double getSubtractions() { return m_Subtractions; }

	/**
	 * obtain the initial quantity of BTC transferred into BTC chips in tranches
	 * @return the total aggregate initial funding of all tranches on BTC side
	 */
	public double getBTCChipStart() { return m_BTCChipStart; }

	/**
	 * obtain the currently configured tranche size
	 * @return the amount (in BTC) with which each new tranche will be funded
	 */
	public float getTrancheSize() { return m_TrancheSize; }

	/**
	 * obtain the first toss on which the account participated in betting
	 * @return the toss number in YYMMDDHHmm format, or 0 if not betting
	 */
	public int getFirstToss() { return m_FirstToss; }

	/**
	 * obtain the most recent toss on which the account participated in betting
	 * @return the toss number in YYMMDDHHmm format, or 0 if not betting
	 */
	public int getLastToss() { return m_LastToss; }

	/**
	 * obtain the highest redemption cash-out value observed during betting
	 * @return the highest value calculated, or 0.0 if not betting
	 */
	public double getHighValue() { return m_HighValue; }

	/**
	 * obtain the toss on which the highest redemption value occurred
	 * @return the toss number in YYMMDDHHmm format
	 */
	public int getHighToss() { return m_HighToss; }

	/**
	 * obtain the lowest redemption cash-out value observed during betting
	 * @return the lowest value calculated, or 0.0 if not betting
	 */
	public double getLowValue() { return m_LowValue; }

	/**
	 * obtain the toss on which the lowest redemption value occurred
	 * @return the toss number in YYMMDDHHmm format
	 */
	public int getLowToss() { return m_LowToss; }

	/**
	 * obtain the number of tosses after which account is eligible for wholeness
	 * @return the number of tosses separating wholeness checks
	 */
	public int getWholenessTosses() { return m_WholenessTosses; }

	/**
	 * obtain the toss on which the account was last checked for wholeness
	 * @return the toss number in YYMMDDHHmm format
	 */
	public int getLastWhChkToss() { return m_LastWhChkToss; }

	/**
	 * obtain the upper stop out profit target
	 * @return the point at which betting will cease, in percent
	 */
	public float getProfitTarget() { return m_ProfitTarget; }

	/**
	 * obtain the lower stop out loss limit
	 * @return the point at which betting will cease, in percent
	 */
	public float getStopLoss() { return m_StopLoss; }

	/**
	 * obtain the sale discount policy specification
	 * @return the spec, in format:
	 * policyID##discount1 (%): tosses to wait :: discount2 (%) : tosses 2 :: ...
	 */
	public String getSalePolicy() { return m_SalePolicy; }

	/**
	 * obtain the purchase discount policy specification
	 * @return the spec, in format:
	 * policyID##discount1 (%): tosses to wait :: discount2 (%) : tosses 2 :: ...
	 */
	public String getPurchPolicy() { return m_PurchPolicy; }

	/**
	 * obtain setting for using VFQ to pay for IB
	 * @return true if owner permits this
	 */
	public boolean isVFQUsed() { return m_UseVFQ; }

	/**
	 * obtain the number of bettors currently betting
	 * @return the total number of accounts actively betting at least 1 tranche
	 */
	public int getBettorCount() { return m_BettorCount; }

	/**
	 * obtain the scheduled time of the next toss
	 * @return the date/time, as a Unix timestamp in UTC
	 */
	public long getNextToss() { return m_NextToss; }

	/**
	 * obtain the number of seconds until the next toss
	 * @return the seconds which should be reflected in the countdown timer
	 */
	public int getTossSeconds() { return m_TossSeconds; }

	/**
	 * obtain the rate at which tosses are scheduled to occur
	 * @return the number of tosses taking place per hour
	 */
	public byte getTossRate() { return m_TossRate; }

	/**
	 * obtain the current backing ratio of IB at the Issuer
	 * @return the ratio of BTC held to IB issued
	 */
	public double getIBRatio() { return m_IBRatio; }

	/**
	 * obtain the current ratio of VFQ to IB at the Issuer
	 * @return the ratio of VFQ held to IB issued
	 */
	public double getVFQRatio() { return m_VFQRatio; }

	/**
	 * obtain the open orders to be returned
	 * @return the list
	 */
	public ArrayList<OpenOrder> getOpenOrders() { return m_OpenOrders; }

	/**
	 * obtain the available discount policies 
	 * @return the list
	 */
	public ArrayList<DiscountPolicy> getDiscountPolicies() { return m_Policies; }

	/**
	 * obtain the list of account history values
	 * @return the list (may be empty)
	 */
	public ArrayList<AccountHistory> getAccountHistory() {
		return m_AccountHistory;
	}

	/**
	 * obtain the list of tranche history values for a particular tranche
	 * @return the list (may be empty)
	 */
	public ArrayList<TrancheHistory> getTrancheHistory() {
		return m_TrancheHistory;
	}

	/**
	 * obtain the list of queued deployments of money to tranches
	 * @return the list (may be empty)
	 */
	public ArrayList<QueuedDeploy> getQueuedDeployments() {
		return m_QueuedDeployments;
	}


	// set accessor methods
	/**
	 * set the bettor ID
	 * @param bId the unique public number assigned to the account
	 */
	public void setBettorId(int bId) {
		if (bId > 0) {
			m_BettorId = bId;
		}
	}

	/**
	 * set the total BTC held by the account
	 * @param btc the amount of BTC, including chips in tranches
	 */
	public void setBTCTotal(double btc) {
		if (btc >= 0.0) {
			m_BTCTotal = btc;
		}
	}

	/**
	 * set the total IB held by the account
	 * @param ib the amount of IB in all tranches
	 */
	public void setIBTotal(double ib) {
		if (ib >= 0.0) {
			m_IBTotal = ib;
		}
	}

	/**
	 * set the total VFQ held by the account
	 * @param vfq the amount of VFQ (always outside of tranches)
	 */
	public void setVFQTotal(double vfq) {
		if (vfq >= 0.0) {
			m_VFQTotal = vfq;
		}
	}

	/**
	 * set the amount of "free" BTC held in the account
	 * @param btc the amount of BTC not deployed to tranches
	 */
	public void setBTCBalance(double btc) {
		if (btc >= 0.0) {
			m_BTCBalance = btc;
		}
	}

	/**
	 * add a tranche to the list
	 * @param tranche the tranche to add (assumed to be completely filled in)
	 */
	public void addTranche(Tranche tranche) {
		if (tranche != null) {
			m_Tranches.add(tranche);
		}
	}

	/**
	 * set flag whether account is actively betting or not
	 * @param betting true if account betting currently
	 */
	public void setActive(boolean betting) {
		m_Active = betting;
	}

	/**
	 * set the initial bankroll transferred from wallet into game account
	 * @param roll the total bankroll (in BTC) with which account was funded
	 */
	public void setBankroll(double roll) {
		if (roll >= 0.0) {
			m_Bankroll = roll;
		}
	}

	/**
	 * set the additions made to the bankroll subsequent to starting betting
	 * @param adds the sum total of all additions made, in BTC
	 */
	public void setAdditions(double adds) {
		if (adds >= 0.0) {
			m_Additions = adds;
		}
	}

	/**
	 * set the subtractions made from the bankroll subsequent to starting betting
	 * @param subs the sum total of all subtractions made, in BTC
	 */
	public void setSubtractions(double subs) {
		if (subs >= 0.0) {
			m_Subtractions = subs;
		}
	}

	/**
	 * set the total amount of BTC chips deployed to tranches
	 * @param chips the amount of chips in all tranches at start of betting
	 */
	public void setBTCChipStart(double chips) {
		if (chips > 0.0) {
			m_BTCChipStart = chips;
		}
	}

	/**
	 * set the tranche size
	 * @param size the size (in BTC) to use per tranche
	 */
	public void setTrancheSize(float size) {
		// NB: in actuality there is a larger minimum here
		if (size > 0.0) {
			m_TrancheSize = size;
		}
	}

	/**
	 * set the first toss on which the account bet
	 * @param toss the toss number in YYMMDDHHmm format
	 */
	public void setFirstToss(int toss) {
		if (toss > 0) {
			m_FirstToss = toss;
		}
	}

	/**
	 * set the last toss on which the account bet
	 * @param toss the toss number in YYMMDDHHmm format
	 */
	public void setLastToss(int toss) {
		if (toss > 0) {
			m_LastToss = toss;
		}
	}

	/**
	 * set the high-water mark for cash-out value
	 * @param value the highest redemption value observed during betting
	 */
	public void setHighValue(double value) {
		if (value > 0.0) {
			m_HighValue = value;
		}
	}

	/**
	 * set the toss number on which the high value occurred
	 * @param toss the toss number in YYMMDDHHmm format
	 */
	public void setHighToss(int toss) {
		if (toss > 0) {
			m_HighToss = toss;
		}
	}

	/**
	 * set the low-water mark for cash-out value
	 * @param value the lowest redemption value observed during betting
	 */
	public void setLowValue(double value) {
		if (value >= 0.0) {
			m_LowValue = value;
		}
	}

	/**
	 * set the toss number on which the low value occurred
	 * @param toss the toss number in YYMMDDHHmm format
	 */
	public void setLowToss(int toss) {
		if (toss > 0) {
			m_LowToss = toss;
		}
	}

	/**
	 * set the number of tosses separating checks for wholeness
	 * @param tosses the count of tosses which separate wholeness checks
	 */
	public void setWholenessTosses(int tosses) {
		if (tosses >= -1) {
			m_WholenessTosses = tosses;
		}
	}

	/**
	 * set the toss on which the account was last checked for wholeness
	 * @param toss the toss number in YYMMDDHHmm format
	 */
	public void setLastWhChkToss(int toss) {
		if (toss >= 0) {
			m_LastWhChkToss = toss;
		}
	}

	/**
	 * set the profit target (upper stop)
	 * @param target the percentage at which betting will automatically stop
	 */
	public void setProfitTarget(float target) {
		// NB: target could conceivably be negative
		m_ProfitTarget = target;
	}

	/**
	 * set the loss limit (lower stop)
	 * @param loss the percentage at which betting will automatically stop
	 */
	public void setStopLoss(float loss) {
		// NB: loss is very likely negative, but could be positive too
		m_StopLoss = loss;
	}

	/**
	 * set the sale discount policy
	 * @param policy the policy specification, in format:
	 * policyID##discount1 (%): tosses to wait :: discount2 (%) : tosses 2 :: ...
	 * @return true if set successfully, false if not well-formed
	 */
	public boolean setSalePolicy(String policy) {
		if (policy != null && policy.indexOf("##") != -1) {
			m_SalePolicy = new String(policy);
			return true;
		}
		return false;
	}

	/**
	 * set the purchase discount policy
	 * @param policy the policy specification, in format:
	 * policyID##discount1 (%): tosses to wait :: discount2 (%) : tosses 2 :: ...
	 * @return true if set successfully, false if not well-formed
	 */
	public boolean setPurchPolicy(String policy) {
		if (policy != null && policy.indexOf("##") != -1) {
			m_PurchPolicy = new String(policy);
			return true;
		}
		return false;
	}

	/**
	 * set the value of the flag indicating whether any VFQ balance should be
	 * "bled in" to purchase IB
	 * @param use true if VFQ should be used, else false
	 */
	public void setUseVFQ(boolean use) { m_UseVFQ = use; }

	/**
	 * set the bettor count
	 * @param bCnt the number of accounts currently betting actively
	 */
	public void setBettorCount(int bCnt) {
		if (bCnt >= 0) {
			m_BettorCount = bCnt;
		}
	}

	/**
	 * set the timestamp for the next next scheduled toss
	 * @param tstamp the time as a Unix timestamp in UTC
	 */
	public void setNextToss(long tstamp) {
		if (tstamp > 0L) {
			m_NextToss = tstamp;
		}
	}

	/**
	 * set the number of seconds until the next toss
	 * @param secs the count of seconds remaining before the next toss
	 */
	public void setTossSeconds(int secs) {
		if (secs >= 0) {
			m_TossSeconds = secs;
		}
	}

	/**
	 * set the rate at which tosses are occurring
	 * @param rate the number of tosses per hour
	 */
	public void setTossRate(byte rate) {
		if (rate > 0) {
			m_TossRate = rate;
		}
	}

	/**
	 * set the current BTC IB ratio
	 * @param ratio the ratio of BTC held in reserve to IB issued
	 */
	public void setIBRatio(double ratio) {
		if (ratio > 0.0) {
			m_IBRatio = ratio;
		}
	}

	/**
	 * set the current VFQ IB ratio
	 * @param ratio the ratio of VFQ held in reserve to IB issued
	 */
	public void setVFQRatio(double ratio) {
		if (ratio >= 0.0) {
			m_VFQRatio = ratio;
		}
	}

	/**
	 * add an open order to the list to be returned
	 * @param ord the order details to be added
	 */
	public void addOpenOrder(OpenOrder ord) {
		if (ord != null) {
			m_OpenOrders.add(ord);
		}
	}

	/**
	 * add a policy to the list to be returned
	 * @param pol the discount policy definition to be added
	 */
	public void addDiscountPolicy(DiscountPolicy pol) {
		if (pol != null) {
			m_Policies.add(pol);
		}
	}

	/**
	 * add an account history entry to the list to be returned
	 * @param hist the history entry
	 */
	public void addAccountHistory(AccountHistory hist) {
		if (hist != null) {
			m_AccountHistory.add(hist);
		}
	}

	/**
	 * add a tranche history entry to the list to be returned
	 * @param hist the history entry
	 */
	public void addTrancheHistory(TrancheHistory hist) {
		if (hist != null) {
			m_TrancheHistory.add(hist);
		}
	}

	/**
	 * add a queued deployment entry to the list to be returned
	 * @param qEntry the queue entry
	 */
	public void addQueuedDeployment(QueuedDeploy qEntry) {
		if (qEntry != null) {
			m_QueuedDeployments.add(qEntry);
		}
	}


	/**
	 * method to parse a decrypted block from the ABC, and set field values
	 * @param svmBlock the decrypted text, including delimiters
	 * @return true on successful parse, false otherwise
	 */
	@Override
	public boolean parseBlock(String svmBlock) {
		if (svmBlock == null || !svmBlock.startsWith("<svmBlock>")) {
			Log.error("Missing or improper ABC block, cannot parse");
			return false;
		}
		Element savElt = EncodingUtils.buildElement(svmBlock);
		if (savElt == null) {
			Log.error("<svmBlock/> parse failure");
			return false;
		}

		// <opcode/> must always be present
		Element opElt = savElt.element("opcode");
		if (opElt == null) {
			Log.error("Missing svmBlock opcode");
			return false;
		}
		String opcode = opElt.getTextTrim();
		setOpcode(opcode);

		// try to determine if we're an error (will have errcode and/or errmsg)
		Element eCodeElt = savElt.element("errcode");
		if (eCodeElt != null) {
			setErrCode(Integer.parseInt(eCodeElt.getTextTrim()));
		}
		else if (m_IsError) {
			Log.error("Missing errcode for error " + opcode);
			return false;
		}

		Element eMsgElt = savElt.element("errmsg");
		if (eMsgElt != null) {
			setErrMsg(eMsgElt.getTextTrim());
		}
		else if (m_IsError) {
			Log.error("Missing errmsg for error " + opcode);
			return false;
		}
		if (m_IsError) {
			// done after these fields
			return true;
		}

		// otherwise examine the opcode for further fields
		if (opcode.equalsIgnoreCase("REP_create_acct1")
			 || opcode.equalsIgnoreCase("REP_login_phase1"))
		{
			// need the challenge data
			Element chalElt = savElt.element("challenge");
			if (chalElt == null) {
				Log.error("Missing challenge for " + opcode);
				return false;
			}
			setChallenge(chalElt.getTextTrim());

			// need a plugin class name
			Element plugElt = savElt.element("pluginClass");
			if (plugElt == null) {
				Log.error("Missing pluginClass for " + opcode);
				return false;
			}
			setPluginClass(plugElt.getTextTrim());

			// optionally need a plugin class version (which can be empty)
			Element versElt = savElt.element("pluginVersion");
			if (versElt == null) {
				Log.warning("Missing pluginVersion for " + opcode);
			}
			else {
				setPluginVersion(versElt.getTextTrim());
			}
		}
		else if (opcode.equalsIgnoreCase("REP_acct_created")
					|| opcode.equalsIgnoreCase("REP_logged_in"))
		{
			// need session Id
			Element sessElt = savElt.element("sessionId");
			if (sessElt == null) {
				Log.error("Missing sessionId for " + opcode);
				return false;
			}
			setSessionId(sessElt.getTextTrim());

			// need ABC's signature
			Element sigElt = savElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_logged_out")) {
			// no fields here
		}
		else if (opcode.equalsIgnoreCase("REP_wallet_xfer")) {
			// need entered transId
			Element tIdElt = savElt.element("transId");
			if (tIdElt == null) {
				Log.error("Missing transId for " + opcode);
				return false;
			}
			setTransId(Integer.parseInt(tIdElt.getTextTrim()));

			// need wallet Id (destination or source)
			Element walletElt = savElt.element("walletId");
			if (walletElt == null) {
				Log.error("Missing walletId for " + opcode);
				return false;
			}
			setWalletId(walletElt.getTextTrim());

			// need ABC's signature
			Element sigElt = savElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_xfer_confirmed")) {
			// need echoed transId
			Element tIdElt = savElt.element("transId");
			if (tIdElt == null) {
				Log.error("Missing transId for " + opcode);
				return false;
			}
			setTransId(Integer.parseInt(tIdElt.getTextTrim()));

			// need new balance after transfer
			Element balElt = savElt.element("newBal");
			if (balElt == null) {
				Log.error("Missing newBal for " + opcode);
				return false;
			}
			setNewBal(Double.parseDouble(balElt.getText()));

			// need SVM's signature
			Element sigElt = savElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_balances")) {
			// need bettor ID
			Element bIdElt = savElt.element("bettorId");
			if (bIdElt == null) {
				Log.error("Missing bettorId for " + opcode);
				return false;
			}
			setBettorId(Integer.parseInt(bIdElt.getTextTrim()));

			// need BTC total
			Element btcElt = savElt.element("btcTotal");
			if (btcElt == null) {
				Log.error("Missing btcTotal for " + opcode);
				return false;
			}
			setBTCTotal(Double.parseDouble(btcElt.getText()));

			// need IB total
			Element ibElt = savElt.element("ibTotal");
			if (ibElt == null) {
				Log.error("Missing ibTotal for " + opcode);
				return false;
			}
			setIBTotal(Double.parseDouble(ibElt.getText()));

			// need VFQ total
			Element vfqElt = savElt.element("vfqTotal");
			if (vfqElt == null) {
				Log.error("Missing vfqTotal for " + opcode);
				return false;
			}
			setVFQTotal(Double.parseDouble(vfqElt.getText()));

			// need ABC's signature
			Element sigElt = savElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_tranches")) {
			// need BTC balance
			Element btcElt = savElt.element("btcBalance");
			if (btcElt == null) {
				Log.error("Missing btcBalance for " + opcode);
				return false;
			}
			setBTCBalance(Double.parseDouble(btcElt.getText()));

			// get all <tranche/> elements, if user has any
			Iterator trancheIt = savElt.elementIterator("tranche");
			while (trancheIt.hasNext()) {
				Element trancheElt = (Element) trancheIt.next();

				// get the ID attribute
				String tId = trancheElt.attributeValue("id");
				if (tId == null || tId.isEmpty()) {
					Log.error("Missing tranche Id for " + opcode);
					continue;
				}
				Tranche tra = new Tranche(Integer.parseInt(tId.trim()));

				// get the BTC chip balance
				Element balElt = trancheElt.element("btc");
				if (balElt == null) {
					Log.error("Missing BTC chip balance in tranche Id " + tId);
					continue;
				}
				tra.m_BTC = Double.parseDouble(balElt.getText());

				// get the IB balance
				Element ibElt = trancheElt.element("ib");
				if (ibElt == null) {
					Log.error("Missing BTC chip balance in tranche Id " + tId);
					continue;
				}
				tra.m_IB = Double.parseDouble(ibElt.getText());

				// get the starting BTC funding
				Element startElt = trancheElt.element("startBtc");
				if (startElt == null) {
					Log.error("Missing starting BTC balance in tranche Id " + tId);
					continue;
				}
				tra.m_StartBTC = Double.parseDouble(startElt.getText());

				// get the net additions, if any
				Element addsElt = trancheElt.element("additions");
				if (addsElt == null) {
					Log.error("Missing BTC additions in tranche Id " + tId);
					continue;
				}
				tra.m_Additions = Double.parseDouble(addsElt.getText());

				// get the active betting status
				Element actElt = trancheElt.element("active");
				if (actElt == null) {
					Log.error("Missing active status in tranche Id " + tId);
					continue;
				}
				tra.m_Active = Boolean.parseBoolean(actElt.getTextTrim());

				// get the current Martingale betting iteration
				Element itElt = trancheElt.element("iteration");
				if (itElt == null) {
					Log.error("Missing iteration in tranche Id " + tId);
					continue;
				}
				tra.m_Iteration = Byte.parseByte(itElt.getTextTrim());

				// get the current betting increment
				Element betElt = trancheElt.element("betIncrement");
				if (betElt == null) {
					Log.error("Missing betIncrement in tranche Id " + tId);
					continue;
				}
				tra.m_BetIncrement = Double.parseDouble(betElt.getText());

				// get the number of consecutive losses
				Element lossElt = trancheElt.element("losses");
				if (lossElt == null) {
					Log.error("Missing losses in tranche Id " + tId);
					continue;
				}
				tra.m_Losses = Byte.parseByte(lossElt.getTextTrim());

				// get the open order, if any
				Element ordElt = trancheElt.element("openOrder");
				if (ordElt == null) {
					Log.error("Missing openOrder in tranche Id " + tId);
					continue;
				}
				String oId = ordElt.getTextTrim();
				if (!oId.isEmpty()) {
					tra.m_OpenOrder = Integer.parseInt(oId);
				}

				// get the bonus subsidy flag
				Element bonElt = trancheElt.element("bonus");
				if (bonElt == null) {
					Log.error("Missing bonus flag in tranche Id " + tId);
					continue;
				}
				tra.m_Bonus = Boolean.parseBoolean(bonElt.getTextTrim());

				addTranche(tra);
			}

			// need ABC's signature
			Element sigElt = savElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_account_details")) {
			// need bettor ID
			Element bIdElt = savElt.element("bettorId");
			if (bIdElt == null) {
				Log.error("Missing bettorId for " + opcode);
				return false;
			}
			setBettorId(Integer.parseInt(bIdElt.getTextTrim()));

			// need the active betting status
			Element actElt = savElt.element("active");
			if (actElt == null) {
				Log.error("Missing active status for " + opcode);
				return false;
			}
			setActive(Boolean.parseBoolean(actElt.getTextTrim()));

			// need initial bankroll
			Element bankElt = savElt.element("bankroll");
			if (bankElt == null) {
				Log.error("Missing bankroll for " + opcode);
				return false;
			}
			setBankroll(Double.parseDouble(bankElt.getText()));

			// need total net additions to bankroll
			Element addElt = savElt.element("additions");
			if (addElt == null) {
				Log.error("Missing additions for " + opcode);
				return false;
			}
			setAdditions(Double.parseDouble(addElt.getText()));

			// need total net subtractions from bankroll
			Element subElt = savElt.element("subtractions");
			if (subElt == null) {
				Log.error("Missing subtractions for " + opcode);
				return false;
			}
			setSubtractions(Double.parseDouble(subElt.getText()));

			// need starting BTC chip total
			Element btcElt = savElt.element("btcChipStart");
			if (btcElt == null) {
				Log.error("Missing btcChipStart for " + opcode);
				return false;
			}
			setBTCChipStart(Double.parseDouble(btcElt.getText()));

			// need tranche size (in BTC)
			Element szElt = savElt.element("trancheSize");
			if (szElt == null) {
				Log.error("Missing trancheSize for " + opcode);
				return false;
			}
			setTrancheSize(Float.parseFloat(szElt.getText()));

			// need VFQ policy
			Element vfqElt = savElt.element("useVFQ");
			if (vfqElt == null) {
				Log.error("Missing useVFQ for " + opcode);
				return false;
			}
			setUseVFQ(Boolean.parseBoolean(vfqElt.getTextTrim()));

			// need first toss
			Element firstElt = savElt.element("firstToss");
			if (firstElt == null) {
				Log.error("Missing firstToss for " + opcode);
				return false;
			}
			setFirstToss(Integer.parseInt(firstElt.getTextTrim()));
			
			// need last toss
			Element lastElt = savElt.element("lastToss");
			if (lastElt == null) {
				Log.error("Missing lastToss for " + opcode);
				return false;
			}
			setLastToss(Integer.parseInt(lastElt.getTextTrim()));

			// need high value
			Element highElt = savElt.element("highValue");
			if (highElt == null) {
				Log.error("Missing highValue for " + opcode);
				return false;
			}
			setHighValue(Double.parseDouble(highElt.getText()));

			// need high toss
			Element hTossElt = savElt.element("highToss");
			if (hTossElt == null) {
				Log.error("Missing highToss for " + opcode);
				return false;
			}
			setHighToss(Integer.parseInt(hTossElt.getTextTrim()));

			// need low value
			Element lowElt = savElt.element("lowValue");
			if (lowElt == null) {
				Log.error("Missing lowValue for " + opcode);
				return false;
			}
			setLowValue(Double.parseDouble(lowElt.getText()));

			// need low toss
			Element lTossElt = savElt.element("lowToss");
			if (lTossElt == null) {
				Log.error("Missing lowToss for " + opcode);
				return false;
			}
			setLowToss(Integer.parseInt(lTossElt.getTextTrim()));

			// need wholeness tosses
			Element whTossesElt = savElt.element("wholenessTosses");
			if (whTossesElt == null) {
				Log.error("Missing wholenessTosses for " + opcode);
				return false;
			}
			setWholenessTosses(Integer.parseInt(whTossesElt.getTextTrim()));

			// need last wholeness toss, if any
			Element lastWhElt = savElt.element("lastWhChkToss");
			if (lastWhElt == null) {
				Log.error("Missing lastWhChkToss for " + opcode);
				return false;
			}
			setLastWhChkToss(Integer.parseInt(lastWhElt.getTextTrim()));

			// need profit target
			Element profElt = savElt.element("profitTarget");
			if (profElt == null) {
				Log.error("Missing profitTarget for " + opcode);
				return false;
			}
			setProfitTarget(Float.parseFloat(profElt.getText()));

			// need stop loss
			Element stopElt = savElt.element("stopLoss");
			if (stopElt == null) {
				Log.error("Missing stopLoss for " + opcode);
				return false;
			}
			setStopLoss(Float.parseFloat(stopElt.getText()));

			// need sale policy descriptor
			Element saleElt = savElt.element("salePolicy");
			if (saleElt == null) {
				Log.error("Missing salePolicy for " + opcode);
				return false;
			}
			setSalePolicy(saleElt.getTextTrim());

			// need purchase policy descriptor
			Element purchElt = savElt.element("purchPolicy");
			if (purchElt == null) {
				Log.error("Missing purchPolicy for " + opcode);
				return false;
			}
			setPurchPolicy(purchElt.getTextTrim());

			// need list of available discount policies
			// get all <policy/> elements
			Iterator policyIt = savElt.elementIterator("policy");
			while (policyIt.hasNext()) {
				Element policyElt = (Element) policyIt.next();

				// get the ID attribute
				String pId = policyElt.attributeValue("id");
				if (pId == null || pId.isEmpty()) {
					Log.error("Missing policy Id for " + opcode);
					continue;
				}
				DiscountPolicy pol
					= new DiscountPolicy(Short.parseShort(pId.trim()));

				// get the type of the order
				Element typeElt = policyElt.element("type");
				if (typeElt == null) {
					Log.error("Missing type for pId " + pId);
					continue;
				}
				String pTyp = typeElt.getTextTrim();
				if (pTyp.equals(M_PolicyPurch) || pTyp.equals(M_PolicySale)) {
					pol.m_Type = pTyp;
				}
				else {
					Log.error("Illegal policy type for pId " + pId);
					continue;
				}

				// get the descriptive name for the policy
				Element nameElt = policyElt.element("name");
				if (nameElt == null) {
					Log.error("Missing name for pId " + pId);
					continue;
				}
				pol.m_Name = nameElt.getTextTrim();

				// get the policy specification
				Element specElt = policyElt.element("spec");
				if (specElt == null) {
					Log.error("Missing spec for pid " + pId);
					continue;
				}
				pol.m_Spec = specElt.getTextTrim();

				addDiscountPolicy(pol);
			}

			// get all <pendDeploy/> elements
			Iterator deployIt = savElt.elementIterator("pendDeploy");
			while (deployIt.hasNext()) {
				Element deployElt = (Element) deployIt.next();

				// get the toss attribute
				String tossId = deployElt.attributeValue("toss");
				if (tossId == null || tossId.isEmpty()) {
					Log.error("Missing pendDeploy toss for " + opcode);
					continue;
				}
				int toss = Integer.parseInt(tossId.trim());
				QueuedDeploy qEntry = new QueuedDeploy(toss);

				// get the bankroll to be applied
				Element amtElt = deployElt.element("bankroll");
				if (amtElt == null) {
					Log.error("Missing pendDeploy bankroll for " + opcode);
					continue;
				}
				double amt = 0.0;
				try {
					amt = Double.parseDouble(amtElt.getText());
					qEntry.m_Bankroll = amt;
				}
				catch (NumberFormatException nfe) {
					Log.error("Illegal pendDeploy bankroll value, " + amtElt, nfe);
					continue;
				}

				// get the required min. IB ratio
				Element ratioElt = deployElt.element("IBratio");
				if (ratioElt == null) {
					Log.error("Missing pendDeploy IBratio for " + opcode);
					continue;
				}
				float ratio = 0.0F;
				try {
					ratio = Float.parseFloat(ratioElt.getText());
					qEntry.m_IBRatio = ratio;
				}
				catch (NumberFormatException nfe) {
					Log.error("Illegal pendDeploy IBratio value, " + ratioElt, nfe);
					continue;
				}

				addQueuedDeployment(qEntry);
			}

			// need ABC's signature
			Element sigElt = savElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_details_config")) {
			// need tranche size (in BTC)
			Element szElt = savElt.element("trancheSize");
			if (szElt == null) {
				Log.error("Missing trancheSize for " + opcode);
				return false;
			}
			setTrancheSize(Float.parseFloat(szElt.getText()));

			// need profit target
			Element profElt = savElt.element("profitTarget");
			if (profElt == null) {
				Log.error("Missing profitTarget for " + opcode);
				return false;
			}
			setProfitTarget(Float.parseFloat(profElt.getText()));

			// need stop loss
			Element stopElt = savElt.element("stopLoss");
			if (stopElt == null) {
				Log.error("Missing stopLoss for " + opcode);
				return false;
			}
			setStopLoss(Float.parseFloat(stopElt.getText()));

			// need sale policy descriptor
			Element saleElt = savElt.element("salePolicy");
			if (saleElt == null) {
				Log.error("Missing salePolicy for " + opcode);
				return false;
			}
			setSalePolicy(saleElt.getTextTrim());

			// need purchase policy descriptor
			Element purchElt = savElt.element("purchPolicy");
			if (purchElt == null) {
				Log.error("Missing purchPolicy for " + opcode);
				return false;
			}
			setPurchPolicy(purchElt.getTextTrim());

			// need VFQ policy
			Element vfqElt = savElt.element("useVFQ");
			if (vfqElt == null) {
				Log.error("Missing useVFQ for " + opcode);
				return false;
			}
			setUseVFQ(Boolean.parseBoolean(vfqElt.getTextTrim()));

			// need ABC's signature
			Element sigElt = savElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_betting_init")) {
			// need bettor ID
			Element bIdElt = savElt.element("bettorId");
			if (bIdElt == null) {
				Log.error("Missing bettorId for " + opcode);
				return false;
			}
			setBettorId(Integer.parseInt(bIdElt.getTextTrim()));

			// need initial bankroll
			Element bankElt = savElt.element("bankroll");
			if (bankElt == null) {
				Log.error("Missing bankroll for " + opcode);
				return false;
			}
			setBankroll(Double.parseDouble(bankElt.getText()));

			// need ABC's signature
			Element sigElt = savElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_betting_ended")) {
			// need BTC balance
			Element btcElt = savElt.element("btcBalance");
			if (btcElt == null) {
				Log.error("Missing btcBalance for " + opcode);
				return false;
			}
			setBTCBalance(Double.parseDouble(btcElt.getText()));

			// need VFQ total
			Element vfqElt = savElt.element("vfqTotal");
			if (vfqElt == null) {
				Log.error("Missing vfqTotal for " + opcode);
				return false;
			}
			setVFQTotal(Double.parseDouble(vfqElt.getText()));

			// need ABC's signature
			Element sigElt = savElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_tranche_modified")) {
			// need BTC balance
			Element btcElt = savElt.element("btcBalance");
			if (btcElt == null) {
				Log.error("Missing btcBalance for " + opcode);
				return false;
			}
			setBTCBalance(Double.parseDouble(btcElt.getText()));

			// need ABC's signature
			Element sigElt = savElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_stats")) {
			// need bettor count
			Element cntElt = savElt.element("bettorCount");
			if (cntElt == null) {
				Log.error("Missing bettorCount for " + opcode);
				return false;
			}
			setBettorCount(Integer.parseInt(cntElt.getTextTrim()));

			// need next scheduled toss
			Element nextElt = savElt.element("nextToss");
			if (nextElt == null) {
				Log.error("Missing nextToss for " + opcode);
				return false;
			}
			setNextToss(Long.parseLong(nextElt.getTextTrim()));

			// need seconds till next toss
			Element secsElt = savElt.element("tossSeconds");
			if (secsElt == null) {
				Log.error("Missing tossSeconds for " + opcode);
				return false;
			}
			setTossSeconds(Integer.parseInt(secsElt.getTextTrim()));

			// need toss rate (per hour)
			Element rateElt = savElt.element("tossRate");
			if (rateElt == null) {
				Log.error("Missing tossRate for " + opcode);
				return false;
			}
			setTossRate(Byte.parseByte(rateElt.getTextTrim()));

			// need IB backing ratio
			Element ratioElt = savElt.element("IBratio");
			if (ratioElt == null) {
				Log.error("Missing IBratio for " + opcode);
				return false;
			}
			setIBRatio(Double.parseDouble(ratioElt.getText()));

			// need VFQ backing ratio
			Element vratioElt = savElt.element("VFQratio");
			if (vratioElt == null) {
				Log.error("Missing VFQratio for " + opcode);
				return false;
			}
			setVFQRatio(Double.parseDouble(vratioElt.getText()));

			// need number of tosses separating wholeness checks
			Element whTossesElt = savElt.element("wholenessTosses");
			if (whTossesElt == null) {
				Log.error("Missing wholenessTosses for " + opcode);
				return false;
			}
			setWholenessTosses(Integer.parseInt(whTossesElt.getTextTrim()));

			// need ABC's signature
			Element sigElt = savElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_order_book")) {
			// get all <order/> elements, if user has any
			Iterator orderIt = savElt.elementIterator("order");
			while (orderIt.hasNext()) {
				Element orderElt = (Element) orderIt.next();

				// get the ID attribute
				String oId = orderElt.attributeValue("id");
				if (oId == null || oId.isEmpty()) {
					Log.error("Missing order Id for " + opcode);
					continue;
				}
				OpenOrder ord = new OpenOrder(Integer.parseInt(oId.trim()));

				// get the tranche Id whose order it is
				Element trancheElt = orderElt.element("trancheId");
				if (trancheElt == null) {
					Log.error("Missing trancheId for oId " + oId);
					continue;
				}
				ord.m_TrancheId = Integer.parseInt(trancheElt.getTextTrim());

				// get the amount of IB subject to the order
				Element ibElt = orderElt.element("ib");
				if (ibElt == null) {
					Log.error("Missing ib for oId " + oId);
					continue;
				}
				ord.m_IB = Double.parseDouble(ibElt.getText());

				// get the first toss where order was entered
				Element firstElt = orderElt.element("firstToss");
				if (firstElt == null) {
					Log.error("Missing firstToss for oId " + oId);
					continue;
				}
				ord.m_FirstToss = Integer.parseInt(firstElt.getTextTrim());

				// get the age of the order, in tosses
				Element ageElt = orderElt.element("age");
				if (ageElt == null) {
					Log.error("Missing age for oId " + oId);
					continue;
				}
				ord.m_Age = Byte.parseByte(ageElt.getTextTrim());

				// get the applicable sale discount policy
				Element saleElt = orderElt.element("salePolicy");
				if (saleElt == null) {
					Log.error("Missing salePolicy for oId " + oId);
					continue;
				}
				ord.m_SalePolicy = saleElt.getTextTrim();

				// get the applicable buy discount policy
				Element buyElt = orderElt.element("purchPolicy");
				if (buyElt == null) {
					Log.error("Missing purchPolicy for oId " + oId);
					continue;
				}
				ord.m_PurchPolicy = buyElt.getTextTrim();

				addOpenOrder(ord);
			}

			// need ABC's signature
			Element sigElt = savElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_account_hist")) {
			// get all <history/> elements, if any
			Iterator histIt = savElt.elementIterator("history");
			while (histIt.hasNext()) {
				Element histElt = (Element) histIt.next();

				// get the toss attribute
				String tss = histElt.attributeValue("toss");
				if (tss == null || tss.isEmpty()) {
					Log.error("Missing toss for " + opcode);
					continue;
				}
				int toss = Integer.parseInt(tss.trim());
				AccountHistory hist = new AccountHistory(toss);

				// get the free BTC balance on this toss
				Element btcElt = histElt.element("btcBal");
				if (btcElt == null) {
					Log.error("Missing btcBal for toss " + tss);
					continue;
				}
				hist.m_BtcBal = Double.parseDouble(btcElt.getText());

				// get the BTC chip balance on this toss
				Element btcChipElt = histElt.element("btcChipBal");
				if (btcChipElt == null) {
					Log.error("Missing btcChipBal for toss " + tss);
					continue;
				}
				hist.m_BtcChipBal = Double.parseDouble(btcChipElt.getText());

				// get the IB balance on this toss
				Element ibElt = histElt.element("ibBal");
				if (ibElt == null) {
					Log.error("Missing ibBal for toss " + tss);
					continue;
				}
				hist.m_IBBal = Double.parseDouble(ibElt.getText());

				// get the VFQ balance on this toss
				Element vfqElt = histElt.element("vfqBal");
				if (vfqElt == null) {
					Log.error("Missing vfqBal for toss " + tss);
					continue;
				}
				hist.m_VFQBal = Double.parseDouble(vfqElt.getText());

				// get the bankroll additions made during this toss, if any
				Element addsElt = histElt.element("additions");
				if (addsElt == null) {
					Log.error("Missing BTC additions for toss " + tss);
					continue;
				}
				hist.m_Additions = Double.parseDouble(addsElt.getText());

				// get the bankroll subtractions made during this toss, if any
				Element subElt = histElt.element("subtractions");
				if (subElt == null) {
					Log.error("Missing subtractions for toss " + tss);
					continue;
				}
				hist.m_Subtractions = Double.parseDouble(subElt.getText());

				// get the IB ratio which existed on this toss
				Element ratioElt = histElt.element("IBratio");
				if (ratioElt == null) {
					Log.error("Missing IBratio for toss " + tss);
					continue;
				}
				hist.m_IBRatio = Double.parseDouble(ratioElt.getText());

				// get the profit target which was set as of this toss
				Element profElt = histElt.element("profitTarget");
				if (profElt == null) {
					Log.error("Missing profitTarget for toss " + tss);
					continue;
				}
				hist.m_ProfitTarget = Float.parseFloat(profElt.getText());

				// get the stop loss which was set as of this toss
				Element stopElt = histElt.element("stopLoss");
				if (profElt == null) {
					Log.error("Missing stopLoss for toss " + tss);
					continue;
				}
				hist.m_StopLoss = Float.parseFloat(stopElt.getText());

				// get the count of tranches on this toss
				Element tCntElt = histElt.element("trancheCount");
				if (tCntElt == null) {
					Log.error("Missing trancheCount for toss " + tss);
					continue;
				}
				hist.m_TrancheCount = Integer.parseInt(tCntElt.getTextTrim());

				// get the tranche size which was set as of this toss
				Element szElt = histElt.element("trancheSize");
				if (szElt == null) {
					Log.error("Missing trancheSize for toss " + tss);
					continue;
				}
				hist.m_TrancheSize = Float.parseFloat(szElt.getText());

				addAccountHistory(hist);
			}

			// need ABC's signature
			Element sigElt = savElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else if (opcode.equalsIgnoreCase("REP_tranche_hist")) {
			// get all <tranche/> elements, if any (in theory there should be one)
			Iterator trancheIt = savElt.elementIterator("tranche");
			while (trancheIt.hasNext()) {
				Element trancheElt = (Element) trancheIt.next();

				// get the ID attribute
				String tId = trancheElt.attributeValue("id");
				if (tId == null || tId.isEmpty()) {
					Log.error("Missing tranche Id for " + opcode);
					continue;
				}
				int trId = Integer.parseInt(tId.trim());

				// get all <history/> elements, if any
				Iterator histIt = trancheElt.elementIterator("history");
				while (histIt.hasNext()) {
					Element histElt = (Element) histIt.next();

					// get the toss attribute
					String tss = histElt.attributeValue("toss");
					if (tss == null || tss.isEmpty()) {
						Log.error("Missing toss for " + opcode);
						continue;
					}
					int toss = Integer.parseInt(tss.trim());
					TrancheHistory hist = new TrancheHistory(trId, toss);

					// get the free BTC balance on this toss
					Element btcElt = histElt.element("btcBal");
					if (btcElt == null) {
						Log.error("Missing btcBal for tId " + tId + " toss " + tss);
						continue;
					}
					hist.m_BtcBal = Double.parseDouble(btcElt.getText());

					// get the BTC chip balance on this toss
					Element btcChipElt = histElt.element("btcChipBal");
					if (btcChipElt == null) {
						Log.error("Missing btcChipBal for tId " + tId + " toss "
									+ tss);
						continue;
					}
					hist.m_BtcChipBal = Double.parseDouble(btcChipElt.getText());

					// get the IB balance on this toss
					Element ibElt = histElt.element("ibBal");
					if (ibElt == null) {
						Log.error("Missing ibBal for tId " + tId + " toss " + tss);
						continue;
					}
					hist.m_IBBal = Double.parseDouble(ibElt.getText());

					// get the bankroll additions made during this toss, if any
					Element addsElt = histElt.element("additions");
					if (addsElt == null) {
						Log.error("Missing BTC additions for tId " + tId + " toss "
									+ tss);
						continue;
					}
					hist.m_Additions = Double.parseDouble(addsElt.getText());

					// get the IB ratio which existed on this toss
					Element ratioElt = histElt.element("IBratio");
					if (ratioElt == null) {
						Log.error("Missing IBratio for tId " + tId + " toss " + tss);
						continue;
					}
					hist.m_IBRatio = Double.parseDouble(ratioElt.getText());

					// get the bonus status flag as of this toss
					Element bonusElt = histElt.element("bonus");
					if (bonusElt == null) {
						Log.error("Missing bonus for tId " + tId + " toss " + tss);
						continue;
					}
					hist.m_Bonus = Boolean.parseBoolean(bonusElt.getTextTrim());

					// get the active betting flag as of this toss
					Element activeElt = histElt.element("active");
					if (activeElt == null) {
						Log.error("Missing active for tId " + tId + " toss " + tss);
						continue;
					}
					hist.m_Active = Boolean.parseBoolean(activeElt.getTextTrim());

					addTrancheHistory(hist);
				}
			}

			// need ABC's signature
			Element sigElt = savElt.element("signature");
			if (sigElt == null) {
				Log.error("Missing signature for " + opcode);
				return false;
			}
			setSignature(sigElt.getTextTrim());
		}
		else {
			Log.error("Unrecognized reply opcode, " + opcode);
			return false;
		}
		return true;
	}

	/**
	 * method to construct the block XML, based on opcode and member values.
	 * Any signature should be checked using the string returned by this method.
	 * Missing required output fields generate log errors, but do not stop the
	 * generation of output.
	 * @return the complete block for the indicated reply, absent the delimiters
	 * and the ABC's signature (if required)
	 */
	@Override
	public String getXML() {
		StringBuilder xml = new StringBuilder(2048);

		// if we're an error message, this is just the code and error text
		xml.append("<opcode>" + m_Opcode + "</opcode>");
		if (m_IsError) {
			xml.append("<errcode>" + m_ErrCode + "</errcode>");
			xml.append("<errmsg>" + m_ErrMsg + "</errmsg>");
			return xml.toString();
		}

		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(0);

		// base our output on the opcode
		if (m_Opcode.equals("REP_create_acct1")
			 || m_Opcode.equals("REP_login_phase1"))
		{
			if (m_Challenge.isEmpty()) {
				Log.error("Missing challenge for " + m_Opcode);
			}
			xml.append("<challenge>" + m_Challenge + "</challenge>");
			if (m_PluginClass.isEmpty()) {
				Log.error("Missing pluginClass for " + m_Opcode);
			}
			xml.append("<pluginClass>" + m_PluginClass + "</pluginClass>");
			xml.append("<pluginVersion>" + m_PluginVersion + "</pluginVersion>");
		}
		else if (m_Opcode.equals("REP_acct_created")
					|| m_Opcode.equals("REP_logged_in"))
		{
			if (m_SessionId.isEmpty()) {
				Log.error("Missing sessionId for " + m_Opcode);
			}
			xml.append("<sessionId>" + m_SessionId + "</sessionId>");
		}
		else if (m_Opcode.equals("REP_logged_out")) {
			// no special fields for this
		}
		else if (m_Opcode.equals("REP_wallet_xfer")) {
			xml.append("<transId>" + m_TransId + "</transId>");
			xml.append("<walletId>" + m_WalletId + "</walletId>");
		}
		else if (m_Opcode.equals("REP_xfer_confirmed")) {
			xml.append("<transId>" + m_TransId + "</transId>");
			xml.append("<newBal>" + df.format(m_NewBal) + "</newBal>");
		}
		else if (m_Opcode.equals("REP_balances")) {
			xml.append("<bettorId>" + m_BettorId + "</bettorId>");
			xml.append("<btcTotal>" + df.format(m_BTCTotal) + "</btcTotal>");
			xml.append("<ibTotal>" + df.format(m_IBTotal) + "</ibTotal>");
			xml.append("<vfqTotal>" + df.format(m_VFQTotal) + "</vfqTotal>");
		}
		else if (m_Opcode.equals("REP_tranches")) {
			// NB: tranches could be empty, so do not report an error
			xml.append("<btcBalance>" + df.format(m_BTCBalance) + "</btcBalance>");
			for (Tranche tra : m_Tranches) {
				xml.append("<tranche id=\"" + tra.m_TrancheId + "\">");
				xml.append("<btc>" + df.format(tra.m_BTC) + "</btc>");
				xml.append("<ib>" + df.format(tra.m_IB) + "</ib>");
				xml.append("<startBtc>" + df.format(tra.m_StartBTC)
							+ "</startBtc>");
				xml.append("<additions>" + df.format(tra.m_Additions)
							+ "</additions>");
				xml.append("<active>" + Boolean.toString(tra.m_Active)
							+ "</active>");
				xml.append("<iteration>" + tra.m_Iteration + "</iteration>");
				xml.append("<betIncrement>" + df.format(tra.m_BetIncrement)
							+ "</betIncrement>");
				xml.append("<losses>" + tra.m_Losses + "</losses>");
				xml.append("<openOrder>" + tra.m_OpenOrder + "</openOrder>");
				xml.append("<bonus>" + Boolean.toString(tra.m_Bonus) + "</bonus>");
				xml.append("</tranche>");
			}
		}
		else if (m_Opcode.equals("REP_account_details")) {
			xml.append("<bettorId>" + m_BettorId + "</bettorId>");
			xml.append("<active>" + Boolean.toString(m_Active) + "</active>");
			xml.append("<bankroll>" + df.format(m_Bankroll) + "</bankroll>");
			xml.append("<additions>" + df.format(m_Additions) + "</additions>");
			xml.append("<subtractions>" + df.format(m_Subtractions)
						+ "</subtractions>");
			xml.append("<btcChipStart>" + df.format(m_BTCChipStart)
						+ "</btcChipStart>");
			xml.append("<trancheSize>" + df.format(m_TrancheSize)
						+ "</trancheSize>");
			xml.append("<useVFQ>" + Boolean.toString(m_UseVFQ) + "</useVFQ>");
			xml.append("<firstToss>" + m_FirstToss + "</firstToss>");
			xml.append("<lastToss>" + m_LastToss + "</lastToss>");
			xml.append("<highValue>" + df.format(m_HighValue) + "</highValue>");
			xml.append("<highToss>" + m_HighToss + "</highToss>");
			xml.append("<lowValue>" + df.format(m_LowValue) + "</lowValue>");
			xml.append("<lowToss>" + m_LowToss + "</lowToss>");
			xml.append("<wholenessTosses>" + m_WholenessTosses
						+ "</wholenessTosses>");
			xml.append("<lastWhChkToss>" + m_LastWhChkToss + "</lastWhChkToss>");
			df.setMinimumFractionDigits(2);
			xml.append("<profitTarget>" + df.format(m_ProfitTarget)
						+ "</profitTarget>");
			xml.append("<stopLoss>" + df.format(m_StopLoss) + "</stopLoss>");
			xml.append("<salePolicy>" + m_SalePolicy + "</salePolicy>");
			xml.append("<purchPolicy>" + m_PurchPolicy + "</purchPolicy>");
			for (DiscountPolicy dPol : m_Policies) {
				xml.append("<policy id=\"" + dPol.m_PolicyId + "\">");
				xml.append("<type>" + dPol.m_Type + "</type>");
				xml.append("<name>" + dPol.m_Name + "</name>");
				xml.append("<spec>" + dPol.m_Spec + "</spec>");
				xml.append("</policy>");
			}
			for (QueuedDeploy qEntry : m_QueuedDeployments) {
				xml.append("<pendDeploy toss=\"" + qEntry.m_Toss + "\">");
				df.setMinimumFractionDigits(4);
				xml.append("<bankroll>" + df.format(qEntry.m_Bankroll)
							+ "</bankroll>");
				df.setMinimumFractionDigits(2);
				xml.append("<IBratio>" + df.format(qEntry.m_IBRatio)
							+ "</IBratio>");
				xml.append("</pendDeploy>");
			}
		}
		else if (m_Opcode.equals("REP_details_config")) {
			xml.append("<trancheSize>" + df.format(m_TrancheSize)
						+ "</trancheSize>");
			xml.append("<profitTarget>" + df.format(m_ProfitTarget)
						+ "</profitTarget>");
			xml.append("<stopLoss>" + df.format(m_StopLoss) + "</stopLoss>");
			xml.append("<salePolicy>" + m_SalePolicy + "</salePolicy>");
			xml.append("<purchPolicy>" + m_PurchPolicy + "</purchPolicy>");
			xml.append("<useVFQ>" + Boolean.toString(m_UseVFQ) + "</useVFQ>");
		}
		else if (m_Opcode.equals("REP_betting_init")) {
			xml.append("<bettorId>" + m_BettorId + "</bettorId>");
			xml.append("<bankroll>" + df.format(m_Bankroll) + "</bankroll>");
		}
		else if (m_Opcode.equals("REP_betting_ended")) {
			xml.append("<btcBalance>" + df.format(m_BTCBalance) + "</btcBalance>");
			xml.append("<vfqTotal>" + df.format(m_VFQTotal) + "</vfqTotal>");
		}
		else if (m_Opcode.equals("REP_stats")) {
			xml.append("<bettorCount>" + m_BettorCount + "</bettorCount>");
			xml.append("<nextToss>" + m_NextToss + "</nextToss>");
			xml.append("<tossSeconds>" + m_TossSeconds + "</tossSeconds>");
			xml.append("<tossRate>" + m_TossRate + "</tossRate>");
			xml.append("<IBratio>" + df.format(m_IBRatio) + "</IBratio>");
			xml.append("<VFQratio>" + df.format(m_VFQRatio) + "</VFQratio>");
			xml.append("<wholenessTosses>" + m_WholenessTosses
						+ "</wholenessTosses>");
		}
		else if (m_Opcode.equals("REP_tranche_modified")) {
			xml.append("<btcBalance>" + df.format(m_BTCBalance) + "</btcBalance>");
		}
		else if (m_Opcode.equals("REP_order_book")) {
			// NB: open orders could be empty, so do not report an error
			for (OpenOrder oOrd : m_OpenOrders) {
				xml.append("<order id=\"" + oOrd.m_OrderId + "\">");
				xml.append("<trancheId>" + oOrd.m_TrancheId + "</trancheId>");
				xml.append("<ib>" + df.format(oOrd.m_IB) + "</ib>");
				xml.append("<firstToss>" + oOrd.m_FirstToss + "</firstToss>");
				xml.append("<age>" + oOrd.m_Age + "</age>");
				xml.append("<salePolicy>" + oOrd.m_SalePolicy + "</salePolicy>");
				xml.append("<purchPolicy>" + oOrd.m_PurchPolicy + "</purchPolicy>");
				xml.append("</order>");
			}
		}
		else if (m_Opcode.equals("REP_account_hist")) {
			xml.ensureCapacity(256000);
			// NB: history could be empty, so do not report an error
			for (AccountHistory hist : m_AccountHistory) {
				xml.append("<history toss=\"" + hist.m_Toss + "\">");
				xml.append("<btcBal>" + df.format(hist.m_BtcBal) + "</btcBal>");
				xml.append("<btcChipBal>" + df.format(hist.m_BtcChipBal)
							+ "</btcChipBal>");
				xml.append("<ibBal>" + df.format(hist.m_IBBal) + "</ibBal>");
				xml.append("<vfqBal>" + df.format(hist.m_VFQBal) + "</vfqBal>");
				xml.append("<additions>" + df.format(hist.m_Additions)
							+ "</additions>");
				xml.append("<subtractions>" + df.format(hist.m_Subtractions)
							+ "</subtractions>");
				xml.append("<IBratio>" + df.format(hist.m_IBRatio) + "</IBratio>");
				xml.append("<profitTarget>" + df.format(hist.m_ProfitTarget)
							+ "</profitTarget>");
				xml.append("<stopLoss>" + df.format(hist.m_StopLoss)
							+ "</stopLoss>");
				xml.append("<trancheCount>" + hist.m_TrancheCount
							+ "</trancheCount>");
				xml.append("<trancheSize>" + df.format(hist.m_TrancheSize)
							+ "</trancheSize>");
				xml.append("</history>");
			}
		}
		else if (m_Opcode.equals("REP_tranche_hist")) {
			xml.ensureCapacity(256000);
			// NB: history could be empty, so do not report an error
			boolean firstH = true;
			/* NB: in theory, we could return data from more than one tranche here,
			 * 	 given a wildcard trancheId with the request, but this isn't
			 * 	 supported yet and would require a HashMap instead of an
			 * 	 ArrayList, indexed by the tranche ID values.  Presently we
			 * 	 just use the <tranche id=NN> label to echo back the tranche ID.
			 */
			for (TrancheHistory hist : m_TrancheHistory) {
				if (firstH) {
					xml.append("<tranche id=\"" + hist.m_TrancheId + "\">");
					firstH = false;
				}
				xml.append("<history toss=\"" + hist.m_Toss + "\">");
				xml.append("<btcBal>" + df.format(hist.m_BtcBal) + "</btcBal>");
				xml.append("<btcChipBal>" + df.format(hist.m_BtcChipBal)
							+ "</btcChipBal>");
				xml.append("<ibBal>" + df.format(hist.m_IBBal) + "</ibBal>");
				xml.append("<additions>" + df.format(hist.m_Additions)
							+ "</additions>");
				xml.append("<IBratio>" + df.format(hist.m_IBRatio) + "</IBratio>");
				xml.append("<bonus>" + Boolean.toString(hist.m_Bonus) + "</bonus>");
				xml.append("<active>" + Boolean.toString(hist.m_Active)
							+ "</active>");
				xml.append("</history>");
			}
			xml.append("</tranche>");
		}
		else {
			Log.error("Unrecognized reply opcode, " + m_Opcode);
		}

		return xml.toString();
	}

	// END methods
}
