/*
 * last modified---
 * 	02-14-18/kbw/ add <units/> to REQ_wallet_xfer
 * 	10-26-16/jet/ add <IBratio/> to REQ_init_betting
 * 	08-06-16/jet/ add REQ_account_hist and REQ_tranche_hist, add m_UseVFQ
 * 	04-30-15/jet/ Spark plugin version
 * 	01-16-15/jet/ rename from SAVClientBlock
 * 	11-06-14/jet/ new
 *
 * purpose---
 * 	class to instantiate a message from the VSC to the ABC
 *
 * notes---
 * 	maintain in parallel with org.silentvault.market.ClientBlock
 *
 * 	these messages are delivered embedded inside a <msg_blob/> element
 * 	within a REQ_svx_message message routed through the OFS-SVG
 * 	
 * 	see underlying message definitions in vsc_sav.xml
 */

package org.silentvault.spark;

import org.jivesoftware.smack.util.Base64;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.SVMClientBlock;

import java.util.Date;
import java.util.Locale;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.security.PublicKey;
import java.security.PrivateKey;
import java.security.Signature;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidKeyException;
import java.security.SignatureException;


/**
 * This class instantiates client requests sent to the ABC from a connected
 * VSC via the OFS gateway.  It knows how to emit the various message fields
 * based on the contained opcode, and affix client signatures.
 *
 * @author Justin Turrell
 */
public final class ABCClientBlock extends SVMClientBlock {
	// BEGIN data members
	/**
	 * tranche modification type 1: topup (add more BTC)
	 */
	public static final String	M_TrancheModTopup = "topup";
	/**
	 * tranche modification type 2: liquidate (withdraw all value and close)
	 */
	public static final String	M_TrancheModLiq = "liquidate";

	/**
	 * size (in BTC) configured per tranche (used with opcode
	 * REQ_config_details)
	 */
	private Float				m_TrancheSize;

	/**
	 * profit goal (upper stop) set by owner, in percent, or zero if none set
	 * (used with opcode REQ_config_details)
	 */
	private Float				m_ProfitTarget;

	/**
	 * stop loss limit set by owner, in percent, or zero if none set
	 * (used with opcode REQ_config_details)
	 */
	private Float				m_StopLoss;

	/**
	 * owner's selected sale discount policy (string expansion, in this format:
	 * policyID##discount1 (%): tosses to wait :: discount2 (%) : tosses to wait)
	 * (used with opcode REQ_config_details)
	 */
	private String				m_SalePolicy;

	/**
	 * owner's selected purchase discount policy (string expansion, in format:)
	 * policyID##discount1 (%): tosses to wait :: discount2 (%) : tosses to wait)
	 * (used with opcode REQ_config_details)
	 */
	private String				m_PurchPolicy;

	/**
	 * owner's selection whether or not to use any VFQ balance to buy IB when
	 * appropriate, according to the rules (used with opcode REQ_config_details)
	 */
	private Boolean			m_UseVFQ;

	/**
	 * initial bankroll in BTC wallet transferred into account to deploy to
	 * tranches (used with opcode REQ_init_betting)
	 */
	private Double				m_Bankroll;

	/**
	 * required minimum IB ratio in order to proceed with bankroll deployment
	 * (used with REQ_init_betting)
	 */
	private Float				m_IBRatio;

	/**
	 * tranche ID to be modified (used with opcodes REQ_modify_tranche and
	 * REQ_tranche_hist)
	 */
	private Integer			m_TrancheId;

	/**
	 * nature of modification to a tranche: "topup" or "liquidate" (used with
	 * opcode REQ_modify_tranche)
	 */
	private String				m_Mod;

	/**
	 * amount of tranche topup (used with opcode REQ_modify_tranche), can be 0.0
	 */
	private Double				m_TopUp;


	// END data members

	// BEGIN methods
	/**
	 * constructor
	 */
	public ABCClientBlock() {
		super();
		m_SalePolicy = m_PurchPolicy = m_Mod = "";
		m_TrancheSize = new Float(0.0F);
		m_ProfitTarget = new Float(0.0F);
		m_StopLoss = new Float(0.0F);
		m_Bankroll = new Double(0.0);
		m_IBRatio = new Float(0.0F);
		m_TrancheId = new Integer(0);
		m_TopUp = new Double(0.0);
		m_UseVFQ = new Boolean(false);
	}


	// get access methods
	/**
	 * obtain the tranche size configured
	 * @return the quantity (in BTC) to be deployed to tranches
	 */
	public float getTrancheSize() { return m_TrancheSize.floatValue(); }

	/**
	 * obtain the upper stop profit target
	 * @return the target, in percent
	 */
	public float getProfitTarget() { return m_ProfitTarget.floatValue(); }

	/**
	 * obtain the lower stop loss target
	 * @return the stop loss, in percent
	 */
	public float getStopLoss() { return m_StopLoss.floatValue(); }

	/**
	 * obtain the sale discount policy specification
	 * @return the spec, in format:
	 * policyID##discount1 (%): tosses to wait :: discount2 (%) : tosses 2 :: ...
	 */
	public String getSalePolicy() { return m_SalePolicy; }

	/**
	 * obtain the purchase discount policy specification
	 * @return the spec, in format:
	 * policyID##discount1 (%): tosses to wait :: discount2 (%) : tosses 2 :: ...
	 */
	public String getPurchPolicy() { return m_PurchPolicy; }

	/**
	 * obtain the value of the flag indicating whether VFQ balance is "bled in"
	 * in order to purchase IB
	 */
	public boolean isVFQUsed() { return m_UseVFQ.booleanValue(); }

	/**
	 * obtain the bankroll to be deployed to tranches
	 * @return the amount (in BTC) with which account was funded
	 */
	public double getBankroll() { return m_Bankroll.doubleValue(); }

	/**
	 * obtain the minimum required IB ratio to deploy money to tranches
	 * @return the necessary backing ratio, in percent
	 */
	public float getIBRatio() { return m_IBRatio.floatValue(); }

	/**
	 * obtain the tranche ID to be modified
	 * @return the ID of the tranche being reconfigured
	 */
	public int getTrancheId() { return m_TrancheId.intValue(); }

	/**
	 * obtain the tranche modification operation
	 * @return the modification (one of the two constants above)
	 */
	public String getMod() { return m_Mod; }

	/**
	 * obtain the amount of a tranche topup
	 * @return the amount (in BTC) to be added to the indicated tranche
	 */
	public double getTopUp() { return m_TopUp.doubleValue(); }


	// set access methods
	/**
	 * set the tranche size
	 * @param size the size (in BTC) to use per tranche
	 */
	public void setTrancheSize(float size) {
		// NB: in actuality there is a larger minimum here
		if (size > 0.0) {
			m_TrancheSize = new Float(size);
		}
	}

	/**
	 * set the profit target
	 * @param target the percentage profit at which betting will be stopped
	 */
	public void setProfitTarget(float target) {
		m_ProfitTarget = new Float(target);
	}

	/**
	 * set the stop loss target
	 * @param stop the percentage loss at which betting will be stopped
	 */
	public void setStopLoss(float stop) {
		m_StopLoss = new Float(stop);
	}

	/**
	 * set the sale discount policy
	 * @param policy the policy specification, in format:
	 * policyID##discount1 (%): tosses to wait :: discount2 (%) : tosses 2 :: ...
	 * @return true if set successfully, false if not well-formed
	 */
	public boolean setSalePolicy(String policy) {
		if (policy != null && policy.indexOf("##") != -1) {
			m_SalePolicy = new String(policy);
			return true;
		}
		return false;
	}

	/**
	 * set the purchase discount policy
	 * @param policy the policy specification, in format:
	 * policyID##discount1 (%): tosses to wait :: discount2 (%) : tosses 2 :: ...
	 * @return true if set successfully, false if not well-formed
	 */
	public boolean setPurchPolicy(String policy) {
		if (policy != null && policy.indexOf("##") != -1) {
			m_PurchPolicy = new String(policy);
			return true;
		}
		return false;
	}

	/**
	 * set the value of the flag indicating whether any VFQ balance should be
	 * "bled in" to purchase IB
	 * @param use true if VFQ should be used, else false
	 */
	public void setUseVFQ(boolean use) {
		m_UseVFQ = new Boolean(use);
	}

	/**
	 * set the initial bankroll transferred from wallet into game account
	 * @param roll the total bankroll (in BTC) with which account was funded
	 */
	public void setBankroll(double roll) {
		if (roll >= 0.0) {
			m_Bankroll = new Double(roll);
		}
	}

	/**
	 * set the required IB ratio to executed a tranche deployment
	 * @param ratio the minimum ratio, in percent
	 */
	public void setIBRatio(float ratio) {
		if (ratio >= 0.0) {
			m_IBRatio = new Float(ratio);
		}
	}

	/**
	 * set the tranche ID to be modified
	 * @param tId the ID of the tranche being reconfigured
	 */
	public void setTrancheId(int tId) {
		if (tId >= 0) {
			m_TrancheId = new Integer(tId);
		}
	}

	/**
	 * set the tranche modification
	 * @param mod the modification to be done to the tranche
	 */
	public void setMod(String mod) {
		if (mod != null && (mod.equals(M_TrancheModTopup)
								  || mod.equals(M_TrancheModLiq)))
		{
			m_Mod = new String(mod);
		}
		else {
			Log.error("Illegal tranche modification, " + mod);
		}
	}

	/**
	 * set the topup quantity
	 * @param amt the amount of topup being added to the affected tranche
	 */
	public void setTopUp(double amt) {
		// NB: can be zero, which triggers a rebalance
		if (amt >= 0.0) {
			m_TopUp = new Double(amt);
		}
	}


	/**
	 * generate the data which would be encrypted (and possibly signed)
	 */
	@Override
	protected void buildSigData() {
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance(Locale.US);
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(0);

		// build data based on the opcode
		StringBuilder sigData = new StringBuilder(1024);
		sigData.append("<opcode>" + m_Opcode + "</opcode>");
		// anything signed will always have a sessionId
		sigData.append("<sessionId>" + m_SessionId + "</sessionId>");
		if (m_Opcode.equalsIgnoreCase("REQ_config_details")) {
			sigData.append("<trancheSize>" + df.format(m_TrancheSize.doubleValue())
								+ "</trancheSize>");
			df.setMinimumFractionDigits(2);
			sigData.append("<profitTarget>"
								+ df.format(m_ProfitTarget.doubleValue())
								+ "</profitTarget>");
			sigData.append("<stopLoss>" + df.format(m_StopLoss.doubleValue())
								+ "</stopLoss>");
			sigData.append("<salePolicy>" + m_SalePolicy + "</salePolicy>");
			sigData.append("<purchPolicy>" + m_PurchPolicy + "</purchPolicy>");
			sigData.append("<useVFQ>" + Boolean.toString(m_UseVFQ) + "</useVFQ>");
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_init_betting")) {
			sigData.append("<bankroll>" + df.format(m_Bankroll.doubleValue())
								+ "</bankroll>");
			df.setMinimumFractionDigits(2);
			sigData.append("<IBratio>" + df.format(m_IBRatio.doubleValue())
								+ "</IBratio>");
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_end_betting")) {
			// no additional fields here
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_modify_tranche")) {
			sigData.append("<trancheId>" + m_TrancheId + "</trancheId>");
			sigData.append("<mod>" + m_Mod + "</mod>");
			sigData.append("<topUp>" + df.format(m_TopUp.doubleValue())
								+ "</topUp>");
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_wallet_xfer")) {
			sigData.append("<direction>" + m_Direction + "</direction>");
			sigData.append("<seriesId>" + m_SeriesId + "</seriesId>");
			sigData.append("<quantity>" + df.format(m_Quantity.doubleValue()));
			sigData.append("</quantity>");
			sigData.append("<units>" + m_Units + "</units>");
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_confirm_xfer")) {
			sigData.append("<transId>" + m_TransId + "</transId>");
		}
		else if (m_Opcode.equalsIgnoreCase("REQ_tranche_hist")) {
			// this one isn't signed, but does specify a tranche Id
			sigData.append("<trancheId>" + m_TrancheId + "</trancheId>");
		}
		// else: opcode is not signed by client, requires only sessionId
		m_ClientBlockXML = sigData.toString();
	}

	/**
	 * apply the signature to the block, given a private key
	 * @param privKey the private key required to create the signature
	 * @return true if the signature was applied, false otherwise
	 */
	@Override
	public boolean signXML(PrivateKey privKey) {
		if (m_ClientBlockXML.isEmpty()) {
			Log.error("No client block XML to sign");
			return false;
		}
		// these opcodes are unsigned
		if (m_Opcode.equalsIgnoreCase("REQ_create_acct1")
			 || m_Opcode.equalsIgnoreCase("REQ_create_acct2")
			 || m_Opcode.equalsIgnoreCase("REQ_login_phase1")
			 || m_Opcode.equalsIgnoreCase("REQ_login_phase2")
			 || m_Opcode.equalsIgnoreCase("REQ_logout")
			 || m_Opcode.equalsIgnoreCase("REQ_balances")
			 || m_Opcode.equalsIgnoreCase("REQ_tranches")
			 || m_Opcode.equalsIgnoreCase("REQ_account_details")
			 || m_Opcode.equalsIgnoreCase("REQ_stats")
			 || m_Opcode.equalsIgnoreCase("REQ_order_book")
			 || m_Opcode.equalsIgnoreCase("REQ_account_hist")
			 || m_Opcode.equalsIgnoreCase("REQ_tranche_hist"))
		{
			return true;
		}

		Signature vSig = null;
		byte[] sig = null;
		try {
			vSig = Signature.getInstance("SHA1withRSA");
			vSig.initSign(privKey);
			vSig.update(m_ClientBlockXML.getBytes());
			sig = vSig.sign();
		}
		catch (NoSuchAlgorithmException nsae) {
			// impossible
			Log.error("No such sig algorithm", nsae);
		}
		catch (InvalidKeyException ike) {
			Log.error("Invalid pubkey for signature validation", ike);
		}
		catch (SignatureException se) {
			Log.error("Unable to validate client signature", se);
		}
		if (sig == null) {
			return false;
		}
		setSignature(Base64.encodeBytes(sig, Base64.DONT_BREAK_LINES));
		return true;
	}

	// END methods
}
