/*
 * last modified---
 * 	08-15-18/kbw/ explicitly cancel PacketCollectors after use
 * 	08-17-16/jet/ add debug dump for acct hist
 * 	08-06-16/jet/ add REQ_account_hist and REQ_tranche_hist
 * 	07-08-15/jet/ stand-alone ABC plugin version
 * 	04-30-15/jet/ Spark plugin version
 * 	04-14-15/jet/ eliminate popup messages; update countdown timer instead
 * 	04-02-15/jet/ display popup messages via other threads
 * 	02-17-15/jet/ new
 *
 * purpose---
 * 	handle interface to the OFS-SVG plugin which proxies the SVM/ABC
 *
 * notes---
 * 	see underlying messages in vsc_ofs-svm.xml, vsc_svm.xml, and vsc_sav.xml
 */

package org.silentvault.spark;

import javax.swing.SwingUtilities;

import org.jivesoftware.smack.PacketListener;
import org.jivesoftware.smack.PacketCollector;
import org.jivesoftware.smack.util.Base64;
import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.packet.*;
import org.jivesoftware.smack.filter.PacketIDFilter;
import org.jivesoftware.smack.filter.OrFilter;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.ui.svm.*;
import org.silentvault.spark.ui.svm.abc.*;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Random;
import java.security.PublicKey;
import java.net.HttpURLConnection;
import java.text.NumberFormat;
import java.text.DecimalFormat;


/**
 * Handles messaging to/from the ABC, which is proxied by the SVG plugin at
 * the OFS.  The actual namespace is jabber:iq:silent-market#vsc-ofs, but the
 * 'virtual namespace' wrapped by REQ_svm_message / REP_svm_message exchanges
 * is jabber:iq:silent-market#vsc-svm.  This class acts as a listener for
 * requests sent from the ABC via the OFS, for replies to its own requests
 * made to the ABC, and for broadcast messages sent from the ABC.
 *
 * @author Justin Turrell
 */
public class ABCListener extends SVMListener {
	// unique data members

	// methods
	/**
	 * constructor
	 * @param plugin the reference to the app of which we are a part
	 */
	public ABCListener(WalletClient plugin) {
		super(plugin);
	}

	// implement the PacketListener interface
	@Override
	public void processPacket(Packet packet) {
		final ATabManager atm = (ATabManager) m_Plugin.getMarketplaceTabManager();

		// check for a Message notify
		if (packet instanceof Message) {
			/* Look for the extension we recognize.  Due to the filtering, *only*
			 * messages with this extension should be sent here.
			 */
			PacketExtension pExt
				= packet.getExtension("vsc_ofs",
											"jabber:message:silent-market#vsc-ofs");
			if (pExt == null || !(pExt instanceof SVGSVMMessage)) {
				Log.error("Message from ABC without valid extension");
				return;
			}
			// the extension will be embodied in a SVGSVMMessage
			final SVGSVMMessage broadCst = (SVGSVMMessage) pExt;
			String opcode = broadCst.getOpcode();
			Log.debug("Got Message, level " + broadCst.getLevel() + ", text = "
						+ broadCst.getText());
			if (opcode.equalsIgnoreCase("NFY_sys_broadcast")) {
				// based on the severity level, display an appropriate popup
				String level = broadCst.getLevel();
				if (level.equalsIgnoreCase("info")) {
					/* check for a toss completed message, and if found handle via
					 * a call to re-init the countdown by requesting latest stats
					 */
					if (broadCst.getText().contains(" has completed toss ")) {
						Runnable getStats = new Runnable() {
							public void run() {
								// wait random 0-3 seconds
								Random rng = new Random();
								int wait = rng.nextInt(4);
								try {
									Thread.sleep(wait
												* javax.management.timer.Timer.ONE_SECOND);
								}
								catch (InterruptedException ie) { /* ignore */ }

								// get new stats (reply will update countdown timer)
								atm.queryStats(false);
							}
						};
						SwingUtilities.invokeLater(getStats);
					}
					else {
						Runnable shInfo = new Runnable() {
							public void run() {
								atm.showInfo(broadCst.getText());
							}
						};
						SwingUtilities.invokeLater(shInfo);
					}
				}
				else if (level.equalsIgnoreCase("warning")) {
					Runnable shWarn = new Runnable() {
						public void run() {
							atm.showWarning(broadCst.getText());
						}
					};
					SwingUtilities.invokeLater(shWarn);
				}
				else {	// "alert"
					Runnable shErr = new Runnable() {
						public void run() {
							// if this is a shutdown, disconnect forcibly from the SVM
							if (broadCst.getText().endsWith("shutting down NOW")) {
								// this does not do normal logout messaging,
								// which might fail
								atm.getHoldingsPane().forceLogout();
							}

							atm.showError(broadCst.getText());
						}
					};
					SwingUtilities.invokeLater(shErr);
				}
			}
			else {
				Log.error("Weird opcode on Message packet extension, " + opcode);
			}
		}
		// check for IQ reply
		else if (packet instanceof IQ) {
			if (!(packet instanceof SVGSVMIQ)) {
				Log.error("Unexpected IQ packet, " + packet.toXML());
				return;
			}
			SVGSVMIQ svmMsg = (SVGSVMIQ) packet;

			// get the ABC's reply blob
			String repBlob = svmMsg.getRepBlob();
			if (repBlob.isEmpty()) {
				// the only time this is okay is if it's an error packet
				if (svmMsg.getType() == IQ.Type.ERROR) {
					// show error to user
					atm.showError(svmMsg.getErrMsg(),
										"code " + svmMsg.getErrCode());
				}
				else {
					Log.error("Empty ABC reply blob, cannot process");
				}
				return;
			}
			ABCBlock svmReply = null;
			String reply = null;
			String opcode = null;

			// attempt decrypt
			VsSecrets secrets = m_Plugin.getLoginSecrets();
			String decBlob = EncodingUtils.getStrFromBase64PubkeyEnc(repBlob,
																			secrets.getPrivKey());
			if (decBlob == null) {
				/* Only way we will not have any encryption is if an error
				 * occurred before the ABC had our key available, or a timeout.
				 */
				if (svmMsg.getType() == IQ.Type.ERROR) {
					// try to parse as error in clear
					svmReply = new ABCBlock(true);
					reply = "<svmBlock>" + repBlob + "</svmBlock>";
					if (!svmReply.parseBlock(reply)) {
						Log.error("Unable to parse ABC error reply block");
						// show error to user
						atm.showError(svmMsg.getErrMsg(),
											"code " + svmMsg.getErrCode());
						return;
					}
					opcode = svmReply.getOpcode();

					// if this is a REQ_ value, then we got a timeout
					if (opcode.startsWith("REQ_")) {
						// show error to user
						atm.showError(svmReply.getErrMsg(),
											"code " + svmReply.getErrCode());
						return;
					}
				}
				else {
					Log.error("Unable to decrypt ABC reply");
					return;
				}
			}
			else {
				// parse what the ABC sent in its block -- might be an error
				svmReply = new ABCBlock(false);
				reply = "<svmBlock>" + decBlob + "</svmBlock>";
				if (!svmReply.parseBlock(reply)) {
					Log.error("Unable to parse ABC reply block");
					atm.showError("ABC encrypted reply message did not parse");
					return;
				}
				opcode = svmReply.getOpcode();
			}

			// errors and certain reply opcodes are not signed
			if (!svmReply.isError()
				 && !(opcode.equalsIgnoreCase("REP_create_acct1")
				 		|| opcode.equalsIgnoreCase("REP_login_phase1")
				 		|| opcode.equalsIgnoreCase("REP_logged_out")))
			{
				// validate ABC's signature
				PublicKey svmKey = secrets.getSVMKey();
				if (svmKey == null) {
					Log.error("No ABC pubkey available for sig check");
					return;
				}
				if (!svmReply.sigVerify(svmKey)) {
					Log.error("ABC signature check failed on " + opcode);
					return;
				}
			}
			//Log.debug("Valid ABC reply seen, opcode " + opcode);

			// evaluate opcode and pass to appropriate ui handler
			// generic SVM opcodes (code copied from SVMListener)
			if (opcode.equalsIgnoreCase("REP_create_acct1")) {
				// pass reply (or error) to LoginPane
				LoginPane lp = atm.getLoginPane();
				lp.recordCreate1(svmReply);
			}
			else if (opcode.equalsIgnoreCase("REP_acct_created")) {
				// pass reply (or error) to LoginPane
				LoginPane lp = atm.getLoginPane();
				lp.recordCreate2(svmReply);
			}
			else if (opcode.equalsIgnoreCase("REP_login_phase1")) {
				// pass reply (or error) to LoginPane
				LoginPane lp = atm.getLoginPane();
				lp.recordLogin1(svmReply);
			}
			else if (opcode.equalsIgnoreCase("REP_logged_in")) {
				// pass reply (or error) to LoginPane
				LoginPane lp = atm.getLoginPane();
				lp.recordLogin2(svmReply);
			}
			else if (opcode.equalsIgnoreCase("REP_logged_out")) {
				// pass reply (or error) to LoginPane
				LoginPane lp = atm.getLoginPane();
				lp.recordLogout(svmReply);
			}
			else if (opcode.equalsIgnoreCase("REP_wallet_xfer")) {
				// pass data to TransfersPane (including errors)
				TransfersPane tp = atm.getTransfersPane();
				tp.recordXfer(svmReply);
			}
			else if (opcode.equalsIgnoreCase("REP_xfer_confirmed")) {
				// pass reply to TransfersPane (including errors)
				TransfersPane tp = atm.getTransfersPane();
				tp.recordXconf(svmReply);
			}
			// ABC-specific opcodes
			else if (opcode.equalsIgnoreCase("REP_balances")) {
				HoldingsPane hp = atm.getHoldingsPane();
				if (svmReply.isError()) {
					hp.setCursor(null);
					atm.showError("unable to obtain balances",
										svmReply.getErrMsg());
				}
				else {
					// pass data to HoldingsPane
					hp.recordBalances(svmReply);
				}
			}
			else if (opcode.equalsIgnoreCase("REP_tranches")) {
				HoldingsPane hp = atm.getHoldingsPane();
				if (svmReply.isError()) {
					hp.setCursor(null);
					atm.showError("unable to obtain tranche list",
										svmReply.getErrMsg());
				}
				else {
					// pass data to HoldingsPane
					hp.recordTranches(svmReply);
				}
			}
			else if (opcode.equalsIgnoreCase("REP_account_details")) {
				ConfigPane cp = atm.getConfigPane();
				if (svmReply.isError()) {
					cp.setCursor(null);
					atm.showError("unable to obtain account details",
										svmReply.getErrMsg());
				}
				else {
					// pass to ConfigPane
					cp.recordDetails(svmReply);
				}
			}
			else if (opcode.equalsIgnoreCase("REP_details_config")) {
				ConfigPane cp = atm.getConfigPane();
				if (svmReply.isError()) {
					cp.setCursor(null);
					atm.showError("unable to modify account details",
										svmReply.getErrMsg());
				}
				else {
					// pass to ConfigPane
					cp.recordChanges(svmReply);
				}
			}
			else if (opcode.equalsIgnoreCase("REP_betting_init")) {
				PlayPane pp = atm.getPlayPane();
				if (svmReply.isError()) {
					pp.setCursor(null);
					pp.resetButtons();
					atm.showError("unable to initialize betting",
										svmReply.getErrMsg());
				}
				else {
					// pass to PlayPane
					pp.recordStart(svmReply);
				}
			}
			else if (opcode.equalsIgnoreCase("REP_betting_ended")) {
				PlayPane pp = atm.getPlayPane();
				if (svmReply.isError()) {
					pp.setCursor(null);
					pp.resetButtons();
					atm.showError("unable to cease betting",
										svmReply.getErrMsg());
				}
				else {
					// pass to PlayPane
					pp.recordStop(svmReply);
				}
			}
			else if (opcode.equalsIgnoreCase("REP_stats")) {
				StatsPane sp = atm.getStatsPane();
				if (svmReply.isError()) {
					sp.setCursor(null);
					atm.showError("unable to obtain game stats",
										svmReply.getErrMsg());
				}
				else {
					// pass to StatsPane
					sp.recordStats(svmReply);
				}
			}
			else if (opcode.equalsIgnoreCase("REP_tranche_modified")) {
				HoldingsPane hp = atm.getHoldingsPane();
				if (svmReply.isError()) {
					hp.setCursor(null);
					atm.showError("unable to modify tranche",
										svmReply.getErrMsg());
				}
				else {
					// pass data to HoldingsPane
					hp.recordTrancheMod(svmReply);
				}
			}
			else if (opcode.equalsIgnoreCase("REP_order_book")) {
				MarketPane mp = atm.getMarketPane();
				if (svmReply.isError()) {
					mp.setCursor(null);
					atm.showError("unable to obtain order list",
										svmReply.getErrMsg());
				}
				else {
					// pass data to MarketPane
					mp.recordOrders(svmReply);
				}
			}
			else if (opcode.equalsIgnoreCase("REP_account_hist")) {
				HistoryPane hp = atm.getHistoryPane();
				if (svmReply.isError()) {
					hp.setCursor(null);
					atm.showError("unable to obtain account history",
				 						svmReply.getErrMsg());
				}
				else {
					// pass data to HistoryPane
					hp.recordAcctHistory(svmReply);
				}
			}
			else if (opcode.equalsIgnoreCase("REP_tranche_hist")) {
				HistoryPane hp = atm.getHistoryPane();
				if (svmReply.isError()) {
				 	hp.setCursor(null);
				 	atm.showError("unable to obtain account history",
				 						svmReply.getErrMsg());
				}
				else {
					// pass data to HistPane
					hp.recordTrancheHistory(svmReply);
				}
			}
			else {
				Log.error("Unrecognized ABC reply opcode, " + opcode);
			}
		}
		else {
			// must be Presence, not supported in this namespace
			Log.error("Unexpected (Presence?) packet: " + packet.toXML());
		}

		// retire this ID from the list of active packet filters
		removeFilter(packet.getPacketID());
	}

	/**
	 * Prepare and send a request message to the ABC (via OFS), given a client
	 * block.  Overrides the method in SVMListener, which only deals with 
	 * generic SVMClientBlocks.
	 * NB: this method blocks on the reply from the OFS/ABC.
	 * @param cBlock the client block to send
	 * @return true if message sent successfully
	 */
	public boolean sendRequest(ABCClientBlock cBlock) {
		if (cBlock == null || !cBlock.getOpcode().startsWith("REQ_")) {
			Log.error("Missing or improper ABC client block");
			return false;
		}

		// build output XML and sign it (if required)
		cBlock.buildSigData();
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		if (!cBlock.signXML(secrets.getPrivKey())) {
			Log.error("Unable to sign output XML");
			return false;
		}

		// encrypt block, including any signature, with SVM's pubkey
		StringBuilder clearBlock = new StringBuilder(5120);
		clearBlock.append("<clientBlock>");
		clearBlock.append(cBlock.getClientBlockXML());
		if (!cBlock.getSignature().isEmpty()) {
			clearBlock.append("<signature>" + cBlock.getSignature());
			clearBlock.append("</signature>");
		}
		clearBlock.append("</clientBlock>");
		String encBlock
			= EncodingUtils.makeBase64PubkeyEncStr(clearBlock.toString(),
																secrets.getSVMKey());
		if (encBlock == null || encBlock.isEmpty()) {
			Log.error("Error encrypting client block for ABC");
			return false;
		}

		// prepare wrapper message to OFS
		XMPPConnection conn = m_Plugin.getOFSConnection();
		VscState stData = m_Plugin.getVscState();
		String opcode = cBlock.getOpcode();
		SVGSVMIQ svmReq = new SVGSVMIQ();
		if (opcode.equals("REQ_balances")
			 || opcode.equals("REQ_tranches")
			 || opcode.equals("REQ_account_details")
			 || opcode.equals("REQ_stats")
			 || opcode.equals("REQ_order_book")
			 || opcode.equals("REQ_logout")
			 || opcode.equals("REQ_account_hist")
			 || opcode.equals("REQ_tranche_hist"))
		{
			svmReq.setType(IQ.Type.GET);
		}
		else {
			svmReq.setType(IQ.Type.SET);
		}
		svmReq.setFrom(conn.getUser());
		svmReq.setTo(conn.getServiceName());
		String pId = "svm_" + stData.getNextOFSid();
		svmReq.setPacketID(pId);
		svmReq.setOpcode("REQ_svm_message");
		svmReq.setSVMId(secrets.getSVMId());
		svmReq.setMsgBlob(encBlock);

		// construct a packet collector to listen for reply
		PacketIDFilter pFilter = new PacketIDFilter(pId);
		PacketCollector repColl = conn.createPacketCollector(pFilter);

		// send it to the server
		//Log.debug("Sending " + opcode + " to SVG, " + clearBlock.toString());
		conn.sendPacket(svmReq);
		stData.setLastActivity(svmReq);

		// wait for reply
		SVGSVMIQ rep = null;
		IQ reply = (IQ) repColl.nextResult(m_Plugin.getTimeout());
		if (reply == null) {
			// got timeout; build error
			rep = new SVGSVMIQ();
			rep.setOpcode("REP_svm_message");
			rep.setFrom(svmReq.getTo());
			rep.setTo(svmReq.getFrom());
			rep.setPacketID(pId);
			rep.setType(IQ.Type.ERROR);
			SVMBlock fakeRep = new SVMBlock(true);
			fakeRep.setErrCode(HttpURLConnection.HTTP_GATEWAY_TIMEOUT);
			fakeRep.setErrMsg("request timed out");
			fakeRep.setOpcode(opcode);
			rep.setRepBlob(fakeRep.getXML());
		}
		else {
			rep = (SVGSVMIQ) reply;
		}
		repColl.cancel();
		this.processPacket(rep);
		return true;
	}

	/**
	 * Prepare and send a request message to the ABC (via OFS), given a client
	 * block.
	 * NB: this method does not block on the reply from the OFS/ABC
	 * @param cBlock the client block to send
	 * @return true if message sent successfully
	 */
	public boolean sendAsyncRequest(ABCClientBlock cBlock) {
		if (cBlock == null || !cBlock.getOpcode().startsWith("REQ_")) {
			Log.error("Missing or improper ABC client block");
			return false;
		}

		// build output XML and sign it (if required)
		cBlock.buildSigData();
		VsSecrets secrets = m_Plugin.getLoginSecrets();
		if (!cBlock.signXML(secrets.getPrivKey())) {
			Log.error("Unable to sign output XML");
			return false;
		}

		// encrypt block, including any signature, with SVM's pubkey
		StringBuilder clearBlock = new StringBuilder(5120);
		clearBlock.append("<clientBlock>");
		clearBlock.append(cBlock.getClientBlockXML());
		if (!cBlock.getSignature().isEmpty()) {
			clearBlock.append("<signature>" + cBlock.getSignature());
			clearBlock.append("</signature>");
		}
		clearBlock.append("</clientBlock>");
		String encBlock
			= EncodingUtils.makeBase64PubkeyEncStr(clearBlock.toString(),
																secrets.getSVMKey());
		if (encBlock == null || encBlock.isEmpty()) {
			Log.error("Error encrypting client block for ABC");
			return false;
		}

		// prepare wrapper message to OFS
		XMPPConnection conn = m_Plugin.getOFSConnection();
		VscState stData = m_Plugin.getVscState();
		String opcode = cBlock.getOpcode();
		SVGSVMIQ svmReq = new SVGSVMIQ();
		if (opcode.equals("REQ_balances")
			 || opcode.equals("REQ_tranches")
			 || opcode.equals("REQ_account_details")
			 || opcode.equals("REQ_stats")
			 || opcode.equals("REQ_order_book")
			 || opcode.equals("REQ_logout")
			 || opcode.equals("REQ_account_hist")
			 || opcode.equals("REQ_tranche_hist"))
		{
			svmReq.setType(IQ.Type.GET);
		}
		else {
			svmReq.setType(IQ.Type.SET);
		}
		svmReq.setFrom(conn.getUser());
		svmReq.setTo(conn.getServiceName());
		String pId = "svm_" + stData.getNextOFSid();
		svmReq.setPacketID(pId);
		svmReq.setOpcode("REQ_svm_message");
		svmReq.setSVMId(secrets.getSVMId());
		svmReq.setMsgBlob(encBlock);

		// construct a packet filter to listen for reply
		PacketIDFilter pFilter = new PacketIDFilter(pId);
		addFilter(pId, pFilter);

		// send it to the server
	/*
		Log.debug("Sending non-blocking " + opcode + " to SVG, "
					+ clearBlock.toString());
	 */
		conn.sendPacket(svmReq);
		stData.setLastActivity(svmReq);
		return true;
	}
}
