/*
 * last modified---
 * 	07-09-15/jet/ make an independent Spark plugin
 *
 * purpose---
 * 	class describing plugin support for ABC client-side classes
 */

package org.silentvault.spark;

import org.jivesoftware.spark.plugin.Plugin;
import org.jivesoftware.spark.PluginManager;
import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.ui.svm.abc.ATabManager;
import org.silentvault.spark.ui.svm.abc.AHomePane;


/**
 * This class describes the classes which must be supplied client-side in
 * order to extend the functionality of an SVM beyond that provided by the
 * base classes.  The ABC will pass the name of this class (i.e. the string
 * "org.silentvault.client.ABCPlugin" in the <pluginClass/> field of a reply
 * to a login or account creation request.  The handling code must determine
 * that this class exists (by asking Spark for it as a Plugin), and then
 * if found assign it to WalletClient.m_SVMPlugin.
 *
 * @author Justin Turrell
 */
public class ABCPlugin extends SVMPlugin implements Plugin {
	// data members

	// methods
	/**
	 * nullary constructor (required)
	 */
	public ABCPlugin() {
		super(null);
	}

	// methods to implement the Plugin interface
	/**
	 * called after Spark is loaded to initialize the plugin
	 */
	public void initialize() {
		// turn on Log.debug() output
		//System.setProperty("debug.mode", "true");

		// obtain our "parent" plugin, the SV wallet client (which must exist)
		PluginManager plugMgr = PluginManager.getInstance();
		ClassLoader plugLoader = plugMgr.getPluginClassLoader();
		Class plugClass = null;
		try {
			plugClass = plugLoader.loadClass("org.silentvault.spark.WalletClient");
		}
		catch (ClassNotFoundException cnfe) {
			Log.error("Could not find WalletClient plugin class, abort ABC init",
						cnfe);
			return;
		}

		// using our parent, init the rest of the supporting roles
		Plugin parentClass = plugMgr.getPlugin(plugClass);
		if (parentClass == null) {
			Log.error("Could not find WalletClient plugin class, abort ABC init");
			return;
		}
		if (parentClass instanceof WalletClient) {
			WalletClient wc = (WalletClient) parentClass;
			setPlugin(wc);
		}
		else {
			Log.error("Plugin was not of type WalletClient, abort ABC init");
			return;
		}
		// specify the other classes we wish to use for messaging
		setClientBlock("org.silentvault.spark.ABCClientBlock");
		setReplyBlock("org.silentvault.spark.ABCBlock");
		setSVMListener(new ABCListener(m_Plugin));
		ATabManager abcTab
			= new org.silentvault.spark.ui.svm.abc.ATabManager(m_Plugin);
		setTabManager(abcTab);
		setHomePane(
			new org.silentvault.spark.ui.svm.abc.AHomePane(m_Plugin, abcTab));
		//Log.debug("ABC plugin initialized");
	}

	/**
	 * called when Spark is shutting down
	 */
	public void shutdown() {
		m_SVMTabManager.deallocatePanes();
	}

	/**
	 * return true if Spark can shutdown on user's request
	 * @return true always
	 */
	public boolean canShutDown() { return true; }

	/**
	 * called when a user explicitly asked to uninstall this plugin
	 */
	public void uninstall() {
	}

	// end methods
}
