/*
 * last modified---
 * 	08-25-16/jet/ add m_BTCAdditions
 * 	04-30-15/jet/ Spark plugin version
 * 	01-23-15/jet/ new
 *
 * purpose---
 * 	provide a class representation of a tranche stored in the ABC game
 */

package org.silentvault.spark;


/**
 * This class supplies an instantiation of a tranche held in a wallet's ABC
 * account as a Java class.  It has methods to get and set all fields, along
 * with some minimal sanity checking logic.
 *
 * @author Justin Turrell
 */
public final class ABCTranche {
	// BEGIN data members
	/**
	 * the ID of the tranche
	 */
	private int							m_TrancheId;

	/**
	 * current balance of BTC chips
	 */
	private double						m_BTCChipBalance;

	/**
	 * current balance in InsurBucks
	 */
	private double						m_IBBalance;

	/**
	 * original balance in BTC when tranche was initialized
	 */
	private double						m_BTCStart;

	/**
	 * additions to the starting balance, if any (in BTC)
	 */
	private double						m_BTCAdditions;

	/**
	 * flag whether tranche is actively betting
	 */
	private boolean					m_Active;

	/**
	 * current Martingale iteration (where tranche is in cycle)
	 */
	private byte						m_Iteration;

	/**
	 * current Martingale betting increment (in BTC chips)
	 */
	private double						m_BetIncrement;

	/**
	 * current count of consecutive losses within Martingale sequence
	 */
	private byte						m_LossCount;

	/**
	 * open order ID (index into of_sav_orders), 0 if none !0 implies not betting
	 */
	private int							m_OrderId;

	/**
	 * flag indicating this tranch is currently receiving extra IB bonusing
	 */
	private boolean					m_Subsidized;

	// END data members

	// BEGIN methods
	/**
	 * nullary constructor
	 */
	public ABCTranche() {
	}

	// get methods
	/**
	 * obtain the tranche ID
	 * @return the ID of the tranche within the account
	 */
	public int getTrancheId() { return m_TrancheId; }

	/**
	 * obtain the current balance in BTC chips
	 * @return the quantity of BTC chips held by this tranche
	 */
	public double getBTCChipBalance() { return m_BTCChipBalance; }

	/**
	 * obtain the current balance in IB
	 * @return the quantity of InsurBucks held by this tranche
	 */
	public double getIBBalance() { return m_IBBalance; }

	/**
	 * obtain the starting BTC (at init)
	 * @return the amount of BTC with which the tranche was initialized
	 */
	public double getBTCStart() { return m_BTCStart; }

	/**
	 * obtain the additions to starting BTC since init
	 * @return the amount of BTC which has been added to the tranche, if any
	 */
	public double getBTCAdditions() { return m_BTCAdditions; }

	/**
	 * determine whether tranche is actively betting
	 * @return true if betting
	 */
	public boolean isActive() { return m_Active; }

	/**
	 * obtain the current Martingale iteration in betting sequence
	 * @return the bet iteration, from 1 - Max
	 */
	public byte getIteration() { return m_Iteration; }

	/**
	 * obtain the current Martingale betting increment
	 * @return the base amount in BTC chips for the Martingale cycle
	 */
	public double getBetIncrement() { return m_BetIncrement; }

	/**
	 * obtain the count of consecutive losses within Martingale cycle
	 * @return the number of consecutive losses the tranche has experienced
	 */
	public byte getLossCount() { return m_LossCount; }

	/**
	 * obtain the ID of any open order the tranche might have
	 * @return the order ID, 0 if none, non-zero implies not currently betting
	 */
	public int getOrderId() { return m_OrderId; }

	/**
	 * determine whether the tranche is presently subsidized
	 * @return true if tranche receiving extra bonus IB on losing bets
	 */
	public boolean isSubsidized() { return m_Subsidized; }


	// set methods
	/**
	 * set the tranche ID (this is an unique index with m_AcctId)
	 * @param trancheId the tranche ID to set
	 */
	public void setTrancheId(int trancheId) {
		if (trancheId > 0) {
			// this is an immutable value which should be set only once
			if (m_TrancheId == 0) {
				m_TrancheId = trancheId;
			}
		}
	}

	/**
	 * set the current balance of BTC chips
	 * @param chips the amount of BTC chips held
	 */
	public void setBTCChipBalance(double chips) {
		m_BTCChipBalance = chips;
	}

	/**
	 * set the current balance of InsurBucks
	 * @param ib the amount of IB held
	 */
	public void setIBBalance(double ib) {
		m_IBBalance = ib;
	}

	/**
	 * set the starting BTC balance
	 * @param btc the amount of starting BTC when tranche was initialized
	 */
	public void setBTCStart(double btc) {
		if (btc > 0.0) {
			m_BTCStart = btc;
		}
	}

	/**
	 * set the additions to the starting BTC balance
	 * @param btc the amount of additions BTC since tranche was initialized
	 */
	public void setBTCAdditions(double btc) {
		if (btc > 0.0) {
			m_BTCAdditions = btc;
		}
	}

	/**
	 * set whether or not the tranche is actively betting
	 * @param active true if betting, false if not
	 */
	public void setActive(boolean active) { m_Active = active; }

	/**
	 * set the Martingale iteration
	 * @param iter the current place within the betting sequence
	 */
	public void setIteration(byte iter) {
		if (iter >= 0) {
			m_Iteration = iter;
		}
	}

	/**
	 * set the Martingale base betting increment
	 * @param incr the Martingale betting increment, in BTC chips
	 */
	public void setBetIncrement(double incr) {
		if (incr >= 0.0) {
			m_BetIncrement = incr;
		}
	}

	/**
	 * set the loss count
	 * @param lcnt the count of consecutive losses within a Martingale sequence
	 */
	public void setLossCount(byte lcnt) {
		if (lcnt >= 0) {
			m_LossCount = lcnt;
		}
	}

	/**
	 * set ID of any open order
	 * @param oId the ID of the currently open order, or 0 if none
	 */
	public void setOrderId(int oId) {
		if (oId >= 0) {
			m_OrderId = oId;
		}
	}

	/**
	 * set flag for subsidy bonusing
	 * @param sub true if subsidized, false if not
	 */
	public void setSubsidized(boolean sub) { m_Subsidized = sub; }

	// END methods
}
