/*
 * last modified---
 * 	04-30-15/jet/ Spark plugin version
 * 	02-09-15/jet/ add policy parsing methods
 * 	02-04-15/jet/ new, based on ABCBlock.DiscountPolicy
 *
 * purpose---
 * 	provide a class representation of a sale or purchase discount policy
 */

package org.silentvault.spark;

import org.jivesoftware.spark.util.log.Log;


/**
 * This class supplies an instantiation of a sale or purchase discount policy
 * It has methods to get and set all fields, along with some minimal sanity
 * checking logic.  It also contains static policy parsing methods.
 *
 * @author Justin Turrell
 */
public final class DiscountPolicy {
	// BEGIN data members
	/**
	 * policy ID (unique index)
	 */
	private short			m_PolicyId;

	/**
	 * type of policy (sale or purchase)
	 */
	private String			m_Type;

	/**
	 * descriptive name for this policy
	 */
	private String			m_Name;

	/**
	 * specification of discount strategy, formatted like this:
	 * discount1 (%): tosses to wait :: discount2 (%) : tosses to wait :: etc.
	 */
	private String			m_Specification;

	// END data members

	// BEGIN methods
	/**
	 * nullary constructor
	 */
	public DiscountPolicy() {
		m_Name = m_Specification = m_Type = "";
	}

	// get methods
	/**
	 * obtain the policy code
	 * @return the policy ID (unique index)
	 */
	public short getPolicyId() { return m_PolicyId; }

	/**
	 * obtain the policy type
	 * @return the type (sale or purchase)
	 */
	public String getType() { return m_Type; }

	/**
	 * obtain the textual policy name
	 * @return the descriptive name by which the policy is known
	 */
	public String getName() { return m_Name; }

	/**
	 * obtain the specification of the discount strategy
	 * @return the formatted discount specification
	 */
	public String getSpecification() { return m_Specification; }


	// set methods
	/**
	 * set the policy code
	 * @param policy the policy code
	 */
	public void setPolicyId(short policy) {
		if (policy >= 0) {
			// this is an immutable value which should be set only once
			if (m_PolicyId == 0) {
				m_PolicyId = policy;
			}
		}
	}

	/**
	 * set the policy type
	 * @param type the type (sale or purchase)
	 */
	public void setType(String type) {
		if (type != null && (type.equals(ABCBlock.M_PolicySale)
									|| type.equals(ABCBlock.M_PolicyPurch)))
		{
			m_Type = new String(type);
		}
	}

	/**
	 * set the policy name
	 * @param name the long descriptive name of the policy
	 */
	public void setName(String name) {
		if (name != null && !name.isEmpty()) {
			m_Name = new String(name);
		}
	}

	/**
	 * set the policy specification
	 * @param spec the formatted specification of the policy definition
	 */
	public void setSpecification(String spec) {
		if (spec != null && !spec.isEmpty()) {
			// this is an immutable value which should be set only once
			if (m_Specification.isEmpty()) {
				m_Specification = new String(spec);
			}
		}
	}

	/**
	 * create a discount policy from constituent parts
	 * @param pId the policy Id
	 * @param type the type (sale or purchase)
	 * @param name the descriptive name
	 * @param spec the detailed specification
	 * @return the policy constructed, or null on illegal input
	 */
	public static DiscountPolicy genPolicy(short pId,
														String type,
														String name,
														String spec)
	{
		if (pId < 0 || type == null || name == null || spec == null) {
			return null;
		}
		if (!(type.equals(ABCBlock.M_PolicyPurch)
			   || type.equals(ABCBlock.M_PolicySale)))
		{
			return null;
		}
		if (name.isEmpty()) {
			return null;
		}
		DiscountPolicy pol = new DiscountPolicy();
		pol.setPolicyId(pId);
		pol.setType(type);
		pol.setName(name);
		pol.setSpecification(spec);
		return pol;
	}

	/**
	 * method to compute the discount on a sale policy, at a certain age
	 * @param pSpec the policy specification
	 * @param age the age in tosses
	 * @param reserve the current IB reserve ratio
	 * @return the discount price applicable at that toss for the policy.  This
	 * is a fraction of the difference between the current redemption rate and
	 * 1.0.  For example a 10% discount with 80% reserve ratio is 10% off of a
	 * 20% difference, hence a discount price of 98%.
	 */
	public static double parseSellPolicy(String pSpec, byte age, double reserve)
	{
		if (pSpec == null || age < 0 || reserve < 0.0) {
			Log.error("parseSellPolicy(): bad input");
			return 0.0;
		}
		double discount = 1.0;
		double redRate = reserve;

		// an empty spec means any discount, hence sell for reserve ratio
		if (pSpec.isEmpty()) {
			return redRate;
		}
		double dRate = 1.0 - redRate;

		// parse into N discount stages
		String[] stages = pSpec.split("::");
		byte usedUp = 0;
		for (int stage = 0; stage < stages.length; stage++) {
			// parse the discount, which will be a double, and the tosses, an int
			String[] stgSpec = stages[stage].split(":");
			if (stgSpec == null || stgSpec.length != 2) {
				Log.error("Sale discount policy " + pSpec + " does not parse at "
							+ "stage " + stage);
				continue;
			}
			Double disc = null;
			Byte appliesFor = null;
			try {
				disc = Double.parseDouble(stgSpec[0]);
				appliesFor = Byte.parseByte(stgSpec[1].trim());
			}
			catch (NumberFormatException nfe) {
				Log.error("Sale discount policy " + pSpec + " does not parse at "
							+ "stage " + stage, nfe);
				continue;
			}
			// if this stage applies for the given age, then return the discount
			usedUp += appliesFor.byteValue();
			if (age < usedUp) {
				// offer this percentage of the reserve shortfall
				dRate *= (disc.doubleValue() / 100.0);
				break;
			}
		}
		discount -= dRate;
		return Math.min(1.0, discount);
	}

	/**
	 * method to compute the discount on a purchase policy, at a certain age
	 * @param pSpec the policy specification
	 * @param age the age in tosses
	 * @param reserve the current IB reserve ratio
	 * @return the discount price demanded at that toss for the policy.  This
	 * is a fraction of the difference between the current redemption rate and
	 * 1.0.  For example a 50% discount with 80% reserve ratio is 50% above a
	 * 20% difference, hence a discount price demanded of 90%.
	 */
	public static double parseBuyPolicy(String pSpec, byte age, double reserve) {
		if (pSpec == null || age < 0 || reserve < 0.0) {
			Log.error("parseBuyPolicy(): bad input");
			return 0.0;
		}
		double discount = 1.0;
		double redRate = reserve;

		// an empty spec means any discount, hence sell for reserve ratio
		if (pSpec.isEmpty()) {
			return redRate;
		}
		double dRate = 1.0 - redRate;

		// parse into N discount stages
		String[] stages = pSpec.split("::");
		byte usedUp = 0;
		for (int stage = 0; stage < stages.length; stage++) {
			// parse the discount, which will be a double, and the tosses, an int
			String[] stgSpec = stages[stage].split(":");
			if (stgSpec == null || stgSpec.length != 2) {
				Log.error("Purch discount policy " + pSpec + " does not parse at "
							+ "stage " + stage);
				continue;
			}
			Double disc = null;
			Byte appliesFor = null;
			try {
				disc = Double.parseDouble(stgSpec[0]);
				appliesFor = Byte.parseByte(stgSpec[1].trim());
			}
			catch (NumberFormatException nfe) {
				Log.error("Purch discount policy " + pSpec + " does not parse at "
							+ "stage " + stage, nfe);
				continue;
			}
			// if this stage applies for the given age, then return the discount
			usedUp += appliesFor.byteValue();
			if (age < usedUp) {
				// demand this percentage of the reserve shortfall
				dRate *= (disc.doubleValue() / 100.0);
				break;
			}
		}
		if (age >= usedUp) {
			// buy at full price from Issuer
			dRate = 0.0;
		}
		discount -= dRate;
		return Math.min(1.0, discount);
	}

	/**
	 * compute the max number of tosses which a sale or purchase discount policy
	 * supports
	 * @param pSpec the policy specification
	 * @return the number of tosses after which the discount price will be 100%
	 * of the reserve shortfall (i.e. equal to the IB reserve rate), or 1.0
	 */
	public static int policyMaxTosses(String pSpec) {
		if (pSpec == null) {
			Log.error("policyMaxTosses(): missing spec");
			return 0;
		}
		// an empty spec means go to reserve ratio redemption value at once
		if (pSpec.isEmpty()) {
			return 0;
		}

		// parse into N discount stages
		String[] stages = pSpec.split("::");
		int tossTally = 0;
		for (int stage = 0; stage < stages.length; stage++) {
			// parse the discount to get the second item, # of tosses
			String[] stgSpec = stages[stage].split(":");
			if (stgSpec == null || stgSpec.length != 2) {
				Log.error("Discount policy " + pSpec + " does not parse at stage "
							+ stage);
				break;
			}
			Byte appliesFor = null;
			try {
				appliesFor = Byte.parseByte(stgSpec[1].trim());
			}
			catch (NumberFormatException nfe) {
				Log.error("Discount policy " + pSpec + " does not parse at stage "
							+ stage, nfe);
				break;
			}
			tossTally += appliesFor.intValue();
		}
		return tossTally;
	}

	// END methods
}
