/*
 * last modified---
 * 	04-30-15/jet/ Spark plugin version
 * 	02-10-15/jet/ add m_Age; make m_Policy a DiscountPolicy
 * 	01-23-15/jet/ new
 *
 * purpose---
 * 	provide a class representation of an ABC order to buy or sell IB
 */

package org.silentvault.spark;


/**
 * This class supplies an instance of an order to buy or sell InsurBucks in an
 * ABC account as a Java class.  It has methods to get and set all fields, along
 * with some minimal sanity checking logic.
 *
 * @author Justin Turrell
 */
public final class IBOrder {
	// BEGIN data members
	// order type constants
	/**
	 * order type to buy IB
	 */
	public static final String		M_TypeBuy = "buy";
	/**
	 * order type to sell IB
	 */
	public static final String		M_TypeSell = "sell";

	/**
	 * the order ID sequence number (unique index)
	 */
	private int							m_OrderId;

	/**
	 * the owning tranche ID within the account (unique within account)
	 */
	private int							m_TrancheId;

	/**
	 * original amount of InsurBucks offered or wanted
	 */
	private double						m_Quantity;

	/**
	 * type of order (buy or sell)
	 */
	private String						m_OrderType;

	/**
	 * the toss number after which the order was entered (YYMMDDHHmm format)
	 */
	private int							m_EntryToss;

	/**
	 * the age of the order, in tosses
	 */
	private byte						m_Age;

	/**
	 * discount policy to be used (a buy policy or a sell policy,
	 * depending on m_OrderType)
	 */
	private DiscountPolicy			m_Policy;

	// END data members

	// BEGIN methods
	/**
	 * nullary constructor
	 */
	public IBOrder() {
		m_OrderType = M_TypeBuy;
	}

	// get methods
	/**
	 * obtain the orderId ID
	 * @return the ID
	 */
	public int getOrderId() { return m_OrderId; }

	/**
	 * obtain the owning tranche ID
	 * @return the ID
	 */
	public int getTrancheId() { return m_TrancheId; }

	/**
	 * obtain the starting quantity of the IB order
	 * @return the initial quantity of IB to be bought or sold
	 */
	public double getQuantity() { return m_Quantity; }

	/**
	 * obtain the toss after which order was entered
	 * @return toss in YYMMDDHHmm format
	 */
	public int getEntryToss() { return m_EntryToss; }

	/**
	 * obtain the age of the order
	 * @return the number of tosses which have elapsed since entry
	 */
	public byte getAge() { return m_Age; }

	/**
	 * obtain the discount policy set
	 * @return the policy (points to entry like id ## disc1 : tosses ... )
	 */
	public DiscountPolicy getPolicy() { return m_Policy; }

	/**
	 * obtain the type of order
	 * @return the type (buy or sell)
	 */
	public String getOrderType() { return m_OrderType; }


	// set methods
	/**
	 * set the order ID (this is an unique index)
	 * @param orderId the order ID to set
	 */
	public void setOrderId(int orderId) {
		if (orderId > 0) {
			// this is an immutable value which should be set only once
			if (m_OrderId == 0) {
				m_OrderId = orderId;
			}
		}
	}

	/**
	 * set the tranche ID
	 * @param trancheId the tranche ID to set
	 */
	public void setTrancheId(int trancheId) {
		if (trancheId >= 0) {
			// this is an immutable value which should be set only once
			if (m_TrancheId == 0) {
				m_TrancheId = trancheId;
			}
		}
	}

	/**
	 * set the starting amount of IB
	 * @param ib the amount of IB initially wanted/offered
	 */
	public void setQuantity(double ib) {
		if (ib > 0.0) {
			m_Quantity = ib;
		}
	}

	/**
	 * set the order type
	 * @param type the type of order (buy or sell)
	 */
	public void setOrderType(String type) {
		if (type != null && (type.equals(M_TypeBuy) || type.equals(M_TypeSell))) {
			m_OrderType = new String(type);
		}
	}

	/**
	 * set the toss after which order was originally placed
	 * @param toss the toss in YYMMDDHHmm format
	 */
	public void setEntryToss(int toss) {
		if (toss >= 0) {
			m_EntryToss = toss;
		}
	}

	/**
	 * set the age of the order
	 * @param age the number of tosses which have elapsed since order entry
	 */
	public void setAge(byte age) {
		if (age >= 0) {
			m_Age = age;
		}
	}

	/**
	 * set the discount policy
	 * @param policy the policy to set (entry like id ## disc1 : tosses ... )
	 */
	public void setPolicy(DiscountPolicy policy) {
		if (policy != null) {
			m_Policy = policy;
		}
	}

	// END methods
}
