/*
 * last modified---
 * 	04-30-15/jet/ Spark plugin version
 * 	02-09-15/jet/ add Age column
 * 	01-26-15/jet/ new
 *
 * purpose---
 * 	provide a cell renderer for all columns in the IB order table display
 */

package org.silentvault.spark.ui;

import javax.swing.border.EmptyBorder;
import javax.swing.border.Border;
import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.SwingConstants;
import javax.swing.UIManager;
import javax.swing.table.TableCellRenderer;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.svm.abc.StatsPane;
import org.silentvault.spark.ui.svm.abc.ATabManager;

import java.util.Date;
import java.text.NumberFormat;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;


/**
 * This class is used as the DefaultRenderer for all IB order table columns.
 * Tooltips are set appropriately to describe the data.
 *
 * @author Justin Turrell
 */
public final class IBOrderRenderer extends JLabel implements TableCellRenderer {
	// data members
	/**
	 * the border to use for the cell
	 */
	private Border					m_Border = null;

	/**
	 * a helpful date formatter for tosses
	 */
	private SimpleDateFormat	m_DateFormatter = null;

	/**
	 * another date formatter, for human-readable tosses
	 */
	private SimpleDateFormat	m_TossFormatter = null;

	/**
	 * the offer table model in support of which we render the table
	 */
	private IBOrderTableModel	m_OrderModel;

	/**
	 * the application or plugin of which we're a part
	 */
	private WalletClient			m_Plugin;

	// methods
	/**
	 * constructor
	 * @param orderModel the order table model we are supporting
	 * @param plugin the plugin or app that we're part of
	 */
	public IBOrderRenderer(IBOrderTableModel orderModel, WalletClient plugin) {
		super();
		m_OrderModel = orderModel;
		m_Plugin = plugin;
		setOpaque(true);			// to make background show up
		m_Border = new EmptyBorder(5, 3, 5, 3);
		m_TossFormatter = new SimpleDateFormat("yyMMddHHmm");
		m_DateFormatter = new SimpleDateFormat("yyyy-MMM-dd HH:mm");
	}

	/**
	 * method to implement the TableCellRenderer interface
	 * @param orderTable the order table we're rendering
	 * @param colValue the value of the particular column (String, Date, etc.)
	 * @param isSelected whether this cell (order) is currently selected (ignore)
	 * @param hasFocus whether this cell has the focus (we ignore this)
	 * @param row the table row (i.e. which order in the model)
	 * @param column the table column (we use this to distinguish cases)
	 */
	public Component getTableCellRendererComponent(	JTable orderTable,
																	Object colValue,
																	boolean isSelected,
																	boolean hasFocus,
																	int row,
																	int column )
	{
		// obtain the offer data, allowing for possibility table is sorted
		int realRow = orderTable.convertRowIndexToModel(row);
		IBOrder order = m_OrderModel.getOrderAtRow(realRow);
		if (order == null) {
			return null;
		}
		if (colValue == null) {
			return null;
		}

		// set defaults
		setFont(m_Plugin.getWalletTabManager().getHomePane().M_TableFont);
		setForeground(UIManager.getColor("Table.foreground"));
		setHorizontalAlignment(SwingConstants.TRAILING);

		// use default background color (off-white) if row is even
		Color offWhite = UIManager.getColor("Table.alternateRowColor");
		if (row % 2 != 0) {
			setBackground(Color.WHITE);
		}
		else {
			setBackground(offWhite);
		}

		/* Examine the column type:
		 * Double - it's an order quantity
		 * Integer - it's an order ID number, tranche ID number, or entered toss
		 * String - it's an order type or a discount policy
		 */
		if (colValue instanceof Double) {
			// amount of order
			Double val = (Double) colValue;
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
			df.setGroupingUsed(false);
			df.setMaximumFractionDigits(XMLVoucher.M_Precision);
			df.setMinimumFractionDigits(XMLVoucher.M_Precision);
			setText(df.format(val.doubleValue()));
			// use bold
			setFont(m_Plugin.getWalletTabManager().getHomePane().M_LabelFont);

			// now determine the current discount value for the tooltip
			ATabManager atm = (ATabManager) m_Plugin.getMarketplaceTabManager();
			StatsPane sp = atm.getStatsPane();
			double curDisc = 0.0;
			DiscountPolicy discPol = order.getPolicy();
			if (discPol == null) {
				setToolTipText("Amount of IB subject to order");
			}
			else {
				String polSpec = discPol.getSpecification();
				byte ordAge = order.getAge();
				df.setMaximumFractionDigits(2);
				df.setMinimumFractionDigits(2);
				if (order.getOrderType().equals(IBOrder.M_TypeBuy)) {
					curDisc = DiscountPolicy.parseBuyPolicy(polSpec,
																		 ordAge,
																		 sp.getIBRatio());
					setToolTipText("Current offer price: "
										+ df.format(curDisc * 100.0) + "%");
				}
				else {
					curDisc = DiscountPolicy.parseSellPolicy(polSpec,
																		  ordAge,
																		  sp.getIBRatio());
					setToolTipText("Current asking price: "
										+ df.format(curDisc * 100.0) + "%");
				}
			}
		}
		else if (colValue instanceof Integer) {
			// this is either tranche ID, Order #, Entered toss (date), or Age
			Integer oId = (Integer) colValue;
			Date tossTime = null;
			switch (column) {
				case 0:
					// tranche ID
					setText(oId.toString());
					setToolTipText("Tranche which owns this order");
					break;

				case 1:
					// order ID
					setText(oId.toString());
					setToolTipText("Unique order sequence number");
					break;

				case 3:
					// first toss on which order was entered
					setText(oId.toString());
					try {
						tossTime = m_TossFormatter.parse(oId.toString());
						setToolTipText(m_DateFormatter.format(tossTime) + " UTC");
					}
					catch (ParseException pe) {
						Log.error("Invalid toss data format, " + oId, pe);
					}
					break;

				default:
					setToolTipText("unexpected Integer column, " + column);
					break;
			}
		}
		else if (colValue instanceof Byte) {
			Byte aId = (Byte) colValue;
			if (column == 4) {
				// age of order in tosses
				setText(aId.toString());
				// get the max possible age and show it on the tooltip
				Integer maxAge = new Integer(DiscountPolicy.policyMaxTosses(
													order.getPolicy().getSpecification()));
				 setToolTipText(aId.toString() + " / " + maxAge.toString());
			}
			else {
				setToolTipText("unexpected Byte column, " + column);
			}
		}
		else if (colValue instanceof String) {
			// this is either Type or discount Policy
			String sVal = (String) colValue;
			switch (column) {
				case 5:
					// order type
					setText(sVal.toUpperCase());
					if (order.getOrderType().equals(IBOrder.M_TypeBuy)) {
						setToolTipText("Your tranche is buying IB with BTC");
					}
					else {
						setToolTipText("Your tranche is selling IB for BTC");
					}
					break;

				case 6:
					// discount policy name
					setText(sVal);

					// for tooltip, display policy definition
					DiscountPolicy discPol = order.getPolicy();
					if (discPol == null) {
						setToolTipText("Unknown discount policy");
						break;
					}
					if (!discPol.getSpecification().isEmpty()) {
						setToolTipText(discPol.getSpecification());
					}
					else {
						setToolTipText("(immediate settlement on next toss)");
					}
					break;

				default:
					setToolTipText("unexpected String column, " + column);
					break;
			}
		}
		else {
			Log.error("Unexpected column data type in IB order renderer, "
						+ colValue.getClass());
			setText("column type error");
		}

		// fix up the border to impose inset requirements
		setBorder(m_Border);

		return this;
	}

	// end methods
}
