/*
 * last modified---
 * 	04-40-15/jet/ Spark plugin version
 * 	03-31-15/jet/ add getBuyTotal()
 * 	02-09-15/jet/ add Age column
 * 	01-23-15/jet/ new
 *
 * purpose---
 * 	provide a TableModel for the IB order JTable display in the OrdersPane
 */

package org.silentvault.spark.ui;

import javax.swing.table.AbstractTableModel;
import javax.swing.JRadioButton;
import javax.swing.ButtonGroup;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.svm.abc.ATabManager;
import org.silentvault.spark.ui.svm.abc.StatsPane;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.List;
import java.util.Date;
import java.text.SimpleDateFormat;
import java.text.ParseException;


/**
 * Extends AbstractTableModel to provide a table model which can be used to
 * display a series of IB buy/sell orders in a scrolling window
 *
 * @author Justin Turrell
 */
public final class IBOrderTableModel extends AbstractTableModel {
	// BEGIN data members
	/**
	 * the column names (header values)
	 */
	private final String[]	M_ColumnNames = {	"ID",
															"Order #",
															"Amount",
															"Entered",
															"Age",
															"Type",
															"Policy" };

	/**
	 * the list of orders we're representing (our data)
	 */
	private ArrayList<IBOrder>		m_OrderList;

	/**
	 * the owning plugin/application
	 */
	private WalletClient				m_Plugin;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param orders the list of orders which we are to represent
	 * @param plugin the plugin that owns our display
	 */
	public IBOrderTableModel(ArrayList<IBOrder> orders, WalletClient plugin) {
		super();
		m_OrderList = orders;
		m_Plugin = plugin;
	}

	// methods required to implement AbstractTableModel
	/**
	 * get the row count
	 * @return the count (number of tranches)
	 */
	public int getRowCount() {
		return m_OrderList.size();
	}

	/**
	 * get the column count
	 * @return the count
	 */
	public int getColumnCount() { return M_ColumnNames.length; }

	/**
	 * get the name of a given column
	 * @param col the column number
	 * @return the name
	 */
	public String getColumnName(int col) {
		if (col < 0 || col > M_ColumnNames.length) {
			return null;
		}
		return M_ColumnNames[col];
	}

	/**
	 * get the value of a particular cell
	 * @param row the row (order) number
	 * @param col the column number
	 * @return the item at that row,col cell
	 */
	public Object getValueAt(int row, int col) {
		if (row >= m_OrderList.size() || row < 0 || col < 0) {
			// request input was invalid
			return null;
		}
		IBOrder order = m_OrderList.get(row);

		// this should never happen, but...
		if (order == null) {
			return null;
		}

		Object ret = null;
		switch (col) {
			case 0:	// tranche Id number
				Integer tId = new Integer(order.getTrancheId());
				ret = tId;
				break;

			case 1:	// order Id number
				Integer oId = new Integer(order.getOrderId());
				ret = oId;
				break;

			case 2:	// amount of order
				Double amt = new Double(order.getQuantity());
				ret = amt;
				break;

			case 3:	// toss on which entered
				Integer toss = new Integer(order.getEntryToss());
				ret = toss;
				break;

			case 4:	// age of order, in tosses
				// obtain the next scheduled toss from the stats pane
				Byte age = new Byte(order.getAge());
				ret = age;
				break;

			case 5:	// order type (buy/sell)
				String otype = new String(order.getOrderType());
				ret = otype;
				break;

			case 6:	// discount policy name
				String policy = new String(order.getPolicy().getName());
				ret = policy;
				break;

			default:	// impossible
				Log.error("Impossible order column number," + col);
				break;
		}
		return ret;
	}

	/**
	 * determine the default renderer to use for a column
	 * @param col the column whose Class we are interested in
	 * @return the Class of object stored in that column (see getValueAt())
	 */
	public Class getColumnClass(int col) {
		Object val = getValueAt(0, col);
		if (val != null) {
			return val.getClass();
		}
		String dummy = "";
		return dummy.getClass();
	}


	// other get methods
	/**
	 * get the whole order list
	 * @return a copy of the list (to avoid side-effects)
	 */
	public ArrayList<IBOrder> getOrderList() {
		return new ArrayList<IBOrder>(m_OrderList);
	}

	/**
	 * get the order at a specific row
	 * @param row the row whose tranche we're interested in
	 * @return the order (null if not found or out of range)
	 */
	public IBOrder getOrderAtRow(int row) {
		if (row >= m_OrderList.size() || row < 0) {
			return null;
		}
		return m_OrderList.get(row);
	}

	/**
	 * obtain the sum of all buy orders in the list
	 * @return the aggregate amount of IB to be bought
	 */
	public double getBuyTotal() {
		double bTot = 0.0;
		for (IBOrder ord : m_OrderList) {
			if (ord.getOrderType().equals(IBOrder.M_TypeBuy)) {
				bTot += ord.getQuantity();
			}
		}
		return bTot;
	}


	// write operations
	/**
	 * insert a new order into the list
	 * @param order the order to be added
	 */
	public void addOrder(IBOrder order) {
		if (order != null) {
			if (!m_OrderList.contains(order)) {
				m_OrderList.add(order);

				// notify any listeners
				fireTableDataChanged();
			}
		}
	}

	/**
	 * destructively merge a list of orders into the list
	 * @param orderList the orders we need to merge
	 */
	public void mergeOrders(List<IBOrder> orderList) {
		m_OrderList.clear();
		if (orderList == null || orderList.isEmpty()) {
			fireTableDataChanged();
			return;
		}
		m_OrderList.addAll(orderList);

		// notify listeners
		fireTableDataChanged();
	}

	/**
	 * remove a list of orders
	 * @param orderList the orders we're to remove
	 */
	public void removeOrders(ArrayList<IBOrder> orderList) {
		if (orderList != null && !orderList.isEmpty()) {
			for (IBOrder order : orderList) {
				m_OrderList.remove(order);
			}

			// notify listeners
			fireTableDataChanged();
		}
	}

	/**
	 * compute the age of an order, based on the next scheduled toss
	 * @param order the order whose age we want
	 * @return the age, in tosses (0 on failure)
	 */
	public int computeOrderAge(IBOrder order) {
		// obtain the next scheduled toss from the stats pane
		int age = 0;
		SimpleDateFormat sdf = new SimpleDateFormat("yyMMddHHmm");
		ATabManager atm = (ATabManager) m_Plugin.getMarketplaceTabManager();
		StatsPane sp = atm.getStatsPane();
		int nextToss = sp.getNextToss();
		if (nextToss <= 0) {
			// we don't have a good value
			Log.error("No next scheduled toss available from StatsPane");
			return age;
		}
		Integer nToss = new Integer(nextToss);
		Integer eToss = new Integer(order.getEntryToss());
		Date entryToss = null;
		Date nextTossTime = null;
		// find the difference in minutes between this and our entry toss
		try {
			entryToss = sdf.parse(eToss.toString());
			nextTossTime = sdf.parse(nToss.toString());
			long tDiff = nextTossTime.getTime() - entryToss.getTime();
			Long minDiff
				= new Long(tDiff / javax.management.timer.Timer.ONE_MINUTE);
			int tossesPerMin = 60 / sp.getTossRate();
			age = minDiff.intValue() / tossesPerMin;
		}
		catch (ParseException pe) {
			Log.error("Invalid toss data format, " + eToss + " or " + nToss,
						pe);
		}
		return age;
	}

	/**
	 * finalize object when garbage-collected, for data security
	 */
	@Override
	protected void finalize() throws Throwable {
		try {
			// nix sensitive data if there is any
			if (m_OrderList != null && !m_OrderList.isEmpty()) {
				m_OrderList.clear();
			}
		} finally {
			super.finalize();
		}
	}

	// END methods
}
