/*
 * last modified---
 * 	04-30-15/jet/ Spark plugin version
 * 	04-14-15/jet/ use colors on up/down
 * 	04-07-15/jet/ show next wagers in the Incr column tooltips
 * 	01-26-15/jet/ new
 *
 * purpose---
 * 	provide a cell renderer for all columns in the tranche table display
 */

package org.silentvault.spark.ui;

import javax.swing.border.EmptyBorder;
import javax.swing.border.Border;
import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.JRadioButton;
import javax.swing.SwingConstants;
import javax.swing.UIManager;
import javax.swing.table.TableCellRenderer;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.ABCTranche;
import org.silentvault.spark.XMLVoucher;
import org.silentvault.spark.WalletClient;

import java.text.NumberFormat;
import java.text.DecimalFormat;


/**
 * This class is used as the DefaultRenderer for all tranche columns.
 * Tooltips are set appropriately to describe the data.
 *
 * @author Justin Turrell
 */
public final class TrancheRenderer extends JLabel implements TableCellRenderer {
	// data members
	/**
	 * the border to use for the cell
	 */
	private Border					m_Border = null;

	/**
	 * the tranche table model in support of which we render the table
	 */
	private TrancheTableModel	m_TrancheModel;

	/**
	 * the application or plugin of which we're a part
	 */
	private WalletClient			m_Plugin;

	// methods
	/**
	 * constructor
	 * @param trancheModel the order table model we are supporting
	 * @param plugin the plugin or app that we're part of
	 */
	public TrancheRenderer(TrancheTableModel trancheModel, WalletClient plugin) {
		super();
		m_TrancheModel = trancheModel;
		m_Plugin = plugin;
		setOpaque(true);			// to make background show up
		m_Border = new EmptyBorder(3, 3, 3, 3);
	}

	/**
	 * method to implement the TableCellRenderer interface
	 * @param trancheTable the order table we're rendering
	 * @param colValue the value of the particular column (String, Date, etc.)
	 * @param isSelected whether this cell (tranche) is currently selected
	 * @param hasFocus whether this cell has the focus (we ignore this)
	 * @param row the table row (i.e. which tranche in the model)
	 * @param column the table column (we use this to distinguish cases)
	 */
	public Component getTableCellRendererComponent(	JTable trancheTable,
																	Object colValue,
																	boolean isSelected,
																	boolean hasFocus,
																	int row,
																	int column )
	{
		// obtain the offer data, allowing for possibility table is sorted
		int realRow = trancheTable.convertRowIndexToModel(row);
		ABCTranche tranche = m_TrancheModel.getTrancheAtRow(realRow);
		if (tranche == null) {
			return null;
		}
		if (colValue == null) {
			return null;
		}

		// set defaults
		setFont(m_Plugin.getWalletTabManager().getHomePane().M_TableFont);
		setForeground(UIManager.getColor("Table.foreground"));
		setHorizontalAlignment(SwingConstants.TRAILING);

		// use default background color (off-white) if row is even
		Color offWhite = UIManager.getColor("Table.alternateRowColor");
		if (row % 2 != 0) {
			setBackground(Color.WHITE);
		}
		else {
			setBackground(offWhite);
		}

		/* Examine the column type:
		 * Double - it's a Funding, BTC, IB, or Bet value
		 * Float - it's a Value (%) value
		 * Integer - it's an order ID number or tranche ID number
		 * Byte - it's an Iteration or Losses value
		 * Boolean - it's a Subsidy flag value
		 * String - it's an order type or a discount policy
		 */
		if (colValue instanceof JRadioButton) {
			// unless we're in single-select mode, return null here
			JRadioButton rb = (JRadioButton) colValue;
			if (m_TrancheModel.getSelectMode()
				 == TrancheTableModel.M_SelectSingle)
			{
				rb.setToolTipText("Click here to select this tranche");
				rb.setText("");
				rb.setOpaque(true);
				rb.setHorizontalAlignment(SwingConstants.CENTER);
				if (row % 2 != 0) {
					rb.setBackground(Color.WHITE);
				}
				else {
					rb.setBackground(offWhite);
				}
				return rb;
			}
			return null;
		}
		else if (colValue instanceof Double) {
			// this is either the starting funding, current BTC, IB,
			// or Bet increment value
			Double val = (Double) colValue;
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
			df.setGroupingUsed(false);
			df.setMaximumFractionDigits(XMLVoucher.M_Precision);
			df.setMinimumFractionDigits(XMLVoucher.M_Precision);
			setText(df.format(val.doubleValue()));
			switch (column) {
				case 2:
					// starting funding in BTC allocated to this tranche
					setToolTipText("Total starting funding (in BTC)");
					break;
				case 3:
					// current BTC chip balance
					setToolTipText("Current balance in BTC");
					break;
				case 4:
					// current IB balance
					setToolTipText("Current balance in IB");
					break;
				case 7:
					if (tranche.getOrderId() != 0) {
						setToolTipText("Betting suspended until open market order "
											+ "is resolved");
					}
					else {
						double btcBet = tranche.getBetIncrement();
						double doublings
							= Math.pow(2.0, (double) tranche.getLossCount());
						btcBet *= doublings;
						btcBet = Math.min(btcBet, tranche.getBTCChipBalance());
						double ibBet = btcBet / 1.6180339887;
						setToolTipText("Next wager: " + df.format(btcBet) + " BTC, "
											+ df.format(ibBet) + " IB");
					}
					break;
				default:
					Log.error("Unexpected Double column, " + column);
					break;
			}
		}
		else if (colValue instanceof Float) {
			Float val = (Float) colValue;
			DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
			df.setGroupingUsed(false);
			df.setMaximumFractionDigits(2);
			df.setMinimumFractionDigits(2);
			// use bold
			setFont(m_Plugin.getWalletTabManager().getHomePane().M_LabelFont);
			setText(df.format(val.doubleValue()) + "%");
			setToolTipText("Ratio of current cash-out value to starting funding");
			if (val.doubleValue() > 100.0) {
				setBackground(Color.GREEN.brighter());
			}
			else if (val.doubleValue() < 100.0) {
				setBackground(Color.RED);
			}
		}
		else if (colValue instanceof Integer) {
			// this is either tranche ID or Order ID
			Integer oId = (Integer) colValue;
			setText(oId.toString());
			switch (column) {
				case 1:
					setToolTipText("Tranche ID within your account");
					break;
				case 9:
					if (oId.intValue() == 0) {
						setToolTipText("No open order for this tranche");
					}
					else {
						setToolTipText("ID of open order to buy or sell IB "
											+ "(see Market)");
					}
					break;
				default:
					setToolTipText("Unexpected Integer column, " + column);
					break;
			}
		}
		else if (colValue instanceof Byte) {
			// this is either the current iteration, or the consecutive loss count
			Byte bCnt = (Byte) colValue;
			setText(bCnt.toString());
			switch (column) {
				case 6:
					// Martingale iteration
					if (bCnt.intValue() == 0) {
						setToolTipText("Tranche is not betting");
					}
					else {
						setToolTipText("Martingale betting iteration");
					}
					break;
				case 8:
					// loss count
					if (bCnt.intValue() == 0) {
						if (tranche.getIteration() == (byte)1) {
							setToolTipText("Loss count was reset");
						}
						else {
							setToolTipText("Last toss was a win");
						}
					}
					else {
						setToolTipText("Consecutive losses");
					}
					break;
				default:
					setToolTipText("Unexpected Byte column, " + column);
					break;
			}
		}
		else if (colValue instanceof Boolean) {
			// subsidy flag
			Boolean flag = (Boolean) colValue;
			if (flag.booleanValue()) {
				setText("Yes");
				setToolTipText("Tranche is receiving bonus IB subsidy");
				setBackground(Color.PINK);
			}
			else {
				setText("No");
				setToolTipText("Tranche is not receiving bonus IB");
			}
		}
		else {
			Log.error("Unexpected column data type in tranche renderer, "
						+ colValue.getClass());
			setText("column type error");
		}

		// fix up the border to impose inset requirements
		setBorder(m_Border);

		return this;
	}

	// end methods
}
