/*
 * last modified---
 * 	04-30-15/jet/ Spark plugin version
 * 	01-27-15/jet/ new
 *
 * purpose---
 * 	provide a cell editor for the selection column in a tranche table
 */

package org.silentvault.spark.ui;

import javax.swing.JRadioButton;
import javax.swing.JToggleButton;
import javax.swing.JTable;
import javax.swing.table.TableCellEditor;
import javax.swing.AbstractCellEditor;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.WalletClient;
import org.silentvault.spark.ui.svm.*;
import org.silentvault.spark.ui.svm.abc.*;

import java.util.EventObject;


/**
 * This class is used as a specialty editor for tranche table cells which are
 * in the 0th (Select) column of the table.  It is necessary because we need
 * radio button support (and DefaultCellEditor only support JCheckBox).
 *
 * @author Justin Turrell
 */
public final class TrancheSelector extends AbstractCellEditor
		implements TableCellEditor, ActionListener
{
	// data members
	/**
	 * the editor activation command
	 */
	protected static final String	SELECT = "select";

	/**
	 * the editor component
	 */
	private JRadioButton			m_CellEditor;

	/**
	 * the value in the table
	 */
	private JRadioButton			m_CellValue;

	/**
	 * the application or plugin of which we're a part
	 */
	private WalletClient			m_Plugin;

	/**
	 * the tranche table model in support of which we edit the cells
	 */
	private TrancheTableModel	m_TrancheModel;


	// methods
	/**
	 * constructor
	 * @param plugin the plugin or app that we're part of
	 * @param trancheModel the tranche table model we are supporting
	 */
	public TrancheSelector(WalletClient plugin, TrancheTableModel trancheModel) {
		super();
		m_Plugin = plugin;
		m_TrancheModel = trancheModel;

		// configure the buttons
		m_CellEditor = new JRadioButton();
		m_CellEditor.setVisible(false);
		m_CellEditor.setActionCommand(SELECT);
		m_CellEditor.addActionListener(this);
		m_CellEditor.setBorderPainted(false);
	}


	// method required to implement ActionListener
	public void actionPerformed(ActionEvent e) {
		if (SELECT.equals(e.getActionCommand())) {
			if (m_CellValue == null) {
				Log.error("Null cell value on click to start edit");
			}
			else {
				// only allow selection if corresponding cell value button enabled
				if (m_CellValue.isEnabled()) {
					m_CellEditor.setSelected(!m_CellValue.isSelected());
				}
			}

			// tell JTable to switch back to renderer for this cell
			fireEditingStopped();
		}
		else {
			Log.error("Unexpected action on select button");
		}
	}


	// method required to implement CellEditor
	/**
	 * obtain the value contained in the editor
	 * @return the value
	 */
	public Object getCellEditorValue() {
		// propogate the selected status if the corresponding button is enabled
		if (m_CellValue.isEnabled()) {
			m_CellValue.setSelected(m_CellEditor.isSelected());
		}

		// redraw the table to force repainting any now de-selected button
		// (the buttons are in a ButtonGroup, so side-effects are desired)

		MTabManager mtm = m_Plugin.getSVMPlugin().getTabManager();
		ATabManager atm = null;
		AHomePane hp = null;
		if (mtm instanceof ATabManager) {
			atm = (ATabManager) mtm;
		}
		else {
			Log.error("Tab manager is not of type ATabManager");
			return null;
		}
		MHomePane mhp = atm.getHomePane();
		if (mhp instanceof AHomePane) {
			hp = (AHomePane) mhp;
		}
		else {
			Log.error("Home pane is not of type AHomePane");
			return null;
		}
		final JTable tt = atm.getHoldingsPane().getTranchesTable();

		/* There are certain panes which might want to know about the change of
		 * selection.  Notify them.
		 */
		int mode = m_TrancheModel.getSelectMode();
		if (mode == TrancheTableModel.M_SelectSingle) {
			// pass the event to whichever screen is up
			int curScreen = atm.getCurrentState();
			if (curScreen == ATabManager.M_GO_MAIN_MENU) {
				HoldingsPane holdp = atm.getHoldingsPane();
				holdp.trancheSelectChanged();
			}
			else {
				Log.error("Unexpected screen up (" + curScreen
							+ ") on tranche select");
			}
		}

		/* redraw the table to force repainting any now de-selected button
		 * (the buttons are in a ButtonGroup, so side-effects are desired)
		 */
		tt.revalidate();
		tt.repaint();
		return m_CellEditor;
	}


	// method required to implement TableCellEditor
	/**
	 * obtain the cell editor Component
	 * @param trancheTable the tranche table we're rendering
	 * @param colValue the value of the particular column we are editing
	 * (NB: this value is itself a JCheckBox or JRadioButton)
	 * @param isSelected whether this cell (offer) is currently selected
	 * @param row the table row (i.e. which offer in the model, ignored)
	 * @param column the table column, which we validate
	 */
	public Component getTableCellEditorComponent(JTable trancheTable,
																Object colValue,
																boolean isSelected,
																int row,
																int column)
	{
		if (column == 0) {
			m_CellValue = (JRadioButton) colValue;
		}
		else {
			Log.error("getTableCellEditorComponent() invoked on column " + column);
			return null;
		}
		m_CellEditor.setSelected(false);
		return m_CellEditor;
	}

	// optional method extending AbstractCellEditor
	/**
	 * see whether the cell can be edited
	 * @param anEvent the event which starts cell editing
	 * @return always true unless we're in no-select mode
	 */
	@Override
	public boolean isCellEditable(EventObject anEvent) {
		// check with the model
		if (m_TrancheModel.getSelectMode() == TrancheTableModel.M_SelectNone) {
			return false;
		}
		return true;
	}

	// optional method extending CellEditor
	/**
	 * state whether cell should be selected when edited
	 * @return always false
	 */
	@Override
	public boolean shouldSelectCell(EventObject anEvent) { return false; }

	// end methods
}
