/*
 * last modified---
 * 	04-30-15/jet/ Spark plugin version
 * 	01-30-15/jet/ add getBTCTotal(), getIBTotal(), getBTCStartTotal()
 * 	01-23-15/jet/ new
 *
 * purpose---
 * 	provide a TableModel for the tranche JTable display in the HoldingsPane
 */

package org.silentvault.spark.ui;

import javax.swing.table.AbstractTableModel;
import javax.swing.JRadioButton;
import javax.swing.ButtonGroup;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.svm.abc.ATabManager;
import org.silentvault.spark.ui.svm.abc.StatsPane;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.List;


/**
 * Extends AbstractTableModel to provide a table model which can be used to
 * display a series of tranches in a scrolling window.
 *
 * @author Justin Turrell
 */
public final class TrancheTableModel extends AbstractTableModel {
	// BEGIN data members
	/**
	 * the column names (header values)
	 */
	private final String[]	M_ColumnNames = { "",	// select has no title
															"T",
															"Start",
															"BTC",
															"IB",
															"Value",
															"M",
															"Incr",
															"L",
															"Ord",
															"B" };
	
	// selection modes
	/**
	 * selection mode constant for when tranches are not selectable
	 */
	public static final int			M_SelectNone = 0;
	/**
	 * selection mode constant for when tranches are single-selectable
	 */
	public static final int			M_SelectSingle = 1;

	/**
	 * the list of tranches we're representing (our data)
	 */
	private ArrayList<ABCTranche>	m_TrancheList;

	/**
	 * the owning plugin/application
	 */
	private WalletClient				m_Plugin;

	/**
	 * the list of radio buttons which correspond to the tranches and are shown
	 * in the first column
	 */
	private Hashtable<ABCTranche, JRadioButton>	m_SelectButtons;

	/**
	 * the selection mode (either M_SelectNone or M_SelectSingle)
	 */
	private int							m_SelectMode;

	/**
	 * the button group for the radio buttons, to ensure only one at a time can
	 * be checked
	 */
	private ButtonGroup				m_ButtonGroup;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param tranches the list of tranches which we are to represent
	 * @param plugin the plugin that owns our display
	 */
	public TrancheTableModel(ArrayList<ABCTranche> tranches,
									 WalletClient plugin)
	{
		super();
		m_TrancheList = tranches;
		m_Plugin = plugin;
		m_SelectButtons = new Hashtable<ABCTranche, JRadioButton>();
		m_ButtonGroup = new ButtonGroup();
		m_SelectMode = M_SelectNone;

		// make a button for each tranche
		for (ABCTranche ttt : m_TrancheList) {
			JRadioButton rb = new JRadioButton();
			rb.setEnabled(true);
			rb.setRolloverEnabled(true);
			m_SelectButtons.put(ttt, rb);
			m_ButtonGroup.add(rb);
		}
	}

	// methods required to implement AbstractTableModel
	/**
	 * get the row count
	 * @return the count (number of tranches)
	 */
	public int getRowCount() {
		return m_TrancheList.size();
	}

	/**
	 * get the column count
	 * @return the count
	 */
	public int getColumnCount() { return M_ColumnNames.length; }

	/**
	 * get the name of a given column
	 * @param col the column number
	 * @return the name
	 */
	public String getColumnName(int col) {
		if (col < 0 || col > M_ColumnNames.length) {
			return null;
		}
		return M_ColumnNames[col];
	}

	/**
	 * get the value of a particular cell
	 * @param row the row (tranche) number
	 * @param col the column number
	 * @return the item at that row,col cell
	 */
	public Object getValueAt(int row, int col) {
		if (row >= m_TrancheList.size() || row < 0 || col < 0) {
			// request input was invalid
			return null;
		}
		ABCTranche tranche = m_TrancheList.get(row);

		// this should never happen, but...
		if (tranche == null) {
			return null;
		}

		Object ret = null;
		switch (col) {
			case 0:	// select radio button
				if (m_SelectMode == M_SelectSingle) {
					ret = m_SelectButtons.get(tranche);
				}
				// else: null is fine
				break;

			case 1:	// tranche Id number
				Integer tId = new Integer(tranche.getTrancheId());
				ret = tId;
				break;

			case 2:	// starting funding in BTC
				Double btcStart = new Double(tranche.getBTCStart()
														+ tranche.getBTCAdditions());
				ret = btcStart;
				break;

			case 3:	// current BTC chips balance
				Double btc = new Double(tranche.getBTCChipBalance());
				ret = btc;
				break;

			case 4:	// current IB balance
				Double ib = new Double(tranche.getIBBalance());
				ret = ib;
				break;

			case 5:	// value up or down, in percent
				// to compute this correctly, we'll need the IB ratio
				double sum = tranche.getBTCChipBalance();
				ATabManager atm = (ATabManager) m_Plugin.getMarketplaceTabManager();
				StatsPane spn = atm.getStatsPane();
				sum += (tranche.getIBBalance() * spn.getIBRatio());
				Double value = new Double(100.0 * (sum / tranche.getBTCStart()));
				Float valPct = value.floatValue();
				ret = valPct;
				break;

			case 6:	// current Martingale iteration
				Byte iter = new Byte(tranche.getIteration());
				ret = iter;
				break;

			case 7:	// current betting increment
				Double bet = new Double(tranche.getBetIncrement());
				ret = bet;
				break;

			case 8:	// total consecutive losses this cycle
				Byte losses = new Byte(tranche.getLossCount());
				ret = losses;
				break;

			case 9:	// order Id, if any
				Integer oId = tranche.getOrderId();
				ret = oId;
				break;

			case 10:	// flag whether tranche is receiving subsidy bonus IB
				Boolean sub = new Boolean(tranche.isSubsidized());
				ret = sub;
				break;

			default:	// impossible
				Log.error("Impossible tranche column number, " + col);
				break;
		}
		return ret;
	}

	/**
	 * method to determine the default renderer to use for a column
	 * @param col the column whose Class we are interested in
	 * @return the Class of object stored in that column (see getValueAt())
	 */
	public Class getColumnClass(int col) {
		Object val = getValueAt(0, col);
		if (val != null) {
			return val.getClass();
		}
		String dummy = "";
		return dummy.getClass();
	}


	// other get methods
	/**
	 * get the whole tranche list
	 * @return a copy of the list (to avoid side-effects)
	 */
	public ArrayList<ABCTranche> getTrancheList() {
		return new ArrayList<ABCTranche>(m_TrancheList);
	}

	/**
	 * get the tranche at a specific row
	 * @param row the row whose tranche we're interested in
	 * @return the tranche (null if not found or out of range)
	 */
	public ABCTranche getTrancheAtRow(int row) {
		if (row >= m_TrancheList.size() || row < 0) {
			return null;
		}
		return m_TrancheList.get(row);
	}

	/**
	 * return all selected tranches
	 * @return the tranches whose associated radio is checked (or null)
	 */
	public ABCTranche getSelectedTranche() {
		ABCTranche ret = null;
		if (m_SelectMode != M_SelectSingle) {
			return ret;
		}
		for (ABCTranche tranche : m_TrancheList) {
			JRadioButton rad = m_SelectButtons.get(tranche);
			if (rad != null && rad.isSelected()) {
				ret = tranche;
				break;
			}
		}
		return ret;
	}

	/**
	 * obtain the current selection mode
	 * @return the mode (one of M_SelectNone, M_SelectSingle)
	 */
	public int getSelectMode() { return m_SelectMode; }


	// write operations
	/**
	 * insert a new tranche into the list
	 * @param tranche the tranche to be added
	 */
	public void addTranche(ABCTranche tranche) {
		if (tranche != null) {
			if (!m_TrancheList.contains(tranche)) {
				m_TrancheList.add(tranche);
				// this was added
				JRadioButton rb = new JRadioButton();
				rb.setEnabled(true);
				rb.setRolloverEnabled(true);
				m_SelectButtons.put(tranche, rb);
				m_ButtonGroup.add(rb);

				// notify any listeners
				fireTableDataChanged();
			}
		}
	}

	/**
	 * destructively merge a list of tranches into the list
	 * @param trancheList the tranches we need to merge
	 */
	public void mergeTranches(List<ABCTranche> trancheList) {
		m_TrancheList.clear();
		if (trancheList == null || trancheList.isEmpty()) {
			fireTableDataChanged();
			return;
		}
		m_TrancheList.addAll(trancheList);

		// rebuild select controls
		for (ABCTranche ttt : m_TrancheList) {
			if (m_SelectButtons.get(ttt) == null) {
				JRadioButton rb = new JRadioButton();
				rb.setEnabled(true);
				rb.setRolloverEnabled(true);
				m_SelectButtons.put(ttt, rb);
				m_ButtonGroup.add(rb);
			}
			else {
				// it was already there, make sure buttons are enabled
				JRadioButton rb = m_SelectButtons.get(ttt);
				rb.setEnabled(true);
			}
		}

		// notify listeners
		fireTableDataChanged();
	}

	/**
	 * remove a list of tranches
	 * @param trancheList the tranches we're to remove
	 */
	public void removeTranches(ArrayList<ABCTranche> trancheList) {
		if (trancheList != null && !trancheList.isEmpty()) {
			for (ABCTranche tranche : trancheList) {
				m_TrancheList.remove(tranche);
				JRadioButton rb = m_SelectButtons.get(tranche);
				m_SelectButtons.remove(tranche);
				m_ButtonGroup.remove(rb);
			}

			// notify listeners
			fireTableDataChanged();
		}
	}

	/**
	 * set the selection mode: 0 == none; 1 == radios
	 * @param mode the offer selection mode
	 */
	public void setSelectMode(int mode) {
		if (mode >= M_SelectNone && mode <= M_SelectSingle) {
			m_SelectMode = mode;
		}
		else {
			Log.error("Invalid offer selection mode, " + mode);
		}
	}

	/**
	 * deselect everything (clear all radio buttons)
	 */
	public void deselectAll() {
		for (ABCTranche tranche : m_SelectButtons.keySet()) {
			JRadioButton rad = m_SelectButtons.get(tranche);
			if (rad != null) {
				rad.setSelected(false);
			}
		}
		m_ButtonGroup.clearSelection();

		// notify listeners
		fireTableDataChanged();
	}

	/**
	 * de-select a particular tranche
	 * @param unselected the tranche to unselect
	 */
	public void deselectTranche(ABCTranche unselected) {
		if (unselected == null) {
			return;
		}
		JRadioButton rad = m_SelectButtons.get(unselected);
		if (rad != null) {
			rad.setSelected(false);
		}
		if (m_SelectMode == M_SelectSingle) {
			m_ButtonGroup.clearSelection();
		}

		// notify listeners
		fireTableDataChanged();
	}

	/**
	 * obtain the sum of BTC chips in all tranches
	 * @return the aggregate amount of BTC chips held
	 */
	public double getBTCTotal() {
		double btcSum = 0.0;
		for (ABCTranche tra : m_TrancheList) {
			btcSum += tra.getBTCChipBalance();
		}
		return btcSum;
	}

	/**
	 * obtain the sum of IB in all tranches
	 * @return the aggregate amount of IB held
	 */
	public double getIBTotal() {
		double ibSum = 0.0;
		for (ABCTranche tra : m_TrancheList) {
			ibSum += tra.getIBBalance();
		}
		return ibSum;
	}

	/**
	 * obtain the sum of initial BTC start funding in all tranches
	 * @return the aggregate amount of starting BTC funding
	 */
	public double getBTCStartTotal() {
		double btcSum = 0.0;
		for (ABCTranche tra : m_TrancheList) {
			btcSum += tra.getBTCStart();
		}
		return btcSum;
	}

	/**
	 * obtain the sum of additional BTC funding in all tranches
	 * @return the aggregate amount of additional BTC funding
	 */
	public double getBTCAdditionsTotal() {
		double btcSum = 0.0;
		for (ABCTranche tra : m_TrancheList) {
			btcSum += tra.getBTCAdditions();
		}
		return btcSum;
	}

	/**
	 * finalize object when garbage-collected, for data security
	 */
	@Override
	protected void finalize() throws Throwable {
		try {
			// nix sensitive data if there is any
			if (m_TrancheList != null && !m_TrancheList.isEmpty()) {
				m_TrancheList.clear();
			}
			if (m_SelectButtons != null && !m_SelectButtons.isEmpty()) {
				m_SelectButtons.clear();
			}
		} finally {
			super.finalize();
		}
	}

	// END methods
}
