/*
 * last modified---
 * 	08-22-16/jet/ add m_AvailVFQ and methods
 * 	08-16-16/jet/ add Hist menu bar item and action
 * 	04-30-15/jet/ Spark plugin version
 * 	04-14-15/jet/ modify countdown bar mechanism; add localUTCdate()
 * 	02-06-15/jet/ add count-down bar for next toss
 * 	01-16-15/jet/ new
 *
 * purpose---
 * 	provide a UI panel for displaying the SVM screen tab once a connection
 * 	to the ABC game has been established
 */

package org.silentvault.spark.ui.svm.abc;

import javax.swing.*;
import java.awt.GridLayout;
import java.awt.Font;
import java.awt.Color;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;
import org.silentvault.spark.ui.svm.*;
import org.silentvault.spark.ui.wallet.HomePane;

import java.util.Timer;
import java.util.TimerTask;
import java.util.ArrayList;
import java.util.Date;
import java.util.Calendar;


/**
 * Extends MHomePane to provide a screen for displaying the configuration of a
 * connected ABC, and presenting interface elements to examine those contents.
 * This class also manages configuration data obtained from the SVM during the
 * connection / discovery phase, but that discovery step will be handled by
 * org.silentvault.spark.ui.svm.MHomePane before we ever get into this class.
 *
 * @author Justin Turrell
 */
public final class AHomePane extends MHomePane {
	// BEGIN extended data members
	/**
	 * a toolbar for the 6 standard menu buttons
	 */
	private JToolBar				m_MenuToolbar;

	/**
	 * action to return to the HoldingsPane
	 */
	private AbstractAction		m_HomeAction;

	/**
	 * action to move to the TransfersPane
	 */
	private AbstractAction		m_TransfersAction;

	/**
	 * action to move to the ConfigPane
	 */
	private AbstractAction		m_ConfigAction;

	/**
	 * action to move to the PlayPane
	 */
	private AbstractAction		m_PlayAction;

	/**
	 * action to move to the MarketPane
	 */
	private AbstractAction		m_MarketAction;

	/**
	 * action to move to the StatsPane
	 */
	private AbstractAction		m_StatsAction;

	/**
	 * action to move to HistPane
	 */
	private AbstractAction		m_HistAction;

	/**
	 * the table model for the tranches table, which backs m_TranchesTable
	 */
	private TrancheTableModel	m_TranchesModel;

	/**
	 * our bettor Id assigned, if we have one (else 0)
	 */
	private int						m_BettorId;

	/**
	 * the total BTC in account, equal to free SBC plus BTC in tranches
	 */
	private double					m_AccountBTC;

	/**
	 * the quantity of BTC (SBC) not in tranches and available for withdrawal
	 * or deployment to tranches
	 */
	private double					m_AvailSBC;

	/**
	 * the quantity of VFQ not in tranches and available for buying IB
	 */
	private double					m_AvailVFQ;

	/**
	 * the total quantity of IB held in the account (in tranches)
	 */
	private double					m_AccountIB;

	/**
	 * default label font
	 */
	public Font						M_LabelFont;

	/**
	 * count-down bar to next toss
	 */
	private JProgressBar			m_TossCountdown;

	/**
	 * timer to trigger toss countdown updates
	 */
	private Timer					m_CountdownTimer;

	/**
	 * inner class to implement updates for toss countdown
	 */
	protected class CountdownUpdater extends TimerTask {
		// method to extend TimerTask
		public void run() {
			if (m_TossCountdown == null) {
				Log.error("No progress bar to manipulate");
				return;
			}
			// subtract 1 (second) from the progress bar's value
			int cur = m_TossCountdown.getValue();
			int upd = Math.max(0, cur - 1);
			m_TossCountdown.setValue(upd);
			m_TossCountdown.setString("Seconds until next toss: " + upd);
			if (upd == 0) {
				// stop the timer; we need to call initCountdown() again to reset
				m_CountdownTimer.cancel();
				ATabManager atm = (ATabManager) m_TabManager;
				int nToss = atm.getStatsPane().getNextToss();
				m_TossCountdown.setString("Toss " + nToss + " in progress.");
				m_TossCountdown.setToolTipText("<html>After countdown resets, use "
														+ "Refresh button<br/>on Holdings "
														+ "or Market screens to refresh "
														+ "data.</html>");
				m_TossCountdown.setForeground(Color.DARK_GRAY.brighter());
			}
			else if (upd <= 10) {
				// set red color
				m_TossCountdown.setForeground(Color.RED);
			}
			else if (upd <= 30) {
				// set yellow color
				m_TossCountdown.setForeground(Color.ORANGE.darker());
			}
			else {
				// set normal color
				m_TossCountdown.setForeground(Color.GREEN.darker());
			}
		}
	}

	/**
	 * toss countdown task
	 */
	private CountdownUpdater	m_CountdownUpdate;

	// END extended data members

	// BEGIN methods
	/**
	 * constructor
	 * @param client the application which owns us
	 * @param mgr the tab manager which created us
	 */
	public AHomePane(WalletClient client, MTabManager mgr) {
		super(client, mgr);
		m_Plugin = client;
		M_LabelFont = new Font(Font.SANS_SERIF, Font.BOLD, 13);

		// create (empty) tranches data model
		m_TranchesModel
			= new TrancheTableModel(new ArrayList<ABCTranche>(), m_Plugin);

		// create toolbar actions, displayed on all screens
		setHomeAction();
		setTransfersAction();
		setConfigAction();
		setPlayAction();
		setMarketAction();
		setStatsAction();
		setHistAction();

		// populate the toolbar
		final Font buttonFont = new Font(Font.SANS_SERIF, Font.BOLD, 13);
		m_MenuToolbar = new JToolBar(SwingConstants.HORIZONTAL);
		final Dimension sepSz = new Dimension(10, 30);
		m_MenuToolbar.setFloatable(false);
		m_MenuToolbar.setRollover(true);
		m_MenuToolbar.setOpaque(true);
		m_MenuToolbar.setBorderPainted(false);
		m_MenuToolbar.setBackground(new Color(0xD2D2D2));

		// add button for return to HoldingsPane (main menu)
		JButton goHome = m_MenuToolbar.add(m_HomeAction);
		goHome.setToolTipText("Return to ABC Holdings");
		goHome.setFont(buttonFont);
		goHome.setRolloverEnabled(true);
		m_MenuToolbar.addSeparator(sepSz);

		// add button for transfer to Transfers
		JButton transfers = m_MenuToolbar.add(m_TransfersAction);
		transfers.setToolTipText("Transfer value to or from your wallet");
		transfers.setFont(buttonFont);
		transfers.setRolloverEnabled(true);
		m_MenuToolbar.add(transfers);
		m_MenuToolbar.addSeparator(sepSz);

		// add button for transfer to Config
		JButton goConfig = m_MenuToolbar.add(m_ConfigAction);
		goConfig.setToolTipText("Configure your account settings");
		goConfig.setFont(buttonFont);
		goConfig.setRolloverEnabled(true);
		m_MenuToolbar.addSeparator(sepSz);

		// add button for transfer to Play
		JButton goPlay = m_MenuToolbar.add(m_PlayAction);
		goPlay.setToolTipText("Start or stop game play");
		goPlay.setFont(buttonFont);
		goPlay.setRolloverEnabled(true);
		m_MenuToolbar.addSeparator(sepSz);

		// add button for transfer to Market
		JButton goMarket = m_MenuToolbar.add(m_MarketAction);
		goMarket.setToolTipText("View status of your market orders");
		goMarket.setFont(buttonFont);
		goMarket.setRolloverEnabled(true);
		m_MenuToolbar.addSeparator(sepSz);

		// add button for transfer to Stats
		JButton goStats = m_MenuToolbar.add(m_StatsAction);
		goStats.setToolTipText("View global game statistics");
		goStats.setFont(buttonFont);
		goStats.setRolloverEnabled(true);
		m_MenuToolbar.addSeparator(sepSz);

		// add button for transfer to Hist
		JButton goHist = m_MenuToolbar.add(m_HistAction);
		goHist.setToolTipText("View your account history");
		goHist.setFont(buttonFont);
		goHist.setRolloverEnabled(true);
		m_MenuToolbar.add(Box.createHorizontalStrut(10));

		// install the 7-button toolbar into the work bar of the AHomePane
		m_WorkBar.removeAll();
		m_WorkBar.add(Box.createHorizontalGlue());
		m_WorkBar.add(m_MenuToolbar);
		m_WorkBar.add(Box.createHorizontalGlue());
		m_WorkBar.setVisible(true);
	}

	/**
	 * helper method to fully initialize the main home pane (which goes in the
	 * scroll pane which we actually display)
	 */
	@Override
	public void buildHomePane() {
		GridBagLayout gbl = new GridBagLayout();
		m_HomePane = new JPanel(gbl);
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.weighty = 1.0;
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		if (m_TossCountdown != null) {
			gbc.gridheight = 5;
		}
		else {
			gbc.gridheight = 4;
		}
		gbc.anchor = GridBagConstraints.CENTER;
		gbc.fill = GridBagConstraints.NONE;
		m_HomePane.add(Box.createVerticalGlue(), gbc);
		m_HomePane.add(m_WorkBar, gbc);
		m_HomePane.add(m_WorkPane, gbc);
		if (m_TossCountdown != null) {
			m_HomePane.add(m_TossCountdown, gbc);
		}
		m_HomePane.add(Box.createVerticalGlue(), gbc);
		m_HomePane.setBorder(null);
		// use teal background frame
		m_HomePane.setBackground(new Color(0x008080));
	}

	/**
	 * initialize the toss countdown
	 * @param maxVal the maximum value the bar can have (in seconds)
	 * @param curVal the starting value (between 0 and maxVal)
	 */
	public void initCountdown(int maxVal, int curVal) {
		// server may return next interval if currently in mid-toss
		if (maxVal <= 0 || curVal > maxVal) {
			if (curVal > (maxVal * 2)) {
				Log.error("Illegal countdown params, " + maxVal + ", " + curVal);
			}
			curVal = maxVal;
		}
		boolean doingInit = false;
		if (m_TossCountdown == null) {
			m_TossCountdown = new JProgressBar(0, maxVal);
			doingInit = true;
			m_TossCountdown.setBorderPainted(true);
			m_TossCountdown.setStringPainted(true);
			m_TossCountdown.setForeground(Color.DARK_GRAY);
			m_TossCountdown.setPreferredSize(new Dimension(490, 20));
		}
		m_TossCountdown.setValue(curVal);
		m_TossCountdown.setString("Seconds until next toss: " + curVal);
		m_TossCountdown.setToolTipText("");
		if (m_CountdownTimer != null) {
			m_CountdownTimer.cancel();
		}
		m_CountdownTimer = new Timer("countdown");
		m_CountdownUpdate = new CountdownUpdater();
		m_CountdownTimer.scheduleAtFixedRate(m_CountdownUpdate, 1000L, 1000L);
		if (doingInit) {
			buildHomePane();
		}
		revalidate();
	}

	/**
	 * stop the toss countdown
	 */
	public void stopCountdown() {
		if (m_CountdownTimer != null) {
			m_CountdownTimer.cancel();
		}
		m_CountdownTimer = null;
		m_CountdownUpdate = null;
	}


	// set methods
	/**
	 * set the contents of the work pane (called by ATabManager)
	 * @param workPane a subclass object of WorkPane
	 */
	@Override
	public void setWorkPane(WorkPane workPane) {
		if (workPane != null) {
			workPane.prepDisplay();
			m_WorkPane.removeAll();
			m_WorkPane.add(workPane);
		}

		/* certain objects in our home pane might have been changed by the call
		 * to the subclass's prepDisplay()
		 */
		m_HomePane.revalidate();
		m_ScrollPane.setViewportView(m_HomePane);
	}

	/**
	 * utility method to establish the Action which moves to the HoldingsPane
	 */
	private void setHomeAction() {
		m_HomeAction = new AbstractAction("Holdings") {
			public void actionPerformed(ActionEvent ae) {
				// switch to HoldingsPane
				m_TabManager.queueTransition(
													new Integer(MTabManager.M_GO_MAIN_MENU));
				m_TabManager.makeTransition();
			}
		};
	}

	/**
	 * utility method to establish the Action which moves to the TransfersPane
	 */
	private void setTransfersAction() {
		m_TransfersAction = new AbstractAction("Transfers") {
			public void actionPerformed(ActionEvent ae) {
				// switch to TransfersPane
				m_TabManager.queueTransition(
													new Integer(ATabManager.M_GO_TRANSFERS));
				m_TabManager.makeTransition();
			}
		};
	}

	/**
	 * utility method to establish the Action which moves to the ConfigPane
	 */
	private void setConfigAction() {
		m_ConfigAction = new AbstractAction("Config") {
			public void actionPerformed(ActionEvent ae) {
				// switch to ConfigPane
				m_TabManager.queueTransition(new Integer(ATabManager.M_GO_CONFIG));
				m_TabManager.makeTransition();
			}
		};
	}

	/**
	 * utility method to establish the Action which moves to the PlayPane
	 */
	private void setPlayAction() {
		m_PlayAction = new AbstractAction("Play") {
			public void actionPerformed(ActionEvent ae) {
				// switch to ConfigPane
				m_TabManager.queueTransition(new Integer(ATabManager.M_GO_PLAY));
				m_TabManager.makeTransition();
			}
		};
	}

	/**
	 * utility method to establish the Action which moves to the MarketPane
	 */
	private void setMarketAction() {
		m_MarketAction = new AbstractAction("Market") {
			public void actionPerformed(ActionEvent ae) {
				// switch to MarketPane
				m_TabManager.queueTransition(new Integer(ATabManager.M_GO_MARKET));
				m_TabManager.makeTransition();
			}
		};
	}

	/**
	 * utility method to establish the Action which moves to the StatsPane
	 */
	private void setStatsAction() {
		m_StatsAction = new AbstractAction("Stats") {
			public void actionPerformed(ActionEvent ae) {
				// switch to StatsPane
				m_TabManager.queueTransition(new Integer(ATabManager.M_GO_STATS));
				m_TabManager.makeTransition();
			}
		};
	}

	/**
	 * utility method to establish the Action which moves to the HistPane
	 */
	private void setHistAction() {
		m_HistAction = new AbstractAction("Hist") {
			public void actionPerformed(ActionEvent ae) {
				// tell it the mode and switch to HistoryPane
				ATabManager atm = (ATabManager) m_TabManager;
				HistoryPane hPane = atm.getHistoryPane();
				hPane.setHistMode(true);
				hPane.setTrancheId(-1);
				m_TabManager.queueTransition(new Integer(ATabManager.M_GO_HIST));
				m_TabManager.makeTransition();
			}
		};
	}

	/**
	 * set the total BTC held by the account
	 * @param btc the total of free SBC and BTC deployed to tranches
	 */
	public void setAccountBTC(double btc) {
		if (btc >= 0.0) {
			m_AccountBTC = btc;
		}
	}

	/**
	 * set the free SBC held by the account
	 * @param sbc the amount of SBC available and not yet deployed to tranches
	 */
	public void setAvailSBC(double sbc) {
		if (sbc >= 0.0) {
			m_AvailSBC = sbc;
		}
	}

	/**
	 * set the VFQ held by the account
	 * @param vfq the amount of VFQ available to buy IB via bleed-ins
	 */
	public void setAvailVFQ(double vfq) {
		if (vfq >= 0.0) {
			m_AvailVFQ = vfq;
		}
	}

	/**
	 * set the total IB held by the account
	 * @param ib the total of IB deployed in tranches
	 */
	public void setAccountIB(double ib) {
		if (ib >= 0.0) {
			m_AccountIB = ib;
		}
	}

	/**
	 * set the bettor ID
	 * @param bId the ID to be set (can be zero)
	 */
	public void setBettorId(int bId) {
		if (bId >= 0) {
			m_BettorId = bId;
		}
	}


	// get methods
	/**
	 * obtain the data model which underlies tranches table
	 * @return the model
	 */
	public TrancheTableModel getTrancheModel() { return m_TranchesModel; }

	/**
	 * obtain the total BTC/SBC in the account, both in tranches and free balance
	 * @return the total BTC held
	 */
	public double getAccountBTC() { return m_AccountBTC; }

	/**
	 * obtain the total IB in the account
	 * @return the sum total of all IB in all tranches
	 */
	public double getAccountIB() { return m_AccountIB; }

	/**
	 * obtain the total free SBC in the account
	 * @return the available balance not deployed to tranches
	 */
	public double getAvailSBC() { return m_AvailSBC; }

	/**
	 * obtain the total VFQ in the account
	 * @return the available balance not spent on IB bleed-ins
	 */
	public double getAvailVFQ() { return m_AvailVFQ; }

	/**
	 * obtain the bettor ID we've been assigned
	 * @return the ID, or zero if not betting
	 */
	public int getBettorId() { return m_BettorId; }

	/**
	 * obtain the countdown timer
	 * @return the progress bar used for the countdown timer
	 */
	public JProgressBar getCountdown() { return m_TossCountdown; }

	/**
	 * utility method to interpolate a date from UTC to local time
	 * @param utc the date/time in UTC
	 * @return the date/time this represents in local time
	 */
	public Date localUTCdate(Date utc) { 
		if (utc == null) {
			return new Date();
		}
		Calendar cal = Calendar.getInstance();
		int gmtOffsetMillis = cal.get(Calendar.ZONE_OFFSET);
		int dstOffsetMillis = cal.get(Calendar.DST_OFFSET);
		Date local = new Date(utc.getTime() + gmtOffsetMillis + dstOffsetMillis);
		return local;
	}

	/**
	 * finalize object when garbage-collected
	 */
	@Override
	protected void finalize() throws Throwable {
		try {
			if (m_CountdownTimer != null) {
				m_CountdownTimer.cancel();
				m_CountdownTimer = null;
			}
		} finally {
			super.finalize();
		}
	}

	// END methods
}
