/*
 * last modified---
 * 	08-22-16/jet/ add M_GO_HIST and HistoryPane
 * 	04-30-15/jet/ Spark plugin version
 * 	01-16-15/jet/ new
 *
 *	purpose---
 *		provide a manager for the contents of the marketplace tab for ABC game
 */

package org.silentvault.spark.ui.svm.abc;

import javax.swing.*;
import java.awt.Cursor;
import java.awt.Color;
import java.awt.Font;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Desktop;

import org.jivesoftware.smack.XMPPConnection;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;
import org.silentvault.spark.ui.svm.*;

import java.util.Hashtable;
import java.util.ArrayList;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.net.URI;
import java.net.URISyntaxException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.io.IOException;


/**
 * This class is used to manage the graphical contents of the marketplace tab.
 * It handles creating and destroying
 * the embedded pane subclasses and manages state transitions between them.
 * It also provides utility functions such as error popups and wait cursors.
 *
 * This class should be overridden for client-side support for a given SVM.
 *
 * @author Justin Turrell
 */
public class ATabManager extends MTabManager {
	// BEGIN data members
	// states defined in MTabManager base class
	//public static final int	M_LOGIN = 1;
	//public static final int	M_HOME_INIT = 2;
	//public static final int	M_GO_MAIN_MENU = 3;
	//public static final int	M_TUTORIALS = 4;
	//public static final int	M_LOGOUT_MARKET = 5;

	// other states specific to our derived class
	/**
	 * go to the Transfers screen
	 */
	public static final int		M_GO_TRANSFERS = 6;

	/**
	 * go to the Config screen
	 */
	public static final int		M_GO_CONFIG = 7;

	/**
	 * go to the Play screen
	 */
	public static final int		M_GO_PLAY = 8;

	/**
	 * go to the Market screen
	 */
	public static final int		M_GO_MARKET = 9;

	/**
	 * go to the Stats screen
	 */
	public static final int		M_GO_STATS = 10;

	/**
	 * go to the History screen
	 */
	public static final int		M_GO_HIST = 11;


	// display panes we control
	/**
	 * the Home pane, used after connect and disco have occurred
	 */
	private AHomePane			m_HomePane;

	/**
	 * the Holdings pane, used to show totals and tranches
	 */
	private HoldingsPane		m_HoldingsPane;

	/**
	 * the Transfers pane, used to move SBC to and from the wallet
	 */
	private TransfersPane	m_TransfersPane;

	/**
	 * the Configuration pane, used to show and update account settings
	 */
	private ConfigPane		m_ConfigPane;

	/**
	 * the Play pane, used to start and stop betting
	 */
	private PlayPane			m_PlayPane;

	/**
	 * the Market pane, used to show IB orders for tranches
	 */
	private MarketPane		m_MarketPane;

	/**
	 * the Stats pane, used to report global game statistics
	 */
	private StatsPane			m_StatsPane;

	/**
	 * the History pane, used to select account and tranche history for graphing
	 */
	private HistoryPane		m_HistoryPane;

	// END data members
	
	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the containing plugin
	 */
	public ATabManager(WalletClient plugin) {
		super(plugin);
		m_HomePane = new AHomePane(m_Plugin, this);
	}

	/**
	 * state transition method, called by threads that remove states from queue
	 * @param state the state to transition to
	 */
	@Override
	public void doState(int state) {
		//Log.debug("ATabManager state transition to " + state);
		switch (state) {
			case M_LOGIN:
				removeAll();
				add(m_LoginPane);
				m_LoginPane.prepDisplay();
				break;

			case M_HOME_INIT:
				// get rid of login pane and tab containing it
	 			removeAll();
				// make a default home pane and recreate the tab using it
				m_HomePane.updateTitleText();
				add(m_HomePane);

				// prepare all the WorkPane subclass objects
				m_HoldingsPane = new HoldingsPane(m_Plugin, m_HomePane, this);
				m_TransfersPane = new TransfersPane(m_Plugin, m_HomePane, this);
				// the TransfersPane wants to receive escrow confirmations
				m_EscrowHandler = m_TransfersPane;
				m_ConfigPane = new ConfigPane(m_Plugin, m_HomePane, this);
				m_PlayPane = new PlayPane(m_Plugin, m_HomePane, this);
				m_MarketPane = new MarketPane(m_Plugin, m_HomePane, this);
				m_StatsPane = new StatsPane(m_Plugin, m_HomePane, this);
				m_HistoryPane = new HistoryPane(m_Plugin, m_HomePane, this);

				// synchronously request balances, tranches, stats, and config data
				queryAcctData(true);
				break;

			case M_GO_MAIN_MENU:
				m_HomePane.setWorkPane(m_HoldingsPane);
				break;

			case M_TUTORIALS:
				// get the desktop browser for this system and try to open the URI
				URI tutURI = null;
				boolean needManual = false;
				try {
					// TBD: ABC-related tutorials
					tutURI = new URI("https://www.voucher-safe.com");
				}
				catch (URISyntaxException use) {
					Log.error("Bad URI for tutorials");
					needManual = true;
				}
				Desktop browser = null;
				try {
					browser = Desktop.getDesktop();
				}
				catch (UnsupportedOperationException uoe) {
					needManual = true;
				}
				finally {
					if (!needManual) {
						if (!browser.isSupported(Desktop.Action.BROWSE)) {
							needManual = true;
						}
					}
				}
				if (needManual) {
					// just put up the URL in a popup window with instructions
					JFrame tutPop = new JFrame("ABC Tutorials");
					tutPop.setDefaultCloseOperation(
															WindowConstants.DISPOSE_ON_CLOSE);
					JEditorPane tutViewer = new JEditorPane("text/html", null);
					tutViewer.setBackground(Color.LIGHT_GRAY);
					StringBuilder content = new StringBuilder(512);
					content.append("<html><body bgcolor=\"silver\">");
					content.append("<center><br/><br/><p>");
					content.append("For tutorials help, paste the following URL ");
					content.append("into your browser:</p><br/><br/>");
					content.append("https://www.voucher-safe.com<br/>");
					content.append("</center></body></html>");
					tutViewer.setText(content.toString());
					tutPop.add(tutViewer);
					tutPop.setSize(new Dimension(400, 200));
					tutPop.setLocationRelativeTo(this);
					tutPop.setVisible(true);
					m_Plugin.registerPopup(tutPop);
				}
				else {
					try {
						// actually invoke the default browser
						browser.browse(tutURI);
					}
					catch (IOException ioe) {
						Log.error("Could not invoke browser on tutorials page", ioe);
						showError("Could not display tutorials",
									"Error opening default browser",
								"Point your browser to: https://www.voucher-safe.com");
					}
				}
				break;

			case M_LOGOUT_MARKET:
				// kill the coundown timer in the HomePane
				m_HomePane.stopCountdown();

				String sessId = m_LoginPane.getSessionId();
				if (sessId.isEmpty()) {
					// we weren't logged in, so just move back to login pane
					Runnable goLogin = new Runnable() {
						public void run() {
							doState(M_LOGIN);
						}
					};
					SwingUtilities.invokeLater(goLogin);
				}
				else {
					VsSecrets sekrits = m_Plugin.getLoginSecrets();
					// access the LoginPane's method for doing this
					m_LoginPane.sendLogout(sekrits.getSVMId());
					/* NB: on reply, we'll switch back to generic listener and
					 * 	 tab manager, and go to generic login pane.
					 */
				}
				break;

			case M_GO_TRANSFERS:
				m_HomePane.setWorkPane(m_TransfersPane);
				break;

			case M_GO_CONFIG:
				// synchronously request config settings
				m_ConfigPane.queryAcctConfig(true);

				// advance to configuration pane
				m_HomePane.setWorkPane(m_ConfigPane);
				break;

			case M_GO_PLAY:
				// advance to play pane
				m_HomePane.setWorkPane(m_PlayPane);
				break;

			case M_GO_MARKET:
				if (m_ConfigPane.getSalePolicies().isEmpty()) {
					// synchronously request config settings (need to parse policies)
					m_ConfigPane.queryAcctConfig(true);
				}

				// synchronously request order list
				m_MarketPane.queryOrders(true);

				// advance to market pane
				m_HomePane.setWorkPane(m_MarketPane);
				break;

			case M_GO_STATS:
				// advance to stats pane
				m_HomePane.setWorkPane(m_StatsPane);
				break;

			case M_GO_HIST:
				// synchronously request config settings, to setup range sliders
				m_ConfigPane.queryAcctConfig(true);

				// advance to history pane
				m_HomePane.setWorkPane(m_HistoryPane);
				break;

			default:
				Log.error("Unsupported ABC UI state " + state);
				break;
		}
		if (state != M_TUTORIALS) {
			m_State = state;
			revalidate();
			repaint();
		}
	}


	// get methods
	/**
	 * obtain the home pane
	 * @return the pane used for home safe display
	 */
	@Override
	public MHomePane getHomePane() { return m_HomePane; }

	/**
	 * obtain the work pane (a component in the home pane which contains the
	 * WorkPane subclass)
	 * @return the pane used for work area display
	 */
	@Override
	public WorkPane getWorkPane() {
		if (m_HomePane == null) {
			return null;
		}
		return m_HomePane.getWorkPane();
	}

	/**
	 * obtain the latest balances, tranche data, stats, and order list from
	 * the ABC server
	 * @param synchronous true if we should do these requests synchronously
	 */
	public void queryAcctData(boolean synchronous) {
		// double-check we're connected
		XMPPConnection conn = m_Plugin.getOFSConnection();
		if (conn == null || !conn.isConnected()) {
			Log.error("Cannot log out of SVM without OFS gateway connection");
			showError("cannot log out of SVM",
						"no OFS gateway connection",
						"connect to OFS first");
			return;
		}

		// double-check we're logged in, to both wallet and ABC account
		VscState state = m_Plugin.getVscState();
		String sessId = m_LoginPane.getSessionId();
		String abcId = m_Plugin.getLoginSecrets().getSVMId();
		if (!state.isLoggedIn() || sessId.isEmpty() || abcId.isEmpty()) {
			Log.error("Cannot obtain ABC account data while not logged in!");
			showError("cannot obtain ABC account data while not logged in");
			return;
		}

		boolean gotOk = true;
		m_HomePane.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		// need to query for all of these, in order
		if (queryBalances(synchronous)) {
			if (queryTranches(synchronous)) {
				if (queryStats(synchronous)) {
					if (!m_ConfigPane.queryAcctConfig(synchronous)) {
						gotOk = false;
					}
				}
				else {
					gotOk = false;
				}
			}
			else {
				gotOk = false;
			}
		}
		else {
			gotOk = false;
		}
		m_HomePane.setCursor(null);

		// if we didn't return early due to an error, go to HoldingsPane display
		if (gotOk) {
			Runnable goMain = new Runnable() {
				public void run() {
					doState(M_GO_MAIN_MENU);
				}
			};
			SwingUtilities.invokeLater(goMain);
		}
	}

	/**
	 * obtain latest balances
	 * @param synchronous true if we should do request synchronously
	 * @return true on success (may not mean successful reply)
	 */
	public boolean queryBalances(boolean synchronous) {
		// prepare account balance request message
		String sessId = m_LoginPane.getSessionId();
		String abcId = m_Plugin.getLoginSecrets().getSVMId();
		ABCClientBlock balBlock = new ABCClientBlock();
		balBlock.setOpcode("REQ_balances");
		balBlock.setSessionId(sessId);

		// send to server; answer will go to HoldingsPane.recordBalances()
		ABCListener abcListener = (ABCListener) m_Plugin.getSVMListener();
		if (synchronous) {
			if (!abcListener.sendRequest(balBlock)) {
				Log.error("Account balances request failed to ABC Id " + abcId);
				showError("cannot get account balances at ABC " + abcId,
							"failure sending balances request message");
				return false;
			}
		}
		else {
			if (!abcListener.sendAsyncRequest(balBlock)) {
				Log.error("Account balances request failed to ABC Id " + abcId);
				showError("cannot get account balances at ABC " + abcId,
							"failure sending balances request message");
				return false;
			}
		}
		return true;
	}

	/**
	 * obtain latest tranches list
	 * @param synchronous true if we should do request synchronously
	 * @return true on success (may not mean successful reply)
	 */
	public boolean queryTranches(boolean synchronous) {
		// prepare account tranches request message
		String sessId = m_LoginPane.getSessionId();
		String abcId = m_Plugin.getLoginSecrets().getSVMId();
		ABCClientBlock traBlock = new ABCClientBlock();
		traBlock.setOpcode("REQ_tranches");
		traBlock.setSessionId(sessId);

		// send to server; answer will go to HoldingsPane.recordTranches()
		ABCListener abcListener = (ABCListener) m_Plugin.getSVMListener();
		if (synchronous) {
			if (!abcListener.sendRequest(traBlock)) {
				Log.error("Account tranches request failed to ABC Id " + abcId);
				showError("cannot get account tranches at ABC " + abcId,
							"failure sending tranches request message");
				return false;
			}
		}
		else {
			if (!abcListener.sendAsyncRequest(traBlock)) {
				Log.error("Account tranches request failed to ABC Id " + abcId);
				showError("cannot get account tranches at ABC " + abcId,
							"failure sending tranches request message");
				return false;
			}
		}
		return true;
	}

	/**
	 * obtain latest global server stats
	 * @param synchronous true if we should do request synchronously
	 * @return true on success (may not mean successful reply)
	 */
	public boolean queryStats(boolean synchronous) {
		// prepare global stats query message
		String sessId = m_LoginPane.getSessionId();
		String abcId = m_Plugin.getLoginSecrets().getSVMId();
		ABCClientBlock statsBlock = new ABCClientBlock();
		statsBlock.setOpcode("REQ_stats");
		statsBlock.setSessionId(sessId);

		// send to server; answer will go to StatsPane.recordStats()
		ABCListener abcListener = (ABCListener) m_Plugin.getSVMListener();
		if (synchronous) {
			if (!abcListener.sendRequest(statsBlock)) {
				Log.error("Stats request failed to ABC Id " + abcId);
				showError("cannot get global stats at ABC " + abcId,
							"failure sending stats request message");
				return false;
			}
		}
		else {
			if (!abcListener.sendAsyncRequest(statsBlock)) {
				Log.error("Stats request failed to ABC Id " + abcId);
				showError("cannot get global stats at ABC " + abcId,
							"failure sending stats request message");
				return false;
			}
		}
		return true;
	}


	// get methods
	/**
	 * get the main menu pane, which is the home state after login
	 * @return the pane used for the holdings / main menu
	 */
	public HoldingsPane getHoldingsPane() { return m_HoldingsPane; }

	/**
	 * get the transfers pane, which is used to move SBC to and from the wallet
	 * @return the pane used for the transfers menu
	 */
	public TransfersPane getTransfersPane() { return m_TransfersPane; }

	/**
	 * get the config pane, which is used to show and update account settings
	 * @return the pane used for the config menu
	 */
	public ConfigPane getConfigPane() { return m_ConfigPane; }

	/**
	 * get the play pane, which is used to start and stop betting
	 * @return the pane used for betting changes
	 */
	public PlayPane getPlayPane() { return m_PlayPane; }

	/**
	 * get the market pane, which is used to show open IB buy/sell orders
	 * @return the pane used to show market orders
	 */
	public MarketPane getMarketPane() { return m_MarketPane; }

	/**
	 * get the stats pane, which is used to report global game statistics
	 * @return the pane used for server-side stats
	 */
	public StatsPane getStatsPane() { return m_StatsPane; }

	/**
	 * get the history pane, which is used to select data for graphical display
	 * @return the pane used to produce history graphs
	 */
	public HistoryPane getHistoryPane() { return m_HistoryPane; }

	/**
	 * method to deallocate all sub-panes (called by WalletClient destructor)
	 */
	@Override
	public void deallocatePanes() {
		if (m_LoginPane != null) {
			m_LoginPane.removeAll();
			m_LoginPane = null;
		}
		if (m_HomePane != null) {
			m_HomePane.removeAll();
			m_HomePane = null;
		}
		if (m_HoldingsPane != null) {
			m_HoldingsPane.removeAll();
			m_HoldingsPane = null;
		}
		if (m_TransfersPane != null) {
			m_TransfersPane.removeAll();
			m_TransfersPane = null;
		}
		if (m_ConfigPane != null) {
			m_ConfigPane.removeAll();
			m_ConfigPane = null;
		}
		if (m_PlayPane != null) {
			m_PlayPane.removeAll();
			m_PlayPane = null;
		}
		if (m_MarketPane != null) {
			m_MarketPane.removeAll();
			m_MarketPane = null;
		}
		if (m_StatsPane != null) {
			m_StatsPane.removeAll();
			m_StatsPane = null;
		}
		if (m_HistoryPane != null) {
			m_HistoryPane.removeAll();
			m_HistoryPane = null;
		}
	}

	// END methods
}
