/*
 * last modified---
 * 	10-29-16/jet/ pass queued deployments to PlayPane
 * 	09-16-16/jet/ add tooltip warnings about VFQ bonus eligibility
 * 	09-07-16/jet/ use info stack to show config changes separately
 * 	08-25-16/jet/ other mods for valuation computation, wholeness, VFQ, etc.
 * 	08-06-16/jet/ add useVFQ to REQ_config_details (pass flag from UI later)
 * 	04-30-15/jet/ Spark plugin version
 * 	04-29-15/jet/ allow for lengthy (real) policy list, drop-downs on own line
 * 	04-21-15/jet/ add Status line
 * 	03-12-15/jet/ add queryAcctConfig(), sendConfigMods()
 * 	01-30-15/jet/ new
 *
 * purpose---
 * 	provide a UI screen to represent ABC account configuration
 */

package org.silentvault.spark.ui.svm.abc;

import javax.swing.*;
import javax.swing.border.EmptyBorder;
import java.awt.event.ActionEvent;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Font;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;
import org.silentvault.spark.ui.svm.*;

import java.util.Date;
import java.util.ArrayList;
import java.util.Locale;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.text.ParsePosition;
import javax.management.timer.Timer;


/**
 * Extends AWorkPane to provide a screen for the ABC configuration settings and
 * updates to those settings.
 *
 * @author Justin Turrell
 */
public final class ConfigPane extends AWorkPane {
	// BEGIN data members
	/**
	 * action to Update settings with changes
	 */
	private AbstractAction			m_UpdateAction;

	/**
	 * button for update action
	 */
	private JButton					m_UpdateButton;

	/**
	 * the panel containing the config information
	 */
	private JPanel						m_ConfigPanel;

	/**
	 * flag indicating active betting
	 */
	private boolean					m_Active;

	/**
	 * total amount (in BTC) originally deployed to tranches, for both BTC and IB
	 */
	private double						m_Bankroll;

	/**
	 * total additions (in BTC) to bankroll since betting started
	 */
	private double						m_Additions;

	/**
	 * total subtractions (in BTC) from bankroll since betting started
	 */
	private double						m_Subtractions;

	/**
	 * initial BTC chip total across all tranches (at start betting)
	 */
	private double						m_BTCChipStart;

	/**
	 * configured tranche size (in BTC)
	 */
	private float						m_TrancheSize;

	/**
	 * first toss (on which betting began, if it did)
	 */
	private int							m_FirstToss;

	/**
	 * last toss (on which betting ceased, if it did)
	 */
	private int							m_LastToss;

	/**
	 * high value cash-out calculation (in BTC)
	 */
	private double						m_HighValue;

	/**
	 * toss on which high value was seen
	 */
	private int							m_HighToss;

	/**
	 * low value cash-out calculation (in BTC)
	 */
	private double						m_LowValue;

	/**
	 * toss on which low value was seen
	 */
	private int							m_LowToss;

	/**
	 * number of tosses separating wholeness awards, for this account
	 * (may differ from system default reported in REP_stats)
	 */
	private int							m_WholenessTosses;

	/**
	 * toss on which account was last checked for wholeness bonus (0 if never)
	 */
	private int							m_LastWhChkToss;

	/**
	 * configured profit target (in percent)
	 */
	private float						m_ProfitTarget;

	/**
	 * configured stop loss target (in percent)
	 */
	private float						m_StopLoss;

	/**
	 * configured IB sale discount policy
	 */
	private int							m_SalePolicy;

	/**
	 * configured IB purchase discount policy
	 */
	private int							m_PurchPolicy;

	/**
	 * configured VFQ application policy (true if VFQ balance used to pay for IB)
	 */
	private boolean					m_VFQPolicy;

	/**
	 * date formatter for parsing tosses
	 */
	private SimpleDateFormat		m_TossFormatter;

	/**
	 * date formatter for human-readable toss display
	 */
	private SimpleDateFormat		m_DateFormatter;

	/**
	 * storage for all available sale discount policies
	 */
	private ArrayList<DiscountPolicy>	m_SDiscPolicies;

	/**
	 * storage for all available sale discount policies
	 */
	private ArrayList<DiscountPolicy>	m_PDiscPolicies;

	/**
	 * UI element for entering new profit target value
	 */
	private JTextField				m_NewProfitTarget;

	/**
	 * UI element for entering new stop loss value
	 */
	private JTextField				m_NewStopLoss;

	/**
	 * UI element for entering new tranche size value
	 */
	private JTextField				m_NewTrancheSize;

	/**
	 * UI element for selecting sale discount policies
	 */
	private JComboBox<String>		m_SaleDiscDropdown;

	/**
	 * UI element for selecting purchase discount policies
	 */
	private JComboBox<String>		m_PurchDiscDropdown;

	/**
	 * UI element for toggling VFQ bleed-in
	 */
	private JCheckBox					m_UseVFQ;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the AHomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public ConfigPane(WalletClient plugin, AHomePane home, ATabManager mgr) {
		super(plugin, home, mgr);
		m_InsetBorder = new EmptyBorder(3, 5, 3, 5);

		// define instructions
		String configTxt = "This panel shows various statistics about your ABC "
								+ "account and its configuration and results.<br/><br/>"
								+ "You may make changes to the items in the bottom "
								+ "section.";
		m_UserInstructs = configTxt;
		m_TossFormatter = new SimpleDateFormat("yyMMddHHmm", Locale.US);
		m_DateFormatter = new SimpleDateFormat("yyyy-MMM-dd HH:mm", Locale.US);
		m_SDiscPolicies = new ArrayList<DiscountPolicy>();
		m_PDiscPolicies = new ArrayList<DiscountPolicy>();

		// define rebalance action and button
		setUpdateAction();
		m_UpdateButton = new JButton(m_UpdateAction);
		m_UpdateButton.setFont(m_TabManager.M_ButtonFont);
		m_UpdateButton.setToolTipText("Submit any changes you have made above");
		m_UpdateButton.setEnabled(false);

		// popuplate the button pane
		m_ButtonPane.removeAll();
		m_ButtonPane.add(Box.createHorizontalStrut(320));	// spacer
		m_ButtonPane.add(m_UpdateButton);

		// init main panel area
		m_ConfigPanel = new JPanel();
		m_ConfigPanel.setOpaque(true);
		m_ConfigPanel.setBorder(null);
		m_ConfigPanel.setLayout(new GridBagLayout());
		m_ConfigPanel.setPreferredSize(new Dimension(460, 680));
		m_ConfigPanel.setBackground(Color.CYAN.darker());

		// init profit target entry
		Dimension entrySz = new Dimension(60, 20);
		m_NewProfitTarget = new JTextField(6);
		m_NewProfitTarget.setDropTarget(null);
		m_NewProfitTarget.setHorizontalAlignment(JTextField.LEFT);
		m_NewProfitTarget.setToolTipText("Specify new profit target "
												+ "(0 with Stop also 0 means none)");
		m_NewProfitTarget.setPreferredSize(entrySz);

		// init stop loss entry
		m_NewStopLoss = new JTextField(6);
		m_NewStopLoss.setDropTarget(null);
		m_NewStopLoss.setHorizontalAlignment(JTextField.LEFT);
		m_NewStopLoss.setToolTipText("Specify new stop loss level "
											+ "(0 with Target also 0 means none)");
		m_NewStopLoss.setPreferredSize(entrySz);

		// init tranche size entry
		m_NewTrancheSize = new JTextField(6);
		m_NewTrancheSize.setDropTarget(null);
		m_NewTrancheSize.setHorizontalAlignment(JTextField.LEFT);
		m_NewTrancheSize.setToolTipText(
											"Specify new tranche size and click Update");
		m_NewTrancheSize.setPreferredSize(entrySz);

		// init sale discount policy drop-down list
		Dimension polSz = new Dimension(450, 25);
		m_SaleDiscDropdown = new JComboBox<String>();
		m_SaleDiscDropdown.setDropTarget(null);
		m_SaleDiscDropdown.setPreferredSize(polSz);
		m_SaleDiscDropdown.setToolTipText(
							"Select a new sale policy from the list and click Update");

		// init purchase discount policy drop-down list
		m_PurchDiscDropdown = new JComboBox<String>();
		m_PurchDiscDropdown.setDropTarget(null);
		m_PurchDiscDropdown.setPreferredSize(polSz);
		m_PurchDiscDropdown.setToolTipText(
					"Select a new purchase policy from the list and click Update");

		// init VFQ bleed-in checkbox
		m_UseVFQ = new JCheckBox("Utilize VFQ", false);
		m_UseVFQ.setToolTipText("<html><b>Note:</b> if you set this flag while "
										+ "betting, then to be eligible <br>for VFQ "
										+ "bonuses when you stop betting, you must either"
										+ ":<br>1. Bet continuously for 4000 tosses "
										+ "(about 333 hours), -or-<br>2. Achieve a net "
										+ "profit level of +38.2%.<br>"
										+ "(You must also reach one or the other "
										+ "criterion before<br>you can turn this "
										+ "setting off, while playing.)</html>");
	}

	/**
	 * do display of configuration
	 */
	public void prepDisplay() {
		// set screen label
		buildScreenLabel("Configuration for your account");
		AHomePane ahp = (AHomePane) m_ParentHome;

		// init config info panel
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(3);
		m_ConfigPanel.removeAll();
		m_GBC.anchor = GridBagConstraints.NORTH;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_GBC.fill = GridBagConstraints.HORIZONTAL;
		m_GBC.gridheight = 32;
		m_ConfigPanel.add(Box.createVerticalGlue(), m_GBC);

		// row 1: begin with whether account is betting
		JLabel actLab = new JLabel("Actively betting: ");
		actLab.setFont(ahp.M_LabelFont);
		actLab.setHorizontalAlignment(SwingConstants.TRAILING);
		actLab.setBorder(m_InsetBorder);
		JTextField actFld = new JTextField(3);
		actFld.setDropTarget(null);
		actFld.setEditable(false);
		actFld.setOpaque(false);
		actFld.setHorizontalAlignment(JTextField.LEFT);
		actFld.setBorder(m_InsetBorder);
		actFld.setBackground(m_Plugin.m_LabelBackColor);
		actFld.setFocusable(false);
		if (m_Active) {
			actFld.setText("Yes");
			actFld.setToolTipText("Your account is currently betting");
		}
		else {
			actFld.setText("No");
			actFld.setToolTipText("Your account is not betting");
		}
		actLab.setLabelFor(actFld);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(actLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(actFld, m_GBC);

		// row 2: bettor ID
		JLabel betLab = new JLabel("Bettor ID: ");
		betLab.setFont(ahp.M_LabelFont);
		betLab.setHorizontalAlignment(SwingConstants.TRAILING);
		betLab.setBorder(m_InsetBorder);
		JTextField bId = new JTextField(8);
		bId.setDropTarget(null);
		bId.setEditable(false);
		bId.setOpaque(false);
		bId.setHorizontalAlignment(JTextField.LEFT);
		bId.setBorder(m_InsetBorder);
		bId.setBackground(m_Plugin.m_LabelBackColor);
		bId.setFocusable(false);
		Integer betId = new Integer(ahp.getBettorId());
		bId.setText(betId.toString());
		if (betId.intValue() == 0) {
			bId.setToolTipText("Your account is not currently betting");
		}
		else {
			bId.setToolTipText("Bettor ID currently assigned to your account");
		}
		betLab.setLabelFor(bId);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(betLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(bId, m_GBC);

		// compute current value and position, for use below
		double position = 0.0;
		double btcNow = 0.0;
		double ibNow = 0.0;
		double valNow = 0.0;
		double bankNet = 0.0;
		if (m_Bankroll > 0.0) {
			bankNet = m_Bankroll + m_Additions - m_Subtractions;
		}
		TrancheTableModel tModel = ahp.getTrancheModel();
		int trancheCnt = tModel.getRowCount();
		ATabManager atm = (ATabManager) m_TabManager;
		StatsPane sp = atm.getStatsPane();
		if (trancheCnt > 0) {
			// compute current valuation
			btcNow = tModel.getBTCTotal();
			ibNow = tModel.getIBTotal();
			// take into account actual IB reserve valuation
			double ibVal = ibNow * sp.getIBRatio();
			valNow = btcNow + ibVal;
			position = (valNow / bankNet) - 1.0;
		}
		else {
			// after betting, examine subtractions as % of all that was put in
			double sum = m_Bankroll + m_Additions;
			if (sum > 0.0) {
				position = m_Subtractions / sum - 1.0;
			}
		}
		double posDiff = 0.0;
		if (m_Active) {
			posDiff = valNow - bankNet;
		}
		else {
			// NB: whatever was cashed out counts as a subtraction now
			posDiff = -bankNet;
		}

		// row 3: results status line
		JLabel statLab = new JLabel("Status: ");
		statLab.setFont(ahp.M_LabelFont);
		statLab.setHorizontalAlignment(SwingConstants.TRAILING);
		statLab.setBorder(m_InsetBorder);
		statLab.setToolTipText("Summary of results since you last started play");
		JTextField statFld = new JTextField(16);
		statFld.setDropTarget(null);
		statFld.setEditable(false);
		statFld.setOpaque(false);
		statFld.setHorizontalAlignment(JTextField.LEFT);
		statFld.setBorder(m_InsetBorder);
		statFld.setBackground(m_Plugin.m_LabelBackColor);
		statFld.setFocusable(false);
		statLab.setLabelFor(statFld);
		// determine proper text for status field, based on other values
		StringBuffer statRes = new StringBuffer(128);
		df.setMaximumFractionDigits(2);
		df.setMinimumFractionDigits(2);
		if (m_Active) {
			statRes.append("To view performance, use Hist page");
			statFld.setToolTipText("Also see high/low values and tosses (below) "
										+ "for more details");
		}
		else {
			// try to determine what happened
			if (m_LastToss == 0) {
				statRes.append("No results yet");
				statFld.setToolTipText("Use Start Betting on Play screen to play");
			}
			else if (m_LastToss == m_LowToss) {
				// cashed out with a loss
				statRes.append("Hit stop loss of ");
				statRes.append(df.format(m_StopLoss) + "%");
				statFld.setForeground(Color.RED);
				statFld.setToolTipText("Sorry, better luck next time!");
			}
			else if (m_LastToss == m_HighToss) {
				// cashed out with a profit
				statRes.append("Hit profit target of ");
				statRes.append(df.format(m_ProfitTarget) + "%");
				statFld.setForeground(Color.GREEN.darker());
				statFld.setToolTipText("Congratulations!");
			}
			else {
				// may have cashed out manually
				statRes.append("Cashed out on toss " + m_LastToss);
				statFld.setToolTipText("Use Start Betting on Play screen to "
											+ "resume play");
			}
		}
		statFld.setText(statRes.toString());
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(statLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(statFld, m_GBC);
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);

		// row 4: bankroll deployed
		JLabel bankLab = new JLabel("Bankroll: ");
		bankLab.setFont(ahp.M_LabelFont);
		bankLab.setHorizontalAlignment(SwingConstants.TRAILING);
		bankLab.setBorder(m_InsetBorder);
		bankLab.setToolTipText("Total amount you committed at start of betting, "
									+ "plus any additions, less any subtractions");
		JTextField bankFld = new JTextField(12);
		bankFld.setDropTarget(null);
		bankFld.setEditable(false);
		bankFld.setOpaque(false);
		bankFld.setHorizontalAlignment(JTextField.LEFT);
		bankFld.setBorder(m_InsetBorder);
		bankFld.setBackground(m_Plugin.m_LabelBackColor);
		bankFld.setFocusable(false);
		df.applyPattern("####0.0000");
		if (m_Active) {
			bankFld.setText(df.format(bankNet) + " (across " + trancheCnt
								+ " tranches)");
		}
		else {
			bankFld.setText("No bankroll presently deployed");
		}
		if (trancheCnt > 0) {
			bankFld.setToolTipText("Total present value: " + df.format(valNow)
										+ " BTC");
		}
		bankLab.setLabelFor(bankFld);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(bankLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(bankFld, m_GBC);

		// row 5: net position across all tranches
		JLabel posLab = new JLabel("Profit or Loss: ");
		posLab.setFont(ahp.M_LabelFont);
		posLab.setHorizontalAlignment(SwingConstants.TRAILING);
		posLab.setBorder(m_InsetBorder);
		posLab.setToolTipText("Net return on bankroll to date");
		JTextField pos = new JTextField(12);
		pos.setDropTarget(null);
		pos.setEditable(false);
		pos.setOpaque(false);
		pos.setHorizontalAlignment(JTextField.LEFT);
		pos.setBorder(m_InsetBorder);
		pos.setBackground(m_Plugin.m_LabelBackColor);
		pos.setFocusable(false);
		String cellTxt = df.format(posDiff) + " (";
		df.applyPattern("+###0.00##%;-###0.00##%");
		cellTxt += df.format(position) + ")";
		pos.setText(cellTxt);
		if (position < 0.0) {
			pos.setForeground(Color.RED);
			if (m_Active) {
				pos.setToolTipText("Approximate net loss");
			}
			else {
				pos.setToolTipText("Approximate loss on last betting session");
			}
		}
		else {
			pos.setForeground(Color.GREEN.darker());
			if (m_Active) {
				pos.setToolTipText("Approximate net gain");
			}
			else {
				pos.setToolTipText("Approximate gain on last betting session");
			}
		}
		posLab.setLabelFor(pos);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(posLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(pos, m_GBC);

		// row 6: total BTC in all tranches
		JLabel btcLab = new JLabel("BTC total: ");
		btcLab.setFont(ahp.M_LabelFont);
		btcLab.setHorizontalAlignment(SwingConstants.TRAILING);
		btcLab.setBorder(m_InsetBorder);
		JTextField btc = new JTextField(12);
		btc.setDropTarget(null);
		btc.setEditable(false);
		btc.setOpaque(false);
		btc.setHorizontalAlignment(JTextField.LEFT);
		btc.setBorder(m_InsetBorder);
		btc.setBackground(m_Plugin.m_LabelBackColor);
		btc.setFocusable(false);
		df.applyPattern("###0.0000");
		btc.setText(df.format(btcNow));
		btc.setToolTipText("Total BTC held in tranches");
		btcLab.setLabelFor(btc);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(btcLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(btc, m_GBC);

		// row 7: total IB in all tranches
		JLabel ibLab = new JLabel("IB total: ");
		ibLab.setFont(ahp.M_LabelFont);
		ibLab.setHorizontalAlignment(SwingConstants.TRAILING);
		ibLab.setBorder(m_InsetBorder);
		JTextField ib = new JTextField(12);
		ib.setDropTarget(null);
		ib.setEditable(false);
		ib.setOpaque(false);
		ib.setHorizontalAlignment(JTextField.LEFT);
		ib.setBorder(m_InsetBorder);
		ib.setBackground(m_Plugin.m_LabelBackColor);
		ib.setFocusable(false);
		ib.setText(df.format(ibNow));
		ib.setToolTipText("Total IB held in tranches");
		ibLab.setLabelFor(ib);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(ibLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(ib, m_GBC);

		// row 8: first toss
		JLabel ftLab = new JLabel("First toss: ");
		ftLab.setFont(ahp.M_LabelFont);
		ftLab.setHorizontalAlignment(SwingConstants.TRAILING);
		ftLab.setBorder(m_InsetBorder);
		JTextField ft = new JTextField(30);
		ft.setDropTarget(null);
		ft.setEditable(false);
		ft.setOpaque(false);
		ft.setHorizontalAlignment(JTextField.LEFT);
		ft.setBorder(m_InsetBorder);
		ft.setBackground(m_Plugin.m_LabelBackColor);
		ft.setFocusable(false);
		Integer fToss = null;
		Date tossTime = null;
		if (m_FirstToss > 0) {
			fToss = new Integer(m_FirstToss);
			tossTime = null;
			try {
				tossTime = m_TossFormatter.parse(fToss.toString());
				// convert from UTC
				Date tossAt = ahp.localUTCdate(tossTime);
				ft.setText(fToss + " (" + m_DateFormatter.format(tossAt) + ")");
			}
			catch (ParseException pe) {
				Log.error("Invalid toss data format, " + fToss, pe);
			}
		}
		else {
			ft.setText("--none--");
		}
		ft.setToolTipText("First toss on which your account bet");
		ftLab.setLabelFor(ft);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(ftLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(ft, m_GBC);

		// row 9: last toss
		JLabel ltLab = new JLabel("Last toss: ");
		ltLab.setFont(ahp.M_LabelFont);
		ltLab.setHorizontalAlignment(SwingConstants.TRAILING);
		ltLab.setBorder(m_InsetBorder);
		JTextField lt = new JTextField(30);
		lt.setDropTarget(null);
		lt.setEditable(false);
		lt.setOpaque(false);
		lt.setHorizontalAlignment(JTextField.LEFT);
		lt.setBorder(m_InsetBorder);
		lt.setBackground(m_Plugin.m_LabelBackColor);
		lt.setFocusable(false);
		if (m_LastToss > 0) {
			fToss = new Integer(m_LastToss);
			tossTime = null;
			try {
				tossTime = m_TossFormatter.parse(fToss.toString());
				// convert from UTC
				Date tossAt = ahp.localUTCdate(tossTime);
				lt.setText(fToss + " (" + m_DateFormatter.format(tossAt) + ")");
			}
			catch (ParseException pe) {
				Log.error("Invalid toss data format, " + fToss, pe);
			}
		}
		else {
			lt.setText("--none--");
		}
		if (m_Active) {
			lt.setToolTipText("Most recent toss on which your account bet");
		}
		else {
			lt.setToolTipText("Last toss on which your account was betting");
		}
		ltLab.setLabelFor(lt);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(ltLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(lt, m_GBC);

		// row 10: low value
		JLabel lowLab = new JLabel("Low value: ");
		lowLab.setFont(ahp.M_LabelFont);
		lowLab.setHorizontalAlignment(SwingConstants.TRAILING);
		lowLab.setBorder(m_InsetBorder);
		JTextField low = new JTextField(12);
		low.setDropTarget(null);
		low.setEditable(false);
		low.setOpaque(false);
		low.setHorizontalAlignment(JTextField.LEFT);
		low.setBorder(m_InsetBorder);
		low.setBackground(m_Plugin.m_LabelBackColor);
		low.setFocusable(false);
		low.setText(df.format(m_LowValue));
		low.setToolTipText("Lowest valuation seen during betting (in BTC)");
		lowLab.setLabelFor(low);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(lowLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(low, m_GBC);

		// row 11: low toss
		JLabel lowTossLab = new JLabel("Low toss: ");
		lowTossLab.setFont(ahp.M_LabelFont);
		lowTossLab.setHorizontalAlignment(SwingConstants.TRAILING);
		lowTossLab.setBorder(m_InsetBorder);
		JTextField lowValToss = new JTextField(30);
		lowValToss.setDropTarget(null);
		lowValToss.setEditable(false);
		lowValToss.setOpaque(false);
		lowValToss.setHorizontalAlignment(JTextField.LEFT);
		lowValToss.setBorder(m_InsetBorder);
		lowValToss.setBackground(m_Plugin.m_LabelBackColor);
		lowValToss.setFocusable(false);
		if (m_LowToss > 0) {
			fToss = new Integer(m_LowToss);
			tossTime = null;
			try {
				tossTime = m_TossFormatter.parse(fToss.toString());
				// convert from UTC
				Date tossAt = ahp.localUTCdate(tossTime);
				lowValToss.setText(fToss + " (" + m_DateFormatter.format(tossAt)
										+ ")");
			}
			catch (ParseException pe) {
				Log.error("Invalid toss data format, " + fToss, pe);
			}
		}
		else {
			lowValToss.setText("--none--");
		}
		lowValToss.setToolTipText("Time of toss at which low value occurred");
		lowTossLab.setLabelFor(lowValToss);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(lowTossLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(lowValToss, m_GBC);

		// row 12: high value
		JLabel highLab = new JLabel("High value: ");
		highLab.setFont(ahp.M_LabelFont);
		highLab.setHorizontalAlignment(SwingConstants.TRAILING);
		highLab.setBorder(m_InsetBorder);
		JTextField high = new JTextField(12);
		high.setDropTarget(null);
		high.setEditable(false);
		high.setOpaque(false);
		high.setHorizontalAlignment(JTextField.LEFT);
		high.setBorder(m_InsetBorder);
		high.setBackground(m_Plugin.m_LabelBackColor);
		high.setFocusable(false);
		high.setText(df.format(m_HighValue));
		high.setToolTipText("Highest valuation seen during betting (in BTC)");
		highLab.setLabelFor(high);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(highLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(high, m_GBC);

		// row 13: high toss
		JLabel highTossLab = new JLabel("High toss: ");
		highTossLab.setFont(ahp.M_LabelFont);
		highTossLab.setHorizontalAlignment(SwingConstants.TRAILING);
		highTossLab.setBorder(m_InsetBorder);
		JTextField highValToss = new JTextField(30);
		highValToss.setDropTarget(null);
		highValToss.setEditable(false);
		highValToss.setOpaque(false);
		highValToss.setHorizontalAlignment(JTextField.LEFT);
		highValToss.setBorder(m_InsetBorder);
		highValToss.setBackground(m_Plugin.m_LabelBackColor);
		highValToss.setFocusable(false);
		if (m_HighToss > 0) {
			fToss = new Integer(m_HighToss);
			tossTime = null;
			try {
				tossTime = m_TossFormatter.parse(fToss.toString());
				// convert from UTC
				Date tossAt = ahp.localUTCdate(tossTime);
				highValToss.setText(fToss + " (" + m_DateFormatter.format(tossAt)
											+ ")");
			}
			catch (ParseException pe) {
				Log.error("Invalid toss data format, " + fToss, pe);
			}
		}
		else {
			highValToss.setText("--none--");
		}
		highValToss.setToolTipText("Time of toss at which high value occurred");
		highTossLab.setLabelFor(highValToss);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(highTossLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(highValToss, m_GBC);

		// row 14: wholeness tosses for this account
		JLabel whLab = new JLabel("Wholeness bonus every: ");
		whLab.setFont(ahp.M_LabelFont);
		whLab.setHorizontalAlignment(SwingConstants.TRAILING);
		whLab.setBorder(m_InsetBorder);
		whLab.setToolTipText("Number of tosses your account must bet in order "
									+ "to be eligible");
		JTextField whTosses = new JTextField(4);
		whTosses.setDropTarget(null);
		whTosses.setEditable(false);
		whTosses.setOpaque(false);
		whTosses.setHorizontalAlignment(JTextField.LEFT);
		whTosses.setBorder(m_InsetBorder);
		whTosses.setBackground(m_Plugin.m_LabelBackColor);
		whTosses.setFocusable(false);
		whTosses.setText(m_WholenessTosses + " tosses");
		if (m_WholenessTosses > -1) {
			float hours = (float) m_WholenessTosses / sp.getTossRate();
			whTosses.setToolTipText("Equals " + df.format(hours)
											+ " hours; may differ from system default");
		}
		else {
			whTosses.setToolTipText("Your account is ineligible for wholeness "
											+ "bonuses");
		}
		whLab.setLabelFor(whTosses);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(whLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(whTosses, m_GBC);

		// row 15: toss of last wholeness check for this particular account
		JLabel lwhLab = new JLabel("Last wholeness check: ");
		lwhLab.setFont(ahp.M_LabelFont);
		lwhLab.setHorizontalAlignment(SwingConstants.TRAILING);
		lwhLab.setBorder(m_InsetBorder);
		lwhLab.setToolTipText("The last time your account was checked for a "
									+ "wholeness bonus");
		JTextField lastWhToss = new JTextField(8);
		lastWhToss.setDropTarget(null);
		lastWhToss.setEditable(false);
		lastWhToss.setOpaque(false);
		lastWhToss.setHorizontalAlignment(JTextField.LEFT);
		lastWhToss.setBorder(m_InsetBorder);
		lastWhToss.setBackground(m_Plugin.m_LabelBackColor);
		lastWhToss.setFocusable(false);
		if (m_LastWhChkToss == 0) {
			lastWhToss.setText("0 (never)");
		}
		else {
			Integer wToss = new Integer(m_LastWhChkToss);
			try {
				tossTime = m_TossFormatter.parse(wToss.toString());
				// convert from UTC
				Date tossAt = ahp.localUTCdate(tossTime);
				lastWhToss.setText(wToss + " (" + m_DateFormatter.format(tossAt)
										+ ")");
			}
			catch (ParseException pe) {
				Log.error("Invalid toss data format, " + wToss, pe);
			}
		}
		if (!m_Active) {
			lastWhToss.setToolTipText("Your account must be betting for "
										+ m_WholenessTosses + " tosses to be eligible");
		}
		else if (m_WholenessTosses > 0) {
			/* Add in tooltip the estimate of the next wholeness check.  Use as a
			 * reference the later of the last wholeness check toss, or the first
			 * toss of the current betting session.
			 */
			int refToss = Math.max(m_LastWhChkToss, m_FirstToss);
			ParsePosition parsepos = new ParsePosition(0);
			Date lastTossTime
				= m_TossFormatter.parse(Integer.toString(refToss), parsepos);
			long tossesHence = lastTossTime.getTime()
							+ (Timer.ONE_HOUR / sp.getTossRate()) * m_WholenessTosses;
			Date eligibleToss = new Date(tossesHence);
			String tossStr = m_TossFormatter.format(eligibleToss);
			int wholenessToss = Integer.parseInt(tossStr);
			lastWhToss.setToolTipText("Next wholeness check should be on toss "
											+ wholenessToss);
		}
		else {
			lastWhToss.setToolTipText("Your account is ineligible for wholeness "
											+ "bonuses");
		}
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(lwhLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(lastWhToss, m_GBC);

		// spacer to demark user-configurable section
		JSeparator demark = new JSeparator();
		demark.setToolTipText("You may change values below this line");
		m_ConfigPanel.add(demark, m_GBC);
		m_ConfigPanel.add(Box.createVerticalStrut(5), m_GBC);

		// row 16: profit target
		JLabel profLab = new JLabel("Profit target: ");
		profLab.setFont(ahp.M_LabelFont);
		profLab.setHorizontalAlignment(SwingConstants.TRAILING);
		profLab.setBorder(m_InsetBorder);
		JTextField profTarg = new JTextField(6);
		profTarg.setDropTarget(null);
		profTarg.setEditable(false);
		profTarg.setOpaque(false);
		profTarg.setHorizontalAlignment(JTextField.LEFT);
		profTarg.setBorder(m_InsetBorder);
		profTarg.setBackground(m_Plugin.m_LabelBackColor);
		profTarg.setFocusable(false);
		df.applyPattern("+##0.00##;-##0.00##");
		profTarg.setText(df.format(m_ProfitTarget) + "%");
		profTarg.setToolTipText("Upper bound at which your account will stop "
										+ "betting and cash out (0.0 means none set)");
		profLab.setLabelFor(profTarg);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(profLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(profTarg, m_GBC);

		// row 17: new target (entry)
		JLabel nProfLab = new JLabel("New target: ");
		nProfLab.setFont(ahp.M_LabelFont);
		nProfLab.setHorizontalAlignment(SwingConstants.TRAILING);
		nProfLab.setBorder(m_InsetBorder);
		nProfLab.setLabelFor(m_NewProfitTarget);
		m_NewProfitTarget.setText(df.format(m_ProfitTarget));
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(nProfLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(m_NewProfitTarget, m_GBC);

		// row 18: stop loss target
		JLabel stopLab = new JLabel("Stop loss: ");
		stopLab.setFont(ahp.M_LabelFont);
		stopLab.setHorizontalAlignment(SwingConstants.TRAILING);
		stopLab.setBorder(m_InsetBorder);
		JTextField stop = new JTextField(6);
		stop.setDropTarget(null);
		stop.setEditable(false);
		stop.setOpaque(false);
		stop.setHorizontalAlignment(JTextField.LEFT);
		stop.setBorder(m_InsetBorder);
		stop.setBackground(m_Plugin.m_LabelBackColor);
		stopLab.setLabelFor(stop);
		stop.setText(df.format(m_StopLoss) + "%");
		stop.setToolTipText("Lower bound at which your account will stop "
									+ "betting and cash out (0.0 means none set)");
		stopLab.setLabelFor(stop);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(stopLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(stop, m_GBC);

		// row 19: new stop (entry)
		JLabel nStopLab = new JLabel("New stop: ");
		nStopLab.setFont(ahp.M_LabelFont);
		nStopLab.setHorizontalAlignment(SwingConstants.TRAILING);
		nStopLab.setBorder(m_InsetBorder);
		m_NewStopLoss.setText(df.format(m_StopLoss));
		nStopLab.setLabelFor(m_NewStopLoss);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(nStopLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(m_NewStopLoss, m_GBC);

		// row 20: IB sale discount policy
		JLabel saleLab = new JLabel("IB Sale Policy: ");
		saleLab.setFont(ahp.M_LabelFont);
		saleLab.setHorizontalAlignment(SwingConstants.TRAILING);
		saleLab.setBorder(m_InsetBorder);
		JTextField sPolicy = new JTextField(8);
		sPolicy.setDropTarget(null);
		sPolicy.setEditable(false);
		sPolicy.setOpaque(false);
		sPolicy.setHorizontalAlignment(JTextField.LEFT);
		sPolicy.setBorder(m_InsetBorder);
		sPolicy.setBackground(m_Plugin.m_LabelBackColor);
		sPolicy.setFocusable(false);
		DiscountPolicy salePol = findPolicy(m_SalePolicy, ABCBlock.M_PolicySale);
		if (salePol != null) {
			sPolicy.setText(salePol.getName());
			if (salePol.getSpecification().isEmpty()) {
				sPolicy.setToolTipText("(immediate settlement on next toss)");
			}
			else {
				sPolicy.setToolTipText(salePol.getSpecification());
			}
		}
		else {
			sPolicy.setText("--none--");
			sPolicy.setToolTipText("not found");
		}
		saleLab.setLabelFor(sPolicy);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(saleLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(sPolicy, m_GBC);

		// row 21: new sale policy (drop-down)
		JLabel nSpolLab = new JLabel("New Sale Policy: ");
		nSpolLab.setFont(ahp.M_LabelFont);
		nSpolLab.setHorizontalAlignment(SwingConstants.TRAILING);
		nSpolLab.setBorder(m_InsetBorder);
		m_SaleDiscDropdown.removeAllItems();
		for (DiscountPolicy spol : m_SDiscPolicies) {
			String thisPol = spol.getName() + ": " + spol.getSpecification();
			m_SaleDiscDropdown.addItem(thisPol);
		}
		if (salePol != null) {
			m_SaleDiscDropdown.setSelectedItem(salePol.getName() + ": "
															+ salePol.getSpecification());
		}
		nSpolLab.setLabelFor(m_SaleDiscDropdown);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(nSpolLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(Box.createHorizontalStrut(50), m_GBC);
		m_ConfigPanel.add(m_SaleDiscDropdown, m_GBC);

		// row 22: IB purchase discount policy
		JLabel purchLab = new JLabel("IB Purchase Policy: ");
		purchLab.setFont(ahp.M_LabelFont);
		purchLab.setHorizontalAlignment(SwingConstants.TRAILING);
		purchLab.setBorder(m_InsetBorder);
		JTextField pPolicy = new JTextField(8);
		pPolicy.setDropTarget(null);
		pPolicy.setEditable(false);
		pPolicy.setOpaque(false);
		pPolicy.setHorizontalAlignment(JTextField.LEFT);
		pPolicy.setBorder(m_InsetBorder);
		pPolicy.setBackground(m_Plugin.m_LabelBackColor);
		pPolicy.setFocusable(false);
		DiscountPolicy purchPol
			= findPolicy(m_PurchPolicy, ABCBlock.M_PolicyPurch);
		if (purchPol != null) {
			pPolicy.setText(purchPol.getName());
			if (purchPol.getSpecification().isEmpty()) {
				pPolicy.setToolTipText("(immediate settlement on next toss)");
			}
			else {
				pPolicy.setToolTipText(purchPol.getSpecification());
			}
		}
		else {
			pPolicy.setText("--none--");
			pPolicy.setToolTipText("not found");
		}
		purchLab.setLabelFor(pPolicy);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(purchLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(pPolicy, m_GBC);

		// row 23: new purchase policy (drop-down)
		JLabel nPpolLab = new JLabel("New Purchase Policy: ");
		nPpolLab.setFont(ahp.M_LabelFont);
		nPpolLab.setHorizontalAlignment(SwingConstants.TRAILING);
		nPpolLab.setBorder(m_InsetBorder);
		m_PurchDiscDropdown.removeAllItems();
		for (DiscountPolicy ppol : m_PDiscPolicies) {
			String thisPol = ppol.getName() + ": " + ppol.getSpecification();
			m_PurchDiscDropdown.addItem(thisPol);
		}
		if (purchPol != null) {
			m_PurchDiscDropdown.setSelectedItem(purchPol.getName() + ": "
															+ purchPol.getSpecification());
		}
		nPpolLab.setLabelFor(m_PurchDiscDropdown);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(nPpolLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(Box.createHorizontalStrut(50), m_GBC);
		m_ConfigPanel.add(m_PurchDiscDropdown, m_GBC);

		// row 24: tranche size
		JLabel tSizeLab = new JLabel("Tranche size: ");
		tSizeLab.setFont(ahp.M_LabelFont);
		tSizeLab.setHorizontalAlignment(SwingConstants.TRAILING);
		tSizeLab.setBorder(m_InsetBorder);
		JTextField tSize = new JTextField(8);
		tSize.setDropTarget(null);
		tSize.setEditable(false);
		tSize.setOpaque(false);
		tSize.setHorizontalAlignment(JTextField.LEFT);
		tSize.setBorder(m_InsetBorder);
		tSize.setBackground(m_Plugin.m_LabelBackColor);
		tSize.setFocusable(false);
		df.applyPattern("###0.0000");
		tSize.setText(df.format(m_TrancheSize));
		tSize.setToolTipText("Amount of BTC deployed to each new tranche");
		tSizeLab.setLabelFor(tSize);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(tSizeLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(tSize, m_GBC);

		// row 25: new tranche size (entry)
		JLabel nSzLab = new JLabel("New tranche size: ");
		nSzLab.setFont(ahp.M_LabelFont);
		nSzLab.setHorizontalAlignment(SwingConstants.TRAILING);
		nSzLab.setBorder(m_InsetBorder);
		m_NewTrancheSize.setText(df.format(m_TrancheSize));
		nSzLab.setLabelFor(m_NewTrancheSize);
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(nSzLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(m_NewTrancheSize, m_GBC);

		// row 26: checkbox to toggle VFQ usage
		JLabel vfqLab = new JLabel("VFQ bleed-in setting: ");
		vfqLab.setFont(ahp.M_LabelFont);
		vfqLab.setHorizontalAlignment(SwingConstants.TRAILING);
		vfqLab.setBorder(m_InsetBorder);
		vfqLab.setLabelFor(m_UseVFQ);
		vfqLab.setToolTipText("Check this box to apply your VFQ balance to pay "
									+ "for a portion of your IB purchases");
		firstItemConstraints(m_GBC);
		m_ConfigPanel.add(vfqLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_ConfigPanel.add(m_UseVFQ);
		m_ConfigPanel.add(Box.createVerticalGlue(), m_GBC);

		// put entire form into the main scroll area
		m_WorkArea.setViewportView(m_ConfigPanel);
		m_UpdateButton.setEnabled(true);
		revalidate();
	}

	/**
	 * method to define update action
	 */
	private void setUpdateAction() {
		m_UpdateAction = new AbstractAction("Update Settings") {
			public void actionPerformed(ActionEvent ae) {
				// profit target must be a valid float
				String newTarg = m_NewProfitTarget.getText();
				Float nTarg = null;
				try {
					nTarg = Float.parseFloat(newTarg);
				}
				catch (NumberFormatException nfe) {
					Log.error("Illegal profit target value, " + newTarg, nfe);
					m_TabManager.showError("cannot set new profit target",
													newTarg + " is an illegal value",
													"enter a valid percentage");
					return;
				}
				final float newProfitTarget = nTarg.floatValue();

				// stop loss must be a valid float
				String newStop = m_NewStopLoss.getText();
				Float nStop = null;
				try {
					nStop = Float.parseFloat(newStop);
				}
				catch (NumberFormatException nfe) {
					Log.error("Illegal stop loss value, " + newStop, nfe);
					m_TabManager.showError("cannot set new stop loss",
													newStop + " is an illegal value",
													"enter a valid percentage");
					return;
				}
				final float newStopLoss = nStop.floatValue();

				// stop must be less than target, unless both are zero
				if (!(newProfitTarget == 0.0F && newStopLoss == 0.0F)
					 && newProfitTarget <= newStopLoss)
				{
					m_TabManager.showError("cannot set new exit targets",
													"profit target cannot be less than stop",
										"enter stop below target, or target above stop");
					return;
				}

				// tranche size must be a positive float (min/max checked @server)
				String newTSz = m_NewTrancheSize.getText();
				Float nTsz = null;
				try {
					nTsz = Float.parseFloat(newTSz);
				}
				catch (NumberFormatException nfe) {
					Log.error("Illegal tranche size value, " + newTSz, nfe);
					m_TabManager.showError("cannot set new tranche size",
													newTSz + " is an illegal value",
													"enter a positive number");
					return;
				}
				final float newTrancheSize = nTsz.floatValue();
				if (newTrancheSize <= 0.0F) {
					Log.error("Illegal tranche size value, " + newTSz);
					m_TabManager.showError("cannot set new tranche size",
													newTSz + " is an illegal value",
													"enter a positive number");
					return;
				}

				final int newSalePolicy = m_SaleDiscDropdown.getSelectedIndex();
				final int newPurchPolicy = m_PurchDiscDropdown.getSelectedIndex();

				// input looks okay, so attempt update at server (do in new thread)
				Runnable sendMods = new Runnable() {
					public void run() {
						if (!sendConfigMods(newProfitTarget,
												  newStopLoss,
												  newTrancheSize,
												  newSalePolicy,
												  newPurchPolicy,
												  m_UseVFQ.isSelected()))
						{
							Log.error("Unable to send account config update to "
										+ "ABC server");
						}
					}
				};
				new Thread(sendMods).start();

				/* NB: If we get an error from the server (e.g. our tranche size
				 *		 is below minimum or above maximum), then we see an error
				 *		 popup describing what's wrong, without altering inputs.
				 *		 If we get a success reply from the server, it goes to
				 *		 recordChanges() and causes an update of the parameters,
				 *		 and a redraw of the screen to reflect any changes.
				 */
			}
		};
	}

	/**
	 * helper method to lookup the indicated purchase or sale discount policy
	 * @param pId the policy ID wanted
	 * @param type the type (sale or purchase)
	 * @return the policy, or null if not found
	 */
	public DiscountPolicy findPolicy(int pId, String type) {
		DiscountPolicy dPol = null;
		if (pId < 0 || type == null || type.isEmpty()) {
			return dPol;
		}
		if (type.equals(ABCBlock.M_PolicySale)) {
			for (DiscountPolicy pol : m_SDiscPolicies) {
				if (pol.getPolicyId() == (short) pId) {
					dPol = pol;
					break;
				}
			}
		}
		else if (type.equals(ABCBlock.M_PolicyPurch)) {
			for (DiscountPolicy pol : m_PDiscPolicies) {
				if (pol.getPolicyId() == (short) pId) {
					dPol = pol;
					break;
				}
			}
		}
		return dPol;
	}

	/**
	 * helper method to prepare GridBagConstraints for use with the
	 * first item on a line (such as a prompt/label)
	 */
	private void firstItemConstraints(GridBagConstraints gbc) {
		if (gbc == null) {
			return;
		}
		gbc.gridwidth = GridBagConstraints.RELATIVE;
		gbc.fill = GridBagConstraints.NONE;
	}

	/**
	 * helper method to prepare GridBagConstraints for use with the
	 * last item on a line (such as an entry field)
	 */
	private void lastItemConstraints(GridBagConstraints gbc) {
		if (gbc == null) {
			return;
		}
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.fill = GridBagConstraints.HORIZONTAL;
	}

	/**
	 * record account details as returned by the server
	 * @param accData the details data from the server
	 */
	public void recordDetails(ABCBlock accData) {
		if (accData == null
			 || !accData.getOpcode().equalsIgnoreCase("REP_account_details"))
		{
			Log.error("recordDetails(): not an account details message");
			return;
		}

		// bettor ID is recorded at our home pane
		AHomePane ahp = (AHomePane) m_ParentHome;
		ahp.setBettorId(accData.getBettorId());

		// other stuff is recorded here
		m_Active = accData.isActive();
		m_Bankroll = accData.getBankroll();
		m_Additions = accData.getAdditions();
		m_Subtractions = accData.getSubtractions();
		m_BTCChipStart = accData.getBTCChipStart();
		m_TrancheSize = accData.getTrancheSize();
		m_VFQPolicy = accData.isVFQUsed();
		m_UseVFQ.setSelected(m_VFQPolicy);
		m_FirstToss = accData.getFirstToss();
		m_LastToss = accData.getLastToss();
		m_HighValue = accData.getHighValue();
		m_HighToss = accData.getHighToss();
		m_LowValue = accData.getLowValue();
		m_LowToss = accData.getLowToss();
		m_WholenessTosses = accData.getWholenessTosses();
		if (m_WholenessTosses == 0) {
			// use system default
			ATabManager atm = (ATabManager) m_TabManager;
			StatsPane sp = atm.getStatsPane();
			m_WholenessTosses = sp.getWholenessTosses();
		}
		m_LastWhChkToss = accData.getLastWhChkToss();
		m_ProfitTarget = accData.getProfitTarget();
		m_StopLoss = accData.getStopLoss();

		// rebuild list of all possible IB discount policies
		ArrayList<ABCBlock.DiscountPolicy> allDiscPols
			= accData.getDiscountPolicies();
		if (!allDiscPols.isEmpty()) {
			m_SDiscPolicies.clear();
			m_PDiscPolicies.clear();
		}
		for (ABCBlock.DiscountPolicy discPol : allDiscPols) {
			DiscountPolicy dPol = new DiscountPolicy();
			dPol.setPolicyId(discPol.m_PolicyId);
			dPol.setType(discPol.m_Type);
			dPol.setName(discPol.m_Name);
			dPol.setSpecification(discPol.m_Spec);
			if (dPol.getType().equals(ABCBlock.M_PolicySale)) {
				m_SDiscPolicies.add(dPol);
			}
			else {
				m_PDiscPolicies.add(dPol);
			}
		}

		// pick out our given discount policies
		String purchPol = accData.getPurchPolicy();
		String salePol = accData.getSalePolicy();
		// use zero (immediate buy/sell) unless we find something valid below
		m_PurchPolicy = 0;
		m_SalePolicy = 0;

		// the strings returned by the server contain the Id as first element
		final String delim = "##";
		int dIdx = purchPol.indexOf(delim);
		if (dIdx != -1) {
			String pPol = purchPol.substring(0, dIdx);
			try {
				Short polId = Short.parseShort(pPol.trim());
				m_PurchPolicy = polId.shortValue();
				// make sure this policy exists
				DiscountPolicy policy
					= findPolicy(m_PurchPolicy, ABCBlock.M_PolicyPurch);
				if (policy == null) {
					Log.error("Unknown purchase policy set, " + polId);
					m_PurchPolicy = 0;
				}
			}
			catch (NumberFormatException nfe) {
				Log.error("Could not parse out purch policy Id from " + purchPol,
							nfe);
			}
		}
		dIdx = salePol.indexOf(delim);
		if (dIdx != -1) {
			String sPol = salePol.substring(0, dIdx);
			try {
				Short polId = Short.parseShort(sPol.trim());
				m_SalePolicy = polId.shortValue();
				// make sure this policy exists
				DiscountPolicy policy
					= findPolicy(m_SalePolicy, ABCBlock.M_PolicySale);
				if (policy == null) {
					Log.error("Unknown sale policy set, " + polId);
					m_SalePolicy = 0;
				}
			}
			catch (NumberFormatException nfe) {
				Log.error("Could not parse out sale policy Id from " + salePol,
							nfe);
			}
		}

		// the queued deployments, if any, are stored in the PlayPane
		ATabManager atm = (ATabManager) m_TabManager;
		ArrayList<ABCBlock.QueuedDeploy> deployQ = accData.getQueuedDeployments();
		PlayPane pp = atm.getPlayPane();
		if (pp != null) {
			pp.recordDeployments(deployQ);
		}

		// tell the history pane about our first and last tosses, to set sliders
		HistoryPane hp = atm.getHistoryPane();
		if (hp != null) {
			hp.configTossRange(m_FirstToss, m_LastToss);
		}

		// update display to reflect this data
		prepDisplay();
	}

	/**
	 * record (successful) changes to configurable items
	 */
	public void recordChanges(ABCBlock accData) {
		if (accData == null
			 || !accData.getOpcode().equalsIgnoreCase("REP_details_config"))
		{
			Log.error("recordChanges(): not an details config message");
			return;
		}

		// check echoed value against what we have, and note changes
		boolean tSizeUpd = false;
		float newTSz = accData.getTrancheSize();
		if (newTSz != m_TrancheSize) {
			m_TrancheSize = newTSz;
			tSizeUpd = true;
		}
		boolean profTargUpd = false;
		float newProfTarg = accData.getProfitTarget();
		if (newProfTarg != m_ProfitTarget) {
			m_ProfitTarget = newProfTarg;
			profTargUpd = true;
		}
		boolean stopUpd = false;
		float newStop = accData.getStopLoss();
		if (newStop != m_StopLoss) {
			m_StopLoss = newStop;
			stopUpd = true;
		}
		boolean sPolUpd = false;
		String newSPol = accData.getSalePolicy();
		// look for this policy
		DiscountPolicy existSPol = null;
		for (DiscountPolicy dp : m_SDiscPolicies) {
			String fullPolicy = dp.getPolicyId() + "##" + dp.getSpecification();
			if (fullPolicy.equals(newSPol)) {
				existSPol = dp;
				break;
			}
		}
		if (existSPol == null) {
			Log.error("New sale discount policy not found, " + newSPol);
		}
		else {
			if (m_SalePolicy != existSPol.getPolicyId()) {
				m_SalePolicy = existSPol.getPolicyId();
				sPolUpd = true;
			}
		}
		boolean pPolUpd = false;
		String newPPol = accData.getPurchPolicy();
		// look for this policy
		DiscountPolicy existPPol = null;
		for (DiscountPolicy dp : m_PDiscPolicies) {
			String fullPolicy = dp.getPolicyId() + "##" + dp.getSpecification();
			if (fullPolicy.equals(newPPol)) {
				existPPol = dp;
				break;
			}
		}
		if (existPPol == null) {
			Log.error("New purchase discount policy not found, " + newPPol);
		}
		else {
			if (m_PurchPolicy != existPPol.getPolicyId()) {
				m_PurchPolicy = existPPol.getPolicyId();
				pPolUpd = true;
			}
		}
		boolean oldVFQPolicy = m_VFQPolicy;
		m_VFQPolicy = accData.isVFQUsed();
		boolean vfqUpd = oldVFQPolicy != m_VFQPolicy;

		// show changes to the user, by building a stack of string objects
		Object[] confStack = new Object[7];
		int confCnt = 0;
		confStack[confCnt++]
			= "The following configuration values have been updated:";
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(3);
		if (tSizeUpd) {
			confStack[confCnt++] = "Tranche size: " + df.format(m_TrancheSize);
		}
		if (profTargUpd) {
			confStack[confCnt++]
				= "Profit target: " + df.format(m_ProfitTarget) + "%";
		}
		if (stopUpd) {
			confStack[confCnt++] = "Stop loss: " + df.format(m_StopLoss) + "%";
		}
		if (sPolUpd) {
			String disc = "Sale discount policy: " + m_SalePolicy + "##";
			String spec = existSPol.getSpecification();
			if (!spec.isEmpty()) {
				disc += spec;
			}
			else {
				disc += "default (immediate)";
			}
			confStack[confCnt++] = disc;
		}
		if (pPolUpd) {
			String purch = "Purchase discount policy: " + m_PurchPolicy + "##";
			String spec = existPPol.getSpecification();
			if (!spec.isEmpty()) {
				purch += spec;
			}
			else {
				purch += "default (immediate)";
			}
			confStack[confCnt++] = purch;
		}
		if (vfqUpd) {
			confStack[confCnt++]
				= "VFQ utilization policy: " + Boolean.toString(m_VFQPolicy);
		}
		if (tSizeUpd || profTargUpd || stopUpd || sPolUpd || pPolUpd || vfqUpd) {
			JOptionPane.showMessageDialog(this, confStack, "Information",
													JOptionPane.INFORMATION_MESSAGE);
		}
		else {
			m_TabManager.showInfo("No config changes recorded");
		}
		prepDisplay();
	}

	/**
	 * query the server for the current config data
	 * @param synchronous whether or not we should do this synchronously
	 * @return true on success
	 */
	public boolean queryAcctConfig(boolean synchronous) {
		// prepare account config request message
		String sessId = m_TabManager.getLoginPane().getSessionId();
		String abcId = m_Plugin.getLoginSecrets().getSVMId();
		ABCClientBlock confBlock = new ABCClientBlock();
		confBlock.setOpcode("REQ_account_details");
		confBlock.setSessionId(sessId);

		// send to server; answer will go to recordDetails() above
		ABCListener abcListener = (ABCListener) m_Plugin.getSVMListener();
		if (synchronous) {
			if (!abcListener.sendRequest(confBlock)) {
				Log.error("Account config request failed to ABC Id " + abcId);
				m_TabManager.showError("cannot get account config at ABC " + abcId,
												"failure sending config request message");
				return false;
			}
		}
		else {
			if (!abcListener.sendAsyncRequest(confBlock)) {
				Log.error("Account config request failed to ABC Id " + abcId);
				m_TabManager.showError("cannot get account config at ABC " + abcId,
												"failure sending config request message");
				return false;
			}
		}
		return true;
	}

	/**
	 * helper method to submit modifications to the server
	 * @param newProfTarget new upper exit percentage
	 * @param newStopLoss new lower exit percentage
	 * @param newTrancheSize new size for tranches
	 * @param newSalePolicy new sale discount policy to set
	 * @param newPurchPolicy new purchase discount policy to set
	 * @param newVfqPolicy new value for applying VFQ to IB purchases
	 * @return true on success
	 */
	private boolean sendConfigMods(float newProfTarget,
											 float newStopLoss,
											 float newTrancheSize,
											 int newSalePolicy,
											 int newPurchPolicy,
											 boolean newVfqPolicy)
	{
		// prepare config modification message
		String sessId = m_TabManager.getLoginPane().getSessionId();
		String abcId = m_Plugin.getLoginSecrets().getSVMId();
		ABCClientBlock confBlock = new ABCClientBlock();
		confBlock.setOpcode("REQ_config_details");
		confBlock.setSessionId(sessId);
		confBlock.setTrancheSize(newTrancheSize);
		confBlock.setProfitTarget(newProfTarget);
		confBlock.setStopLoss(newStopLoss);
		DiscountPolicy sPolicy = m_SDiscPolicies.get(newSalePolicy);
		confBlock.setSalePolicy(newSalePolicy + "##"
										+ sPolicy.getSpecification());
		DiscountPolicy pPolicy = m_PDiscPolicies.get(newPurchPolicy);
		confBlock.setPurchPolicy(newPurchPolicy + "##"
										+ pPolicy.getSpecification());
		confBlock.setUseVFQ(newVfqPolicy);

		// send to server; answer will go to recordChanges() above
		ABCListener abcListener = (ABCListener) m_Plugin.getSVMListener();
		if (!abcListener.sendRequest(confBlock)) {
			Log.error("Config mod request failed to ABC Id " + abcId);
			m_TabManager.showError("cannot update account config at ABC " + abcId,
											"failure sending config update message");
			return false;
		}
		return true;
	}


	// get methods
	/**
	 * determine whether or not an account is currently betting
	 * @return true if betting
	 */
	public boolean isBetting() { return m_Active; }

	/**
	 * obtain list of all defined sale discount policies
	 * @return the list
	 */
	public ArrayList<DiscountPolicy>	getSalePolicies() {
		return m_SDiscPolicies;
	}

	/**
	 * obtain list of all defined purchase discount policies
	 * @return the list
	 */
	public ArrayList<DiscountPolicy> getPurchPolicies() {
		return m_PDiscPolicies;
	}

	/**
	 * obtain the first toss account bet
	 * @return the toss, in YYMMDDHHmm format (can be 0)
	 */
	public int getFirstToss() { return m_FirstToss; }

	/**
	 * obtain the last toss account bet
	 * @return the toss, in YYMMDDHHmm format (can be 0)
	 */
	public int getLastToss() { return m_LastToss; }

	/**
	 * obtain the account's starting bankroll
	 * @return the amount of initial BTC funding at start of betting
	 */
	public double getBankroll() { return m_Bankroll; }

	/**
	 * obtain the account's additions to bankroll
	 * @return the total additions to BTC funding since betting started, if any
	 */
	public double getAdditions() { return m_Additions; }

	/**
	 * obtain the account's subtractions from bankroll
	 * @return the total subtractions from BTC funding since betting started
	 */
	public double getSubtractions() { return m_Subtractions; }

	// END methods
}
