/*
 * last modified---
 * 	11-03-16/jet/ for now, comment out button actions and supply TBD popup
 * 	10-22-16/jet/ rework to include scroll pane
 * 	09-22-16/jet/ begin adding actual graphing features
 * 	08-22-16/jet/ new
 *
 * purpose---
 * 	provide a UI screen to deal with selecting account and tranche history
 * 	for graphical display
 */

package org.silentvault.spark.ui.svm.abc;

import javax.swing.*;
import javax.swing.border.EmptyBorder;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowAdapter;
import java.awt.*;
import java.awt.geom.GeneralPath;
import java.awt.geom.AffineTransform;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;
import org.silentvault.spark.ui.svm.*;
import org.silentvault.spark.ui.wallet.HomePane;
import org.silentvault.spark.ui.wallet.WTabManager;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;


/**
 * Extends AWorkPane to provide a screen for selecting a range of historical
 * toss data for a tranche or the whole account, to be graphed as specified.
 *
 * @author Justin Turrell
 */
public final class HistoryPane extends AWorkPane {
	// BEGIN data members
	/**
	 * constant for left and right margin spacing (axes and labels)
	 */
	private final int					M_HorizMargins;

	/**
	 * constant for bottom margin spacing (axis and label)
	 */
	private final int					M_VertMargins;

	/**
	 * action to fetch history data from server and graph it
	 */
	private AbstractAction			m_GraphAction;

	/**
	 * button for Graph action
	 */
	private JButton					m_GraphButton;

	/**
	 * action to refresh existing graph with new params
	 */
	private AbstractAction			m_RedrawAction;

	/**
	 * button for Redraw action
	 */
	private JButton					m_RedrawButton;

	/**
	 * the panel containing the search and graph config
	 */
	private JPanel						m_HistPanel;

	/**
	 * the history mode: true if doing account history, false if doing a tranche
	 */
	private boolean					m_HistMode;

	/**
	 * the ID of the tranche whose history we're working with (valid only when
	 * m_HistMode = false)
	 */
	private int							m_TrancheId;

	/**
	 * storage for account history, after retrieval via Graph action
	 */
	private ArrayList<ABCBlock.AccountHistory>	m_AcctHistory;

	/**
	 * storage for tranche history, after retrieval via Graph action
	 */
	private ArrayList<ABCBlock.TrancheHistory>	m_TrancheHistory;

	/**
	 * radio button for selecting all tosses
	 */
	private JRadioButton				m_AllTosses;

	/**
	 * radio button for selecting a range of tosses
	 */
	private JRadioButton				m_RangeTosses;

	/**
	 * button group for tosses selection buttons
	 */
	private ButtonGroup				m_TossesGroup;

	/**
	 * slider for selecting starting toss of a range
	 */
	private JSlider					m_StartToss;

	/**
	 * slider for selecting ending toss of a range
	 */
	private JSlider					m_EndToss;

	/**
	 * flag for whether or not stops are shown on graph (true by default)
	 */
	private JCheckBox					m_ShowStops;

	/**
	 * flag for whether or not tranche count should be shown (acct mode only)
	 */
	private JCheckBox					m_ShowTrancheCnt;

	/**
	 * flag for whether or not BTC chips should also be graphed separately
	 */
	private JCheckBox					m_ShowBTCChips;

	/**
	 * flag for whether or not IB value should also be graphed separately
	 */
	private JCheckBox					m_ShowIBValue;

	/**
	 * flag for whether or not total value shown, as percentage of bankroll
	 * (true by default)
	 */
	private JCheckBox					m_ShowTotalValue;

	/**
	 * flag for whether or not VFQ balance should also be graphed separately
	 */
	private JCheckBox					m_ShowVFQ;

	/**
	 * flag for whether additions should be shown with arrows
	 */
	private JCheckBox					m_ShowAdditions;

	/**
	 * flag for whether subtractions should be shown with arrows (acct mode only)
	 */
	private JCheckBox					m_ShowSubtractions;

	/**
	 * flag for whether or not IB ratio should be shown on graph (default true)
	 */
	private JCheckBox					m_ShowIBRatio;

	/**
	 * flag for whether or not subsidy status is indicated (tranche mode only)
	 */
	private JCheckBox					m_ShowSubsidy;

	/**
	 * floating point number formatter
	 */
	private DecimalFormat			m_Df;

	/**
	 * background color for this panel
	 */
	private Color						m_HistBack;

	/**
	 * default font for labels and legend
	 */
	private Font						m_DefaultFont;

	/**
	 * flag set indicating that an X-axis scrollbar will be required
	 */
	private boolean					m_XScrolling;

	/**
	 * X dimension for m_AcctHistPopup
	 */
	private int							m_AcctX;

	/**
	 * Y dimension for m_AcctHistPopup
	 */
	private int							m_AcctY;

	/**
	 * X dimension for m_TrancheHistPopup
	 */
	private int							m_TrancheX;

	/**
	 * Y dimension for m_TrancheHistPopup
	 */
	private int							m_TrancheY;

	/**
	 * the current width of the graphics pane graph area in pixels
	 */
	private int							m_GraphWidth;

	/**
	 * starting history index (in m_AcctHistory or m_TrancheHistory)
	 */
	private int							m_StartIdx;

	/**
	 * ending history index (in m_AcctHistory or m_TrancheHistory)
	 */
	private int							m_EndIdx;

	/**
	 * max Y percent value seen in our data set
	 */
	private double						m_MaxYPct;

	/**
	 * max Y value seen in our data set
	 */
	private double						m_MaxYValue;

	/**
	 * inner class for handling popup window closures
	 */
	private class PopupCloser extends WindowAdapter {
		public void windowClosing(WindowEvent we) {
			JFrame popup = (JFrame) we.getWindow();
			m_Plugin.deregisterPopup(popup);
			if (popup == m_AcctHistPopup) {
				m_AcctHistPopup = null;
			}
			else if (popup == m_TrancheHistPopup) {
				m_TrancheHistPopup = null;
			}
		}
	}

	/**
	 * inner class to define graphs for account history data
	 */
	protected class AccountGraph extends JFrame {
		// constants
		/**
		 * constant for max char height in pixels of labeling
		 */
		private final int				M_MaxLabelHeight;

		/**
		 * constant for the minimum font size we're willing to use
		 */
		private final int				M_MinFontSize;

		// data members
		/**
		 * panel in which we want to generate graphics
		 */
		private JPanel					m_GraphPane;

		/**
		 * scroll pane to host graphics pane
		 */
		private JScrollPane			m_ScrollPane;

		/**
		 * the current scaling factor for the X-axis
		 */
		private double					m_ScaleX;

		/**
		 * the current scaling factor for the left side Y-axis (in BTC)
		 */
		private double					m_LtScaleY;

		/**
		 * the current scaling factor for the right side Y-axis (in percent)
		 */
		private double					m_RtScaleY;

		/**
		 * stroke for drawing lines on graph
		 */
		private BasicStroke			m_SolidStroke;

		/**
		 * the calculated font metrics required for labeling
		 */
		private FontMetrics			m_FontMetrics;

		/**
		 * horizontal font for axis labeling
		 */
		private Font					m_HorizFont;

		/**
		 * vertical font for axis labeling
		 */
		private Font					m_VertFont;

		/**
		 * point set for total percent value graph
		 */
		private GeneralPath			m_PctValue;

		/**
		 * point set for BTC chip value
		 */
		private GeneralPath			m_BTCChipValue;

		/**
		 * point set for IB ratio
		 */
		private GeneralPath			m_IBRatio;

		/**
		 * point set for IB value
		 */
		private GeneralPath			m_IBValue;

		/**
		 * point set for VFQ value
		 */
		private GeneralPath			m_VFQValue;

		/**
		 * point set for bottom stop values (%)
		 */
		private GeneralPath			m_BottomStop;

		/**
		 * point set for top stop (target) values (%)
		 */
		private GeneralPath			m_TopStop;

		/**
		 * point set for tranche count
		 */
		private GeneralPath			m_TrancheCnt;

		/**
		 * point set for axes (X, and Y on both sides)
		 */
		private GeneralPath			m_Axes;

		/**
		 * our graphics drawing object (associated with m_GraphPane)
		 */
		private Graphics2D			m_G2;

		/**
		 * constructor
		 * @param title the label for the window frame
		 */
		public AccountGraph(String title) {
			super(title);
			setOpaque(true);
			M_MaxLabelHeight = 15;
			M_MinFontSize = 6;
			setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
			setBackground(Color.WHITE);
			addWindowListener(new PopupCloser());
			//setLocationRelativeTo(m_TabManager);
			m_SolidStroke = new BasicStroke(1.0F);
			m_GraphPane = new JPanel();
			m_GraphPane.setBorder(null);
			m_GraphPane.setOpaque(true);
			m_ScrollPane = new JScrollPane(m_GraphPane);
			m_ScrollPane.setBorder(null);
			m_ScrollPane.setOpaque(true);
			m_ScrollPane.setVerticalScrollBarPolicy(
											ScrollPaneConstants.VERTICAL_SCROLLBAR_NEVER);
			m_ScrollPane.setHorizontalScrollBarPolicy(
									ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		}

		/**
		 * initializer
		 */
		public void init() {
			// init all paths, whether or not displayed
			int sz = m_AcctHistory.size();
			m_GraphWidth = m_AcctX;
			m_PctValue = new GeneralPath(GeneralPath.WIND_EVEN_ODD, sz);
			m_BTCChipValue = new GeneralPath(GeneralPath.WIND_EVEN_ODD, sz);
			m_IBRatio = new GeneralPath(GeneralPath.WIND_EVEN_ODD, sz);
			m_VFQValue = new GeneralPath(GeneralPath.WIND_EVEN_ODD, sz);
			m_BottomStop = new GeneralPath(GeneralPath.WIND_EVEN_ODD, 10);
			m_TopStop = new GeneralPath(GeneralPath.WIND_EVEN_ODD, 10);
			m_TrancheCnt = new GeneralPath(GeneralPath.WIND_EVEN_ODD, 5);
			m_Axes = new GeneralPath(GeneralPath.WIND_EVEN_ODD, 4);

			// init display size and contents
			Dimension dim = new Dimension(m_AcctX, m_AcctY);
			setSize(dim);
			m_ScrollPane.setPreferredSize(dim);
			setContentPane(m_ScrollPane);
		}

		/**
		 * method to draw the required graphs
		 */
		@Override
		public void paint(Graphics g) {
			Log.debug("In paint()");
			// draw to m_GraphPane, not in the JFrame (so we can scroll)
			m_G2 = (Graphics2D) m_GraphPane.getGraphics();
			if (m_G2 == null) {
				Log.debug("m_G2 null!");
				return;
			}
			m_G2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
										 RenderingHints.VALUE_ANTIALIAS_ON);
			m_G2.setStroke(m_SolidStroke);
			m_G2.setFont(m_DefaultFont);
			setVisible(true);

			// obtain data scaling factors
			m_ScaleX = getXscaling();
			m_RtScaleY = getYPercentScaling();
			m_LtScaleY = getYValueScaling();
			Dimension dim2 = new Dimension(m_GraphWidth, m_AcctY);
			Log.debug("Dim for graph = " + dim2);
			m_GraphPane.setPreferredSize(dim2);

			// draw X-axis (tosses), and both Y-axes, using scale of 1:1
			m_G2.scale(1.0, 1.0);
			if (m_Axes == null) {
				Log.debug("Scaled 1:1 but m_Axes null");
				return;
			}
			m_Axes.moveTo(M_HorizMargins, M_VertMargins);
			m_Axes.lineTo(M_HorizMargins, m_AcctY - 2 * M_VertMargins);
			m_Axes.lineTo(m_GraphWidth - M_HorizMargins,
							  m_AcctY - 2 * M_VertMargins);
			m_Axes.lineTo(m_GraphWidth - M_HorizMargins, M_VertMargins);
			m_G2.setPaint(Color.BLACK);
			m_G2.draw(m_Axes);

			// now label the axes
			// highest % on Y-axis right
			m_Df.setMinimumFractionDigits(0);
			m_Df.setMaximumFractionDigits(0);
			String topYrt = " +" + m_Df.format(m_MaxYPct) + "%";
			m_FontMetrics = pickFont(m_G2, topYrt, M_HorizMargins - 10);
			m_G2.setFont(m_HorizFont);
			final int lnHeight = m_FontMetrics.getHeight() / 3;
			int labX = m_GraphWidth - M_HorizMargins;
			int labY = M_VertMargins + lnHeight;
			m_G2.drawString(topYrt, labX, labY);

			// lowest possible % on Y-axis right
			final int ySpace = m_AcctY - 3 * M_VertMargins;
			String botYrt = " -100.0%";
			labY = ySpace + M_VertMargins + lnHeight;
			m_G2.drawString(botYrt, labX, labY);

			// zero point on Y-axis right
			String zeroY = " +0.0%";
			double zeroLoc = m_MaxYPct / (m_MaxYPct + 100.0);
			labY = (int) Math.round(ySpace * zeroLoc) + lnHeight;
			m_G2.drawString(zeroY, labX, labY);

			// draw ticks from max down to 0, stepping by 10% intervals
			int pixInterval = (labY - lnHeight - M_VertMargins) / 10;
			int pixDiff = 0;
			double yCnt = 0;
			while (yCnt < m_MaxYPct) {
				yCnt += 10.0;
				pixDiff += pixInterval;
				labY = M_VertMargins + pixDiff + lnHeight;
				String tick = " +" + m_Df.format(m_MaxYPct - yCnt) + "%";
				m_G2.drawString(tick, labX, labY);
			}
			m_Df.setMinimumFractionDigits(XMLVoucher.M_Precision);
			m_Df.setMaximumFractionDigits(XMLVoucher.M_Precision);

			// overall label down right Y-axis
			m_G2.setFont(m_DefaultFont);
			String labelY = "Value in percent";
			m_FontMetrics = pickFont(m_G2, labelY, ySpace);
			m_VertFont = genVertFont();
			m_G2.setFont(m_VertFont);
			labX = m_AcctX - M_MaxLabelHeight;
			labY = m_AcctY / 2 - m_FontMetrics.stringWidth(labelY);
			m_G2.drawString(labelY, labX, labY);
			m_G2.setFont(m_HorizFont);

			// if total percentage value is wanted, generate and display the curve
			ATabManager atm = (ATabManager) m_TabManager;
			ConfigPane cp = atm.getConfigPane();
			if (m_ShowTotalValue.isSelected()) {
				m_G2.scale(m_ScaleX, m_RtScaleY);
				Log.debug("Using scale " + m_Df.format(m_ScaleX) + ", "
							+ m_Df.format(m_RtScaleY));
				double totBank = cp.getBankroll();
				final double yInterval = m_MaxYPct + 100.0;
				// loop through history entries, calculating pixel location at each
				for (ABCBlock.AccountHistory aHist : m_AcctHistory) {
					int histIdx = m_AcctHistory.indexOf(aHist);
					if (histIdx < m_StartIdx) {
						continue;
					}
					if (histIdx > m_EndIdx) {
						break;
					}
					double totVal = aHist.m_BtcBal + aHist.m_BtcChipBal
										+ (aHist.m_IBBal * aHist.m_IBRatio);
					totBank = totBank + aHist.m_Additions - aHist.m_Subtractions;
					double value = (totVal / totBank) * 100.0;
					// NB: Y pixel values run from 0 downward
					Long pixY = Math.round(yInterval - value);
					Long pixX = Math.round((double) M_HorizMargins / m_ScaleX)
									+ (histIdx - m_StartIdx);
					if (histIdx == m_StartIdx) {
						m_PctValue.moveTo(pixX.intValue(), pixY.intValue());
					}
					else {
						m_PctValue.lineTo(pixX.intValue(), pixY.intValue());
					}
				/*
					Log.debug("Plotted " + (histIdx - m_StartIdx) + ", "
								+ m_Df.format(value) + "% at (" + pixX + ", "
								+ pixY + ")");
				 */
				}

				// now paint point set in black
				m_G2.setPaint(Color.BLACK);
				m_G2.draw(m_PctValue);
			}
			Log.debug("AHist: paint() completed");
		}

		/**
		 * helper font selector, by size
		 * @param g2 the graphics display
		 * @param strText the text for which we want a suitably sized font
		 * @param width the number of X pixels we're trying to stay within
		 * @return the new font metrics (also sets m_HorizFont to the picked font)
		 */
		private FontMetrics pickFont(Graphics2D g2, String strText, int width) {
			boolean fontFits = false;
			Font font = g2.getFont();
			FontMetrics fontMetrics = g2.getFontMetrics();
			int size = font.getSize();
			String name = font.getName();
			int style = font.getStyle();

			while (!fontFits) {
				if ((fontMetrics.getHeight() <= M_MaxLabelHeight)
					 && (fontMetrics.stringWidth(strText) <= width)) {
					fontFits = true;
				}
				else {
					if (size <= M_MinFontSize) {
						// can't go any smaller and stay readable
						fontFits = true;
					}
					else {
						// move down to next smallest font size
						g2.setFont(m_HorizFont = new Font(name, style, --size));
						fontMetrics = g2.getFontMetrics();
						g2.setFont(font);
					}
				}
			}
			return fontMetrics;
		}

		/**
		 * helper method to generate a vertical font corresponding to the
		 * current horizontal one
		 * @return the font generated
		 */
		private Font genVertFont() {
			if (m_HorizFont == null) {
				Log.error("No horizontal font, cannot generate matching vertical");
				return null;
			}
			AffineTransform fontAT = new AffineTransform();
			fontAT.rotate(Math.PI / 2.0);
			return m_HorizFont.deriveFont(fontAT);
		}
	}

	/**
	 * account history popup graphical display window
	 */
	private AccountGraph				m_AcctHistPopup;

	/**
	 * inner class to define graphs for tranche history data
	 */
	protected class TrancheGraph extends JFrame {
		/**
		 * constructor
		 * @param title the label for the window frame
		 */
		public TrancheGraph(String title) {
			super(title);
			setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
			setSize(new Dimension(m_TrancheX, m_TrancheY));
			m_GraphWidth = m_TrancheX;
			addWindowListener(new PopupCloser());
			setBackground(Color.WHITE.darker());
			//setLocationRelativeTo(m_TabManager);
			setVisible(true);
		}

		/**
		 * method to draw the required graphs
		 */
		@Override
		public void paint(Graphics g) {
			// obtain scaling factors
			double scaleX = getXscaling();
			double scaleY = getYPercentScaling();
			double ltScaleY = getYValueScaling();

			// TBD
		}
	}

	/**
	 * tranche history popup graphical display window
	 */
	private TrancheGraph				m_TrancheHistPopup;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the AHomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public HistoryPane(WalletClient plugin, AHomePane home, ATabManager mgr) {
		super(plugin, home, mgr);
		m_TrancheId = -1;
		m_InsetBorder = new EmptyBorder(3, 5, 3, 5);
		m_AcctX = m_TrancheX = 1000;
		m_AcctY = m_TrancheY = 700;
		M_HorizMargins = 50;
		M_VertMargins = 50;
		m_DefaultFont = new Font(Font.SANS_SERIF, Font.PLAIN, 20);

		// define instructions
		String histTxt = "This panel is used to select the data ranges and "
							+ "configuration options for retrieving historical value "
							+ "data from the server, and graphing it in a popup "
							+ "window.  This can be done for your entire account, or "
							+ "for an individual tranche.  Note that if your account "
							+ "is not currently betting, then any history data will "
							+ "refer to the previous betting session.<br/><br/>"
							+ "To graph all available data, select All tosses.  To "
							+ "zoom in on a subset, select Range and use the sliders "
							+ "to estimate the starting and ending tosses.  "
							+ "Tosses are shown in YYMMDDHHmm format.<br/><br/>"
							+ "Use the checkboxes to configure which data you want "
							+ "displayed.  The Graph button will query the ABC game "
							+ "server for your history data, and generate the graph "
							+ "in a separate window.  The Redraw button will update "
							+ "the graph according to your checkbox settings, but "
							+ "does not refresh history data from the server.";
		m_UserInstructs = histTxt;

		// define UI elements to select tosses
		m_AllTosses = new JRadioButton("All", true);
		m_AllTosses.setBorder(m_InsetBorder);
		m_AllTosses.setSelected(true);
		m_AllTosses.addActionListener(this);
		m_RangeTosses = new JRadioButton("Range (use sliders)", false);
		m_RangeTosses.setBorder(m_InsetBorder);
		m_RangeTosses.setSelected(false);
		m_RangeTosses.addActionListener(this);
		m_TossesGroup = new ButtonGroup();
		m_TossesGroup.add(m_AllTosses);
		m_TossesGroup.add(m_RangeTosses);
		ConfigPane cp = mgr.getConfigPane();
		int firstT = 0;
		int lastT = 0;
		if (cp != null) {
			firstT = cp.getFirstToss();
			lastT = cp.getLastToss();
		}
		m_StartToss = new JSlider(firstT, lastT, firstT);
		m_StartToss.setToolTipText("Select starting toss for range");
		m_StartToss.setMajorTickSpacing(5);
		m_StartToss.setSnapToTicks(true);
		m_StartToss.setEnabled(false);
		m_EndToss = new JSlider(firstT, lastT, lastT);
		m_EndToss.setToolTipText("Select ending toss for range");
		m_EndToss.setMajorTickSpacing(5);
		m_EndToss.setSnapToTicks(true);
		m_EndToss.setEnabled(false);

		// define UI elements to control graph configuration
		m_ShowStops = new JCheckBox("Show stops", true);
		m_ShowStops.setToolTipText("Display target and stop loss settings");
		m_ShowTrancheCnt = new JCheckBox("Show tranche count", false);
		m_ShowTrancheCnt.setToolTipText("Graph count of tranches in gray");
		m_ShowBTCChips = new JCheckBox("Show BTC chips", false);
		m_ShowBTCChips.setToolTipText("Graph BTC chips separately in blue");
		m_ShowIBValue = new JCheckBox("Show IB value", false);
		m_ShowIBValue.setToolTipText("Graph IB value in light green");
		m_ShowTotalValue = new JCheckBox("Show total value", true);
		m_ShowTotalValue.setToolTipText("Graph value in percent, in black");
		m_ShowVFQ = new JCheckBox("Show VFQ", false);
		m_ShowVFQ.setToolTipText("Graph VFQ separately in magenta");
		m_ShowAdditions = new JCheckBox("Show additions", false);
		m_ShowAdditions.setToolTipText("Show points where value was added");
		m_ShowSubtractions = new JCheckBox("Show subtractions", false);
		m_ShowSubtractions.setToolTipText("Show points where value was removed");
		m_ShowIBRatio = new JCheckBox("Show IB ratio", true);
		m_ShowIBRatio.setToolTipText("Graph IB reserve backing ratio in yellow");
		m_ShowSubsidy = new JCheckBox("Show subsidy", false);
		m_ShowSubsidy.setToolTipText("Show when tranche was receiving bonus IB");

		// define graph action and button
		setGraphAction();
		m_GraphButton = new JButton(m_GraphAction);
		m_GraphButton.setFont(m_TabManager.M_ButtonFont);
		m_GraphButton.setToolTipText("Retrieve the indicated history range and "
											+ "graph data in a separate window");
		m_GraphButton.setEnabled(true);

		// define refresh action and button
		setRedrawAction();
		m_RedrawButton = new JButton(m_RedrawAction);
		m_RedrawButton.setFont(m_TabManager.M_ButtonFont);
		m_RedrawButton.setToolTipText("Redisplay the same data set using the "
												+ "current option settings");
		m_RedrawButton.setEnabled(true);

		// populate the button pane
		m_ButtonPane.removeAll();
		m_ButtonPane.add(Box.createHorizontalStrut(110));	// spacer
		m_ButtonPane.add(m_GraphButton);
		m_ButtonPane.add(Box.createHorizontalStrut(120));	// spacer
		m_ButtonPane.add(m_RedrawButton);

		// init main panel area
		m_HistPanel = new JPanel();
		m_HistPanel.setOpaque(true);
		m_HistPanel.setBorder(null);
		m_HistPanel.setLayout(new GridBagLayout());
		m_HistPanel.setPreferredSize(new Dimension(460, 250));
		m_HistBack = Color.CYAN;
		m_HistPanel.setBackground(m_HistBack);

		m_Df = (DecimalFormat) NumberFormat.getInstance();
		m_Df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		m_Df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		m_Df.setGroupingSize(3);
	}

	/**
	 * do display of history get/graph
	 */
	public void prepDisplay() {
		// set screen label
		buildScreenLabel("Fetch and graph historical value data.");
		//Color front = new Color(0xEEEECC);

		// inits
		AHomePane ahp = (AHomePane) m_ParentHome;
		ATabManager atm = (ATabManager) m_TabManager;
		m_HistPanel.removeAll();
		m_GBC.anchor = GridBagConstraints.NORTH;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_GBC.fill = GridBagConstraints.HORIZONTAL;
		if (m_HistMode) {
			m_GBC.gridheight = 13;
		}
		else {
			m_GBC.gridheight = 11;
		}
		m_HistPanel.add(Box.createHorizontalGlue(), m_GBC);

		// row 1: show whether this is for a given tranche or the entire account
		JLabel histLab = new JLabel("Graphically display value history for: ");
		histLab.setFont(ahp.M_LabelFont);
		histLab.setHorizontalAlignment(SwingConstants.TRAILING);
		histLab.setBorder(m_InsetBorder);
		//histLab.setForeground(front);
		JTextField histFld = new JTextField(3);
		histFld.setDropTarget(null);
		histFld.setEditable(false);
		histFld.setOpaque(false);
		histFld.setHorizontalAlignment(JTextField.LEFT);
		histFld.setBorder(m_InsetBorder);
		histFld.setFont(ahp.M_LabelFont);
		histFld.setBackground(m_HistBack);
		histFld.setFocusable(false);
		//histFld.setForeground(front);
		if (m_HistMode) {
			histFld.setText("your account.");
			histFld.setToolTipText("Graphs will display overall account value");
		}
		else {
			histFld.setText("tranche ID " + m_TrancheId);
			histFld.setToolTipText("Graphs will display value for this tranche");
		}
		histLab.setLabelFor(histFld);
		firstItemConstraints(m_GBC);
		m_HistPanel.add(histLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_HistPanel.add(histFld, m_GBC);

		// row 2: toss mode label
		JLabel tossLab = new JLabel("Tosses to display: ");
		tossLab.setFont(ahp.M_LabelFont);
		tossLab.setHorizontalAlignment(SwingConstants.TRAILING);
		tossLab.setBorder(m_InsetBorder);
		tossLab.setLabelFor(m_AllTosses);
		firstItemConstraints(m_GBC);
		m_HistPanel.add(tossLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_HistPanel.add(Box.createHorizontalStrut(100), m_GBC);

		// row 3-4: toss selector
		firstItemConstraints(m_GBC);
		m_HistPanel.add(m_AllTosses, m_GBC);
		lastItemConstraints(m_GBC);
		m_HistPanel.add(m_RangeTosses, m_GBC);

		// row 5: start selection slider label
		JLabel startLab = new JLabel("Starting toss: ");
		startLab.setFont(ahp.M_LabelFont);
		startLab.setHorizontalAlignment(SwingConstants.TRAILING);
		startLab.setBorder(m_InsetBorder);
		startLab.setLabelFor(m_StartToss);
		startLab.setToolTipText("Defaults to the first toss on which you bet");
		firstItemConstraints(m_GBC);
		m_HistPanel.add(startLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_HistPanel.add(m_StartToss, m_GBC);

		// row 6: end selection slider
		JLabel endLab = new JLabel("Ending toss: ");
		endLab.setFont(ahp.M_LabelFont);
		endLab.setHorizontalAlignment(SwingConstants.TRAILING);
		endLab.setBorder(m_InsetBorder);
		endLab.setLabelFor(m_EndToss);
		endLab.setToolTipText("Defaults to the last toss on which you bet");
		firstItemConstraints(m_GBC);
		m_HistPanel.add(endLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_HistPanel.add(m_EndToss, m_GBC);

		// row 7: toss range
		JLabel rangeLab = new JLabel("Toss range available: ");
		rangeLab.setFont(ahp.M_LabelFont);
		rangeLab.setHorizontalAlignment(SwingConstants.TRAILING);
		if (m_EndToss.getMaximum() == 0) {
			rangeLab.setToolTipText("To establish range, request All history "
											+ "first, or visit the Config page.");
		}
		else {
			rangeLab.setToolTipText("Format is YYMMDDHHmm, e.g. 1608242105 = "
											+ "2016 Aug 24th 21:05 UTC");
		}
		rangeLab.setBorder(m_InsetBorder);
		JTextField rangeFld = new JTextField(10);
		rangeFld.setDropTarget(null);
		rangeFld.setEditable(false);
		rangeFld.setOpaque(false);
		rangeFld.setHorizontalAlignment(JTextField.LEFT);
		rangeFld.setBorder(m_InsetBorder);
		rangeFld.setFont(new Font(Font.SANS_SERIF, Font.PLAIN, 10));
		rangeFld.setBackground(m_HistBack);
		rangeFld.setFocusable(false);
		rangeLab.setLabelFor(rangeFld);
		rangeFld.setText(m_StartToss.getMinimum() + " : "
							+ m_EndToss.getMaximum());
		rangeFld.setToolTipText("Values correspond to left and right ends of the "
										+ "slider range");
		firstItemConstraints(m_GBC);
		m_HistPanel.add(rangeLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_HistPanel.add(rangeFld, m_GBC);

		// row 8: total value and IB ratio checkboxes
		firstItemConstraints(m_GBC);
		m_HistPanel.add(m_ShowTotalValue, m_GBC);
		lastItemConstraints(m_GBC);
		m_HistPanel.add(m_ShowIBRatio, m_GBC);

		// row 9: btc chips and IB value checkboxes
		firstItemConstraints(m_GBC);
		m_HistPanel.add(m_ShowBTCChips, m_GBC);
		lastItemConstraints(m_GBC);
		m_HistPanel.add(m_ShowIBValue, m_GBC);

		if (!m_HistMode) {
			// in tranche mode, just one more row (10) for additions and subsidy
			firstItemConstraints(m_GBC);
			m_HistPanel.add(m_ShowAdditions, m_GBC);
			lastItemConstraints(m_GBC);
			m_HistPanel.add(m_ShowSubsidy, m_GBC);
		}
		else {
			// row 10: stops and VFQ checkboxes
			firstItemConstraints(m_GBC);
			m_HistPanel.add(m_ShowStops, m_GBC);
			lastItemConstraints(m_GBC);
			m_HistPanel.add(m_ShowVFQ, m_GBC);

			// row 11: additions and removals checkboxes
			firstItemConstraints(m_GBC);
			m_HistPanel.add(m_ShowAdditions, m_GBC);
			lastItemConstraints(m_GBC);
			m_HistPanel.add(m_ShowSubtractions, m_GBC);

			// row 12: tranche count checkbox
			firstItemConstraints(m_GBC);
			m_HistPanel.add(m_ShowTrancheCnt, m_GBC);
			lastItemConstraints(m_GBC);
			m_HistPanel.add(Box.createHorizontalStrut(100), m_GBC);
		}
		m_HistPanel.add(Box.createVerticalGlue(), m_GBC);

		// put entire form into the main scroll area
		m_WorkArea.setViewportView(m_HistPanel);
		revalidate();
	}

	/**
	 * method to define Graph action
	 */
	private void setGraphAction() {
		m_GraphAction = new AbstractAction("Graph") {
			public void actionPerformed(ActionEvent ae) {
				String abcId = m_Plugin.getLoginSecrets().getSVMId();
				m_TabManager.showError("cannot show history at " + abcId,
											"this feature is TBD",
									"try again using a later version of the ABC plugin");
			/*
				String sessId = m_TabManager.getLoginPane().getSessionId();
				ABCClientBlock histBlock = new ABCClientBlock();
				histBlock.setSessionId(sessId);
				ABCListener abcListener = (ABCListener) m_Plugin.getSVMListener();
				setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

				if (m_HistMode) {
					histBlock.setOpcode("REQ_account_hist");
					// send to server; answer will go to recordAcctHistory()
					if (!abcListener.sendAsyncRequest(histBlock)) {
						Log.error("Acct history request failed to ABC Id " + abcId);
						m_TabManager.showError("cannot get account history at ABC Id "
														+ abcId,
												"failure sending history request message");
					}
				}
				else {
					histBlock.setOpcode("REQ_tranche_hist");
					histBlock.setTrancheId(m_TrancheId);
					if (!abcListener.sendAsyncRequest(histBlock)) {
						Log.error("Tranche hist request failed to ABC Id " + abcId);
						m_TabManager.showError("cannot get tranche history at ABC Id "
														+ abcId,
												"failure sending tranche history request");
					}
				}
			 */
			}
		};
	}

	/**
	 * method to define Redraw action
	 */
	private void setRedrawAction() {
		m_RedrawAction = new AbstractAction("Redraw") {
			public void actionPerformed(ActionEvent ae) {
				String abcId = m_Plugin.getLoginSecrets().getSVMId();
				m_TabManager.showError("cannot show history at " + abcId,
												"this feature is TBD",
									"try again using a later version of the ABC plugin");
			/*
				if (m_HistMode) {
					if (m_AcctHistPopup != null) {
						m_AcctHistPopup.dispose();
					}
					graphAcctHistory();
				}
				else {
					if (m_TrancheHistPopup != null) {
						m_TrancheHistPopup.dispose();
					}
					graphTrancheHistory();
				}
			 */
			}
		};
	}

	/**
	 * record account history results (called by ABCListener)
	 * @param acctHist the message returned from the ABC server
	 */
	public void recordAcctHistory(ABCBlock acctHist) {
		setCursor(null);
		if (acctHist == null) {
			Log.error("Missing account history data");
			return;
		}
		m_AcctHistory = acctHist.getAccountHistory();
		if (m_AcctHistory.isEmpty()) {
			// account has never been betting
			m_TabManager.showInfo("Your account has no betting history.");
			return;
		}

		if (!m_RangeTosses.isSelected()) {
			// extract first and last tosses, and configure slider ranges
			ABCBlock.AccountHistory fHist = m_AcctHistory.get(0);
			ABCBlock.AccountHistory lHist
				= m_AcctHistory.get(m_AcctHistory.size()-1);
			configTossRange(fHist.m_Toss, lHist.m_Toss);
		}

		// graph the data in a separate Swing thread
		Runnable aDraw = new Runnable() {
			public void run() {
				graphAcctHistory();
			}
		};
		SwingUtilities.invokeLater(aDraw);
	}

	/**
	 * record tranche history results (called by ABCListener)
	 * @param trancheHist the message returned from the ABC server
	 */
	public void recordTrancheHistory(ABCBlock trancheHist) {
		setCursor(null);
		if (trancheHist == null) {
			Log.error("Missing tranche history data");
			return;
		}
		m_TrancheHistory = trancheHist.getTrancheHistory();
		if (m_TrancheHistory.isEmpty()) {
			// tranche has never been betting (odd, since it can be inactive)
			m_TabManager.showWarning("This tranche has no betting history.");
			return;
		}

		// extract first and last tosses, and configure slider ranges
		ABCBlock.TrancheHistory fHist = m_TrancheHistory.get(0);
		ABCBlock.TrancheHistory lHist
			= m_TrancheHistory.get(m_TrancheHistory.size()-1);
		m_StartToss.setMinimum(fHist.m_Toss);
		m_StartToss.setMaximum(lHist.m_Toss);
		m_StartToss.setValue(fHist.m_Toss);
		m_EndToss.setMinimum(fHist.m_Toss);
		m_EndToss.setMaximum(lHist.m_Toss);
		m_EndToss.setValue(lHist.m_Toss);

		// graph the data in a separate Swing thread
		Runnable tDraw = new Runnable() {
			public void run() {
				graphTrancheHistory();
			}
		};
		SwingUtilities.invokeLater(tDraw);
	}

	/**
	 * set the tranche ID selected
	 * @param tId the tranche Id (from the HoldingsPane)
	 */
	public void setTrancheId(int tId) {
		m_TrancheId = tId;
	}

	/**
	 * set the history display mode
	 * @param mode true if account mode, false if tranche mode
	 */
	public void setHistMode(boolean mode) { m_HistMode = mode; }

	/**
	 * set the sliders for the start and end tosses
	 * @param first the first toss we were betting
	 * @param last the last toss we were betting
	 */
	public void configTossRange(int first, int last) {
		// only applicable for account toss ranges
		if (m_HistMode && first > 0 && last >= first) {
			// before updating values, remove any labels
			m_StartToss.setMinimum(first);
			m_StartToss.setMaximum(last);
			m_StartToss.setValue(first);
			m_StartToss.revalidate();
			m_EndToss.setMinimum(first);
			m_EndToss.setMaximum(last);
			m_EndToss.setValue(last);
			m_EndToss.revalidate();
		}
	}

	/**
	 * method to implement ActionListener (overriding WorkPane), used to
	 * activate sliders whenever m_RangeTosses is selected
	 */
	@Override
	public void actionPerformed(ActionEvent ae) {
		Object src = ae.getSource();
		if (src instanceof JRadioButton) {
			JRadioButton radSrc = (JRadioButton) src;
			if (radSrc == m_RangeTosses) {
				m_StartToss.setEnabled(true);
				m_EndToss.setEnabled(true);
			}
			else {
				m_StartToss.setEnabled(false);
				m_EndToss.setEnabled(false);
			}
		}
	}

	/**
	 * helper method to prepare GridBagConstraints for use with the
	 * first item on a line (such as a prompt/label)
	 */
	private void firstItemConstraints(GridBagConstraints gbc) {
		if (gbc == null) {
			return;
		}
		gbc.gridwidth = GridBagConstraints.RELATIVE;
		gbc.fill = GridBagConstraints.NONE;
	}

	/**
	 * helper method to prepare GridBagConstraints for use with the
	 * last item on a line (such as an entry field)
	 */
	private void lastItemConstraints(GridBagConstraints gbc) {
		if (gbc == null) {
			return;
		}
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.fill = GridBagConstraints.HORIZONTAL;
	}

	/**
	 * helper method to scale and draw the popup window for account history
	 */
	private void graphAcctHistory() {
		// if we don't already have one, create popup window
		if (m_AcctHistPopup == null) {
			m_AcctHistPopup = new AccountGraph("Account Betting History");
			// register with wallet plugin for closure on logout
			m_Plugin.registerPopup(m_AcctHistPopup);
		}
		m_AcctHistPopup.init();
		m_AcctHistPopup.pack();
		m_AcctHistPopup.setVisible(true);
	}

	/**
	 * helper method to scale and draw the popup graph for tranche history
	 */
	private void graphTrancheHistory() {
		// if we don't already have one, create popup window
		if (m_TrancheHistPopup == null) {
			m_TrancheHistPopup = new TrancheGraph("Tranche ID " + m_TrancheId
																+ " Betting History");
			m_Plugin.registerPopup(m_TrancheHistPopup);
		}
		m_TrancheHistPopup.pack();
	}

	/**
	 * helper method to determine the X-axis scaling in pixels per toss
	 * @return the number of pixels allocated for toss (min. 1.0)
	 */
	private double getXscaling() {
		double pixPerToss = 1.0;
		// take into account how many tosses we have, possibly modified by sliders
		int graphedTossCnt = numGraphedTosses();
		// allow space on each side for margins and axes
		int xSpace = 2 * M_HorizMargins;
		int limit = m_HistMode ? m_AcctX : m_TrancheX;
		if (graphedTossCnt > limit - xSpace) {
			m_XScrolling = true;
			m_GraphWidth = xSpace + graphedTossCnt;
		}
		else {
			m_XScrolling = false;
			pixPerToss = (double) (limit - xSpace) / (double) graphedTossCnt;
			m_GraphWidth = limit + xSpace;
		}
		return Math.max(1.0, pixPerToss);
	}

	/**
	 * helper method to count the number of tosses in history data between
	 * the values of the sliders.  Also sets start and end indices correctly.
	 * @return the number of tosses whose data we'll be graphing
	 */
	private int numGraphedTosses() {
		int rawTosses = 0;
		int graphedTosses = 0;
		int minToss = m_StartToss.getMinimum();
		int maxToss = m_EndToss.getMaximum();
		int startToss = minToss;
		int endToss = maxToss;
		// if range selector mode, override with slider values
		if (m_RangeTosses.isSelected()) {
			startToss = m_StartToss.getValue();
			endToss = m_EndToss.getValue();
		}
		if (startToss > endToss) {
			// silently reverse them
			m_StartToss.setValue(endToss);
			m_EndToss.setValue(startToss);
			endToss = m_EndToss.getValue();
			startToss = m_StartToss.getValue();
		}

		if (m_HistMode) {
			// doing account history
			rawTosses = m_AcctHistory.size();
			if (startToss != minToss || endToss != maxToss) {
				// user specified a range, so loop through
				boolean started = false;
				m_EndIdx = m_AcctHistory.size() - 1;
				for (ABCBlock.AccountHistory aHist : m_AcctHistory) {
					if (aHist.m_Toss >= startToss) {
						graphedTosses++;
						if (!started) {
							m_StartIdx = m_AcctHistory.indexOf(aHist);
							started = true;
						}
						if (aHist.m_Toss >= endToss) {
							m_EndIdx = m_AcctHistory.indexOf(aHist);
							break;
						}
					}
				}
			}
			else {
				// graph it all
				graphedTosses = rawTosses;
				m_StartIdx = 0;
				m_EndIdx = m_AcctHistory.size() - 1;
			}
			Log.debug("AHist: graphing " + graphedTosses + " from " + startToss
						+ " (index " + m_StartIdx + ") to " + endToss + " (index "
						+ m_EndIdx + ")");
		}
		else {
			// doing tranche history
			rawTosses = m_TrancheHistory.size();

			/* Because tranches are created and destroyed during betting, we
			 * cannot assume that this tranche starts or stops at a particular
			 * point corresponding to the first and last tosses for the account.
			 * So always loop through.
			 */
			m_StartIdx = 0;
			m_EndIdx = m_TrancheHistory.size() - 1;
			boolean started = false;
			for (ABCBlock.TrancheHistory tHist : m_TrancheHistory) {
				if (tHist.m_Toss >= startToss) {
					graphedTosses++;
					if (!started) {
						m_StartIdx = m_TrancheHistory.indexOf(tHist);
						started = true;
					}
					if (tHist.m_Toss >= endToss) {
						m_EndIdx = m_TrancheHistory.indexOf(tHist);
						break;
					}
				}
			}
			Log.debug("THist: graphing " + graphedTosses + " from " + startToss
						+ " (index " + m_StartIdx + ") to " + endToss + " (index "
						+ m_EndIdx + ")");
		}
		return graphedTosses;
	}

	/**
	 * helper method to determine the right Y-axis scaling in pixels per percent
	 * @return the number of pixels allocated for each percentage point (min. 1)
	 * (also sets m_MaxYPct)
	 */
	private double getYPercentScaling() {
		double pixPerPct = 1.0;
		m_MaxYPct = getMaxYPercent();
		// allow space on top and bottom for margins and X-axis
		double limit = m_HistMode ? m_AcctY : m_TrancheY;
		double calcPct = (limit - 2 * M_VertMargins) / (m_MaxYPct + 100.0);
		pixPerPct = Math.max(pixPerPct, calcPct);
		return pixPerPct;
	}

	/**
	 * helper method to determine the top end of the scale for the Y axis
	 * @return the highest percentage value we'll need to graph
	 */
	private double getMaxYPercent() {
		double rawMax = 0.0;
		ATabManager atm = (ATabManager) m_TabManager;
		if (m_HistMode) {
			ConfigPane cp = atm.getConfigPane();
			double startingBank = cp.getBankroll();
			// loop through history entries, calculating % value at each
			for (ABCBlock.AccountHistory aHist : m_AcctHistory) {
				double totVal = aHist.m_BtcBal + aHist.m_BtcChipBal
									+ (aHist.m_IBBal * aHist.m_IBRatio);
				double totBank
					= startingBank + aHist.m_Additions - aHist.m_Subtractions;
				rawMax = Math.max(rawMax, totVal / totBank);
			}
			Log.debug("AHist: calc max Y % = " + m_Df.format(rawMax * 100.0));
		}
		else {
			// find the tranche we're working with
			HoldingsPane hp = atm.getHoldingsPane();
			ABCTranche tranche = hp.getSelectedTranche();
			if (tranche == null) {
				Log.error("No tranche selected graphing tranche history!");
				return 100.0;
			}
			double startingBank = tranche.getBTCStart();

			// loop through history entries, calculating % value at each
			for (ABCBlock.TrancheHistory tHist : m_TrancheHistory) {
				double totVal = tHist.m_BtcBal + tHist.m_BtcChipBal
									+ (tHist.m_IBBal * tHist.m_IBRatio);
				double totBank = startingBank + tHist.m_Additions;
				rawMax = Math.max(rawMax, totVal / totBank);
			}
			Log.debug("THist: calc max Y % = " + m_Df.format(rawMax * 100.0));
		}

		// never return less than 100%
		return Math.max(rawMax * 100.0, 100.0);
	}

	/**
	 * helper method to determine the left Y-axis scaling in pixels per BTC
	 * @return the number of pixels allocated for each BTC unit
	 * (also sets m_MaxYValue)
	 */
	private double getYValueScaling() {
		double pixPerBtc = 1.0;
		// allow 50 on bottom for margin and X-axis
		double m_MaxYValue = getMaxYValue();
		if (m_HistMode) {
			pixPerBtc = (m_AcctY - 2 * M_VertMargins) / m_MaxYValue;
		}
		else {
			pixPerBtc = (m_TrancheY - 2 * M_VertMargins) / m_MaxYValue;
		}
		return pixPerBtc;
	}

	/**
	 * helper method to calculate the max value in BTC which needs to be shown
	 * @return the highest BTC value we need to represent on the graph
	 */
	private double getMaxYValue() {
		double rawMax = 0.0;
		if (m_HistMode) {
			// loop through history entries, calculating value at each
			for (ABCBlock.AccountHistory aHist : m_AcctHistory) {
				double totVal = aHist.m_BtcBal + aHist.m_BtcChipBal
									+ (aHist.m_IBBal * aHist.m_IBRatio);
				rawMax = Math.max(rawMax, totVal);
			}
			Log.debug("AHist: calc max Y = " + m_Df.format(rawMax));
		}
		else {
			// loop through history entries, calculating value at each
			for (ABCBlock.TrancheHistory tHist : m_TrancheHistory) {
				double totVal = tHist.m_BtcBal + tHist.m_BtcChipBal
									+ (tHist.m_IBBal * tHist.m_IBRatio);
				rawMax = Math.max(rawMax, totVal);
			}
			Log.debug("THist: calc max Y = " + m_Df.format(rawMax));
		}
		return rawMax;
	}

	// END methods
}
