/*
 * last modified---
 * 	08-26-16/jet/ add display of VFQ balance
 * 	08-16-16/jet/ add History button
 * 	04-30-15/jet/ Spark plugin version
 * 	04-15-15/jet/ add m_TossPosition
 * 	04-14-15/jet/ appearance changes
 * 	04-03-15/jet/ complete tranche manipulation sections
 * 	03-31-15/jet/ include BTC allocated for open IB buy orders when computing
 * 					  net position standing
 * 	01-22-15/jet/ new
 *
 * purpose---
 * 	provide a UI screen to represent ABC account holdings
 */

package org.silentvault.spark.ui.svm.abc;

import javax.swing.*;
import java.awt.event.MouseEvent;
import java.awt.event.ActionEvent;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Font;
import java.awt.Point;
import java.awt.Component;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;
import org.silentvault.spark.ui.svm.*;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;


/**
 * Extends AWorkPane to provide a screen for the ABC holdings (main menu) page.
 *
 * @author Justin Turrell
 */
public final class HoldingsPane extends AWorkPane {
	// BEGIN data members
	/**
	 * tool tips for tranche table header columns
	 */
	protected final String[]		M_TrancheTooltips =
		{	"Click on a radio button to select a tranche",
			"The unique number of each tranche within your account",
			"The starting amount of BTC with which you funded each tranche",
			"The current quantity of BTC held by each tranche",
			"The current quantity of InsurBucks held by each tranche",
			"The estimated cash-out value, as a percentage of starting funding",
			"The current Martingale betting iteration",
			"The current Martingale betting increment",
			"The number of consecutive losses experienced this Martingale cycle",
			"ID of any open order to buy or sell IB (0 means no order)",
			"Whether the tranche is eligible for bonus IB on losses"	};

	/**
	 * font for use in table columns
	 */
	public final Font					M_TableFont;

	/**
	 * action to Logout from the user's account
	 */
	private AbstractAction			m_LogoutAction;

	/**
	 * action to Rebalance a selected tranche
	 */
	private AbstractAction			m_RebalanceAction;

	/**
	 * action to TopUp a selected tranche
	 */
	private AbstractAction			m_TopUpAction;

	/**
	 * action to Liquidate a selected tranche
	 */
	private AbstractAction			m_LiquidateAction;

	/**
	 * action to request History for a selected tranche
	 */
	private AbstractAction			m_HistoryAction;

	/**
	 * action to Refresh holdings
	 */
	private AbstractAction			m_RefreshAction;

	/**
	 * button for logout action
	 */
	private JButton					m_LogoutButton;

	/**
	 * button for rebalance action
	 */
	private JButton					m_RebalanceButton;

	/**
	 * button for TopUp action
	 */
	private JButton					m_TopUpButton;

	/**
	 * button for Liquidate action
	 */
	private JButton					m_LiquidateButton;

	/**
	 * button for History action
	 */
	private JButton					m_HistoryButton;

	/**
	 * button for Refresh action
	 */
	private JButton					m_RefreshButton;

	/**
	 * the panel containing the general holdings information
	 */
	private JPanel						m_HoldingsPanel;

	/**
	 * the table which contains the tranches shown
	 */
	private JTable						m_TranchesTable;

	/**
	 * the currently selected tranche, or null if one is not selected
	 */
	private ABCTranche				m_SelTranche;

	/**
	 * mode flag controlling behavior of prepDisplay() to show amount to deploy
	 * for tranche top ups (if true), or regular display (if false)
	 */
	private boolean					m_TopUpMode;

	/**
	 * the amount (in SBC) currently available for deployment to top up tranches
	 */
	private double						m_TopUpAvail;

	/**
	 * the UI element for entering topup amount to deploy
	 */
	private JTextField				m_TopUp;

	/**
	 * record of the latest computed total position as of current toss
	 */
	private double						m_TossPosition;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the AHomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public HoldingsPane(WalletClient plugin, AHomePane home, ATabManager mgr) {
		super(plugin, home, mgr);

		// define instructions
		String holdingsTxt = "This panel shows various statistics about your ABC "
								+ "account.  If you have established any tranches, "
								+ "these will be listed in the table.  See also the "
								+ "Config screen for more information.<br/><br/>"
								+ "You may select a tranche to TopUp or Liquidate, or "
								+ "view its History.<br/><br/>Use the Refresh button "
								+ "after a toss to obtain the latest values.  The "
								+ "Logout button will log you out of your ABC account.";
		m_UserInstructs = holdingsTxt;

		// init font value
		M_TableFont = new Font(Font.SANS_SERIF, Font.PLAIN, 13);

		// set up tranches table model (empty) and table
		m_TranchesTable = mkNewTranchesTable();

		// define rebalance action and button
		setRebalanceAction();
		m_RebalanceButton = new JButton(m_RebalanceAction);
		m_RebalanceButton.setFont(m_TabManager.M_ButtonFont);
		m_RebalanceButton.setToolTipText("Rebalance the BTC vs. IB in the "
													+ "selected tranche");
		m_RebalanceButton.setEnabled(false);

		// define topup action and button
		setTopUpAction();
		m_TopUpButton = new JButton(m_TopUpAction);
		m_TopUpButton.setFont(m_TabManager.M_ButtonFont);
		m_TopUpButton.setToolTipText("Add more BTC to the selected tranche");
		m_TopUpButton.setEnabled(false);

		// define liquidate action and button
		setLiquidateAction();
		m_LiquidateButton = new JButton(m_LiquidateAction);
		m_LiquidateButton.setFont(m_TabManager.M_ButtonFont);
		m_LiquidateButton.setToolTipText("Cash out the selected tranche");
		m_LiquidateButton.setEnabled(false);

		// define history action and button
		setHistoryAction();
		m_HistoryButton = new JButton(m_HistoryAction);
		m_HistoryButton.setFont(m_TabManager.M_ButtonFont);
		m_HistoryButton.setToolTipText("View history for the selected tranche");
		m_HistoryButton.setEnabled(false);

		// define refresh action and button
		setRefreshAction();
		m_RefreshButton = new JButton(m_RefreshAction);
		m_RefreshButton.setFont(m_TabManager.M_ButtonFont);
		m_RefreshButton.setToolTipText("Query the game server for latest data");

		// define logout action
		m_LogoutAction = new AbstractAction("Logout") {
			public void actionPerformed(ActionEvent ae) {
				// logout and return to login pane
				m_TabManager.queueTransition(
												new Integer(MTabManager.M_LOGOUT_MARKET));
				m_TabManager.makeTransition();
			}
		};
		// define matching button
		m_LogoutButton = new JButton(m_LogoutAction);
		m_LogoutButton.setFont(m_TabManager.M_ButtonFont);
		m_LogoutButton.setToolTipText("Log out of your ABC account");

		// popuplate the button pane
		m_ButtonPane.removeAll();
		m_ButtonPane.add(Box.createHorizontalStrut(5));	// spacer
		//m_ButtonPane.add(m_RebalanceButton);
		m_ButtonPane.add(Box.createHorizontalStrut(10));	// spacer
		m_ButtonPane.add(m_TopUpButton);
		m_ButtonPane.add(Box.createHorizontalStrut(10));	// spacer
		m_ButtonPane.add(m_LiquidateButton);
		m_ButtonPane.add(Box.createHorizontalStrut(10));	// spacer
		m_ButtonPane.add(m_HistoryButton);
		m_ButtonPane.add(Box.createHorizontalStrut(10));	// spacer
		m_ButtonPane.add(m_RefreshButton);
		m_ButtonPane.add(Box.createHorizontalStrut(10));	// spacer
		m_ButtonPane.add(m_LogoutButton);
		m_ButtonPane.add(Box.createHorizontalStrut(5));	// spacer

		m_HoldingsPanel = new JPanel();
		m_HoldingsPanel.setOpaque(true);
		m_HoldingsPanel.setBorder(null);
		m_HoldingsPanel.setLayout(new GridLayout(3, 4, 5, 5));
		m_HoldingsPanel.setPreferredSize(new Dimension(460, 65));
		m_HoldingsPanel.setBackground(Color.ORANGE);

		// topup amount UI
		m_TopUp = new JTextField(8);
		m_TopUp.setDropTarget(null);
		m_TopUp.setHorizontalAlignment(JTextField.LEFT);
		Dimension entrySz = new Dimension(30, 20);
		m_TopUp.setPreferredSize(entrySz);
		m_TopUp.setText("0.0");
		m_TopUp.setToolTipText("<html>Enter the specific amount you wish<br/>"
										+ "to deploy to top up a tranche.</html>");
	}

	/**
	 * do display of holdings
	 */
	public void prepDisplay() {
		// set screen label
		buildScreenLabel("Holdings for your account, including tranches");

		// deactivate proper buttons
		AHomePane ahp = (AHomePane) m_ParentHome;
		final TrancheTableModel ttm = ahp.getTrancheModel();
		if (!m_TopUpMode) {
			ttm.setSelectMode(TrancheTableModel.M_SelectSingle);
			ttm.deselectAll();
			trancheSelectChanged();
		}

		// inits
		m_TopUpAvail = ahp.getAvailSBC();
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(3);
		m_HoldingsPanel.removeAll();

		if (m_TopUpMode) {
			// now add the topup entry item
			m_TopUp.setText("0.0");
			JLabel topupLab = new JLabel("BTC to use: ");
			topupLab.setFont(ahp.M_LabelFont);
			topupLab.setHorizontalAlignment(SwingConstants.TRAILING);
			topupLab.setBorder(m_InsetBorder);
			topupLab.setLabelFor(m_TopUp);
			final Dimension entrySz = new Dimension(30, 20);
			JTextField topAvail = new JTextField(8);
			topAvail.setDropTarget(null);
			topAvail.setEditable(false);
			topAvail.setOpaque(false);
			topAvail.setHorizontalAlignment(JTextField.LEFT);
			topAvail.setBorder(m_InsetBorder);
			topAvail.setBackground(m_Plugin.m_LabelBackColor);
			topAvail.setFocusable(false);
			topAvail.setPreferredSize(entrySz);
			topAvail.setText("( " + df.format(m_TopUpAvail) + " BTC )");
			topAvail.setToolTipText("The amount of BTC available for top up");
			m_HoldingsPanel.setLayout(new GridLayout(1, 3, 5, 5));
			m_HoldingsPanel.setPreferredSize(new Dimension(460, 25));
			m_HoldingsPanel.add(topupLab);
			m_HoldingsPanel.add(m_TopUp);
			m_HoldingsPanel.add(topAvail);
			m_LiquidateButton.setEnabled(false);
			m_RebalanceButton.setEnabled(false);
			m_HistoryButton.setEnabled(false);
			if (m_TopUpAvail > 0.0) {
				m_TopUpButton.setEnabled(true);
			}
			else {
				m_TopUpButton.setEnabled(false);
			}
		}
		else {
			// begin with amount of free SBC
			m_HoldingsPanel.setLayout(new GridLayout(3, 4, 5, 5));
			m_HoldingsPanel.setPreferredSize(new Dimension(460, 65));
			JLabel sbcLab = new JLabel("SBC: ");
			sbcLab.setFont(ahp.M_LabelFont);
			sbcLab.setHorizontalAlignment(SwingConstants.TRAILING);
			sbcLab.setBorder(m_InsetBorder);
			JTextField sbcAmt = new JTextField(8);
			sbcAmt.setDropTarget(null);
			sbcAmt.setEditable(false);
			sbcAmt.setOpaque(false);
			sbcAmt.setHorizontalAlignment(JTextField.LEFT);
			sbcAmt.setBorder(m_InsetBorder);
			sbcAmt.setBackground(m_Plugin.m_LabelBackColor);
			sbcAmt.setFocusable(false);
			sbcAmt.setText(df.format(ahp.getAvailSBC()));
			sbcAmt.setToolTipText("Available for withdrawal or deployment to "
										+ "tranches");
			sbcLab.setLabelFor(sbcAmt);
			m_HoldingsPanel.add(sbcLab);
			m_HoldingsPanel.add(sbcAmt);

			// next, net position across all tranches
			JLabel posLab = new JLabel("Position: ");
			posLab.setFont(ahp.M_LabelFont);
			posLab.setHorizontalAlignment(SwingConstants.TRAILING);
			posLab.setBorder(m_InsetBorder);
			posLab.setToolTipText("Aggregate of all tranches");
			JTextField pos = new JTextField(8);
			pos.setDropTarget(null);
			pos.setEditable(false);
			pos.setOpaque(false);
			pos.setHorizontalAlignment(JTextField.LEFT);
			pos.setFont(ahp.M_LabelFont);
			pos.setBorder(m_InsetBorder);
			pos.setBackground(m_Plugin.m_LabelBackColor);
			pos.setFocusable(false);
			posLab.setLabelFor(pos);
			double position = 0.0;
			double btcStart = 0.0;
			double btcNow = 0.0;
			double ibNow = 0.0;
			double valNow = 0.0;
			double ibVal = 0.0;
			double posDelta = 0.0;
			TrancheTableModel tModel = ahp.getTrancheModel();
			ATabManager atm = (ATabManager) m_TabManager;
			StatsPane sp = atm.getStatsPane();
			ConfigPane cp = atm.getConfigPane();
			int trancheCnt = tModel.getRowCount();
			if (trancheCnt > 0) {
				// compute current valuation
				btcStart = tModel.getBTCStartTotal()
								+ tModel.getBTCAdditionsTotal();
				btcNow = tModel.getBTCTotal();
				ibNow = tModel.getIBTotal();
				// take into account actual IB reserve valuation
				ibVal = ibNow * sp.getIBRatio();
				valNow = btcNow + ibVal;
				position = (valNow / btcStart) - 1.0;
				// note the latest difference, if any
				if (m_TossPosition != 0.0 && m_TossPosition != position) {
					posDelta = position - m_TossPosition;
				}
				m_TossPosition = position;
			}
			else {
				m_TossPosition = 0.0;
			}
			df.applyPattern("+###0.00##%;-###0.00##%");
			pos.setToolTipText("Net change since previous refresh: "
									+ df.format(posDelta));
			String valTxt = df.format(position);
			pos.setText(valTxt);
			if (position < 0.0) {
				pos.setForeground(Color.RED);
			}
			else {
				pos.setForeground(Color.GREEN.darker());
			}
			m_HoldingsPanel.add(posLab);
			m_HoldingsPanel.add(pos);

			// next, total BTC in all tranches
			JLabel btcLab = new JLabel("BTC total: ");
			btcLab.setFont(ahp.M_LabelFont);
			btcLab.setHorizontalAlignment(SwingConstants.TRAILING);
			btcLab.setBorder(m_InsetBorder);
			JTextField btc = new JTextField(8);
			btc.setDropTarget(null);
			btc.setEditable(false);
			btc.setOpaque(false);
			btc.setHorizontalAlignment(JTextField.LEFT);
			btc.setBorder(m_InsetBorder);
			btc.setBackground(m_Plugin.m_LabelBackColor);
			btc.setFocusable(false);
			df.applyPattern("###0.0000");
			btc.setText(df.format(btcNow));
			btc.setToolTipText("Total BTC held in tranches");
			btcLab.setLabelFor(btc);
			m_HoldingsPanel.add(btcLab);
			m_HoldingsPanel.add(btc);

			// next, total IB in all tranches
			JLabel ibLab = new JLabel("IB total: ");
			ibLab.setFont(ahp.M_LabelFont);
			ibLab.setHorizontalAlignment(SwingConstants.TRAILING);
			ibLab.setBorder(m_InsetBorder);
			JTextField ib = new JTextField(8);
			ib.setDropTarget(null);
			ib.setEditable(false);
			ib.setOpaque(false);
			ib.setHorizontalAlignment(JTextField.LEFT);
			ib.setBorder(m_InsetBorder);
			ib.setBackground(m_Plugin.m_LabelBackColor);
			ib.setFocusable(false);
			ib.setText(df.format(ibNow));
			ib.setToolTipText("Redemption value: " + df.format(ibVal) + " BTC");
			ibLab.setLabelFor(ib);
			m_HoldingsPanel.add(ibLab);
			m_HoldingsPanel.add(ib);

			// next, amount of free VFQ available for bleed-in on purchases
			JLabel vfqLab = new JLabel("VFQ: ");
			vfqLab.setFont(ahp.M_LabelFont);
			vfqLab.setHorizontalAlignment(SwingConstants.TRAILING);
			vfqLab.setBorder(m_InsetBorder);
			JTextField vfqAmt = new JTextField(8);
			vfqAmt.setDropTarget(null);
			vfqAmt.setEditable(false);
			vfqAmt.setOpaque(false);
			vfqAmt.setHorizontalAlignment(JTextField.LEFT);
			vfqAmt.setBorder(m_InsetBorder);
			vfqAmt.setBackground(m_Plugin.m_LabelBackColor);
			vfqAmt.setFocusable(false);
			vfqAmt.setText(df.format(ahp.getAvailVFQ()));
			vfqAmt.setToolTipText("Available for partial payment on IB bought by "
										+ "tranches");
			vfqLab.setLabelFor(vfqAmt);
			m_HoldingsPanel.add(vfqLab);
			m_HoldingsPanel.add(vfqAmt);

			// lastly, the count of tranches
			JLabel traLab = new JLabel("Tranches: ");
			traLab.setFont(ahp.M_LabelFont);
			traLab.setHorizontalAlignment(SwingConstants.TRAILING);
			traLab.setBorder(m_InsetBorder);
			JTextField tCnt = new JTextField(8);
			tCnt.setDropTarget(null);
			tCnt.setEditable(false);
			tCnt.setOpaque(false);
			tCnt.setHorizontalAlignment(JTextField.LEFT);
			tCnt.setBorder(m_InsetBorder);
			tCnt.setBackground(m_Plugin.m_LabelBackColor);
			tCnt.setFocusable(false);
			Integer tCount = new Integer(trancheCnt);
			df.applyPattern("##0");
			tCnt.setText(tCount.toString());
			if (trancheCnt == 0) {
				tCnt.setToolTipText("Your account controls no tranches");
			}
			else {
				tCnt.setToolTipText("Number of tranches controlled by "
										+ "your account");
			}
			traLab.setLabelFor(tCnt);
			m_HoldingsPanel.add(traLab);
			m_HoldingsPanel.add(tCnt);
		}
		
		// place the tranche list into a scroll pane
		JScrollPane trancheSelect = new JScrollPane(m_TranchesTable);
		trancheSelect.setOpaque(true);
		trancheSelect.setBorder(null);

		// place both the holdings data pane and the tranche list scroll pane
		// into another pane
		JPanel holdPane = new JPanel(new GridBagLayout());
		holdPane.setOpaque(true);
		holdPane.setBorder(null);
		m_GBC.fill = GridBagConstraints.NONE;
		m_GBC.gridheight = 4;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_GBC.anchor = GridBagConstraints.NORTH;
		holdPane.add(Box.createVerticalStrut(5), m_GBC);
		holdPane.add(m_HoldingsPanel, m_GBC);
		holdPane.add(Box.createVerticalStrut(10), m_GBC);
		holdPane.add(trancheSelect, m_GBC);
		holdPane.setBackground(new Color(0x008080));

		// put entire form into the main scroll area
		m_WorkArea.setViewportView(holdPane);
		revalidate();
	}

	/**
	 * helper method to create a JTable (for tranches), also providing column
	 * tooltip definitions for the header
	 */
	private JTable mkNewTranchesTable() {
		AHomePane ahp = (AHomePane) m_ParentHome;
		final TrancheTableModel ttm = ahp.getTrancheModel();
		if (ttm == null) {
			Log.error("Null tranche table model, cannot build table");
			return null;
		}
		JTable newTab = new JTable(ttm) {
			// provide table header tool tips
			protected JTableHeader createDefaultTableHeader() {
				return new JTableHeader(columnModel) {
					public String getToolTipText(MouseEvent e) {
						String tip = null;
						Point ppp = e.getPoint();
						int index = columnModel.getColumnIndexAtX(ppp.x);
						int realIndex = columnModel.getColumn(index).getModelIndex();
						return M_TrancheTooltips[realIndex];
					}
				};
			}
			public boolean isCellEditable(int row, int col) {
				// objects in the select column can be editable, others are not
				if (col == 0) {
					if (ttm.getSelectMode() != TrancheTableModel.M_SelectNone) {
						return true;
					}
				}
				return false;
			}
		};
		newTab.setCellSelectionEnabled(false);
		newTab.setPreferredScrollableViewportSize(new Dimension(460, 140));
		newTab.setFillsViewportHeight(true);

		// set special renderer for the all columns
		TrancheRenderer tr = new TrancheRenderer(ttm, m_Plugin);
		newTab.setDefaultRenderer(JRadioButton.class, tr);
		newTab.setDefaultRenderer(Double.class, tr);
		newTab.setDefaultRenderer(Float.class, tr);
		newTab.setDefaultRenderer(Integer.class, tr);
		newTab.setDefaultRenderer(Byte.class, tr);
		newTab.setDefaultRenderer(Boolean.class, tr);

		// configure the table columns and renderers
		configTableColumns(newTab);

		// set default sorter
		newTab.setAutoCreateRowSorter(true);
		newTab.setAutoCreateColumnsFromModel(true);

		// set spacing preferences
		newTab.setRowHeight(20);
		newTab.setIntercellSpacing(new Dimension(2, 2));
		newTab.setShowGrid(true);
		newTab.setGridColor(UIManager.getColor("TableHeader.background"));

		// establish table as listener for model
		ttm.addTableModelListener(newTab);
		return newTab;
	}

	/**
	 * Helper method to configure the columns and renderers of the tranches
	 * table.  This needs to be called when the table is created,
	 * and after each change to the table.
	 * @param tab the table to configure (== m_TranchesTable)
	 */
	private void configTableColumns(JTable tab) {
		// set max width on select column, and set cell editor
		TableColumnModel tcm = tab.getColumnModel();
		TableColumn tc = tcm.getColumn(0);
		tc.setMaxWidth(25);
		AHomePane ahp = (AHomePane) m_ParentHome;
		TrancheTableModel ttm = ahp.getTrancheModel();
		TrancheSelector ts = new TrancheSelector(m_Plugin, ttm);
		tc.setCellEditor(ts);
		// provide special column header renderer for right justification
		TraColRenderer tcr = new TraColRenderer();
		tc.setHeaderRenderer(tcr);

		// tranche ID
		tc = tcm.getColumn(1);
		tc.setMinWidth(25);
		tc.setMaxWidth(30);
		tc.setHeaderRenderer(tcr);

		// starting BTC funding (plus additions)
		tc = tcm.getColumn(2);
		tc.setMinWidth(50);
		tc.setHeaderRenderer(tcr);

		// BTC balance
		tc = tcm.getColumn(3);
		tc.setMinWidth(50);
		tc.setHeaderRenderer(tcr);

		// IB balance
		tc = tcm.getColumn(4);
		tc.setMinWidth(50);
		tc.setHeaderRenderer(tcr);

		// percentage status
		tc = tcm.getColumn(5);
		tc.setMinWidth(70);
		tc.setHeaderRenderer(tcr);

		// Martingale iteration
		tc = tcm.getColumn(6);
		tc.setMinWidth(25);
		tc.setMaxWidth(25);
		tc.setHeaderRenderer(tcr);

		// betting increment
		tc = tcm.getColumn(7);
		tc.setMinWidth(50);
		tc.setHeaderRenderer(tcr);

		// losses
		tc = tcm.getColumn(8);
		tc.setMinWidth(20);
		tc.setMaxWidth(25);
		tc.setHeaderRenderer(tcr);

		// order number
		tc = tcm.getColumn(9);
		tc.setMinWidth(45);
		tc.setHeaderRenderer(tcr);

		// subsidy status
		tc = tcm.getColumn(10);
		tc.setMinWidth(35);
		tc.setMaxWidth(35);
		tc.setHeaderRenderer(tcr);
	}

	/**
	 * method to receive notification of a change in tranche selection
	 */
	public void trancheSelectChanged() {
		AHomePane ahp = (AHomePane) m_ParentHome;
		final TrancheTableModel ttm = ahp.getTrancheModel();
		m_SelTranche = ttm.getSelectedTranche();
		if (m_SelTranche == null) {
			// nothing is selected, deactivate buttons related to tranches
			m_RebalanceButton.setEnabled(false);
			m_TopUpButton.setEnabled(false);
			m_LiquidateButton.setEnabled(false);
			m_HistoryButton.setEnabled(false);
		}
		else {
			// a tranche is selected, activate buttons related to tranches
			m_RebalanceButton.setEnabled(true);
			m_TopUpButton.setEnabled(true);
			m_LiquidateButton.setEnabled(true);
			m_HistoryButton.setEnabled(true);
		}
	}

	/**
	 * method to define rebalance action
	 */
	private void setRebalanceAction() {
		m_RebalanceAction = new AbstractAction("Rebalance") {
			public void actionPerformed(ActionEvent ae) {
				String abcId = m_Plugin.getLoginSecrets().getSVMId();
				final String err = "cannot rebalance tranche at ABC " + abcId;
				// error-check
				if (m_SelTranche == null) {
					// (this should be impossible due to button enabling code)
					m_TabManager.showError(err,
												"no tranche selected",
											"select one of your tranches and try again");
					return;
				}

				// prepare tranche modification message
				String sessId = m_TabManager.getLoginPane().getSessionId();
				ABCClientBlock traBlock = new ABCClientBlock();
				traBlock.setOpcode("REQ_modify_tranche");
				traBlock.setSessionId(sessId);
				traBlock.setTrancheId(m_SelTranche.getTrancheId());
				traBlock.setMod(ABCClientBlock.M_TrancheModTopup);
				// (amount is zero here)

				// send to server; answer will go to recordTrancheMod() below
				ABCListener abcListener
					= (ABCListener) m_Plugin.getSVMListener();
				setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
				if (!abcListener.sendAsyncRequest(traBlock)) {
					Log.error("Tranche mod request failed to ABC Id " + abcId);
					m_TabManager.showError(err,
									"failure sending tranche modification message");
				}
			}
		};
	}

	/**
	 * method to define topup action
	 */
	private void setTopUpAction() {
		m_TopUpAction = new AbstractAction("TopUp") {
			public void actionPerformed(ActionEvent ae) {
				if (!m_TopUpMode) {
					// check that they have a free SBC balance
					if (m_TopUpAvail <= 0.0) {
						m_TabManager.showError("cannot top up selected tranche",
													"you do not have any BTC available",
													"add money from your wallet or "
													+ "liquidate one or more tranches");
						return;
					}

					// obtain the amount to be applied to the topup
					m_TopUpMode = true;
					prepDisplay();
				}
				else {
					String abcId = m_Plugin.getLoginSecrets().getSVMId();
					final String err = "cannot top up tranche at ABC " + abcId;
					// error-check
					if (m_SelTranche == null) {
						// (this should be impossible due to button enabling code)
						m_TabManager.showError(err,
													"no tranche selected",
											"select one of your tranches and try again");
						return;
					}
					if (m_SelTranche.getOrderId() != 0) {
						m_TabManager.showError(err,
													"this tranche has an open market order",
													"wait for order to settle first");
						return;
					}

					// default means to do a rebalance, so we require positive value
					String amtTxt = m_TopUp.getText();
					Double amt = null;
					double topAmt = 0.0;
					AHomePane ahp = (AHomePane) m_ParentHome;
					m_TopUpAvail = ahp.getAvailSBC();
					// parse entered amount
					try {
						amt = Double.parseDouble(amtTxt);
						topAmt = amt.doubleValue();
					}
					catch (NumberFormatException nfe) {
						Log.error("Illegal topup amount, " + amtTxt, nfe);
					}
					if (topAmt <= 0.0) {
						m_TabManager.showError(err,
													"invalid top up amount",
													"enter a positive number");
						return;
					}

					// make sure user has this much
					if (topAmt > m_TopUpAvail) {
						m_TabManager.showError(err,
													"amount is more than you have available",
													"enter an amount not more than "
													+ m_TopUpAvail);
						return;
					}

					// reset display mode
					m_TopUpMode = false;

					// prepare tranche modification message
					String sessId = m_TabManager.getLoginPane().getSessionId();
					ABCClientBlock traBlock = new ABCClientBlock();
					traBlock.setOpcode("REQ_modify_tranche");
					traBlock.setSessionId(sessId);
					traBlock.setTrancheId(m_SelTranche.getTrancheId());
					traBlock.setMod(ABCClientBlock.M_TrancheModTopup);
					traBlock.setTopUp(topAmt);

					// send to server; answer will go to recordTrancheMod() below
					ABCListener abcListener
						= (ABCListener) m_Plugin.getSVMListener();
					setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
					if (!abcListener.sendAsyncRequest(traBlock)) {
						Log.error("Tranche mod request failed to ABC Id " + abcId);
						m_TabManager.showError(err,
										"failure sending tranche modification message");
					}
				}
			}
		};
	}

	/**
	 * method to define liquidate action
	 */
	private void setLiquidateAction() {
		m_LiquidateAction = new AbstractAction("Liquidate") {
			public void actionPerformed(ActionEvent ae) {
				m_TopUpMode = false;
				String abcId = m_Plugin.getLoginSecrets().getSVMId();
				final String err = "cannot liquidate tranche at ABC " + abcId;
				// error-check
				if (m_SelTranche == null) {
					// (this should be impossible due to button enabling code)
					m_TabManager.showError(err,
												"no tranche selected",
											"select one of your tranches and try again");
					return;
				}

				// prepare tranche modification message
				String sessId = m_TabManager.getLoginPane().getSessionId();
				ABCClientBlock traBlock = new ABCClientBlock();
				traBlock.setOpcode("REQ_modify_tranche");
				traBlock.setSessionId(sessId);
				traBlock.setTrancheId(m_SelTranche.getTrancheId());
				traBlock.setMod(ABCClientBlock.M_TrancheModLiq);

				// send to server; answer will go to recordTrancheMod() below
				ABCListener abcListener
					= (ABCListener) m_Plugin.getSVMListener();
				setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
				if (!abcListener.sendAsyncRequest(traBlock)) {
					Log.error("Tranche liq. request failed to ABC Id " + abcId);
					m_TabManager.showError(err,
											"failure sending tranche liquidation message");
				}
			}
		};
	}

	/**
	 * method to define history action
	 */
	private void setHistoryAction() {
		m_HistoryAction = new AbstractAction("History") {
			public void actionPerformed(ActionEvent ae) {
				m_TopUpMode = false;
				// error-check
				if (m_SelTranche == null) {
					// (this should be impossible due to button enabling code)
					m_TabManager.showError("cannot examine tranche value history",
												"no tranche selected",
											"select one of your tranches and try again");
					return;
				}

				// tell history pane which tranche to load, and switch there
				ATabManager atm = (ATabManager) m_TabManager;
				HistoryPane hPane = atm.getHistoryPane();
				hPane.setHistMode(false);
				hPane.setTrancheId(m_SelTranche.getTrancheId());
				m_TabManager.queueTransition(new Integer(ATabManager.M_GO_HIST));
				m_TabManager.makeTransition();
			}
		};
	}

	/**
	 * method to define refresh action
	 */
	private void setRefreshAction() {
		m_RefreshAction = new AbstractAction("Refresh") {
			public void actionPerformed(ActionEvent ae) {
				m_TopUpMode = false;
				// do sync refresh of balances, tranches, stats, and config
				final ATabManager atm = (ATabManager) m_TabManager;
				Runnable doSync = new Runnable() {
					public void run() {
						atm.queryAcctData(true);
					}
				};
				new Thread(doSync).start();
			}
		};
	}

	/**
	 * handle the server reply giving our balance statistics
	 * @param balData the message as returned from the ABC server
	 */
	public void recordBalances(ABCBlock balData) {
		if (balData == null
			 || !balData.getOpcode().equalsIgnoreCase("REP_balances"))
		{
			Log.error("recordBalances(): not a balances message");
			return;
		}

		// record data in our home pane
		AHomePane ahp = (AHomePane) m_ParentHome;
		ahp.setAccountBTC(balData.getBTCTotal());
		ahp.setAccountIB(balData.getIBTotal());
		ahp.setBettorId(balData.getBettorId());
		ahp.setAvailVFQ(balData.getVFQTotal());

		// subtract tranche BTC total from total returned, to get free SBC
		TrancheTableModel tModel = ahp.getTrancheModel();
		double trancheBTC = tModel.getBTCTotal();
		ahp.setAvailSBC(Math.max(0.0, balData.getBTCTotal() - trancheBTC));
		// NB: the TransfersPane queries AHomePane's data for free SBC
	}

	/**
	 * handle the server's reply giving our tranche list
	 * @param trancheData the message as received from the ABC server
	 */
	public void recordTranches(ABCBlock trancheData) {
		if (trancheData == null
			 || !trancheData.getOpcode().equalsIgnoreCase("REP_tranches"))
		{
			Log.error("recordTranches(): not a tranches message");
			return;
		}

		// record the free SBC total and list of tranches
		AHomePane ahp = (AHomePane) m_ParentHome;
		ahp.setAvailSBC(trancheData.getBTCBalance());
		ArrayList<ABCBlock.Tranche> traList = trancheData.getTranches();
		ArrayList<ABCTranche> tranches
			= new ArrayList<ABCTranche>(Math.max(1, traList.size()));
		TrancheTableModel tModel = ahp.getTrancheModel();
		double traBTC = 0.0;
		for (ABCBlock.Tranche tra : traList) {
			ABCTranche abcTra = new ABCTranche();
			abcTra.setTrancheId(tra.m_TrancheId);
			abcTra.setBTCChipBalance(tra.m_BTC);
			traBTC += tra.m_BTC;
			abcTra.setIBBalance(tra.m_IB);
			abcTra.setBTCStart(tra.m_StartBTC);
			abcTra.setBTCAdditions(tra.m_Additions);
			abcTra.setActive(tra.m_Active);
			abcTra.setIteration(tra.m_Iteration);
			abcTra.setBetIncrement(tra.m_BetIncrement);
			abcTra.setLossCount(tra.m_Losses);
			abcTra.setOrderId(tra.m_OpenOrder);
			abcTra.setSubsidized(tra.m_Bonus);
			tranches.add(abcTra);
		}
		tModel.mergeTranches(tranches);

		// add tranche BTC total to free SBC to get account total
		ahp.setAccountBTC(traBTC + trancheData.getBTCBalance());
	}

	/**
	 * record success results of tranche modification
	 * @param traMod the message returned from the ABC server
	 */
	public void recordTrancheMod(ABCBlock traMod) {
		setCursor(null);
		if (traMod == null
			 || !traMod.getOpcode().equalsIgnoreCase("REP_tranche_modified"))
		{
			Log.error("recordTrancheMod(): not a tranche mod message");
			return;
		}

		// record new balance (which becomes new available SBC balance)
		AHomePane ahp = (AHomePane) m_ParentHome;
		ahp.setAvailSBC(traMod.getBTCBalance());
		m_TopUpMode = false;

		// a get should now be performed for account data and market orders
		final ATabManager atm = (ATabManager) m_TabManager;
		Runnable doUpdate = new Runnable() {
			public void run() {
				MarketPane mp = atm.getMarketPane();
				mp.queryOrders(false);
				// NB: does implicit refresh and sends us back here
				atm.queryAcctData(true);
			}
		};
		new Thread(doUpdate).start();
	}

	/**
	 * trigger a logout programmatically, without messaging the server
	 */
	public void forceLogout() {
		// logout and return to login pane (generic version)
		m_TopUpMode = false;
		SVMBlock forceLog = new SVMBlock();
		forceLog.setOpcode("REP_logged_out");
		m_TabManager.getLoginPane().recordLogout(forceLog);
	}


	// get methods
	/**
	 * obtain the table which displays tranches
	 * @return the table
	 */
	public JTable getTranchesTable() { return m_TranchesTable; }

	/**
	 * obtain the selected tranche
	 * @return the tranche currently selected, or null
	 */
	public ABCTranche getSelectedTranche() { return m_SelTranche; }

	// END methods
}
