/*
 * last modified---
 * 	04-30-15/jet/ Spark plugin version
 * 	03-13-15/jet/ add Refresh button
 * 	02-09-15/jet/ add age column
 * 	02-06-15/jet/ new
 *
 * purpose---
 * 	provide a UI screen to represent ABC market orders
 */

package org.silentvault.spark.ui.svm.abc;

import javax.swing.*;
import java.awt.event.MouseEvent;
import java.awt.event.ActionEvent;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Font;
import java.awt.Point;
import java.awt.Component;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;
import org.silentvault.spark.ui.svm.*;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;


/**
 * Extends AWorkPane to provide a screen for the ABC market orders page.
 *
 * @author Justin Turrell
 */
public final class MarketPane extends AWorkPane {
	// BEGIN data members
	/**
	 * tool tips for tranche table header columns
	 */
	protected final String[]		M_OrderTooltips =
		{ "The unique number of your tranche that placed this order",
			"The unique order number",
			"The amount of InsurBucks being bought or sold",
			"The toss on which the order was entered (YYMMDDHHmm format)",
			"The number of tosses this order has been in the system",
			"The type of order (buy or sell)",
			"The discount policy which applied when the order was entered" };

	/**
	 * the table which contains the orders shown
	 */
	private JTable						m_OrdersTable;

	/**
	 * the data model backing the orders table
	 */
	private IBOrderTableModel		m_OrderTableModel;

	/**
	 * font for use in table columns
	 */
	public final Font					M_TableFont;

	/**
	 * action to Refresh orders with latest list
	 */
	private AbstractAction			m_RefreshAction;

	/**
	 * button for refresh action
	 */
	private JButton					m_RefreshButton;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the AHomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public MarketPane(WalletClient plugin, AHomePane home, ATabManager mgr) {
		super(plugin, home, mgr);

		// define instructions
		String marketTxt = "This panel shows all orders placed by tranches in "
								+ "your ABC account.<br/><br/>"
								+ "These orders will automatically resolve according "
								+ "to the applicable discount policy in effect.  It is "
								+ "not possible to withdraw or modify them.<br/><br/>"
								+ "Note that "
								+ "you cannot begin betting while sell orders "
								+ "from a previous betting session still exist.";
		m_UserInstructs = marketTxt;

		// init font value
		M_TableFont = new Font(Font.SANS_SERIF, Font.PLAIN, 13);

		// define refresh action and button
		m_RefreshAction = new AbstractAction("Refresh") {
			public void actionPerformed(ActionEvent ae) {
				// query orders, synchronously
				m_RefreshButton.setEnabled(false);
				setCursor(M_WaitCursor);
				Runnable getOrds = new Runnable() {
					public void run() {
						ATabManager atm = (ATabManager) m_TabManager;
						// need stats first, for proper interpretation of order prices
						if (atm.queryStats(true)) {
							if (!queryOrders(true)) {
								Log.error("Unable to query orders at ABC server");
							}
						}
						else {
							Log.error("Unable to query stats at ABC server");
						}
						setCursor(null);
						m_RefreshButton.setEnabled(true);
					}
				};
				new Thread(getOrds).start();
			}
		};
		m_RefreshButton = new JButton(m_RefreshAction);
		m_RefreshButton.setFont(m_TabManager.M_ButtonFont);
		m_RefreshButton.setToolTipText("Query the game server for latest list");
		m_RefreshButton.setEnabled(true);

		// set up tranches table model (empty) and table
		ArrayList<IBOrder> noOrders = new ArrayList<IBOrder>(100);
		m_OrderTableModel = new IBOrderTableModel(noOrders, m_Plugin);
		m_OrdersTable = mkNewOrdersTable();

		// popuplate the button pane
		m_ButtonPane.removeAll();
		m_ButtonPane.add(Box.createHorizontalStrut(380));	// spacer
		m_ButtonPane.add(m_RefreshButton);
	}

	/**
	 * do display of holdings
	 */
	public void prepDisplay() {
		// set screen label
		buildScreenLabel("Orders by your tranches to buy or sell IB");

		// put the tranche list into the scroll area
		m_WorkArea.setViewportView(m_OrdersTable);
		revalidate();
	}

	/**
	 * helper method to create a JTable (for orders), also providing column
	 * tooltip definitions for the header
	 */
	private JTable mkNewOrdersTable() {
		AHomePane ahp = (AHomePane) m_ParentHome;
		if (m_OrderTableModel == null) {
			Log.error("Null order table model, cannot build table");
			return null;
		}
		JTable newTab = new JTable(m_OrderTableModel) {
			// provide table header tool tips
			protected JTableHeader createDefaultTableHeader() {
				return new JTableHeader(columnModel) {
					public String getToolTipText(MouseEvent e) {
						String tip = null;
						Point ppp = e.getPoint();
						int index = columnModel.getColumnIndexAtX(ppp.x);
						int realIndex = columnModel.getColumn(index).getModelIndex();
						return M_OrderTooltips[realIndex];
					}
				};
			}
			public boolean isCellEditable(int row, int col) {
				// no objects are selectable
				return false;
			}
		};
		newTab.setCellSelectionEnabled(false);
		newTab.setPreferredScrollableViewportSize(new Dimension(460, 340));
		newTab.setFillsViewportHeight(true);

		// set special renderer for the all columns
		IBOrderRenderer or = new IBOrderRenderer(m_OrderTableModel, m_Plugin);
		newTab.setDefaultRenderer(Double.class, or);
		newTab.setDefaultRenderer(Integer.class, or);
		newTab.setDefaultRenderer(Byte.class, or);
		newTab.setDefaultRenderer(String.class, or);

		// configure the table columns and renderers
		configTableColumns(newTab);

		// set default sorter
		newTab.setAutoCreateRowSorter(true);
		newTab.setAutoCreateColumnsFromModel(true);

		// set spacing preferences
		newTab.setRowHeight(20);
		newTab.setIntercellSpacing(new Dimension(2, 2));
		newTab.setShowGrid(true);
		newTab.setGridColor(UIManager.getColor("TableHeader.background"));

		// establish table as listener for model
		m_OrderTableModel.addTableModelListener(newTab);
		return newTab;
	}

	/**
	 * Helper method to configure the columns and renderers of the tranches
	 * table.  This needs to be called when the table is created,
	 * and after each change to the table.
	 * @param tab the table to configure (== m_TranchesTable)
	 */
	private void configTableColumns(JTable tab) {
		// provide special column header renderer for right justification
		OrdColRenderer ocr = new OrdColRenderer();
		TableColumnModel tcm = tab.getColumnModel();
		TableColumn tc = tcm.getColumn(0);
		tc.setHeaderRenderer(ocr);

		// tranche ID (column 0)
		tc.setMinWidth(25);
		tc.setMaxWidth(40);
		tc.setHeaderRenderer(ocr);

		// order number
		tc = tcm.getColumn(1);
		tc.setMinWidth(65);
		tc.setHeaderRenderer(ocr);

		// amount of order
		tc = tcm.getColumn(2);
		tc.setMinWidth(90);
		tc.setHeaderRenderer(ocr);

		// toss entered
		tc = tcm.getColumn(3);
		tc.setMinWidth(90);
		tc.setHeaderRenderer(ocr);

		// age of order in tosses
		tc = tcm.getColumn(4);
		tc.setMinWidth(35);
		tc.setMaxWidth(40);
		tc.setHeaderRenderer(ocr);

		// order type
		tc = tcm.getColumn(5);
		tc.setMinWidth(45);
		tc.setMaxWidth(50);
		tc.setHeaderRenderer(ocr);

		// discount policy
		tc = tcm.getColumn(6);
		tc.setMinWidth(100);
		tc.setHeaderRenderer(ocr);
	}

	/**
	 * record orders retrieved from game server
	 * @param book the order book returned from ABC
	 */
	public void recordOrders(ABCBlock book) {
		if (book == null
			 || !book.getOpcode().equalsIgnoreCase("REP_order_book"))
		{
			Log.error("recordOrders(): not a order book reply message");
			return;
		}

		// loop through list of orders and build list for submission to model
		ArrayList<ABCBlock.OpenOrder> orderBook = book.getOpenOrders();
		ArrayList<IBOrder> openOrders
			= new ArrayList<IBOrder>(Math.max(1, orderBook.size()));
		ATabManager atm = (ATabManager) m_Plugin.getMarketplaceTabManager();
		ConfigPane cp = atm.getConfigPane();
		for (ABCBlock.OpenOrder ord : orderBook) {
			IBOrder opOrd = new IBOrder();
			opOrd.setOrderId(ord.m_OrderId);
			opOrd.setTrancheId(ord.m_TrancheId);
			opOrd.setQuantity(ord.m_IB);
			opOrd.setEntryToss(ord.m_FirstToss);
			opOrd.setAge(ord.m_Age);
			// validate the sale policy
			if (!ord.m_SalePolicy.isEmpty()) {
				opOrd.setOrderType(IBOrder.M_TypeSell);
				DiscountPolicy sdp = null;
				for (DiscountPolicy dp : cp.getSalePolicies()) {
					String policy = dp.getPolicyId() + "##" + dp.getSpecification();
					if (policy.equals(ord.m_SalePolicy)) {
						sdp = dp;
						break;
					}
				}
				if (sdp == null) {
					Log.error("Unrecognized sale policy for oId "
								+ opOrd.getOrderId() + ", " + ord.m_SalePolicy);
				}
				else {
					opOrd.setPolicy(sdp);
				}
			}
			// validate the purchase policy
			else if (!ord.m_PurchPolicy.isEmpty()) {
				DiscountPolicy pdp = null;
				for (DiscountPolicy dp : cp.getPurchPolicies()) {
					String policy = dp.getPolicyId() + "##" + dp.getSpecification();
					if (policy.equals(ord.m_PurchPolicy)) {
						pdp = dp;
						break;
					}
				}
				if (pdp == null) {
					Log.error("Unrecognized purchase policy for oId "
								+ opOrd.getOrderId() + ", " + ord.m_PurchPolicy);
				}
				else {
					opOrd.setPolicy(pdp);
				}
			}
			else {
				Log.error("Missing sale/purchase policy for oId "
							+ opOrd.getOrderId());
			}
			openOrders.add(opOrd);
		}

		// destructively merge received order book into table model
		m_OrderTableModel.mergeOrders(openOrders);
		// NB: this fires table model listener which redraws our table
	}

	/**
	 * request all orders from the server
	 * @param sync if true, make a synchronous query
	 * @return true on success
	 */
	public boolean queryOrders(boolean sync) {
		// prepare orders query message
		String sessId = m_TabManager.getLoginPane().getSessionId();
		String abcId = m_Plugin.getLoginSecrets().getSVMId();
		ABCClientBlock ordersBlock = new ABCClientBlock();
		ordersBlock.setOpcode("REQ_order_book");
		ordersBlock.setSessionId(sessId);

		// send to server; answer will go to recordOrders()
		ABCListener abcListener = (ABCListener) m_Plugin.getSVMListener();
		if (sync) {
			if (!abcListener.sendRequest(ordersBlock)) {
				Log.error("Orders request failed to ABC Id " + abcId);
				m_TabManager.showError("cannot get orders at ABC " + abcId,
												"failure sending orders request message");
				return false;
			}
		}
		else {
			if (!abcListener.sendAsyncRequest(ordersBlock)) {
				Log.error("Async orders request failed to ABC Id " + abcId);
				m_TabManager.showError("cannot get orders at ABC " + abcId,
												"failure sending orders request message");
				return false;
			}
		}
		return true;
	}


	// get methods
	/**
	 * obtain the table which displays orders
	 * @return the table
	 */
	public JTable getOrdersTable() { return m_OrdersTable; }

	/**
	 * obtain the model which backs the orders table
	 * @return the order table model
	 */
	public IBOrderTableModel getOrderModel() { return m_OrderTableModel; }

	// END methods
}
