/*
 * last modified---
 * 	10-29-16/jet/ add support for specifying a minimum IB ratio before doing
 * 					  the deployment, and for showing any queued deployments
 * 	08-26-16/jet/ allow play start even though already betting
 * 	04-30-15/jet/ Spark plugin version
 * 	03-19-15/jet/ complete
 * 	02-05-15/jet/ new
 *
 * purpose---
 * 	provide a UI screen to deal with initializing and terminating betting
 */

package org.silentvault.spark.ui.svm.abc;

import javax.swing.*;
import java.awt.event.ActionEvent;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Font;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.GridBagConstraints;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;
import org.silentvault.spark.ui.svm.*;
import org.silentvault.spark.ui.wallet.HomePane;
import org.silentvault.spark.ui.wallet.WTabManager;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;


/**
 * Extends AWorkPane to provide a screen for starting and stopping betting by
 * a game account.
 *
 * @author Justin Turrell
 */
public final class PlayPane extends AWorkPane {
	// BEGIN data members
	/**
	 * action to Init betting
	 */
	private AbstractAction				m_StartAction;

	/**
	 * button for Init action
	 */
	private JButton						m_StartButton;

	/**
	 * action to Stop betting
	 */
	private AbstractAction				m_StopAction;

	/**
	 * button for Stop action
	 */
	private JButton						m_StopButton;

	/**
	 * the panel containing the transfer information
	 */
	private JPanel							m_PlayPanel;

	/**
	 * the amount (in SBC) currently available for deployment to tranches
	 */
	private double							m_BankrollAvail;

	/**
	 * the UI element for entering bankroll to deploy
	 */
	private JTextField					m_Bankroll;

	/**
	 * the required level (in percent) for the IB ratio before tranche 
	 * deployments are to be processed by the server
	 */
	private float							m_IBRatioReq;

	/**
	 * the UI element for entering the required IB ratio
	 */
	private JTextField					m_IBRatio;

	/**
	 * inner class to store a pending queued tranche deployment
	 */
	protected final class QueuedDeploy {
		/**
		 * toss on which this entry was queued (YYMMDDHHmm)
		 */
		public int			m_Toss;

		/**
		 * the amount to be deployed when the ratio is satisfied
		 */
		public double		m_Bankroll;

		/**
		 * the required IB reserve ratio which must be attained before processing
		 */
		public float		m_IBRatio;

		/**
		 * nullary constructor
		 */
		public QueuedDeploy() {
		}
	}

	/**
	 * current (most recently fetched) list of pending tranche deployments
	 */
	private ArrayList<QueuedDeploy>	m_QueuedDeployments;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the AHomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public PlayPane(WalletClient plugin, AHomePane home, ATabManager mgr) {
		super(plugin, home, mgr);
		m_QueuedDeployments = new ArrayList<QueuedDeploy>();

		// define instructions
		String playTxt = "This panel is used to start or stop betting.  "
							+ "In order to start betting, you must have a bankroll "
							+ "available to deploy to tranches.  This bankroll will "
							+ "be divided into tranches according to the tranche size "
							+ "that you set in your Config.  Each tranche will "
							+ "purchase the appropriate quantity of InsurBucks (IB), "
							+ "according to your selected purchase discount policy. "
							+ " Tranches will begin betting on the toss following the "
							+ "completion of their IB purchase order.<br/><br/>"
							+ "When you order betting stopped, all of your tranches "
							+ "will liquidate their IB according to your sale discount"
							+ " policy, and all BTC will end up back in your bankroll."
							+ "<br/><br/>To move BTC in or out of your bankroll, use "
							+ "the Transfers screen.  Once your account is betting, "
							+ "you can examine and modify existing tranches from the "
							+ "Holdings screen.<br/><br/>"
							+ "Note you may use the Start function multiple times, "
							+ "while already betting, to deploy additional bankroll "
							+ "into more tranches.<br/><br/>"
							+ "You may optionally specify a minimum reserve ratio for "
							+ "IB, up to 99%.  If the current ratio is less than what "
							+ "you entered, your tranche deployment order will be "
							+ "queued until the "
							+ "ratio meets or exceeds the level you specified.  This "
							+ "mechanism applies to both initial and additional "
							+ "deployments.<br/><br/>"
							+ "Any queued deployments will be shown on this page.  "
							+ "Note that you cannot withdraw or alter your queued "
							+ "deployments, except by using the stop betting "
							+ "function, which cancels them.";
		m_UserInstructs = playTxt;

		// define start action and button
		setStartAction();
		m_StartButton = new JButton(m_StartAction);
		m_StartButton.setFont(m_TabManager.M_ButtonFont);
		m_StartButton.setToolTipText("Deploy the indicated bankroll to tranches");
		m_StartButton.setBackground(Color.GREEN);

		// define stop action and button
		setStopAction();
		m_StopButton = new JButton(m_StopAction);
		m_StopButton.setFont(m_TabManager.M_ButtonFont);
		m_StopButton.setToolTipText("Cease playing and cash out all tranches "
											+ "back into your bankroll");
		m_StopButton.setEnabled(false);
		m_StopButton.setBackground(Color.RED);

		// populate the button pane
		m_ButtonPane.removeAll();
		m_ButtonPane.add(Box.createHorizontalStrut(80));	// spacer
		m_ButtonPane.add(m_StartButton);
		m_ButtonPane.add(Box.createHorizontalStrut(95));	// spacer
		m_ButtonPane.add(m_StopButton);

		// init main panel area
		m_PlayPanel = new JPanel();
		m_PlayPanel.setOpaque(true);
		m_PlayPanel.setBorder(null);
		m_PlayPanel.setLayout(new GridBagLayout());
		m_PlayPanel.setBackground(Color.WHITE);

		// bankroll amount UI
		m_Bankroll = new JTextField(8);
		m_Bankroll.setDropTarget(null);
		m_Bankroll.setHorizontalAlignment(JTextField.LEFT);
		final Dimension entrySz = new Dimension(30, 20);
		m_Bankroll.setPreferredSize(entrySz);
		m_Bankroll.setText("0.0");
		m_Bankroll.setToolTipText("<html>Enter the specific amount you wish to "
										+ "deploy to tranches for<br/>betting, or leave "
										+ "set to zero to use all BTC available.</html>");

		// IB ratio UI
		m_IBRatio = new JTextField(6);
		m_IBRatio.setDropTarget(null);
		m_IBRatio.setHorizontalAlignment(JTextField.LEFT);
		m_IBRatio.setPreferredSize(entrySz);
		m_IBRatio.setText("0.0");
		m_IBRatio.setToolTipText("<html>Enter the minimum IB backing ratio you "
										+ "require before your tranche<br/>deployment "
										+ "will be executed.  Valid values are "
										+ "between 0% and 99%.<br/>The 0.0% default "
										+ "means that deployments will be executed "
										+ "<br/>immediately, regardless of the current "
										+ "ratio.</html>");
	}

	/**
	 * do display of play start/stop
	 */
	public void prepDisplay() {
		// set screen label
		buildScreenLabel("Start or stop betting by your account.");

		// inits
		AHomePane ahp = (AHomePane) m_ParentHome;
		ATabManager atm = (ATabManager) m_TabManager;
		m_BankrollAvail = ahp.getAvailSBC();
		StatsPane sp = atm.getStatsPane();
		Double nowRatio = null;
		if (sp != null) {
			nowRatio = new Double(sp.getIBRatio() * 100.0);
		}
		else {
			nowRatio = 100.0;
		}
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(3);
		df.applyPattern("####0.0000");
		m_PlayPanel.removeAll();
		m_GBC.anchor = GridBagConstraints.NORTH;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_GBC.fill = GridBagConstraints.HORIZONTAL;
		m_GBC.gridheight = 9 + (m_QueuedDeployments.size() * 2);
		m_PlayPanel.setPreferredSize(new Dimension(470,
											200 + m_QueuedDeployments.size() * 35));
		m_PlayPanel.add(Box.createVerticalGlue(), m_GBC);

		// first, current betting status
		boolean betting = atm.getConfigPane().isBetting();
		JPanel betPane = new JPanel(new GridLayout(1, 4, 5, 5));
		betPane.setOpaque(true);
		betPane.setBorder(null);
		final Dimension lineSz = new Dimension(465, 25);
		betPane.setPreferredSize(lineSz);
		JLabel betLab = new JLabel("Betting: ");
		betLab.setFont(ahp.M_LabelFont);
		betLab.setHorizontalAlignment(SwingConstants.TRAILING);
		betLab.setBorder(m_InsetBorder);
		JTextField betFld = new JTextField(3);
		betFld.setDropTarget(null);
		betFld.setEditable(false);
		betFld.setOpaque(false);
		betFld.setHorizontalAlignment(JTextField.LEFT);
		betFld.setBorder(m_InsetBorder);
		betFld.setBackground(m_Plugin.m_LabelBackColor);
		betFld.setFocusable(false);
		if (betting) {
			betFld.setText("Yes");
			betFld.setToolTipText("Your account is currently betting");
		}
		else {
			betFld.setText("No");
			betFld.setToolTipText("Your account is not betting");
		}
		betLab.setLabelFor(betFld);
		betPane.add(betLab);
		betPane.add(betFld);

		// next, bettor ID
		JLabel idLab = new JLabel("Bettor ID: ");
		idLab.setFont(ahp.M_LabelFont);
		idLab.setHorizontalAlignment(SwingConstants.TRAILING);
		idLab.setBorder(m_InsetBorder);
		JTextField bId = new JTextField(6);
		bId.setDropTarget(null);
		bId.setEditable(false);
		bId.setOpaque(false);
		bId.setHorizontalAlignment(JTextField.LEFT);
		bId.setBorder(m_InsetBorder);
		bId.setBackground(m_Plugin.m_LabelBackColor);
		bId.setFocusable(false);
		Integer betId = new Integer(ahp.getBettorId());
		bId.setText(betId.toString());
		if (betId.intValue() == 0) {
			bId.setToolTipText("Your account is not currently betting");
		}
		else {
			bId.setToolTipText("Bettor ID currently assigned to your account");
		}
		idLab.setLabelFor(bId);
		betPane.add(idLab);
		betPane.add(bId);
		m_PlayPanel.add(betPane, m_GBC);

		// display any and all queued deployments
		for (QueuedDeploy qDeploy : m_QueuedDeployments) {
			m_PlayPanel.add(Box.createVerticalStrut(10), m_GBC);
			JPanel qPane = new JPanel(new GridLayout(1, 6, 5, 5));
			qPane.setOpaque(true);
			qPane.setBorder(null);
			qPane.setPreferredSize(lineSz);
			qPane.setBackground(Color.PINK);

			// show amount
			JLabel qLab = new JLabel("Queue: ");
			qLab.setFont(ahp.M_LabelFont);
			qLab.setHorizontalAlignment(SwingConstants.TRAILING);
			qLab.setBorder(m_InsetBorder);
			JTextField qAmt = new JTextField(6);
			qAmt.setDropTarget(null);
			qAmt.setEditable(false);
			qAmt.setOpaque(false);
			qAmt.setHorizontalAlignment(JTextField.LEFT);
			qAmt.setBorder(m_InsetBorder);
			qAmt.setBackground(Color.PINK);
			qAmt.setFocusable(false);
			qAmt.setText(df.format(qDeploy.m_Bankroll));
			qAmt.setToolTipText("The amount of BTC which will be deployed to "
									+ "tranches");
			qLab.setLabelFor(qAmt);
			qPane.add(qLab);
			qPane.add(qAmt);

			// show required ratio
			JLabel rLab = new JLabel("Ratio: ");
			rLab.setFont(ahp.M_LabelFont);
			rLab.setHorizontalAlignment(SwingConstants.TRAILING);
			rLab.setBorder(m_InsetBorder);
			JTextField rAmt = new JTextField(6);
			rAmt.setDropTarget(null);
			rAmt.setEditable(false);
			rAmt.setOpaque(false);
			rAmt.setHorizontalAlignment(JTextField.LEFT);
			rAmt.setBorder(m_InsetBorder);
			rAmt.setBackground(Color.PINK);
			rAmt.setFocusable(false);
			df.setMinimumFractionDigits(2);
			rAmt.setText(df.format(qDeploy.m_IBRatio) + " %");
			df.setMinimumFractionDigits(XMLVoucher.M_Precision);
			rAmt.setToolTipText("The minimum IB reserve ratio needed in order to "
									+ "activate this queued deployment");
			rLab.setLabelFor(qAmt);
			qPane.add(rLab);
			qPane.add(rAmt);

			// show toss when entered
			JLabel pLab = new JLabel("Toss: ");
			pLab.setFont(ahp.M_LabelFont);
			pLab.setHorizontalAlignment(SwingConstants.TRAILING);
			pLab.setBorder(m_InsetBorder);
			JTextField toss = new JTextField(6);
			toss.setDropTarget(null);
			toss.setEditable(false);
			toss.setOpaque(false);
			toss.setHorizontalAlignment(JTextField.LEFT);
			toss.setBorder(m_InsetBorder);
			toss.setBackground(Color.PINK);
			toss.setFocusable(false);
			toss.setFont(new Font(Font.SANS_SERIF, Font.PLAIN, 8));
			toss.setText(Integer.toString(qDeploy.m_Toss));
			toss.setToolTipText("The toss on which you queued this deployment "
									+ "(YYMMDDHHmm)");
			pLab.setLabelFor(toss);
			qPane.add(pLab);
			qPane.add(toss);

			m_PlayPanel.add(qPane, m_GBC);
		}

		// spacer
		JSeparator demark = new JSeparator();
		m_PlayPanel.add(Box.createVerticalStrut(25), m_GBC);
		m_PlayPanel.add(demark, m_GBC);
		m_PlayPanel.add(Box.createVerticalStrut(25), m_GBC);

		// add the bankroll entry item
		JPanel rollPane = new JPanel(new GridLayout(1, 3, 5, 5));
		rollPane.setOpaque(true);
		rollPane.setBorder(null);
		rollPane.setPreferredSize(lineSz);
		rollPane.setBackground(Color.GREEN.darker());
		JLabel rollLab = new JLabel("Bankroll to use: ");
		rollLab.setFont(ahp.M_LabelFont);
		rollLab.setHorizontalAlignment(SwingConstants.TRAILING);
		rollLab.setBorder(m_InsetBorder);
		rollLab.setLabelFor(m_Bankroll);
		final Dimension entrySz = new Dimension(30, 20);
		JTextField rollAvail = new JTextField(8);
		rollAvail.setDropTarget(null);
		rollAvail.setEditable(false);
		rollAvail.setOpaque(false);
		rollAvail.setHorizontalAlignment(JTextField.LEFT);
		rollAvail.setBorder(m_InsetBorder);
		rollAvail.setBackground(m_Plugin.m_LabelBackColor);
		rollAvail.setFocusable(false);
		rollAvail.setPreferredSize(entrySz);
		rollAvail.setText("( " + df.format(m_BankrollAvail) + " BTC )");
		rollAvail.setToolTipText("The amount of BTC available for betting");
		rollPane.add(rollLab);
		rollPane.add(m_Bankroll);
		rollPane.add(rollAvail);
		m_PlayPanel.add(rollPane, m_GBC);
		m_PlayPanel.add(Box.createVerticalStrut(15), m_GBC);

		// now add the minimum backing ratio item
		JPanel ratioPane = new JPanel(new GridLayout(1, 3, 5, 5));
		ratioPane.setOpaque(true);
		ratioPane.setBorder(null);
		ratioPane.setPreferredSize(lineSz);
		ratioPane.setBackground(Color.GREEN.darker());
		JLabel ratioLab = new JLabel("Minimum ratio: ");
		ratioLab.setFont(ahp.M_LabelFont);
		ratioLab.setHorizontalAlignment(SwingConstants.TRAILING);
		ratioLab.setBorder(m_InsetBorder);
		ratioLab.setLabelFor(m_IBRatio);
		// and show the current figure
		JTextField curRatio = new JTextField(6);
		curRatio.setDropTarget(null);
		curRatio.setEditable(false);
		curRatio.setOpaque(false);
		curRatio.setHorizontalAlignment(JTextField.LEFT);
		curRatio.setBorder(m_InsetBorder);
		curRatio.setBackground(m_Plugin.m_LabelBackColor);
		curRatio.setFocusable(false);
		curRatio.setPreferredSize(entrySz);
		df.setMinimumFractionDigits(2);
		curRatio.setText("( ratio now: " + df.format(nowRatio.doubleValue())
							+ "% )");
		curRatio.setToolTipText("If you enter a value higher than this, your "
										+ "deployment will be queued.");
		m_IBRatio.setText(df.format(m_IBRatioReq));
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		ratioPane.add(ratioLab);
		ratioPane.add(m_IBRatio);
		ratioPane.add(curRatio);
		m_PlayPanel.add(ratioPane, m_GBC);
		m_PlayPanel.add(Box.createVerticalGlue(), m_GBC);

		// activate appropriate button
		if (betting || !m_QueuedDeployments.isEmpty()) {
			m_StopButton.setEnabled(true);
		}
		else {
			m_StopButton.setEnabled(false);
		}

		// put entire form into a into the main scroll area
		m_WorkArea.setBorder(null);
		m_WorkArea.setViewportView(m_PlayPanel);
		m_WorkArea.revalidate();
		revalidate();
	}

	/**
	 * method to define start action
	 */
	private void setStartAction() {
		m_StartAction = new AbstractAction("Start Betting") {
			public void actionPerformed(ActionEvent ae) {
				// initialize betting
				// by default we deploy all available, but can also use less

				String amtTxt = m_Bankroll.getText();
				Double amt = null;
				double bankAmt = 0.0;
				AHomePane ahp = (AHomePane) m_ParentHome;
				ATabManager atm = (ATabManager) m_TabManager;
				m_BankrollAvail = ahp.getAvailSBC();
				try {
					amt = Double.parseDouble(amtTxt);
					bankAmt = amt.doubleValue();
				}
				catch (NumberFormatException nfe) {
					Log.error("Illegal starting bankroll amount, " + amtTxt, nfe);
				}
				if (bankAmt < 0.0) {
					m_TabManager.showError("cannot start betting",
													"invalid bankroll amount",
													"enter a positive number");
					return;
				}

				// make sure user has this much
				if (bankAmt > m_BankrollAvail) {
					m_TabManager.showError("cannot start betting",
													"amount is more than you have available",
									"enter an amount not more than " + m_BankrollAvail);
					return;
				}

				// check specified value for min. IB ratio
				String ratioTxt = m_IBRatio.getText();
				Float reqRatio = null;
				try {
					reqRatio = Float.parseFloat(ratioTxt);
					m_IBRatioReq = reqRatio.floatValue();
				}
				catch (NumberFormatException nfe) {
					Log.error("Illegal ratio percentage, " + ratioTxt, nfe);
					m_TabManager.showError("cannot start betting",
													"invalid required IB ratio, " + ratioTxt,
													"enter a percentage between 0 and 99");
					return;
				}
				if (m_IBRatioReq < 0.0F || m_IBRatioReq > 99.0F) {
					Log.error("Illegal ratio percentage, " + ratioTxt);
					m_TabManager.showError("cannot start betting",
													"invalid required IB ratio, " + ratioTxt,
													"enter a percentage between 0 and 99");
					m_IBRatioReq = 0.0F;
					return;
				}

				// prepare start message
				String sessId = m_TabManager.getLoginPane().getSessionId();
				final String abcId = m_Plugin.getLoginSecrets().getSVMId();
				final ABCClientBlock startBlock = new ABCClientBlock();
				startBlock.setOpcode("REQ_init_betting");
				startBlock.setSessionId(sessId);
				if (bankAmt != 0.0 && bankAmt != m_BankrollAvail) {
					// pass specific amount instead of zero default to deploy all
					startBlock.setBankroll(bankAmt);
				}
				startBlock.setIBRatio(m_IBRatioReq);

				// prevent a concurrent transaction, using either button
				m_StopButton.setEnabled(false);
				m_StartButton.setEnabled(false);
				setCursor(M_WaitCursor);

				Runnable sendStart = new Runnable() {
					public void run() {
						// send to server; success answer will go to recordStart()
						ABCListener abcListener
							= (ABCListener) m_Plugin.getSVMListener();
						if (!abcListener.sendRequest(startBlock)) {
							Log.error("Start request failed to ABC Id " + abcId);
							m_TabManager.showError("cannot start betting at ABC "
															+ abcId,
															"failure sending start message");
							m_StartButton.setEnabled(true);
							m_StopButton.setEnabled(true);
							setCursor(null);
						}
					}
				};
				new Thread(sendStart).start();
			}
		};
	}

	/**
	 * method to define stop action
	 */
	private void setStopAction() {
		m_StopAction = new AbstractAction("Stop Betting") {
			public void actionPerformed(ActionEvent ae) {
				// prepare stop message
				String sessId = m_TabManager.getLoginPane().getSessionId();
				final String abcId = m_Plugin.getLoginSecrets().getSVMId();
				final ABCClientBlock stopBlock = new ABCClientBlock();
				stopBlock.setOpcode("REQ_end_betting");
				stopBlock.setSessionId(sessId);

				// prevent a concurrent transaction, using either button
				m_StopButton.setEnabled(false);
				m_StartButton.setEnabled(false);
				setCursor(M_WaitCursor);

				Runnable sendStop = new Runnable() {
					public void run() {
						// send to server; answer will go to recordStop()
						ABCListener abcListener
							= (ABCListener) m_Plugin.getSVMListener();
						if (!abcListener.sendRequest(stopBlock)) {
							Log.error("Stop request failed to ABC Id " + abcId);
							m_TabManager.showError("cannot stop betting at ABC "
															+ abcId,
															"failure sending stop message");
							m_StartButton.setEnabled(true);
							m_StopButton.setEnabled(true);
							setCursor(null);
						}
					}
				};
				new Thread(sendStop).start();
			}
		};
	}

	/**
	 * record success results of betting initialization
	 * @param betDat the message returned from the ABC server
	 */
	public void recordStart(ABCBlock betDat) {
		m_StopButton.setEnabled(true);
		m_StartButton.setEnabled(true);
		setCursor(null);
		if (betDat == null
			 || !betDat.getOpcode().equalsIgnoreCase("REP_betting_init"))
		{
			Log.error("recordStart(): not a betting init message");
			return;
		}

		// record the bettor ID
		AHomePane ahp = (AHomePane) m_ParentHome;
		ahp.setBettorId(betDat.getBettorId());

		// recompute available SBC figure, based on bankroll actually deployed
		ahp.setAvailSBC(ahp.getAccountBTC() - betDat.getBankroll());

		// perform sync gets for balances, tranches, stats, configs, and go to
		// the holdings pane
		ATabManager atm = (ATabManager) m_TabManager;
		atm.queryAcctData(true);
		// make background fetch for orders
		MarketPane mp = atm.getMarketPane();
		mp.queryOrders(false);
	}

	/**
	 * record success results of betting cessation
	 * @param stopDat the message returned from the ABC server
	 */
	public void recordStop(ABCBlock stopDat) {
		m_StopButton.setEnabled(false);
		m_StartButton.setEnabled(true);
		setCursor(null);
		if (stopDat == null
			 || !stopDat.getOpcode().equalsIgnoreCase("REP_betting_ended"))
		{
			Log.error("recordStop(): not a betting stop message");
			return;
		}

		// record new balance (which becomes new available SBC balance)
		AHomePane ahp = (AHomePane) m_ParentHome;
		ahp.setAvailSBC(stopDat.getBTCBalance());
		m_QueuedDeployments.clear();
		// NB: by implication, the remainder is still in tranches
		double vfqDiff = stopDat.getVFQTotal() - ahp.getAvailVFQ();
		ahp.setAvailVFQ(stopDat.getVFQTotal());

		// perform sync gets for balances, tranches, stats, configs, and go to
		// the holdings pane
		ATabManager atm = (ATabManager) m_TabManager;
		atm.queryAcctData(true);
		// make background fetch for orders
		MarketPane mp = atm.getMarketPane();
		mp.queryOrders(false);

		// if we got extra VFQ, display popup showing award
		if (vfqDiff > 0.0) {
			m_TabManager.showInfo("You have received a cashout bonus of "
										+ vfqDiff + " VFQ!");
		}
	}

	/**
	 * method to reset buttons after error (called by ABCListener)
	 */
	public void resetButtons() {
		m_StopButton.setEnabled(true);
		m_StartButton.setEnabled(true);
	}

	/**
	 * method to record the current list of queued deployments
	 * (called by ConfigPane on a REP_account_details message)
	 */
	public void recordDeployments(ArrayList<ABCBlock.QueuedDeploy> queue) {
		m_QueuedDeployments.clear();
		if (queue == null || queue.isEmpty()) {
			return;
		}

		// add all that we received
		for (ABCBlock.QueuedDeploy qEntry : queue) {
			QueuedDeploy deploy = new QueuedDeploy();
			deploy.m_Toss = qEntry.m_Toss;
			deploy.m_Bankroll = qEntry.m_Bankroll;
			deploy.m_IBRatio = qEntry.m_IBRatio;
			m_QueuedDeployments.add(deploy);
		}
	}

	// END methods
}
