/*
 * last modified---
 * 	09-13-16/jet/ add VFQ ratio to display
 * 	08-25-16/jet/ add wholeness tosses to display
 * 	04-30-15/jet/ Spark plugin version
 * 	04-15-15/jet/ add net change since last refresh field (m_PosDelta)
 * 	04-14-15/jet/ use tossSeconds value in stats reply
 * 	03-11-15/jet/ update countdown timer correctly on stats replies
 * 	02-06-15/jet/ new
 *
 * purpose---
 * 	provide a UI screen to represent ABC global statistics
 */

package org.silentvault.spark.ui.svm.abc;

import javax.swing.*;
import java.awt.event.ActionEvent;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.WalletClient;
import org.silentvault.spark.XMLVoucher;
import org.silentvault.spark.ABCBlock;

import java.util.Date;
import java.util.Calendar;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import javax.management.timer.Timer;


/**
 * Extends AWorkPane to provide a screen for the ABC game stats (read-only).
 *
 * @author Justin Turrell
 */
public final class StatsPane extends AWorkPane {
	// BEGIN data members
	/**
	 * action to Refresh statistics with latest changes
	 */
	private AbstractAction			m_RefreshAction;

	/**
	 * button for refresh action
	 */
	private JButton					m_RefreshButton;

	/**
	 * the panel containing the stats information
	 */
	private JPanel						m_StatsPanel;

	/**
	 * current total number of actively betting accounts
	 */
	private int							m_NumBettors;

	/**
	 * next scheduled toss (YYMMDDHHmm format)
	 */
	private int							m_NextToss;

	/**
	 * toss rate, in tosses conducted per hour
	 */
	private byte						m_TossRate;

	/**
	 * current IB BTC reserve ratio
	 */
	private double						m_IBReserve;

	/**
	 * current IB VFQ reserve ratio
	 */
	private double						m_VFQRatio;

	/**
	 * current wholeness tosses (min betting duration)
	 */
	private int							m_WholenessTosses;

	/**
	 * date formatter for parsing tosses
	 */
	private SimpleDateFormat		m_TossFormatter;

	/**
	 * date formatter for human-readable toss display
	 */
	private SimpleDateFormat		m_DateFormatter;

	/**
	 * time delta in seconds until next scheduled toss
	 */
	private int							m_TossDelta;

	/**
	 * percentage change in valuation experienced between last and current tosses
	 */
	private double						m_PosDelta;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the AHomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public StatsPane(WalletClient plugin, AHomePane home, ATabManager mgr) {
		super(plugin, home, mgr);

		// define instructions
		String statsTxt = "This panel shows various statistics about the ABC "
								+ "game.  To get the latest values, click Refresh.";
		m_UserInstructs = statsTxt;
		m_TossFormatter = new SimpleDateFormat("yyMMddHHmm");
		m_DateFormatter = new SimpleDateFormat("yyyy-MMM-dd HH:mm");

		// define refresh action and button
		setRefreshAction();
		m_RefreshButton = new JButton(m_RefreshAction);
		m_RefreshButton.setFont(m_TabManager.M_ButtonFont);
		m_RefreshButton.setToolTipText("Query the game server for latest values");
		m_RefreshButton.setEnabled(true);

		// popuplate the button pane
		m_ButtonPane.removeAll();
		m_ButtonPane.add(Box.createHorizontalStrut(380));	// spacer
		m_ButtonPane.add(m_RefreshButton);
		Color back = new Color(0x007000);
		m_ButtonPane.setBackground(back);

		// init main panel area
		m_StatsPanel = new JPanel();
		m_StatsPanel.setOpaque(true);
		m_StatsPanel.setBorder(null);
		m_StatsPanel.setLayout(new GridBagLayout());
		m_StatsPanel.setPreferredSize(new Dimension(470, 200));
		// use dark green
		m_StatsPanel.setBackground(back);
	}

	/**
	 * do display of stats
	 */
	public void prepDisplay() {
		// set screen label
		buildScreenLabel("Global ABC game statistics");
		AHomePane ahp = (AHomePane) m_ParentHome;
		Color back = new Color(0x007000);
		Color front = new Color(0xEEEE33);

		// init config info panel
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(3);
		m_StatsPanel.removeAll();
		m_GBC.anchor = GridBagConstraints.NORTH;
		m_GBC.gridheight = 13;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_GBC.fill = GridBagConstraints.HORIZONTAL;
		m_StatsPanel.add(Box.createVerticalGlue(), m_GBC);

		// row 1: bettor count
		JLabel betLab = new JLabel("Number of bettors: ");
		betLab.setFont(ahp.M_LabelFont);
		betLab.setHorizontalAlignment(SwingConstants.TRAILING);
		betLab.setBorder(m_InsetBorder);
		betLab.setForeground(front);
		JTextField bCnt = new JTextField(8);
		bCnt.setDropTarget(null);
		bCnt.setEditable(false);
		bCnt.setOpaque(false);
		bCnt.setHorizontalAlignment(JTextField.LEFT);
		bCnt.setBorder(m_InsetBorder);
		bCnt.setBackground(back);
		bCnt.setForeground(front);
		bCnt.setFocusable(false);
		Integer betCnt = new Integer(m_NumBettors);
		bCnt.setText(betCnt.toString());
		bCnt.setToolTipText("The number of accounts currently betting");
		bCnt.setFont(ahp.M_LabelFont);
		betLab.setLabelFor(bCnt);
		firstItemConstraints(m_GBC);
		m_StatsPanel.add(betLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_StatsPanel.add(bCnt, m_GBC);
		m_StatsPanel.add(Box.createVerticalStrut(17), m_GBC);

		// row 2: next toss
		JLabel ntLab = new JLabel("Next toss: ");
		ntLab.setFont(ahp.M_LabelFont);
		ntLab.setHorizontalAlignment(SwingConstants.TRAILING);
		ntLab.setBorder(m_InsetBorder);
		ntLab.setForeground(front);
		JTextField nt = new JTextField(10);
		nt.setDropTarget(null);
		nt.setEditable(false);
		nt.setOpaque(false);
		nt.setHorizontalAlignment(JTextField.LEFT);
		nt.setBorder(m_InsetBorder);
		nt.setBackground(back);
		nt.setForeground(front);
		nt.setFocusable(false);
		nt.setFont(ahp.M_LabelFont);
		Integer nToss = null;
		Date tossTime = null;
		if (m_NextToss > 0) {
			nToss = new Integer(m_NextToss);
			tossTime = null;
			try {
				tossTime = m_TossFormatter.parse(nToss.toString());
				nt.setText(nToss.toString());
				// convert from UTC
				Date tossAt = ahp.localUTCdate(tossTime);
				nt.setToolTipText("Next scheduled toss ("
										+ m_DateFormatter.format(tossAt) + ")");
			}
			catch (ParseException pe) {
				Log.error("Invalid toss data format, " + nToss, pe);
			}
		}
		else {
			nt.setText("--unknown--");
		}
		ntLab.setLabelFor(nt);
		firstItemConstraints(m_GBC);
		m_StatsPanel.add(ntLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_StatsPanel.add(nt, m_GBC);
		m_StatsPanel.add(Box.createVerticalStrut(17), m_GBC);

		// row 3: toss rate (per hour)
		JLabel rateLab = new JLabel("Tosses per hour: ");
		rateLab.setFont(ahp.M_LabelFont);
		rateLab.setHorizontalAlignment(SwingConstants.TRAILING);
		rateLab.setBorder(m_InsetBorder);
		rateLab.setForeground(front);
		JTextField tRate = new JTextField(8);
		tRate.setDropTarget(null);
		tRate.setEditable(false);
		tRate.setOpaque(false);
		tRate.setHorizontalAlignment(JTextField.LEFT);
		tRate.setBorder(m_InsetBorder);
		tRate.setBackground(back);
		tRate.setForeground(front);
		tRate.setFocusable(false);
		tRate.setFont(ahp.M_LabelFont);
		Byte tph = new Byte(m_TossRate);
		int tMin = 0;
		if (m_TossRate > 0) {
			tMin = 60 / m_TossRate;
		}
		tRate.setText(tph.toString());
		tRate.setToolTipText("Every " + tMin + " minutes");
		rateLab.setLabelFor(tRate);
		firstItemConstraints(m_GBC);
		m_StatsPanel.add(rateLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_StatsPanel.add(tRate, m_GBC);
		m_StatsPanel.add(Box.createVerticalStrut(17), m_GBC);

		// row 4: IB backing ratio
		JLabel ratioLab = new JLabel("IB backing ratio: ");
		ratioLab.setFont(ahp.M_LabelFont);
		ratioLab.setHorizontalAlignment(SwingConstants.TRAILING);
		ratioLab.setBorder(m_InsetBorder);
		ratioLab.setForeground(front);
		ratioLab.setToolTipText("The reserve status of InsurBucks");
		JTextField ratio = new JTextField(6);
		ratio.setDropTarget(null);
		ratio.setEditable(false);
		ratio.setOpaque(false);
		ratio.setHorizontalAlignment(JTextField.LEFT);
		ratio.setBorder(m_InsetBorder);
		ratio.setBackground(back);
		ratio.setForeground(front);
		ratio.setFocusable(false);
		ratio.setFont(ahp.M_LabelFont);
		df.applyPattern("##0.00##%");
		ratio.setText(df.format(m_IBReserve));
		ratio.setToolTipText("The percentage of BTC backing each unit of IB");
		ratioLab.setLabelFor(ratio);
		firstItemConstraints(m_GBC);
		m_StatsPanel.add(ratioLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_StatsPanel.add(ratio, m_GBC);
		m_StatsPanel.add(Box.createVerticalStrut(17), m_GBC);

		// row 5: VFQ backing ratio
		JLabel vRatioLab = new JLabel("VFQ bonus ratio: ");
		vRatioLab.setFont(ahp.M_LabelFont);
		vRatioLab.setHorizontalAlignment(SwingConstants.TRAILING);
		vRatioLab.setBorder(m_InsetBorder);
		vRatioLab.setForeground(front);
		vRatioLab.setToolTipText("The percentage of VFQ in the bonus pool for "
										+ "each unit of IB");
		JTextField vRatio = new JTextField(6);
		vRatio.setDropTarget(null);
		vRatio.setEditable(false);
		vRatio.setOpaque(false);
		vRatio.setHorizontalAlignment(JTextField.LEFT);
		vRatio.setBorder(m_InsetBorder);
		vRatio.setBackground(back);
		vRatio.setForeground(front);
		vRatio.setFocusable(false);
		vRatio.setFont(ahp.M_LabelFont);
		vRatio.setText(df.format(m_VFQRatio));
		vRatio.setToolTipText("<html>If eligible, on cashout you would receive a "
									+ "special VFQ bonus<br/>"
									+ "of this figure multiplied by the "
									+ "quantity of IB in your account.</html>");
		vRatioLab.setLabelFor(vRatio);
		firstItemConstraints(m_GBC);
		m_StatsPanel.add(vRatioLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_StatsPanel.add(vRatio, m_GBC);
		m_StatsPanel.add(Box.createVerticalStrut(17), m_GBC);

		// row 6: wholeness tosses
		JLabel wholeLab = new JLabel("Default wholeness tosses: ");
		wholeLab.setFont(ahp.M_LabelFont);
		wholeLab.setHorizontalAlignment(SwingConstants.TRAILING);
		wholeLab.setBorder(m_InsetBorder);
		wholeLab.setForeground(front);
		JTextField wholeness = new JTextField(4);
		wholeness.setDropTarget(null);
		wholeness.setEditable(false);
		wholeness.setOpaque(false);
		wholeness.setHorizontalAlignment(JTextField.LEFT);
		wholeness.setBorder(m_InsetBorder);
		wholeness.setBackground(back);
		wholeness.setForeground(front);
		wholeness.setFocusable(false);
		wholeness.setFont(ahp.M_LabelFont);
		wholeness.setText(Integer.toString(m_WholenessTosses));
		wholeness.setToolTipText("The number of tosses an account must "
										+ "bet between wholeness bonus awards");
		wholeLab.setLabelFor(wholeness);
		firstItemConstraints(m_GBC);
		m_StatsPanel.add(wholeLab, m_GBC);
		lastItemConstraints(m_GBC);
		m_StatsPanel.add(wholeness, m_GBC);
		m_StatsPanel.add(Box.createVerticalGlue(), m_GBC);

		// put entire form into the main scroll area
		m_WorkArea.setViewportView(m_StatsPanel);
		revalidate();
	}

	/**
	 * method to define refresh action
	 */
	private void setRefreshAction() {
		m_RefreshAction = new AbstractAction("Refresh") {
			public void actionPerformed(ActionEvent ae) {
				// query stats, asynchronously
				ATabManager atm = (ATabManager) m_TabManager;
				atm.queryStats(false);
			}
		};
	}

	/**
	 * helper method to prepare GridBagConstraints for use with the
	 * first item on a line (such as a prompt/label)
	 */
	private void firstItemConstraints(GridBagConstraints gbc) {
		if (gbc == null) {
			return;
		}
		gbc.gridwidth = GridBagConstraints.RELATIVE;
		gbc.fill = GridBagConstraints.NONE;
	}

	/**
	 * helper method to prepare GridBagConstraints for use with the
	 * last item on a line (such as an entry field)
	 */
	private void lastItemConstraints(GridBagConstraints gbc) {
		if (gbc == null) {
			return;
		}
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.fill = GridBagConstraints.HORIZONTAL;
	}

	/**
	 * record new stat values as returned by the game server
	 * @param stats the stats reply returned from the server
	 */
	public void recordStats(ABCBlock stats) {
		if (stats == null
			 || !stats.getOpcode().equalsIgnoreCase("REP_stats"))
		{
			Log.error("recordStats(): not a stats reply message");
			return;
		}

		// record what server sent, with minimal sanity checking
		if (stats.getBettorCount() >= 0) {
			m_NumBettors = stats.getBettorCount();
		}
		if (stats.getTossRate() > (byte)0) {
			m_TossRate = stats.getTossRate();
		}
		if (stats.getNextToss() > 0L) {
			// convert from Unix timestamp (interpreted locally) to desired format
			Date tossTime = new Date(stats.getNextToss());
			Calendar cal = Calendar.getInstance();
			int gmtOffsetMillis = cal.get(Calendar.ZONE_OFFSET);
			int dstOffsetMillis = cal.get(Calendar.DST_OFFSET);
			Date utc
				= new Date(tossTime.getTime() - gmtOffsetMillis - dstOffsetMillis);
			String nToss = m_TossFormatter.format(utc);
			m_NextToss = Integer.parseInt(nToss);

			m_TossDelta = (60 / m_TossRate) * 60;
			int curVal = stats.getTossSeconds();
			// update the toss countdown timer
			AHomePane ahp = (AHomePane) m_ParentHome;
			ahp.initCountdown(m_TossDelta, curVal);
			JProgressBar countDown = ahp.getCountdown();
			if (countDown != null) {
				countDown.setToolTipText("Time until toss " + m_NextToss);
			}
		}
		if (stats.getIBRatio() > 0.0) {
			m_IBReserve = stats.getIBRatio();
		}
		if (stats.getVFQRatio() > 0.0) {
			m_VFQRatio = stats.getVFQRatio();
		}
		m_WholenessTosses = stats.getWholenessTosses();

		// and redraw display to show any changes
		prepDisplay();
	}

	/**
	 * record the percentage change between latest toss and the previous toss
	 * @param delta the amount of change, as a percentage of total position value
	 */
	public void recordTossDelta(double delta) {
		m_PosDelta = delta;
		prepDisplay();
	}


	// get methods
	/**
	 * obtain the current IB reserve ratio
	 * @return the ratio of BTC backing IB
	 */
	public double getIBRatio() { return m_IBReserve; }

	/**
	 * obtain the current VFQ reserve ratio
	 * @return the ratio of VFQ held per unit of IB
	 */
	public double getVFQRatio() { return m_VFQRatio; }

	/**
	 * obtain the number of seconds until the next toss
	 * @return the delta in seconds
	 */
	public int getTossDelta() { return m_TossDelta; }

	/**
	 * obtain the number of tosses per hour
	 * @return the count of tosses occurring in one hour
	 */
	public byte getTossRate() { return m_TossRate; }

	/**
	 * obtain the next scheduled toss
	 * @return the toss, in YYMMDDHHmm format
	 */
	public int getNextToss() { return m_NextToss; }

	/**
	 * obtain the system default for wholeness tosses
	 * @return the separation in tosses
	 */
	public int getWholenessTosses() { return m_WholenessTosses; }

	// END methods
}
