/*
 * last modified---
 * 	02-14-18/kbw/ supply appropriate units with REQ_wallet_xfer requests
 * 	08-26-16/jet/ add deposit and withdraw forms for VFQ also
 * 	04-30-15/jet/ Spark plugin version
 * 	03-27-15/jet/ add m_PickupMode
 * 	03-17-15/jet/ complete
 * 	02-04-15/jet/ new
 *
 * purpose---
 * 	provide a UI screen to represent ABC wallet transfers
 */

package org.silentvault.spark.ui.svm.abc;

import javax.swing.*;
import java.awt.event.ActionEvent;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Font;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.GridBagConstraints;

import org.jivesoftware.spark.util.log.Log;

import org.silentvault.spark.*;
import org.silentvault.spark.ui.*;
import org.silentvault.spark.ui.svm.*;
import org.silentvault.spark.ui.wallet.HomePane;
import org.silentvault.spark.ui.wallet.WTabManager;
import org.silentvault.spark.ui.wallet.PaymentPane;
import org.silentvault.spark.ui.wallet.ConfirmPane;

import java.text.DecimalFormat;
import java.text.NumberFormat;


/**
 * Extends WorkPane to provide a screen for making transfers between the ABC
 * game account and the associated wallet.
 *
 * @author Justin Turrell
 */
public final class TransfersPane extends AWorkPane {
	// BEGIN data members
	/**
	 * the first voucher asset to be deposited or withdrawn
	 */
	private final String					M_XferAsset1;

	/**
	 * the second alternate voucher asset to be deposited or withdrawn
	 */
	private final String					M_XferAsset2;

	/**
	 * action to Deposit value from wallet to ABC account
	 */
	private AbstractAction				m_DepositAction;

	/**
	 * button for Deposit action
	 */
	private JButton						m_DepositButton;

	/**
	 * action to Withdraw value from the ABC account back to the wallet
	 */
	private AbstractAction				m_WithdrawAction;

	/**
	 * button for Withdraw action
	 */
	private JButton						m_WithdrawButton;

	/**
	 * action to acknowledge and pick up incoming transfer
	 */
	private AbstractAction				m_PickupAction;

	/**
	 * button for Pickup action
	 */
	private JButton						m_PickupButton;

	/**
	 * the panel containing the transfer information
	 */
	private JPanel							m_TransferPanel;

	/**
	 * the amount (in SBC) currently available for deposit
	 */
	private double							m_DepositAvail1;

	/**
	 * the amount (in SBC) currently available for withdrawal
	 */
	private double							m_WithdrawAvail1;

	/**
	 * the amount (in VFQ) currently available for deposit
	 */
	private double							m_DepositAvail2;

	/**
	 * the amount (in VFQ) currently available for withdrawal
	 */
	private double							m_WithdrawAvail2;

	/**
	 * entry field for deposit amount, type 1
	 */
	private JTextField					m_DepAmt1;

	/**
	 * entry field for deposit amount, type 2
	 */
	private JTextField					m_DepAmt2;

	/**
	 * entry field for withdrawal amount, type 1
	 */
	private JTextField					m_WithAmt1;

	/**
	 * entry field for withdrawal amount, type 2
	 */
	private JTextField					m_WithAmt2;

	/**
	 * button for selecting SBC as the transfer currency for deposits
	 */
	private JRadioButton					m_DepositSBC;

	/**
	 * button for selecting VFQ as the transfer currency for deposits
	 */
	private JRadioButton					m_DepositVFQ;

	/**
	 * button group for deposit transfer currency selectors
	 */
	private ButtonGroup					m_DepositGroup;

	/**
	 * button for selecting SBC as the transfer currency for withdrawals
	 */
	private JRadioButton					m_WithdrawSBC;

	/**
	 * button for selecting VFQ as the transfer currency for withdrawals
	 */
	private JRadioButton					m_WithdrawVFQ;

	/**
	 * button group for withdrawal transfer currency selectors
	 */
	private ButtonGroup					m_WithdrawGroup;

	/**
	 * the actual currency chosen to deposit or withdraw
	 */
	private String							m_Currency;

	/**
	 * the actual amount of the transfer
	 */
	private double							m_XferAmt;

	/**
	 * flag indicating whether last transfer was in or out (true if in)
	 */
	private boolean						m_XferWasIn;

	/**
	 * the transaction Id returned by the ABC server
	 */
	private int								m_TransId;

	/**
	 * the wallet Id on the ABC side
	 */
	private String							m_ABCWalletId;

	/**
	 * flag indicating that we're about to pick up an escrow payment made to us
	 */
	private boolean						m_PickupMode;

	// END data members

	// BEGIN methods
	/**
	 * constructor
	 * @param plugin the owning plugin or application
	 * @param home the AHomePane in which we're installed
	 * @param mgr the tab manager which creates us
	 */
	public TransfersPane(WalletClient plugin, AHomePane home, ATabManager mgr) {
		super(plugin, home, mgr);
		M_XferAsset1 = "SBC";
		M_XferAsset2 = "VFQ";
		m_ABCWalletId = m_Currency = "";

		// define instructions
		String transTxt = "This panel is used to transfer value from your "
							+ "SilentVault wallet to your ABC game account (Deposit), "
							+ "or to transfer it from your ABC game account back to "
							+ "your wallet (Withdraw).<br/><br/>"
							+ "Enter any amount up to the quantity available and "
							+ "click the appropriate button.  If you have no value "
							+ "available, the corresponding button will be inactive.";
		m_UserInstructs = transTxt;

		// define deposit action and button
		setDepositAction();
		m_DepositButton = new JButton(m_DepositAction);
		m_DepositButton.setFont(m_TabManager.M_ButtonFont);
		m_DepositButton.setToolTipText("Deposit the entered amount to your game "
												+ "account from your wallet");
		m_DepositButton.setEnabled(false);
		m_DepositButton.setBackground(Color.GREEN);

		// define withdraw action and button
		setWithdrawAction();
		m_WithdrawButton = new JButton(m_WithdrawAction);
		m_WithdrawButton.setFont(m_TabManager.M_ButtonFont);
		m_WithdrawButton.setToolTipText("Withdraw the entered amount from your "
												+ "game account to your wallet");
		m_WithdrawButton.setEnabled(false);
		m_WithdrawButton.setBackground(Color.RED);

		// define pickup action and button
		setPickupAction();
		m_PickupButton = new JButton(m_PickupAction);
		m_PickupButton.setFont(m_TabManager.M_ButtonFont);
		m_PickupButton.setToolTipText("Pick up the voucher payment sent to "
												+ "your wallet");

		// init main panel area
		m_TransferPanel = new JPanel();
		m_TransferPanel.setOpaque(true);
		m_TransferPanel.setBorder(null);
		m_TransferPanel.setLayout(new GridBagLayout());
		m_TransferPanel.setPreferredSize(new Dimension(470, 200));
		m_TransferPanel.setBackground(Color.YELLOW);

		// init entry controls
		m_DepAmt1 = new JTextField(8);
		m_DepAmt1.setDropTarget(null);
		m_DepAmt1.setHorizontalAlignment(JTextField.LEFT);
		final Dimension entrySz = new Dimension(30, 20);
		m_DepAmt1.setPreferredSize(entrySz);
		m_DepAmt1.setText("0.0");
		m_DepAmt1.setToolTipText("enter the amount you wish to transfer from "
										+ "your wallet to your game account, in mBTC");
		m_DepAmt2 = new JTextField(8);
		m_DepAmt2.setDropTarget(null);
		m_DepAmt2.setHorizontalAlignment(JTextField.LEFT);
		m_DepAmt2.setPreferredSize(entrySz);
		m_DepAmt2.setText("0.0");
		m_DepAmt2.setToolTipText("enter the amount you wish to transfer from "
										+ "your wallet to your game account");

		m_WithAmt1 = new JTextField(8);
		m_WithAmt1.setDropTarget(null);
		m_WithAmt1.setHorizontalAlignment(JTextField.LEFT);
		m_WithAmt1.setPreferredSize(entrySz);
		m_WithAmt1.setText("0.0");
		m_WithAmt1.setToolTipText("enter the amount you wish to transfer from "
										+ "your game account to your wallet, in BTC");
		m_WithAmt2 = new JTextField(8);
		m_WithAmt2.setDropTarget(null);
		m_WithAmt2.setHorizontalAlignment(JTextField.LEFT);
		m_WithAmt2.setPreferredSize(entrySz);
		m_WithAmt2.setText("0.0");
		m_WithAmt2.setToolTipText("enter the amount you wish to transfer from "
										+ "your game account to your wallet");

		final Dimension radioSz = new Dimension(20, 20);
		m_DepositSBC = new JRadioButton(M_XferAsset1, true);
		m_DepositSBC.setToolTipText("Select to deposit SBC");
		m_DepositSBC.addActionListener(this);
		m_DepositSBC.setPreferredSize(radioSz);
		m_DepositVFQ = new JRadioButton(M_XferAsset2, false);
		m_DepositVFQ.setToolTipText("Select to deposit VFQ");
		m_DepositVFQ.addActionListener(this);
		m_DepositVFQ.setPreferredSize(radioSz);
		m_DepositGroup = new ButtonGroup();
		m_DepositGroup.add(m_DepositSBC);
		m_DepositGroup.add(m_DepositVFQ);
		m_WithdrawSBC = new JRadioButton(M_XferAsset1, true);
		m_WithdrawSBC.setToolTipText("Select to withdraw SBC");
		m_WithdrawSBC.addActionListener(this);
		m_WithdrawSBC.setPreferredSize(radioSz);
		m_WithdrawVFQ = new JRadioButton(M_XferAsset2, false);
		m_WithdrawVFQ.setToolTipText("Select to withdraw VFQ");
		m_WithdrawVFQ.addActionListener(this);
		m_WithdrawVFQ.setPreferredSize(radioSz);
		m_WithdrawGroup = new ButtonGroup();
		m_WithdrawGroup.add(m_WithdrawSBC);
		m_WithdrawGroup.add(m_WithdrawVFQ);
	}

	/**
	 * do display of transfers
	 */
	public void prepDisplay() {
		// set screen label
		buildScreenLabel("Perform transfers between your wallet and account");

		// inits
		AHomePane ahp = (AHomePane) m_ParentHome;
		m_WithdrawAvail1 = ahp.getAvailSBC();
		m_WithdrawAvail2 = ahp.getAvailVFQ();
		DecimalFormat df = (DecimalFormat) NumberFormat.getInstance();
		df.setMaximumFractionDigits(XMLVoucher.M_Precision);
		df.setMinimumFractionDigits(XMLVoucher.M_Precision);
		df.setGroupingSize(3);
		m_TransferPanel.removeAll();
		m_GBC.anchor = GridBagConstraints.NORTH;
		m_GBC.gridwidth = GridBagConstraints.REMAINDER;
		m_GBC.fill = GridBagConstraints.HORIZONTAL;

		// determine how much SBC is available in owning wallet, for deposit
		HomePane whp = m_Plugin.getWalletTabManager().getHomePane();
		VoucherTableModel vtm = whp.getVoucherModel();
		m_DepositAvail1 = vtm.getVoucherAssetSum(M_XferAsset1);
		m_DepositAvail2 = vtm.getVoucherAssetSum(M_XferAsset2);

		if (!m_PickupMode) {
			// populate the button bar for this mode
			m_ButtonPane.removeAll();
			m_DepositButton.setVisible(true);
			m_WithdrawButton.setVisible(true);
			m_ButtonPane.add(Box.createHorizontalStrut(105));	// spacer
			m_ButtonPane.add(m_DepositButton);
			m_ButtonPane.add(Box.createHorizontalStrut(100));	// spacer
			m_ButtonPane.add(m_WithdrawButton);
			m_PickupButton.setVisible(false);

			// first, deposit amount for SBC
			m_GBC.gridheight = 9;
			JPanel depPane = new JPanel(new GridLayout(1, 4, 5, 5));
			depPane.setOpaque(true);
			depPane.setBorder(null);
			Dimension lineSz = new Dimension(470, 25);
			depPane.setPreferredSize(lineSz);
			depPane.setBackground(Color.GREEN);
			JLabel depLab = new JLabel("Deposit : ");
			depLab.setFont(ahp.M_LabelFont);
			depLab.setHorizontalAlignment(SwingConstants.TRAILING);
			depLab.setBorder(m_InsetBorder);
			depLab.setLabelFor(m_DepAmt1);
			JTextField depAvail = new JTextField(8);
			depAvail.setDropTarget(null);
			depAvail.setEditable(false);
			depAvail.setOpaque(false);
			depAvail.setHorizontalAlignment(JTextField.LEFT);
			depAvail.setBorder(m_InsetBorder);
			depAvail.setBackground(m_Plugin.m_LabelBackColor);
			depAvail.setFocusable(false);
			final Dimension entrySz = new Dimension(50, 20);
			depAvail.setPreferredSize(entrySz);
			depAvail.setText("( " + df.format(m_DepositAvail1) + " "
								+ M_XferAsset1 + " )");
			depAvail.setToolTipText("the amount of " + M_XferAsset1
											+ " available in your wallet, in mBTC");
			depPane.add(m_DepositSBC);
			depPane.add(depLab);
			depPane.add(m_DepAmt1);
			depPane.add(depAvail);
			m_TransferPanel.add(depPane, m_GBC);
			if (m_DepositSBC.isSelected()) {
				if (m_DepositAvail1 > 0.0) {
					m_DepositButton.setEnabled(true);
				}
				else {
					m_DepositButton.setEnabled(false);
				}
			}
			m_TransferPanel.add(Box.createVerticalStrut(10), m_GBC);

			// second, deposit amount for VFQ
			JPanel depPane2 = new JPanel(new GridLayout(1, 4, 5, 5));
			depPane2.setOpaque(true);
			depPane2.setBorder(null);
			depPane2.setPreferredSize(lineSz);
			depPane2.setBackground(Color.GREEN);
			JLabel depLab2 = new JLabel("Deposit : ");
			depLab2.setFont(ahp.M_LabelFont);
			depLab2.setHorizontalAlignment(SwingConstants.TRAILING);
			depLab2.setBorder(m_InsetBorder);
			depLab2.setLabelFor(m_DepAmt2);
			JTextField depAvail2 = new JTextField(8);
			depAvail2.setDropTarget(null);
			depAvail2.setEditable(false);
			depAvail2.setOpaque(false);
			depAvail2.setHorizontalAlignment(JTextField.LEFT);
			depAvail2.setBorder(m_InsetBorder);
			depAvail2.setBackground(m_Plugin.m_LabelBackColor);
			depAvail2.setFocusable(false);
			depAvail2.setPreferredSize(entrySz);
			depAvail2.setText("( " + df.format(m_DepositAvail2) + " "
									+ M_XferAsset2 + " )");
			depAvail2.setToolTipText("the amount of " + M_XferAsset2
											+ " available in your wallet");
			depPane2.add(m_DepositVFQ);
			depPane2.add(depLab2);
			depPane2.add(m_DepAmt2);
			depPane2.add(depAvail2);
			m_TransferPanel.add(depPane2, m_GBC);
			if (m_DepositVFQ.isSelected()) {
				if (m_DepositAvail2 > 0.0) {
					m_DepositButton.setEnabled(true);
				}
				else {
					m_DepositButton.setEnabled(false);
				}
			}
			m_TransferPanel.add(Box.createVerticalStrut(10), m_GBC);

			// spacer
			JSeparator demark = new JSeparator();
			m_TransferPanel.add(Box.createVerticalStrut(40), m_GBC);
			m_TransferPanel.add(demark, m_GBC);
			m_TransferPanel.add(Box.createVerticalStrut(40), m_GBC);

			// next, withdrawal amount for SBC
			JPanel withPane = new JPanel(new GridLayout(1, 4, 5, 5));
			withPane.setOpaque(true);
			withPane.setBorder(null);
			withPane.setPreferredSize(lineSz);
			withPane.setBackground(Color.RED);
			JLabel withLab = new JLabel("Withdraw : ");
			withLab.setFont(ahp.M_LabelFont);
			withLab.setHorizontalAlignment(SwingConstants.TRAILING);
			withLab.setBorder(m_InsetBorder);
			withLab.setLabelFor(m_WithAmt1);
			JTextField withAvail = new JTextField(8);
			withAvail.setDropTarget(null);
			withAvail.setEditable(false);
			withAvail.setOpaque(false);
			withAvail.setHorizontalAlignment(JTextField.LEFT);
			withAvail.setBorder(m_InsetBorder);
			withAvail.setBackground(m_Plugin.m_LabelBackColor);
			withAvail.setFocusable(false);
			withAvail.setPreferredSize(entrySz);
			withAvail.setText("( " + df.format(m_WithdrawAvail1) + " "
									+ M_XferAsset1 + " )");
			withAvail.setToolTipText("the amount of " + M_XferAsset1
											+ " available in your account, in BTC");
			withPane.add(m_WithdrawSBC);
			withPane.add(withLab);
			withPane.add(m_WithAmt1);
			withPane.add(withAvail);
			m_TransferPanel.add(withPane, m_GBC);
			if (m_WithdrawSBC.isSelected()) {
				if (m_WithdrawAvail1 > 0.0) {
					m_WithdrawButton.setEnabled(true);
				}
				else {
					m_WithdrawButton.setEnabled(false);
				}
			}
			m_TransferPanel.add(Box.createVerticalStrut(10), m_GBC);

			// next, withdrawal amount for VFQ
			JPanel withPane2 = new JPanel(new GridLayout(1, 4, 5, 5));
			withPane2.setOpaque(true);
			withPane2.setBorder(null);
			withPane2.setPreferredSize(lineSz);
			withPane2.setBackground(Color.RED);
			JLabel withLab2 = new JLabel("Withdraw : ");
			withLab2.setFont(ahp.M_LabelFont);
			withLab2.setHorizontalAlignment(SwingConstants.TRAILING);
			withLab2.setBorder(m_InsetBorder);
			withLab2.setLabelFor(m_WithAmt2);
			JTextField withAvail2 = new JTextField(8);
			withAvail2.setDropTarget(null);
			withAvail2.setEditable(false);
			withAvail2.setOpaque(false);
			withAvail2.setHorizontalAlignment(JTextField.LEFT);
			withAvail2.setBorder(m_InsetBorder);
			withAvail2.setBackground(m_Plugin.m_LabelBackColor);
			withAvail2.setFocusable(false);
			withAvail2.setPreferredSize(entrySz);
			withAvail2.setText("( " + df.format(m_WithdrawAvail2) + " "
									+ M_XferAsset2 + " )");
			withAvail2.setToolTipText("the amount of " + M_XferAsset2
											+ " available in your account");
			withPane2.add(m_WithdrawVFQ);
			withPane2.add(withLab2);
			withPane2.add(m_WithAmt2);
			withPane2.add(withAvail2);
			m_TransferPanel.add(withPane2, m_GBC);
			if (m_WithdrawVFQ.isSelected()) {
				if (m_WithdrawAvail2 > 0.0) {
					m_WithdrawButton.setEnabled(true);
				}
				else {
					m_WithdrawButton.setEnabled(false);
				}
			}
		}
		else {
			// populate the button bar for this mode
			m_ButtonPane.removeAll();
			m_ButtonPane.add(Box.createHorizontalStrut(215));	// spacer
			m_PickupButton.setVisible(true);
			m_ButtonPane.add(m_PickupButton);
			m_DepositButton.setVisible(false);
			m_WithdrawButton.setVisible(false);
			m_ButtonPane.revalidate();

			// display language about escrow pickup, place in transfer panel
			m_GBC.gridheight = 1;
			JPanel pickPane = new JPanel(new GridLayout(1, 1, 10, 10));
			pickPane.setOpaque(true);
			pickPane.setBorder(null);
			StringBuilder pickTxt = new StringBuilder(256);
			pickTxt.append("Your withdrawal of " + df.format(m_XferAmt));
			pickTxt.append(" " + m_Currency);
			pickTxt.append(" has been assigned\ntransaction ID " + m_TransId);
			// if there's no escrow wallet Id, this means manual support processing
			if (!m_ABCWalletId.isEmpty()) {
				pickTxt.append(", and has been sent from:\n\n" + m_ABCWalletId);
				pickTxt.append("\n\nClick on the Pick Up button to pick up your ");
				pickTxt.append("payment.\n");
				m_PickupButton.setEnabled(true);
			}
			else {
				pickTxt.append(".\nDue to automated withdrawal limits, this ");
				pickTxt.append("transaction will be processed manually by ABC ");
				pickTxt.append("Customer Support.\nManual withdrawals are ");
				pickTxt.append("usually completed within one business day.\n");
				// nothing to pick up
				m_PickupButton.setEnabled(false);
			}
			JTextArea pickBlurb = new JTextArea(pickTxt.toString(), 6, 50);
			pickBlurb.setEditable(false);
			pickPane.add(pickBlurb);
			m_TransferPanel.add(pickPane, m_GBC);
		}

		// put entire form into the main scroll area
		m_WorkArea.setViewportView(m_TransferPanel);
		revalidate();
	}

	/**
	 * method to define deposit action
	 */
	private void setDepositAction() {
		m_DepositAction = new AbstractAction("Deposit") {
			// do spend from wallet to ABC escrow wallet
			public void actionPerformed(ActionEvent ae) {
				// examine which currency is selected
				String amtTxt = "";
				double avail = 0.0;
				if (m_DepositSBC.isSelected()) {
					amtTxt = m_DepAmt1.getText();
					m_Currency = M_XferAsset1;
					avail = m_DepositAvail1;
				}
				else {
					amtTxt = m_DepAmt2.getText();
					m_Currency = M_XferAsset2;
					avail = m_DepositAvail2;
				}
				String units = VPAssetConfig.getBaseUnits(m_Currency);

				// validate amount
				Double amt = null;
				m_XferAmt = 0.0;
				try {
					amt = Double.parseDouble(amtTxt);
					m_XferAmt = amt.doubleValue();
				}
				catch (NumberFormatException nfe) {
					Log.error("Illegal deposit amount, " + amtTxt, nfe);
				}
				if (m_XferAmt <= 0.0) {
					m_TabManager.showError("cannot make deposit",
													"invalid amount",
													"enter a positive number");
					return;
				}

				// make sure user has this much
				if (m_XferAmt > avail) {
					m_TabManager.showError("cannot make deposit",
													"amount is more than you have available",
												"enter an amount not more than " + avail);
					return;
				}

				// prepare wallet xfer message
				String sessId = m_TabManager.getLoginPane().getSessionId();
				final String abcId = m_Plugin.getLoginSecrets().getSVMId();
				final ABCClientBlock depBlock = new ABCClientBlock();
				depBlock.setOpcode("REQ_wallet_xfer");
				depBlock.setSessionId(sessId);
				depBlock.setDirection("in");
				depBlock.setSeriesId(m_Currency);
				depBlock.setQuantity(m_XferAmt);
				depBlock.setUnits(units);

				// provide for correct interpretation of ABC's reply
				m_XferWasIn = true;

				// prevent a concurrent transaction, using either button
				m_DepositButton.setEnabled(false);
				m_WithdrawButton.setEnabled(false);
				setCursor(M_WaitCursor);
				m_TransId = 0;

				Runnable sendDep = new Runnable() {
					public void run() {
						// send to server; answer will go to recordXfer()
						ABCListener abcListener
							= (ABCListener) m_Plugin.getSVMListener();
						if (!abcListener.sendRequest(depBlock)) {
							Log.error("Deposit request failed to ABC Id " + abcId);
							m_TabManager.showError("cannot make deposit at ABC "
															+ abcId,
														"failure sending transfer message");
							m_DepositButton.setEnabled(true);
							m_WithdrawButton.setEnabled(true);
							setCursor(null);
						}
					}
				};
				new Thread(sendDep).start();
			}
		};
	}

	/**
	 * method to define withdraw action
	 */
	private void setWithdrawAction() {
		m_WithdrawAction = new AbstractAction("Withdraw") {
			// request spend from ABC escrow wallet to wallet
			public void actionPerformed(ActionEvent ae) {
				// examine which currency is selected
				String amtTxt = "";
				double avail = 0.0;
				String units = "";
				if (m_WithdrawSBC.isSelected()) {
					amtTxt = m_WithAmt1.getText();
					m_Currency = M_XferAsset1;
					avail = m_WithdrawAvail1;
					units = "BTC";
				}
				else {
					amtTxt = m_WithAmt2.getText();
					m_Currency = M_XferAsset2;
					avail = m_WithdrawAvail2;
					units = VPAssetConfig.getBaseUnits(m_Currency);
				}

				// validate amount
				Double amt = null;
				m_XferAmt = 0.0;
				try {
					amt = Double.parseDouble(amtTxt);
					m_XferAmt = amt.doubleValue();
				}
				catch (NumberFormatException nfe) {
					Log.error("Illegal withdrawal amount, " + amtTxt, nfe);
				}
				if (m_XferAmt <= 0.0) {
					m_TabManager.showError("cannot make withdrawal",
													"invalid amount",
													"enter a positive number");
					return;
				}

				// make sure user has this much
				if (m_XferAmt > avail) {
					m_TabManager.showError("cannot make withdrawal",
													"amount is more than you have available",
												"enter an amount not more than " + avail);
					return;
				}

				// prepare wallet xfer message
				String sessId = m_TabManager.getLoginPane().getSessionId();
				final String abcId = m_Plugin.getLoginSecrets().getSVMId();
				final ABCClientBlock withBlock = new ABCClientBlock();
				withBlock.setOpcode("REQ_wallet_xfer");
				withBlock.setSessionId(sessId);
				withBlock.setDirection("out");
				withBlock.setSeriesId(m_Currency);
				withBlock.setQuantity(m_XferAmt);
				withBlock.setUnits(units);

				// provide for correct interpretation of ABC's reply
				m_XferWasIn = false;

				// prevent a concurrent transaction, using either button
				m_DepositButton.setEnabled(false);
				m_WithdrawButton.setEnabled(false);
				setCursor(M_WaitCursor);
				m_TransId = 0;

				Runnable sendWith = new Runnable() {
					public void run() {
						// send to server; answer will go to recordXfer()
						ABCListener abcListener
							= (ABCListener) m_Plugin.getSVMListener();
						if (!abcListener.sendRequest(withBlock)) {
							Log.error("Withdrawal request failed to ABC Id " + abcId);
							m_TabManager.showError("cannot make withdrawal from ABC "
														+ abcId,
														"failure sending transfer message");
							m_DepositButton.setEnabled(true);
							m_WithdrawButton.setEnabled(true);
							setCursor(null);
						}
						// NB: reply goes to recordXfer() below
					}
				};
				new Thread(sendWith).start();
			}
		};
	}

	/**
	 * method to define pickup action
	 */
	private void setPickupAction() {
		m_PickupAction = new AbstractAction("Pick Up") {
			public void actionPerformed(ActionEvent ae) {
				// attempt to pick up payment via housekeeping
				WTabManager wtm = m_Plugin.getWalletTabManager();
				m_Plugin.setSelectedTab(wtm);
				wtm.queryDHTrecords(false);

				// then in this tab, reset for next transfer
				AHomePane ahp = (AHomePane) m_ParentHome;
				double newAvail = 0.0;
				if (m_Currency.equals(M_XferAsset1)) {
					newAvail = ahp.getAvailSBC() - m_XferAmt;
					ahp.setAvailSBC(newAvail);
					m_WithdrawAvail1 = newAvail;
				}
				else {
					newAvail = ahp.getAvailVFQ() - m_XferAmt;
					ahp.setAvailVFQ(newAvail);
					m_WithdrawAvail2 = newAvail;
				}
				m_XferAmt = 0.0;
				m_TransId = 0;
				m_ABCWalletId = "";
				m_WithAmt1.setText("0.0");
				m_WithAmt2.setText("0.0");
				m_PickupMode = false;
				m_Currency = "";

				/* Now go to holdings pane, so that whenever we return to transfers
				 * (by clicking on Transfers) the available deposit number will show
				 * the correct value, post-pickup.  Safe to do in this UI thread.
				 */
				m_TabManager.queueTransition(
													new Integer(MTabManager.M_GO_MAIN_MENU));
				m_TabManager.makeTransition();
			}
		};
	}

	/**
	 * record wallet transfer results
	 * @param xfer the transfer reply returned from the server
	 */
	public void recordXfer(ABCBlock xfer) {
		setCursor(null);
		// re-enable buttons
		m_DepositButton.setEnabled(true);
		m_WithdrawButton.setEnabled(true);
		// sanity check
		if (xfer == null
			 || !xfer.getOpcode().equalsIgnoreCase("REP_wallet_xfer"))
		{
			Log.error("recordXfer(): not a transfer reply message");
			return;
		}

		// see whether we got an error
		if (xfer.isError()) {
			Log.error("wallet xfer init failed, " + xfer.getErrCode() + ": "
						+ xfer.getErrMsg());
			m_TabManager.showError("cannot initiate wallet transfer",
											xfer.getErrMsg() + " ( code "
											+ xfer.getErrCode() + " )");
			return;
		}

		// note transfer Id and record wallet Id
		m_TransId = xfer.getTransId();
		if (m_TransId <= 0) {
			Log.error("Missing transId, " + m_TransId);
			m_TabManager.showError("cannot initiate wallet transfer",
										"missing transaction Id");
			return;
		}
		m_ABCWalletId = xfer.getWalletId();
		if (m_ABCWalletId.isEmpty() && m_XferWasIn) {
			Log.error("Missing wallet Id on " + xfer.getOpcode());
			m_TabManager.showError("cannot initiate wallet transfer",
											"missing counterparty wallet Id");
			return;
		}

		// if we're doing a deposit, go to payment screen with prefill
		final WTabManager wtm = m_Plugin.getWalletTabManager();
		final HomePane whp = wtm.getHomePane();
		final VoucherTableModel vtm = whp.getVoucherModel();
		if (m_XferWasIn) {
			WebMessage webPass = new WebMessage();
			webPass.setAmount(m_XferAmt);
			webPass.setVSnumber(m_ABCWalletId);
			webPass.setIssuer(m_Currency);
			webPass.setCurrency(XMLVoucher.getDisplayUnits(m_Currency));
			webPass.setBaggageFields("transId");
			webPass.addOrderRecord("transId", new Integer(m_TransId).toString());
			webPass.setDecoded(true);
			m_Plugin.getWebListener().addWebMessage(webPass);
			PaymentPane pp = wtm.getPaymentPane();
			pp.setSelectedSiteMsg(webPass);
			// reset issuer in case user has a different one selected
			vtm.selectIssuer(m_Currency);

			// re-disable buttons to prevent multiple entry
			m_DepositButton.setEnabled(true);
			m_WithdrawButton.setEnabled(true);
			m_DepositButton.setToolTipText("Please make your deposit payment "
													+ "in the Wallet tab.");
			// set that we want to receive escrow notifications when payment done
			m_DoingEscrow = true;

			// send user to payment screen in WTabManager, in another thread
			Runnable goPay = new Runnable() {
				public void run() {
					wtm.queueTransition(new Integer(WTabManager.M_PAYMENTS1));
					m_Plugin.setSelectedTab(wtm);
					wtm.makeTransition();
				}
			};
			SwingUtilities.invokeLater(goPay);
		}
		else {
			// tell user to pickup withdrawal via housekeeping
			m_PickupMode = true;
			Runnable goPickup = new Runnable() {
				public void run() {
					ATabManager atm = (ATabManager) m_TabManager;
					atm.queueTransition(new Integer(ATabManager.M_GO_TRANSFERS));
					atm.makeTransition();
				}
			};
			SwingUtilities.invokeLater(goPickup);
		}
	}

	/**
	 * register confirmation that we have successfully completed escrow payment
	 * (sent to us by the wallet tab's PaymentPane)
	 */
	@Override
	public void confirmEscrowFunding() {
		// clear flags
		m_DoingEscrow = false;
		m_DepositButton.setToolTipText("Confirming ABC has seen our payment");
		if (m_TransId == 0) {
			Log.error("No transfer in progress");
			return;
		}

		// prepare a REQ_confirm_xfer to ABC to nudge it to pick up our payment
		String sessId = m_TabManager.getLoginPane().getSessionId();
		final String abcId = m_Plugin.getLoginSecrets().getSVMId();
		ABCClientBlock confBlock = new ABCClientBlock();
		confBlock.setOpcode("REQ_confirm_xfer");
		confBlock.setSessionId(sessId);
		confBlock.setTransId(m_TransId);

		// return focus to our tab
		m_Plugin.setSelectedTab(m_TabManager);

		// put up info saying what we are doing (on SVM tab)
		WTabManager wtm = m_Plugin.getWalletTabManager();
		PaymentPane pp = wtm.getPaymentPane();
		final StringBuilder popTxt = new StringBuilder(256);
		// obtain last confirmation recorded by ConfirmPane off the DHT
		XMLConfirmPayment conf = pp.getLatestConfirmation();
		if (conf != null) {
			/* we deferred showing payment success, and so must recapitulate
			 * what the ConfirmPane would have displayed as popup text.
			 */
			popTxt.append("Your payment of " + conf.getQuantity());
			popTxt.append(" " + conf.getUnits() + " of " + conf.getAsset());
			if (!conf.getAsset().equals(conf.getIssuer())) {
				popTxt.append(" at " + conf.getIssuer());
			}
			popTxt.append("\nwas sent to ");
			popTxt.append(conf.getPayee() + " successfully.");
			// show that change voucher was received if there was one
			if (conf.getVoucher() != null) {
				popTxt.append("\nYou have received a voucher back as change.");
			}
		}
		else {
			Log.error("Unexpected: no XMLConfirmPayment found in PaymentPane");
			// show something generic
			popTxt.append("Your escrow payment to fund transfer ID ");
			popTxt.append(m_TransId + " succeeded.");
		}

		// add generic indication that we're now telling ABC to pick it up
		popTxt.append("\n\nThe ABC server is now processing the escrow payment.");
		popTxt.append("  This may take a few moments.\nYou will receive a ");
		popTxt.append("receipt confirming your payment.");

		// for safety, show notice in separate thread allowing message to proceed
		Runnable notice = new Runnable() {
			public void run() {
				JOptionPane.showMessageDialog(m_TabManager, popTxt.toString(),
														"Escrow Payment Completed",
														JOptionPane.INFORMATION_MESSAGE);
			}
		};
		SwingUtilities.invokeLater(notice);

		// reset ConfirmPane to default behavior
		ConfirmPane cp = wtm.getConfirmPane();
		cp.setDeferPopup(false);

		// send async confirmation request to ABC
		ABCListener abcListener = (ABCListener) m_Plugin.getSVMListener();
		setCursor(M_WaitCursor);
		if (!abcListener.sendAsyncRequest(confBlock)) {
			Log.error("Deposit confirmation failed to ABC Id " + abcId);
			Runnable err = new Runnable() {
				public void run() {
					m_TabManager.showError("cannot confirm deposit at ABC " + abcId,
													"failure sending confirmation request");
				}
			};
			SwingUtilities.invokeLater(err);
			m_DepositButton.setEnabled(true);
			m_WithdrawButton.setEnabled(true);
			setCursor(null);
		}
		// NB: reply goes to recordXconf() below
	}

	/**
	 * record wallet transfer confirmation results
	 * @param conf the transfer confirmation reply returned from the server
	 */
	public void recordXconf(ABCBlock conf) {
		setCursor(null);
		if (conf == null
			 || !conf.getOpcode().equalsIgnoreCase("REP_xfer_confirmed"))
		{
			Log.error("recordXconf(): not a transfer confirm reply message");
			return;
		}
		// reset flags
		m_DoingEscrow = false;
		m_DepositButton.setToolTipText("Deposit the entered amount to your game "
												+ "account from your wallet");
		m_DepositButton.setEnabled(true);
		m_WithdrawButton.setEnabled(true);

		if (conf.isError()) {
			final String err = conf.getErrMsg();
			Log.error("recordXconf(): error confirming deposit, " + err);
			// show notice in separate thread to allow message to return
			Runnable showErr = new Runnable() {
				public void run() {
					m_TabManager.showError("error on deposit transfer", err);
				}
			};
			SwingUtilities.invokeLater(showErr);
			return;
		}

		// verify transfer Id echo
		final int tId = conf.getTransId();
		if (tId != m_TransId) {
			Log.error("recordXconf(): transId mismatch, " + tId + " vs. "
						+ m_TransId);
			// show notice in separate thread to allow message to return
			Runnable showErr = new Runnable() {
				public void run() {
					m_TabManager.showError("possible error on deposit transfer",
												"confirmation transId " + tId
												+ " not as expected (" + m_TransId + ")",
												"re-confirm account balances in Holdings");
				}
			};
			SwingUtilities.invokeLater(showErr);
			return;
		}

		// record new balances
		AHomePane ahp = (AHomePane) m_ParentHome;
		double newAvail = conf.getNewBal();
		if (m_Currency.equals(M_XferAsset1)) {
			ahp.setAvailSBC(newAvail);
			m_WithdrawAvail1 = newAvail;
		}
		else {
			ahp.setAvailVFQ(newAvail);
			m_WithdrawAvail2 = newAvail;
		}
		// can't just subtract, because a token purchase might have occurred
		HomePane whp = m_Plugin.getWalletTabManager().getHomePane();
		VoucherTableModel vtm = whp.getVoucherModel();
		m_DepositAvail1 = vtm.getVoucherAssetSum(M_XferAsset1);
		m_DepositAvail2 = vtm.getVoucherAssetSum(M_XferAsset2);
		// reset for next
		m_XferAmt = 0.0;
		m_TransId = 0;
		m_ABCWalletId = "";
		m_DepAmt1.setText("0.0");
		m_DepAmt2.setText("0.0");
		m_Currency = "";
		prepDisplay();
	}

	/**
	 * method to implement ActionListener (overriding WorkPane), used to
	 * activate buttons appropriately whenever currencies are switched
	 */
	@Override
	public void actionPerformed(ActionEvent ae) {
		Object src = ae.getSource();
		if (src instanceof JRadioButton) {
			JRadioButton radSrc = (JRadioButton) src;
			if (radSrc == m_DepositSBC) {
				if (m_DepositAvail1 > 0.0) {
					m_DepositButton.setEnabled(true);
				}
				else {
					m_DepositButton.setEnabled(false);
				}
			}
			else if (radSrc == m_DepositVFQ) {
				if (m_DepositAvail2 > 0.0) {
					m_DepositButton.setEnabled(true);
				}
				else {
					m_DepositButton.setEnabled(false);
				}
			}
			else if (radSrc == m_WithdrawSBC) {
				if (m_WithdrawAvail1 > 0.0) {
					m_WithdrawButton.setEnabled(true);
				}
				else {
					m_WithdrawButton.setEnabled(false);
				}
			}
			else if (radSrc == m_WithdrawVFQ) {
				if (m_WithdrawAvail2 > 0.0) {
					m_WithdrawButton.setEnabled(true);
				}
				else {
					m_WithdrawButton.setEnabled(false);
				}
			}
		}
	}


	// get methods
	/**
	 * obtain the transfer Id recorded
	 * @return the latest value returned by the game server
	 */
	public int getTransId() { return m_TransId; }

	/**
	 * obtain the ABC wallet Id
	 * @return the wallet Id which will be the counterparty on the ABC side
	 */
	public String getABCWallet() { return m_ABCWalletId; }

	// END methods
}
